# Adds an optional component to the activate components list.
function add_optional_component() {
  local -r component="$1"

  if [[ "${OPTIONAL_COMPONENTS_VALUE}" == *${component}* ]]; then
    loginfo "${component} is already added to the list of optional components: ${OPTIONAL_COMPONENTS_VALUE}."
    return 0
  fi
  if [[ -z "${OPTIONAL_COMPONENTS_VALUE}" ]]; then
    OPTIONAL_COMPONENTS_VALUE=${component}
    echo "dataproc.components.activate=${component}" >>"${DATAPROC_ETC_DIR}/dataproc.properties"
  else
    OPTIONAL_COMPONENTS_VALUE+=" ${component}"
    sed -i \
      "s/dataproc.components.activate=.*/dataproc.components.activate=${OPTIONAL_COMPONENTS_VALUE}/g" \
      "${DATAPROC_ETC_DIR}/dataproc.properties"
  fi
}

# Removes a component from the activate components list.
function remove_optional_component() {
  local -r component="$1"

  OPTIONAL_COMPONENTS_VALUE="${OPTIONAL_COMPONENTS_VALUE/${component}/''}"
  sed -i \
    "s/dataproc.components.activate=.*/dataproc.components.activate=${OPTIONAL_COMPONENTS_VALUE}/g" \
    "${DATAPROC_ETC_DIR}/dataproc.properties"
}

# Returns a list of a default components.
#
# Names are in lower case and separated by a space, e.g.,
# "hbase presto".
function get_default_components() {
  # Use the runtime property if provided, otherwise fall back to the env
  # variable set at image build time.
  get_dataproc_property_or_default \
      dataproc.components.default \
      "${DATAPROC_COMPONENTS_DEFAULT//,/ }" \
    | tr '[:upper:]' '[:lower:]'
}

# Returns additional components to activate.
#
# Names are in lower case and separated by a space, e.g.,
# "hbase presto".
function get_components_to_activate() {
  get_dataproc_property dataproc.components.activate |
    tr '[:upper:]' '[:lower:]'
}

# Returns components to deactivate.
#
# Names are in lower case and separated by a space, e.g.,
# "hdfs mysql hive-metastore".
function get_components_to_deactivate() {
  get_dataproc_property dataproc.components.deactivate |
    tr '[:upper:]' '[:lower:]'
}

# Checks whether optional component is selected for activation.
function is_component_selected() {
  local -r component="$1"

  local activated_components
  activated_components=$(get_components_to_activate)
  [[ ${activated_components} == *${component}* ]]
}

# Checks whether an optional component is explicitly unselected.
#
# Note that this function is NOT equal to `! is_component_selected`, some
# components (e.g., hive-metastore) are selected by default, but could be
# explicitly unselected by adding them to the `dataproc.components.deactivate`
# property.
function is_component_explicitly_unselected() {
  local -r component="$1"

  local deactivated_components
  deactivated_components=$(get_components_to_deactivate)
  [[ ${deactivated_components} == *${component}* ]]
}

# Add default components if they are not explicitly unselected.
function add_default_components() {
  local default_components=("$@")
  for component in "${default_components[@]}"; do
    if ! is_component_explicitly_unselected "${component}"; then
      add_optional_component "${component}"
    fi
  done
}

# Waits until the component has been activated or time out.
function wait_until_component_activated() {
  local -r component="$1"
  local -r max_retry_time="${2:-300}"
  local -r retry_delay="${3:-1}"
  local -r done_file="/tmp/dataproc/components/activate/${component}.done"

  if retry_constant_custom \
    "${max_retry_time}" "${retry_delay}" test -f "${done_file}"; then
    loginfo "Component ${component} is activated"
    return 0
  fi

  loginfo "Timed out waiting for component ${component} to be activated"
  return 1
}
