# Copyright 2021 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

function download_golang_tools() {
  local download_dir
  download_dir="$(mktemp -d)"

  local -r go_version="1.21.1"
  local arch_suffix
  arch_suffix="$(get_arch_suffix)"
  local -r go_download_file="go${go_version}.linux-${arch_suffix}.tar.gz"
  local -r download_source="https://go.dev/dl/${go_download_file}"
  local -r download_location="${download_dir}/${go_download_file}"

  curl -fsSL --retry-connrefused --retry 10 --retry-max-time 30 \
      "${download_source}" -o "${download_location}"
  tar -C "${download_dir}" -zxf "${download_location}"

  rm -Rf "${download_location}"
  export GOLANG_TOOLS_DIR="${download_dir}"
}

function cleanup_golang_tools() {
  rm -Rf "${GOLANG_TOOLS_DIR}"
  unset GOLANG_TOOLS_DIR
}

function install_golang_binary() {
  local -r binary_url="$1"
  local path_dir
  path_dir="$(mktemp -d)"
  readonly path_dir
  local cache_dir
  cache_dir="$(mktemp -d)"
  readonly cache_dir
  GOCACHE="${cache_dir}" GOPATH="${path_dir}" \
    "${GOLANG_TOOLS_DIR}/go/bin/go" install -v "${binary_url}"

  local binary
  for binary in $(ls "${path_dir}/bin"); do
    chmod a+rx "${path_dir}/bin/${binary}"
    cp "${path_dir}/bin/${binary}" "${DATAPROC_DIR}/"
    ln -s "${DATAPROC_DIR}/${binary}" "/usr/local/bin/${binary}"
  done

  rm -Rf "${path_dir}" "${cache_dir}"
  echo "${binary_url}" >> "${DATAPROC_DIR}/installed_go_binaries.txt"
}

function write_go_licenses() {
  # Run the remaining code in a subshell so we can safely override the GOPATH,
  # PATH, and HOME environment variables and change directories without it
  # affecting the outer script.
  #
  # This is necessary because the go-licenses tool is extremely picky about the
  # environment that it runs in.
  #
  # In particular, the `go-licenses` tool invokes the `go` binaries, so it
  # needs them to be in the PATH when it runs. Those go binaries, in turn,
  # require the GOPATH and HOME environment variables to be set when they run.
  (
    local working_dir
    working_dir="$(mktemp -d)"
    readonly working_dir
    local -r licenses_output_dir="${working_dir}/licenses"
    export GOPATH="${working_dir}"
    export HOME="${working_dir}"
    export PATH="${GOLANG_TOOLS_DIR}/go/bin:${GOPATH}/bin:${PATH}"
    # Install the license check tool
    go install -v github.com/google/go-licenses@latest

    # N.B. Some of the subsequent commands generate lint warnings about command
    # expansions not being quoted. The lack of quotes is both intentional and
    # required, so the corresponding lint warnings have been disabled.
    #
    # The affected commands expand to multiple values that we want to be treated
    # as separate arguments to the invoked binaries. The only way for us to Do
    # that is to explicitly *not* quote the output so that it is interpreted
    # by the shell as multiple, separate arguments.

    # Download the codebase for every installed tool and add them to a local
    # go.mod file as required by the go-licenses tool...
    local -r licenses_mod_dir="${working_dir}/src/licenses"
    mkdir -p "${licenses_mod_dir}"
    cd "${licenses_mod_dir}"
    go mod init "licenses"
    # shellcheck disable=SC2046
    go get $(cat "${DATAPROC_DIR}/installed_go_binaries.txt")

    # Add all the license files to our license directory
    # shellcheck disable=SC2046
    go-licenses save \
      --save_path="${licenses_output_dir}/go-licenses" \
      $(cat "${DATAPROC_DIR}/installed_go_binaries.txt" | cut -d "@" -f 1)
    local dir
    for dir in $(ls "${licenses_output_dir}/go-licenses"); do
      mv "${licenses_output_dir}/go-licenses/${dir}" "${DATAPROC_DIR}/licenses"
    done
    rm -Rf "${working_dir}"
  )
}