# Use different exit_code to differentiate internal errors and client errors.
readonly EXIT_CODE_INTERNAL_ERROR=1
readonly EXIT_CODE_CLIENT_ERROR=2

# Simple wrapper around "echo"
function loginfo() {
  echo "$@" >&2
}

# Emits failure message in a structured format that can be extracted
# and reported to user as part of cluster creation error.
# Default error code is 1, which indicates internal error.
# Caller of this function can exit 2 instead, which indicates client error.
function log_and_fail() {
  local component=$1
  local message=$2
  local error_code="${3:-$EXIT_CODE_INTERNAL_ERROR}"

  local client_error_indicator=""
  if [[ "${error_code}" -eq "${EXIT_CODE_CLIENT_ERROR}" ]]; then
    client_error_indicator="(client-error)"
  fi
  # Record error in a format that can be extracted and reported.
  echo "StructuredError{${component}${client_error_indicator}, ${message}}"
  exit "${error_code}"
}

# Log call stack trace on exit.
function logstacktrace() {
  local err=$?
  local code="${1:-$EXIT_CODE_INTERNAL_ERROR}"

  set +o xtrace

  # Print out the stack trace described by $function_stack
  echo "ERROR: '${BASH_COMMAND}' exited with status $err"
  echo "Call tree:"
  for ((i = 0; i < ${#FUNCNAME[@]} - 1; i++)); do
    echo " $i: ${BASH_SOURCE[$i + 1]}:${BASH_LINENO[$i]} ${FUNCNAME[$i + 1]}"
  done

  echo "Exiting with status ${code}"
  exit "${code}"
}

function set_log_tag() {
  local -r tag="$1"
  exec 1> >(logger -s -t "${tag}[${BASHPID}]") 2>&1
}
