# Checks if the given metric source is enabled.
# First check dataproc.monitoring.stackdriver.enable property, if it is set
# return 0. This property is used as back-compat and will be deprecated soon.
# If dataproc.monitoring.stackdriver.enable property is false, then
# dataproc.monitoring.metric.sources property is checked if it has given
# metric source in list, then return 0.
function is_metric_source_enabled() {
  local -r input_metric_source=$1
  local stack_driver_monitoring_enabled
  stack_driver_monitoring_enabled=$(get_dataproc_property dataproc.monitoring.stackdriver.enable)

  # If stack driver monitoring is enabled
  if [[ "${stack_driver_monitoring_enabled}" == "true" ]]; then
    # only hdfs, yarn and monitoring_agent_defaults are applicable via dataproc.monitoring.stackdriver.enable cluster property
    # TODO: Remove hadoop from this list once b/211579403 is done
    metric_sources_applicable_array=( 'hadoop' 'hdfs' 'yarn')
    # for obtaining monitoringAgentDefaults in 2.2+ images with UCP agents, customers must install Ops Agent
    if ! use_ucp_observability_agents; then
      metric_sources_applicable_array+=('monitoringAgentDefaults')
    fi
    if [[ " ${metric_sources_applicable_array[*]} " =~ " ${input_metric_source} " ]]; then
      return 0
    fi
    return 1
  fi

  local metric_sources_enabled_array
  get_enabled_metric_sources_array metric_sources_enabled_array
  # Below statement does contains check: https://stackoverflow.com/a/15394738/3227693
  if [[ " ${metric_sources_enabled_array[*]} " =~ " ${input_metric_source} " ]]; then
    return 0
  fi

  return 1
}

# Checks if the any metric source is enabled.
# First check dataproc.monitoring.stackdriver.enable property, if it is set
# return 0. This property is used as back-compat and will be deprecated soon.
# If dataproc.monitoring.stackdriver.enable property is false, then
# dataproc.monitoring.metric.sources property is checked if it has any
# metric source in list, then return 0.
function is_any_metric_source_enabled() {
  local stack_driver_monitoring_enabled
  stack_driver_monitoring_enabled=$(get_dataproc_property dataproc.monitoring.stackdriver.enable)

  if [[ "${stack_driver_monitoring_enabled}" == "true" ]]; then
    return 0
  fi

  local metric_sources_enabled_array
  get_enabled_metric_sources_array metric_sources_enabled_array

  if [[ "${#metric_sources_enabled_array[@]}" -gt 0 ]]; then
    return 0
  else
    return 1
  fi
}

function is_default_metrics_enabled() {
  local -r default_metrics_enabled=$(get_dataproc_property dataproc.monitoring.default.metrics.enable)
  [[ "${default_metrics_enabled}" == "true" ]]
}

function is_default_system_metrics_enabled() {
  local -r default_metrics_system_enabled=$(get_dataproc_property dataproc.monitoring.default.metrics.system.enable)
  [[ "${default_metrics_system_enabled}" == "true" ]]
}

function get_enabled_metric_sources_array() {
  # Using nameref to return array https://stackoverflow.com/a/49971213
  local -n ms_array=$1
  local metric_sources_enabled_delimited_by_comma
  metric_sources_enabled_delimited_by_comma="$(get_dataproc_property dataproc.monitoring.metric.sources)"
  # for obtaining monitoringAgentDefaults in 2.2+ images with UCP agents, customers must install Ops Agent
  if use_ucp_observability_agents; then
    metric_sources_enabled_delimited_by_comma="${metric_sources_enabled_delimited_by_comma/monitoringAgentDefaults/}"
  fi
  ms_array=(${metric_sources_enabled_delimited_by_comma//,/ })
}
