source "${BDUTIL_DIR}/bdutil_gcs.sh"
source "${BDUTIL_DIR}/bdutil_os.sh"

# Get the agent memory on this node in mb.
function get_agent_memory_mb() {
  local role
  role="$(get_metadata_role)"

  if [[ "${DATAPROC_IMAGE_TYPE}" == "serverless-spark" ]]; then
    if [[ "${role}" == "Master" ]]; then
      echo 512
    else
      echo 160
    fi
  elif is_version_at_least "${DATAPROC_IMAGE_VERSION}" "2.0" && [[ "${role}" != "Master" ]]; then
    # LINT.IfChange
    echo 150
  else
    get_memory_mb 0.05
  fi
}

# Get directories to store local data. If additional disks are mounted it will
# be one directory be mountpoint. Otherwise it will be one on the boot
# partition.
# The naming of mounted directories is assumed to to be consistent with
# mount_disks.sh
function get_data_dirs() {
  local -a mount_points
  mapfile -t mount_points < <(find /mnt/[0-9]*/ -maxdepth 0 || true)
  if ! (("${#mount_points[@]}")); then
    echo /
    return
  fi
  for mount_point in "${mount_points[@]}"; do
    # Validate mount point
    mountpoint "${mount_point}" >&2
    echo "${mount_point}"
  done
}

# Space separated list of directories
function get_yarn_nm_local_dirs() {
  local -r data_dirs=($(get_data_dirs))
  echo "${data_dirs[*]/%/hadoop/yarn/nm-local-dir}"
}

# Downloads the artifact to ${DATAPROC_ARTIFACTS_DIR}, then creates a symbolic
# link to it, deletes the old artifact. If the artifact is local, skip the
# downloading step.
function replace_artifact() {
  local -r new_artifact="$1"
  local -r local_link="$2"
  local -r old_artifact="$3"
  local -r artifact_name="${new_artifact##*/}"

  if [[ "${new_artifact}" == gs://* ]]; then
    local -r local_artifact="${DATAPROC_ARTIFACTS_DIR}/${artifact_name}"
  else
    local -r local_artifact="${new_artifact}"
  fi

  # Check if the desired local artifact and local link are already created,
  # avoid redoing after restart.
  if [[ -f "${local_artifact}" &&
        -f "${local_link}" &&
        $(realpath "${local_artifact}") == $(realpath "${local_link}") ]]; then
    return 0
  fi

  # Download the new artifact if it is in GCS.
  mkdir -p "${DATAPROC_ARTIFACTS_DIR}"
  if [[ "${new_artifact}" == gs://* ]]; then
    retry_constant_custom 5 2 \
      download_resource "${new_artifact}" "${local_artifact}"
  fi

  # Create symbolic link.
  ln -s -f "${local_artifact}" "${local_link}"

  # Delete the old artifact.
  if [[ -n "${old_artifact}" ]]; then
    rm -f "${old_artifact}"
  fi
}

function sha256sum_10() {
  echo "$1" | /usr/bin/sha256sum | cut -f1 -d" " | head -c 10
}

function is_hermetic_vm() {
  local -r sentinel_file="${DATAPROC_ETC_DIR}/hermetic_vm"

  if [[ -f "${sentinel_file}" ]]; then
    return 0
  fi

  local hermetic_vm
  hermetic_vm=$(/usr/share/google/get_metadata_value attributes/hermetic-vm ||
    echo "false")

  if [[ "${hermetic_vm}" == "true" ]]; then
    touch "${sentinel_file}"
    return 0
  fi

  return 1
}

function is_java8() {
  [[ "${JAVA_HOME}" == *'-8-'* ]]
}

function get_instance_name_prefix() {
  local optional_prefix
  optional_prefix="$(get_dataproc_property dataproc.gce-resource-prefix)"
  local cluster_name
  cluster_name="$(get_metadata_cluster_name)"
  echo "${optional_prefix}${cluster_name}"
}

function get_size_in_bytes() {
  # Capitalize: 100m => 100M
  local size=${1^^}
  # Remove trailing 'B': 100MIB => 100MI
  size=${size%B}
  # Remove trailing 'I': 100MI => 100M
  size=${size%I}
  # Convert size to bytes:
  # https://www.gnu.org/software/coreutils/manual/html_node/numfmt-invocation.html
  numfmt --from=iec "${size}"
}

function get_include_file_path() {
  local include_path
  # New images will have include file path set in the VM metadata, we fallback
  # to figuring out the path here otherwise.
  include_path=$(/usr/share/google/get_metadata_value \
    'attributes/dataproc-include-file-location' || echo "")

  if [[ -z "${include_path}" ]]; then
    include_path="$(get_include_file_path_deprecated)"
  fi

  echo "${include_path}"
}

function create_or_validate_include_file_path() {
  local include_path
  include_path="$(get_include_file_path)"
  if [[ "${include_path}" == gs://* ]]; then
    loginfo "Checking if include membership files exist in GCS: ${include_path}"
    if ! gcs_file_exists "${include_path}"; then
          log_and_fail "${include_path} does not exist in GCS!"
    fi
  else
    if [[ ! -f "${include_path}" ]]; then
      loginfo "Creating local membership files: ${include_path}"

      # Create empty membership files initially to indicate all nodes are allowed.
      # The include file will always be empty. The exclude file is populated with
      # names of nodes to remove during decommission.
      touch "${include_path}"

      # The yarn resourcemanager and hdfs namenode need to read these files.
      chown root:hadoop "${include_path}"
      chmod ugo+rw "${include_path}"
    fi
  fi
  echo "${include_path}"
}

# Eventually we will consolidate this so that the include file is always in GCS.
function get_include_file_path_deprecated() {
  local include_path
  if [[ "${NUM_MASTERS}" -eq 1 ]]; then
    include_path="${HADOOP_CONF_DIR}/nodes_include"
  else
    include_path="gs://${CONFIGBUCKET}/google-cloud-dataproc-metainfo/${CLUSTER_UUID}/nodes_include"
  fi
  echo "${include_path}"
}

function get_exclude_file_path() {
  local exclude_path
  exclude_path=$(/usr/share/google/get_metadata_value \
    'attributes/dataproc-exclude-file-location' || echo "")

  if [[ -z "${exclude_path}" ]]; then
    exclude_path="$(get_exclude_file_path_deprecated)"
  fi

  echo "${exclude_path}"
}

function create_or_validate_exclude_file_path() {
  local exclude_path
  exclude_path="$(get_exclude_file_path)"
  if [[ "${exclude_path}" == gs://* ]]; then
    loginfo "Checking if exclude membership files exist in GCS: ${exclude_path}"
    if ! gcs_file_exists "${exclude_path}"; then
          log_and_fail "${exclude_path} does not exist in GCS!"
    fi
  else
    if [[ ! -f "${exclude_path}" ]]; then
      loginfo "Creating local membership files: ${exclude_path}"

      # Create empty membership files initially to indicate all nodes are allowed.
      # The include file will always be empty. The exclude file is populated with
      # names of nodes to remove during decommission.
      touch "${exclude_path}"
      # If the exclude file path ends in .xml, initially populate it with <hosts/>
      # so that it is valid xml. XML-formatted include files are not supported.
      if [[ "${exclude_path}" == *'.xml' ]]; then
        cat > "${exclude_path}" <<EOF
<hosts/>
EOF
      fi

      # The yarn resourcemanager and hdfs namenode need to read these files.
      chown root:hadoop "${exclude_path}"
      chmod uro+rw "${exclude_path}"
    fi
  fi
  echo "${exclude_path}"
} 

# Eventually we will consolidate this so that the exclude file is always in GCS
# and always has the .xml file extension.
function get_exclude_file_path_deprecated() {
  local exclude_path
  MASTER_RUN_DRIVER_LOCATION=$(/usr/share/google/get_metadata_value \
    'attributes/master-run-driver-location' || echo "")

  if [[ "${MASTER_RUN_DRIVER_LOCATION}" == 'YARN' ]]; then
    if [[ "${NUM_MASTERS}" -eq 1 ]]; then
      exclude_path="${HADOOP_CONF_DIR}/nodes_exclude.xml"
    else
      exclude_path="gs://${CONFIGBUCKET}/google-cloud-dataproc-metainfo/${CLUSTER_UUID}/nodes_exclude.xml"
      # Fall back to old non-xml exclude path if xml one does not exist.
      # This prevents breaking private preview driver pools with HA clusters.
      # This fallback can be removed once dispatcher change to create xml-based
      # exclude file in gcs is fully rolled out to prod.
      if ! gcs_file_exists "${exclude_path}"; then
        exclude_path="gs://${CONFIGBUCKET}/google-cloud-dataproc-metainfo/${CLUSTER_UUID}/nodes_exclude"
      fi
    fi
  else
    if [[ "${NUM_MASTERS}" -eq 1 ]]; then
      exclude_path="${HADOOP_CONF_DIR}/nodes_exclude"
    else
      exclude_path="gs://${CONFIGBUCKET}/google-cloud-dataproc-metainfo/${CLUSTER_UUID}/nodes_exclude"
    fi
  fi
  echo "${exclude_path}"
}

# Generate a random password
function random_password() {
  # Add a fixed prefix to make sure it meets the strength requirements.
  #
  # Replace '/' in the password with '_' to avoid being interpreted as special
  # character in some cases, e.g., used as part of `sed`.
  echo "Pw1+$(cat /dev/urandom | base64 | tr '/' '_' | head -c 8)"
}

# Get the architecture suffix most commonly used.
function get_arch_suffix() {
  local architecture
  architecture="$(uname -m)"

  local arch_suffix
  case "${architecture}" in
    "aarch64" | "arm64" )
      arch_suffix="arm64"
      ;;
    "x86_64" | "amd64" )
      arch_suffix="amd64"
      ;;
    *)
      loginfo "Unsupported arch : ${architecture}"
      exit 1
  esac
  echo "${arch_suffix}"
}

function is_arm64() {
  [[ "$(uname -m)" == 'aarch64' ]]
}

function is_containerised_legacy_agents_supported() {
  if [[ "${DATAPROC_IMAGE_TYPE}" == "standard" ]] &&
    is_version_at_least "${DATAPROC_IMAGE_VERSION}" "2.2"; then
    return 0
  fi
  return 1
}

function is_containerised_legacy_agents_supported_for_s8s() {
  if [[ "${DATAPROC_IMAGE_TYPE}" == 'serverless-spark' ]] && is_debian12; then
    return 0
  fi
  return 1
}

function use_ucp_observability_agents() {
  if [[ "${DATAPROC_IMAGE_TYPE}" == 'serverless-spark' ]]; then
    # not supported for s8s yet
    return 1
  fi

  local -r legacy_agents_enabled_prop="$(get_dataproc_property dataproc.observability.containerised.legacy.agents.enable)"
  if ! is_containerised_legacy_agents_supported; then
    # not supported on older image versions.
    return 1
  fi
  if [[ "${legacy_agents_enabled_prop}" == "true" ]]; then
    # UCP agents are not supported if legacy containerised agents are enabled.
    return 1
  fi
  return 0
}

function join_by_char() {
  local IFS="$1"
  shift
  echo "$*"
}
