readonly IPTABLES_RESTORE_SCRIPT='/usr/local/share/google/dataproc/iptables_restore.sh'
readonly IPTABLES_SAVE_LOCATION='/usr/local/share/google/dataproc/iptables_saved'

function download_resource() {
  local -r remote_uri=$1
  local -r local_destination=$2
  local -r local_symlink=${3:-}

  if [[ $remote_uri == gs://* ]]; then
    gsutil -q cp "${remote_uri}" "${local_destination}"
  else
    curl -fsSL --retry-connrefused --retry 10 --retry-max-time 30 \
      "${remote_uri}" -o "${local_destination}"
  fi

  if [[ -n $local_symlink ]]; then
    ln -s "${local_destination}" "${local_symlink}"
  fi
}

# Waits for service on a given port to come up.
function wait_for_port() {
  local -r name="$1"
  local -r host="$2"
  local -r port="$3"
  local -r timeout="${4:-300}"

  # We only respect timeouts up to 1800 seconds (30 minutes).
  local -r capped_timeout=$((timeout > 1800 ? 1800 : timeout))

  loginfo "Waiting ${capped_timeout} seconds for service to come up on host=${host} port=${port} name=${name}."
  # TODO: replace `nc` everywhere with something
  # that doesn't send requests; maybe `fuser`?
  retry_constant_custom \
    "${capped_timeout}" 1 nc -v -z -w 1 "${host}" "${port}"
  loginfo "Service up on host=${host} port=${port} name=${name}."
}

# Print a line of diagnostic for each TCP connection on a given port.
# Strips the header. Lines should include pid=<PID>.
function get_tcp_port_info() {
  port=$1
  ss -apt "( sport = :${port} )" | sed '1d'
}

# True if the specified single IPv4 tcp port is available.
function tcp_port_is_free() {
  port=$1
  ss_lines=$(get_tcp_port_info "${port}" | wc -l)
  return $((ss_lines > 0))
}

# True if all specified IPv4 tcp ports are available.
function all_tcp_ports_are_free() {
  local -r ports=("$@")
  for p in "${ports[@]}"; do
    if ! tcp_port_is_free "$p"; then
      return 1
    fi
  done
  return 0
}

# Prints the list of ports in use.
function tcp_ports_in_use() {
  local -r ports=("$@")
  local -a in_use
  for p in "${ports[@]}"; do
    if ! tcp_port_is_free "$p"; then
      in_use+=("$p")
    fi
  done
  echo "${in_use[@]}"
}

# Wait until all of the specified IPv4 tcp ports are unused.
# Returns 0 (true) if all ports are free, or 1 (false) if we timed out
# before all ports are free. The timeout can be set by specifying it
# with the --max_wait_seconds option (must be the first argument).
# All other arguments are port names or numbers.
function wait_until_ports_free() {
  local start_time
  start_time=$(date '+%s')
  if [[ "$1" == '--max_wait_seconds' ]]; then
    local -r max_wait_seconds=$2
    shift 2
  else
    local -r max_wait_seconds=130
  fi
  local -r ports=("$@")
  local -r seconds_between_checks=1
  local now=${start_time}
  while ((now < start_time + max_wait_seconds)); do
    if all_tcp_ports_are_free "${ports[@]}"; then
      return 0
    fi
    sleep "$seconds_between_checks"
    now=$(date '+%s')
  done
  # Timed out before all ports are free.
  return 1
}

# Wait for ports to be free, and output some info about it.
# $1 is an array of port numbers.
function wait_until_ports_free_and_report() {
  local -r ports=("$@")
  loginfo "Testing for available ports: ${ports[*]}"
  local in_use
  in_use=$(tcp_ports_in_use "${ports[@]}")
  if [[ "${in_use[*]}" == "" ]]; then
    loginfo "All tested ports are available."
    return 0
  fi
  loginfo "Waiting for ports that are not yet available: ${in_use[*]}."
  local port_wait_start_time
  port_wait_start_time=$(date '+%s')
  local port_wait_status=0
  wait_until_ports_free "${ports[@]}" || port_wait_status=$?
  local port_wait_end_time
  port_wait_end_time=$(date '+%s')
  local port_wait_time=$((port_wait_end_time - port_wait_start_time))
  if [[ ${port_wait_time} -gt 0 ]]; then
    loginfo "Wait time in seconds for ports ${ports[*]} was ${port_wait_time}."
    local still_in_use
    still_in_use=$(tcp_ports_in_use "${ports[@]}")
    if [[ "${still_in_use}" != "" ]]; then
      loginfo "Ports still in use: ${still_in_use}."
      for port in ${still_in_use}; do
        echo "${port} is still used by these processes:"
        get_tcp_port_info "${port}" |
          sed -n 's/.*pid=\([0-9]*\).*/\1/p' |
          xargs -r ps -fwwp
      done
    fi
  fi
  if [[ "${still_in_use}" == "" ]]; then
    loginfo "All tested ports are now available."
  fi
  return ${port_wait_status}
}

# Create a script to restore the saved IP tables rules when VMs are rebooted.
#
# The restore script is located at the path specified by the
# `IPTABLES_RESTORE_SCRIPT` variable.
#
# In order to be restored, the IP table rules must be saved to the location
# specified by the `iptables_save_location` function, which is distribution
# specific. That can be done by invoking the `save_iptables` method below.
function create_iptables_restore_script() {
  cat >"${IPTABLES_RESTORE_SCRIPT}" <<EOF
#!/bin/bash
if [[ -f "${IPTABLES_SAVE_LOCATION}" ]]; then
  iptables-restore < "${IPTABLES_SAVE_LOCATION}"
fi
EOF
  chmod u+rwx "${IPTABLES_RESTORE_SCRIPT}"
}

# This function generates and saves the snapshot of the iptables rules.
#
# The restore script automatically restores the IP table rules from that
# snapshot using the `iptables-restore` command.
#
# Every component that updates IP tables rules should invoke this function
# (it's safe to run more than once).
function save_iptables() {
  mkdir -p "$(dirname ${IPTABLES_SAVE_LOCATION})"
  iptables-save >"${IPTABLES_SAVE_LOCATION}" || err "Failed to save iptables"
}

# Wait until a remote server:port/path is ready
function wait_until_url_ready() {
  local -r url=$1
  # Default sleep 5
  local -r retry_delay_sec=${2:-5}
  local -r timeout_sec=300
  curl --head --retry-delay "${retry_delay_sec}" --retry-max-time "${timeout_sec}" --output /dev/null --silent "${url}"
}
