function os_id() {
  grep '^ID=' /etc/os-release | cut -d= -f2 | xargs
}

function os_version() {
  grep '^VERSION_ID=' /etc/os-release | cut -d= -f2 | xargs
}

function is_debian() {
  [[ "$(os_id)" == 'debian' ]]
}

function is_debian10() {
  is_debian && [[ "$(os_version)" == '10'* ]]
}

function is_debian11() {
  is_debian && [[ "$(os_version)" == '11'* ]]
}

function is_debian12() {
  is_debian && [[ "$(os_version)" == '12'* ]]
}

function is_rocky() {
  [[ "$(os_id)" == 'rocky' ]]
}

function is_rocky8() {
  is_rocky && [[ "$(os_version)" == 8\.* ]]
}

function is_rocky9() {
  is_rocky && [[ "$(os_version)" == 9\.* ]]
}

function is_ubuntu() {
  [[ "$(os_id)" == 'ubuntu' ]]
}

function is_ubuntu18() {
  is_ubuntu && [[ "$(os_version)" == 18\.* ]]
}

function is_ubuntu20() {
  is_ubuntu && [[ "$(os_version)" == 20\.* ]]
}

function is_ubuntu22() {
  is_ubuntu && [[ "$(os_version)" == 22\.* ]]
}

# Helper to overwrite the contents of file specified by $1 with strings
# specified as ${@:2}, using "echo". This helper is convenient for cases
# where a command string may be passed around in a way that makes precedence
# of stdin/stdout redirection characters unclear.
# Example: overwrite_file_with_strings foo.txt hello world
function overwrite_file_with_strings() {
  local -r filename=$1
  local -r contents=${*:2}
  echo "Overwriting ${filename} with contents '${contents}'"
  echo "${contents}" >"${filename}"
}

# Get the requested fraction of total memory on this node in mb.
function get_memory_mb() {
  local -r fraction=$1
  # Calculate the memory allocations, MB, using 'free -m'. Floor to nearest MB.
  local total_mem
  total_mem=$(free -m | awk '/^Mem:/{print $2}')
  python -c "print(int(${total_mem} * ${fraction}))"
}

function mv_and_link() {
  local -r src=$1
  local -r dst=$2
  mv "$src" "$dst"
  ln -s "$dst" "$src"
}

function unlink_and_mv() {
  local -r link=$1
  local target
  target=$(readlink "$link")
  unlink "$link"
  mv "$target" "$link"
}

function user_exists() {
  local -r user="$1"
  id "${user}" >/dev/null 2>&1
}

function group_exists() {
  local -r group="$1"
  grep -q "^${group}:" /etc/group
}

# Copy everything into the syslog for debugging. Redirects stderr to stdout,
# and then both to logger, which tees between the syslog and stdout.
# Logs the PID. Does not propagate the error to stderr to avoid duplicate
# entries.
function run_with_logger() {
  local tag=''
  local pid="${BASHPID}"
  if [[ "$1" == "--tag" ]]; then
    tag="$2"
    shift 2
  fi

  if [[ "$#" -eq 0 ]]; then
    exec 1> >(logger -s -t "${tag}[${pid}]") 2>&1
  else
    "$@" 1> >(logger -s -t "${tag}[${pid}]") 2>&1
  fi
}

function run_in_background() {
  { trap 'echo "$?" >"${COMMANDS_TMP_DIR}/${BASHPID}.exitcode"' EXIT; run_with_logger "$@"; } &
  local -r pid=$!

  # Strip --tag <tag> from the command that we save. Otherwise, we'll end up
  # saving "--tag <tag> <cmd> <arg1>". This must match the behavior in
  # run_with_logger.
  if [[ "$1" == "--tag" ]]; then
    shift 2
  fi

  # Make sure that background process running file does not exist already
  [[ ! -f "${COMMANDS_TMP_DIR}/${pid}.running" ]]
  # Save background process PID and command in the running file for tracking
  echo "$*" >"${COMMANDS_TMP_DIR}/${pid}.running"

  echo "Started background process [$*] as pid ${pid}"
}

function wait_on_async_processes() {
  loginfo 'Waiting on async processes'
  local running_file
  for running_file in "${COMMANDS_TMP_DIR}/"*'.running'; do
    local pid
    pid="$(basename "${running_file%'.running'}")"
    local cmd
    cmd="$(<"${running_file}")"

    # Wait on it regardless to get its exit status.
    loginfo "Waiting on pid=${pid} cmd=[${cmd}]"
    echo "${cmd}" >>"${COMMANDS_TMP_DIR}/${pid}.waiting"

    local exitcode_file="${COMMANDS_TMP_DIR}/${pid}.exitcode"
    # Wait until async process finishes and get its status (exit code)
    while [[ ! -f "${exitcode_file}" ]]; do sleep 1; done
    local status
    status="$(<"${exitcode_file}")"

    if ((status != 0)); then
      # Log a more descriptive error message
      echo "ERROR: Command cmd=[${cmd}] pid=${pid} exited with ${status}" |
        tee "${COMMANDS_TMP_DIR}/${pid}.error"
      exit "$status"
    fi
    echo "Command cmd=[${cmd}] pid=${pid} exited with ${status}" |
      tee "${COMMANDS_TMP_DIR}/${pid}.done"
    # Remove background process running file to not wait for this process again
    rm "${exitcode_file}" "${running_file}"
  done
}

if is_rocky; then
  source "${BDUTIL_DIR}/os/rocky/bdutil_helpers.sh"
else
  source "${BDUTIL_DIR}/os/debian/bdutil_helpers.sh"
fi
