# Retrieves the last occurrence of the requested property_name from
# the specified file, unescaping any escaped characters found. Usage:
#   get_java_property <filename> <property_name>
# TODO: don't use java properties format in this file since
# keys can have '=' in them as well, so regex-based approaches written
# by hand are not a long-term answer.
function get_java_property() {
  set +x
  local -r property_file=$1
  local -r property_name=$2
  local property_value
  property_value=$(grep "^${property_name}=" "${property_file}" |
    tail -n 1 | cut -d '=' -f 2- | sed -r 's/\\([#!=:])/\1/g')
  echo "${property_value}"
  set -x
}

function get_java_property_or_default() {
  set +x
  local -r property_file="$1"
  local -r property_name="$2"
  local -r default_value="$3"
  local actual_value
  actual_value=$(get_java_property  "${property_file}" "${property_name}")
  if [[ -n "${actual_value}" ]]; then
    echo "${actual_value}"
  else
    echo "${default_value}"
  fi
  set -x
}

# Sets the property with the value in the property file based on the mode.
#
# Modes:
# 1) overwrite: overwrite any existing property value with the new value.
# 2) append: append the new value to the existing value.
# 3) skip: skip the new value if there is already an non-empty value.
function set_java_property() {
  set +x
  local -r property_file="$1"
  local -r property_name="$2"
  local -r property_value="$3"
  local -r mode="${4:-overwrite}"
  local -r delimiter="${5:-,}"

  local skip="false"
  local old_value
  local new_value

  if [[ "${mode}" == "overwrite" ]]; then
    new_value="${property_value}"
  elif [[ "${mode}" == "append" ]]; then
    old_value=$(get_java_property "${property_file}" "${property_name}")
    if [[ -n "${old_value}" ]]; then
      new_value="${old_value}${delimiter}${property_value}"
    else
      new_value="${property_value}"
    fi
  elif [[ "${mode}" == "skip" ]]; then
    old_value=$(get_java_property "${property_file}" "${property_name}")
    if [[ -n "${old_value}" ]]; then
      skip="true"
    else
      new_value=${property_value}
    fi
  fi

  if [[ "${skip}" == "false" ]]; then
    cat >>"${property_file}" <<EOF

${property_name}=${new_value}
EOF

  fi
  set -x
}

# Retrieves the value of the requested dataproc property from
# the /etc/google-dataproc/dataproc.properties file.
# Property names should not contain the dataproc: prefix.
function get_dataproc_property() {
  set +x
  # Remove "dataproc:" prefix if provided.
  local -r property_name="${1#dataproc:}"
  local property_value
  property_value=$(get_java_property \
    /etc/google-dataproc/dataproc.properties "${property_name}")
  echo "${property_value}"
  set -x
}

function get_dataproc_property_or_default() {
  set +x
  # Remove "dataproc:" prefix if provided.
  local -r property_name="${1#dataproc:}"
  local -r default_value="$2"
  local actual_value
  actual_value=$(get_dataproc_property "${property_name}")
  if [[ -n "${actual_value}" ]]; then
    echo "${actual_value}"
  else
    echo "${default_value}"
  fi
  set -x
}

# Sets the property with the value in the XML file based on the mode.
#
# Modes:
# 1) overwrite: overwrite any existing property value with the new value.
# 2) append: append the new value to the existing value.
# 3) skip: skip the new value if there is already an non-empty value.
function set_property_in_xml() {
  local -r xml_file=$1
  local -r name=$2
  local -r value=$3
  local -r mode="${4:-overwrite}"
  local -r delimiter="${5:-,}"

  local skip="false"
  local old_value
  local new_value

  if [[ "${mode}" == "overwrite" ]]; then
    new_value="${value}"
  elif [[ "${mode}" == "append" ]]; then
    old_value=$(get_property_in_xml "${xml_file}" "${name}")
    if [[ -n "${old_value}" ]]; then
      new_value="${old_value}${delimiter}${value}"
    else
      new_value="${value}"
    fi
  elif [[ "${mode}" == "skip" ]]; then
    old_value=$(get_property_in_xml "${xml_file}" "${name}")
    if [[ -n "${old_value}" ]]; then
      skip="true"
    else
      new_value=${value}
    fi
  fi

  if [[ "${skip}" == "false" ]]; then
    bdconfig set_property \
      --configuration_file "${xml_file}" \
      --name "${name}" \
      --value "${new_value}" \
      --create_if_absent \
      --clobber
  fi
}

function get_property_in_xml() {
  set +x
  local -r path="$1"
  local -r property="$2"
  local -r default_value="${3:-}"

  local val
  val=$(bdconfig get_property_value \
    --configuration_file "${path}" \
    --name "${property}" 2>/dev/null)
  if [[ "${val}" == "None" ]]; then
    val="${default_value}"
  fi
  echo "${val}"
  set -x
}

# Sets the property with the value in the YAML file based on the mode.
#
# Modes:
# 1) overwrite: overwrite any existing property value with the new value.
# 2) append: append the new value to the existing value.
# 3) skip: skip the new value if there is already an non-empty value.
function set_property_in_yaml() {
  local -r yaml_file=$1
  local -r name=$2
  local -r value=$3
  local -r mode="${4:-overwrite}"
  local -r delimiter="${5:-,}"

  local skip="false"
  local old_value
  local new_value

  if [[ "${mode}" == "overwrite" ]]; then
    new_value="${value}"
  elif [[ "${mode}" == "append" ]]; then
    old_value=$(get_property_in_yaml "${yaml_file}" "${name}")
    if [[ -n "${old_value}" ]]; then
      new_value="${old_value}${delimiter}${value}"
    else
      new_value="${value}"
    fi
  elif [[ "${mode}" == "skip" ]]; then
    old_value=$(get_property_in_yaml "${yaml_file}" "${name}")
    if [[ -n "${old_value}" ]]; then
      skip="true"
    else
      new_value=${value}
    fi
  fi

  if [[ "${skip}" == "false" ]]; then
    bdconfig yaml_set_property \
      --configuration_file "${yaml_file}" \
      --name "${name}" \
      --value "${new_value}" \
      --create_if_absent \
      --clobber
  fi
}

function get_property_in_yaml() {
  set +x
  local -r path="$1"
  local -r property="$2"
  local -r default_value="${3:-}"

  local val
  val=$(bdconfig yaml_get_property_value \
    --configuration_file "${path}" \
    --name "${property}" 2>/dev/null)
  if [[ "${val}" == "None" ]]; then
    val="${default_value}"
  fi
  echo "${val}"
  set -x
}

function get_property_core_site_or_default() {
  set +x
  local -r property="$1"
  local -r default_value="${2:-}"

  get_property_in_xml \
    "${HADOOP_CONF_DIR}/core-site.xml" \
    "${property}" \
    "${default_value}"
  set -x
}

function set_property_core_site() {
  local -r name=$1
  local -r value=$2
  set_property_in_xml "${HADOOP_CONF_DIR}/core-site.xml" "${name}" "${value}"
}

function set_property_hive_site() {
  local -r name=$1
  local -r value=$2
  set_property_in_xml "/etc/hive/conf/hive-site.xml" "${name}" "${value}"
}

function set_property_hdfs_site() {
  local -r name=$1
  local -r value=$2
  set_property_in_xml "${HADOOP_CONF_DIR}/hdfs-site.xml" "${name}" "${value}"
}

function set_property_yarn_site() {
  local -r name=$1
  local -r value=$2
  set_property_in_xml "${HADOOP_CONF_DIR}/yarn-site.xml" "${name}" "${value}"
}

function set_property_mapred_site() {
  local -r name=$1
  local -r value=$2
  set_property_in_xml "${HADOOP_CONF_DIR}/mapred-site.xml" "${name}" "${value}"
}

function set_property_tez_site() {
  local -r name=$1
  local -r value=$2
  set_property_in_xml "/etc/tez/conf/tez-site.xml" "${name}" "${value}"
}

function set_property_dpms() {
  local -r name="$1"
  local -r value="$2"
  set_property_in_xml "${DPMS_PROXY_CONF_DIR}/proxy-config.xml" "${name}" "${value}"
}

# Merges xml conf properties files.
function merge_xml_properties() {
  local src=$1
  local dest=$2

  if [[ ! -f ${src} ]]; then
    loginfo "Skipping merging ${src}, file does not exist."
    return 0
  fi

  bdconfig merge_configurations \
    --configuration_file "${dest}" \
    --source_configuration_file "${src}" \
    --resolve_environment_variables \
    --create_if_absent \
    --clobber
  loginfo "Merged ${src}."
}

# Merges yaml conf properties files.
function merge_yaml_properties() {
  local src=$1
  local dest=$2

  if [[ ! -f ${src} ]]; then
    loginfo "Skipping merging ${src}, file does not exist."
    return 0
  fi

  bdconfig yaml_merge_configurations \
    --configuration_file "${dest}" \
    --source_configuration_file "${src}" \
    --resolve_environment_variables \
    --create_if_absent \
    --clobber
  loginfo "Merged ${src}."
}

# Merges java properties files.
function merge_java_properties() {
  local -r src=$1
  local -r dest=$2
  local -r header=${3:-"\n# User-supplied properties."}

  if [[ ! -f ${src} ]]; then
    loginfo "Skipping merging ${src}, file does not exist."
    return 0
  fi

  echo -e "${header}" >>"${dest}"
  cat "${src}" >>"${dest}"

  loginfo "Merged ${src}."
}

# Merges sh env files.
function merge_sh_env_vars() {
  local src=$1
  local dest=$2

  if [[ ! -f ${src} ]]; then
    loginfo "Skipping merging ${src}, file does not exist."
    return 0
  fi

  echo -e "\n# User-supplied properties." >>"${dest}"
  cat "${src}" >>"${dest}"

  loginfo "Merged ${src}."
}

# function requires a hadoop property file as a parameter.
# It will replace the values of hostname (0.0.0.0) for specific keys to the hostname
function update_hostname_for_property {
  local xml_file="${1}"
  local prop_name="${2}"
  local host_name="${3}"
  local updated_value
  local prop_value
  prop_value="$(bdconfig get_property_value \
    --configuration_file "${xml_file}" \
    --name "${prop_name}")"

  if [[ "${prop_value}" != "None" ]]; then
    loginfo "Reading ${prop_name} in ${xml_file}"
    # replace the ip address 0.0.0.0 with hostname
    updated_value="$(echo "${prop_value}" | sed -e 's/0.0.0.0/'"${host_name}"'/g')"

    if [[ "${updated_value}" != "${prop_value}" ]]; then
      bdconfig set_property \
        --configuration_file "${xml_file}" --name "${prop_name}" \
        --value "${updated_value}"
      loginfo "Updating ${prop_name} in ${xml_file} to ${updated_value}"
    fi
  fi
}

# function requires a yaml config file as a parameter.
# It will replace the values of hostname (0.0.0.0) for
# specific keys to the hostname.
function yaml_update_hostname_for_property {
  local yaml_file="${1}"
  local prop_name="${2}"
  local host_name="${3}"
  if ! is_version_at_least "${DATAPROC_IMAGE_VERSION}" "2.0"; then
    log_and_fail "yaml_update" \
       "YAML config update is not supported in ${DATAPROC_IMAGE_VERSION}!" \
       "${EXIT_CODE_INTERNAL_ERROR}"
  fi

  local updated_value
  local prop_value
  prop_value="$(bdconfig yaml_get_property_value \
    --configuration_file "${yaml_file}" \
    --name "${prop_name}")"

  if [[ "${prop_value}" != "None" ]]; then
    loginfo "Reading ${prop_name} in ${yaml_file}"
    # replace the ip address 0.0.0.0 with hostname
    updated_value="$(echo "${prop_value}" | sed -e 's/0.0.0.0/'"${host_name}"'/g')"

    if [[ "${updated_value}" != "${prop_value}" ]]; then
      bdconfig yaml_set_property \
        --configuration_file "${yaml_file}" --name "${prop_name}" \
        --value "${updated_value}"
      loginfo "Updating ${prop_name} in ${yaml_file} to ${updated_value}"
    fi
  fi
}
