# Helper to run any command with default bdutil settings with Fibonacci
# backoff. If all retries fail, returns last attempt's exit code.
# Args: "$@" is the command to run.
function retry_fibonacci() {
  local -r cmd=("$@")
  local -r retry_backoff=(1 1 2 3 5 8 13 21 34 55 89 144)
  local -r max_retries="${#retry_backoff[@]}"

  loginfo "About to run '${cmd[*]}' with retries..."
  for ((i = 0; i < ${max_retries}; i++)); do
    if "${cmd[@]}"; then
      loginfo "${cmd[*]} succeeded."
      return
    fi

    local retry_delay=${retry_backoff[$i]}
    loginfo "'${cmd[*]}' attempt $((i + 1))/${max_retries} failed! Sleeping ${retry_delay}s." >&2
    sleep "${retry_delay}"
  done

  loginfo "Final attempt of '${cmd[*]}'..."
  # Let any final error propagate all the way out to any error traps.
  "${cmd[@]}"
}

# Retries command until successful or up to a variable number of seconds
# (excluding command runtime).
# Sleeps for N seconds between the attempts.
# Max Num Attempts = Retry Time / Retry Delay
function retry_constant_custom() {
  local -r max_retry_time="$1"
  local -r retry_delay="$2"
  local -r cmd=("${@:3}")

  local -r max_retries=$((max_retry_time / retry_delay))

  # Disable debug logs (if enabled) to not polute logs with retry attempts
  local reenable_x=false
  if [[ -o xtrace ]]; then
    set +x
    reenable_x=true
  fi

  local last_log_timestamp="0"
  loginfo "About to run '${cmd[*]}' with retries..."
  for ((i = 1; i < ${max_retries}; i++)); do
    if "${cmd[@]}"; then
      loginfo "'${cmd[*]}' succeeded after ${i} execution(s)."
      if [[ "${reenable_x}" == "true" ]]; then
        set -x
      fi
      return 0
    fi

    local timestamp
    timestamp=$(date +%s)
    # Log at most once per 10 seconds
    if ((timestamp - last_log_timestamp > 10)); then
      loginfo "'${cmd[*]}' attempt ${i}/${max_retries} failed! Sleeping ${retry_delay}s." >&2
      last_log_timestamp="${timestamp}"
    fi
    sleep "${retry_delay}"
  done

  loginfo "Final attempt of '${cmd[*]}'..."
  if [[ "${reenable_x}" == "true" ]]; then
    set -x
  fi
  # Let any final error propagate all the way out to any error traps.
  "${cmd[@]}"
}

# Retries command until successful or up to 300 attempts.
# Sleeps for 1 second between the attempts.
function retry_constant() {
  retry_constant_custom 300 1 "$@"
}

# Retries command until successful or up to 30 attempts.
# Sleeps for 1 seconds between the attempts.
function retry_constant_short() {
  retry_constant_custom 30 1 "$@"
}

# Retries command until successful or up to 300 attempts.
# Sleeps for 1 second between the attempts.
# Suitable for retrying commands with credentials because
# it doesn't log any information
function retry_quiet() {
  local -r cmd=("$@")
  local -r max_retries=300
  local -r retry_delay=1

  for ((i = 1; i < ${max_retries}; i++)); do
    if "${cmd[@]}"; then
      return 0
    fi
    sleep "${retry_delay}"
  done

  # Let any final error propagate all the way out to any error traps.
  "${cmd[@]}"
}
