COMMON_RESTART_DROP_IN="/etc/systemd/system/common/restart.conf"
WORKER_RESTART_DROP_IN="/etc/systemd/system/common/worker-restart.conf"

function create_restart_drop_in_config() {
  local -r conf_path="$1"
  local -r restart_type="$2"
  local -r interval_sec="$3"

  if [[ ! -f "${conf_path}" ]]; then
    mkdir -p "$(dirname "${conf_path}")"
  fi

  cat >"${conf_path}" <<EOF
[Unit]
StartLimitIntervalSec=${interval_sec}

[Service]
Restart=${restart_type}
EOF
}

function create_common_restart_drop_in_configs() {
  create_restart_drop_in_config "${COMMON_RESTART_DROP_IN}" on-failure 0
  create_restart_drop_in_config "${WORKER_RESTART_DROP_IN}" always 0
}

function enable_service() {
  local -r service="$1"
  local -r unit="${service}.service"

  retry_constant_short systemctl enable "${unit}"

  local -r drop_in_dir="/etc/systemd/system/${unit}.d"
  mkdir -p "${drop_in_dir}"

  # Check the existing systemd unit to determine if the service unit should be
  # amended to enable restart on failure.
  # If RemainAfterExit is not "no" then systemd will assume the service is
  # still running after all processes exit. Therefore restarting such a service
  # on exit would not make sense.
  # Note that systemd-sysv-generator will set RemainAfterExit=yes if it cannot
  # detect the pidfile in the LSB headers of the init.d script.
  local props
  props="$(retry_constant_short systemctl show "${unit}" -p Restart,RemainAfterExit)"
  if [[ "${service}" != 'hadoop-hdfs-datanode' &&
        "${service}" != 'hadoop-yarn-nodemanager' &&
        "${props}" == *"Restart=no"* &&
        "${props}" == *"RemainAfterExit=no"* ]]; then
    ln -s -f "${COMMON_RESTART_DROP_IN}" "${drop_in_dir}"
  fi

  local -r gate_start_on_agent_success_drop_in="/etc/systemd/system/common/agent-gate.conf"
  if [[ ! -f "${gate_start_on_agent_success_drop_in}" ]]; then
    mkdir -p "$(dirname "${gate_start_on_agent_success_drop_in}")"
    cat >"${gate_start_on_agent_success_drop_in}" <<EOF
[Unit]
ConditionPathExists=/var/lib/google/dataproc/ready_to_start_worker_services
EOF
  fi

  # The datanode and node manager should not start until the agent has completed
  # initialization successfully.
  if [[ "${service}" == 'hadoop-hdfs-datanode' ||
        "${service}" == 'hadoop-yarn-nodemanager' ]]; then
    local -r worker_count="$(get_metadata_worker_count)"
    if [[ "${worker_count}" != 0 ]]; then
      # On regular clusters, worker services are started later by the agent.
      # This drop in prevents worker services from being started by systemd
      # in the case the VM is rebooted and the agent has not yet run its
      # initialization actions.
      # On single node clusters (where worker count is 0), worker services are
      # started early in the startup script so that optional components that
      # depend on HDFS can be activated.
      ln -s -f "${gate_start_on_agent_success_drop_in}" "${drop_in_dir}"
    fi
    ln -s -f "${WORKER_RESTART_DROP_IN}" "${drop_in_dir}"
  fi
}

function start_service() {
  local -r service="$1"
  local -r unit="${service}.service"

  retry_constant_short systemctl start "${unit}" ||
    {
      journalctl -u "${unit}"
      return 1
    }
}

function enable_and_start_service() {
  local -r service="$1"

  enable_service "${service}"
  start_service "${service}"
}

# Checks whether the service unit file exists (but doesn't care about the status
# is enabled or not).
function is_service_installed() {
  local -r service="$1"

  local output
  local status
  for ((i = 0; i < 10; i++)); do
    output=$(systemctl cat "${service}.service" 2>&1)
    status=$?

    if [[ "${status}" -eq 0 ]]; then
      return 0
    fi

    if echo "${output}" | grep -q "No files found"; then
      return 1
    fi

    # The command failed for other reasons than the service doesn't exist, need
    # to retry.
    sleep 1
  done

  # Cannot determine whether the service is installed or not after exhausting
  # all retries.
  return 2
}

function is_service_enabled() {
  local -r service="$1"
  systemctl is-enabled "${service}.service"
}
