#!/bin/bash

set -euo pipefail

source "$(dirname "$0")/../../bdutil_components.sh"
source "$(dirname "$0")/../../bdutil_logging.sh"
source "$(dirname "$0")/../../bdutil_metadata.sh"
source "$(dirname "$0")/../../bdutil_properties.sh"
source "$(dirname "$0")/../../bdutil_retry.sh"
source "$(dirname "$0")/../../bdutil_services.sh"
source "$(dirname "$0")/../shared/containerd.sh"
source "$(dirname "$0")/../shared/nvidia-drivers.sh"

set -x

readonly CONTAINERD_ROOT='/var/dataproc/containerd'
DOWNLOADER_SA=$(get_dataproc_property 'spark.byosa.downloader.service-account')
readonly DOWNLOADER_SA

function is_custom_container_image_configured() {
  local custom_image
  custom_image=$(get_dataproc_property 'dataproc.spark.container.image')
  [[ -n "${custom_image}" ]]
}

# Set containerd root directory. This is used to store images, snapshots, etc
# so it should be on the data disk if one is mounted
function configure_containerd_root() {
  # This needs to be part of the root table of the toml file, so we need to
  # insert it as the first line in case there are already other sections
  sed -i "1s:^:root = \"${CONTAINERD_ROOT}\"\n:" "${CONTAINERD_CONFIG_FILE}"

  # Because we preload some container images, we need to link existing image
  # cache to the new Dataproc runtime directory when default images is used (we
  # can not just move image cache because it takes a significant amount of time)
  if is_custom_container_image_configured; then
    mkdir -p "${CONTAINERD_ROOT}"
    rm -rf '/var/lib/containerd'
  else
    mkdir -p "$(dirname "${CONTAINERD_ROOT}")"
    # TODO: Moving containerd root directory with default images between boot
    # disk and data disk (i.e. local SSD) takes time. We need to figure out
    # whether there is a way to link this directory instead of moving it and
    # configure containerd to write all new data (tmpmounts, etc) to the data
    # disk. See b/314989065 for context.
    mv '/var/lib/containerd' "${CONTAINERD_ROOT}"
  fi
}

function configure_nvidia_gpus() {
  sed -i -e '/\[plugins."io.containerd.grpc.v1.cri".containerd\]/r'<(
    cat <<EOF
    default_runtime_name = "nvidia"
EOF
  ) "${CONTAINERD_CONFIG_FILE}"

  sed -i -e '/\[plugins\]/r'<(
    cat <<EOF
  [plugins."io.containerd.grpc.v1.cri".containerd.runtimes.nvidia]
    privileged_without_host_devices = false
    runtime_engine = ""
    runtime_root = ""
    runtime_type = "io.containerd.runc.v1"
    [plugins."io.containerd.grpc.v1.cri".containerd.runtimes.nvidia.options]
      BinaryName = "/usr/bin/nvidia-container-runtime"
EOF
  ) "${CONTAINERD_CONFIG_FILE}"
}

function configure_snapshotter() {
  sed -i -e '/\[plugins."io.containerd.grpc.v1.cri".containerd\]/r'<(
    cat <<EOF
    snapshotter = "gcfs"
    disable_snapshot_annotations = false
EOF
  ) "${CONTAINERD_CONFIG_FILE}"

  cat >>"${CONTAINERD_CONFIG_FILE}" <<EOF
[proxy_plugins]
  [proxy_plugins.gcfs]
    type = "snapshot"
    address = "/run/containerd-gcfs-grpc/containerd-gcfs-grpc.sock"
EOF
}

# This is needed for BYOSA (go/dataproc-bigspark-byosa-design) feature
# on Batches. This updates the gcfsd config to add service account flag.
# If the spark downloader service account is set, it is used by gcfsd
# for image streaming authentication using impersonation.
function configure_gcfsd_downloader_service_account() {
  sed -i -e "s/--log_level=/--service_account=${DOWNLOADER_SA} --log_level=/g" \
    '/etc/systemd/system/gcfsd.service'
}

# This is needed for BYOSA (go/dataproc-bigspark-byosa-design) feature
# on Batches. The custom container is prefetched during cluster creation
# because we need to impersonate the downloader SA as root.
function prefetch_custom_container_using_downloader() {
  set +x

  # Use root token to impersonate and retrieve downloader service account tokens
  local token
  token=$(get_auth_token)
  local -r impersonate_api="https://iamcredentials.googleapis.com/v1/projects/-/serviceAccounts/${DOWNLOADER_SA}:generateAccessToken"
  token=$(curl -fsS --retry-connrefused --retry 10 --retry-max-time 30 \
    -H "Authorization: Bearer ${token}" \
    -H 'Content-Type:application/json' \
    -d '{"scope":["https://www.googleapis.com/auth/cloud-platform"]}' "${impersonate_api}" |
    jq -r ".accessToken")

  echo "Pulling custom container image ${image} using downloader service account ${DOWNLOADER_SA}."

  # Create and grant permissions to log file before starting the async image
  # pull process. This is read by executors during the job process to verify
  # process completion.
  touch "${BYOSA_IMAGE_PULL_STATUS_FILE}"
  chmod a+r "${BYOSA_IMAGE_PULL_STATUS_FILE}"

  # Make the image pull request for image streaming. In most cases,
  # this should respond back immediately, however, if something goes wrong with
  # Riptide, it falls back to downloading the entire image on local
  # disk, which can take a long time. To avoid holding up cluster startup,
  # `crictl pull` is ran as an asynchronous process here. The job process later
  # waits and reads the status file to determine if this process was completed.
  pull_image_request_async "${token}" "${image}" &

  set -x
}

function pull_image_request_async() {
  local -r token="$1"
  local -r image="$2"

  retry_quiet crictl pull --creds="oauth2accesstoken:${token}" "${image}" \
    &>>"${BYOSA_IMAGE_PULL_STATUS_FILE}"
  # Indicate that process was complete regardless of its status
  echo "DONE" >>"${BYOSA_IMAGE_PULL_STATUS_FILE}"
}

function main() {
  configure_containerd_root

  if [[ -n "${DOWNLOADER_SA}" ]]; then
    configure_gcfsd_downloader_service_account
  fi

  if is_component_selected 'nvidia-drivers' && is_gpu_attached; then
    configure_nvidia_gpus
  fi

  local riptide_enabled
  if [[ "${DATAPROC_IMAGE_TYPE}" == "serverless-spark" ]] &&
    ([[ "${DATAPROC_IMAGE_VERSION}" == '1.2' ]] ||
      is_version_at_least "${DATAPROC_IMAGE_VERSION}" '2.2'); then
    riptide_enabled='true'
  else
    local custom_image
    custom_image=$(get_dataproc_property 'dataproc.spark.container.image')
    local riptide_enabled_default='true'
    # Do not enable image streaming for GCR-hosted images by default
    if [[ "${custom_image}" =~ ^(.*://)?gcr.io/.*$ ]]; then
      riptide_enabled_default='false'
    fi
    riptide_enabled=$(get_dataproc_property_or_default \
      'internal.riptide.enabled' "${riptide_enabled_default}")
  fi
  if is_custom_container_image_configured &&
    [[ "${riptide_enabled}" == 'true' ]]; then
    configure_snapshotter

    systemctl daemon-reload
    systemctl enable gcfsd gcfs-snapshotter containerd
    systemctl start gcfsd gcfs-snapshotter containerd
  else
    enable_and_start_service containerd
  fi

  if is_custom_container_image_configured; then
    ctr --namespace=k8s.io images import --digests=true "${PAUSE_IMAGE_FILE}"
  fi

  # We only need to prefetch the container image during initialization if
  # downloader service account is set and a custom container image is specified.
  # Else, the image pull request is to be made during the job process.
  if [[ -n "${DOWNLOADER_SA}" ]] && is_custom_container_image_configured; then
    prefetch_custom_container_using_downloader
  fi
}

main
