#!/bin/bash

set -euo pipefail

source "$(dirname "$0")/../../bdutil_logging.sh"
source "$(dirname "$0")/../../bdutil_properties.sh"
source "$(dirname "$0")/../../bdutil_versions.sh"
source "$(dirname "$0")/../../bdutil_components.sh"
source "$(dirname "$0")/../../bdutil_retry.sh"
source "$(dirname "$0")/../../bdutil_os.sh"

set -x

YARN_DOCKER_IMAGE=$(get_dataproc_property_or_default yarn.docker.image '')
YARN_DOCKER_IMAGE_REGISTRY="${YARN_DOCKER_IMAGE%%/*}"
REGIONAL_ARTIFACTS_REGISTRY="${REGION}-docker.pkg.dev"

# Configure Docker access to GCR / GAR for a specific Linux user.
function configure_gcr_auth_for_user() {
  local -r user="$1"

  # The default GCR registries.
  sudo -u "${user}" gcloud -q auth configure-docker

  # The registry of the user-provided Docker image.
  if [[ -n "${YARN_DOCKER_IMAGE_REGISTRY}" ]]; then
    sudo -u "${user}" \
      gcloud -q auth configure-docker "${YARN_DOCKER_IMAGE_REGISTRY}"
  fi

  # The regional GAR registry. Ignore errors b/c it might not be needed.
  if [[ "${REGIONAL_ARTIFACTS_REGISTRY}" != "${YARN_DOCKER_IMAGE_REGISTRY}" ]];
  then
    sudo -u "${user}" \
      gcloud -q auth configure-docker "${REGIONAL_ARTIFACTS_REGISTRY}" || true
  fi
}

# Configure Docker access to GCR / GAR.
function configure_gcr_auth() {
  # The `gcloud auth configure-docker` command configures Docker on a per-user
  # basis. Therefore we configure the root user, as well as the yarn user which
  # is part of the docker group. If additional users are added to the docker
  # group later, this command will need to be run for them as well.

  configure_gcr_auth_for_user "root"

  # If Docker-on-YARN is enabled, configure Docker for yarn_docker_user
  if [[ "${ENABLE_DOCKER_YARN}" == "true" ]];  then
    configure_gcr_auth_for_user "yarn_docker_user"
  fi
}

# To allow run Docker containers using YARN, create a wrapper for docker binary
function create_docker_wrapper() {
  if is_ubuntu; then
    # No wrapper required for Ubuntu based images.
    echo "Skip creating docker-wrapper for ubuntu 18"
    return
  fi
  # Create a docker wrapper that add the Google SDK path to PATH to avoid docker
  # auth issue when not finding gcloud and docker-credential-gcloud in PATH
  readonly GCLOUD_PATH="$(which gcloud)"
  readonly GCLOUD_LIB_PATH="$("${GCLOUD_PATH}" info --format="value(installation.sdk_root)")"
  mv /usr/bin/docker /usr/bin/docker.real
  cat >"/usr/bin/docker" <<EOF
#!/bin/bash
readonly DOCKER_BINARY="/usr/bin/docker.real"
readonly GCLOUD_LIB_PATH="${GCLOUD_LIB_PATH}"
export PATH="/usr/bin:\${GCLOUD_LIB_PATH}/bin:\${GCLOUD_LIB_PATH}"
exec "\${DOCKER_BINARY}" "\$@"
EOF
  chmod +x /usr/bin/docker
}

function pull_image() {
  if [[ -n "${YARN_DOCKER_IMAGE}" ]]; then
    echo "Pulling image ${YARN_DOCKER_IMAGE}"
    retry_constant_custom 30 5 docker pull "${YARN_DOCKER_IMAGE}"
  fi
}

# TODO: fix this for Ubuntu and Rocky Linux
ENABLE_DOCKER_YARN="$(get_dataproc_property yarn.docker.enable)"
if [[ -z "${ENABLE_DOCKER_YARN}" ]]; then
  # This is kept for backwards compatiblitiy.
  ENABLE_DOCKER_YARN="$(get_dataproc_property docker.yarn.enable)"
fi

if [[ "${ENABLE_DOCKER_YARN}" == "true" ]]; then
  if is_component_selected 'kerberos'; then
    log_and_fail "docker-ce" "Docker on Yarn can not be enabled with Kerberos"
  fi

  if ! is_version_at_least "${DATAPROC_IMAGE_VERSION}" 2.0; then
    log_and_fail "docker-ce" "Docker on Yarn can only be enabled with Dataproc 2.0 or higher version"
  fi

  create_docker_wrapper
fi

# Docker is configured before any service starts using configure_docker.sh
systemctl daemon-reload

if (systemctl -q status docker.socket); then
  systemctl enable docker.socket
  systemctl start docker.socket
fi

systemctl enable docker
systemctl start docker

configure_gcr_auth
pull_image
