#!/bin/bash

set -euo pipefail

source "$(dirname "$0")/../../bdutil_env.sh"
source "$(dirname "$0")/../../bdutil_logging.sh"
source "$(dirname "$0")/../../bdutil_retry.sh"
source "$(dirname "$0")/../../bdutil_gcs.sh"
source "$(dirname "$0")/../../bdutil_metadata.sh"
source "$(dirname "$0")/../../bdutil_properties.sh"
source "$(dirname "$0")/../../bdutil_versions.sh"
source "$(dirname "$0")/../../bdutil_components.sh"
source "$(dirname "$0")/../../bdutil_misc.sh"
source "$(dirname "$0")/../shared/druid.sh"
source "$(dirname "$0")/../shared/proxy.sh"
source "$(dirname "$0")/../shared/mysql.sh"
source "$(dirname "$0")/../shared/zookeeper.sh"


set -x

readonly CLUSTER_PROPERTIES_DIR="/tmp/cluster/properties"
readonly DRUID_COMMON_PROPERTIES_CONFIG_FILE="${DRUID_DIR}/conf/druid/_common/common.runtime.properties"
readonly CLOUD_SQL_INSTANCE=$(get_dataproc_property druid.cloud-sql.instance.connection.name)

readonly CLUSTER_DRUID_BROKER_JVM_PROPERTIES_FILE="/tmp/cluster/properties/druid-broker-jvm.properties"
readonly CLUSTER_DRUID_BROKER_RUNTIME_PROPERTIES_FILE="/tmp/cluster/properties/druid-broker-runtime.properties"
readonly CLUSTER_DRUID_COORDINATOR_JVM_PROPERTIES_FILE="/tmp/cluster/properties/druid-coordinator-jvm.properties"
readonly CLUSTER_DRUID_COORDINATOR_RUNTIME_PROPERTIES_FILE="/tmp/cluster/properties/druid-coordinator-runtime.properties"
readonly CLUSTER_DRUID_HISTORICAL_JVM_PROPERTIES_FILE="/tmp/cluster/properties/druid-historical-jvm.properties"
readonly CLUSTER_DRUID_HISTORICAL_RUNTIME_PROPERTIES_FILE="/tmp/cluster/properties/druid-historical-runtime.properties"
readonly CLUSTER_DRUID_MIDDLEMANAGER_JVM_PROPERTIES_FILE="/tmp/cluster/properties/druid-middleManager-jvm.properties"
readonly CLUSTER_DRUID_MIDDLEMANAGER_RUNTIME_PROPERTIES_FILE="/tmp/cluster/properties/druid-middleManager-runtime.properties"
readonly CLUSTER_DRUID_OVERLORD_JVM_PROPERTIES_FILE="/tmp/cluster/properties/druid-overlord-jvm.properties"
readonly CLUSTER_DRUID_OVERLORD_RUNTIME_PROPERTIES_FILE="/tmp/cluster/properties/druid-overlord-runtime.properties"
readonly CLUSTER_DRUID_ROUTER_JVM_PROPERTIES_FILE="/tmp/cluster/properties/druid-router-jvm.properties"
readonly CLUSTER_DRUID_ROUTER_RUNTIME_PROPERTIES_FILE="/tmp/cluster/properties/druid-router-runtime.properties"

MIDDLE_MANAGER_HADOOP_WORKING_PATH="var/druid/hadoop-tmp"
DRUID_EXTENSIONS=(druid-avro-extensions druid-basic-security druid-stats
  druid-bloom-filter druid-datasketches druid-distinctcount druid-histogram
  druid-kafka-extraction-namespace druid-kafka-indexing-service
  druid-lookups-cached-global druid-lookups-cached-single
  druid-protobuf-extensions mysql-metadata-storage
  druid-orc-extensions druid-parquet-extensions)

function get_db_root_password() {
  set +x
  local kms_key_uri
  kms_key_uri=$(get_dataproc_property druid.kms.key.uri)

  local db_root_password

  # Retrieve root user password of cloud-sql.
  if [[ -n "${CLOUD_SQL_INSTANCE}" ]]; then
    local db_root_password_uri
    db_root_password_uri=$(get_dataproc_property druid.cloud-sql.root.password.uri)
    if [[ -z "${db_root_password_uri}" ]]; then
      log_and_fail "druid" "When using Cloud SQL instance as the druid database, you must specify the property 'dataproc:druid.cloud-sql.root.password.uri'."
    fi
    if [[ -z "${kms_key_uri}" ]]; then
      log_and_fail "druid" "Failed to read KMS key URI from Dataproc properties. Set it through 'dataproc:druid.kms.key.uri'."
    fi

    db_root_password=$(decrypt_file_with_kms_key "${db_root_password_uri}" "${kms_key_uri}")
    if [[ -z "${db_root_password}" ]]; then
      log_and_fail "druid" "Cloud SQL root password cannot be empty."
    fi
  else
    # read from common function from shared/mysql.sh
    db_root_password="$(read_mysql_password)"
  fi
  echo "${db_root_password}"
  set -x
}

function get_db_admin_password() {
  set +x
  local db_admin_password_uri
  db_admin_password_uri=$(get_dataproc_property druid.db.admin.password.uri)

  local kms_key_uri
  kms_key_uri=$(get_dataproc_property druid.kms.key.uri)

  local db_admin_password
  if [[ -z "${db_admin_password_uri}" ]]; then
    db_admin_password="diurd"
  else
    db_admin_password=$(decrypt_file_with_kms_key "${db_admin_password_uri}" "${kms_key_uri}")
    if [[ -z "${db_admin_password}" ]]; then
      log_and_fail "druid" "Database druid admin password cannot be empty!"
    fi
  fi
  echo "${db_admin_password}"
  set -x
}

function prepare_environment() {
  # prepare environment first so that metadata config can use environment
  cd "${DRUID_DIR}"
  for conf_dir in _common broker coordinator historical middleManager overlord router; do
    rm -rf "conf/druid/${conf_dir}"
    mkdir "conf/druid/${conf_dir}"
  done

  ln -s /usr/share/java/mysql.jar extensions/mysql-metadata-storage/

  ln -s /usr/lib/hadoop/etc/hadoop/core-site.xml conf/druid/_common/core-site.xml
  ln -s /usr/lib/hadoop/etc/hadoop/hdfs-site.xml conf/druid/_common/hdfs-site.xml
  ln -s /usr/lib/hadoop/etc/hadoop/yarn-site.xml conf/druid/_common/yarn-site.xml
  ln -s /usr/lib/hadoop/etc/hadoop/mapred-site.xml conf/druid/_common/mapred-site.xml

  mkdir -p var/tmp
  mkdir -p "${DRUID_LOG_DIR}"
  chown root conf/druid/
  chmod +xr -R conf/druid/
  chmod a+x bin/*
}

function configure_metadata_database() {
  local cloud_sql_private_ip
  cloud_sql_private_ip=$(get_dataproc_property druid.cloud-sql.use-private-ip)

  local fqdn
  fqdn="$(hostname -f)"

  local master
  master="$(get_metadata_master)"

  local master_fqdn
  master_fqdn="${master}.$(dnsdomainname)"

  local -r db_admin_user="druid"

  if [[ -n "${CLOUD_SQL_INSTANCE}" ]]; then
    configure_cloud_sql_proxy_for_component \
      druid "${CLOUD_SQL_INSTANCE}" "${DRUID_CLOUD_SQL_PORT}" "${cloud_sql_private_ip}"
  fi

  # Set 127.0.0.1 as mysql-host if cloud-sql-proxy is configured
  # Otherwise for master node set mysql-host to localhost
  # And for worker node set mysql-host to worker nodes.
  if [[ -n "${CLOUD_SQL_INSTANCE}" ]]; then
    local sql_port="${DRUID_CLOUD_SQL_PORT}"
    local sql_host="127.0.0.1"
    local sql_user_host="%"
  else
    if [[ "${master_fqdn}" == "${fqdn}" ]]; then
      local sql_host="localhost"
      local sql_user_host="localhost"
    else
      local sql_host="${master}"
      local sql_user_host="${master}"
    fi
    local sql_port="3306"
  fi

  set +x
  local db_root_password
  db_root_password="$(get_db_root_password)"

  local db_admin_password
  db_admin_password="$(get_db_admin_password)"

  # Only use master node to host metadata database when Cloud SQL is not configured
  # In HA cluster deployment, just configure MySQL on first master node.
  if [[ "${master_fqdn}" == "${fqdn}" ]]; then
    retry_quiet \
      mysqladmin -u root -p"${db_root_password}" --port "${sql_port}" -h "${sql_host}" status
    # Create the database 'druid' if it does not already exist.
    if ! mysql -u root -p"${db_root_password}" --port "${sql_port}" -h "${sql_host}" -e 'use druid'; then
      mysql -u root -p"${db_root_password}" --port "${sql_port}" -h "${sql_host}" \
        -e "CREATE DATABASE druid DEFAULT CHARACTER SET utf8;"
    fi

    # Create the user 'druid' if it does not already exist.
    if ! mysql -u "${db_admin_user}" -p"${db_admin_password}" --port "${sql_port}" -h "${sql_host}" -e ''; then
      mysql -u root -p"${db_root_password}" --port "${sql_port}" -h "${sql_host}" \
        -e "CREATE USER '${db_admin_user}'@'${sql_user_host}' IDENTIFIED BY '${db_admin_password}';"
      mysql -u root -p"${db_root_password}" --port "${sql_port}" -h "${sql_host}" \
        -e "GRANT ALL PRIVILEGES ON druid.* TO 'druid'@'${sql_user_host}';"
    fi
  fi

  # configure metadata runtime properties
  cat >"${DRUID_COMMON_PROPERTIES_CONFIG_FILE}" <<EOF
# Metadata storage
#
# For MySQL:
druid.metadata.storage.type=mysql
druid.metadata.storage.connector.connectURI=jdbc:mysql://${sql_host}:${sql_port}/druid
druid.metadata.storage.connector.user=${db_admin_user}
druid.metadata.storage.connector.password=${db_admin_password}
druid.sql.enable=true
EOF
  set -x
}

function kerberize() {
  # Define keytab path based on role
  local -r druid_keytab="/etc/security/keytab/druid.service.keytab"
  local realm
  realm=$(grep "default_realm = " /etc/krb5.conf | awk '{print $NF}')
  local principal
  principal="druid/$(hostname -f)@${realm}"

  # Druid Kerberos properties
  cat >>"${DRUID_COMMON_PROPERTIES_CONFIG_FILE}" <<EOF
# Druid Kerberos properties
druid.hadoop.security.kerberos.principal=${principal}
druid.hadoop.security.kerberos.keytab=${druid_keytab}
EOF
}

function add_deep_storage_configs() {
  # Check if GCS deep storage path is provided, if provided configure GCS as
  # deep storage otherwise configure HDFS as deep storage
  local gcs_deep_storage_path
  gcs_deep_storage_path=$(get_dataproc_property druid.deep.storage.gcs.path)
  local gcs_indexer_logs_path
  gcs_indexer_logs_path=$(get_dataproc_property druid.indexer.logs.gcs.path)

  if [[ -z "${gcs_deep_storage_path}" ]] && [[ -z "${gcs_indexer_logs_path}" ]]; then
    DRUID_EXTENSIONS+=(druid-hdfs-storage)
    cat >>"${DRUID_COMMON_PROPERTIES_CONFIG_FILE}" <<EOF
#
# Deep storage
#
# For HDFS (make sure to include the HDFS extension and that your Hadoop config files in the cp):
druid.storage.type=hdfs
druid.storage.storageDirectory=/druid/segments
#
# Indexing service logs
#
# For HDFS (make sure to include the HDFS extension and that your Hadoop config files in the cp):
druid.indexer.logs.type=hdfs
druid.indexer.logs.directory=/druid/indexing-logs
EOF
  else
    # HDFS and GCS extensions does not work together, use extension
    # druid-google-extensions when GCS is used as deep storage
    DRUID_EXTENSIONS+=(druid-google-extensions)
    MIDDLE_MANAGER_HADOOP_WORKING_PATH="/hadoop/tmp"

    # check if indexing logs path is also provided, if not, then fail as
    # indexing logs should also be written to GCS as HDFS and GCS extensions
    # does not work together.
    if [[ -z "${gcs_deep_storage_path}" ]]; then
      log_and_fail "druid" "When using GCS as the druid deep storage, you must specify the property 'dataproc:druid.deep.storage.gcs.path'."
    fi

    if [[ -z "${gcs_indexer_logs_path}" ]]; then
      log_and_fail "druid" "When using GCS as the druid deep storage, you must specify the property 'dataproc:druid.indexer.logs.gcs.path'."
    fi

    # Add GCS connector to druid-google-extensions
    ln -s "${DATAPROC_COMMON_LIB_DIR}/gcs-connector.jar" \
      "${DRUID_DIR}/extensions/druid-google-extensions/"

    # gcs deep storage configs
    # strip off gs
    local -r deep_storage_dir=${gcs_deep_storage_path/gs:\/\//}
    local deep_storage_bucket
    deep_storage_bucket=$(echo "${deep_storage_dir}" | cut -d '/' -f 1)
    # Verify if we can access the deep storage bucket
    check_gcs_bucket_access "gs://${deep_storage_bucket}/" 'druid'
    local -r deep_storage_path=${deep_storage_dir#"${deep_storage_bucket}/"}

    # gcs indexer logs configs
    # strip off gs
    local -r indexer_logs_dir=${gcs_indexer_logs_path/gs:\/\//}
    local indexer_logs_bucket
    indexer_logs_bucket=$(echo "${indexer_logs_dir}" | cut -d '/' -f 1)
    # Verify if we can access the deep storage bucket
    check_gcs_bucket_access "gs://${indexer_logs_bucket}/" 'druid'
    local -r indexer_logs_path=${indexer_logs_dir#"${indexer_logs_bucket}/"}

    cat >>"${DRUID_COMMON_PROPERTIES_CONFIG_FILE}" <<EOF
#
# Deep storage
#
# For HDFS (make sure to include the HDFS extension and that your Hadoop config files in the cp):
druid.storage.type=google
druid.google.bucket=${deep_storage_bucket}
druid.google.prefix=${deep_storage_path}
#
# Indexing service logs
#
# For HDFS (make sure to include the HDFS extension and that your Hadoop config files in the cp):
druid.indexer.logs.type=google
druid.indexer.logs.bucket=${indexer_logs_bucket}
druid.indexer.logs.prefix=${indexer_logs_path}
EOF
  fi

  cat >>"${DRUID_COMMON_PROPERTIES_CONFIG_FILE}" <<EOF
#
# Extensions
#
# This is not the full list of Druid extensions, but common ones that people often use. You may need to change this list
# based on your particular setup.
druid.extensions.loadList=[$(printf ", \"%s\"" "${DRUID_EXTENSIONS[@]}" | sed 's/^, //')]
#
EOF
}

function configure_druid() {
  local zookeeper_list
  zookeeper_list=$(get_zookeeper_list)

  local -r hadoop_conf_dir="/etc/hadoop/conf"
  cd "${DRUID_DIR}"

  cat >>"${DRUID_COMMON_PROPERTIES_CONFIG_FILE}" <<EOF
# Logging
#
# Log all runtime properties on startup. Disable to avoid logging properties on startup:
druid.startup.logging.logProperties=true
#
# Zookeeper
#
druid.zk.service.host=${zookeeper_list}
druid.zk.paths.base=/druid
#
# Service discovery
#
druid.selectors.indexing.serviceName=druid/overlord
druid.selectors.coordinator.serviceName=druid/coordinator
#
# Monitoring
#
druid.monitoring.monitors=["io.druid.java.util.metrics.JvmMonitor"]
druid.emitter=logging
druid.emitter.logging.logLevel=info
# Storage type of double columns
# omitting this will lead to index double as float at the storage layer
druid.indexing.doubleStorage=double
EOF

  add_deep_storage_configs

  if is_component_selected "kerberos"; then
    kerberize
  fi

  cat >conf/druid/_common/log4j2.xml <<'EOF'
<?xml version="1.0" encoding="UTF-8" ?>

<Configuration status="WARN">
    <Appenders>
        <Console name="Console" target="SYSTEM_OUT">
            <PatternLayout pattern="%d{ISO8601} %5p [%t] %c{2}: %m%n"/>
        </Console>
    </Appenders>
    <Loggers>
        <Root level="info">
            <AppenderRef ref="Console"/>
        </Root>
    </Loggers>
</Configuration>
EOF

  # JVM configs for services running on master node
  cat >conf/druid/coordinator/jvm.config <<'EOF'
-server
-Xms2g
-Xmx4g
-Duser.timezone=UTC
-Dfile.encoding=UTF-8
-Djava.io.tmpdir=/tmp
-Djava.util.logging.manager=org.apache.logging.log4j.jul.LogManager
-Dderby.stream.error.file=var/druid/derby.log
EOF

  # Merge server supplied jvm properties
  merge_user_specified_jvm_config conf/druid/coordinator/jvm.config \
    "${CLUSTER_DRUID_COORDINATOR_JVM_PROPERTIES_FILE}"

  cat >conf/druid/overlord/jvm.config <<'EOF'
-server
-Xms2g
-Xmx2g
-XX:MaxDirectMemorySize=1792m
-Duser.timezone=UTC
-Dfile.encoding=UTF-8
-Djava.io.tmpdir=/tmp
-Djava.util.logging.manager=org.apache.logging.log4j.jul.LogManager
EOF

  # Merge server supplied jvm properties
  merge_user_specified_jvm_config conf/druid/overlord/jvm.config \
    "${CLUSTER_DRUID_OVERLORD_JVM_PROPERTIES_FILE}"

  cat >conf/druid/broker/jvm.config <<'EOF'
-server
-Xms2g
-Xmx8g
-XX:MaxDirectMemorySize=8192m
-Duser.timezone=UTC
-Dfile.encoding=UTF-8
-Djava.io.tmpdir=/tmp
-Djava.util.logging.manager=org.apache.logging.log4j.jul.LogManager
EOF

  # Merge server supplied jvm properties
  merge_user_specified_jvm_config conf/druid/broker/jvm.config \
    "${CLUSTER_DRUID_BROKER_JVM_PROPERTIES_FILE}"

  cat >conf/druid/router/jvm.config <<'EOF'
-server
-Xmx256m
-Xms256m
-XX:NewSize=256m
-XX:MaxNewSize=256m
-XX:+UseConcMarkSweepGC
-XX:+PrintGCDetails
-XX:+PrintGCTimeStamps
-XX:+UseLargePages
-XX:+HeapDumpOnOutOfMemoryError
-XX:HeapDumpPath=/mnt/galaxy/deploy/current/
-Duser.timezone=UTC
-Dfile.encoding=UTF-8
-Djava.io.tmpdir=/tmp
-Djava.util.logging.manager=org.apache.logging.log4j.jul.LogManager
-Dcom.sun.management.jmxremote.port=17071
-Dcom.sun.management.jmxremote.rmi.port=17071
-Dcom.sun.management.jmxremote.authenticate=false
-Dcom.sun.management.jmxremote.ssl=false
EOF

  # Merge server supplied jvm properties
  merge_user_specified_jvm_config conf/druid/router/jvm.config \
    "${CLUSTER_DRUID_ROUTER_JVM_PROPERTIES_FILE}"

  cat >conf/druid/historical/jvm.config <<'EOF'
-server
-Xms2g
-XX:MaxDirectMemorySize=4096m
-Duser.timezone=UTC
-Dfile.encoding=UTF-8
-Djava.io.tmpdir=/tmp
-Djava.util.logging.manager=org.apache.logging.log4j.jul.LogManager
EOF

  # Merge server supplied jvm properties
  merge_user_specified_jvm_config conf/druid/historical/jvm.config \
    "${CLUSTER_DRUID_HISTORICAL_JVM_PROPERTIES_FILE}"

  cat >conf/druid/middleManager/jvm.config <<'EOF'
-server
-Xms256m
-Xmx256m
-XX:MaxDirectMemorySize=4096m
-Duser.timezone=UTC
-Dfile.encoding=UTF-8
-Djava.io.tmpdir=/tmp
-Djava.util.logging.manager=org.apache.logging.log4j.jul.LogManager
EOF

  # Merge server supplied jvm properties
  merge_user_specified_jvm_config conf/druid/middleManager/jvm.config \
    "${CLUSTER_DRUID_MIDDLEMANAGER_JVM_PROPERTIES_FILE}"

  cat >conf/druid/broker/runtime.properties <<'EOF'
druid.service=druid/broker
druid.plaintextPort=8082
# HTTP server threads
druid.broker.http.numConnections=5
druid.server.http.numThreads=25
# Query cache
druid.broker.cache.useCache=true
druid.broker.cache.populateCache=true
druid.cache.type=local
druid.cache.sizeInBytes=2000000000
# Processing threads and buffers
druid.processing.buffer.sizeBytes=268435456
druid.processing.numThreads=7
EOF

  # Merge server supplied runtime properties
  cat "${CLUSTER_DRUID_BROKER_RUNTIME_PROPERTIES_FILE}" >>"conf/druid/broker/runtime.properties"

  cat >conf/druid/overlord/runtime.properties <<'EOF'
druid.service=druid/overlord
druid.plaintextPort=8092
druid.indexer.queue.startDelay=PT30S
druid.processing.buffer.sizeBytes=268435456
druid.indexer.runner.type=remote
druid.indexer.storage.type=metadata
EOF

  # Merge server supplied runtime properties
  cat "${CLUSTER_DRUID_OVERLORD_RUNTIME_PROPERTIES_FILE}" >>"conf/druid/overlord/runtime.properties"

  cat >conf/druid/middleManager/runtime.properties <<EOF
druid.service=druid/middleManager
druid.plaintextPort=8091
# Number of tasks per middleManager
druid.worker.capacity=5
# Task launch parameters
druid.indexer.runner.javaOpts=-server -Xmx2g -Duser.timezone=UTC -Dfile.encoding=UTF-8 -Djava.util.logging.manager=org.apache.logging.log4j.jul.LogManager
druid.indexer.task.baseTaskDir=var/druid/task
# HTTP server threads
druid.server.http.numThreads=25
# Processing threads and buffers on Peons
druid.indexer.fork.property.druid.processing.buffer.sizeBytes=268435456
druid.indexer.fork.property.druid.processing.numThreads=2
druid.indexer.fork.property.druid.processing.numMergeBuffers=2
# Hadoop indexing
druid.indexer.task.hadoopWorkingPath=${MIDDLE_MANAGER_HADOOP_WORKING_PATH}
EOF

  # Merge server supplied runtime properties
  cat "${CLUSTER_DRUID_MIDDLEMANAGER_RUNTIME_PROPERTIES_FILE}" >>"conf/druid/middleManager/runtime.properties"

  cat >conf/druid/historical/runtime.properties <<'EOF'
druid.service=druid/historical
druid.plaintextPort=8083
# HTTP server threads
druid.server.http.numThreads=25
# Segment storage
druid.segmentCache.locations=[{"path":"var/druid/segment-cache","maxSize":130000000000}]
druid.server.maxSize=130000000000
# Processing threads and buffers
druid.processing.buffer.sizeBytes=268435456
druid.processing.numThreads=7
EOF

  # Merge server supplied runtime properties
  cat "${CLUSTER_DRUID_HISTORICAL_RUNTIME_PROPERTIES_FILE}" >>"conf/druid/historical/runtime.properties"

  cat >conf/druid/coordinator/runtime.properties <<'EOF'
druid.service=druid/coordinator
druid.plaintextPort=8081
druid.coordinator.startDelay=PT30S
druid.coordinator.period=PT30S
druid.processing.buffer.sizeBytes=268435456
EOF

  # Merge server supplied runtime properties
  cat "${CLUSTER_DRUID_COORDINATOR_RUNTIME_PROPERTIES_FILE}" >>"conf/druid/coordinator/runtime.properties"

  cat >conf/druid/router/runtime.properties <<'EOF'
druid.plaintextPort=8095
druid.service=druid/router
druid.router.defaultBrokerServiceName=druid/broker
druid.router.coordinatorServiceName=druid/coordinator
druid.router.http.numConnections=50
druid.router.http.readTimeout=PT5M
# Number of threads used by the Router proxy http client
druid.router.http.numMaxThreads=100
druid.server.http.numThreads=100
druid.router.managementProxy.enabled=true
EOF

  # Merge server supplied runtime properties
  cat "${CLUSTER_DRUID_ROUTER_RUNTIME_PROPERTIES_FILE}" >>"conf/druid/router/runtime.properties"

  merge_user_specified_properties

  local -r common_classpath="${DRUID_DIR}/conf/druid/_common:${DRUID_DIR}/lib/*"

  cat >/etc/systemd/system/druid-coordinator.service <<EOF
[Unit]
Description=Druid Coordinator
Wants=network-online.target
After=network-online.target
[Service]
WorkingDirectory=${DRUID_DIR}
Environment="JAVA_HOME=${JAVA_HOME}"
Environment="HADOOP_CONF_DIR=${hadoop_conf_dir}"
User=root
Group=root
Type=simple
ExecStart=/bin/bash -c "java $(xargs -a conf/druid/coordinator/jvm.config) -cp "${common_classpath}:${DRUID_DIR}/conf/druid/coordinator" org.apache.druid.cli.Main server coordinator > ${DRUID_LOG_DIR}/druid-coordinator.log"
Restart=on-failure
[Install]
WantedBy=multi-user.target
EOF

  cat >/etc/systemd/system/druid-overlord.service <<EOF
[Unit]
Description=Druid Overlord
Wants=network-online.target
After=network-online.target
[Service]
WorkingDirectory=${DRUID_DIR}
Environment="JAVA_HOME=${JAVA_HOME}"
Environment="HADOOP_CONF_DIR=${hadoop_conf_dir}"
User=root
Group=root
Type=simple
ExecStart=/bin/bash -c "java $(xargs -a conf/druid/overlord/jvm.config) -cp "${common_classpath}:${DRUID_DIR}/conf/druid/overlord" org.apache.druid.cli.Main server overlord > ${DRUID_LOG_DIR}/druid-overlord.log"
Restart=on-failure
[Install]
WantedBy=multi-user.target
EOF

  cat >/etc/systemd/system/druid-broker.service <<EOF
[Unit]
Description=Druid Broker
Wants=network-online.target
After=network-online.target
[Service]
WorkingDirectory=${DRUID_DIR}
Environment="JAVA_HOME=${JAVA_HOME}"
Environment="HADOOP_CONF_DIR=${hadoop_conf_dir}"
User=root
Group=root
Type=simple
ExecStart=/bin/bash -c "java $(xargs -a conf/druid/broker/jvm.config) -cp "${common_classpath}:${DRUID_DIR}/conf/druid/broker" org.apache.druid.cli.Main server broker > ${DRUID_LOG_DIR}/druid-broker.log"
Restart=on-failure
[Install]
WantedBy=multi-user.target
EOF

  cat >/etc/systemd/system/druid-historical.service <<EOF
[Unit]
Description=Druid Historical
Wants=network-online.target
After=network-online.target
[Service]
WorkingDirectory=${DRUID_DIR}
Environment="JAVA_HOME=${JAVA_HOME}"
Environment="HADOOP_CONF_DIR=${hadoop_conf_dir}"
User=root
Group=root
Type=simple
ExecStart=/bin/bash -c "java $(xargs -a conf/druid/historical/jvm.config) -cp "${common_classpath}:${DRUID_DIR}/conf/druid/historical" org.apache.druid.cli.Main server historical > ${DRUID_LOG_DIR}/druid-historical.log"
Restart=on-failure
[Install]
WantedBy=multi-user.target
EOF

  cat >/etc/systemd/system/druid-middle-manager.service <<EOF
[Unit]
Description=Druid MiddleManager
Wants=network-online.target
After=network-online.target
[Service]
WorkingDirectory=${DRUID_DIR}
Environment="JAVA_HOME=${JAVA_HOME}"
Environment="HADOOP_CONF_DIR=${hadoop_conf_dir}"
User=root
Group=root
Type=simple
ExecStart=/bin/bash -c "java $(xargs -a conf/druid/middleManager/jvm.config) -cp "${common_classpath}:${DRUID_DIR}/conf/druid/middleManager" org.apache.druid.cli.Main server middleManager > ${DRUID_LOG_DIR}/druid-middlemanager.log"
Restart=on-failure
[Install]
WantedBy=multi-user.target
EOF

  cat >/etc/systemd/system/druid-router.service <<EOF
[Unit]
Description=Druid Router
Wants=network-online.target
After=network-online.target
[Service]
WorkingDirectory=${DRUID_DIR}
Environment="JAVA_HOME=${JAVA_HOME}"
Environment="HADOOP_CONF_DIR=${hadoop_conf_dir}"
User=root
Group=root
Type=simple
ExecStart=/bin/bash -c "java $(xargs -a conf/druid/router/jvm.config) -cp "${common_classpath}:${DRUID_DIR}/conf/druid/router" org.apache.druid.cli.Main server router > ${DRUID_LOG_DIR}/druid-router.log"
Restart=on-failure
[Install]
WantedBy=multi-user.target
EOF
}

function merge_user_specified_jvm_config() {
  local -r jvm_config_file="${1}"
  local -r user_jvm_properties="${2}"
  while IFS="" read -r property || [[ -n "$property" ]]
  do
    if [[ ! "${property}" == "#"* && "${property}" == *"="* ]]; then
      key=$(echo "${property}" | cut -d\= -f1)
      value=$(echo "${property}" | cut -d\= -f2)
      if grep -q "${key}" "${jvm_config_file}"; then
        sed -i -e "s/\(${key}=*\).*/\1${value}/g" "${jvm_config_file}"
      else
        if [[ "${key}" == "Xms" || "${key}" == "Xmx" ]]; then
          echo "-${key}${value}" >> "${jvm_config_file}"
        else
          echo "-XX:${key}=${value}" >> "${jvm_config_file}"
        fi
      fi
    fi
  done < "${user_jvm_properties}"
}

function merge_user_specified_properties() {
  for service in "broker coordinator historical middleManager overlord"; do
    local propFile="${CLUSTER_PROPERTIES_DIR}/druid-${service}.properties"
    merge_java_properties "${propFile}" \
      "${DRUID_DIR}/conf/druid/${service}/runtime.properties"
  done
}

function start_services() {
  local -r master_services=(
    druid-broker druid-coordinator druid-overlord druid-router)
  local -r worker_services=(druid-historical druid-middle-manager)

  local worker_count
  worker_count=$(get_metadata_worker_count)
  local role
  role=$(get_metadata_role)

  local -a services_to_activate
  # Install all druid services on single configuration
  if [[ "${worker_count}" == '0' ]]; then
    services_to_activate=("${master_services[@]}" "${worker_services[@]}")
  else
    # Isolate management services from middle-manager and historical
    if [[ ${role} == 'Master' ]]; then
      services_to_activate=("${master_services[@]}")
    else
      services_to_activate=("${worker_services[@]}")
    fi
  fi

  systemctl daemon-reload
  systemctl enable "${services_to_activate[@]}"
  systemctl start "${services_to_activate[@]}"
}

function main() {
  prepare_environment
  configure_metadata_database
  configure_druid
  start_services || log_and_fail "druid" "Starting services failed"
}

main
