#!/bin/bash

set -euo pipefail

source "$(dirname "$0")/../shared/flink.sh"
source "$(dirname "$0")/../../bdutil_logging.sh"
source "$(dirname "$0")/../../bdutil_retry.sh"
source "$(dirname "$0")/../../bdutil_networking.sh"
source "$(dirname "$0")/../../bdutil_metadata.sh"
source "$(dirname "$0")/../../bdutil_properties.sh"
source "$(dirname "$0")/../../bdutil_components.sh"
source "$(dirname "$0")/../../bdutil_misc.sh"
source "$(dirname "$0")/../shared/zookeeper.sh"

set -x

readonly YARN_MASTER_HOSTNAME="$(get_metadata_master)"
readonly MASTER_ADDITIONAL="$(/usr/share/google/get_metadata_value attributes/dataproc-master-additional)"
readonly WORKER_COUNT=$(get_metadata_worker_count)
readonly CLUSTER_PROPS="/tmp/cluster/properties"
readonly FLINK_CONF=${FLINK_INSTALL_DIR}/conf

function set_property_in_flink_conf() {
  local -r name=$1
  local -r value=$2
  set_property_in_yaml "${FLINK_CONF}/flink-conf.yaml" "${name}" "${value}"
}

function start_flink_history_server() {

  local -r service_name="flink-history-server"
  local -r exec_start="/usr/lib/flink/bin/historyserver.sh start"
  local -r exec_stop="/usr/lib/flink/bin/historyserver.sh stop"

  cat >"/etc/systemd/system/${service_name}.service" <<EOF
[Unit]
Description=${service_name} ${ROLE}
Wants=network-online.target
After=network-online.target

[Service]
User=root
Group=root
Type=forking
ExecStart=${exec_start}
ExecStop=${exec_stop}
Restart=on-failure
OOMScoreAdjust=0

[Install]
WantedBy=multi-user.target
EOF

  mkdir -p /var/run/${service_name}
  chown root:root /var/run/${service_name}
  chmod 775 /var/run/${service_name}

  retry_constant_short systemctl daemon-reload
  retry_constant_short systemctl enable "${service_name}"
  retry_constant_short systemctl start "${service_name}"

}

# TODO: Flink sets number of Job Managers to worker -1. So --single-node should
# not be used.
function configure_flink() {
  # Number of Flink TaskManagers to use. Reserve 1 node for the JobManager.
  # NB: This assumes > 1 worker node.
  local -r num_taskmanagers="$((WORKER_COUNT - 1))"

  # The `-n` flag has been removed since Flink 1.7.
  if is_version_at_least "${DATAPROC_IMAGE_VERSION}" '2.0'; then
    local -r num_taskmanagers_flag=''
  else
    local -r num_taskmanagers_flag="-n ${num_taskmanagers}"
  fi

  # merge user supplied log4j properties
  merge_java_properties \
  "${CLUSTER_PROPS}/flink-log4j.properties" "${FLINK_CONF}/log4j.properties"

  # Determine the number of task slots per worker.
  local flink_taskmanager_slots
  flink_taskmanager_slots=$(get_property_in_xml /etc/hadoop/conf/yarn-site.xml \
    yarn.nodemanager.resource.cpu-vcores)

  # Determine the default parallelism.
  local flink_parallelism
  flink_parallelism=$(python -c \
    "print(${num_taskmanagers} * ${flink_taskmanager_slots})")

  # Set Flink job manager and task manager memory
  local flink_jobmanager_memory
  flink_jobmanager_memory=$(calculate_flink_jm_tm_memory "FLINK_JOBMANAGER" \
    "${FLINK_JOBMANAGER_MIN_MEMORY_MB}" "${FLINK_JOBMANAGER_MAX_MEMORY_MB}")
  local flink_taskmanager_memory
  flink_taskmanager_memory=$(calculate_flink_jm_tm_memory "FLINK_TASKMANAGER" \
    "${FLINK_TASKMANAGER_MIN_MEMORY_MB}" "${FLINK_TASKMANAGER_MAX_MEMORY_MB}")

  # Set the 3rd party class loader check to False
  local classloader_check_leaked_flag
  classloader_check_leaked_flag=$(python -c \
    "print(str(False))")

  # Apply Flink settings
  set_property_in_flink_conf "env.hadoop.conf.dir" "${HADOOP_CONF_DIR}"
  set_property_in_flink_conf "fs.hdfs.hadoopconf" "${HADOOP_CONF_DIR}"
  set_property_in_flink_conf "jobmanager.rpc.address" "${YARN_MASTER_HOSTNAME}"
  set_property_in_flink_conf "jobmanager.memory.process.size" "${flink_jobmanager_memory}m"
  set_property_in_flink_conf "taskmanager.memory.process.size" "${flink_taskmanager_memory}m"
  set_property_in_flink_conf "taskmanager.numberOfTaskSlots" "${flink_taskmanager_slots}"
  set_property_in_flink_conf "parallelism.default" "${flink_parallelism}"
  set_property_in_flink_conf "historyserver.web.port" "8500"
  set_property_in_flink_conf "env.log.dir" "/var/log/flink"

  # Set tuning parameters
  set_property_in_flink_conf "state.storage.fs.write-buffer-size" "1048576"
  set_property_in_flink_conf "fs.gs.outputstream.buffer.size" "1048576"
  set_property_in_flink_conf "fs.gs.outputstream.pipe.buffer.size" "12582912"
  set_property_in_flink_conf "fs.gs.outputstream.upload.chunk.size" "16777216"

  if [[ -n "${MASTER_ADDITIONAL}" ]]; then
    local zookeeper_nodes
    zookeeper_nodes=$(get_zookeeper_list)
    # HA settings applied by Dataproc
    set_property_in_flink_conf "high-availability" "zookeeper"
    set_property_in_flink_conf "high-availability.zookeeper.quorum" "${zookeeper_nodes}"
    set_property_in_flink_conf "high-availability.storageDir" "hdfs:///flink/recovery"
    set_property_in_flink_conf "high-availability.zookeeper.path.root" "/flink"
    set_property_in_flink_conf "yarn.application-attempts" "10"
  fi

  if is_version_at_least "${DATAPROC_IMAGE_VERSION}" '2.1'; then
    set_property_in_flink_conf "classloader.check-leaked-classloader" "${classloader_check_leaked_flag}"
  fi

  # Configure Hadoop classpath for Flink 1.11+:
  # https://ci.apache.org/projects/flink/flink-docs-release-1.11/ops/deployment/hadoop.html
  if is_version_at_least "${DATAPROC_IMAGE_VERSION}" '2.0'; then
    patch --merge --ignore-whitespace "${FLINK_INSTALL_DIR}/bin/config.sh" <<'EOF'
@@ -336,6 +336,7 @@ if [ -z "$HADOOP_CONF_DIR" ]; then
     fi
 fi

+HADOOP_CLASSPATH="$(hadoop classpath)"
 # if neither HADOOP_CONF_DIR nor HADOOP_CLASSPATH are set, use some common default (if available)
 if [ -z "$HADOOP_CONF_DIR" ] && [ -z "$HADOOP_CLASSPATH" ]; then
     if [ -d "/etc/hadoop/conf" ]; then
EOF
  fi

  # Configure Flink fs-gs jar for Flink 1.15.0+:
  # https://nightlies.apache.org/flink/flink-docs-release-1.15/docs/deployment/filesystems/gcs/
  if is_version_at_least "${DATAPROC_IMAGE_VERSION}" '2.1'; then
    mkdir -p "${FLINK_INSTALL_DIR}/plugins/gs-fs-hadoop"
    ln -s "${FLINK_INSTALL_DIR}"/opt/flink-gs-fs-hadoop-*.jar "${FLINK_INSTALL_DIR}/plugins/gs-fs-hadoop/flink-gs-fs-hadoop.jar"
    ln -s "/usr/lib/hadoop/client/hadoop-client-runtime.jar" "${FLINK_INSTALL_DIR}/plugins/gs-fs-hadoop/hadoop-client-runtime.jar"
    ln -s "/usr/lib/hadoop/client/hadoop-client-api.jar" "${FLINK_INSTALL_DIR}/plugins/gs-fs-hadoop/hadoop-client-api.jar"
    ln -s "/usr/local/share/google/dataproc/lib/gcs-connector.jar" "${FLINK_INSTALL_DIR}/plugins/gs-fs-hadoop/gcs-connector.jar"
  fi

  cat >"${FLINK_YARN_SCRIPT}" <<EOF
#!/bin/bash
set -euxo pipefail

${FLINK_INSTALL_DIR}/bin/yarn-session.sh \
  ${num_taskmanagers_flag} \
  -nm flink-dataproc \
  --detached
EOF

  # Split into 2 HEREDOC because 'EOF' does not substitute variables and EOF does substitution
  cat >>"${FLINK_YARN_SCRIPT}" <<'EOF'

# grab final field from the first yarn application that contains 'flink'
FLINK_MASTER_URL=$(yarn application -list |
    grep -i 'flink' |
    head -n1 |
    awk -F $'\t' '{print $9}' |
    cut -c8-)
export FLINK_MASTER_URL
EOF

  chmod +x "${FLINK_YARN_SCRIPT}"
}

function is_yarn_core_available() {
  curl -s http://localhost:8088/ws/v1/cluster/metrics |
    python3 -c "import sys, json; vcores=json.load(sys.stdin)['clusterMetrics']['availableVirtualCores']; print(0) if (vcores > 0) else print(-1)"
}

function kerberize() {
  # Flink Kerberos configuration
  set_property_in_flink_conf "security.kerberos.login.use-ticket-cache" "true"
  set_property_in_flink_conf "security.kerberos.login.contexts" "Client"

  if [[ -n "${MASTER_ADDITIONAL}" ]]; then
    set_property_in_flink_conf "zookeeper.sasl.service-name" "zookeeper"
    set_property_in_flink_conf "zookeeper.sasl.login-context-name" "Client"
  fi
}

function start_flink_master() {
  local start_yarn_session
  start_yarn_session="$(/usr/share/google/get_metadata_value \
    "attributes/${START_FLINK_YARN_SESSION_METADATA_KEY}" ||
    echo "${START_FLINK_YARN_SESSION_DEFAULT}")"

  # Start Flink master only on the master node ("0"-master in HA mode)
  if [[ "${start_yarn_session}" == "true" ]]; then
    # Make sure YARN RM started and has enough cores available
    wait_for_port "hadoop-yarn-resourcemanager" localhost 8088
    retry_constant is_yarn_core_available
    retry_constant "${FLINK_YARN_SCRIPT}"
  else
    echo "Doing nothing"
  fi
}

configure_flink || err "Flink configuration failed"
if is_component_selected 'kerberos'; then
  kerberize
fi


function configure_flink_history_server_dir() {
  local persist_history_to_gcs
  persist_history_to_gcs=$(get_dataproc_property_or_default \
    job.history.to-gcs.enabled false)

  if [[ "${persist_history_to_gcs}" == "true" ]]; then
    loginfo "Enabling persisting Flink job history files to GCS."
    if [[ -n "${TEMP_BUCKET}" ]]; then
      local gcs_history_dir_path="${CLUSTER_UUID}/completed-jobs"
      archieve_dir="gs://${TEMP_BUCKET}/${gcs_history_dir_path}"
      set_property_in_flink_conf "jobmanager.archive.fs.dir" "${archieve_dir}"
      set_property_in_flink_conf "historyserver.archive.fs.dir" "${archieve_dir}"
    else
      loginfo "[WARNING] Cannot find VALUE for METADATA 'dataproc-temp-bucket'. Will not persist Flink job history files to the temp bucket."
    fi
  fi
}

if [[ "${ROLE}" == 'Master' ]]; then
  configure_flink_history_server_dir
fi

# Merging the cluster properties at the end to be able to override any auto
# calculated configurations by user provided flink cluster properties
merge_yaml_properties "${CLUSTER_PROPERTIES_DIR}/flink.yaml" \
  "${FLINK_CONF}/flink-conf.yaml"

if [[ "${HOSTNAME}" == "${YARN_MASTER_HOSTNAME}" ]]; then
  start_flink_master || err "Unable to start Flink master"
fi

if [[ "${ROLE}" == 'Master' ]]; then
  start_flink_history_server
fi