#!/bin/bash

set -euo pipefail

source "$(dirname "$0")/../../bdutil_env.sh"
source "$(dirname "$0")/../../bdutil_logging.sh"
source "$(dirname "$0")/../../bdutil_retry.sh"
source "$(dirname "$0")/../../bdutil_networking.sh"
source "$(dirname "$0")/../../bdutil_metadata.sh"
source "$(dirname "$0")/../../bdutil_properties.sh"
source "$(dirname "$0")/../../bdutil_components.sh"
source "$(dirname "$0")/../../bdutil_os.sh"
source "$(dirname "$0")/../../bdutil_misc.sh"
source "$(dirname "$0")/../shared/hbase.sh"
source "$(dirname "$0")/../shared/zookeeper.sh"


set -x

readonly CLUSTER_NAME="$(get_metadata_cluster_name)"
readonly INSTANCE_NAME_PREFIX="$(get_instance_name_prefix)"
readonly CLUSTER_PROPERTIES_DIR="/tmp/cluster/properties"
readonly DOMAIN=$(dnsdomainname)
readonly FQDN=$(hostname -f)
readonly DATAPROC_MASTER=$(get_metadata_master)
readonly HBASE_ETC_DIR="/etc/hbase"
readonly MASTER_ADDITIONAL=$(get_metadata_master_additional)
readonly REALM=$(grep "default_realm = " /etc/krb5.conf | awk '{print $NF}')
readonly ROLE=$(get_metadata_role)
readonly WORKER_COUNT="$(get_metadata_worker_count)"
readonly ZOOKEEPER_NODES="$(get_zookeeper_list)"
readonly ZOOKEEPER_NODES_QUORUM="$(get_zookeeper_server_list_for_zk_quorum)"
readonly ZOOKEEPER_PORT_QUORUM="$(get_zookeeper_port_for_zk_quorum)"


is_component_selected 'kerberos' && readonly ENABLE_KERBEROS=true ||
  readonly ENABLE_KERBEROS=false

function set_hbase_property() {
  property_name=$1
  property_value=$2
  bdconfig set_property \
    --configuration_file "${CLUSTER_PROPERTIES_DIR}/hbase.xml" \
    --name "${property_name}" --value "${property_value}" \
    --clobber
}

function configure_hbase_with_kerberos() {
  # Kerberos authentication
  set_hbase_property "hbase.security.authentication" "kerberos"

  # Security authorization
  set_hbase_property "hbase.security.authorization" "true"

  # Kerberos master principal
  set_hbase_property "hbase.master.kerberos.principal" "hbase/_HOST@${REALM}"

  # Kerberos region server principal
  set_hbase_property "hbase.regionserver.kerberos.principal" "hbase/_HOST@${REALM}"

  # Kerberos master server keytab file path
  set_hbase_property "hbase.master.keytab.file" "/etc/security/keytab/hbase.service.keytab"

  # Kerberos region server keytab file path
  set_hbase_property "hbase.regionserver.keytab.file" "/etc/security/keytab/hbase.service.keytab"

  # Zookeeper authentication provider
  set_hbase_property "hbase.zookeeper.property.authProvider.1" "org.apache.zookeeper.server.auth.SASLAuthenticationProvider"

  # HBase coprocessor region classes
  set_hbase_property "hbase.coprocessor.region.classes" "org.apache.hadoop.hbase.security.token.TokenProvider"

  # Zookeeper remove host from principal
  set_hbase_property "hbase.zookeeper.property.kerberos.removeHostFromPrincipal" "true"

  # Zookeeper remove realm from principal
  set_hbase_property "hbase.zookeeper.property.kerberos.removeRealmFromPrincipal" "true"

  # Zookeeper znode
  set_hbase_property "zookeeper.znode.parent" "/hbase-secure"

  # HBase RPC protection
  set_hbase_property "hbase.rpc.protection" "privacy"

  # Kerberos properties for the UI
  set_hbase_property "hbase.security.authentication.ui" "kerberos"
  set_hbase_property "hbase.security.authentication.spnego.kerberos.principal" "HTTP/_HOST@${REALM}"
  set_hbase_property "hbase.security.authentication.spnego.kerberos.keytab" "/etc/security/keytab/http.service.keytab"
}

function configure_hbase() {
  # Prepare and merge configuration values:
  # hbase.rootdir
  local hbase_root_dir
  hbase_root_dir="$(get_property_in_xml "${CLUSTER_PROPERTIES_DIR}/hbase.xml" \
    hbase.rootdir)"
  if [[ -z "${hbase_root_dir}" ]]; then
    if [[ -n "${MASTER_ADDITIONAL}" ]]; then
      hbase_root_dir="hdfs://${CLUSTER_NAME}:8020/hbase"
    else
      hbase_root_dir="hdfs://${INSTANCE_NAME_PREFIX}-m/hbase"
    fi
  else
    # Ensure GCS path ends with slash eg: 'gs://bucket-name/'
    if [[ ! ${hbase_root_dir} =~ ^.*/$ ]]; then
      hbase_root_dir="${hbase_root_dir}/"
    fi
  fi
  bdconfig set_property \
    --configuration_file "${CLUSTER_PROPERTIES_DIR}/hbase.xml" \
    --name 'hbase.rootdir' --value "${hbase_root_dir}" \
    --clobber

  # zookeeper.quorum
  set_hbase_property "hbase.zookeeper.quorum" "${ZOOKEEPER_NODES_QUORUM}"
  set_hbase_property "hbase.zookeeper.property.clientPort" "${ZOOKEEPER_PORT_QUORUM}"

  set_hbase_property "hbase.zookeeper.property.initLimit" 20

  # Direct HBase to run in distributed mode.
  set_hbase_property "hbase.cluster.distributed" "true"

  # Prepare kerberos specific config values for hbase-site.xml
  if [[ "${ENABLE_KERBEROS}" == true ]]; then
    configure_hbase_with_kerberos
  fi

  # Merge all config values to hbase-site.xml
  bdconfig merge_configurations \
    --source_configuration_file "${CLUSTER_PROPERTIES_DIR}/hbase.xml" \
    --configuration_file "${HBASE_ETC_DIR}/conf/hbase-site.xml" \
    --clobber

  # merge user supplied log4j properties
  merge_java_properties ${CLUSTER_PROPERTIES_DIR}/hbase-log4j.properties \
    ${HBASE_ETC_DIR}/conf/log4j.properties
}

configure_hive() {
  # A comma separated list (with no spaces) of the jar files required for
  # Hive-HBase integration.
  local aux_jars_path
  aux_jars_path="$(get_property_in_xml /etc/hive/conf/hive-site.xml \
    hive.aux.jars.path)"
  local hive_version
  hive_version="$(os_installed_package_version hive)"
  if [[ -z "${aux_jars_path}" ]]; then
    aux_jars_path="file:///usr/lib/hive/lib/hive-hbase-handler-${hive_version}.jar"
  else
    aux_jars_path="${aux_jars_path},file:///usr/lib/hive/lib/hive-hbase-handler-${hive_version}.jar"
  fi

  bdconfig set_property \
    --configuration_file /etc/hive/conf/hive-site.xml \
    --name 'hive.aux.jars.path' \
    --value "${aux_jars_path}" \
    --clobber
}

function load_system_daemons() {
  cat >"${HBASE_MASTER_PATH}" <<EOF
[Unit]
Description=HBase Master
Wants=network-online.target
After=network-online.target hadoop-hdfs-namenode.service
StartLimitIntervalSec=0
[Service]
User=root
Group=root
Type=simple
EnvironmentFile=/etc/environment
Environment=HBASE_ETC_DIR=/etc/hbase
ExecStart=/usr/bin/hbase --config ${HBASE_ETC_DIR}/conf/ master start
Restart=on-failure
RestartSec=1
[Install]
WantedBy=multi-user.target
EOF

  cat >"${HBASE_REGION_SERVER_PATH}" <<EOF
[Unit]
Description=HBase Regionserver
Wants=network-online.target
After=network-online.target hadoop-hdfs-datanode.service
[Service]
User=root
Group=root
Type=simple
EnvironmentFile=/etc/environment
Environment=HBASE_ETC_DIR=/etc/hbase
ExecStart=/usr/bin/hbase --config ${HBASE_ETC_DIR}/conf/ regionserver start
[Install]
WantedBy=multi-user.target
EOF

  retry_constant systemctl daemon-reload
}

function kerberize() {
  # Define keytab path based on role
  hbase_keytab_path=/etc/security/keytab/hbase.service.keytab

  # Change regionserver information
  for ((c = 0; c < WORKER_COUNT; c++)); do
    echo "${INSTANCE_NAME_PREFIX}-w-${c}.${DOMAIN}" >>/tmp/regionservers
  done
  mv /tmp/regionservers "${HBASE_ETC_DIR}"/conf/regionservers

  # Add server JAAS
  cat >/tmp/hbase-server.jaas <<EOF
Client {
  com.sun.security.auth.module.Krb5LoginModule required
  useKeyTab=true
  storeKey=true
  useTicketCache=false
  keyTab="${hbase_keytab_path}"
  principal="hbase/${FQDN}";
};
EOF

  # Copy JAAS file to hbase conf directory
  mv /tmp/hbase-server.jaas "${HBASE_ETC_DIR}"/conf/hbase-server.jaas

  # Add client JAAS
  cat >/tmp/hbase-client.jaas <<EOF
Client {
  com.sun.security.auth.module.Krb5LoginModule required
  useKeyTab=false
  useTicketCache=true;
};
EOF

  # Copy JAAS file to hbase conf directory
  mv /tmp/hbase-client.jaas "${HBASE_ETC_DIR}"/conf/hbase-client.jaas

  # Extend hbase environment variable script
  cat "${HBASE_ETC_DIR}"/conf/hbase-env.sh >/tmp/hbase-env.sh
  cat >>/tmp/hbase-env.sh <<EOF
export HBASE_MANAGES_ZK=false
export HBASE_OPTS="\$HBASE_OPTS -Djava.security.auth.login.config=/etc/hbase/conf/hbase-client.jaas"
export HBASE_MASTER_OPTS="\$HBASE_MASTER_OPTS -Djava.security.auth.login.config=/etc/hbase/conf/hbase-server.jaas"
export HBASE_REGIONSERVER_OPTS="\$HBASE_REGIONSERVER_OPTS -Djava.security.auth.login.config=/etc/hbase/conf/hbase-server.jaas"
EOF

  # Copy script to hbase conf directory
  mv /tmp/hbase-env.sh "${HBASE_ETC_DIR}"/conf/hbase-env.sh
}

function main() {
  usermod -a -G hadoop hbase
  load_system_daemons
  configure_hbase
  configure_hive
  if [[ "${ENABLE_KERBEROS}" == true ]]; then
    kerberize
    retry_constant \
      kinit -kt /etc/security/keytab/dataproc.service.keytab "dataproc/${FQDN}"
  fi

  # Start hbase-regionserver on workers, and on the master in single node
  # clusters since they don't have workers.
  if [[ "${ROLE}" == 'Master' && "${WORKER_COUNT}" -eq 0 || "${ROLE}" == 'Worker' ]]; then
    # Wait for zookeeper server ready to connect before starting regionserver.
    local -a ZOOKEEPER_NODE_LIST
    IFS=',' read -r -a ZOOKEEPER_NODE_LIST <<<"${ZOOKEEPER_NODES}"
    for zk_node in "${ZOOKEEPER_NODE_LIST[@]}"; do
      # shellcheck disable=SC2206
      host_and_port=(${zk_node//:/ })
      wait_for_port "zookeeper" "${host_and_port[0]}" "${host_and_port[1]}" "300"
    done
    retry_constant systemctl enable hbase-regionserver
    retry_constant systemctl start hbase-regionserver
  fi

  if [[ "${ROLE}" == 'Master' ]]; then
    if is_component_selected "hdfs"; then
      echo "Delaying starting hbase-master until HDFS is ready"
    else
      start_hbase_master
    fi
  fi
}

main
