#!/bin/bash

set -euo pipefail

source /etc/environment
source "$(dirname "$0")/../../bdutil_env.sh"
source "$(dirname "$0")/../../bdutil_logging.sh"
source "$(dirname "$0")/../../bdutil_retry.sh"
source "$(dirname "$0")/../../bdutil_networking.sh"
source "$(dirname "$0")/../../bdutil_metadata.sh"
source "$(dirname "$0")/../../bdutil_properties.sh"
source "$(dirname "$0")/../../bdutil_components.sh"
source "$(dirname "$0")/../../bdutil_services.sh"
source "$(dirname "$0")/../../bdutil_os.sh"
source "$(dirname "$0")/../shared/hdfs.sh"

set -x

MASTER_HOSTNAMES=(${DATAPROC_MASTER} ${DATAPROC_MASTER_ADDITIONAL//,/ })
CLUSTER_NAME=$(get_metadata_cluster_name)

function start_hdfs_namenode() {
  case "${MASTER_INDEX?}" in
  "0")
    loginfo 'Formatting NameNode'
    # This is the same as running`service hadoop-hdfs-namenode init`, but
    # done explicitly for consistency with bootstrapStandby and ZKC init,
    # which are not in # Bigtop
    retry_constant_short su -s /bin/bash hdfs -c \
      "source /etc/default/hadoop-hdfs-namenode && \
          hdfs namenode -format -nonInteractive"
    ;;
  "1")
    loginfo 'Synchronizing NameNode'
    retry_constant_short su -s /bin/bash hdfs -c \
      "source /etc/default/hadoop-hdfs-namenode && \
           hdfs namenode -bootstrapStandby -nonInteractive"
    ;;
  "2")
    local name_node_list
    name_node_list=$(get_property_in_xml "${HADOOP_CONF_DIR}/hdfs-site.xml" \
             "dfs.ha.namenodes.${CLUSTER_NAME}")
    if [[ "${name_node_list}" == "nn0,nn1,nn2" ]]; then
      loginfo 'Synchronizing NameNode'
      retry_constant_short su -s /bin/bash hdfs -c \
        "source /etc/default/hadoop-hdfs-namenode && \
            hdfs namenode -bootstrapStandby -nonInteractive"
    else
      loginfo 'Not starting unsupported tertiary NameNode'
      return
    fi
    ;;
   *)
    loginfo 'Not starting unsupported tertiary NameNode'
    return
    ;;
  esac

  enable_and_start_service "hadoop-hdfs-namenode"
}

function start_hdfs_zkfc() {
  case "${MASTER_INDEX?}" in
  "0" | "1")
    enable_and_start_service "hadoop-hdfs-zkfc"
    ;;
  "2")
    local name_node_list
    name_node_list=$(get_property_in_xml "${HADOOP_CONF_DIR}/hdfs-site.xml" \
             "dfs.ha.namenodes.${CLUSTER_NAME}")
    if  [[ "${name_node_list}" != "nn0,nn1,nn2" ]]; then
      loginfo 'Not starting unsupported tertiary NameNode'
      return
    fi
    enable_and_start_service "hadoop-hdfs-zkfc"
    ;;
  esac
  return
}

function format_hdfs_zkfc() {
  case "${MASTER_INDEX?}" in
  "0")
    loginfo 'Formatting ZKFC node'
    retry_constant_short su -s /bin/bash hdfs -c \
      "source /etc/default/hadoop-hdfs-zkfc && \
           hdfs zkfc -formatZK -nonInteractive"
    ;;
  "1")
    # continue
    ;;
  "2")
    local name_node_list
    name_node_list=$(get_property_in_xml "${HADOOP_CONF_DIR}/hdfs-site.xml" \
             "dfs.ha.namenodes.${CLUSTER_NAME}")
    if  [[ "${name_node_list}" != "nn0,nn1,nn2" ]]; then
      loginfo 'Not starting unsupported tertiary NameNode'
      return
    fi
    ;;
  *)
    loginfo 'Not starting unsupported tertiary NameNode'
    return
    ;;
  esac
}

function start_hdfs_secondarynamenode() {
  enable_and_start_service "hadoop-hdfs-secondarynamenode"
}

function start_hdfs_journalnode() {
  enable_and_start_service "hadoop-hdfs-journalnode"
}

function start_master_services() {
  login_as_hdfs_if_kerberos_enabled

  # Regular (1 master, n workers) or single-node cluster.
  if [[ "${MASTER_COUNT}" == "1" || "${WORKER_COUNT}" == "0" ]]; then
    start_hdfs_namenode
    start_hdfs_secondarynamenode
    init_hcfs_dirs
    return
  fi

# HA cluster: 3 masters, n worker(s).
  format_hdfs_zkfc
  start_hdfs_journalnode
  start_hdfs_namenode
  start_hdfs_zkfc
  if [[ "${MASTER_INDEX}" == "0" ]]; then
    init_hcfs_dirs
  fi
}

# Enable without starting worker services. They need to be started by the
# agent after initialization (e.g., install custom libraries) is done.
#
# The sentinel file /var/lib/google/dataproc/ready_to_start_worker_services
# is used by worker services to prevent unintended start, the agent will
# generate the file then start the service after initialization is done
# successfully.
function enable_worker_services() {
  # Do not enable the datanode service on secondary workers.
  local -r datanode_enabled=$(get_metadata_datanode_enabled || echo "")
  if [[ "${datanode_enabled}" == 'true' ]]; then
    enable_service "hadoop-hdfs-datanode"
  fi
}

function activate_hdfs() {
  # Call unused command to load NameNode byte code into disk cache.
  sudo -u hdfs hdfs namenode -genclusterid &>/dev/null &

  mkdir -p /var/run/hadoop-hdfs
  chown root:hdfs /var/run/hadoop-hdfs
  chmod 775 /var/run/hadoop-hdfs

  if [[ "${ROLE}" == "Master" ]]; then
    start_master_services
    # On single-node clusters, enable and start the datanode.
    # The datanode must be started in the startup script (as opposed to in the
    # agent) so that optional components that depend on HDFS can be activated.
    if [[ "${WORKER_COUNT}" == 0 ]]; then
      enable_and_start_service "hadoop-hdfs-datanode"
    fi
  else
    enable_worker_services
  fi
}

activate_hdfs
