#!/bin/bash

set -euo pipefail

source "$(dirname "$0")/../../bdutil_logging.sh"
source "$(dirname "$0")/../../bdutil_retry.sh"
source "$(dirname "$0")/../../bdutil_properties.sh"
source "$(dirname "$0")/../../bdutil_components.sh"
source "$(dirname "$0")/../shared/effective-python.sh"

set -x

HIVE_WEBHCAT_CONF_DIR=/etc/hive-webhcat/conf
HIVE_WEBHCAT_ENVFILE=/usr/lib/hive-hcatalog/libexec/hcat-config.sh
readonly CLUSTER_PROPERTIES_DIR="/tmp/cluster/properties"

function gen_passwd() {
  head -c 10 < <(cat /dev/urandom | tr -dc 'a-zA-Z0-9')
}

function set_webhcat_property() {
  bdconfig set_property \
    --configuration_file "${HIVE_WEBHCAT_CONF_DIR}/webhcat-site.xml" \
    --name "$1" --value "$2" \
    --create_if_absent \
    --clobber ||
    fatal "Unable to set $1"
}

function configure_hcat_hbase() {
  # Hide hbase from hcat if we're going to uninstall hbase later.
  #
  # The hcat command looks for hbase libraries at invocation time when
  # determining the classpath to use. Specifying /dev/null for hbase
  # locations prevents hcat from looking for them in its expected places.
  # We assume that normally HBASE_HOME and HBASE_CONF_DIR are *not*
  # specified in hcat-config.sh, so if we do intend to keep hbase we'll
  # unset the variables here to let hcat use its default logic.
  if is_component_selected hbase; then
    echo "unset HBASE_HOME" >>"${HIVE_WEBHCAT_ENVFILE}"
    echo "unset HBASE_CONF_DIR" >>"${HIVE_WEBHCAT_ENVFILE}"
  else
    echo "HBASE_HOME='/dev/null'" >>"${HIVE_WEBHCAT_ENVFILE}"
    echo "HBASE_CONF_DIR='/dev/null'" >>"${HIVE_WEBHCAT_ENVFILE}"
  fi
}

function kerberize() {
  if [[ ! -f "${HIVE_WEBHCAT_CONF_DIR}/webhcat-site.xml" ]]; then
    cat >"${HIVE_WEBHCAT_CONF_DIR}/webhcat-site.xml" <<EOF
<?xml version="1.0" ?>
<configuration>
</configuration>
EOF
  fi
  chown hive:hive "${HIVE_WEBHCAT_CONF_DIR}/webhcat-site.xml"
  chmod 400 "${HIVE_WEBHCAT_CONF_DIR}/webhcat-site.xml"

  set +x
  secret=$(gen_passwd)
  set -x
  fqdn=$(hostname -f)
  realm=$(grep "default_realm = " /etc/krb5.conf | awk '{print $NF}')
  set +x
  set_webhcat_property "templeton.kerberos.secret" "${secret}"
  set -x
  set_webhcat_property "templeton.kerberos.principal" "HTTP/${fqdn}@${realm}"
  set_webhcat_property "templeton.kerberos.keytab" "/etc/security/keytab/http.service.keytab"
}

if is_component_selected 'kerberos'; then
  kerberize
fi

# Determine effective python.
install_effective_python_profile

# Force webhcat to use the effective python.
PYTHON_CMD="$(get_effective_python_path)/python"
set_webhcat_property "templeton.python" "${PYTHON_CMD}"

# /usr/lib/hive-hcatalog/bin/hcat.py makes specific assumptions about
# where logs directory is. We work around this by symlinking to /var/log.
ln -s /var/log/hive-hcatalog/ /usr/lib/hive/logs

rm /usr/lib/hive-hcatalog/etc/webhcat
ln -s "${HIVE_WEBHCAT_CONF_DIR}" /usr/lib/hive-hcatalog/etc/webhcat

# Force hcat to use a standard log format.
sed -E "s/(.*\.layout\.pattern = ).*/\1%d{ISO8601} %5p [%t] %c{2}: %m%n/g" \
  -i "${HIVE_WEBHCAT_CONF_DIR}/webhcat-log4j2.properties"

configure_hcat_hbase

# merge user supplied log4j properties
merge_java_properties ${CLUSTER_PROPERTIES_DIR}/webhcat-log4j2.properties \
  ${HIVE_WEBHCAT_CONF_DIR}/webhcat-log4j2.properties

retry_constant systemctl enable hive-webhcat-server
retry_constant systemctl start hive-webhcat-server
