#!/bin/bash

set -euo pipefail

source "$(dirname "$0")/../../bdutil_env.sh"
source "$(dirname "$0")/../../bdutil_logging.sh"
source "$(dirname "$0")/../../bdutil_networking.sh"
source "$(dirname "$0")/../../bdutil_retry.sh"
source "$(dirname "$0")/../../bdutil_metadata.sh"
source "$(dirname "$0")/../../bdutil_properties.sh"
source "$(dirname "$0")/../../bdutil_versions.sh"
source "$(dirname "$0")/../../bdutil_components.sh"
source "$(dirname "$0")/../../bdutil_gcs.sh"
source "$(dirname "$0")/../shared/jupyter.sh"

set -x

DATAPROC_MASTER="$(get_metadata_master)"
readonly DATAPROC_MASTER
DATAPROC_BUCKET="$(get_metadata_bucket)"
readonly DATAPROC_BUCKET
JUPYTERHUB_ENABLED=$(get_dataproc_property jupyter.hub.enabled)
readonly JUPYTERHUB_ENABLED
ADD_HUB_MENU=$(get_dataproc_property jupyter.hub.menu.enabled)
readonly ADD_HUB_MENU
JUPYTER_DATAPROC_USER_ENABLED=$(get_dataproc_property jupyter.dataprocuser.enabled)
readonly JUPYTER_DATAPROC_USER_ENABLED
install -d "${JUPYTER_ETC_DIR}"
chmod a+r "${JUPYTER_ETC_DIR}"

S8S_JUPYTER_KERNEL_TO_ENABLE=$(get_dataproc_property internal.s8s.jupyter.kernel | tr '[:upper:]' '[:lower:]')
readonly S8S_JUPYTER_KERNEL_TO_ENABLE

function generate_notebook_location() {
  # Default GCS directory in which to store notebooks.
  # DATAPROC_BUCKET must not include gs:// and bucket must exist
  local -r default_notebook_dir="gs://${DATAPROC_BUCKET}/notebooks/jupyter"

  # Check if a user-specified notebook location was provided.
  local notebook_dir
  notebook_dir=$(get_dataproc_property_or_default \
    'jupyter.notebook.gcs.dir' "${default_notebook_dir}")
  # Strip 'gs://' prefix.
  notebook_dir=${notebook_dir/gs:\/\//}

  NOTEBOOK_BUCKET=$(echo "${notebook_dir}" | cut -d '/' -f 1)
  NOTEBOOK_PATH=${notebook_dir#"${NOTEBOOK_BUCKET}/"}

  # Verify we can access the GCS bucket.
  check_gcs_bucket_access "gs://${notebook_dir}" 'jupyter'

  if [[ "${DATAPROC_IMAGE_TYPE}" == "standard" ]]; then
    # Create storage path if it does not exist
    hadoop fs -mkdir -p "gs://${notebook_dir}"
  fi
}

function configure_jupyter() {
  # Get user-provided port.
  local jupyter_port
  if [[ "${DATAPROC_IMAGE_TYPE}" == "serverless-spark" ]]; then
    # Using port 8443 for s8s as we don't use knox here
    jupyter_port=$(get_dataproc_property_or_default 'jupyter.port' '8443')
  else
    jupyter_port=$(get_dataproc_property_or_default 'jupyter.port' '8123')
  fi

  # The following snippet ensures that the `root_dir` parameter for the
  # CombinedContentsManager has a single entry named "Local Disk" that
  # points at the `root_dir` parameter for the FileContentsManager.
  local -r symlink_dir="${JUPYTER_ETC_DIR}/symlinks_for_jupyterlab_widgets"
  mkdir -p "${symlink_dir}"
  ln -s / "${symlink_dir}/Local Disk"

  echo "PATH=${CONDA_BIN_DIR}:$PATH" >"${JUPYTER_ENV_FILE}"

  cat >>"${JUPYTER_CONFIG_FILE}" <<EOF
from gcs_contents_manager import CombinedContentsManager, GCSContentsManager

c.Application.log_level = 'INFO'
c.Application.log_datefmt = '%Y-%m-%d %H:%M:%S'
c.Application.log_format = '%(asctime)s %(levelname)s: %(name)s %(message)s'
c.JupyterApp.answer_yes = True
c.JupyterApp.log_level = 'INFO'
c.JupyterApp.log_datefmt = '%Y-%m-%d %H:%M:%S'
c.JupyterApp.log_format = '%(asctime)s %(levelname)s: %(name)s %(message)s'
c.NotebookApp.allow_root= True
c.NotebookApp.log_level = 'INFO'
c.NotebookApp.log_datefmt = '%Y-%m-%d %H:%M:%S'
c.NotebookApp.log_format = '%(asctime)s %(levelname)s: %(name)s %(message)s'
c.NotebookApp.open_browser = False
c.NotebookApp.port = ${jupyter_port}
c.NotebookApp.token = u''
c.NotebookApp.quit_button = False
c.NotebookApp.terminado_settings = {
    'shell_command': ['bash']
}
c.FileContentsManager.root_dir = '/'
EOF

  if [[ "${DATAPROC_IMAGE_TYPE}" == "standard" ]]; then
    cat >>"${JUPYTER_CONFIG_FILE}" <<EOF
c.GCSContentsManager.bucket_name = '${NOTEBOOK_BUCKET}'
c.GCSContentsManager.bucket_notebooks_path = '${NOTEBOOK_PATH}'
c.NotebookApp.contents_manager_class = CombinedContentsManager
c.CombinedContentsManager.root_dir = '${symlink_dir}'
EOF
  fi

  # If Knox is enabled, set some additional config options to allow proxying
  if ([[ "${DATAPROC_IMAGE_TYPE}" == "serverless-spark" ]] || is_component_selected 'knox')  && [[ "${JUPYTERHUB_ENABLED}" != "true" ]]; then
    local cluster_ui_hostname
    cluster_ui_hostname=$(get_dataproc_property \
      dataproc.proxy.public.hostname |
      sed 's/\\:/:/') # Unescape the colon in the URL
    local cluster_ui_hostname_with_no_protocol
    cluster_ui_hostname_with_no_protocol="${cluster_ui_hostname#https://}"
    cat >>"${JUPYTER_CONFIG_FILE}" <<EOF
c.NotebookApp.base_url = "/gateway/default/jupyter"
c.NotebookApp.allow_origin_pat = "(https?://)?(${cluster_ui_hostname}|localhost:8443)"
c.NotebookApp.disable_check_xsrf = True
c.NotebookApp.local_hostnames = ['${cluster_ui_hostname_with_no_protocol}', 'localhost']
EOF
  fi

  # Apply restrictions to what IPs Jupyter will listen on
  if [[ "${DATAPROC_IMAGE_TYPE}" == "serverless-spark" ]]; then
    echo "c.NotebookApp.ip = '127.0.0.1'" >>"${JUPYTER_CONFIG_FILE}"
  elif [[ "$(get_dataproc_property jupyter.listen.all.interfaces)" == "true" ]] ||
      [[ "${JUPYTERHUB_ENABLED}" == "true" ]]; then
    echo "c.NotebookApp.ip = '*'" >>"${JUPYTER_CONFIG_FILE}"
  elif is_version_at_least "${DATAPROC_IMAGE_VERSION}" "2.0" ||
      is_component_selected 'knox'; then
    echo "c.NotebookApp.ip = '127.0.0.1'" >>"${JUPYTER_CONFIG_FILE}"
  else
    # Legacy behavior for pre-2.0 compatibility: allow to connect from any host
    echo "c.NotebookApp.ip = '*'" >>"${JUPYTER_CONFIG_FILE}"
  fi
}

function install_libraries() {
  if [[ "${DATAPROC_IMAGE_TYPE}" == "serverless-spark" ]]; then
    return 0
  fi

  if ! is_version_at_least "${DATAPROC_IMAGE_VERSION}" "2.0"; then
    # Force installation of jupyter-gcs-contents-manager from local wheel
    echo 'jupyter-gcs-contents-manager' >>"${JUPYTER_EXTRA_PACKAGES}"

    # Install cached wheels.
    "${CONDA_BIN_DIR}/pip" install --no-index --upgrade-strategy only-if-needed \
      --find-links="${JUPYTER_WHEEL}" -r "${JUPYTER_EXTRA_PACKAGES}"
  fi

  # For serverless-spark images, we have moved this phase to post-install
  create_ipython_profile
  enable_jupyter_extensions

  # Enable Spark Monitor for Jupyter for Dataproc 1.5 only,
  # because it does not work with Spark 3 and JupyterLab yet
  if ! is_version_at_least "${DATAPROC_IMAGE_VERSION}" "2.0"; then
    enable_spark_monitor_for_jupyter
  fi
}

function configure_kernels() {
  echo "Creating kernelspec"
  mkdir -p "$(dirname "${PYSPARK_KERNELSPEC}")"
  chmod a+r "$(dirname "${PYSPARK_KERNELSPEC}")"

  # Jupyter requires Conda.
  local -r python="${CONDA_BIN_DIR}/python"

  # {connection_file} is a magic variable that Jupyter fills in for us
  # Note: we can only use it in argv, so cannot use env to set those
  # environment variables.
  echo "Generating ${PYSPARK_KERNELSPEC}"
  if [[ "${DATAPROC_IMAGE_TYPE}" == "serverless-spark" ]]; then
    # s8s will set env variables in entrypoint.sh
    cat >"${PYSPARK_KERNELSPEC}" <<EOF
{
  "argv": [
    "bash",
    "-c",
    "PYSPARK_DRIVER_PYTHON_OPTS='kernel -f {connection_file}' pyspark"
  ],
  "display_name": "PySpark",
  "language": "python"
}
EOF
  else
    cat >"${PYSPARK_KERNELSPEC}" <<EOF
{
  "argv": [
    "bash",
    "-c",
    "PYSPARK_DRIVER_PYTHON_OPTS='kernel -f {connection_file}' pyspark"
  ],
  "display_name": "PySpark",
  "language": "python",
  "env": {
    "PYSPARK_DRIVER_PYTHON": "${CONDA_BIN_DIR}/ipython",
    "PYSPARK_PYTHON": "${python}"
  }
}
EOF
  fi
  # Ensure Jupyter has picked up the PySpark kernel
  "${CONDA_BIN_DIR}/jupyter" kernelspec list | grep kernels/pyspark ||
    log_and_fail "jupyter" "Failed to create PySpark kernelspec"

  if [[ "${DATAPROC_IMAGE_TYPE}" == "serverless-spark" ]]; then
    if [[ -n "${S8S_JUPYTER_KERNEL_TO_ENABLE}" ]] && [[ "${S8S_JUPYTER_KERNEL_TO_ENABLE}" != "kernel_unspecified" ]]; then
      # Get the list of kernelspecs that we need to uninstall
      local uninstall_kernelspecs
      uninstall_kernelspecs=($("${CONDA_BIN_DIR}/jupyter" kernelspec list --json | jq -r --arg ENABLED_KERNEL "${S8S_JUPYTER_KERNEL_TO_ENABLE}" '.kernelspecs | to_entries | .[] | select(.value.spec.language != $ENABLED_KERNEL) | .key'))
      for ks in "${uninstall_kernelspecs[@]}"
      do
        # Here we ignore python3 kernel as it is always shows up in the list
        # result, even after it gets uninstalled. Given that serverless images
        # have already uninstalled python3 kernelspec before, we are ok to skip
        # here.
        if [[ "${ks}" != "python3" ]]; then
          "${CONDA_BIN_DIR}/jupyter" kernelspec uninstall "${ks}" -y
        fi
      done
    fi
  elif [[ "${DATAPROC_IMAGE_TYPE}" == "standard" ]]; then
    # Update Python kernel with needed environment variables for PySpark libs.
    local python_kernel_dir="${CONDA_DIR}/share/jupyter/kernels/python3"
    mkdir -p "${python_kernel_dir}"
    local python_kernelspec
    python_kernelspec=$(ls "${python_kernel_dir}/kernel.json")
    echo "Generating ${python_kernelspec}"
    cat >"${python_kernelspec}" <<EOF
{
  "argv": [
    "${python}",
    "-m",
    "ipykernel_launcher",
    "-f",
    "{connection_file}"
  ],
  "display_name": "Python 3",
  "language": "python",
  "env": {
    "PYSPARK_PYTHON": "${python}",
    "SPARK_HOME": "/usr/lib/spark"
  }
}
EOF

    # Ensure Jupyter has picked up the Python kernel
    "${CONDA_BIN_DIR}/jupyter" kernelspec list | grep kernels/python ||
      log_and_fail "jupyter" "Failed to create Python kernelspec"

    if is_version_at_least "${DATAPROC_IMAGE_VERSION}" "2.1"; then
      # Install Spylon kernel
      "${CONDA_BIN_DIR}/jupyter" toree install --spark_home=/usr/lib/spark/ --sys-prefix
      # Ensure Jupyter has picked up the new kernel
      "${CONDA_BIN_DIR}/jupyter" kernelspec list | grep kernels/apache_toree_scala ||
        log_and_fail "jupyter" "Failed to create Toree kernelspec"
    else
      # Install Spylon kernel
      "${CONDA_BIN_DIR}/python" -m spylon_kernel install --sys-prefix
      # Ensure Jupyter has picked up the new kernel
      "${CONDA_BIN_DIR}/jupyter" kernelspec list | grep kernels/spylon ||
        log_and_fail "jupyter" "Failed to create Spylon kernelspec"
    fi

    "${CONDA_BIN_DIR}/jupyter" kernelspec list | grep kernels/ir ||
      log_and_fail "jupyter" "Failed to create R kernelspec"

    # Fix R kernel to get Spark libraries out of the box.
    echo "Generating ${R_KERNELSPEC}"
    cat >"${R_KERNELSPEC}" <<EOF
{
  "argv": [
    "R",
    "--slave",
    "-e",
    "IRkernel::main()",
    "--args",
    "{connection_file}"
  ],
  "display_name": "R",
  "language": "R",
  "env": {
    "SPARK_HOME": "/usr/lib/spark",
    "R_LIBS_USER": "${CONDA_DIR}/lib/R/library:/usr/lib/spark/R/lib"
  }
}
EOF
  fi
}

function configure_default_workspace() {
  local -r jupyter_username="${1}"
  local -r workspaces_subdir=".jupyter/lab/workspaces"
  local -r root_workspaces_dir="/root/${workspaces_subdir}"

  # First, make sure we are starting with an empty workspaces directory
  mkdir -p "${root_workspaces_dir}"
  rm -f "${root_workspaces_dir}"/*

  # There are two different "default" workspaces; the workspace named
  # "default" which is used when you first connect to JupyterLab and can
  # not be modified using the `jupyter lab workspaces import` command.
  #
  # The other is the "/lab" workspace which is exported by default if a
  # user does not specify a workspace name in the
  # `jupyter lab workspaces export` command.
  #
  # We want to have both default to the "GCS" subdirectory, so we set
  # both with initial values explicitly.

  # The "/lab" workspace is set using the `jupyter lab workspaces import`
  # command.

  local workspace_file
  workspace_file=$(mktemp)
  local metadata_id="lab"
  if ! is_version_at_least "${DATAPROC_IMAGE_VERSION}" "2.1"; then
    metadata_id="/${metadata_id}"
  fi

  cat >"$workspace_file" <<EOF
{
  "data": {
    "file-browser-filebrowser:cwd": {
      "path": "GCS"
    }
  },
  "metadata": {"id": "${metadata_id}"}
}
EOF

  "${CONDA_BIN_DIR}"/jupyter lab workspaces import "$workspace_file"
  rm "$workspace_file"

  # The "default" workspace can only be configured by manually writing
  # the "default-37a8.jupyterlab-workspace" file in the jupyter user's
  # home directory. To do that we copy over the contents of the "/lab"
  # workspace and then change its ID.
  cp "${root_workspaces_dir}/lab-a511.jupyterlab-workspace" \
      "${root_workspaces_dir}/default-37a8.jupyterlab-workspace"
  sed -i -e 's/\/lab/default/g' \
      "${root_workspaces_dir}/default-37a8.jupyterlab-workspace"

  if [[ "${jupyter_username}" != "root" ]]; then
      # Copy the workspaces that we have set up to the jupyter user
      # in case it is different from root.
      local -r jupyter_user_workspaces_dir="/home/${jupyter_username}/${workspaces_subdir}"
      mkdir -p "${jupyter_user_workspaces_dir}"
      cp "${root_workspaces_dir}"/* "${jupyter_user_workspaces_dir}"/.
      chown -R "${jupyter_username}" "${jupyter_user_workspaces_dir}"
  fi
}

function parse_jupyterhub_env() {
  python3 - "${1}" <<END
import json
import sys

env_dict = json.loads(sys.argv[1])

for key, value in env_dict.items():
  print(f'Environment={key}={value}')
END
}

function get_service_username() {
  if [[ "${DATAPROC_IMAGE_TYPE}" == "serverless-spark" ]]; then
    echo "spark"
  elif [[ "${JUPYTER_DATAPROC_USER_ENABLED}" == "true" ]]; then
    # The Jupyter service should run as the user `dataproc` rather than `root`.
    echo "dataproc"
  else
    echo "root"
  fi
}

function set_dir_permissions() {
  local -r service_username="$1"

  # Make sure all of the files required by Jupyter are accessible to the
  # user that will run the Jupyter server.
  if [[ "${service_username}" != "root" ]]; then
    chown -R "${service_username}":"${service_username}" /etc/jupyter
    chown -R "${service_username}":"${service_username}" /etc/default/jupyter
    chown -R "${service_username}":"${service_username}" /opt/dataproc/jupyter
    chown -R "${service_username}":"${service_username}" "${CONDA_DIR}/share/jupyter"

    if [[ "${DATAPROC_IMAGE_TYPE}" == "serverless-spark" ]]; then
      chown -R "${service_username}":"${service_username}" /home/spark/.ipython/
    fi

    mkdir -p "${JUPYTER_LOG_DIR}"
    touch "${JUPYTER_LOG_FILE}"
    chown "${service_username}":"${service_username}" "${JUPYTER_LOG_FILE}"
  fi
}

function generate_systemd_unit_and_start() {
  local -r service_username="$1"

  # For serverless image, we don't need to config workspace, as we want users to
  # run Jupyter outside of the session.
  if [[ "${DATAPROC_IMAGE_TYPE}" == "standard" ]]; then
    configure_default_workspace "${service_username}"
  fi

  # If Jupyterhub is enabled, then pass it the environment and arguments.
  # Frontend validation will ensure that the image being used has python3 as the
  # default and has the expected dependencies before running Jupyterhub-singleuser
  if [[ "${JUPYTERHUB_ENABLED}" == "true" ]] || [[ "${ADD_HUB_MENU}" == "true" ]]; then

    # Formatted as a JSON-parsable string of (key,value) pairs by the
    # DataprocSpawner, will be parsed into a dictionary later
    # Reference: https://jupyterhub.readthedocs.io/en/stable/api/spawner.html#jupyterhub.spawner.Spawner.environment
    local jupyterhub_env
    jupyterhub_env=$(get_dataproc_property jupyter.hub.env)
    # Formatted as ENVIRONMENT={key}={value}
    local jupyterhub_env_parsed
    jupyterhub_env_parsed=$(parse_jupyterhub_env "${jupyterhub_env}")
    local jupyterhub_args
    jupyterhub_args=$(get_dataproc_property jupyter.hub.args)
    local jupyterhub_binary="jupyterhub-singleuser"

    if [[ "${ADD_HUB_MENU}" == "true" ]]; then
      # Patch Jupyterhub-singleuser to not callback to the Hub.
      # The patches: get_user_hack, get_token_hack, oauth_scopes are needed for
      # Dataproc hub to work with a 2.1 cluster yaml.
      cat >"${CONDA_BIN_DIR}/${jupyterhub_binary}-dataproc" <<EOF
#!${CONDA_BIN_DIR}/python
import sys

from jupyterhub.singleuser import SingleUserNotebookApp
from notebook.auth.login import LoginHandler
from notebook.auth.logout import LogoutHandler
from types import SimpleNamespace

def get_user_hack(*args):
  return {'name':args[0].settings['user']}

def get_token_hack(*args):
  return ''

class DataprocSingleUserNotebookApp(SingleUserNotebookApp):
  """Customized single-user notebook that uses Jupyter auth and
     and performs no callbacks to the parent Hub."""

  login_handler_class = LoginHandler
  logout_handler_class = LogoutHandler

  async def check_hub_version(self):
      pass

  def init_hub_auth(self):
      self.hub_auth = SimpleNamespace(
        oauth_redirect_uri=self.hub_host + '/base_url/oauth_callback',
        get_user=get_user_hack,
        oauth_scopes=set(),
        get_token=get_token_hack
      )

  def init_webapp(self):
      super(DataprocSingleUserNotebookApp, self).init_webapp()
      self.jinja_template_vars['logo_url'] = self.base_url + 'static/base/images/logo.png'

if __name__ == '__main__':
    sys.exit(DataprocSingleUserNotebookApp.launch_instance())
EOF
      jupyterhub_binary="jupyterhub-singleuser-dataproc"
      chmod +x "${CONDA_BIN_DIR}/${jupyterhub_binary}"
    fi

    echo "Generating ${JUPYTERHUB_SYSTEMD_UNIT}"
    cat >"${JUPYTERHUB_SYSTEMD_UNIT}" <<EOF
[Unit]
Description=Jupyter Notebook Server for a single user
After=hadoop-yarn-resourcemanager.service
[Service]
Type=simple
User=${service_username}
Group=${service_username}
Restart=on-failure
EnvironmentFile=/etc/environment
EnvironmentFile=${JUPYTER_ENV_FILE}
${jupyterhub_env_parsed}
ExecStart=/bin/bash -c "${CONDA_BIN_DIR}/${jupyterhub_binary} ${jupyterhub_args} &>> ${JUPYTER_LOG_FILE}"
[Install]
WantedBy=multi-user.target
EOF

    retry_constant systemctl daemon-reload
    retry_constant systemctl enable jupyterhub
    retry_constant systemctl start jupyterhub

  else
    local exec_start
    if [[ "${DATAPROC_IMAGE_TYPE}" == "serverless-spark" ]]; then
      chmod +x "${BDUTIL_DIR}/jupyter-containerd-runner.sh"
      exec_start="/bin/bash -c '${BDUTIL_DIR}/jupyter-containerd-runner.sh 2>&1 | tee -a ${JUPYTER_LOG_FILE}'"
    else
      exec_start="/bin/bash -c '${CONDA_BIN_DIR}/jupyter notebook &>> ${JUPYTER_LOG_FILE}'"
    fi
    echo "Generating ${JUPYTER_SYSTEMD_UNIT}"
    cat >"${JUPYTER_SYSTEMD_UNIT}" <<EOF
[Unit]
Description=Jupyter Notebook Server
After=hadoop-yarn-resourcemanager.service
[Service]
Type=simple
User=${service_username}
Group=${service_username}
EnvironmentFile=/etc/environment
EnvironmentFile=${JUPYTER_ENV_FILE}
ExecStart=${exec_start}
Restart=on-failure
[Install]
WantedBy=multi-user.target
EOF

    retry_constant systemctl daemon-reload
    retry_constant systemctl enable jupyter
    retry_constant systemctl start jupyter
  fi
}

function wait_for_jupyter_server() {
  local jupyter_port
  local base_path
  if [[ "${DATAPROC_IMAGE_TYPE}" == "serverless-spark" ]]; then
    # Using port 8443 for s8s as we don't use knox here
    jupyter_port=$(get_dataproc_property_or_default 'jupyter.port' '8443')
    base_path="gateway/default/jupyter/"
  elif is_component_selected 'knox' && [[ "${JUPYTERHUB_ENABLED}" != "true" ]]; then
    jupyter_port=$(get_dataproc_property_or_default 'jupyter.port' '8123')
    base_path="gateway/default/jupyter/"
  else
    jupyter_port=$(get_dataproc_property_or_default 'jupyter.port' '8123')
    base_path=""
  fi

  # We should have at least 1 kernelspec
  retry_constant is_kernelspecs_available "${jupyter_port}" "${base_path}" ||
    log_and_fail "jupyter" "Failed to start jupyter server"
}

function is_kernelspecs_available() {
  local -r port=$1
  local -r base_path=$2
  curl -fs "http://localhost:${port}/${base_path}api/kernelspecs" |
    jq -e ".kernelspecs | length > 0"
}

function activate_jupyter() {
  local service_username
  service_username="$(get_service_username)"

  if [[ "${DATAPROC_IMAGE_TYPE}" == "standard" ]]; then
    generate_notebook_location
  fi
  configure_jupyter
  install_libraries
  configure_kernels
  set_dir_permissions "${service_username}"

  # For s8s where Jupyter component depends on containerd, thus here we need to
  # wait until the containerd is ready.
  if [[ "${DATAPROC_IMAGE_TYPE}" == "serverless-spark" ]]; then
    if ! wait_until_component_activated "containerd"; then
      log_and_fail "jupyter" "Prerequisite component containerd failed to initialize"
    fi
  fi

  generate_systemd_unit_and_start "${service_username}"
  wait_for_jupyter_server
}

activate_jupyter
