#!/bin/bash

set -euo pipefail

source "$(dirname "$0")/../../bdutil_env.sh"
source "$(dirname "$0")/../../bdutil_logging.sh"
source "$(dirname "$0")/../../bdutil_retry.sh"
source "$(dirname "$0")/../../bdutil_networking.sh"
source "$(dirname "$0")/../../bdutil_metadata.sh"
source "$(dirname "$0")/../../bdutil_properties.sh"
source "$(dirname "$0")/../../bdutil_os.sh"
source "$(dirname "$0")/../../bdutil_misc.sh"
source "$(dirname "$0")/../../bdutil_versions.sh"
source "$(dirname "$0")/../shared/zookeeper.sh"

set -x

readonly CLUSTER_PROPERTIES_DIR=/tmp/cluster/properties
readonly KAFKA_CUSTOMER_PROPERTIES=${CLUSTER_PROPERTIES_DIR}/kafka.properties
readonly KAFKA_HOME=/usr/lib/kafka
readonly KAFKA_CONFIG_DIR=/etc/kafka/conf
readonly KAFKA_PROP_FILE=/etc/kafka/conf/server.properties
readonly KAFKA_LOG_DIR=/var/lib/kafka-logs
readonly COLLECTD_DIR=/opt/stackdriver/collectd/etc/collectd.d
readonly DATAPROC_CONFIG_DIR=/usr/local/share/google/dataproc/conf

# The complete list of ZooKeeper server addresses, e.g.,
# "cluster1-m-0:2181,cluster1-m-1:2181,cluster1-m-2:2181".
ZOOKEEPER_LIST=''
# Integer broker ID, e.g., 0
BROKER_ID=''

ROLE="$(get_metadata_role)"

KAFKA_ON_MASTER=$(get_dataproc_property kafka.run.on.master)
if [[ -z "${KAFKA_ON_MASTER}" ]]; then
  KAFKA_ON_MASTER="true"
fi

function err() {
  echo "[$(date +'%Y-%m-%dT%H:%M:%S%z')]: $*" >&2
  return 1
}

ENABLE_STACKDRIVER=$(get_dataproc_property_or_default 'kafka.enable.stackdriver' 'true')
if [[ "${ENABLE_STACKDRIVER}" == "true" ]]; then
  # Check if Cluster has a proper monitoring scope to access Cloud Monitoring
  if ! (instance_has_scope "auth/monitoring" || instance_has_scope "auth/cloud-platform"); then
    err "Cluster does not have proper scope to enable 'Cloud Monitoring' for 'Kafka' component.
         Create the cluster with 'monitoring' or 'cloud-platform' scope to enable 'Cloud Monitoring' for 'Kafka' component".
  fi
fi

function get_broker_id() {
  if [[ "${ROLE}" == "Master" ]]; then
    # For master nodes, broker ID starts from 10,000.
    if [[ "$(hostname)" == *-m ]]; then
      # non-HA
      local -r broker_id=10000
    else
      # HA
      local -r broker_id=$((10000 + $(hostname | sed 's/.*-m-\([0-9]*\)$/\1/g')))
    fi
  else
    # For worker nodes, broker ID is the worker ID.
    local broker_id
    broker_id=$(hostname | sed 's/.*-w-\([0-9]*\)$/\1/g')
  fi
  echo "${broker_id}"
}

function configure_ssl_store() {
  local store_uri="$1"
  local store_path="$2"
  local store_uri_property="$3"
  local store_location_property="$4"

  if [[ -n "${store_uri}" ]]; then
    gsutil cp "${store_uri}" "${store_path}"
    cat >>"${KAFKA_PROP_FILE}" <<EOF
${store_location_property}=${store_path}
EOF
  else
    err "Property '${store_uri_property}' is not specified."
  fi
}

function configure_ssl_password() {
  local password_uri="$1"
  local kms_key_uri="$2"
  local password_uri_property="$3"
  local password_property="$4"

  if [[ -n "${password_uri}" ]]; then
    set +x
    local password
    password=$(decrypt_file_with_kms_key "${password_uri}" "${kms_key_uri}")
    cat >>"${KAFKA_PROP_FILE}" <<EOF
${password_property}=${password}
EOF
    set -x
  else
    err "Property '${password_uri_property}' is not specified."
  fi
}

# Configures SSL for brokers with the user-provided keystore and truststore.
function configure_ssl() {
  local ssl_keystore_uri
  ssl_keystore_uri=$(get_java_property "${KAFKA_CUSTOMER_PROPERTIES}" \
    ssl.keystore.uri)
  local ssl_keystore_password_uri
  ssl_keystore_password_uri=$(get_java_property \
    "${KAFKA_CUSTOMER_PROPERTIES}" ssl.keystore.password.uri)
  local ssl_key_password_uri
  ssl_key_password_uri=$(get_java_property \
    "${KAFKA_CUSTOMER_PROPERTIES}" ssl.key.password.uri)
  local ssl_truststore_uri
  ssl_truststore_uri=$(get_java_property \
    "${KAFKA_CUSTOMER_PROPERTIES}" ssl.truststore.uri)
  local ssl_truststore_password_uri
  ssl_truststore_password_uri=$(get_java_property \
    "${KAFKA_CUSTOMER_PROPERTIES}" ssl.truststore.password.uri)
  local kms_key_uri
  kms_key_uri=$(get_java_property "${KAFKA_CUSTOMER_PROPERTIES}" kms.key.uri)

  # Do nothing, if none of the SSL properties is specified.
  if [[ -z "${ssl_keystore_uri}" || -z "${ssl_keystore_password_uri}" || -z "${ssl_key_password_uri}" || -z "${ssl_truststore_uri}" || -z "${ssl_truststore_password_uri}" ]]; then
    echo "Skip configuring SSL."
    return
  fi

  if [[ -z "${kms_key_uri}" ]]; then
    err "Property 'kafka:kms.key.uri' is not specified."
  fi

  # If user specified kafka:security.inter.broker.protocol to SSL, only
  # configure SSL listener; otherwise, in addition to SSL, configure another
  # listener of that protocol. If the property is unspecified, use SSL by
  # default.
  local inter_broker_protocol
  inter_broker_protocol=$(get_java_property \
    "${KAFKA_CUSTOMER_PROPERTIES}" security.inter.broker.protocol)
  if [[ -z "${inter_broker_protocol}" ]]; then
    inter_broker_protocol="SSL"
  fi

  if [[ "${inter_broker_protocol}" == SSL ]]; then
    cat >>"${KAFKA_PROP_FILE}" <<EOF
listeners=SSL://0.0.0.0:9093
advertised.listeners=SSL://${HOSTNAME}:9093
security.inter.broker.protocol=${inter_broker_protocol}
EOF
  else
    cat >>"${KAFKA_PROP_FILE}" <<EOF
listeners=${inter_broker_protocol}://0.0.0.0:9092,SSL://0.0.0.0:9093
advertised.listeners=${inter_broker_protocol}://${HOSTNAME}:9092,SSL://${HOSTNAME}:9093
security.inter.broker.protocol=${inter_broker_protocol}
EOF
  fi

  local keystore_path="${KAFKA_CONFIG_DIR}"/ssl/broker.keystore.jks
  local truststore_path="${KAFKA_CONFIG_DIR}"/ssl/broker.truststore.jks

  mkdir "${KAFKA_CONFIG_DIR}"/ssl

  # keystore
  configure_ssl_store \
    "${ssl_keystore_uri}" \
    "${keystore_path}" \
    "kafka:ssl.keystore.uri" \
    "ssl.keystore.location"

  # keystore password
  configure_ssl_password \
    "${ssl_keystore_password_uri}" \
    "${kms_key_uri}" \
    "kafka:ssl.keystore.password.uri" \
    "ssl.keystore.password"

  # key password
  configure_ssl_password \
    "${ssl_key_password_uri}" \
    "${kms_key_uri}" \
    "kafka:ssl.key.password.uri" \
    "ssl.key.password"

  # truststore
  configure_ssl_store \
    "${ssl_truststore_uri}" \
    "${truststore_path}" \
    "kafka:ssl.truststore.uri" \
    "ssl.truststore.location"

  # truststore password
  configure_ssl_password \
    "${ssl_truststore_password_uri}" \
    "${kms_key_uri}" \
    "kafka:ssl.truststore.password.uri" \
    "ssl.truststore.password"

  chown -R kafka:kafka "${KAFKA_CONFIG_DIR}"/ssl
  chmod 755 "${KAFKA_CONFIG_DIR}"/ssl
  chmod 440 "${keystore_path}"
  chmod 444 "${truststore_path}"
}

# If local SSDs are attached, Kafka log dirs will be /mnt/<id>/kafka/kafka-logs;
# otherwise /kafka/kafka-logs.
function configure_kafka_log_dirs() {
  local data_dirs=($(get_data_dirs))
  local log_dirs="${data_dirs[*]/%/kafka/kafka-logs}"

  for data_dir in "${data_dirs[@]}"; do
    local log_dir="${data_dir}/kafka/kafka-logs"
    mkdir -p "${log_dir}"
    chown kafka:kafka -R "${log_dir}"
    chmod 755 "${log_dir}"
  done

  cat >>"${KAFKA_PROP_FILE}" <<EOF
log.dirs=${log_dirs/ /,}
EOF
}

function configure_kafka_server() {
  # Verify Kafka is installed.
  os_is_package_installed kafka-server || err 'kafka-server was not installed.'

  ZOOKEEPER_LIST=$(get_zookeeper_list)
  # Configure zookeeper.connect.
  if [[ -z "${ZOOKEEPER_LIST}" ]]; then
    err 'Failed to find configured ZooKeeper list in ${ZOOKEEPER_CONFIG_FILE}.'
  fi
  sed -i "s|^\(zookeeper\.connect=\).*|\1${ZOOKEEPER_LIST}|" \
    "${KAFKA_PROP_FILE}"

  # Configure broker.id
  BROKER_ID="$(get_broker_id)"
  sed -i "s|^\(broker\.id=\).*|\1${BROKER_ID}|" "${KAFKA_PROP_FILE}"

  # Others
  cat >>"${KAFKA_PROP_FILE}" <<EOF


# Dataproc properties.
reserved.broker.max.id=100000
delete.topic.enable=true
EOF

  configure_kafka_log_dirs

  # Add SSL properties before appending user-provided properties to make sure
  # users can always override default and Dataproc generated properties.
  configure_ssl

  # Allow user-provided config to override system generated config.
  merge_java_properties "${CLUSTER_PROPERTIES_DIR}/kafka.properties" \
    "${KAFKA_PROP_FILE}"

  if [[ "${ENABLE_STACKDRIVER}" == 'true' ]]; then
    # Enable JMX (default port 9999) in Kafka server.
    sed -i '/kafka-run-class.sh/i export KAFKA_JMX_OPTS="-Dcom.sun.management.jmxremote=true -Dcom.sun.management.jmxremote.authenticate=false -Dcom.sun.management.jmxremote.ssl=false -Djava.rmi.server.hostname=localhost"' \
      /usr/lib/kafka/bin/kafka-server-start.sh
    sed -i '/kafka-run-class.sh/i export JMX_PORT=${JMX_PORT:-9999}' \
      /usr/lib/kafka/bin/kafka-server-start.sh
  fi
}

function configure_stackdriver() {
  local observability_agents_helpers
  if is_arm64; then
    observability_agents_helpers="${BDUTIL_DIR}/observability_agents/arm.sh"
  elif ! is_containerised_legacy_agents_supported; then
    observability_agents_helpers="${BDUTIL_DIR}/observability_agents/legacy.sh"
  else
    observability_agents_helpers="${BDUTIL_DIR}/observability_agents/containerised_or_ucp.sh"
  fi

  source "$observability_agents_helpers"

  if ! use_ucp_observability_agents; then
    # Enable stackdriver config only for stackdriver agents.
    ln -s -t "${COLLECTD_DIR}" "${DATAPROC_CONFIG_DIR}/kafka-082.conf"
    systemctl restart "$(monitoring_agent_service_name)"
  fi

}

# Returns a list of broker IDs registered in ZooKeeper, e.g., " 0, 2, 1,".
function get_broker_list() {
  # The first ZooKeeper server address, e.g., "cluster1-m-0:2181".
  local zookeeper_address="${ZOOKEEPER_LIST%%,*}"
  ${KAFKA_HOME}/bin/zookeeper-shell.sh "${zookeeper_address}" \
    <<<"ls /brokers/ids" |
    grep '\[.*\]' |
    sed 's/\[/ /' |
    sed 's/\]/,/'
}

function wait_for_kafka() {
  # Wait until the broker is registered or time out.
  for i in {1..10}; do
    local broker_list
    broker_list=$(get_broker_list || true)
    if [[ "${broker_list}" == *" ${BROKER_ID},"* ]]; then
      return 0
    else
      echo "Kafka broker ${BROKER_ID} is not registered yet, retry ${i}..."
      sleep 5
    fi
  done
  echo "Failed to start Kafka broker ${BROKER_ID}." >&2
  exit 1
}

function start_kafka_server() {
  systemctl enable kafka-server
  systemctl restart kafka-server
}

if [[ "${ROLE}" == 'Worker' ]] || [[ "${ROLE}" == 'Master' && "${KAFKA_ON_MASTER}" == 'true' ]]; then
  wait_for_zookeeper
  configure_kafka_server
  start_kafka_server
  wait_for_kafka
  if [[ "${ENABLE_STACKDRIVER}" == 'true' ]]; then
    configure_stackdriver
  fi
fi
