#!/bin/bash

set -euo pipefail

source "$(dirname "$0")/../../bdutil_env.sh"
source "$(dirname "$0")/../../bdutil_logging.sh"
source "$(dirname "$0")/../../bdutil_retry.sh"
source "$(dirname "$0")/../../bdutil_metadata.sh"
source "$(dirname "$0")/../../bdutil_properties.sh"
source "$(dirname "$0")/../../bdutil_versions.sh"
source "$(dirname "$0")/../../bdutil_components.sh"
source "$(dirname "$0")/../../bdutil_gcs.sh"
source "$(dirname "$0")/../../bdutil_misc.sh"
source "$(dirname "$0")/../../bdutil_os.sh"
source "${BDUTIL_DIR}/components/shared/kerberos.sh"

# In general we want to enable debug through -x
# But there are also quite some commands involving passwords/keys
# so make sure you turn it off (set +x) before such commands.
set -x

readonly CONFIGBUCKET=$(get_metadata_bucket)
readonly ROLE=$(get_metadata_role)
readonly INSTANCE_NAME_PREFIX=$(get_instance_name_prefix)
readonly CLUSTER_UUID=$(get_metadata_cluster_uuid)
readonly STAGING_LOCATION="${CONFIGBUCKET}/google-cloud-dataproc-metainfo/${CLUSTER_UUID}/kerberos"
readonly WORKER_COUNT=$(get_metadata_worker_count)
readonly FQDN=$(hostname -f)
readonly DOMAIN=$(dnsdomainname)

readonly ADDITIONAL_MASTERS=($(get_metadata_master_additional | sed 's/,/\n/g'))
readonly MASTER="$(/usr/share/google/get_metadata_value attributes/dataproc-master)"
readonly MASTERS=($MASTER ${ADDITIONAL_MASTERS[*]})
readonly MASTER_COUNT=${#MASTERS[@]}
readonly MASTER_FQDN="${MASTER}.${DOMAIN}"

readonly HADOOP_CONF_DIR="/etc/hadoop/conf"
readonly KEYTAB_DIR="/etc/security/keytab"
mkdir -p "${KEYTAB_DIR}"

function get_root_password_from_control_plane() {
  if [[ -f "${HANDS_OFF_PASSWORD}" ]]; then
    local -r root_password="$(read_hands_off_kerberos_password "$HANDS_OFF_PASSWORD")"
    if [[ -z "${root_password}" ]]; then
      log_and_fail "kerberos" "Root principal password cannot be empty!" "${EXIT_CODE_INTERNAL_ERROR}"
    fi
  else
    log_and_fail "kerberos" "Failed to read  from file containing generated password." "${EXIT_CODE_INTERNAL_ERROR}"
  fi
  echo "${root_password}"
}

REALM=$(get_dataproc_property kerberos.realm)
if [[ -z "${REALM}" ]]; then
  REALM=$(echo "${DOMAIN}" | awk '{print toupper($0)}')
fi

# KMS_KEY might be used to decrypt other secrets, not necessarily the
# root principal password, so extract it first
KMS_KEY_URI=$(get_dataproc_property kerberos.kms.key.uri)

# Either read the generated password, or read from GCS then decrypt with KMS_KEY.
GENERATE_PASSWORD=$(get_dataproc_property kerberos.beta.automatic-config.enable)
if [[ "${GENERATE_PASSWORD}" == "true" ]]; then
  set +x
  ROOT_PRINCIPAL_PASSWORD="$(get_root_password_from_control_plane)"
  set -x
else
  if [[ -z "${KMS_KEY_URI}" ]]; then
    log_and_fail "kerberos" "Kerberos component is enabled but failed to read KMS key URI from Dataproc properties. \
  Set it through 'dataproc:kerberos.kms.key.uri'." "${EXIT_CODE_CLIENT_ERROR}"
  fi

  set +x
  ROOT_PRINCIPAL_PASSWORD=$(get_kerberos_root_password_from_gcs "${KMS_KEY_URI}")
  set -x
fi

KDC_DB_KEY_URI=$(get_dataproc_property kerberos.kdc.db.key.uri)
if [[ -z "${KDC_DB_KEY_URI}" ]]; then
  echo "User does not provide property 'dataproc:kerberos.kdc.db.key.uri'. Generating a database key."
  set +x
  # TODO: think about a more random/secure way to generate the db key, but also keep it consistent
  # among masters in HA mode.
  KDC_DB_KEY=$(sha256sum_10 "${ROOT_PRINCIPAL_PASSWORD}")
  set -x
else
  set +x
  KDC_DB_KEY=$(decrypt_with_kms_key "${KDC_DB_KEY_URI}" "${KMS_KEY_URI}")
  if [[ -z "${KDC_DB_KEY}" ]]; then
    log_and_fail "kerberos" "KDC database key cannot be empty!" "${EXIT_CODE_CLIENT_ERROR}"
  fi
  set -x
fi

KEYSTORE_URI=$(get_dataproc_property kerberos.keystore.uri)
TRUSTSTORE_URI=$(get_dataproc_property kerberos.truststore.uri)
KEYSTORE_PASSWORD_URI=$(get_dataproc_property kerberos.keystore.password.uri)
KEY_PASSWORD_URI=$(get_dataproc_property kerberos.key.password.uri)
TRUSTSTORE_PASSWORD_URI=$(get_dataproc_property kerberos.truststore.password.uri)
# If none of the 5 properties are set
if [[ -z "${KEYSTORE_URI}" && -z "${TRUSTSTORE_URI}" && -z "${KEYSTORE_PASSWORD_URI}" && -z "${KEY_PASSWORD_URI}" && -z "${TRUSTSTORE_PASSWORD_URI}" ]]; then
  echo "User does not provide SSL keystore/truststore related properties. Will use self signed certificate."
  set +x
  # TODO: this solves the problem of keeping the passwords consistent on all nodes (so no need to distribute them)
  # but the passwords are not random. We need to think about a more secure way of generating them.
  KEYSTORE_PASSWORD=$(sha256sum_10 "${KDC_DB_KEY}")
  KEY_PASSWORD=$(sha256sum_10 "${KDC_DB_KEY}")
  TRUSTSTORE_PASSWORD=$(sha256sum_10 "${KDC_DB_KEY}")
  set -x
# If some of the 5 properties, but not all of them, are set
elif [[ -z "${KEYSTORE_URI}" || -z "${TRUSTSTORE_URI}" || -z "${KEYSTORE_PASSWORD_URI}" || -z "${KEY_PASSWORD_URI}" || -z "${TRUSTSTORE_PASSWORD_URI}" ]]; then
  log_and_fail "kerberos" "The SSL keystore/truststore related properties must be all specified or all left blank." "${EXIT_CODE_CLIENT_ERROR}"
# If all 3 properties are set (user provides their own certificate)
else
  echo 'Fetching and decrypting KEYSTORE_PASSWORD'
  set +x
  KEYSTORE_PASSWORD=$(decrypt_with_kms_key "${KEYSTORE_PASSWORD_URI}" "${KMS_KEY_URI}")
  if [[ -z "${KEYSTORE_PASSWORD}" ]]; then
    log_and_fail "kerberos" "Keystore password cannot be empty!" "${EXIT_CODE_CLIENT_ERROR}"
  fi
  set -x
  echo 'Fetching and decrypting KEY_PASSWORD'
  set +x
  KEY_PASSWORD=$(decrypt_with_kms_key "${KEY_PASSWORD_URI}" "${KMS_KEY_URI}")
  if [[ -z "${KEY_PASSWORD}" ]]; then
    log_and_fail "kerberos" "Key password cannot be empty!" "${EXIT_CODE_CLIENT_ERROR}"
  fi
  set -x
  echo 'Fetching and decrypting TRUSTSTORE_PASSWORD'
  set +x
  TRUSTSTORE_PASSWORD=$(decrypt_with_kms_key "${TRUSTSTORE_PASSWORD_URI}" "${KMS_KEY_URI}")
  if [[ -z "${TRUSTSTORE_PASSWORD}" ]]; then
    log_and_fail "kerberos" "Truststore password cannot be empty!" "${EXIT_CODE_CLIENT_ERROR}"
  fi
  set -x
fi

CROSS_REALM_TRUST_REALM=$(get_dataproc_property kerberos.cross-realm-trust.realm)
if [[ -n "${CROSS_REALM_TRUST_REALM}" ]]; then
  CROSS_REALM_TRUST_KDC=$(get_dataproc_property kerberos.cross-realm-trust.kdc)
  if [[ -z "${CROSS_REALM_TRUST_KDC}" ]]; then
    log_and_fail "kerberos" "Cross-realm trust realm is specified but not the KDC address." "${EXIT_CODE_CLIENT_ERROR}"
  fi
  CROSS_REALM_TRUST_ADMIN_SERVER=$(get_dataproc_property kerberos.cross-realm-trust.admin-server)
  CROSS_REALM_TRUST_PASSWORD_URI=$(get_dataproc_property kerberos.cross-realm-trust.shared-password.uri)
  if [[ -z "${CROSS_REALM_TRUST_PASSWORD_URI}" ]]; then
    log_and_fail "kerberos" "Cross-realm trust realm is specified but not the shared trust password URI." "${EXIT_CODE_CLIENT_ERROR}"
  fi
  set +x
  CROSS_REALM_TRUST_PASSWORD=$(decrypt_with_kms_key "${CROSS_REALM_TRUST_PASSWORD_URI}" "${KMS_KEY_URI}")
  if [[ -z "${CROSS_REALM_TRUST_PASSWORD}" ]]; then
    log_and_fail "kerberos" "Cross-realm trust shared password cannot be empty!" "${EXIT_CODE_CLIENT_ERROR}"
  fi
  set -x
fi

TGT_LIFETIME_HOURS="$(get_dataproc_property kerberos.tgt.lifetime.hours)"
if [[ -z "${TGT_LIFETIME_HOURS}" ]]; then
  TGT_LIFETIME_HOURS=10
fi

if ! is_version_at_least "${DATAPROC_IMAGE_VERSION}" "2.1"; then
  cp "${JAVA_HOME}"/jre/lib/security/policy/unlimited/*.jar "${JAVA_HOME}"/jre/lib/security
fi

function krb5_configuration() {
  if [[ -n "${CROSS_REALM_TRUST_REALM}" ]]; then
    CROSS_REALM_TRUST_ADMIN_SERVER_FULL=""
    if [[ -n "${CROSS_REALM_TRUST_ADMIN_SERVER}" ]]; then
      CROSS_REALM_TRUST_ADMIN_SERVER_FULL="\t\tadmin_server = ${CROSS_REALM_TRUST_ADMIN_SERVER}\n"
    fi
    echo "Configuring cross-realm trust to ${CROSS_REALM_TRUST_REALM}"
    sed -i "/\[realms\]/a\ \t${CROSS_REALM_TRUST_REALM} = {\n\t\tkdc = ${CROSS_REALM_TRUST_KDC}\n${CROSS_REALM_TRUST_ADMIN_SERVER_FULL}\t}" "${KERBEROS_USER_CONFIG}"
    sed -i "/\[domain_realm\]/a\ \t${CROSS_REALM_TRUST_KDC} = ${CROSS_REALM_TRUST_REALM}" "${KERBEROS_USER_CONFIG}"
  fi
  echo "Configuring krb5 to add local realm in ${KERBEROS_USER_CONFIG}."
  additional_kdcs=""
  if (("${MASTER_COUNT}" > 1)); then
    for additional_master in "${ADDITIONAL_MASTERS[@]}"; do
      additional_kdcs+="\t\tkdc = ${additional_master}.${DOMAIN}\n"
    done
  fi

  # Rocky Linux has an extra config directory for libdefaults configs,
  # exclude those configs. No-op for Debian/Ubuntu.
  sed -i '/includedir/d' "${KERBEROS_USER_CONFIG}"
  # Rocky Linux has a default_ccache_name property set by default,
  # delete it if exists. No-op for Debian/Ubuntu.
  sed -i '/default_ccache_name/d' "${KERBEROS_USER_CONFIG}"
  # Rocky Linux has a default_realm config commented out after install,
  # delete it if exists. No-op for Debian/Ubuntu.
  sed -i '/default_realm/d' "${KERBEROS_USER_CONFIG}"
  # Configure default_ccache_name to associate with user uid
  sed -i "/\[libdefaults\]/a\ \tdefault_ccache_name = FILE:\/tmp\/krb5cc_%{uid}" "${KERBEROS_USER_CONFIG}"
  # Set ticket lifetime from client side
  sed -i "/\[libdefaults\]/a\ \tticket_lifetime = ${TGT_LIFETIME_HOURS}h\n\trenew_lifetime = 7d" "${KERBEROS_USER_CONFIG}"
  # Make sure default realm is the KDC realm
  sed -i "/\[libdefaults\]/a\ \tdefault_realm = ${REALM}" "${KERBEROS_USER_CONFIG}"
  sed -i "/\[realms\]/a\ \t${REALM} = {\n\t\tkdc = ${MASTER_FQDN}\n${additional_kdcs}\t\tadmin_server = ${MASTER_FQDN}\n\t}" "${KERBEROS_USER_CONFIG}"
  sed -i "/\[domain_realm\]/a\ \t\.${DOMAIN} = ${REALM}" "${KERBEROS_USER_CONFIG}"

  if [[ "${ROLE}" == 'Master' ]]; then
    echo 'Configure KDC'
    local supported_enctypes
    if is_version_at_least "${DATAPROC_IMAGE_VERSION}" "2.1"; then
      supported_enctypes="aes256-cts:normal arcfour-hmac:normal"
    else
      supported_enctypes="aes256-cts:normal arcfour-hmac:normal des3-hmac-sha1:normal des-cbc-crc:normal des:normal des:v4 des:norealm des:onlyrealm des:afs3"
    fi
    cat >"${KDC_CONF_DIR}/kdc.conf" <<EOF
[kdcdefaults]
    kdc_ports = 750,88

[realms]
    ${REALM} = {
        database_name = ${KERBEROS_DATABASE_DIR}/principal
        admin_keytab = FILE:${KDC_CONF_DIR}/kadm5.keytab
        acl_file = ${KDC_CONF_DIR}/kadm5.acl
        key_stash_file = ${KDC_CONF_DIR}/stash
        kdc_ports = 750,88
        max_life = ${TGT_LIFETIME_HOURS}h 0m 0s
        max_renewable_life = 7d 0h 0m 0s
        master_key_type = aes256-cts
        supported_enctypes = ${supported_enctypes}
        default_principal_flags = +preauth
    }
EOF
  fi
}

function start_master_kdc() {
  set +x
  /usr/sbin/kdb5_util create -s -W -P "${KDC_DB_KEY}"
  set -x
  cat >>"${KDC_CONF_DIR}/kadm5.acl" <<EOF
root *
EOF
  systemctl enable "${KDC_SERVICE}"
  retry_constant_short systemctl restart "${KDC_SERVICE}" || log_and_fail "kerberos" 'Cannot restart KDC'
  systemctl enable "${KADMIN_SERVICE}"
  retry_constant_short systemctl restart "${KADMIN_SERVICE}" || log_and_fail "kerberos" 'Cannot restart Kerberos admin server'

  set +x
  /usr/sbin/kadmin.local -q "addprinc -pw \"${ROOT_PRINCIPAL_PASSWORD}\" root"
  set -x
}

function create_cross_realm_trust_principal() {
  set +x
  # Disable pre-auth with cross-realm trust, as recommended by:
  # https://web.mit.edu/kerberos/krb5-1.5/krb5-1.5.4/doc/krb5-admin/Cross_002drealm-Authentication.html
  /usr/sbin/kadmin.local -q "addprinc -requires_preauth -pw \"${CROSS_REALM_TRUST_PASSWORD}\" krbtgt/${REALM}@${CROSS_REALM_TRUST_REALM}"
  set -x
}

function create_service_principals() {
  if [[ "${FQDN}" != "${MASTER_FQDN}" ]]; then
    local server_started=0
    for ((i = 0; i < 5 * 60; i++)); do
      echo "Waiting for KDC and root principal."
      set +x
      if kadmin -p root -w "${ROOT_PRINCIPAL_PASSWORD}" -q "getprinc root"; then
        server_started=1
        break
      fi
      set -x
      sleep 1
    done

    if (("${server_started}" == 0)); then
      log_and_fail "kerberos" 'Kerberos server has not started even after 5 minutes, likely failed.'
    fi
  fi

  if [[ "${MASTER_COUNT}" -gt 1 && "${ROLE}" == 'Master' ]]; then
    echo "Creating host principal on ${FQDN}"
    set +x
    retry_quiet \
      kadmin -p root -w "${ROOT_PRINCIPAL_PASSWORD}" -q "addprinc -randkey host/${FQDN}"
    retry_quiet \
      kadmin -p root -w "${ROOT_PRINCIPAL_PASSWORD}" -q "ktadd host/${FQDN}"
    set -x
  fi

  echo "Creating service principals on ${FQDN}"
  set +x
  # principals: hdfs/<FQDN>, yarn/<FQDN>, mapred/<FQDN> and HTTP/<FQDN>
  retry_quiet \
    kadmin -p root -w "${ROOT_PRINCIPAL_PASSWORD}" -q "addprinc -randkey hdfs/${FQDN}"
  retry_quiet \
    kadmin -p root -w "${ROOT_PRINCIPAL_PASSWORD}" -q "addprinc -randkey yarn/${FQDN}"
  retry_quiet \
    kadmin -p root -w "${ROOT_PRINCIPAL_PASSWORD}" -q "addprinc -randkey mapred/${FQDN}"
  retry_quiet \
    kadmin -p root -w "${ROOT_PRINCIPAL_PASSWORD}" -q "addprinc -randkey HTTP/${FQDN}"

  # keytabs
  retry_quiet \
    kadmin -p root -w "${ROOT_PRINCIPAL_PASSWORD}" -q "ktadd -k hdfs.service.keytab hdfs/${FQDN} HTTP/${FQDN}"
  retry_quiet \
    kadmin -p root -w "${ROOT_PRINCIPAL_PASSWORD}" -q "ktadd -k yarn.service.keytab yarn/${FQDN} HTTP/${FQDN}"
  retry_quiet \
    kadmin -p root -w "${ROOT_PRINCIPAL_PASSWORD}" -q "ktadd -k mapred.service.keytab mapred/${FQDN} HTTP/${FQDN}"
  retry_quiet \
    kadmin -p root -w "${ROOT_PRINCIPAL_PASSWORD}" -q "ktadd -k http.service.keytab HTTP/${FQDN}"
  set -x
  # Move keytabs and set permissions on them
  mv *.keytab "${KEYTAB_DIR}"
  chown hdfs:hdfs "${KEYTAB_DIR}"/hdfs.service.keytab
  chown yarn:yarn "${KEYTAB_DIR}"/yarn.service.keytab
  chown mapred:mapred "${KEYTAB_DIR}"/mapred.service.keytab
  chown hdfs:hadoop "${KEYTAB_DIR}"/http.service.keytab
  chmod 400 "${KEYTAB_DIR}"/*.keytab
  # Allow group 'hadoop' to read the http.service.keytab file.
  chmod 440 "${KEYTAB_DIR}"/http.service.keytab
}

function config_krb5_propagate() {
  local -r service_name="krb5_prop"
  local -r exec_start="/usr/sbin/kpropd"

  cat >>/etc/services <<EOF
${service_name} 754/tcp
EOF

  # Rockylinux 9 have removed xinetd. So configuring as systemd service in rocky 9
  if is_rocky9; then
    cat >"/etc/systemd/system/${service_name}.service" <<EOF
[Unit]
Description=${service_name}
Wants=network-online.target
After=syslog.target network-online.target

[Service]
User=root
Group=root
Type=forking
ExecStart=${exec_start}
Restart=on-failure
OOMScoreAdjust=0

[Install]
WantedBy=multi-user.target
EOF

    mkdir -p /var/run/${service_name}
    chown root:root /var/run/${service_name}
    chmod 775 /var/run/${service_name}

    retry_constant_short systemctl daemon-reload
    retry_constant_short systemctl enable "${service_name}"
    retry_constant_short systemctl start "${service_name}"
  else
    # Using xinet for other OS.
    cat >/etc/xinetd.d/krb5_prop <<EOF
service ${service_name}
{
        disable         = no
        socket_type     = stream
        protocol        = tcp
        user            = root
        wait            = no
        server          = ${exec_start}
}
EOF
    systemctl enable xinetd
    systemctl restart xinetd
  fi

  cat >>"${KDC_CONF_DIR}/kpropd.acl" <<EOF
host/${INSTANCE_NAME_PREFIX}-m-0.${DOMAIN}@${REALM}
host/${INSTANCE_NAME_PREFIX}-m-1.${DOMAIN}@${REALM}
host/${INSTANCE_NAME_PREFIX}-m-2.${DOMAIN}@${REALM}
EOF
  if [[ "${FQDN}" != "${MASTER_FQDN}" ]]; then
    local krb5_prop_finished=0
    for ((i = 0; i < 5 * 60; i++)); do
      echo "Waiting for initial KDC database propagation to finish..."
      if [[ -f "${KERBEROS_DATABASE_DIR}/principal" \
        && -f "${KERBEROS_DATABASE_DIR}/principal.ok" \
        && -f "${KERBEROS_DATABASE_DIR}/principal.kadm5" \
        && -f "${KERBEROS_DATABASE_DIR}/principal.kadm5.lock" ]]; then
        krb5_prop_finished=1
        break
      fi
      sleep 1
    done
    if (("${krb5_prop_finished}" == 0)); then
      loginfo "Failed waiting for propagation."
      local -r files="$(ls "${KERBEROS_DATABASE_DIR}" | tr '\n' ' ')"
      loginfo "Files under ${KERBEROS_DATABASE_DIR}: ${files}"
      log_and_fail "kerberos" 'Initial KDC database propagation did not finish after 5 minutes, likely failed.'
    fi
    set +x
    /usr/sbin/kdb5_util stash -m -P "${KDC_DB_KEY}"
    set -x
    retry_constant_short systemctl restart "${KDC_SERVICE}" || log_and_fail "kerberos" 'Cannot restart KDC'
    retry_constant_short systemctl restart "${KADMIN_SERVICE}" || log_and_fail "kerberos" 'Cannot restart Kerberos admin server'
  else
    cat >/etc/krb5_prop.sh <<EOF
#!/bin/bash
ADDITIONAL_KDCS=(\$(/usr/share/google/get_metadata_value attributes/dataproc-master-additional | sed 's/,/\n/g'))
/usr/sbin/kdb5_util dump "${KDC_CONF_DIR}/kdc_datatrans"
for kdc_host in "\${ADDITIONAL_KDCS[@]}"; do
  /usr/sbin/kprop -f "${KDC_CONF_DIR}/kdc_datatrans" "\${kdc_host}.${DOMAIN}"
done
EOF
    chmod +x /etc/krb5_prop.sh
    (
      crontab -l 2>/dev/null || true
      echo "*/5 * * * * /etc/krb5_prop.sh >/dev/null 2>&1"
    ) | crontab -
    retry_constant /etc/krb5_prop.sh
  fi
}

# TODO: move setting of properties to frontend where possible.
function config_core_site() {
  echo "Setting properties in core-site.xml."
  set_property_core_site 'hadoop.security.authentication' 'kerberos'
  set_property_core_site 'hadoop.security.authorization' 'true'
  set_property_core_site 'hadoop.rpc.protection' 'privacy'
  set_property_core_site 'hadoop.ssl.enabled' 'true'
  set_property_core_site 'hadoop.ssl.require.client.cert' 'false'
  set_property_core_site 'hadoop.ssl.hostname.verifier' 'DEFAULT'
  set_property_core_site 'hadoop.ssl.keystores.factory.class' 'org.apache.hadoop.security.ssl.FileBasedKeyStoresFactory'
  set_property_core_site 'hadoop.ssl.server.conf' 'ssl-server.xml'
  set_property_core_site 'hadoop.ssl.client.conf' 'ssl-client.xml'

  # Required from Hadoop 3.3.0
  if is_version_at_least "${DATAPROC_IMAGE_VERSION}" "2.1"; then
    set_property_core_site "hadoop.http.authentication.type" "kerberos"
    set_property_core_site "hadoop.http.authentication.kerberos.principal" "HTTP/_HOST@${REALM}"
    set_property_core_site "hadoop.http.authentication.kerberos.keytab" "${KEYTAB_DIR}/http.service.keytab"
  fi

  # This will map all principals from all realms to the first part of the
  # principal name.
  # eg: both 'my-name/master.fqdn@MY.REALM' and 'my-name@MY.REALM' will be mapped to
  # 'my-name'.
  if [[ -n "${CROSS_REALM_TRUST_REALM}" ]]; then
    auth_to_local_rules=$(
      cat <<'EOF'
    RULE:[1:$1](.*)s/(.*)/$1/g
    RULE:[2:$1](.*)s/(.*)/$1/g
    DEFAULT
EOF
    )
    set_property_core_site 'hadoop.security.auth_to_local' "${auth_to_local_rules}"
  fi

  if is_component_selected 'hive-webhcat-server'; then
    set_property_core_site 'hadoop.proxyuser.HTTP.hosts' '*'
    set_property_core_site 'hadoop.proxyuser.HTTP.groups' '*'
  fi

  if is_component_selected 'knox'; then
    set_property_core_site 'hadoop.proxyuser.knox.hosts' '*'
    set_property_core_site 'hadoop.proxyuser.knox.groups' '*'
  fi
}

function config_hdfs_site() {
  echo "Setting properties in hdfs-site.xml."
  set_property_hdfs_site 'dfs.block.access.token.enable' 'true'
  set_property_hdfs_site 'dfs.encrypt.data.transfer' 'true'

  if is_version_at_least "${DATAPROC_IMAGE_VERSION}" "2.1"; then
    set_property_hdfs_site 'dfs.encrypt.data.transfer.cipher.suites' 'AES/CTR/NoPadding'
  fi

  # Name Node
  set_property_hdfs_site 'dfs.namenode.keytab.file' "${KEYTAB_DIR}/hdfs.service.keytab"
  set_property_hdfs_site 'dfs.namenode.kerberos.principal' "hdfs/_HOST@${REALM}"
  set_property_hdfs_site 'dfs.namenode.kerberos.internal.spnego.principal' "HTTP/_HOST@${REALM}"
  # Secondary Name Node
  set_property_hdfs_site 'dfs.secondary.namenode.keytab.file' "${KEYTAB_DIR}/hdfs.service.keytab"
  set_property_hdfs_site 'dfs.secondary.namenode.kerberos.principal' "hdfs/_HOST@${REALM}"
  set_property_hdfs_site 'dfs.secondary.namenode.kerberos.internal.spnego.principal' "HTTP/_HOST@${REALM}"
  # Journal Node
  set_property_hdfs_site 'dfs.journalnode.keytab.file' "${KEYTAB_DIR}/hdfs.service.keytab"
  set_property_hdfs_site 'dfs.journalnode.kerberos.principal' "hdfs/_HOST@${REALM}"
  set_property_hdfs_site 'dfs.journalnode.kerberos.internal.spnego.principal' "HTTP/_HOST@${REALM}"
  # Data Node
  # 'dfs.data.transfer.protection' Enable SASL, only supported after version 2.6
  set_property_hdfs_site 'dfs.data.transfer.protection' 'privacy'
  set_property_hdfs_site 'dfs.datanode.keytab.file' "${KEYTAB_DIR}/hdfs.service.keytab"
  set_property_hdfs_site 'dfs.datanode.kerberos.principal' "hdfs/_HOST@${REALM}"
  # WebHDFS
  set_property_hdfs_site 'dfs.webhdfs.enabled' 'true'
  set_property_hdfs_site 'dfs.web.authentication.kerberos.principal' "HTTP/_HOST@${REALM}"
  set_property_hdfs_site 'dfs.web.authentication.kerberos.keytab' "${KEYTAB_DIR}/http.service.keytab"
  # HTTPS
  set_property_hdfs_site 'dfs.http.policy' 'HTTPS_ONLY'
}

function config_yarn_site_and_permission() {
  echo "Setting properties in yarn-site.xml."
  # Resource Manager
  set_property_yarn_site 'yarn.resourcemanager.keytab' "${KEYTAB_DIR}/yarn.service.keytab"
  set_property_yarn_site 'yarn.resourcemanager.principal' "yarn/_HOST@${REALM}"
  # Node Manager
  set_property_yarn_site 'yarn.nodemanager.keytab' "${KEYTAB_DIR}/yarn.service.keytab"
  set_property_yarn_site 'yarn.nodemanager.principal' "yarn/_HOST@${REALM}"
  if (("${MASTER_COUNT}" > 1)); then
    set_property_yarn_site 'yarn.resourcemanager.webapp.https.address.rm0' "${INSTANCE_NAME_PREFIX}-m-0.${DOMAIN}:8090"
    set_property_yarn_site 'yarn.resourcemanager.webapp.https.address.rm1' "${INSTANCE_NAME_PREFIX}-m-1.${DOMAIN}:8090"
    set_property_yarn_site 'yarn.resourcemanager.webapp.https.address.rm2' "${INSTANCE_NAME_PREFIX}-m-2.${DOMAIN}:8090"
  fi
  # Containers need to be run as the users that submitted the job.
  # See https://www.ibm.com/support/knowledgecenter/en/SSPT3X_4.2.0/com.ibm.swg.im.infosphere.biginsights.install.doc/doc/inst_adv_yarn_config.html
  set_property_yarn_site 'yarn.nodemanager.container-executor.class' 'org.apache.hadoop.yarn.server.nodemanager.LinuxContainerExecutor'
  set_property_yarn_site 'yarn.nodemanager.linux-container-executor.group' 'hadoop'
  # The 'yarn.nodemanager.local-dirs' value should always be present.
  # But if for any reason at all it's missing, we always fall back to
  # the default value '/hadoop/yarn/nm-local-dir'.
  local nm_local_dirs_str
  nm_local_dirs_str=$(get_property_in_xml "${HADOOP_CONF_DIR}/yarn-site.xml" \
    yarn.nodemanager.local-dirs /hadoop/yarn/nm-local-dir)
  # This split assumes the str may contains more than one dirs separated by ','
  local nm_local_dirs=(${nm_local_dirs_str//,/ })
  chown yarn "${nm_local_dirs[@]/%//}"
  # HTTPS
  set_property_yarn_site 'yarn.http.policy' 'HTTPS_ONLY'
  set_property_yarn_site 'yarn.log.server.url' "https://${MASTER_FQDN}:19889/jobhistory/logs"
  # YARN App Timeline Server
  set_property_yarn_site 'yarn.timeline-service.http-authentication.type' 'kerberos'
  set_property_yarn_site 'yarn.timeline-service.principal' "yarn/_HOST@${REALM}"
  set_property_yarn_site 'yarn.timeline-service.keytab' "${KEYTAB_DIR}/yarn.service.keytab"
  set_property_yarn_site 'yarn.timeline-service.http-authentication.kerberos.principal' "HTTP/_HOST@${REALM}"
  set_property_yarn_site 'yarn.timeline-service.http-authentication.kerberos.keytab' "${KEYTAB_DIR}/http.service.keytab"
  cat >"${HADOOP_CONF_DIR}"/container-executor.cfg <<EOF
yarn.nodemanager.linux-container-executor.group=hadoop
banned.users=hdfs,yarn,mapred,bin
min.user.id=1000
EOF
  # We need to allow Zeppelin to run containers as a system user
  # Do this here instead of activate/zeppelin.sh since the Zeppelin startup
  # script runs only on the master, while the Kerberos startup script runs
  # on all nodes
  if is_component_selected 'zeppelin'; then
    echo "allowed.system.users=zeppelin" >>"${HADOOP_CONF_DIR}/container-executor.cfg"
  fi

  # container-executor need same owner and permission.
  chown root:hadoop /usr/lib/hadoop-yarn/bin/container-executor
  chmod 6050 /usr/lib/hadoop-yarn/bin/container-executor
}

function config_mapred_site() {
  echo "Setting properties in mapred-site.xml."
  set_property_mapred_site 'mapreduce.jobhistory.keytab' "${KEYTAB_DIR}/mapred.service.keytab"
  set_property_mapred_site 'mapreduce.jobhistory.principal' "mapred/_HOST@${REALM}"
  set_property_mapred_site 'mapreduce.jobhistory.http.policy' 'HTTPS_ONLY'
  set_property_mapred_site 'mapreduce.ssl.enabled' 'true'
  set_property_mapred_site 'mapreduce.shuffle.ssl.enabled' 'true'
  set_property_mapred_site 'mapreduce.jobhistory.address' "${MASTER_FQDN}:10020"
  set_property_mapred_site 'mapreduce.jobhistory.webapp.http.address' "${MASTER_FQDN}:19888"
  set_property_mapred_site 'mapreduce.jobhistory.webapp.https.address' "${MASTER_FQDN}:19889"
}

function copy_or_create_keystore_files() {
  mkdir "${HADOOP_CONF_DIR}"/ssl
  if [[ -n "${KEYSTORE_URI}" ]]; then
    echo "Copying keystore and truststore files from GCS."
    gsutil cp "${KEYSTORE_URI}" "${HADOOP_CONF_DIR}"/ssl/keystore.jks
    gsutil cp "${TRUSTSTORE_URI}" "${HADOOP_CONF_DIR}"/ssl/truststore.jks
  else
    local keystore_location="${STAGING_LOCATION}/keystores"
    if [[ "${FQDN}" == "${MASTER_FQDN}" ]]; then
      echo "Generating self-signed certificate from ${FQDN}"
      local -r cert_conf_file=/tmp/cert.conf
      local -r private_key="private.key"
      local -r cert="dataproc.cert"
      local -r p12_key="dataproc.p12"
      # Openssl allows only 64 chars in the CN field. When WILDCARD is longer
      # than find the subdomain string that is shorter than 64
      local -r wildcard="*.${DOMAIN}"
      local cn
      if [[ ${#wildcard} -lt 65 ]]; then
        cn=${wildcard}
      else
        cn=$(get_short_cn_wildcard)
      fi
        cat >>"${cert_conf_file}" <<EOF
WILDCARD = ${wildcard}
SAN = DNS:\$WILDCARD

[ req ]
default_bits = 2048
prompt = no
distinguished_name = req_distinguished_name
x509_extensions = v3_req

[ req_distinguished_name ]
CN = ${cn}

[ v3_req ]
subjectAltName = \$SAN
EOF
      openssl req -newkey rsa:2048 -x509 -keyout "${private_key}" -out "${cert}" \
        -days 1800 -nodes -config "${cert_conf_file}"
      set +x
      openssl pkcs12 -export -in "${cert}" -inkey "${private_key}" \
        -out "${p12_key}" -name dataproc-cert -password "pass:${KEYSTORE_PASSWORD}"
      keytool -importkeystore -destkeystore keystore.jks -deststorepass "${KEYSTORE_PASSWORD}" \
        -srckeystore "${p12_key}" -srcstoretype PKCS12 -srcstorepass "${KEYSTORE_PASSWORD}"
      keytool -importcert -keystore truststore.jks -alias dataproc-cert \
        -storepass "${TRUSTSTORE_PASSWORD}" -file "${cert}" -noprompt
      if is_component_selected 'knox'; then
        local cacerts_location
        if is_version_at_least "${DATAPROC_IMAGE_VERSION}" "2.1"; then
          cacerts_location="${JAVA_HOME}/lib/security/cacerts"
        else
          cacerts_location="${JAVA_HOME}/jre/lib/security/cacerts"
        fi
        keytool -importcert -keystore "${cacerts_location}" -alias dataproc-cert -storepass changeit -file dataproc.cert -noprompt
      fi
      set -x
      rm "${cert}" "${p12_key}" "${private_key}" "${cert_conf_file}"
      gsutil cp keystore.jks truststore.jks "gs://${keystore_location}"
      mv keystore.jks "${HADOOP_CONF_DIR}"/ssl/
      mv truststore.jks "${HADOOP_CONF_DIR}"/ssl/
    else
      local files_downloaded=0
      for ((i = 0; i < 5 * 60; i++)); do
        echo "Waiting for keystore and truststore files."
        if gsutil -q stat "gs://${keystore_location}/keystore.jks" && gsutil -q stat "gs://${keystore_location}/truststore.jks"; then
          gsutil cp "gs://${keystore_location}/keystore.jks" "${HADOOP_CONF_DIR}"/ssl/keystore.jks
          gsutil cp "gs://${keystore_location}/truststore.jks" "${HADOOP_CONF_DIR}"/ssl/truststore.jks
          files_downloaded=1
          break
        fi
        sleep 1
      done
      if (("${files_downloaded}" == 0)); then
        log_and_fail "kerberos" 'Failed to download keystore and truststore files after 5 minutes.'
      fi
    fi
  fi
  # LINT.IfChange
  # Create the pem file for use by non-jks compatible API usage.
  # Temporarily go through pkcs12 format so that openssl can
  # generate the pem format.
  set +x
  keytool -importkeystore \
    -srckeystore "${HADOOP_CONF_DIR}/ssl/truststore.jks" \
    -srcstorepass "${TRUSTSTORE_PASSWORD}" \
    -destkeystore truststore.p12 \
    -deststoretype pkcs12 \
    -deststorepass "${TRUSTSTORE_PASSWORD}"
  # Use sed to strip out "Bag Attributes" which apparently
  # aren't well supported by keytool or Ruby.
  openssl pkcs12 -nokeys -in truststore.p12 \
    -password "pass:${TRUSTSTORE_PASSWORD}" |
    sed -ne '/-BEGIN CERTIFICATE-/,/-END CERTIFICATE-/p' >"${HADOOP_CONF_DIR}/ssl/truststore.pem"
  rm truststore.p12
  set -x

  chown -R yarn:hadoop "${HADOOP_CONF_DIR}"/ssl
  chmod 755 "${HADOOP_CONF_DIR}"/ssl
  chmod 440 "${HADOOP_CONF_DIR}"/ssl/keystore.jks
  chmod 444 "${HADOOP_CONF_DIR}"/ssl/truststore.jks
  chmod 444 "${HADOOP_CONF_DIR}"/ssl/truststore.pem
}

function set_property_ssl_xml() {
  set_property_in_xml "${HADOOP_CONF_DIR}/ssl-$1.xml" "$2" "$3"
}

function config_ssl_xml() {
  echo "Setting properties in ssl-$1.xml."
  cp "${HADOOP_CONF_DIR}/ssl-$1.xml.example" "${HADOOP_CONF_DIR}/ssl-$1.xml"
  set +x
  set_property_ssl_xml "$1" "ssl.$1.truststore.location" "${HADOOP_CONF_DIR}/ssl/truststore.jks"
  set_property_ssl_xml "$1" "ssl.$1.truststore.password" "${TRUSTSTORE_PASSWORD}"
  set_property_ssl_xml "$1" "ssl.$1.truststore.type" "jks"
  if [[ $1 == 'server' ]]; then
    set_property_ssl_xml "$1" "ssl.$1.keystore.location" "${HADOOP_CONF_DIR}/ssl/keystore.jks"
    set_property_ssl_xml "$1" "ssl.$1.keystore.password" "${KEYSTORE_PASSWORD}"
    set_property_ssl_xml "$1" "ssl.$1.keystore.keypassword" "${KEY_PASSWORD}"
    set_property_ssl_xml "$1" "ssl.$1.keystore.type" "jks"
  fi
  set -x

  if [[ $1 == 'client' ]]; then
    chown yarn:hadoop "${HADOOP_CONF_DIR}/ssl-$1.xml"
    chmod 444 "${HADOOP_CONF_DIR}/ssl-$1.xml"
  else
    chown mapred:hadoop "${HADOOP_CONF_DIR}/ssl-$1.xml"
    chmod 440 "${HADOOP_CONF_DIR}/ssl-$1.xml"
  fi
}

function secure_hive() {
  echo "Securing Hive."
  set +x
  retry_quiet \
    kadmin -p root -w "${ROOT_PRINCIPAL_PASSWORD}" -q "addprinc -randkey hive/${FQDN}"
  retry_quiet \
    kadmin -p root -w "${ROOT_PRINCIPAL_PASSWORD}" -q "ktadd -k hive.service.keytab hive/${FQDN}"
  set -x
  mv ./hive.service.keytab "${KEYTAB_DIR}"
  chown hive:hive "${KEYTAB_DIR}"/hive.service.keytab
  chmod 400 "${KEYTAB_DIR}"/hive.service.keytab
  # Add the user 'hive' to group 'hadoop' so it can access the keystore files
  usermod -a -G hadoop hive
  # HiveServer2
  set_property_hive_site 'hive.server2.authentication' 'KERBEROS'
  set_property_hive_site 'hive.server2.authentication.kerberos.principal' "hive/_HOST@${REALM}"
  set_property_hive_site 'hive.server2.authentication.kerberos.keytab' "${KEYTAB_DIR}/hive.service.keytab"
  set_property_hive_site 'hive.server2.thrift.sasl.qop' 'auth-conf'
  # See https://lists.apache.org/thread/x8334lwynf84qw9bqygsmfmt10k5dpjq
  # Enabling FallbackHiveAuthorizerFactory does the following:
  # 1. Disallow local file location in sql statements except for admin
  # 2. Allow "set" only selected whitelist parameters
  # 3. Disallow dfs commands except for admin
  # 4. Disallow "ADD JAR" statement
  # 5. Disallow "COMPILE" statement
  # 6. Disallow "TRANSFORM" statement
  set_property_hive_site 'hive.security.authorization.enabled' 'true'
  set_property_hive_site 'hive.security.authorization.manager' 'org.apache.hadoop.hive.ql.security.authorization.plugin.fallback.FallbackHiveAuthorizerFactory'
  # Allow setting tags since Dataproc Agent tags YARN apps with job id.
  # These are appended to the hive.security.authorization.sqlstd.confwhitelist
  set_property_hive_site 'hive.security.authorization.sqlstd.confwhitelist.append' 'tez.application.tags'
  # Hive Metastore
  set_property_hive_site 'hive.metastore.sasl.enabled' 'true'
  set_property_hive_site 'hive.metastore.kerberos.principal' "hive/_HOST@${REALM}"
  set_property_hive_site 'hive.metastore.kerberos.keytab.file' "${KEYTAB_DIR}/hive.service.keytab"
  # WebUI SSL
  set_property_hive_site 'hive.server2.webui.use.ssl' 'true'
  set_property_hive_site 'hive.server2.webui.keystore.path' "${HADOOP_CONF_DIR}/ssl/keystore.jks"
  set +x
  set_property_hive_site 'hive.server2.webui.keystore.password' "${KEYSTORE_PASSWORD}"
  set -x
}

function secure_spark() {
  # Dataproc only runs Spark History Server on 'first' master so we only need
  # to Kerberize it on this master.
  if [[ "${FQDN}" == "${MASTER_FQDN}" ]]; then
    echo "Securing Spark."
    set +x
    retry_quiet \
      kadmin -p root -w "${ROOT_PRINCIPAL_PASSWORD}" -q "addprinc -randkey spark/${FQDN}"
    retry_quiet \
      kadmin -p root -w "${ROOT_PRINCIPAL_PASSWORD}" -q "ktadd -k spark.service.keytab spark/${FQDN}"
    set -x
    mv ./spark.service.keytab "${KEYTAB_DIR}"/spark.service.keytab
    chown spark:spark "${KEYTAB_DIR}"/spark.service.keytab
    chmod 400 "${KEYTAB_DIR}"/spark.service.keytab
    # Add the user 'spark' to group 'hadoop' so it can access the keystore files
    usermod -a -G hadoop spark
    cat >>/etc/spark/conf/spark-env.sh <<EOF
SPARK_HISTORY_OPTS="\$SPARK_HISTORY_OPTS
-Dspark.history.kerberos.enabled=true
-Dspark.history.kerberos.principal=spark/${FQDN}@${REALM}
-Dspark.history.kerberos.keytab=${KEYTAB_DIR}/spark.service.keytab"
EOF
  fi
  # All the nodes receive the configurations.
  sed -i "s/spark.yarn.historyServer.address.*/spark.yarn.historyServer.address https:\/\/${MASTER_FQDN}:18480/" /etc/spark/conf/spark-defaults.conf
  # Disable the unmanaged AM setting.
  sed -i "s/spark.yarn.unmanagedAM.enabled.*/spark.yarn.unmanagedAM.enabled false/" /etc/spark/conf/spark-defaults.conf
  set +x
  if [[ "${DATAPROC_IMAGE_TYPE}" == "standard" ]] &&
      is_version_at_least "${DATAPROC_IMAGE_VERSION}" '2.2'; then
  cat >>/etc/spark/conf/spark-defaults.conf <<EOF
spark.ssl.enabled true
spark.ssl.ui.enabled false
EOF
  fi
  cat >>/etc/spark/conf/spark-defaults.conf <<EOF
spark.ssl.protocol tls
spark.ssl.historyServer.enabled true
spark.ssl.trustStore ${HADOOP_CONF_DIR}/ssl/truststore.jks
spark.ssl.keyStore ${HADOOP_CONF_DIR}/ssl/keystore.jks
spark.ssl.trustStorePassword ${TRUSTSTORE_PASSWORD}
spark.ssl.keyStorePassword ${KEYSTORE_PASSWORD}
spark.hadoop.yarn.timeline-service.enabled false
EOF
  set -x

  if [[ "${DATAPROC_IMAGE_TYPE}" == "standard" ]] &&
      ! is_version_at_least "${DATAPROC_IMAGE_VERSION}" '2.0'; then
    cat >> /etc/spark/conf/spark-defaults.conf << EOF
spark.hadoop.hive.security.authorization.enabled=false
spark.hadoop.hive.security.authorization.manager=org.apache.hadoop.hive.ql.security.authorization.DefaultHiveAuthorizationProvider
EOF
  fi
}

function secure_tez() {
  set_property_tez_site "tez.tez-ui.history-url.base" "https://${MASTER_FQDN}:8190/tez-ui/"
  set_property_tez_site "tez.runtime.shuffle.ssl.enable" "true"
  set_property_tez_site "tez.runtime.shuffle.keep-alive.enabled" "true"
}

function secure_pig() {
  cat >/etc/google-dataproc/pig.kerberos.properties <<EOF
java.security.krb5.conf=${KERBEROS_USER_CONFIG}
hadoop.security.krb5.principal=dataproc/${FQDN}@${REALM}
hadoop.security.krb5.keytab=${KEYTAB_DIR}/dataproc.service.keytab
EOF
}

function secure_zookeeper_server() {
  echo "Securing Zookeeper Server"
  set +x
  retry_quiet \
    kadmin -p root -w "${ROOT_PRINCIPAL_PASSWORD}" -q "addprinc -randkey zookeeper/${FQDN}"
  retry_quiet \
    kadmin -p root -w "${ROOT_PRINCIPAL_PASSWORD}" -q "ktadd -k zookeeper.service.keytab zookeeper/${FQDN}"
  set -x
  mv ./zookeeper.service.keytab "${KEYTAB_DIR}"/zookeeper.service.keytab
  chown zookeeper:zookeeper "${KEYTAB_DIR}"/zookeeper.service.keytab
  chmod 400 "${KEYTAB_DIR}"/zookeeper.service.keytab

  local zookeeper_conf_dir="/etc/zookeeper/conf"
  cat >>"${zookeeper_conf_dir}"/zoo.cfg <<EOF
authProvider.1=org.apache.zookeeper.server.auth.SASLAuthenticationProvider
jaasLoginRenew=3600000
EOF

  cat >"${zookeeper_conf_dir}"/jaas.conf <<EOF
Server {
  com.sun.security.auth.module.Krb5LoginModule required
  useKeyTab=true
  keyTab="${KEYTAB_DIR}/zookeeper.service.keytab"
  storeKey=true
  useTicketCache=false
  principal="zookeeper/${FQDN}@${REALM}";
};
EOF

  # See: https://github.com/apache/zookeeper/blob/master/zookeeper-server/src/main/java/org/apache/zookeeper/server/auth/SaslServerCallbackHandler.java#L42-L43
  # And: https://github.com/apache/zookeeper/blob/release-3.4.14/zookeeper-server/src/main/java/org/apache/zookeeper/server/auth/KerberosName.java#L358
  cat >"${zookeeper_conf_dir}"/java.env <<EOF
export JVMFLAGS="-Djava.security.auth.login.config=${zookeeper_conf_dir}/jaas.conf -Dzookeeper.skipACL=no -Dzookeeper.kerberos.removeHostFromPrincipal=true -Dzookeeper.kerberos.removeRealmFromPrincipal=true"
EOF
}

function create_dataproc_principal() {
  echo "Creating principal dataproc/${FQDN}@${REALM}."
  set +x
  retry_quiet \
    kadmin -p root -w "${ROOT_PRINCIPAL_PASSWORD}" -q "addprinc -randkey dataproc/${FQDN}"
  retry_quiet \
    kadmin -p root -w "${ROOT_PRINCIPAL_PASSWORD}" -q "ktadd -k dataproc.service.keytab dataproc/${FQDN}"
  set -x
  mkdir -p /etc/security/keytab
  mv dataproc.service.keytab "${KEYTAB_DIR}"/dataproc.service.keytab
  chown root:dataproc "${KEYTAB_DIR}"/dataproc.service.keytab
  chmod 440 "${KEYTAB_DIR}"/dataproc.service.keytab
}

function create_user_principals() {
  local -r dataproc_users=$(get_dataproc_property dataproc.users)
  if [[ -n "${dataproc_users}" ]]; then
    local -a user_list
    local username
    local keytab
    IFS=',' read -r -a user_list <<< "${dataproc_users}"
    for pair in "${user_list[@]}"; do
      IFS=':' read -r user sa <<< "${pair}"
      username="$(echo "${user}" | cut -d@ -f1)"
      keytab="/etc/security/keytab/${username}.user.keytab"
      set +x
      retry_quiet \
        kadmin -p root -w "${ROOT_PRINCIPAL_PASSWORD}" -q "addprinc -randkey ${username}/${FQDN}"
      retry_quiet \
        kadmin -p root -w "${ROOT_PRINCIPAL_PASSWORD}" -q "ktadd -k ${keytab} ${username}/${FQDN}"
      set -x
      chmod 400 "${keytab}"
      chown "${username}":"${username}" "${keytab}"
    done
  fi
}

function kinit_as_dataproc_principal() {
  kinit -kt /etc/security/keytab/dataproc.service.keytab "dataproc/${FQDN}@${REALM}"
}


function krb5_propagate() {
  if [[ "${FQDN}" == "${MASTER_FQDN}" && "${MASTER_COUNT}" -gt 1 ]]; then
    /etc/krb5_prop.sh
  fi
}

function create_component_keytab() {
  local -r service="$1"
  local -r keytab_owner="$2"

  local -r keytab="/etc/security/keytab/${service}.service.keytab"
  echo "Creating principal for ${service} and keytab: ${keytab}"
  set +x
  retry_quiet \
    kadmin -p root -w "${ROOT_PRINCIPAL_PASSWORD}" -q "addprinc -randkey ${service}/${FQDN}"
  retry_quiet \
    kadmin -p root -w "${ROOT_PRINCIPAL_PASSWORD}" -q "ktadd -k ${keytab} ${service}/${FQDN}"
  useradd "${keytab_owner}" || true
  chown "${keytab_owner}":"${keytab_owner}" "${keytab}"
  chmod 400 "${keytab}"
  set -x
}

function delete_hands_off_password_file() {
  rm -f "${HANDS_OFF_PASSWORD}"
}

function get_short_cn_wildcard() {
  local sub_domains
  local wildcard
  IFS='.' read -r -a sub_domains <<< "${DOMAIN}"
  wildcard=$(IFS=. ; echo "${sub_domains[*]}")
  # check for 62 chars as *. is added to a wildcard
  while [[ ${#wildcard} -gt 62 ]]; do
    # delete first sub_domain so that we can get to 62 or less chars dns.
    sub_domains=( "${sub_domains[@]:1}" )
    wildcard=$(IFS=. ; echo "${sub_domains[*]}")
  done
  echo "*.${wildcard}"
}

function main() {
  krb5_configuration
  if [[ "${FQDN}" == "${MASTER_FQDN}" ]]; then
    start_master_kdc
    if [[ -n "${CROSS_REALM_TRUST_REALM}" ]]; then
      create_cross_realm_trust_principal
    fi
  fi
  create_service_principals
  if [[ "${ROLE}" == 'Master' && "${MASTER_COUNT}" -gt 1 ]]; then
    config_krb5_propagate
  fi

  config_core_site
  config_hdfs_site
  config_yarn_site_and_permission
  config_mapred_site
  copy_or_create_keystore_files
  config_ssl_xml 'server'
  config_ssl_xml 'client'
  secure_hive
  secure_spark
  secure_tez
  secure_pig

  if [[ "${ROLE}" == 'Master' ]] && is_component_selected 'zookeeper'; then
    secure_zookeeper_server
  fi

  if is_component_selected 'presto'; then
    create_component_keytab "presto" "root"
  fi

  if is_component_selected 'trino'; then
    create_component_keytab "trino" "root"
  fi

  if [[ "${ROLE}" == 'Master' ]] && is_component_selected 'zeppelin'; then
    create_component_keytab "zeppelin" "zeppelin"
  fi

  if [[ "${ROLE}" == 'Master' ]] && is_component_selected 'knox'; then
    create_component_keytab "knox" "knox"
  fi

  if is_component_selected 'druid'; then
    create_component_keytab "druid" "root"
  fi

  if [[ "${ROLE}" == 'Master' ]] && is_component_selected 'ranger'; then
    # principal: rangeradmin
    create_component_keytab "rangeradmin" "ranger"

    # principal: rangerlookup
    create_component_keytab "rangerlookup" "ranger"

    # principal: rangerusersync
    create_component_keytab "rangerusersync" "ranger"
  fi

  if is_component_selected 'hbase'; then
    create_component_keytab "hbase" "hbase"
  fi

  create_dataproc_principal
  create_user_principals
  krb5_propagate

  delete_hands_off_password_file
  kinit_as_dataproc_principal
}

main
