#!/bin/bash

set -euo pipefail

source "$(dirname "$0")/../../bdutil_env.sh"
source "$(dirname "$0")/../../bdutil_logging.sh"
source "$(dirname "$0")/../../bdutil_metadata.sh"
source "$(dirname "$0")/../../bdutil_properties.sh"
source "$(dirname "$0")/../../bdutil_retry.sh"
source "$(dirname "$0")/../../bdutil_versions.sh"
source "$(dirname "$0")/../../bdutil_components.sh"
source "${DATAPROC_DIR}/dataproc_env.sh"

set -x

readonly CLUSTER_PROPS="/tmp/cluster/properties"
readonly KNOX_INSTALL_DIRECTORY="/usr/lib/knox"
FQDN=$(hostname -f)
readonly FQDN
readonly GATEWAY_CONF="${KNOX_INSTALL_DIRECTORY}/conf/gateway-site.xml"
readonly KNOX_SERVICE_INIT_SCRIPT="/usr/lib/systemd/system/knox.service"
readonly IS_COMPONENT_GATEWAY_HA_ENABLED=$(get_dataproc_property_or_default dataproc:componentgateway.ha.enabled false)
TOPOLOGY_TEMPLATE="knox/topology-template.xml"

# use HA template if dataproc HA is enabled.
if (("${MASTER_COUNT}" > 1)); then
    TOPOLOGY_TEMPLATE="knox/topology-template-ha.xml"
fi

function generate_config() {
  # Override default settings
  mv "knox/gateway-site-template.xml" "${GATEWAY_CONF}"

  # Merge user-supplied settings
  merge_xml_properties "${CLUSTER_PROPS}/knox.xml" "${GATEWAY_CONF}"
}

function generate_topology() {
  mv "${TOPOLOGY_TEMPLATE}" \
    "${KNOX_INSTALL_DIRECTORY}/conf/topologies/default.xml"
  # Obtain first master hostname and set JobHistoryUI configuration
  # We set the hostname explicitly just for this component because the job history
  # server refuses connections to localhost
  local master_hostname
  master_hostname="$(get_metadata_master)"
  local scheme="http"
  local host="localhost"

  # port numbers
  local apphistory_port=8188
  local yarn_ats_v2_port=8192
  local sparkhistory_port=18080
  local hdfs_port=9870
  local nodeui_port=8042
  local jobhistory_port=19888
  local yarn_port=8088
  local presto_port=8060
  local trino_port=8060
  local hiveserver2ui_port=10002
  local flinkhistory_port=8500

  if is_component_selected 'kerberos'; then
    master_hostname="${FQDN}"
    scheme="https"
    host="${FQDN}"

    # port numbers
    apphistory_port=8190
    yarn_ats_v2_port=8188
    sparkhistory_port=18480
    hdfs_port=9871
    nodeui_port=8044
    jobhistory_port=19889
    yarn_port=8090
    presto_port=7778
    trino_port=7778
    flinkhistory_port=18500
  fi
  # Resolve topology template with ports and hostname
  local -r topology_conf="${KNOX_INSTALL_DIRECTORY}/conf/topologies/default.xml"

  sed -i -e "s/{_MASTER_HOSTNAME_0}/${master_hostname}/g" "${topology_conf}"
  # HA cluster with 3 masters
  if (("${MASTER_COUNT}" > 1)); then
    sed -i -e "s/{_MASTER_HOSTNAME_1}/${MASTER_HOSTNAME_1}/g" "${topology_conf}"
    sed -i -e "s/{_MASTER_HOSTNAME_2}/${MASTER_HOSTNAME_2}/g" "${topology_conf}"
  fi

  sed -i -e "s/{_SCHEME_}/${scheme}/g" "${topology_conf}"
  sed -i -e "s/{_HOST_}/${host}/g" "${topology_conf}"

  # User configurable Jupyter port
  local jupyter_port
  jupyter_port=$(get_dataproc_property_or_default 'jupyter.port' '8123')
  sed -i "s/JUPYTER-PORT/${jupyter_port}/g" "${topology_conf}"

  # User configurable Zeppelin port
  local zeppelin_port
  zeppelin_port=$(get_dataproc_property_or_default 'zeppelin.port' '8080')
  sed -i "s/ZEPPELIN-PORT/${zeppelin_port}/g" "${topology_conf}"

  # APPHISTORYUI
  sed -i -e "s/{_APPHISTORY_PORT_}/${apphistory_port}/g" "${topology_conf}"
  # APPHISTORYUI2
  sed -i -e "s/{_APPTIMELINE_V2_PORT_}/${yarn_ats_v2_port}/g" "${topology_conf}"
  # SPARKHISTORYUI
  sed -i -e "s/{_SPARKHISTORY_PORT_}/${sparkhistory_port}/g" "${topology_conf}"
  # FLINKHISTORYUI
  sed -i -e "s/{_FLINKHISTORY_PORT_}/${flinkhistory_port}/g" "${topology_conf}"
  # HDFSUI and WEBHDFS
  sed -i -e "s/{_HDFS_PORT_}/${hdfs_port}/g" "${topology_conf}"
  # NODEUI
  sed -i -e "s/{_NODEUI_PORT_}/${nodeui_port}/g" "${topology_conf}"
  # JOBHISTORYUI
  sed -i -e "s/{_JOBHISTORY_PORT_}/${jobhistory_port}/g" "${topology_conf}"
  # YARNUI and RESOURCEMANAGER
  sed -i -e "s/{_YARN_PORT_}/${yarn_port}/g" "${topology_conf}"
  # PRESTO
  sed -i -e "s/{_PRESTO_PORT_}/${presto_port}/g" "${topology_conf}"
  # TRINO
  sed -i -e "s/{_TRINO_PORT_}/${trino_port}/g" "${topology_conf}"
  # HIVESERVER2UI
  sed -i -e "s/{_HIVESERVER2UI_PORT_}/${hiveserver2ui_port}/g" "${topology_conf}"
}

function kerberize_knox() {
  # JAAS login conf
  local kerberos_realm
  kerberos_realm=$(grep "default_realm = " /etc/krb5.conf | awk '{print $NF}')
  cat >/etc/knox/conf/krb5JAASLogin.conf <<EOF
com.sun.security.jgss.initiate {
    com.sun.security.auth.module.Krb5LoginModule required
    renewTGT=true
    doNotPrompt=true
    useKeyTab=true
    keyTab="/etc/security/keytab/knox.service.keytab"
    principal="knox/${FQDN}@${kerberos_realm}"
    isInitiator=true
    storeKey=true
    useTicketCache=true
    client=true;
};
EOF

  # Set Kerberos related Gateway configurations
  set_property_in_xml "${GATEWAY_CONF}" "java.security.krb5.conf" "/etc/krb5.conf"
  set_property_in_xml "${GATEWAY_CONF}" "gateway.hadoop.kerberos.secured" "true"
  set_property_in_xml "${GATEWAY_CONF}" "java.security.auth.login.config" "/etc/knox/conf/krb5JAASLogin.conf"
}

function set_websocket_size() {
  local max_ws_size="1024000"
  # Check if user overrode default websocket max in Zeppelin conf
  if is_component_selected zeppelin; then
    max_ws_size=$(get_property_in_xml /tmp/cluster/properties/zeppelin.xml \
      zeppelin.websocket.max.text.message.size "${max_ws_size}")
  fi

  set_property_in_xml "${GATEWAY_CONF}" "gateway.websocket.max.binary.buffer.size" "${max_ws_size}"
  set_property_in_xml "${GATEWAY_CONF}" "gateway.websocket.max.text.buffer.size" "${max_ws_size}"
}

function enable_dependency_to_component_gateway() {
  cat >"${KNOX_SERVICE_INIT_SCRIPT}" <<EOF
[Unit]
Description=Knox gateway service
OnSuccess=google-dataproc-component-gateway.service
Requires=google-dataproc-component-gateway.service
[Service]
User=knox
Type=forking
PIDFile=/usr/lib/knox/pids/gateway.pid
ExecStartPre=/bin/bash -c '/usr/lib/knox/bin/ldap.sh start'
ExecStart=/bin/bash -c '/usr/lib/knox/bin/gateway.sh start'
ExecStop=/bin/bash -c '/usr/lib/knox/bin/gateway.sh stop'
ExecStopPost=/bin/bash -c '/usr/lib/knox/bin/ldap.sh stop'
Restart=on-failure
[Install]
WantedBy=multi-user.target
EOF
}

function generate_runfiles_and_start() {
  # Replace PID symlink with local folder to avoid restart issues
  if [[ -L "${KNOX_INSTALL_DIRECTORY}/pids" ]]; then
    rm "${KNOX_INSTALL_DIRECTORY}/pids"
    mkdir "${KNOX_INSTALL_DIRECTORY}/pids"
  fi

  # Generate the master secret.
  retry_constant_short "${KNOX_INSTALL_DIRECTORY}"/bin/knoxcli.sh \
    create-master --generate
  chown knox:knox "${KNOX_INSTALL_DIRECTORY}/data/security/master"
  chmod 400 "${KNOX_INSTALL_DIRECTORY}/data/security/master"

  # Make sure Knox has access to the files we just created
  chown -R -L knox:knox \
    "${KNOX_INSTALL_DIRECTORY}/pids" \
    "${KNOX_INSTALL_DIRECTORY}/conf" \
    "${KNOX_INSTALL_DIRECTORY}/data/security"

  systemctl daemon-reload
  retry_constant_short systemctl enable knox
  retry_constant_short systemctl start knox
}

generate_config
generate_topology
# Enable Kerberos integration
if is_component_selected 'kerberos'; then
  kerberize_knox
fi
# Fix websocket buffer sizes
set_websocket_size
if [[ "${IS_COMPONENT_GATEWAY_HA_ENABLED}" == "true" ]]; then
  enable_dependency_to_component_gateway
fi
generate_runfiles_and_start
