#!/bin/bash

set -euo pipefail

source "$(dirname "$0")/../../bdutil_env.sh"
source "$(dirname "$0")/../../bdutil_logging.sh"
source "$(dirname "$0")/../../bdutil_retry.sh"
source "$(dirname "$0")/../../bdutil_metadata.sh"
source "$(dirname "$0")/../../bdutil_properties.sh"
source "$(dirname "$0")/../../bdutil_versions.sh"
source "$(dirname "$0")/../../bdutil_components.sh"
source "$(dirname "$0")/../../bdutil_misc.sh"
source "$(dirname "$0")/../shared/hive-metastore.sh"
source "$(dirname "$0")/../shared/presto.sh"
source "$(dirname "$0")/../shared/mysql.sh"
source "$(dirname "$0")/../shared/hive.sh"

set -x

readonly ROLE="$(get_metadata_role)"
readonly PRESTO_MASTER_HOSTNAME="$(get_metadata_master)"
readonly WORKER_COUNT=$(get_metadata_worker_count)
readonly PRESTO_PROPERTIES_FILE="${PRESTO_INSTALL_DIRECTORY}/etc/config.properties"
readonly CLUSTER_PRESTO_PROPERTIES_FILE="/tmp/cluster/properties/presto.properties"
readonly CLUSTER_PRESTO_JVM_PROPERTIES_FILE="/tmp/cluster/properties/presto-jvm.properties"
readonly CLUSTER_PRESTO_CATALOG_PROPERTIES_FILE="/tmp/cluster/properties/presto-catalog.properties"
readonly PRESTO_CATALOG_DIR="${PRESTO_INSTALL_DIRECTORY}/etc/catalog"
readonly CLUSTER_UUID=$(get_metadata_cluster_uuid)
readonly GCS_CONNECTOR_DIR="${DATAPROC_COMMON_LIB_DIR}"

function configure_node_properties() {
  cat >"${PRESTO_INSTALL_DIRECTORY}/etc/node.properties" <<EOF
node.environment=production
node.id=$(uuidgen)
node.data-dir=/var/presto/data
EOF
}

# Get the uri as set in hive-site.xml
function configure_hive() {
  local metastore_uri
  metastore_uri="$(get_hive_metastore_uri)"

  cat >"${PRESTO_CATALOG_DIR}/hive.properties" <<EOF
connector.name=hive-hadoop2
hive.metastore.uri=${metastore_uri}
EOF
}

function configure_jvm() {
  local jvm_config="${PRESTO_INSTALL_DIRECTORY}/etc/jvm.config"

  cat >"${jvm_config}" <<EOF
-server
-Xmn512m
-XX:+UseConcMarkSweepGC
-XX:+ExplicitGCInvokesConcurrent
-XX:ReservedCodeCacheSize=150M
-XX:+ExplicitGCInvokesConcurrent
-XX:+CMSClassUnloadingEnabled
-XX:+AggressiveOpts
-XX:+HeapDumpOnOutOfMemoryError
-XX:OnOutOfMemoryError=kill -9 %p
-Dhive.config.resources=/etc/hadoop/conf/core-site.xml,/etc/hadoop/conf/hdfs-site.xml
-Djava.library.path=/usr/lib/hadoop/lib/native/:/usr/lib/
EOF

  if ! is_version_at_least "${DATAPROC_IMAGE_VERSION}" "2.1"; then
    echo "-Dpresto-temporarily-allow-java8=true" >>"${jvm_config}"
  fi

  # Merge server supplied jvm properties
  cat "${CLUSTER_PRESTO_JVM_PROPERTIES_FILE}" |
    sed -e '1d' -e 's/^/-XX:/' >>"${jvm_config}"
}

function configure_catalogs() {

  cat >"${PRESTO_CATALOG_DIR}/tpch.properties" <<EOF
connector.name=tpch
EOF

  cat >"${PRESTO_CATALOG_DIR}/tpcds.properties" <<EOF
connector.name=tpcds
EOF

  cat >"${PRESTO_CATALOG_DIR}/memory.properties" <<EOF
connector.name=memory
EOF

  if [[ "${MY_HOSTNAME}" != "${DATAPROC_MASTER}" ]]; then
     echo "Skipping configuring MYSQL catalog for Presto on ${MY_HOSTNAME}"
  else
    set +x
    local db_root_password
    # read from common function from shared/mysql.sh
    db_root_password="$(read_mysql_password)"

    cat >"${PRESTO_CATALOG_DIR}/mastersql.properties" <<EOF
connector.name=mysql
connection-url=jdbc:mysql://localhost:3306
connection-user=root
connection-password=${db_root_password}
EOF
    set -x
  fi

cat >"${PRESTO_CATALOG_DIR}/bigquery.properties" <<EOF
connector.name=bigquery
EOF

  cat >"${PRESTO_CATALOG_DIR}/bigquery_public_data.properties" <<EOF
connector.name=bigquery
bigquery.project-id=bigquery-public-data
EOF

  # For each catalog.key.name=value line in PRESTO_CATALOG_PROPERTIES_FILE,
  # create a new file named catalog.properties under PRESTO_CATALOG_DIR with
  # key.name=value line.

  set +e
  cat "${CLUSTER_PRESTO_CATALOG_PROPERTIES_FILE}" | grep -v "^#" |
    awk -v dir="${PRESTO_CATALOG_DIR}" -F"." \
      '{st=index($0, "."); filename=dir"/"$1".properties"; value=substr($0, st+1); system("echo \""value"\" >> "filename)}'

  set -e

}

function configure_master() {
  # Configure master properties
  if (($WORKER_COUNT == 0)); then
    # master on single-node is also worker
    include_coordinator='true'
  else
    include_coordinator='false'
  fi
  cat >>"${PRESTO_PROPERTIES_FILE}" <<EOF
coordinator=true
node-scheduler.include-coordinator=${include_coordinator}
http-server.http.port=${HTTP_PORT}
query.max-memory=999TB
discovery-server.enabled=true
discovery.uri=http://${PRESTO_MASTER_HOSTNAME}:${HTTP_PORT}
EOF

  # Configure web ui to allow proxying by Knox
  if is_component_selected 'knox'; then
    if is_version_at_least "${DATAPROC_IMAGE_VERSION}" "2.0"; then
      echo "web-ui.authentication.type=fixed" >>"${PRESTO_PROPERTIES_FILE}"
      echo "web-ui.user=dataproc" >>"${PRESTO_PROPERTIES_FILE}"
    else
      if is_component_selected 'kerberos'; then
        echo "web-ui.session-cookie-path=/gateway/default/presto/ui" >>"${PRESTO_PROPERTIES_FILE}"
      else
        echo "web-ui.session-cookie-path=/" >>"${PRESTO_PROPERTIES_FILE}"
      fi
    fi
  fi
}

function configure_cli() {
  # Configure CLI to override default port
  cat >"${PRESTO_BIN_PATH}" <<EOF
#!/bin/bash

"${PRESTO_CLI_PATH}" --server "${PRESTO_MASTER_HOSTNAME}:${HTTP_PORT}" "\$@"
EOF
  chmod +x "${PRESTO_BIN_PATH}"
}

function configure_worker() {
  # Remove Presto CLI if not on master
  rm -f "${PRESTO_CLI_PATH}"

  cat >>"${PRESTO_PROPERTIES_FILE}" <<EOF
coordinator=false
http-server.http.port=${HTTP_PORT}
query.max-memory=999TB
discovery.uri=http://${PRESTO_MASTER_HOSTNAME}:${HTTP_PORT}
EOF
}

function merge_cluster_properties() {
  cat "${CLUSTER_PRESTO_PROPERTIES_FILE}" >>"${PRESTO_PROPERTIES_FILE}"
}

function start_presto() {
  local presto_log_dir="/var/log/presto"
  # Remove existing directory or symlink.
  rm -rf ${presto_log_dir}
  mkdir -p /var/presto/data/var/log
  ln -s /var/presto/data/var/log ${presto_log_dir}
  local init_script="/usr/lib/systemd/system/presto.service"

  cat >"${init_script}" <<EOF
[Unit]
Description=Presto DB
[Service]
Type=forking
ExecStart=${PRESTO_INSTALL_DIRECTORY}/bin/launcher.py start
ExecStop=${PRESTO_INSTALL_DIRECTORY}/bin/launcher.py stop
Restart=always
RestartPreventExitStatus=125
[Install]
WantedBy=multi-user.target
EOF

  chmod a+rw "${init_script}"

  systemctl daemon-reload
  systemctl enable presto
  systemctl start presto
  systemctl status presto
}

function kerberize() {
  domain=$(dnsdomainname)
  realm=$(grep "default_realm = " /etc/krb5.conf | awk '{print $NF}')
  presto_master_fqdn="${PRESTO_MASTER_HOSTNAME}.${domain}"

  https_port=7778
  keystore=/etc/hadoop/conf/ssl/keystore.jks
  presto_keytab=/etc/security/keytab/presto.service.keytab
  spnego_keytab=/etc/security/keytab/http.service.keytab
  local  hive_keytab=/etc/security/keytab/hive.service.keytab

  # Everything that touches unencrypted passwords should not be logged
  # Disable logging starting here
  set +x
  keystore_password=$(get_property_in_xml /etc/hadoop/conf/ssl-server.xml \
    ssl.server.keystore.password)

  # Disable global read access to config.properties to protect keystore password
  chmod 640 "${PRESTO_PROPERTIES_FILE}"

  # Presto Kerberos configuration
  cat >>"${PRESTO_PROPERTIES_FILE}" <<EOF
http-server.authentication.type=KERBEROS
http-server.http.enabled=false
http-server.https.enabled=true
http.server.authentication.krb5.service-name=HTTP
http.server.authentication.krb5.keytab=${spnego_keytab}
http.authentication.krb5.config=/etc/krb5.conf
http-server.https.port=${https_port}
http-server.https.keystore.path=${keystore}
http-server.https.keystore.key=${keystore_password}
node.internal-address-source=FQDN
internal-communication.https.required=true
internal-communication.https.keystore.path=${keystore}
internal-communication.https.keystore.key=${keystore_password}
EOF
  # shared_secret has to be the same across all hosts,
  # so cannot be totally random
  shared_secret=$(sha256sum_10 "${CLUSTER_UUID}")
  cat >>"${PRESTO_PROPERTIES_FILE}" <<EOF
internal-communication.shared-secret=${shared_secret}
EOF
  # Re-enable logging
  set -x

  sed -i "s/discovery.uri=.*/discovery.uri=https:\/\/${presto_master_fqdn}:${https_port}/" \
    "${PRESTO_PROPERTIES_FILE}"

  # Access control on coordinator
  if [[ "${HOSTNAME}" == "${PRESTO_MASTER_HOSTNAME}" ]]; then
    cat >"${PRESTO_INSTALL_DIRECTORY}/etc/access-control.properties" <<EOF
access-control.name=allow-all
EOF
  fi

  # Hive connector Kerberos properties
  cat >>"${PRESTO_CATALOG_DIR}/hive.properties" <<EOF
hive.metastore.authentication.type=KERBEROS
hive.metastore.service.principal=hive/_HOST@${realm}
hive.metastore.client.principal=hive/_HOST@${realm}
hive.metastore.client.keytab=${hive_keytab}
hive.hdfs.wire-encryption.enabled=true
hive.hdfs.authentication.type=KERBEROS
hive.hdfs.presto.principal=hive/_HOST@${realm}
hive.hdfs.presto.keytab=${hive_keytab}
EOF
}

function get_node_information() {
  if is_component_selected 'kerberos'; then
    echo "https://${presto_master_fqdn}:${https_port}"
    # Don't log this command since it uses the keystore password
    set +x
    # Don't fail execution if this command fails since
    # we will return its return code from this function
    set +e
    presto \
      --server "https://${presto_master_fqdn}:${https_port}" \
      --krb5-config-path /etc/krb5.conf \
      --krb5-keytab-path "${presto_keytab}" \
      --krb5-remote-service-name HTTP \
      --krb5-principal "presto/$(hostname -f)@${realm}" \
      --keystore-path "${keystore}" \
      --keystore-password "${keystore_password}" \
      --execute 'select * from system.runtime.nodes'
    local -r ret=$?
    set -e
    set -x
    return $ret
  else
    presto --execute 'select * from system.runtime.nodes'
  fi
}

function configure_gcs_connector() {
  # Link GCS connector JAR in Presto plugins
  for plugin in iceberg hive-hadoop2; do
    local plugin_dir="${PRESTO_INSTALL_DIRECTORY}/plugin/${plugin}"
    if [[ -d "${plugin_dir}" ]]; then
      # Delete default GCS connector and GCSIO JARs if exist
      rm -rf "${plugin_dir}/"{gcsio,util}-?.*.jar
      rm -rf "${plugin_dir}/"{gcs-connector,util-hadoop}-hadoop?-?.*.jar
      # Link shaded GCS connector JAR used by Dataproc
      ln -s "${GCS_CONNECTOR_DIR}/gcs-connector.jar" "${plugin_dir}/"
    fi
  done
}

function configure_and_start_presto() {
  # Add necessary symlinks and working directories
  mkdir -p "${PRESTO_CATALOG_DIR}"
  mkdir -p "${PRESTO_DATA_DIRECTORY}"

  ln -s "${PRESTO_INSTALL_DIRECTORY}/etc" "/etc/presto/conf"

  configure_gcs_connector

  # Configure Presto
  configure_node_properties
  configure_hive
  configure_jvm
  configure_catalogs

  local cli_only='false'
  if [[ "${HOSTNAME}" == "${PRESTO_MASTER_HOSTNAME}" ]]; then
    configure_master
    configure_cli
  elif [[ "${ROLE}" == 'Worker' ]]; then
    configure_worker
  else
    # Configure the presto CLI for secondary masters and driver pool nodes
    configure_cli
    cli_only='true'
  fi

  merge_cluster_properties

  # Secondary master and driver pool node should not run coordinator or worker
  if [[ "${cli_only}" == 'true' ]]; then
    return 0
  fi

  if is_component_selected 'kerberos'; then
    kerberize
  fi

  start_presto

  if [[ "${HOSTNAME}" == "${PRESTO_MASTER_HOSTNAME}" ]]; then
    local timeout
    timeout=$(get_dataproc_property_or_default \
      startup.component.service-binding-timeout.presto-coordinator 300)
    retry_constant_custom "$timeout" 10 get_node_information
  fi
}

function main() {
  configure_and_start_presto
}

main
