#!/bin/bash

set -euo pipefail

source "$(dirname "$0")/../../bdutil_env.sh"
source "$(dirname "$0")/../../bdutil_logging.sh"
source "$(dirname "$0")/../../bdutil_retry.sh"
source "$(dirname "$0")/../../bdutil_metadata.sh"
source "$(dirname "$0")/../../bdutil_properties.sh"
source "$(dirname "$0")/../../bdutil_os.sh"
source "$(dirname "$0")/../shared/proxy-agent.sh"
source "$(dirname "$0")/../shared/spark.sh"

set -x

# Obtain the hostname for the proxy endpoint from properties file
# Unescape the colon following the protocol
readonly HOSTNAME=$(get_dataproc_property \
  dataproc.proxy.agent.endpoint |
  sed 's/\\:/:/')

# Hostname assumed to be slash terminated
readonly PROXY="${HOSTNAME}tun/m/4592f092208ecc84946b8f8f8016274df1b36a14/"
readonly HOST="localhost:8443"
readonly SHIM_PATH="websocket-shim"
readonly BACKEND=$(get_dataproc_property \
  dataproc.proxy.backend.id)

readonly ENABLE_WEBSOCKETS=$(get_dataproc_property \
  dataproc.proxy.agent.enablewebsockets)
readonly SESSION_COOKIE_NAME=$(get_dataproc_property \
  dataproc.proxy.agent.sessioncookie)
readonly ENABLE_BANNER=$(get_dataproc_property \
  dataproc.proxy.agent.enablebanner)
readonly SIGNIN_URL="${HOSTNAME}_signin"

SHIM_WEBSOCKETS_FLAG="-shim-websockets"
if [[ "${ENABLE_WEBSOCKETS}" == "true" ]]; then
  SHIM_WEBSOCKETS_FLAG=""
fi
SESSION_COOKIE_NAME_FLAG=""
if [[ -n "${SESSION_COOKIE_NAME}" ]]; then
  SESSION_COOKIE_NAME_FLAG="-session-cookie-name=${SESSION_COOKIE_NAME}"
fi

readonly PROJECT_ID="$(get_metadata_project_id)"
readonly CLUSTER_NAME="$(get_metadata_cluster_name)"
readonly REGION_NAME="$(get_metadata_dataproc_region)"
readonly SIGNOUT_URL="${HOSTNAME}_signout"
readonly BANNER_HTML="/opt/dataproc/proxy-agent/banner.html"

mkdir -p /opt/dataproc/proxy-agent/
cat >"${BANNER_HTML}" <<EOF
<style>
banner-header {
  width:100%;
  height:40;
  font-size:20;
  padding:0px;
  color:white;
  background-color:#1a73e8;
  display: flex;
  display: -webkit-flex;
}
banner-title {
  padding:8px;
  height:24;
  text-align: left;
  flex: 1;
  -ms-flex: 1;
  -webkit-flex: 1;
}
banner-account {
  padding:8px;
  height:24;
  text-align: right;
  flex: 2;
  -ms-flex: 2;
  -webkit-flex: 2;
}
a#signout {
  color:white;
  background-color:#3a89e6;
  text-decoration:none;
  padding:4px;
}
a#project {
  color:white;
  text-decoration:none;
  padding:4px;
}
a#cluster {
  color:white;
  text-decoration:none;
  padding:4px;
}
.forum:hover {
  background:#3a89e6;
}
</style>
<banner-header>
  <banner-title>
    <a id="project" class="forum" href="https://console.cloud.google.com/home/dashboard?project=${PROJECT_ID}">${PROJECT_ID}</a> &gt; <a id="cluster" class="forum" href="https://console.cloud.google.com/dataproc/clusters/${CLUSTER_NAME}?project=${PROJECT_ID}&region=${REGION_NAME}">${CLUSTER_NAME}</a>
  </banner-title>
  <banner-account>
    <a id="signout" href="${SIGNOUT_URL}" role=button>Sign out</a>
  </banner-account>
</banner-header>
<script>
/* Update banner links for third party endpoints */
window.onload = function(){
  var componentUrl = window.location.href;

  if (componentUrl.includes("byoid.googleusercontent")) {
    var project = document.getElementById("project");
    var cluster = document.getElementById("cluster");
    var signout = document.getElementById("signout");

    project.href = project.href.replace("console.cloud.google.com/", "console.cloud.google/");
    cluster.href = cluster.href.replace("console.cloud.google.com/", "console.cloud.google/");
    signout.href = signout.href.replace("dataproc.cloud.google.com/", "dataproc.cloud.google/");
  }
}
</script>
EOF

INJECT_BANNER_FLAG=""
if [[ "${ENABLE_BANNER}" == "true" ]]; then
  INJECT_BANNER_FLAG="-banner-height=40px -inject-banner=\"\$(cat ${BANNER_HTML})\""
fi

# In Ubuntu, google-startup-scripts is after multi-user.target, in Debian it is
# the opposite. Make google-dataproc-component-gateway.service after
# multi-user.target, otherwise there will be ordering cycle.
AFTER="local-fs.target network-online.target"
if is_ubuntu; then
  AFTER="${AFTER} multi-user.target"
fi

readonly IS_COMPONENT_GATEWAY_HA_ENABLED=$(get_dataproc_property_or_default dataproc:componentgateway.ha.enabled false)
REQUIRES_LIST="Requires="
AFTER_LIST="After=google-guest-agent.service google-startup-scripts.service"
if [[ "${IS_COMPONENT_GATEWAY_HA_ENABLED}" == "true" ]]; then
  REQUIRES_LIST="${REQUIRES_LIST}knox.service"
  AFTER_LIST="${AFTER_LIST} knox.service"
fi

FORCE_HTTP2_FLAG=""
if should_start_spark_connect; then
  FORCE_HTTP2_FLAG="--force-http2"
fi

cat >"${PROXY_AGENT_INIT_SCRIPT}" <<EOF
[Unit]
Description=Google Dataproc Component Gateway Service
After=${AFTER}
${AFTER_LIST}
${REQUIRES_LIST}
Before=shutdown.target
[Service]
Type=simple
ExecStart=/bin/bash -c '${PROXY_AGENT_INSTALL_LOCATION} -proxy ${PROXY} -host ${HOST} -backend ${BACKEND} -shim-path ${SHIM_PATH} ${INJECT_BANNER_FLAG} ${SHIM_WEBSOCKETS_FLAG} ${SESSION_COOKIE_NAME_FLAG} ${FORCE_HTTP2_FLAG}'
Restart=on-failure
[Install]
WantedBy=multi-user.target
EOF

retry_constant systemctl daemon-reload
retry_constant systemctl enable "${PROXY_AGENT_SERVICE_NAME}"
retry_constant systemctl start "${PROXY_AGENT_SERVICE_NAME}"
