#!/bin/bash

set -euo pipefail

source "$(dirname "$0")/../../bdutil_logging.sh"
source "$(dirname "$0")/../../bdutil_retry.sh"
source "$(dirname "$0")/../../bdutil_properties.sh"
source "$(dirname "$0")/../../bdutil_metadata.sh"
source "$(dirname "$0")/../../bdutil_components.sh"
source "$(dirname "$0")/../../bdutil_gcs.sh"
source "$(dirname "$0")/../../bdutil_versions.sh"
source "$(dirname "$0")/../../bdutil_services.sh"
source "$(dirname "$0")/../shared/ranger.sh"
source "$(dirname "$0")/../shared/ranger_gcs_plugin.sh"

set -x

DB_ROOT_PASSWORD=""
if [[ -f /etc/mysql/my.cnf ]]; then
  if [[ -n "$(grep 'password=' /etc/mysql/my.cnf)" ]]; then
    # read from common function from shared/mysql.sh
    echo "Reading DB_ROOT_PASSWORD"
    set +x
    DB_ROOT_PASSWORD="$(read_mysql_password)"
    set -x
  fi
else
  echo "/etc/mysql/my.cnf not found in ${FQDN}"
fi
set -x

# Get a property value to indicate whether the same ranger database is to be used
# when the value is false then ranger database will be created.
function get_ranger_reuse_schema_value() {
  local ranger_reuse_schema_prop
  ranger_reuse_schema_prop=$(get_dataproc_property_or_default ranger.reuse.schema "false")
  echo "${ranger_reuse_schema_prop}"
}

RANGER_REUSE_SCHEMA="$(get_ranger_reuse_schema_value)"

function password_is_valid() {
  echo "Checking to see if password is valid"
  local -r password="$1"
  local -r length="${#password}"
  [[ "${password}" =~ [a-zA-Z] && "${password}" =~ [0-9] && "${length}" -ge 8 ]] && return

  false
}

function set_property_in_xml() {
  bdconfig set_property \
    --configuration_file "$1" \
    --name "$2" --value "$3" \
    --create_if_absent \
    --clobber ||
    log_and_fail "ranger" "Unable to set $2"
}

function set_property_hive_audit() {
  set_property_in_xml "${RANGER_INSTALL_DIR}/hive/conf/ranger-hive-audit.xml" "$1" "$2"
}

function set_property_hdfs_audit() {
  set_property_in_xml "${RANGER_INSTALL_DIR}/hadoop/etc/hadoop/ranger-hdfs-audit.xml" "$1" "$2"
}

function set_property_yarn_audit() {
  set_property_in_xml "${RANGER_INSTALL_DIR}/hadoop/etc/hadoop/ranger-yarn-audit.xml" "$1" "$2"
}

function set_property_ranger_admin() {
  set_property_in_xml "${RANGER_INSTALL_DIR}/ranger-admin/ews/webapp/WEB-INF/classes/conf.dist/ranger-admin-site.xml" "$1" "$2"
}

function remove_property_ranger_admin() {
  bdconfig remove_property --configuration_file "${RANGER_INSTALL_DIR}/ranger-admin/ews/webapp/WEB-INF/classes/conf.dist/ranger-admin-site.xml" \
    --name "$1"
}

function create_principals() {
  # add ranger user in group hadoop so it can access the spnego (http) keytab.
  useradd ranger || true
  usermod -a -G hadoop ranger

  set +x
  # add an admin principal using the password provided
  kadmin.local -q "addprinc -pw ${ADMIN_PASSWORD} admin"
  set -x
}

function configure_admin() {
  echo "Configuring Admin"
  set +x
  if ! password_is_valid "${ADMIN_PASSWORD}" ; then
    log_and_fail "ranger" "The Ranger admin password must consist of at least 8 characters \
with a minimum of one alphabetic and one numeric character." "${EXIT_CODE_CLIENT_ERROR}"
  fi

  local -r install_properties="${RANGER_INSTALL_DIR}/ranger-admin/install.properties"
  chmod 750 "${install_properties}"

  if [[ "${RANGER_REUSE_SCHEMA}" == 'true' ]]; then
    echo "Re-using Existing Ranger Schema"
    sed -i "s/^#setup_mode=/setup_mode=/" "${install_properties}"
  fi

  sed -i "s/^db_root_password=/db_root_password=${DB_ROOT_PASSWORD}/" \
    "${install_properties}"
  sed -i "s/^db_password=/db_password=${DB_ADMIN_PASSWORD}/" \
    "${install_properties}"
  sed -i "s/^rangerAdmin_password=/rangerAdmin_password=${ADMIN_PASSWORD}/" \
    "${install_properties}"
  sed -i "s/^rangerTagsync_password=/rangerTagsync_password=${ADMIN_PASSWORD}/" \
    "${install_properties}"
  sed -i "s/^rangerUsersync_password=/rangerUsersync_password=${ADMIN_PASSWORD}/" \
    "${install_properties}"
  sed -i "s/^keyadmin_password=/keyadmin_password=${ADMIN_PASSWORD}/" \
    "${install_properties}"
  set -x
  sed -i 's/^audit_solr_user=/audit_solr_user=solr/' "${install_properties}"
  sed -i "s/^db_host=.*/db_host=127.0.0.1:${MYSQL_PORT}/" "${install_properties}"
  set_property_ranger_admin "ranger.service.http.port" "${RANGER_ADMIN_PORT}"

  set +x
  echo "Checking if mysql is running on ${MYSQL_PORT} using mysqladmin"
  if [[ "${RANGER_REUSE_SCHEMA}" == 'false' ]]; then
    retry_quiet \
      mysqladmin -u root -p"${DB_ROOT_PASSWORD}" --port "${MYSQL_PORT}" -h "${MYSQL_HOST}" status
    # Create the database 'ranger' if it does not already exist.
    if ! mysql -u root -p"${DB_ROOT_PASSWORD}" --port "${MYSQL_PORT}" -h "${MYSQL_HOST}" -e 'use ranger'; then
      echo "Creating Ranger Database"
      mysql -u root -p"${DB_ROOT_PASSWORD}" --port "${MYSQL_PORT}" -h "${MYSQL_HOST}" -e "CREATE DATABASE ranger;"
    fi

    # Create the user 'rangeradmin' if it does not already exist.
    echo "Checking to see if the mysql user Ranger exists."
    if ! mysql -u rangeradmin -p"${DB_ADMIN_PASSWORD}" --port "${MYSQL_PORT}" -h "${MYSQL_HOST}" -e ''; then
      echo "Creating rangeradmin user using mysql"
      mysql -u root -p"${DB_ROOT_PASSWORD}" --port "${MYSQL_PORT}" -h "${MYSQL_HOST}" -e "CREATE USER 'rangeradmin'@'${MYSQL_USER_HOST}' IDENTIFIED BY '${DB_ADMIN_PASSWORD}';"
      echo "Granting privileges to rangeradmins"
      mysql -u root -p"${DB_ROOT_PASSWORD}" --port "${MYSQL_PORT}" -h "${MYSQL_HOST}" -e "GRANT ALL PRIVILEGES ON ranger.* TO 'rangeradmin'@'${MYSQL_USER_HOST}';"
      echo "Ranger database and user created successfully"
    fi
  fi
  set -x

  if is_component_selected "kerberos"; then
    sed -i 's/^authentication_method=/authentication_method=UNIX/' \
      "${install_properties}"

    sed -i "s/^spnego_principal=/spnego_principal=HTTP\/${FQDN}@${REALM}/" \
      "${install_properties}"

    sed -i "s/^spnego_keytab=/spnego_keytab=\/etc\/security\/keytab\/http.service.keytab/" \
      "${install_properties}"

    sed -i "s/^cookie_domain=/cookie_domain=${FQDN}/" \
      "${install_properties}"

    sed -i "s/^admin_principal=/admin_principal=rangeradmin\/${FQDN}@${REALM}/" \
      "${install_properties}"

    sed -i "s/^admin_keytab=/admin_keytab=\/etc\/security\/keytab\/rangeradmin.service.keytab/" \
      "${install_properties}"

    sed -i "s/^lookup_principal=/lookup_principal=rangerlookup\/${FQDN}@${REALM}/" \
      "${install_properties}"

    sed -i "s/^lookup_keytab=/lookup_keytab=\/etc\/security\/keytab\/rangerlookup.service.keytab/" \
      "${install_properties}"

    set_property_ranger_admin "xasecure.audit.destination.solr.force.use.inmemory.jaas.config" "true"
    set_property_ranger_admin "xasecure.audit.jaas.Client.loginModuleControlFlag" "required"
    set_property_ranger_admin "xasecure.audit.jaas.Client.loginModuleName" "com.sun.security.auth.module.Krb5LoginModule"
    set_property_ranger_admin "xasecure.audit.jaas.Client.option.keyTab" "/etc/security/keytab/rangeradmin.service.keytab"
    set_property_ranger_admin "xasecure.audit.jaas.Client.option.principal" "rangeradmin/_HOST@${REALM}"
    set_property_ranger_admin "xasecure.audit.jaas.Client.option.serviceName" "solr"
    set_property_ranger_admin "xasecure.audit.jaas.Client.option.storeKey" "false"
    set_property_ranger_admin "xasecure.audit.jaas.Client.option.useKeyTab" "true"
    remove_property_ranger_admin "ranger.solr.audit.user"
    remove_property_ranger_admin "ranger.solr.audit.user.password"
  fi
}

function ranger_activate() {

  # Add user 'root' to hdfs supergroup 'hadoop'.
  # If Kerberized, add user 'dataproc'.
  usermod -a -G hadoop root
  if is_component_selected "kerberos"; then
    usermod -a -G hadoop dataproc
  fi

  # Some distributions of Python does not have the 'requests' module,
  # which is needed by Ranger, so make sure it is installed.
  # So far we've only seen this on Ubuntu machines, where pip is installed.
  if ! python -c "import requests"; then
    pip install requests
  fi

  local role
  role="$(get_metadata_role)"
  cd "${RANGER_INSTALL_DIR}"

  if [[ -n "${CLOUD_SQL_INSTANCE}" ]]; then
    check_for_cloud_sql_instance
    configure_cloud_sql_proxy_for_component ranger "${CLOUD_SQL_INSTANCE}" "${MYSQL_PORT}" "${CLOUD_SQL_PRIVATE_IP}"
  fi

  if [[ "${role}" == 'Master' && "${NODE_NAME}" =~ ^.*(-m|-m-0)$ ]]; then
    configure_admin
    if is_component_selected "hdfs"; then
      echo "Delaying starting ranger-admin until HDFS is ready"
    else
      run_ranger_admin
      add_usersync_plugin
      add_hdfs_plugin
      add_hive_plugin
      add_yarn_plugin
      if is_component_selected "presto"; then
        add_presto_plugin
      fi
    fi
  fi

  # In HA clusters Namenode is installed also on m-1. We also need to install and configure
  # ranger hdfs plugin on m-1.
  if [[ "${role}" == 'Master' && "${NODE_NAME}" =~ ^.*(-m-1)$ ]]; then
    if is_component_selected "hdfs"; then
       echo "Delaying adding hdfs plugin until HDFS is ready"
    else
        add_hdfs_plugin
    fi
  fi

  # In HA clusters ResourceManager is installed also on m-1 and m-2.
  # We also need to install and configure ranger yarn plugin on additional masters.
  if [[ "${role}" == 'Master' && "${NODE_NAME}" =~ ^.*(-m-1|-m-2)$ ]]; then
    if is_component_selected "hdfs"; then
       echo "Delaying adding hive and yarn plugin until HDFS is ready"
    else
        add_hive_plugin
        add_yarn_plugin
        if is_component_selected "presto"; then
          add_presto_plugin
        fi
    fi
  fi

  echo "Ranger plugins installation complete."
  # Install Ranger GCS plugin on all VMs if it is selected.
  if [[ "$(get_dataproc_property 'ranger.gcs.plugin.enable')" == 'true' ]]; then
    if [[ "${role}" == 'Master' && "${NODE_NAME}" =~ ^.*(-m|-m-0)$ ]]; then
      if is_component_selected "hdfs"; then
        # We can't block master node otherwise post hdfs script won't run
        echo "Delaying adding Ranger GCS plugin until HDFS is ready"
      else
        # Wait until Ranger admin is ready
        wait_until_url_ready "http://${MASTER}:${RANGER_ADMIN_PORT}"
        register_ranger_gcs_service_to_ranger_admin
        create_ranger_gcs_plugin_iam_roles
        install_ranger_gcs_plugin_authorization_service
        install_ranger_gcs_plugin_client
      fi
    elif [[ "${role}" == 'Worker' ]]; then
        # Install auth service regardless of Ranger admin's status.
        install_ranger_gcs_plugin_authorization_service
        install_ranger_gcs_plugin_client
    fi
  fi
}

ranger_activate
