#!/bin/bash

set -euo pipefail

source "$(dirname "$0")/../../bdutil_env.sh"
source "$(dirname "$0")/../../bdutil_logging.sh"
source "$(dirname "$0")/../../bdutil_retry.sh"
source "$(dirname "$0")/../../bdutil_metadata.sh"
source "$(dirname "$0")/../../bdutil_properties.sh"
source "$(dirname "$0")/../../bdutil_components.sh"
source "$(dirname "$0")/../../bdutil_misc.sh"
source "$(dirname "$0")/../../bdutil_versions.sh"
source "$(dirname "$0")/../shared/zookeeper.sh"

set -x

readonly MASTER_ADDITIONAL="$(get_metadata_master_additional)"
readonly CLUSTER_NAME="$(get_metadata_cluster_name)"
readonly FQDN="$(hostname -f)"
readonly DOMAIN="$(dnsdomainname)"
readonly SOLR_CONF_FILE="/etc/default/solr"
readonly SOLR_SERVER_DIR="/usr/lib/solr/server"
readonly SOLR_HOME_DIR="/usr/lib/solr/server/solr"
readonly SOLR_WEBAPP_HOME_DIR="/usr/lib/solr/server/solr-webapp"
readonly SOLR_SECURITY_POLICY_FILE="/usr/lib/solr/server/etc/security.policy"
readonly RANGER_GCS_PLUGIN_CLIENT_JAR="/usr/local/share/google/dataproc/lib/ranger_gcs_plugin_client.jar"

function configure_solr() {
  usermod -a -G hadoop solr

  local zookeeper_nodes
  zookeeper_nodes="$(get_zookeeper_list)"

  sed -i "s/^#SOLR_HOST=\"192.168.1.1\"/SOLR_HOST=\"${FQDN}\"/" "${SOLR_CONF_FILE}"
  if is_version_at_least "${DATAPROC_IMAGE_VERSION}" "2.1"; then
    # Solr > 9.0 exposes server to 127.0.0.1, because of other services (eg. ranger)
    # cannot reach it as they attempt to connect via FQDN
    sed -i "/^SOLR_HOST=\"/aSOLR_JETTY_HOST=\"0.0.0.0\"" "${SOLR_CONF_FILE}"
  fi
  sed -i "s/^#SOLR_LOGS_DIR=/SOLR_LOGS_DIR=/" "${SOLR_CONF_FILE}"
  chown -R solr:solr "/usr/lib/solr/server"

  # Enable SolrCloud setup in HA mode.
  if [[ "${MASTER_ADDITIONAL}" != "" ]]; then
    sed -i "s/^#ZK_HOST=\"\"/ZK_HOST=\"${zookeeper_nodes}\/solr\"/" \
      "${SOLR_CONF_FILE}"
    /usr/lib/solr/bin/solr zk mkroot /solr -z "${FQDN}:2181" || echo "Node already exists for /solr."
  fi

  # Enable hdfs as default backend storage. User can optionally use a GCS path
  # as the 'root' path of HDFS, so that the documents will be preserved
  # after cluster is deleted.
  # TODO  Notice live clusters cannot share the same path,
  # as one cluster will lock the path, so activation of solr servers
  # from other clusters will fail.
  local gcs_path
  gcs_path=$(get_dataproc_property solr.gcs.path)

  if [[ -z "${gcs_path}" ]]; then
    local hdfs_uri
    hdfs_uri=$(hdfs getconf -confKey fs.defaultFS)
  else
    local -r hdfs_uri="${gcs_path}"
    # Workaround for SOLR-11335
    bdconfig set_property \
      --configuration_file /etc/hadoop/conf/core-site.xml \
      --name fs.gs.impl.disable.cache --value true \
      --create_if_absent \
      --clobber
  fi

  ln -s "${DATAPROC_COMMON_LIB_DIR}/gcs-connector.jar" \
    "${SOLR_WEBAPP_HOME_DIR}/webapp/WEB-INF/lib/"

  # Ranger gcs plugin jar should already be linked at common lib path
  # at image build
  if [[ -f "${RANGER_GCS_PLUGIN_CLIENT_JAR}" ]]; then
    ln -s "${RANGER_GCS_PLUGIN_CLIENT_JAR}" \
      "${SOLR_WEBAPP_HOME_DIR}/webapp/WEB-INF/lib/"
  fi

  local solr_modules=""
  if is_version_at_least "${DATAPROC_IMAGE_VERSION}" "2.1"; then
    solr_modules="-Dsolr.modules=hdfs,hadoop-auth"

    # Don't set content security policy for Solr web UI, otherwise it will have
    # problem being embedded as an iframe by Component Gateway, see b/262892210.
    sed -i \
      '/Content-Security-Policy/{n; s#.*#<Set name="value"><\/Set>#g}' \
      "${SOLR_SERVER_DIR}/etc/jetty.xml"
  fi

  cat >>"${SOLR_CONF_FILE}" <<EOF
SOLR_OPTS="\${SOLR_OPTS} -Dsolr.directoryFactory=HdfsDirectoryFactory -Dsolr.lock.type=hdfs \
-Dsolr.hdfs.home=${hdfs_uri}/solr -Dsolr.hdfs.confdir=/etc/hadoop/conf ${solr_modules}"
EOF

  if is_version_at_least "${DATAPROC_IMAGE_VERSION}" "2.2"; then
    # Solr 9.2+ seems to validate permissions based on security policy even if
    # kerberos is not enabled.
    cat >>"${SOLR_SECURITY_POLICY_FILE}" <<EOF
// permissions to access hadoop configurations.
grant {
    permission java.io.FilePermission "\${solr.hdfs.confdir}", "read,readlink";
    permission java.io.FilePermission "\${solr.hdfs.confdir}\${/}-", "read,readlink";
};
EOF
  fi

  # If on a Kerberized cluster, enable Kerberos Authentication Plugin.
  if is_component_selected "kerberos"; then
    local solr_auth_class;
    if is_version_at_least "${DATAPROC_IMAGE_VERSION}" "2.1"; then
      # Solr > 9.0 KerberosPlugin moved to hadoop auth module inside Solr.
      # https://github.com/apache/solr/pull/636
      solr_auth_class="org.apache.solr.security.hadoop.KerberosPlugin"
      # Copying hadoop-auth module jars directly to classpath due to
      # some weird classpath errors with this plugin.
      cp /usr/lib/solr/modules/hdfs/lib/*.jar ${SOLR_WEBAPP_HOME_DIR}/webapp/WEB-INF/lib
      cp /usr/lib/solr/modules/hadoop-auth/lib/*.jar ${SOLR_WEBAPP_HOME_DIR}/webapp/WEB-INF/lib
    else
      solr_auth_class="org.apache.solr.security.KerberosPlugin"
    fi
    if is_version_at_least "${DATAPROC_IMAGE_VERSION}" "2.2"; then
      # Copying commons-codec jar to avoid classpath problem.
      cp /usr/lib/hadoop/lib/commons-codec-*.jar ${SOLR_SERVER_DIR}/lib
      chown solr:solr ${SOLR_SERVER_DIR}/lib/commons-codec-*.jar
    fi
    cat >>"${SOLR_HOME_DIR}/security.json" <<EOF
{"authentication":{"class":"$solr_auth_class"}}
EOF
    chown solr:solr "${SOLR_HOME_DIR}/security.json"
    if [[ "${MASTER_ADDITIONAL}" != "" ]]; then
      /usr/lib/solr/bin/solr zk cp "${SOLR_HOME_DIR}/security.json" zk:/security.json -z "${zookeeper_nodes}"
    fi

    local realm
    realm=$(grep "default_realm = " /etc/krb5.conf | awk '{print $NF}')
    cat >"${SOLR_HOME_DIR}"/jaas-client.conf <<EOF
Client {
  com.sun.security.auth.module.Krb5LoginModule required
  useKeyTab=true
  keyTab="/etc/security/keytab/http.service.keytab"
  storeKey=true
  debug=true
  principal="HTTP/${FQDN}@${realm}";
};
EOF
    chmod 444 "${SOLR_HOME_DIR}"/jaas-client.conf

    cat >>"${SOLR_CONF_FILE}" <<EOF
SOLR_AUTH_TYPE="kerberos"
SOLR_AUTHENTICATION_OPTS="-Djava.security.auth.login.config=${SOLR_HOME_DIR}/jaas-client.conf \
-Dsolr.kerberos.cookie.domain=${FQDN} \
-Dsolr.kerberos.cookie.portaware=true \
-Dsolr.kerberos.principal=HTTP/${FQDN}@${realm} \
-Dsolr.kerberos.keytab=/etc/security/keytab/http.service.keytab \
-Dsolr.kerberos.delegation.token.enabled=true"
EOF
    cat >>"${SOLR_CONF_FILE}" <<EOF
SOLR_OPTS="\${SOLR_OPTS} -Dsolr.hdfs.security.kerberos.enabled=true \
-Dsolr.hdfs.security.kerberos.keytabfile=/etc/security/keytab/http.service.keytab \
-Dsolr.hdfs.security.kerberos.principal=HTTP/${FQDN}@${realm} \
-Dsolr.hdfs.confdir=/etc/hadoop/conf"
EOF
    if is_component_selected "knox"; then
      cat >>"${SOLR_CONF_FILE}" <<EOF
SOLR_AUTHENTICATION_OPTS="\${SOLR_AUTHENTICATION_OPTS} -Dsolr.impersonator.user.name=knox \
-Dsolr.kerberos.impersonator.user.knox.hosts=* \
-Dsolr.kerberos.impersonator.user.knox.groups=*"
EOF
    fi
    if is_version_at_least "${DATAPROC_IMAGE_VERSION}" "2.2"; then
      # Solr 9.2+ seems to validate permissions based on security policy even if
      # kerberos is not enabled.
      cat >>"${SOLR_SECURITY_POLICY_FILE}" <<EOF
// permissions to access kerberos keytab.
grant {
    permission java.io.FilePermission "\${solr.hdfs.security.kerberos.keytabfile}", "read,readlink";
};
EOF
    fi
  fi

  # Make sure solr configuration is universally readable
  chmod +r "${SOLR_CONF_FILE}"

  cat >/etc/systemd/system/solr.service <<EOF
[Unit]
Description=Apache SOLR
ConditionPathExists=/usr/lib/solr/bin
After=syslog.target network.target remote-fs.target nss-lookup.target systemd-journald-dev-log.socket
Before=multi-user.target
Conflicts=shutdown.target

[Service]
User=solr
LimitNOFILE=1048576
LimitNPROC=1048576
Environment=SOLR_INCLUDE=/etc/default/solr
Environment=RUNAS=solr
Environment=SOLR_INSTALL_DIR=/usr/lib/solr

Restart=on-failure
RestartSec=5

ExecStart=/usr/lib/solr/bin/solr start -f
ExecStop=/usr/lib/solr/bin/solr stop
Restart=on-failure

[Install]
WantedBy=multi-user.target
EOF

  retry_constant systemctl daemon-reload
  retry_constant systemctl enable solr.service
}

function main() {
  local role
  role="$(get_metadata_role)"

  if [[ "${role}" == "Master" ]]; then
    configure_solr
    if is_component_selected "kerberos"; then
      kinit -kt /etc/security/keytab/dataproc.service.keytab "dataproc/${FQDN}"
    fi
    if is_component_selected "hdfs"; then
       echo "Delaying starting solr until HDFS is ready"
    else
        retry_constant systemctl start solr
    fi
  fi
}

main
