#!/bin/bash

set -euo pipefail

source "$(dirname "$0")/../../bdutil_components.sh"
source "$(dirname "$0")/../../bdutil_env.sh"
source "$(dirname "$0")/../../bdutil_logging.sh"
source "$(dirname "$0")/../../bdutil_metadata.sh"
source "$(dirname "$0")/../../bdutil_networking.sh"
source "$(dirname "$0")/../../bdutil_properties.sh"
source "$(dirname "$0")/../../bdutil_retry.sh"
source "$(dirname "$0")/../../bdutil_services.sh"
source "$(dirname "$0")/../../bdutil_versions.sh"
source "$(dirname "$0")/../shared/spark.sh"
source "${DATAPROC_DIR}/dataproc_env.sh"

set -x

# We need to set it to lower value than 5 minutes to allow s8s control plane
# detect cluster creation failure earlier than job task acquisition failure.
# This allows to surface root cause of the s8s workload failure.
readonly WAIT_TIMEOUT_SECONDS=200
readonly INITIAL_WORKER_COUNT=1
COMPONENT_GATEWAY_HA_ENABLED=$(get_dataproc_property_or_default dataproc:componentgateway.ha.enabled false)
readonly COMPONENT_GATEWAY_HA_ENABLED

function wait_for_spark_master() {
  wait_for_port \
    "spark-master" \
    "${MASTER_HOSTNAME_0}" \
    "${SPARK_MASTER_PORT}" \
    "${WAIT_TIMEOUT_SECONDS}"
  wait_for_port \
    "spark-master" \
    "${MASTER_HOSTNAME_0}" \
    "${SPARK_MASTER_WEBUI_PORT}" \
    "${WAIT_TIMEOUT_SECONDS}"
}

function wait_for_spark_worker() {
  wait_for_port \
    "spark-worker" \
    "${HOSTNAME}" \
    "${SPARK_WORKER_PORT}" \
    "${WAIT_TIMEOUT_SECONDS}"
}

function start_standalone_master_service() {
  local -r service_name='spark-master'
  local -r exec_start='/usr/lib/spark/sbin/start-master.sh'
  local -r exec_stop='/usr/lib/spark/sbin/stop-master.sh'

  start_standalone_service "${service_name}" "${exec_start}" "${exec_stop}"
}

function start_standalone_worker_service() {
  local -r service_name='spark-worker'
  local -r spark_master_address="spark://${MASTER_HOSTNAME_0}:${SPARK_MASTER_PORT}"
  local -r exec_start="/usr/lib/spark/sbin/start-slave.sh ${spark_master_address}"
  local -r exec_stop='/usr/lib/spark/sbin/stop-slave.sh'

  start_standalone_service "${service_name}" "${exec_start}" "${exec_stop}"
}

function start_standalone_service() {
  local -r service_name="$1"
  local -r exec_start="$2"
  local -r exec_stop="$3"

  # Run Spark worker as user `spark` because executor containers run as `spark`,
  # the staging dirs and files created by the worker need to be accessible by
  # the executor. If run as `root`, executor will hit permission errors.
  cat >"/etc/systemd/system/${service_name}.service" <<EOF
[Unit]
Description=Spark Standalone ${ROLE}
Wants=network-online.target
After=network-online.target

[Service]
User=spark
Group=spark
Type=forking
ExecStart=${exec_start}
ExecStop=${exec_stop}
Restart=on-failure
OOMScoreAdjust=-1000

[Install]
WantedBy=multi-user.target
EOF

  retry_constant_short systemctl daemon-reload
  retry_constant_short systemctl enable "${service_name}"
  retry_constant_short systemctl start "${service_name}"
}

function check_workers_registered() {
  local alive_workers
  alive_workers=$(curl -fsS \
    "http://${HOSTNAME}:${SPARK_MASTER_WEBUI_PORT}/json/" |
    jq ".aliveworkers")
  ((alive_workers >= INITIAL_WORKER_COUNT))
}

function wait_until_worker_registered() {
  retry_constant_custom \
    "${WAIT_TIMEOUT_SECONDS}" 1 check_workers_registered ||
    log_and_fail "spark" "Timed out waiting for at least ${INITIAL_WORKER_COUNT} worker(s) \
registered. This is often caused by firewall rules that prevent Spark workers \
from communicating with the master. Please review your network firewall rules \
and be sure they allow communication on all ports between all nodes. See \
https://cloud.google.com/dataproc-serverless/docs/concepts/network for instructions."
}

function start_standalone_master() {
  start_standalone_master_service
  wait_for_spark_master
  wait_until_worker_registered
}

function start_standalone_worker() {
  start_standalone_worker_service
  wait_for_spark_worker
}

function should_start_history_server() {
  # Start history server only on standard clusters (i.e. non-serverless)
  # on 0-master node if HDFS disabled, otherwise postpone history server startup
  # until after HDFS is initialized on standard clusters
  [[ "${DATAPROC_IMAGE_TYPE}" == "standard" ]] &&
    [[ "${ROLE}" == "Master" ]] &&
    [[ "${COMPONENT_GATEWAY_HA_ENABLED}" == "true" || "${MASTER_INDEX?}" == "0" ]] &&
    ! is_component_selected "hdfs"
}

function start_spark_connect_server() {
  retry_constant systemctl enable spark-connect spark-connect-proxy
  retry_constant systemctl start spark-connect spark-connect-proxy
  # Check whether Spark Connect server is up
  wait_for_port \
    "spark-connect" \
    "localhost" \
    "${SPARK_CONNECT_PORT}" \
    "${WAIT_TIMEOUT_SECONDS}" ||
    log_and_fail "spark-connect" "Failed to start spark-connect server"
}

function activate_spark() {
  if [[ "${DATAPROC_IMAGE_TYPE}" == "serverless-spark" ]] &&
    ! is_version_at_least "${DATAPROC_IMAGE_VERSION}" '3.0'; then
    # Spark master and worker do not run in container, driver and executors do.
    if [[ "${ROLE}" == 'Master' ]]; then
      start_standalone_master
    elif [[ "${ROLE}" == 'Worker' ]]; then
      start_standalone_worker
    fi
  fi

  if should_start_history_server; then
    start_spark_history_server
  fi

  if should_start_spark_connect; then
    start_spark_connect_server
  fi
}

activate_spark
