#!/bin/bash

set -euo pipefail

source "$(dirname "$0")/../../bdutil_env.sh"
source "$(dirname "$0")/../../bdutil_logging.sh"
source "$(dirname "$0")/../../bdutil_retry.sh"
source "$(dirname "$0")/../../bdutil_metadata.sh"
source "$(dirname "$0")/../../bdutil_properties.sh"
source "$(dirname "$0")/../../bdutil_versions.sh"
source "$(dirname "$0")/../../bdutil_components.sh"
source "$(dirname "$0")/../../bdutil_gcs.sh"
source "$(dirname "$0")/../shared/effective-python.sh"

set -x

readonly ZEPPELIN_CONF="/etc/zeppelin/conf"
readonly CLUSTER_PROPS="/tmp/cluster/properties"

CONFIGBUCKET=$(get_metadata_bucket)

readonly SPARK_INTERPRETER_SETTINGS="/usr/lib/zeppelin/interpreter/spark/interpreter-setting.json"
readonly SHELL_INTERPRETER_SETTINGS="/usr/lib/zeppelin/interpreter/sh/interpreter-setting.json"

# Link in hive configuration.
ln -s /etc/hive/conf/hive-site.xml "${ZEPPELIN_CONF}"

# Make Hive JARs available for JDBC Hive interpreter
ln -s /usr/lib/hive/jdbc/hive-jdbc-*-standalone.jar /usr/lib/zeppelin/interpreter/jdbc

# Determine effective python before starting zeppelin service.
# TODO: b/150611572 remove this logic and activating Anaconda/Miniconda
# before Zeppelin.
install_effective_python_profile

# Use conda's python interpreter.
cat >>"${ZEPPELIN_CONF}/zeppelin-env.sh" <<EOF
if [[ -f "${EFFECTIVE_PYTHON_PROFILE}" ]]; then
  source "${EFFECTIVE_PYTHON_PROFILE}"
fi
EOF

if [[ ! -f ${ZEPPELIN_CONF}/zeppelin-site.xml ]]; then
  mv "${ZEPPELIN_CONF}/zeppelin-site.xml.template" "${ZEPPELIN_CONF}/zeppelin-site.xml"
fi

if is_component_selected 'knox' || is_version_at_least "${DATAPROC_IMAGE_VERSION}" "2.0"; then
  # Restrict connections to localhost to prevent security vuln
  bdconfig set_property \
    --configuration_file "${ZEPPELIN_CONF}/zeppelin-site.xml" \
    --name 'zeppelin.server.addr' \
    --value '127.0.0.1' \
    --clobber
else
  # Legacy behavior for pre-2.0 compatibility: allow to connect from any host
  bdconfig set_property \
    --configuration_file "${ZEPPELIN_CONF}/zeppelin-site.xml" \
    --name 'zeppelin.server.addr' \
    --value '0.0.0.0' \
    --clobber
fi

# Verify we can access the GCS bucket.
check_gcs_bucket_access "gs://${CONFIGBUCKET}/" 'zeppelin'

# Store notebooks in GCS by default.
bdconfig set_property \
  --configuration_file "${ZEPPELIN_CONF}/zeppelin-site.xml" \
  --name 'zeppelin.notebook.gcs.dir' \
  --value "gs://${CONFIGBUCKET}/notebooks/zeppelin" \
  --clobber
bdconfig set_property \
  --configuration_file "${ZEPPELIN_CONF}/zeppelin-site.xml" \
  --name 'zeppelin.notebook.storage' \
  --value 'org.apache.zeppelin.notebook.repo.GCSNotebookRepo' \
  --clobber

# Merge any user-supplied properties.
merge_xml_properties \
  "${CLUSTER_PROPS}/zeppelin.xml" "${ZEPPELIN_CONF}/zeppelin-site.xml"
merge_sh_env_vars \
  "${CLUSTER_PROPS}/zeppelin-env.sh" "${ZEPPELIN_CONF}/zeppelin-env.sh"

merge_java_properties ${CLUSTER_PROPS}/zeppelin-log4j.properties \
  ${ZEPPELIN_CONF}/log4j.properties

# Configure port.
ZEPPELIN_PORT=$(get_dataproc_property_or_default zeppelin.port 8080)
echo "export ZEPPELIN_PORT=${ZEPPELIN_PORT}" >>"${ZEPPELIN_CONF}/zeppelin-env.sh"

# Configure properties for Zeppelin interpreters (Spark, shell, etc)
function configure_interpreters() {
  python - <<EOF
import json

def get_matching_interpreter(interpreters, name):
  for interpreter in interpreters:
    if "name" in interpreter and interpreter["name"] == name:
      return interpreter

with open("${SPARK_INTERPRETER_SETTINGS}") as f:
  spark_interpreters = json.load(f)

spark_properties = get_matching_interpreter(spark_interpreters, "spark")["properties"]

# Disable Spark version check because Spark 3.1 is not supported by Zeppelin 0.9
spark_properties["zeppelin.spark.enableSupportedVersionCheck"]["defaultValue"] = False
# Ensure that Spark interpreter defaults to using YARN instead of local mode
spark_properties["spark.master"]["defaultValue"] = "yarn"
spark_properties["spark.submit.deployMode"]["defaultValue"] = "client"

with open("${SPARK_INTERPRETER_SETTINGS}", "w") as f:
  json.dump(spark_interpreters, f, indent=2)
EOF
}

# Configure a `zeppelin` Kerberos principal and configure Zeppelin interpreters
# to run as this principal
function kerberize() {
  local realm
  realm=$(grep "default_realm = " /etc/krb5.conf | awk '{print $NF}')
  local -r zeppelin_keytab=/etc/security/keytab/zeppelin.service.keytab
  local zeppelin_principal
  zeppelin_principal="zeppelin/$(hostname -f)@${realm}"

  # Set interpreter settings for Spark and Shell interpreters
  python - <<EOF
import json

def get_matching_interpreter(interpreters, name):
  for interpreter in interpreters:
    if "name" in interpreter and interpreter["name"] == name:
      return interpreter

with open("${SPARK_INTERPRETER_SETTINGS}") as f:
    spark_interpreters = json.load(f)
spark_properties = get_matching_interpreter(spark_interpreters, "spark")["properties"]
spark_properties["spark.yarn.principal"] = {
  "propertyName": "spark.yarn.principal",
  "defaultValue": "${zeppelin_principal}",
  "type": "textarea"
}
spark_properties["spark.yarn.keytab"] = {
  "propertyName": "spark.yarn.keytab",
  "defaultValue": "${zeppelin_keytab}",
  "type": "textarea"
}
with open("${SPARK_INTERPRETER_SETTINGS}", "w") as f:
    json.dump(spark_interpreters, f, indent=2)

with open("${SHELL_INTERPRETER_SETTINGS}") as f:
    shell_interpreters = json.load(f)
shell_properties = get_matching_interpreter(shell_interpreters, "sh")["properties"]
shell_properties["zeppelin.shell.principal"] = {
  "propertyName": "zeppelin.shell.principal",
  "defaultValue": "${zeppelin_principal}",
  "type": "string"
}
shell_properties["zeppelin.shell.auth.type"] = {
  "propertyName": "zeppelin.shell.auth.type",
  "defaultValue": "KERBEROS",
  "type": "string"
}
shell_properties["zeppelin.shell.keytab.location"] = {
  "propertyName": "zeppelin.shell.keytab.location",
  "defaultValue": "${zeppelin_keytab}",
  "type": "string"
}
with open("${SHELL_INTERPRETER_SETTINGS}", "w") as f:
    json.dump(shell_interpreters, f, indent=2)
EOF
}

if is_version_at_least "${DATAPROC_IMAGE_VERSION}" "2.0"; then
  configure_interpreters
fi

if is_component_selected 'kerberos'; then
  kerberize
fi

retry_constant systemctl daemon-reload
retry_constant systemctl enable zeppelin
retry_constant systemctl start zeppelin
