# Note: this script should be sourced into build-image.sh.

# Verifies the required scripts are available for a component.
function verify_component() {
  local -r component="$1"
  local -r install_script="${BDUTIL_DIR}/components/install/${component}.sh"
  local -r activate_script="${BDUTIL_DIR}/components/activate/${component}.sh"
  local -r pre_uninstall_script="${BDUTIL_DIR}/components/pre-uninstall/${component}.sh"
  local -r uninstall_script="${BDUTIL_DIR}/components/uninstall/${component}.sh"

  if [[ ! -f "${install_script}" ]]; then
    echo "Component ${component} does not have install script"
    return 1
  fi

  if [[ ! -f "${activate_script}" ]]; then
    echo "Component ${component} does not have activate script"
    return 1
  fi

  if [[ ! -f "${pre_uninstall_script}" && ! -f "${uninstall_script}" ]]; then
    echo "Component ${component} has neither pre-uninstall nor uninstall script"
    return 1
  fi

  return 0
}

function init_component() {
  local -r component="$1"

  if ! verify_component "${component}"; then
    exit 1
  fi

  local init_script="${BDUTIL_DIR}/components/initialize/${component}.sh"
  echo "Running component init script: ${init_script}"
  source "${init_script}"
}

# Called from build-image.sh to append optional components to global lists
# based on DATAPROC_IMAGE_VERSION version.
function initialize_components() {
  for script_name in $(ls "${BDUTIL_DIR}/components/initialize/"); do
    local component=${script_name%%.sh}
    if [[ -z "${COMPONENTS_TO_INSTALL}" ||
        "|${COMPONENTS_TO_INSTALL}|" == *"|${component}|"* ]]; then
      init_component "${component}"
    fi
  done
}

# Performs installation of components.
function install_components() {
  local components=("$@")

  # Iterate over all optional components and see if they have install
  # script.
  for component in "${components[@]}"; do
    local install_script="${BDUTIL_DIR}/components/install/${component}.sh"
    echo "Running component install script: ${install_script}"
    run_with_logger \
      --tag "install-component-${component}" \
      bash -e "${install_script}" "${DATAPROC_IMAGE_VERSION}"
    echo "Done with ${install_script}"
  done
}

# Performs post-installation steps on components.
#
# This is an optional step, so not every component should have
# a post-install file.
function post_install_components() {
  local components=("$@")
  run_components_scripts "post-install" "${components[@]}"
}
