# Generates makefile for a task graph.
function generate_task_graph_makefile() {
  local names=()
  read -r -a names <<<"$1"

  local scripts=()
  read -r -a scripts <<<"$2"

  local deps_manifests=()
  read -r -a deps_manifests <<<"$3"

  local -r sentinel_dir="$4"
  local -r task_args=("${@:5}")

  local targets=()
  for ((i = 0; i < "${#names[@]}"; i++)); do
    local name="${names[$i]}"
    local script="${scripts[$i]}"
    local deps_manifest="${deps_manifests[$i]}"

    local deps=()
    if [[ -f "${deps_manifest}" ]]; then
      local dep_names=($(cat "${deps_manifest}"))
      for ((j = 0; j < "${#dep_names[@]}"; j++)); do
        local dep_name="${dep_names[$j]}"
        # Skip dependencies which are not selected (e.g., hive-server2 depends
        # on hive-metastore only if the latter is selected). We don't validate
        # required dependencies here, it is validated at a higher level.
        if [[ " ${names[*]} " =~ " ${dep_name} " ]]; then
          deps+=("${sentinel_dir}/${dep_name}")
        fi
      done
    fi
    local target="${sentinel_dir}/${name}"
    targets+=("${target}")
    echo -e "${target}: ${deps[*]}"
    echo -e "\t@echo 'Running task: ${name}'"
    echo -e "\tbash -e ${script} ${task_args[*]} || { echo \"Error: ${script}\"; exit 1; }"
    echo -e "\ttouch ${target}\n"
  done

  echo -e "all: ${targets[*]}"
}

# Executes a task graph.
#
# Each task has a name, an executable script and an optional dependencies
# manifest. Dependencies manifests are text files with dependent task names one
# on each line.
function execute_task_graph() {
  echo "Generating makefile for the task graph"
  mkdir -p "${DATAPROC_TMP_DIR}/make"
  local makefile
  makefile=$(mktemp "${DATAPROC_TMP_DIR}/make/makefile.XXXXXX")
  generate_task_graph_makefile "$@" >"${makefile}"

  echo "Generated makefile:"
  cat "${makefile}"

  echo "Running task graph:"
  make -f "${makefile}" all
}

function run_components_scripts() {
  local -r script_type="$1"
  local -r all_components=("${@:2}")

  local components=()
  for component in "${all_components[@]}"; do
    local script="${BDUTIL_DIR}/components/${script_type}/${component}.sh"
    if [[ -f "${script}" ]]; then
      components+=("${component}")
    fi
  done
  echo "Components with ${script_type} script: ${components[*]}"

  local -r sentinel_dir="${DATAPROC_TMP_DIR}/sentinel"
  mkdir -p "${sentinel_dir}"

  local names=()
  local scripts=()
  local deps=()
  for cmp in "${components[@]}"; do
    names+=("${cmp}.${script_type}")
    scripts+=("${BDUTIL_DIR}/components/${script_type}/${cmp}.sh")
    deps+=("${BDUTIL_DIR}/components/${script_type}/${cmp}.deps")
  done

  execute_task_graph \
    "${names[*]}" \
    "${scripts[*]}" \
    "${deps[*]}" \
    "${sentinel_dir}" \
    "${DATAPROC_IMAGE_VERSION}"

  echo "All ${script_type} scripts done"
}
