#!/bin/bash

set -euo pipefail

source "${OS_BDUTIL_DIR}/bdutil_helpers.sh"
source "${BDUTIL_DIR}/components/shared/docker.sh"
source "${BDUTIL_DIR}/components/shared/spark.sh"

set -x

function install_docker() {
  os_update_repo
  os_install_packages docker.io
}

function remove_docker() {
  os_uninstall_packages docker.io

  # Make sure to delete network bridge docker0
  if ip link show docker0 2>&1; then
    ip link del docker0
  fi

  rm -rf /etc/docker /run/docker
}

# Creates the default Serverless Spark container image and pre-installs it in
# the VM image, so that at runtime there is no need to pull the image from the
# image registry, unless the user provides a custom container image.
function create_serverless_spark_container_image() {
  local container_build_dir
  container_build_dir=$(mktemp -d)
  local -r dockerfile="${container_build_dir}/spark.dockerfile"
  local -r image_name="dataproc.local/serverless-spark/default:${DATAPROC_IMAGE_VERSION}"

  systemctl daemon-reload
  systemctl enable docker
  systemctl start docker

  cat >"${dockerfile}" <<EOF
FROM debian:${OS_VERSION}-slim

SHELL ["/bin/bash", "-c"]

# Suppress interactive prompts
ENV DEBIAN_FRONTEND=noninteractive

# Install packages required by Spark scripts and default Dataproc packages
RUN apt-get update && apt-get -y install --no-install-recommends \
  apt-transport-https \
  ca-certificates \
  curl \
  gnupg \
  libopenblas0 \
  libssl-dev \
  procps \
  tini \
  libjemalloc2 \
  libgomp1 \
  && echo "deb [signed-by=/usr/share/keyrings/cloud.google.gpg] https://packages.cloud.google.com/apt cloud-sdk main" | \
    tee -a /etc/apt/sources.list.d/google-cloud-sdk.list \
  && curl -fsS --retry-connrefused --retry 10 --retry-max-time 30 https://packages.cloud.google.com/apt/doc/apt-key.gpg | \
    gpg --dearmor -o /usr/share/keyrings/cloud.google.gpg \
  && apt-get update && apt-get -y install --no-install-recommends google-cloud-cli \
  && apt-get clean && rm -rf /var/cache/apt/* /var/lib/apt/lists/* /tmp/*

# Enable jemalloc
ENV LD_PRELOAD=/usr/lib/x86_64-linux-gnu/libjemalloc.so.2

RUN groupadd -g "${SPARK_GID}" spark
RUN useradd -u "${SPARK_UID}" -g "${SPARK_GID}" -d /home/spark -m spark
USER spark
EOF

  docker build -f "${dockerfile}" -t "${image_name}" "${container_build_dir}"
  rm -r "${container_build_dir}"

  mkdir -p "$(dirname "${SPARK_DEFAULT_IMAGE_FILE}")"
  docker save -o "${SPARK_DEFAULT_IMAGE_FILE}" "${image_name}"
  chown "${SPARK_UID}:${SPARK_GID}" "${SPARK_DEFAULT_IMAGE_FILE}"

  systemctl stop docker
  systemctl disable docker

  cat >>/etc/environment <<EOF
SPARK_CONTAINER_IMAGE_DEFAULT=${image_name}
EOF
}

# Install ContainerD in serverless-spark images only.
if [[ "${DATAPROC_IMAGE_TYPE}" == "serverless-spark" ]]; then
  DATAPROC_OPTIONAL_COMPONENTS+=(containerd)
  DATAPROC_COMMON_PACKAGES+=(containerd)

  os_install_packages fuse

  # We can't create images using ctr/crcitl, so we temporarily install docker,
  # create the default image, and then remove docker.
  install_docker
  create_serverless_spark_container_image
  remove_docker
fi
