#!/bin/bash

set -euo pipefail

source "$(dirname "$0")/../shared/anaconda.sh"
source "$(dirname "$0")/../shared/conda.sh"
source "$(dirname "$0")/../../bdutil_env.sh"
source "$(dirname "$0")/../../bdutil_helpers.sh"

set -x

function install_anaconda() {
  local -r installer="${ANACONDA_VERSION}"
  gsutil cp "gs://conda-dist/${installer}" "/tmp/"

  bash "/tmp/${installer}" -b -p "${ANACONDA_INSTALL_PATH}"
  chmod 755 "${ANACONDA_INSTALL_PATH}"

  rm "/tmp/${installer}"
}

function configure_anaconda() {
  # Disable conda auto upgrade
  "${ANACONDA_BIN_DIR}/conda" config --system --set auto_update_conda False

  "${ANACONDA_BIN_DIR}/conda" config --append channels conda-forge

  # Prevent major.minor upgrades of Python and Conda.
  pin_component_version "${ANACONDA_INSTALL_PATH}" conda
  pin_component_version "${ANACONDA_INSTALL_PATH}" python
}

function install_anaconda_packages() {
  local packages=('jupyterhub=1.0' 'pyarrow=0.15' 'spylon-kernel=0.4')

  # ipyparallel and libiconv are required by spylon_kernel
  packages+=('ipyparallel=6.3' 'libiconv=1.15')

  # Install R and IR Jupyter kernels.
  packages+=(
    "r-recommended=${R_VERSION}"
    'r-irkernel=0.8'
    'r-ggplot2=3.1'
    'r-rcurl=1.95_4'
  )

  "${ANACONDA_BIN_DIR}/conda" install "${packages[@]}"
}

function install_conda_forge_packages() {
  # Jupyterlab extensions require recent Nodejs version
  local -r packages=('nodejs>=8')

  retry_constant_custom 4 1 "${ANACONDA_BIN_DIR}/conda" install -c conda-forge "${packages[@]}"
}

function install_pip_packages() {
  if [[ "${DATAPROC_IMAGE_VERSION}" == "1.5" ]]; then
    # The version of anaconda pinned in the 1.5 images comes with a Jupyter
    # notebook build that does not work with custom contents managers.
    #
    # We do not want to update all of Anaconda, as that would change the
    # versions of a lot of packages, so we just update the Jupyter Notebook
    # package by iteself.
    "${ANACONDA_BIN_DIR}/pip" install -U --no-deps \
      'jupyter-core~=4.6.1' 'jupyter-client~=5.3.4' 'notebook~=6.0.3'
  fi

  # Install relevant GCP client libraries by default
  local packages=(
    'google-api-core==1.32.0'
    'google-api-python-client==1.11.*'
    'google-cloud-core==1.4.*'
    'google-cloud-bigquery[pandas]==1.26.*'
    'google-cloud-bigquery-storage==1.0.*'
    'google-cloud-dataproc==1.1.*'
    'google-cloud-datastore==1.15.*'
    'google-cloud-language==1.3.*'
    'google-cloud-logging==1.15.*'
    'google-cloud-spanner==1.18.*'
    'google-cloud-storage==1.31.*'
    'google-cloud-translate==2.0.*'
    'six==1.15.*')

  # Use the sparkmonitor-s package for Tornado 6 compatibility (package name
  # changed due to lack of maintenance).
  # https://github.com/krishnan-r/sparkmonitor/issues/18
  packages+=('asyncio==3.4.*' 'sparkmonitor-s==0.0.*')

  "${ANACONDA_BIN_DIR}/pip" install --upgrade-strategy only-if-needed \
    "${packages[@]}"
}

function configure_packages() {
  # Prevent major.minor upgrades of CRAN R
  pin_component_version "${ANACONDA_INSTALL_PATH}" r-recommended
}

function cleanup_anaconda() {
  "${ANACONDA_BIN_DIR}/conda" clean --all
}

install_anaconda
configure_anaconda
install_anaconda_packages
install_conda_forge_packages
install_pip_packages
configure_packages
cleanup_anaconda
