#!/bin/bash
#
# Based on the go/riptide-client-integration

set -euo pipefail

source "$(dirname "$0")/../../bdutil_env.sh"
source "$(dirname "$0")/../../bdutil_logging.sh"
source "$(dirname "$0")/../../bdutil_networking.sh"
source "$(dirname "$0")/../../bdutil_retry.sh"
source "$(dirname "$0")/../shared/containerd.sh"

set -x

readonly CONTAINERD_VERSION="1.7.11"
readonly RUNC_VERSION="1.1.10"
readonly CNI_VERSION="1.4.0"
readonly CRICTL_VERSION="1.29.0"
readonly GCFSD_VERSION="0.196.0"
readonly GCFS_SNAPSHOTTER_VERSION="1.4-29"

function install_containerd() {
  download_resource \
    "https://github.com/containerd/containerd/releases/download/v${CONTAINERD_VERSION}/containerd-${CONTAINERD_VERSION}-linux-amd64.tar.gz" \
    /tmp/containerd.tgz
  tar -xzf /tmp/containerd.tgz -C /usr
  rm /tmp/containerd.tgz

  # Modified containerd service file based on
  # https://github.com/containerd/containerd/blob/main/containerd.service
  cat >/etc/systemd/system/containerd.service <<EOF
[Unit]
Description=containerd container runtime
Documentation=https://containerd.io
After=network.target local-fs.target

[Service]
ExecStartPre=-/sbin/modprobe overlay
ExecStart=/usr/bin/containerd

Type=notify
Delegate=yes
KillMode=process
Restart=always
RestartSec=5

# Having non-zero Limit*s causes performance problems due to accounting overhead
# in the kernel. We recommend using cgroups to do container-local accounting.
LimitNPROC=infinity
LimitCORE=infinity

# Comment TasksMax if your systemd version does not supports it.
# Only systemd 226 and above support this version.
TasksMax=infinity
OOMScoreAdjust=-999

StandardOutput=file:/var/log/containerd.out
StandardError=file:/var/log/containerd.err

[Install]
WantedBy=multi-user.target
EOF

  # Create config file
  mkdir "$(dirname "${CONTAINERD_CONFIG_FILE}")"
  cat >"${CONTAINERD_CONFIG_FILE}" <<EOF
version = 2
EOF
}

function install_runc() {
  download_resource \
    "https://github.com/opencontainers/runc/releases/download/v${RUNC_VERSION}/runc.amd64" \
    /usr/local/sbin/runc
  chmod 755 /usr/local/sbin/runc
}

function install_cni() {
  download_resource \
    "https://github.com/containernetworking/plugins/releases/download/v${CNI_VERSION}/cni-plugins-linux-amd64-v${CNI_VERSION}.tgz" \
    /tmp/cni.tgz
  mkdir -p /opt/cni/bin
  tar -xzf /tmp/cni.tgz -C /opt/cni/bin
  rm /tmp/cni.tgz

  # Use `install` to create config dirs because it sets correct permissions
  install -d /etc/cni/net.d
  cat >/etc/cni/net.d/10-containerd-net.conflist <<EOF
{
  "cniVersion": "1.0.0",
  "name": "containerd-net",
  "plugins": [
    {
      "type": "bridge",
      "bridge": "cni0",
      "isGateway": true,
      "ipMasq": true,
      "promiscMode": true,
      "ipam": {
        "type": "host-local",
        "ranges": [
          [{
            "subnet": "10.88.0.0/16"
          }],
          [{
            "subnet": "2001:4860:4860::/64"
          }]
        ],
        "routes": [
          { "dst": "0.0.0.0/0" },
          { "dst": "::/0" }
        ]
      }
    },
    {
      "type": "portmap",
      "capabilities": {"portMappings": true}
    }
  ]
}
EOF
}

function install_crictl() {
  download_resource \
    "https://github.com/kubernetes-sigs/cri-tools/releases/download/v${CRICTL_VERSION}/crictl-v${CRICTL_VERSION}-linux-amd64.tar.gz" \
    /tmp/crictl.tgz
  tar -xzf /tmp/crictl.tgz -C /usr/local/sbin
  rm /tmp/crictl.tgz

  crictl config --set runtime-endpoint="unix:///run/containerd/containerd.sock"
  crictl config --set timeout=90
}

function install_gcfsd() {
  download_resource \
    "gs://gke-release/gcfsd/v${GCFSD_VERSION}/gcfsd" \
    /usr/local/sbin/gcfsd
  chmod 755 /usr/local/sbin/gcfsd

  local -r cmd=(
    '/usr/local/sbin/gcfsd'
    '--mount_point=/run/gcfsd/mnt'
    '--metrics_flavor=""'
    '--log_level=info'
    '--max_content_cache_size_mb=256'
    '--event_reporter_type=none'
    '--client_name=dataproc'
    "--client_version=${DATAPROC_IMAGE_TYPE}-${DATAPROC_IMAGE_VERSION}"
    '--layer_cache_dir=/var/dataproc/lib/containerd/io.containerd.snapshotter.v1.gcfs/snapshotter/layers'
    '--images_in_use_db_path=/var/dataproc/lib/containerd/io.containerd.snapshotter.v1.gcfs/gcfsd/images_in_use.db'
  )
  cat >/etc/systemd/system/gcfsd.service <<EOF
[Unit]
Description=Google Container File System service
After=network.target

[Service]
Type=simple
Environment=GOGC=10
ExecStartPre=-/bin/mkdir -p /run/gcfsd/mnt
ExecStartPre=/bin/mkdir -p /var/dataproc/lib/containerd/io.containerd.snapshotter.v1.gcfs/snapshotter/layers
ExecStartPre=/bin/mkdir -p /var/dataproc/lib/containerd/io.containerd.snapshotter.v1.gcfs/gcfsd
ExecStart=${cmd[*]}
ExecStop=/bin/umount /run/gcfsd/mnt
Nice=-20
StandardOutput=file:/var/log/gcfsd.out
StandardError=file:/var/log/gcfsd.err

[Install]
WantedBy=multi-user.target
EOF
}

function install_gcfs_snapshotter() {
  download_resource \
    "gs://gke-release/gcfs-snapshotter/v${GCFS_SNAPSHOTTER_VERSION}/containerd-gcfs-grpc" \
    /usr/local/sbin/containerd-gcfs-grpc
  chmod 755 /usr/local/sbin/containerd-gcfs-grpc

  mkdir /etc/containerd-gcfs-grpc/
  touch /etc/containerd-gcfs-grpc/config.toml

  local -r cmd=(
    '/usr/local/sbin/containerd-gcfs-grpc'
    '--log-level=info'
    '--enable-kubernetes-event-reporting=false'
    '--config=/etc/containerd-gcfs-grpc/config.toml'
  )
  cat >/etc/systemd/system/gcfs-snapshotter.service <<EOF
[Unit]
Description=Google Container File System Snapshotter service
After=network.target
Before=containerd.service

[Service]
Environment=HOME=/root
ExecStart=${cmd[*]}
Restart=always
RestartSec=1
StandardOutput=file:/var/log/gcfs-snapshotter.out
StandardError=file:/var/log/gcfs-snapshotter.err

[Install]
WantedBy=multi-user.target
EOF
}

function download_licenses() {
  mkdir -p "${DATAPROC_DIR}/licenses"
  download_resource \
    "https://raw.githubusercontent.com/containerd/containerd/master/LICENSE" \
    "${DATAPROC_DIR}/licenses/CONTAINERD_LICENSE"
  download_resource \
    "https://raw.githubusercontent.com/kubernetes-sigs/cri-tools/master/LICENSE" \
    "${DATAPROC_DIR}/licenses/CRICTL_LICENSE"
}

install_containerd
install_runc
install_cni
install_crictl
install_gcfsd
install_gcfs_snapshotter
download_licenses
