#!/bin/bash

set -euo pipefail

source "$(dirname "$0")/../../bdutil_logging.sh"
source "$(dirname "$0")/../../bdutil_properties.sh"
source "$(dirname "$0")/../../bdutil_retry.sh"
source "$(dirname "$0")/../shared/dpms-proxy.sh"

set -x

# Proxy jar files are staged into a bucket until the proxy builds onboard to
# CI/CD.  This is the expected structure:
#
# dataproc-image-dependencies/dpms-proxy/
#     <docker-image-version>/
#         <hive-version>/
#             hms-proxy- <docker-image-version>-<hive-version>.jar
#
# Proxy builds are delivered as Docker images.  To unpack the jar from a
# Docker image and stage it into GCS, run a command like this:
#
# DIST_BUCKET=dataproc-image-dependencies/dpms-proxy
# DOCKER_IMAGE_VERSION=v0.0.45
# HIVE2_VERSION=2.3.6
# HIVE3_VERSION=3.1.2
#
# docker run --entrypoint /bin/bash gcr.io/cloud-metastore-public/hms-grpc-proxy/${HIVE2_VERSION}:${DOCKER_IMAGE_VERSION} \
#     -c 'cat /usr/src/hms-proxy/hms-proxy.jar' |
#     gsutil cp - "gs://${DIST_BUCKET}/${DOCKER_IMAGE_VERSION}/${HIVE2_VERSION}/hms-proxy-${DOCKER_IMAGE_VERSION}-${HIVE2_VERSION}.jar"
#
# docker run --entrypoint /bin/bash gcr.io/cloud-metastore-public/hms-grpc-proxy/${HIVE3_VERSION}:${DOCKER_IMAGE_VERSION} \
#     -c 'cat /usr/src/hms-proxy/hms-proxy.jar' |
#     gsutil cp - "gs://${DIST_BUCKET}/${DOCKER_IMAGE_VERSION}/${HIVE3_VERSION}/hms-proxy-${DOCKER_IMAGE_VERSION}-${HIVE3_VERSION}.jar"
function install_dpms_proxy() {
  mkdir -p "${DPMS_PROXY_INSTALL_DIR}"
  local dpms_proxy_bucket
  dpms_proxy_bucket=$(get_dataproc_property_or_default \
      'attributes/dataproc_dpms_proxy_bucket' \
      'gs://dataproc-image-dependencies/dpms-proxy')

  if [[ ! -v DPMS_PROXY_HIVE_VERSIONS || -z "${DPMS_PROXY_HIVE_VERSIONS[@]}" ]]; then
    loginfo "Variable DPMS_PROXY_HIVE_VERSIONS not found in image_config."
    exit 1
  fi
  local -r hive_versions=("${DPMS_PROXY_HIVE_VERSIONS[@]}")

  for hive_version in "${hive_versions[@]}"; do
    retry_constant_custom 5 1 gsutil -q cp -r \
        "${dpms_proxy_bucket}/${DPMS_PROXY_DOCKER_IMAGE_VERSION}/${hive_version}" \
        "${DPMS_PROXY_INSTALL_DIR}"
    # Create a symlink without Docker image version for backwards compatibility.
    ln -s \
        "${DPMS_PROXY_INSTALL_DIR}/${hive_version}/hms-proxy-${DPMS_PROXY_DOCKER_IMAGE_VERSION}-${hive_version}.jar" \
        "${DPMS_PROXY_INSTALL_DIR}/${hive_version}/hms-proxy-${hive_version}.jar"
  done

  # If something is wrong with the bucket path and no files are found, then
  # gsutil will fail fast with a non-zero exit code, so we don't need any
  # additional checks after copying.
}

install_dpms_proxy
