#!/bin/bash

set -euo pipefail

source "$(dirname "$0")/../../bdutil_env.sh"
source "$(dirname "$0")/../../bdutil_helpers.sh"
source "$(dirname "$0")/../shared/conda.sh"
source "$(dirname "$0")/../shared/miniconda3.sh"

set -x

CONDA_BIN="${MINICONDA3_BIN_DIR}/conda"

function install_miniconda() {
  local installer="${MINICONDA_VERSION}"
  if is_arm64; then
    installer="${MINICONDA_VERSION/x86_64/aarch64}"
  fi

  download_resource \
    "gs://dataproc-image-dependencies/conda/${installer}" "/tmp/"

  # `HOME=none` is a workaround for Mambaforge installation
  HOME=none bash "/tmp/${installer}" -b -p "${MINICONDA3_INSTALL_PATH}"
  chmod 755 "${MINICONDA3_INSTALL_PATH}"

  rm "/tmp/${installer}"
}

function configure_miniconda() {
  "${CONDA_BIN}" config --system --set always_yes True
  "${CONDA_BIN}" config --system --set auto_update_conda False
  "${CONDA_BIN}" config --system --set changeps1 False

  if [[ "${DATAPROC_IMAGE_TYPE}" == "standard" ]] &&
      is_version_at_least "${DATAPROC_IMAGE_VERSION}" "2.0"; then
    # Use conda-forge channel by default because it has most up-to-date packages
    "${CONDA_BIN}" config --system --prepend channels conda-forge
    # Use strict channel priority to prevent
    # slow cross-channel dependencies resolution
    "${CONDA_BIN}" config --system --set channel_priority strict
  fi
}

function install_miniconda_packages() {
  local packages=()

  if [[ "${DATAPROC_IMAGE_TYPE}" == "serverless-spark" ]]; then
    packages+=("${CONDA_PACKAGES[@]}")
    readonly CONDA_BIN="${MINICONDA3_BIN_DIR}/mamba"
  elif is_version_at_least "${DATAPROC_IMAGE_VERSION}" "2.1"; then
    configure_mamba "mamba=${MAMBA_VERSION}"
    packages+=("${CONDA_PACKAGES[@]}")
  elif is_version_at_least "${DATAPROC_IMAGE_VERSION}" "2.0"; then
    configure_mamba 'mamba=0.15'
    packages+=(
      "conda=${CONDA_VERSION}"
      'cython=0.29'
      'dask=2020.12'
      'fastavro=1.4'
      'fastparquet=0.5'
      'gcsfs=0.8'
      'ipykernel=6.14'
      'ipyparallel=6.3'
      'ipython-sql=0.3'
      'jupyterhub=1.3'
      'jupyterlab=3.0'
      'koalas=1.4'
      'matplotlib=3.4'
      'mistune=0.8.4' # b/209528071 Pinning as recent release has breaking change.
      'nltk=3.6'
      'notebook=6.2'
      'nodejs=16.13'
      'numba=0.53'
      'numpy=1.19'
      'openblas=0.3'
      'orc=1.6'
      'pandas=1.2'
      'pyarrow=3.0'
      'pyhive=0.6'
      'pysal=2.4'
      'pytables=3.6'
      "python=${PYTHON_VERSION}"
      'r-ggplot2=3.3'
      'r-irkernel=1.1'
      'r-rcurl=1.98_1'
      "r-recommended=${R_VERSION}"
      'regex=2021.4'
      'requests=2.25'
      'rtree=0.9'
      'scikit-image=0.18'
      'scikit-learn=0.24'
      'scipy=1.6'
      'seaborn=0.11'
      'spyder=5.0'
      'spylon-kernel=0.4'
      'sqlalchemy=1.4'
      'sympy=1.8'
      'virtualenv=20.4'
      # Jupyter-specific packages: install these with default conda env to make
      # solving the environment easier.
      'bcolz=1.2'
      'cookiecutter=1.7'
      'httplib2=0.19'
      'ipywidgets=7.6'
      'jinja2=3.0' # b/228060250 pinning as jinja2 3.1 breaks with nbconvert 6.0
      'jupyter_contrib_nbextensions=0.5'
      'jupyter_core=4.10.0' # b/239566378 jupyter_core=4.11.0 causes jupyter module not found
      'jupyter_http_over_ws=0.0'
      'jupyterlab-git=0.30'
      'jupyterlab_widgets=1.0'
      'markdown=3.3'
      'nbdime=3.0'
      'nbconvert=6.0'
      'oauth2client=4.1'
      'opencv=4.5'
      'pandas-profiling=2.11'
      'papermill[gcs]=2.2'
      'pydot=1.4'
      'tornado=6.1'
      'uritemplate=3.0'
    )
  else
    packages+=(
      'conda=4.7'
      'importlib-metadata=4.11.4' # b/250769257 pinning as importlib-metadata has removed deprecated endpoint
      'numpy=1.19'
      'pyarrow=0.15'
      'python=3.7'
      "r-base=${R_VERSION}"
    )
  fi

  retry_constant_custom 4 1 "${CONDA_BIN}" install -q "${packages[@]}"

  # Prevent major.minor upgrades of some packages
  local pinned_packages=('conda' 'python' 'r-base')
  if [[ "${DATAPROC_IMAGE_TYPE}" == "serverless-spark" ]] ||
      is_version_at_least "${DATAPROC_IMAGE_VERSION}" "2.0"; then
    pinned_packages+=('r-recommended')
  fi
  for package in "${pinned_packages[@]}"; do
    pin_component_version "${MINICONDA3_INSTALL_PATH}" "${package%%[[~=<>]*}"
  done
}

function install_pip_packages() {
  local packages=()
  if [[ "${DATAPROC_IMAGE_TYPE}" == "standard" ]] && is_version_at_least "${DATAPROC_IMAGE_VERSION}" "2.0"; then
    if is_version_at_least "${DATAPROC_IMAGE_VERSION}" "2.1"; then
      packages=("${PIP_PACKAGES[@]}")
    elif [[ "${DATAPROC_IMAGE_VERSION}" == "2.0" ]]; then
      packages=(
        'google-auth-oauthlib==0.5.*'
        'google-cloud-bigquery[pandas]==2.6.*'
        'google-cloud-bigquery-storage==2.1.*'
        'google-cloud-bigtable==1.6.*'
        'google-cloud-container==2.3.*'
        'google-cloud-datacatalog==3.0.*'
        'google-cloud-dataproc==2.2.*'
        'google-cloud-datastore==2.1.*'
        'google-cloud-language==2.0.*'
        'google-cloud-logging==2.1.*'
        'google-cloud-monitoring==2.0.*'
        'google-cloud-pubsub==2.2.*'
        'google-cloud-redis==2.0.*'
        'google-cloud-spanner==2.1.*'
        'google-cloud-speech==2.0.*'
        'google-cloud-storage==1.35.*'
        'google-cloud-texttospeech==2.2.*'
        'google-cloud-translate==3.0.*'
        'google-cloud-vision==2.0.*'
      )
    fi
  elif [[ "${DATAPROC_IMAGE_TYPE}" == "serverless-spark" ]]; then
    packages=("${PIP_PACKAGES[@]}")
  fi
  if [[ ${#packages[@]} != 0 ]]; then
    retry_constant_custom 4 1 "${MINICONDA3_BIN_DIR}/pip" install \
      --progress-bar off --upgrade-strategy only-if-needed "${packages[@]}"
  fi
}

function configure_conda_dir_permission() {
  if [[ "${DATAPROC_IMAGE_TYPE}" == "serverless-spark" ]]; then
    # Allow user "spark" could install/uninstall pip and conda packages in
    # serverless workloads
    chown -R spark:spark "${MINICONDA3_INSTALL_PATH}"
  fi
}

function cleanup_miniconda() {
  "${CONDA_BIN}" clean --all
}

function configure_mamba() {
  # Use Mamba to install conda packages because it's much faster
  local -r mamba_package=$1
  retry_constant_custom 4 1 "${CONDA_BIN}" install "${mamba_package}" -n base \
    -c conda-forge
  readonly CONDA_BIN="${MINICONDA3_BIN_DIR}/mamba"
}

install_miniconda
configure_miniconda
install_miniconda_packages
install_pip_packages
configure_conda_dir_permission
cleanup_miniconda
