#!/bin/bash

set -euo pipefail

source "$(dirname "$0")/../../bdutil_env.sh"
source "$(dirname "$0")/../../bdutil_helpers.sh"
source "$(dirname "$0")/../shared/spark.sh"

set -x

readonly DATAPROC_MAVEN_REPO_URL='https://us-central1-maven.pkg.dev/cloud-dataproc-prod/dataproc/com/google/cloud/dataproc'

function is_spark_version_at_least_3_4() {
  [[ "${DATAPROC_IMAGE_VERSION}" == '1.2' ]] || is_version_at_least "${DATAPROC_IMAGE_VERSION}" '2.1'
}

function install_spark_packages() {
  local spark_packages=(spark-core spark-datanucleus spark-python spark-r)
  if [[ "${DATAPROC_IMAGE_TYPE}" == "serverless-spark" ]]; then
    if is_spark_version_at_least_3_4; then
      spark_packages+=(spark-connector)
    else
      spark_packages+=(spark-external)
    fi
  else
    spark_packages+=(spark-external spark-history-server spark-yarn-shuffle)
  fi

  os_install_packages "${spark_packages[@]}" ||
    err "Failed to install packages for Spark (${DATAPROC_IMAGE_TYPE})"
}

# Install Spark BigQuery connector
install_bq_connector() {
  if [[ ! -v SPARK_BIGQUERY_CONNECTOR_VERSION ]]; then
    return
  fi

  if [[ "${DATAPROC_IMAGE_TYPE}" == "serverless-spark" ]] ||
      is_version_at_least "${DATAPROC_IMAGE_VERSION}" '2.2'; then
    local -r download_dir="${SPARK_JARS_DIR}"
  else
    local -r download_dir="${DATAPROC_COMMON_LIB_DIR}"
  fi

  if [[ "${DATAPROC_IMAGE_TYPE}" == 'standard' ]] &&
      is_version_at_least "${DATAPROC_IMAGE_VERSION}" '2.2'; then
    local -r jar_name="spark-3.4-bigquery-${SPARK_BIGQUERY_CONNECTOR_VERSION}.jar"
  elif [[ "${DATAPROC_IMAGE_TYPE}" == "serverless-spark" ]] &&
      ([[ "${DATAPROC_IMAGE_VERSION}" == '1.2' ]] ||
        is_version_at_least "${DATAPROC_IMAGE_VERSION}" '2.2'); then
    local spark_version
    spark_version=$(dpkg -s spark-core | grep "^Version: " | cut -d' ' -f2 | cut -d'-' -f1)
    local -r jar_name="spark-${spark_version%.*}-bigquery-${SPARK_BIGQUERY_CONNECTOR_VERSION}.jar"
  else
    local -r jar_name="spark-bigquery-with-dependencies_${SPARK_BIGQUERY_CONNECTOR_VERSION}.jar"
  fi

  download_resource \
    "gs://spark-lib/bigquery/${jar_name}" \
    "${download_dir}/${jar_name}"
}

# Installs Dataproc Spark Plugins
#
# The plugin registers GCS connector metrics with Spark metrics system.
function install_dataproc_spark_plugins() {
  local -r name='dataproc-spark-plugins'
  local -r version='1.0.0'
  local -r jar_name="${name}-${version}.jar"

  download_resource \
    "${DATAPROC_MAVEN_REPO_URL}/${name}/${version}/${jar_name}" \
    "${SPARK_JARS_DIR}/${jar_name}"
}

function install_spark_metrics_listener() {
  local -r name='spark-metrics-listener'
  local -r base_version='1.8.0'

  if [[ "${DATAPROC_IMAGE_TYPE}" == "serverless-spark" ]]; then
    local -r version="dataproc-s8s-${DATAPROC_IMAGE_VERSION}-${base_version}"
  else
    if is_version_at_least "${DATAPROC_IMAGE_VERSION}" "2.0"; then
      local -r version="dataproc-${DATAPROC_IMAGE_VERSION}-${base_version}"
    else
      # for older images (ex: 1.5.x) use older version of spark metric listener
      local -r version='1.1.0'
    fi
  fi

  local -r jar_name="${name}-${version}.jar"
  if [[ "${DATAPROC_IMAGE_TYPE}" == 'serverless-spark' ]] ||
      is_version_at_least "${DATAPROC_IMAGE_VERSION}" '2.2'; then
    download_resource \
      "${DATAPROC_MAVEN_REPO_URL}/${name}/${version}/${jar_name}" \
      "${SPARK_JARS_DIR}/dataproc-${jar_name}"
  else
    download_resource \
      "${DATAPROC_MAVEN_REPO_URL}/${name}/${version}/${jar_name}" \
      "${DATAPROC_COMMON_LIB_DIR}/${jar_name}" \
      "${DATAPROC_COMMON_LIB_DIR}/${name}.jar"
  fi
}

function install_spark_dataproc_buffer() {
  download_resource \
    "${SPARK_DATAPROC_BUFFER_URI}" \
    "${SPARK_CONNECTOR_DIR}/$(basename "${SPARK_DATAPROC_BUFFER_URI}")"
}

function install_rapids() {
  local -r scala_version="${SPARK_RAPIDS_VERSION%-*}"

  download_resource \
    "gs://maven-central/maven2/com/nvidia/rapids-4-spark_${scala_version}/${SPARK_RAPIDS_VERSION#*-}/rapids-4-spark_${SPARK_RAPIDS_VERSION}.jar" \
    "${SPARK_CONNECTOR_DIR}"
}

function install_xgboost() {
  local -r scala_version="${XGBOOST_VERSION%-*}"

  # TODO: Remove this `if` condition when XGBoost GPU supports Scala 2.13
  if [[ "${scala_version}" == '2.12' ]]; then
    download_resource \
      "gs://maven-central/maven2/ml/dmlc/xgboost4j-gpu_${scala_version}/${XGBOOST_VERSION#*-}/xgboost4j-gpu_${XGBOOST_VERSION}.jar" \
      "${SPARK_CONNECTOR_DIR}"
    download_resource \
      "gs://maven-central/maven2/ml/dmlc/xgboost4j-spark-gpu_${scala_version}/${XGBOOST_VERSION#*-}/xgboost4j-spark-gpu_${XGBOOST_VERSION}.jar" \
      "${SPARK_CONNECTOR_DIR}"
  fi
}

install_spark_packages
if [[ "${DATAPROC_IMAGE_TYPE}" == "standard" ]]; then
  systemctl disable spark-history-server.service
fi
install_bq_connector
install_dataproc_spark_plugins

if [[ "${DATAPROC_IMAGE_TYPE}" == "serverless-spark" ]]; then
  install_rapids
  install_xgboost
fi

# Install spark-dataproc-buffer for spark application context ingestion in serverless.
if [[ "${DATAPROC_IMAGE_TYPE}" == "serverless-spark" ]] &&
    [[ -v SPARK_DATAPROC_BUFFER_URI ]]; then
  install_spark_dataproc_buffer
fi

# Do not install Spark listener on 3.0+ s8s runtime as it's not supported, yet
if [[ "${DATAPROC_IMAGE_TYPE}" != "serverless-spark" ]] ||
    ! is_version_at_least "${DATAPROC_IMAGE_VERSION}" '3.0'; then
  install_spark_metrics_listener
fi

# Install the proxy that allows us to bridge TCP connections to Spark Connect.
if [[ "${DATAPROC_IMAGE_TYPE}" == "serverless-spark" ]] &&
    is_spark_version_at_least_3_4; then
  install_golang_binary "github.com/google/inverting-proxy/utils/tcpbridge/tcp-bridge-backend@latest"
fi
