#!/bin/bash

set -euo pipefail

source "$(dirname "$0")/../../bdutil_metadata.sh"
source "$(dirname "$0")/../../bdutil_misc.sh"
source "$(dirname "$0")/../shared/containerd.sh"
source "$(dirname "$0")/../shared/spark.sh"

set -x

# Latest pause image version can be checked in
# https://github.com/kubernetes/kubernetes/blob/v1.30.0/build/dependencies.yaml#L194-L195
readonly PAUSE_IMAGE_NAME='pause:3.9'
readonly PAUSE_REMOTE_IMAGE="registry.k8s.io/${PAUSE_IMAGE_NAME}"
readonly PAUSE_LOCAL_IMAGE="dataproc.local/${PAUSE_IMAGE_NAME}"

readonly MONITORING_AGENT_IMAGE="gcr.io/cloud-dataproc/observability/monitoring-agent:latest"
readonly MONITORING_AGENT_LOCAL_IMAGE="dataproc.local/monitoring-agent"
readonly MONITORING_AGENT_IMAGE_FILE="${DATAPROC_DIR}/images/monitoring-agent-image.tar"

function create_grant_log_read_script() {
  # Create script that grants everyone access to containerd log. Currently,
  # containerd creates the log such that only root has read permission
  local -r grant_log_read_script="/usr/local/bin/grant_container_log_read.sh"
  cat >"${grant_log_read_script}" <<EOF
#!/bin/bash
chmod a+r ${CONTAINERD_LOG_DIR}/${CONTAINERD_LOG_FILE}
EOF

  chmod u+x "${grant_log_read_script}"

  # Give spark user permission to run this script as root
  echo "spark ALL=(root) NOPASSWD:${grant_log_read_script}" | sudo EDITOR='tee -a' visudo
}

# Configure containerd socket to allow spark user to issue commands
function configure_containerd_spark_user_access() {
  cat >>"${CONTAINERD_CONFIG_FILE}" <<EOF

[grpc]
  address = "/run/containerd/containerd.sock"
  uid = ${SPARK_UID}
  gid = ${SPARK_GID}
EOF
}

function import_default_images() {
  systemctl enable containerd
  systemctl start containerd

  # Pull the sandbox/pause image and re-tag it with local tag. This is done to
  # make sure that we always use a local container image and never fallback to
  # fetching it from the remote repo as it may not be available at runtime.
  local token
  token=$(get_auth_token)
  crictl pull --creds="oauth2accesstoken:${token}" "${PAUSE_REMOTE_IMAGE}"
  ctr --namespace=k8s.io images tag "${PAUSE_REMOTE_IMAGE}" "${PAUSE_LOCAL_IMAGE}"

  # Install containerised agents for s8s debian-12 onwards
  if is_containerised_legacy_agents_supported_for_s8s; then
    ctr --namespace=k8s.io images pull "${MONITORING_AGENT_IMAGE}"
    ctr --namespace=k8s.io images tag "${MONITORING_AGENT_IMAGE}" "${MONITORING_AGENT_LOCAL_IMAGE}"
    mkdir -p "$(dirname "${MONITORING_AGENT_IMAGE_FILE}")"
    ctr --namespace=k8s.io images export "${MONITORING_AGENT_IMAGE_FILE}" "${MONITORING_AGENT_LOCAL_IMAGE}"
  fi

  # Export pulled sandbox/pause image to the tar file
  mkdir -p "$(dirname "${PAUSE_IMAGE_FILE}")"
  ctr --namespace=k8s.io images export "${PAUSE_IMAGE_FILE}" "${PAUSE_LOCAL_IMAGE}"

  # Remove all the cached images because we want to import only local images
  # that are necessary at runtime
  local -a images_to_remove
  images_to_remove=($(ctr --namespace=k8s.io images list -q))
  ctr --namespace=k8s.io images remove "${images_to_remove[@]}"

  # Import local images that are necessary at runtime
  ctr --namespace=k8s.io images import --digests=true "${PAUSE_IMAGE_FILE}"
  ctr --namespace=k8s.io images import --digests=true "${SPARK_DEFAULT_IMAGE_FILE}"
  rm "${SPARK_DEFAULT_IMAGE_FILE}"

  if is_containerised_legacy_agents_supported_for_s8s; then
    ctr --namespace=k8s.io images import --digests=true "${MONITORING_AGENT_IMAGE_FILE}"
  fi

  systemctl stop containerd
  systemctl disable containerd

  # Configure containerd to use the pulled pause image
  cat >>"${CONTAINERD_CONFIG_FILE}" <<EOF
[plugins]
  [plugins."io.containerd.grpc.v1.cri"]
    sandbox_image = "${PAUSE_LOCAL_IMAGE}"
  [plugins."io.containerd.grpc.v1.cri".containerd]
EOF
}

create_grant_log_read_script
configure_containerd_spark_user_access
import_default_images
