#!/bin/bash

set -euo pipefail

source "$(dirname "$0")/../../bdutil_logging.sh"
source "$(dirname "$0")/../../bdutil_os.sh"
source "$(dirname "$0")/../../bdutil_retry.sh"
source "$(dirname "$0")/../shared/dpms-proxy.sh"
source "$(dirname "$0")/../shared/hive.sh"

set -x

readonly DPMS_PROXY_STDOUT="${DPMS_PROXY_LOG_DIR}/dpms-proxy.out"
readonly DPMS_PROXY_STDERR="${DPMS_PROXY_LOG_DIR}/dpms-proxy.err"

# The proxy runs as the Hive user.  If Hive is installed, then this user already
# exists.  If Hive is not installed (e.g. serverless), then create the user.
function create_dpms_proxy_user_if_not_exists() {
  if user_exists "${DPMS_PROXY_USER}"; then
    loginfo "Dataproc Metastore Proxy user ${DPMS_PROXY_USER} already exists."
  else
    useradd -d "/var/lib/${DPMS_PROXY_USER}" \
        -s /bin/false \
        -c 'Dataproc Metastore Proxy User' \
        "${DPMS_PROXY_USER}"
    loginfo "Dataproc Metastore Proxy user ${DPMS_PROXY_USER} created."
  fi
}

# Generate Log4J configuration.
function configure_dpms_proxy_log4j() {
  mkdir -p "${DPMS_PROXY_CONF_DIR}"
  cat >"${DPMS_PROXY_CONF_DIR}/log4j2.properties" <<EOF
rootLogger.level = info
rootLogger.appenderRef.rolling.ref = rolling

# Set up a rolling file.
appender.rolling.type = RollingFile
appender.rolling.name = rolling
appender.rolling.fileName = ${DPMS_PROXY_LOG_DIR}/dpms-proxy.log
appender.rolling.filePattern = ${DPMS_PROXY_LOG_DIR}/dpms-proxy-%d{MM-dd-yy}-%i.log.gz
appender.rolling.layout.type = PatternLayout
appender.rolling.layout.pattern = %d{ISO8601} %5p [%t] %c{2}: %m%n
appender.rolling.policies.type = Policies

# Rollover on process restart.
appender.rolling.policies.startup.type = OnStartupTriggeringPolicy

# Rollover on start of new day.
appender.rolling.policies.time.type = TimeBasedTriggeringPolicy
appender.rolling.policies.time.interval = 1
appender.rolling.policies.time.modulate = true

# Rollover if file size grows too large.
appender.rolling.policies.size.type = SizeBasedTriggeringPolicy
appender.rolling.policies.size.size=10MB

# Retain no more than 5 rolled files (plus current active file), so maximum
# disk consumption is 60 MB.
appender.rolling.strategy.type = DefaultRolloverStrategy
appender.rolling.strategy.max = 5
EOF
}

# The proxy should write all logs through Log4J.  Just in case something slips
# out to stdout/stderr, cover it with logrotate.
function configure_dpms_proxy_logrotate() {
  cat >"${DPMS_PROXY_LOGROTATE_CONF}" <<EOF
$DPMS_PROXY_STDOUT
$DPMS_PROXY_STDERR
{
    daily
    copytruncate
    rotate 5
    compress
    delaycompress
    dateext
    dateformat -%m-%d-%Y
}
EOF
}

# Generate systemd service file.
function create_dpms_proxy_systemd_service_file() {
  if [[ "${DATAPROC_IMAGE_TYPE}" == "serverless-spark" ]]; then
    local -r max_heap='-Xmx512m'
  fi

  cat >"${DPMS_PROXY_SYSTEMD_SERVICE_FILE}" <<EOF
[Unit]
Description=Dataproc Metastore Proxy
Wants=network-online.target
After=network-online.target

[Service]
WorkingDirectory=${DPMS_PROXY_INSTALL_DIR}
User=${DPMS_PROXY_USER}
Group=${DPMS_PROXY_USER}
Type=exec

ExecStart=${JAVA_HOME}/bin/java \
    -cp ${DPMS_PROXY_CONF_DIR}:${HADOOP_CONF_DIR}:${HIVE_CONF_DIR}:${DPMS_PROXY_INSTALL_DIR}/hms-proxy.jar \
    ${max_heap:-} com.google.App

StandardOutput=file:${DPMS_PROXY_STDOUT}
StandardError=file:${DPMS_PROXY_STDERR}

Restart=on-failure
RestartSec=5

[Install]
WantedBy=multi-user.target
EOF

  retry_constant_short systemctl daemon-reload
}

function post_install_dpms_proxy() {
  create_dpms_proxy_user_if_not_exists
  configure_dpms_proxy_log4j
  configure_dpms_proxy_logrotate
  create_dpms_proxy_systemd_service_file
}

post_install_dpms_proxy
