#!/bin/bash

set -euo pipefail

source "$(dirname "$0")/../../bdutil_env.sh"
source "$(dirname "$0")/../../bdutil_helpers.sh"
source "$(dirname "$0")/../../bdutil_versions.sh"
source "$(dirname "$0")/../shared/hive.sh"

set -x

set_log_tag "post-install-component-hive-server2"

readonly HIVE_CONF_DIR="/etc/hive/conf"
readonly HIVE_SCRATCH_DIR="/var/tmp/hive-scratch"

function configure_hive_hbase() {
  # If /usr/lib/hbase exists, /usr/lib/hive/bin/hive automatically adds HBase
  # classpath to Hive during HiveServer2 startup, which could cause problems if
  # HBase is not selected but the debian packages are not uninstalled yet. We
  # want to avoid the race condition with the SKIP_HBASECP env variable, i.e.,
  # set it to true by default at image build time, update it to false in
  # hive-server2 activate script if HBase is selected at runtime.
  sed -i 's#^SKIP_HBASECP=false$#SKIP_HBASECP=${SKIP_HBASECP:-false}#' /usr/lib/hive/bin/hive

  local -r config_dir=/etc/systemd/system/hive-server2.service.d
  mkdir -p "${config_dir}"
  cat >"${config_dir}/hive-hbase.conf" <<EOF
[Service]
Environment=SKIP_HBASECP=true
EOF
  systemctl daemon-reload
}

# Configures Hive scratch dir, so that when driver processes run as non-root
# (e.g., run as `dataproc` in hermetic VM mode), they have write access to it.
function configure_hive_scratch_dir() {
  set_property_in_xml \
    "${HIVE_CONF_DIR}/hive-site.xml" \
    "hive.exec.local.scratchdir" \
    "${HIVE_SCRATCH_DIR}"
  set_property_in_xml \
    "${HIVE_CONF_DIR}/hive-site.xml" \
    "hive.exec.scratchdir" \
    "${HIVE_SCRATCH_DIR}"
  mkdir -p "${HIVE_SCRATCH_DIR}"
  chown hive:hive "${HIVE_SCRATCH_DIR}"
  chmod 1777 "${HIVE_SCRATCH_DIR}"
}

function configure_hive_classpath() {
  cat >>/etc/hive/conf/hive-env.sh <<EOF

# Add Tez to classpath
HADOOP_CLASSPATH=\\
"\${HADOOP_CLASSPATH}:/usr/lib/tez/*:/usr/lib/tez/lib/*:/etc/tez/conf:"

# Add Dataproc common libs (e.g., connectors) to Hadoop class path.
export HADOOP_CLASSPATH=\${HADOOP_CLASSPATH}:${DATAPROC_COMMON_LIB_DIR}/*
EOF
}

function configure_hive_server2() {
  bdconfig set_property \
    --configuration_file /etc/hive/conf/hive-site.xml \
    --name 'hive.localize.resource.num.wait.attempts' \
    --value "25" \
    --clobber

  # Hive on Tez
  bdconfig set_property \
    --configuration_file /etc/hive/conf/hive-site.xml \
    --name 'hive.execution.engine' \
    --value "tez" \
    --clobber

  # blacklist reflect and reflect2 because of vulnerability
  bdconfig set_property \
    --configuration_file /etc/hive/conf/hive-site.xml \
    --name 'hive.server2.builtin.udf.blacklist' \
    --value "reflect,reflect2" \
    --clobber

  configure_hive_classpath

  configure_hive_gc "hiveserver2"

  # Configure Tez
  # TODO: Move to Bigtop
  local tez_libs
  tez_libs="file:/usr/lib/tez,file:/usr/lib/tez/lib,file:${DATAPROC_COMMON_LIB_DIR}"
  bdconfig set_property \
    --configuration_file /etc/tez/conf/tez-site.xml \
    --name 'tez.lib.uris' \
    --value "${tez_libs}" \
    --clobber

  # Configure Pig to use HCatalog
  if os_is_package_installed hive-hcatalog && [[ -f /etc/pig/conf/pig-env.sh ]]; then
    echo "includeHCatalog=true" >>/etc/pig/conf/pig-env.sh
  fi

  # TODO: Fix in BigTop
  # delete Log4j2 bindings for Slf4j to prevent warning in Hive
  rm -f /usr/lib/hive/lib/log4j-slf4j-impl*.jar

  configure_hive_scratch_dir
  configure_hive_hbase
}

function disable_hive_server2_on_startup() {
  systemctl stop hive-server2
  systemctl disable hive-server2
}

configure_hive_server2
disable_hive_server2_on_startup
