#!/bin/bash

set -euo pipefail

source "$(dirname "$0")/../../bdutil_env.sh"
source "$(dirname "$0")/../../bdutil_helpers.sh"
source "$(dirname "$0")/../shared/jupyter.sh"

set -x

set_log_tag "post-install-component-jupyter"

readonly JUPYTER_GCS_CONTENTS_MANAGER_REVISION="67f08e19469494ace1b953c515b09ae960c1a4ec"
readonly JUPYTER_GCS_CONTENTS_MANAGER_REPO="jupyter-extensions"
readonly JUPYTER_GCS_CONTENTS_MANAGER_DIR="jupyter-gcs-contents-manager"
readonly JUPYTER_GCS_CONTENTS_MANAGER_ARCHIVE_URL="https://github.com/GoogleCloudPlatform/${JUPYTER_GCS_CONTENTS_MANAGER_REPO}/archive/${JUPYTER_GCS_CONTENTS_MANAGER_REVISION}"

install -d "${JUPYTER_DATAPROC_DIR}"

# Install additional packages for alignment with AI Platform notebooks
if [[ "${DATAPROC_IMAGE_TYPE}" == "serverless-spark" ]] ||
    is_version_at_least "${DATAPROC_IMAGE_VERSION}" "2.0"; then
  # For 2.0 images, Jupyter packages are installed along with Miniconda
  # Install Jupyter GCS Contents Manager
  JUPYTER_GCS_CONTENTS_MANAGER_PATH="${JUPYTER_GCS_CONTENTS_MANAGER_REPO}-${JUPYTER_GCS_CONTENTS_MANAGER_REVISION}/${JUPYTER_GCS_CONTENTS_MANAGER_DIR}"
  curl -fsSL --retry-connrefused --retry 10 --retry-max-time 30 \
      "${JUPYTER_GCS_CONTENTS_MANAGER_ARCHIVE_URL}.tar.gz" |
    tar -zxf - -C "${JUPYTER_DATAPROC_DIR}" \
      "${JUPYTER_GCS_CONTENTS_MANAGER_PATH}"
  "${CONDA_BIN_DIR}/pip" install --no-deps -e \
    "${JUPYTER_DATAPROC_DIR}/${JUPYTER_GCS_CONTENTS_MANAGER_PATH}"
else
  # b/238899570
  if is_rocky; then
    os_install_packages libffi-devel
  else
    os_install_packages libffi-dev
  fi
  cat >"${JUPYTER_EXTRA_PACKAGES}" <<EOF
bcolz==1.2.*
confuse==1.4.*
cookiecutter==1.7.*
fairing==0.5.*
grpcio==1.31.*
httplib2==0.18.*
ipython-sql==0.4.*
ipywidgets==7.5.*
jupyter-contrib-nbextensions==0.5.*
jupyter-http-over-ws==0.0.*
jupyterlab-git==0.11.*
markdown==3.2.*
missingno==0.4.*
nbdime==1.0.*
oauth2client==4.1.*
opencv-python==4.2.0.*
pandas-profiling==3.0.*
pydantic==1.8.*
papermill[gcs]==2.1.*
phik==0.12.0
pillow==8.3.*
pillow-simd==7.0.*
protobuf==3.13.*
pandas==1.2.*
pyarrow==0.15.*
pyasn1==0.4.*
pyasn1-modules==0.2.*
pydot==1.4.*
rsa==4.6.*
tenacity==7.0.*
tornado==6.0.*
uritemplate==3.0.*
virtualenv==20.0.*
pyzmq==23.2.*
EOF
  # TODO: pin the transitive dependencies for 1.5
  touch "${JUPYTER_EXTRA_PACKAGES_DEPS}"
fi

# Install JupyterLab extensions
# In Dataproc 2.0 all Jupyter packages are pre-installed in default
# Conda environment so we can install Jupyter extensions from them
if [[ "${DATAPROC_IMAGE_TYPE}" == 'standard' ]]; then
  if is_version_at_least "${DATAPROC_IMAGE_VERSION}" '2.0'; then
    "${CONDA_BIN_DIR}/nbdime" extensions --enable
  else
    # In Dataproc 1.5 it's not feasible to package Jupyter extensions locally
    # and install at activation time like we do for notebook extensions, so
    # preinstall them now, and if Jupyter isn't activated uninstall them at
    # cluster creation time
    PATH="${CONDA_BIN_DIR}:${PATH}" "${CONDA_BIN_DIR}/jupyter" labextension install \
      "@jupyter-widgets/jupyterlab-manager@1.0" \
      "@jupyterlab/git" \
      "nbdime-jupyterlab@1.0.0"
  fi
fi

"${CONDA_BIN_DIR}/jupyter" lab build
"${CONDA_BIN_DIR}/jupyter" lab clean

# Generate the config file in a well known location.
"${CONDA_BIN_DIR}/jupyter" notebook --allow-root --generate-config -y \
  --config "${JUPYTER_CONFIG_FILE}"
# TODO: remove this workaround when
#  above command works as intended in JupyterLab 4.0
if [[ "${DATAPROC_IMAGE_TYPE}" == "serverless-spark" ]] &&
    [[ "${DATAPROC_IMAGE_VERSION}" =~ [12]\.2 ]]; then
  mkdir -p "$(dirname "${JUPYTER_CONFIG_FILE}")"
  mv '/root/.jupyter/jupyter_notebook_config.py' "${JUPYTER_CONFIG_FILE}"
fi

# Allow uninstalling the default kernel(python) for s8s image
if [[ "${DATAPROC_IMAGE_TYPE}" == "serverless-spark" ]]; then
  echo "c.KernelSpecManager.ensure_native_kernel = False" >>"${JUPYTER_CONFIG_FILE}"
  # Remove Python kernel for s8s images during image build
  # so batches and sessions have the same env/Jupyter configuration
  "${CONDA_BIN_DIR}/jupyter" kernelspec uninstall "python3" -y

  # Install toree/scala kernel for s8s images
  "${CONDA_BIN_DIR}/jupyter" toree install --spark_home=/usr/lib/spark/ --sys-prefix
fi

if [[ "${DATAPROC_IMAGE_TYPE}" == "standard" ]] &&
    ! is_version_at_least "${DATAPROC_IMAGE_VERSION}" "2.0"; then
  "${CONDA_BIN_DIR}/pip" wheel \
      -r "${JUPYTER_EXTRA_PACKAGES}" \
      -r "${JUPYTER_EXTRA_PACKAGES_DEPS}" \
      "${JUPYTER_GCS_CONTENTS_MANAGER_ARCHIVE_URL}.zip#subdirectory=${JUPYTER_GCS_CONTENTS_MANAGER_DIR}" \
      -w "${JUPYTER_WHEEL}"
fi

if [[ ! -f "${JUPYTER_CONFIG_FILE}" ]]; then
  log_and_fail "jupyter" "Jupyter config file not found: ${JUPYTER_CONFIG_FILE}"
fi

if [[ "${DATAPROC_IMAGE_TYPE}" == "serverless-spark" ]]; then
  create_ipython_profile
  enable_jupyter_extensions
fi

# Apply patches.
if [[ "${DATAPROC_IMAGE_TYPE}" == "standard" ]] &&
    [[ "${DATAPROC_IMAGE_VERSION}" == '2.1' ]]; then
  # Patching might fail if the target file changes, in this case,
  # need to regenerate the patch file, see b/291976789#comment26.
  patch -f \
    /opt/conda/miniconda3/lib/python3.10/site-packages/notebook/static/notebook/js/main.min.js \
    "${BDUTIL_DIR}/patches/jupyter/2.1/patch-main.min.js.diff"
fi
