#!/bin/bash

set -euo pipefail

source "$(dirname "$0")/../../bdutil_helpers.sh"
source "$(dirname "$0")/../shared/spark.sh"

set -x

set_log_tag "post-install-component-spark"

readonly SPARK_CONF_DIR='/etc/spark/conf'
readonly STATSD_METRIC_SINK_PORT='8125'

function configure_spark_standalone_defaults() {
# Spark namespace (spark.metrics.namespace) can have either colon(:) or double-hash(##) as delimiter.
# Default delimiter is colon, though double-hashes are used when we're using OpenTelemetry monitoring
# agent on Dataproc 2.2+ image track (snippet: http://shortn/_2gnNQebQgL).

  cat >>"${SPARK_CONF_DIR}/spark-defaults.conf" <<EOF
# Dataproc configuration
spark.submit.deployMode=client
spark.dataproc.metrics.listener.max.stages=5000

# Configure dynamic allocation
spark.dynamicAllocation.enabled=true
spark.dynamicAllocation.minExecutors=2
spark.dynamicAllocation.maxExecutors=1000

# Enable graceful decommission for Spark workers
spark.decommission.enabled=true
spark.dynamicAllocation.shuffleTracking.enabled=true
spark.storage.decommission.enabled=true
spark.storage.decommission.rddBlocks.enabled=true
spark.storage.decommission.shuffleBlocks.enabled=true

spark.rpc.message.maxSize=512

# Enable adaptive query execution, which re-optimizes the query plan
# in the middle of query execution, based on accurate runtime statistics.
spark.sql.adaptive.enabled=true

# Reorder joins when cost-based optimization enabled to improve performance.
spark.sql.cbo.joinReorder.enabled=true

# Explicitly set Spark driver UI port
spark.ui.port=${SPARK_UI_PORT}

# Configure default namespace, so that metrics can be queried using namespace.
spark.metrics.namespace=app_name:\${spark.app.name}.app_id:\${spark.app.id}

spark.hadoop.mapreduce.fileoutputcommitter.algorithm.version=2

# Configure EventLog compression to reduce event logs size on GCS
spark.eventLog.compress=true

# GCS connector configuration
spark.hadoop.fs.AbstractFileSystem.gs.impl=com.google.cloud.hadoop.fs.gcs.GoogleHadoopFS
spark.hadoop.fs.gs.impl=com.google.cloud.hadoop.fs.gcs.GoogleHadoopFileSystem
spark.hadoop.fs.gs.reported.permissions=777
EOF

  if [[ "${DATAPROC_IMAGE_VERSION}" != '2.0' ]]; then
    cat >>"${SPARK_CONF_DIR}/spark-defaults.conf" <<EOF

# Enable and configure bloom filter
spark.sql.optimizer.runtime.bloomFilter.enabled=true
EOF
  fi

  if [[ "${DATAPROC_IMAGE_VERSION}" == '1.1' ]] ||
      [[ "${DATAPROC_IMAGE_VERSION}" == '2.0' ]]; then
    cat >>"${SPARK_CONF_DIR}/spark-defaults.conf" <<EOF

# Enables vectorized parquet reader for complex types
spark.sql.parquet.enableNestedColumnVectorizedReader=true
EOF
  fi

  if [[ "${DATAPROC_IMAGE_VERSION}" == '1.2' ]] ||
      is_version_at_least "${DATAPROC_IMAGE_VERSION}" '2.2'; then
    cat >>"${SPARK_CONF_DIR}/spark-defaults.conf" <<EOF

spark.sql.shuffle.partitions=1000

# Enable Dataproc perf improvements
spark.dataproc.advanced.infer.filter.enabled=true
spark.dataproc.aqe.remove.redundant.groupby.keys.enabled=true
spark.dataproc.enhanced.execution.enabled=true
spark.dataproc.enhanced.optimizer.enabled=true
spark.dataproc.sql.optimizer.join.fusion.enabled=true
spark.dataproc.sql.optimizer.scalar.subquery.fusion.enabled=true
EOF
  else
    # Before s8s 1.2 and 2.2 runtimes, compression was enabled and zstd was used as
    # compression codec for limiting the disk footprint. But it degrades the
    # performance of jobs. Going forward, the default spark
    # configuration will be used (lz4 compression codec and compression disabled).
    cat >>"${SPARK_CONF_DIR}/spark-defaults.conf" <<EOF

# Configure all possible Spark compression
spark.checkpoint.compress=true
spark.rdd.compress=true

# Use Zstandard as a default compression codec because it provides
# better compression ratios with similar throughput to the default LZ4
spark.io.compression.codec=zstd
EOF
  fi
}

function configure_spark_standalone_env() {
  cat >>"${SPARK_CONF_DIR}/spark-env.sh" <<EOF

# Dataproc configuration
export SPARK_COMMAND_RUNNER=${BDUTIL_DIR}/spark-executor-runner.sh
EOF

  if ! is_version_at_least "${DATAPROC_IMAGE_VERSION}" '3.0'; then
    cat >>"${SPARK_CONF_DIR}/spark-env.sh" <<EOF
export SPARK_MASTER_PORT=${SPARK_MASTER_PORT}
export SPARK_MASTER_WEBUI_PORT=${SPARK_MASTER_WEBUI_PORT}
export SPARK_WORKER_PORT=${SPARK_WORKER_PORT}
export SPARK_WORKER_WEBUI_PORT=${SPARK_WORKER_WEBUI_PORT}
EOF
  fi

  cat >>/etc/environment <<EOF
SPARK_UID=${SPARK_UID}
SPARK_GID=${SPARK_GID}
EOF
}

function configure_spark_yarn_defaults() {
  cat >>"${SPARK_CONF_DIR}/spark-defaults.conf" <<EOF
# Configure Spark on YARN
spark.master=yarn

# Dynamic allocation on YARN
spark.dynamicAllocation.enabled=true
spark.dynamicAllocation.minExecutors=1
spark.executor.instances=10000
spark.dynamicAllocation.maxExecutors=10000
spark.shuffle.service.enabled=true
spark.scheduler.minRegisteredResourcesRatio=0.0

# This undoes setting hive.execution.engine to tez in hive-site.xml
# It is not used by Spark
spark.hadoop.hive.execution.engine=mr

spark.rpc.message.maxSize=512

# Adding namespace to extract app_name and app_id for spark metrics
spark.metrics.namespace=app_name:\${spark.app.name}.app_id:\${spark.app.id}
EOF

  if is_version_at_least "${DATAPROC_IMAGE_VERSION}" '2.0'; then
    cat >>"${SPARK_CONF_DIR}/spark-defaults.conf" <<EOF

# On clusters without Kerberos integration, use unmanaged AM to speed-up Spark
# Context initialization and free-up 1 YARN container per-AM.
# On clusters integrated with Kerberos, YARN unmanaged AM setting will be
# disabled though.
spark.yarn.unmanagedAM.enabled=true

# Enable adaptive query execution, which re-optimizes the query plan
# in the middle of query execution, based on accurate runtime statistics.
spark.sql.adaptive.enabled=true
# Reorder joins when cost-based optimization enabled to improve performance.
spark.sql.cbo.joinReorder.enabled=true

spark.hadoop.mapreduce.fileoutputcommitter.algorithm.version=2
EOF
  fi

  if [[ "${DATAPROC_IMAGE_VERSION}" == '2.1' ]]; then
    cat >>"${SPARK_CONF_DIR}/spark-defaults.conf" <<EOF

# Enables vectorized parquet reader for complex types
spark.sql.parquet.enableNestedColumnVectorizedReader=true
EOF
  fi

  if is_version_at_least "${DATAPROC_IMAGE_VERSION}" '2.2'; then
    cat >>"${SPARK_CONF_DIR}/spark-defaults.conf" <<EOF

spark.sql.shuffle.partitions=1000
EOF
  fi
}

function configure_metrics_sink_standalone() {
  cat >>"${SPARK_CONF_DIR}/metrics.properties" <<EOF

# Enable MetricsServletSink for all instances
*.sink.servlet.class=org.apache.spark.metrics.sink.MetricsServlet
# Polling period for MetricsServletSink
*.sink.servlet.period=5
*.sink.servlet.unit=seconds
# Polling path for MetricsServletSink for all instances and master and applications
# This provides metrics as JSON on http://<spark-master>:<app-port>/metrics/json endpoint
# e.g. http://localhost:4040/metrics/json/ on a master node
*.sink.servlet.path=/metrics/json
master.sink.servlet.path=/metrics/master/json
applications.sink.servlet.path=metrics/applications/json
# Enable jvm source for instance master, worker, driver and executor
master.source.jvm.class=org.apache.spark.metrics.source.JvmSource
worker.source.jvm.class=org.apache.spark.metrics.source.JvmSource
driver.source.jvm.class=org.apache.spark.metrics.source.JvmSource
executor.source.jvm.class=org.apache.spark.metrics.source.JvmSource

# Enable statsd as sink for all instances.
master.sink.statsd.class=org.apache.spark.metrics.sink.StatsdSink
worker.sink.statsd.class=org.apache.spark.metrics.sink.StatsdSink
driver.sink.statsd.class=org.apache.spark.metrics.sink.StatsdSink
executor.sink.statsd.class=org.apache.spark.metrics.sink.StatsdSink
shuffleService.sink.statsd.class=org.apache.spark.metrics.sink.StatsdSink

master.sink.statsd.prefix=spark.master
worker.sink.statsd.prefix=spark.worker
driver.sink.statsd.prefix=spark.driver
executor.sink.statsd.prefix=spark.executor
shuffleService.sink.statsd.prefix=spark.shuffleService

*.sink.statsd.host=localhost
*.sink.statsd.port=${STATSD_METRIC_SINK_PORT}
EOF
}

function configure_metrics_sink_yarn() {
  cat >>"${SPARK_CONF_DIR}/metrics.properties" <<EOF
# Master and worker is not applicable for dataproc on GCE because Spark runs
# on yarn for GCE flavour
driver.sink.statsd.class=org.apache.spark.metrics.sink.StatsdSink
executor.sink.statsd.class=org.apache.spark.metrics.sink.StatsdSink
shuffleService.sink.statsd.class=org.apache.spark.metrics.sink.StatsdSink
applicationMaster.sink.statsd.class=org.apache.spark.metrics.sink.StatsdSink

driver.sink.statsd.prefix=spark.driver
executor.sink.statsd.prefix=spark.executor
shuffleService.sink.statsd.prefix=spark.shuffleService
applicationMaster.sink.statsd.prefix=spark.applicationMaster

*.source.jvm.class=org.apache.spark.metrics.source.JvmSource

*.sink.statsd.host=localhost
*.sink.statsd.port=${STATSD_METRIC_SINK_PORT}
EOF

}

function configure_fairscheduler() {
  cat >>"${SPARK_CONF_DIR}/fairscheduler.xml" <<'EOF'
<?xml version="1.0"?>

<!--
   Licensed to the Apache Software Foundation (ASF) under one or more
   contributor license agreements.  See the NOTICE file distributed with
   this work for additional information regarding copyright ownership.
   The ASF licenses this file to You under the Apache License, Version 2.0
   (the "License"); you may not use this file except in compliance with
   the License.  You may obtain a copy of the License at
       http://www.apache.org/licenses/LICENSE-2.0
   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
-->

<allocations>
  <pool name="default">
    <schedulingMode>FAIR</schedulingMode>
    <weight>1</weight>
    <minShare>0</minShare>
  </pool>
</allocations>
EOF
}

function configure_log4j_yarn() {
  # Set up log4j.properties to be less spammy, and use full package-qualified
  # classnames, since there are too many generically named classes like "Server"
  cat >>"${SPARK_CONF_DIR}/log4j.properties" <<'EOF'
# Set everything to be logged to the console
log4j.rootCategory=INFO, console
log4j.appender.console=org.apache.log4j.ConsoleAppender
log4j.appender.console.target=System.err
log4j.appender.console.layout=org.apache.log4j.PatternLayout
log4j.appender.console.layout.ConversionPattern=%d{yy/MM/dd HH:mm:ss} %p %c: %m%n

# Settings to quiet third party logs that are too verbose
log4j.logger.org.apache.spark.repl.SparkILoop$SparkILoopInterpreter=INFO
log4j.logger.org.apache.spark.repl.SparkIMain$exprTyper=INFO
log4j.logger.org.eclipse.jetty.util.component.AbstractLifeCycle=ERROR
log4j.logger.org.eclipse.jetty=WARN

# Allow INFO logging from Spark Env for EFM
log4j.logger.org.apache.spark.SparkEnv=INFO

# Spark 3.x
log4j.logger.org.sparkproject.jetty.server.handler.ContextHandler=WARN

# Spark 2.x
log4j.logger.org.spark_project.jetty.server.handler.ContextHandler=WARN
EOF
}

function configure_log4j2_yarn() {
  # Set up log4j2.properties to be less spammy
  cat >>"${SPARK_CONF_DIR}/log4j2.properties" <<'EOF'
# Set everything to be logged to the console
rootLogger.level = info
rootLogger.appenderRef.file.ref = console

appender.console.type = Console
appender.console.name = console
appender.console.target = SYSTEM_ERR
appender.console.layout.type = PatternLayout
appender.console.layout.pattern = %d{yy/MM/dd HH:mm:ss} %p %c{1}: %m%n

# Settings to quiet third party logs that are too verbose
logger.replSparkILoopInterpreter.name = org.apache.spark.repl.SparkILoop$SparkILoopInterpreter
logger.replSparkILoopInterpreter.level = info
logger.replexprTyper.name = org.apache.spark.repl.SparkIMain$exprTyper
logger.replexprTyper.level = info
logger.jetty2.name = org.sparkproject.jetty.util.component.AbstractLifeCycle
logger.jetty2.level = error
logger.jetty1.name = org.sparkproject.jetty
logger.jetty1.level = warn


# Allow INFO logging from Spark Env for EFM
logger.sparkEnv.name = org.apache.spark.SparkEnv
logger.sparkEnv.level = info

logger.jettyContextHandler.name = org.sparkproject.jetty.server.handler.ContextHandler
logger.jettyContextHandler.level = warn
EOF
}

# This uses default Spark configuration that configures all Spark logging to
# stdout/stderr. Spark driver and executor runners (Dataproc Agent and
# spark-executor-runner.sh respectively) take care of redirecting them
# to the local FS files that are picked up by fluentd and sent to Cloud Logging
# TODO: Remove NativeCodeLoader log level after native libs ship
function configure_log4j2_standalone() {
  cp "${SPARK_CONF_DIR}/log4j2.properties.template" \
    "${SPARK_CONF_DIR}/log4j2.properties"

  cat >>"${SPARK_CONF_DIR}/log4j2.properties" <<EOF

# Dataproc Serverless-specific configuration
logger.hadoopNative.name = org.apache.hadoop.util.NativeCodeLoader
logger.hadoopNative.level = error

# Reduce verbosity for Dataproc Hadoop connectors (GCS and BQ) classes
logger.gcpConnectors.name = com.google.cloud.hadoop
logger.gcpConnectors.level = warn
EOF
}

function configure_log4j() {
  cat >>"${SPARK_CONF_DIR}/log4j.properties" <<'EOF'

# Reduce verbosity for other spammy core classes
log4j.logger.org.apache.hadoop.conf.Configuration.deprecation=WARN
log4j.logger.org.apache.hadoop.mapreduce.lib.output.FileOutputCommitter=WARN
log4j.logger.org.apache.spark.ExecutorAllocationManager=ERROR
log4j.logger.org.apache.spark=WARN
EOF
}

function configure_log4j2() {
  cat >>"${SPARK_CONF_DIR}/log4j2.properties" <<'EOF'

# Reduce verbosity for other spammy core classes
logger.hadoopConfDeprecation.name = org.apache.hadoop.conf.Configuration.deprecation
logger.hadoopConfDeprecation.level = warn

logger.fileOutputCommitter.name = org.apache.hadoop.mapreduce.lib.output.FileOutputCommitter
logger.fileOutputCommitter.level = warn

logger.executorAllocationManager.name = org.apache.spark.ExecutorAllocationManager
logger.executorAllocationManager.level = error

logger.sparkRoot.name = org.apache.spark
logger.sparkRoot.level = warn
EOF
}

function update_permissions() {
  chmod a+x \
    "${BDUTIL_DIR}/spark-containerd-runner.sh" \
    "${BDUTIL_DIR}/spark-entrypoint.sh" \
    "${BDUTIL_DIR}/spark-executor-runner.sh"
  # Spark container entrypoint.sh running as user `spark` needs to update the
  # configs mounted from the host.
  chown -L -R spark:spark "${SPARK_CONF_DIR}"
}

function configure_env() {
  cat >>/etc/environment <<EOF
SPARK_HOME=${SPARK_HOME}
EOF
}

function generate_spark_connect_systemd_unit() {
  local -r service_username="$1"
  install -d -m 0755 -o "${service_username}" -p "${service_username}" "${SPARK_WORK_DIR}"

  cat >"${SPARK_CONNECT_SYSTEMD_UNIT_PATH}" <<EOF
[Unit]
Description=Spark Connect Server
Wants=spark-master.service
[Service]
Type=simple
User=${service_username}
Group=${service_username}
Restart=on-failure
Environment=CONTAINER_NAME="spark-connect-\$(date +%%Y%%m%%d%%H%%M%%S)"
EnvironmentFile=/etc/environment
ExecStart=/bin/bash -c "${BDUTIL_DIR}/spark-containerd-runner.sh \
          ${SPARK_WORK_DIR} \
          \${CONTAINER_NAME} \
          /bin/bash -c '/usr/lib/spark/sbin/start-connect-server.sh \
          --conf spark.connect.grpc.binding.address=localhost \
          --conf spark.connect.grpc.binding.port=${SPARK_CONNECT_PORT} \
          --jars ${SPARK_CONNECTOR_DIR}/spark-connect.jar \
          && sleep infinity'"
[Install]
WantedBy=multi-user.target
EOF

  cat >"${SPARK_CONNECT_PROXY_SYSTEMD_UNIT_PATH}" <<EOF
[Unit]
Description=Spark Connect Server Proxy
Wants=spark-connect.service
[Service]
Type=simple
User=${service_username}
Group=${service_username}
Restart=on-failure
EnvironmentFile=/etc/environment
ExecStart=/usr/local/bin/tcp-bridge-backend \
          --frontend-port=${SPARK_CONNECT_PROXY_PORT} \
          --backend-port=${SPARK_CONNECT_PORT} \
          --force-http2
[Install]
WantedBy=multi-user.target
EOF
}

function post_install_spark() {
  if [[ "${DATAPROC_IMAGE_TYPE}" == "serverless-spark" ]]; then
    configure_log4j2_standalone
    configure_log4j2
    configure_spark_standalone_defaults
    configure_spark_standalone_env
    configure_metrics_sink_standalone
    update_permissions
    generate_spark_connect_systemd_unit "spark"
  else
    if is_version_at_least "${DATAPROC_IMAGE_VERSION}" "2.1"; then
      configure_log4j2_yarn
      configure_log4j2
    else
      configure_log4j_yarn
      configure_log4j
    fi
    configure_spark_yarn_defaults
    configure_metrics_sink_yarn
    if is_version_at_least "${DATAPROC_IMAGE_VERSION}" '2.0'; then
      # TODO: fix in Bigtop
      cp /usr/lib/hadoop/lib/hadoop-lzo-*.jar "${SPARK_JARS_DIR}/"
    fi
  fi

  configure_env
  configure_fairscheduler
}

post_install_spark
