#!/bin/bash

set -euo pipefail

source "$(dirname "$0")/../../bdutil_components.sh"
source "$(dirname "$0")/../../bdutil_logging.sh"
source "$(dirname "$0")/../../bdutil_metadata.sh"
source "$(dirname "$0")/../../bdutil_misc.sh"
source "$(dirname "$0")/../../bdutil_properties.sh"
source "$(dirname "$0")/../shared/dpms-proxy.sh"
source "$(dirname "$0")/../shared/hive-metastore.sh"

set -x

set_log_tag "pre-activate-component-dpms-proxy"

# Generate proxy configuration.  The proxy code requires a complete
# configuration file (all properties, even if empty value).
function configure_dpms_proxy() {
  if is_component_selected "kerberos"; then
    local -r keytab_file='/etc/security/keytab/hive.service.keytab'
    local realm
    realm=$(grep "default_realm = " /etc/krb5.conf | awk '{print $NF}')
    local -r kerberos_principal="hive/_HOST@${realm}"
  fi

  # By default, the proxy sets the maximum inbound gRPC message size to 60 MiB,
  # and this can be overridden by a property.
  local max_message_size_bytes
  max_message_size_bytes=$(get_dataproc_property_or_default \
    dataproc.dpms-proxy.proxy-max-message-size-bytes 62914560)

  cat >"${DPMS_PROXY_CONF_DIR}/proxy-config.xml" <<EOF
<?xml version="1.0"?>
<?xml-stylesheet type="text/xsl" href="configXML.xsl"?>

<configuration>

    <property>
        <name>proxy.mode</name>
        <value>thrift</value>
        <description>Proxy will accept requests and produce responses using protocol found provided in value</description>
    </property>

    <property>
        <name>grpc.listening.port</name>
        <value>9085</value>
        <description>gRPC -> Thrift listening port</description>
    </property>

    <property>
        <name>thrift.listening.port</name>
        <value>9083</value>
        <description>Thrift -> gRPC listening port</description>
    </property>

    <property>
        <name>hive.metastore.uri</name>
        <value>localhost:9083</value>
        <description>Hive Metastore uri</description>
    </property>

    <property>
        <name>keytab.file</name>
        <value>${keytab_file:-}</value>
        <description>Path to the keytab file</description>
    </property>

    <property>
        <name>kerberos.principal</name>
        <value>${kerberos_principal:-}</value>
        <description>Principal from keytab file</description>
    </property>

    <property>
        <name>hive.version</name>
        <value></value>
        <description></description>
    </property>

    <property>
        <name>google.service.account.key.file</name>
        <value></value>
        <description>Google Service account JSON file to use</description>
    </property>

    <property>
        <name>google.credentials.applicationdefault.enabled</name>
        <value>true</value>
        <description>Whether to use Google Application Default credentials</description>
    </property>

    <property>
        <name>proxy.grpc.auth.interceptor.class</name>
        <value>com.google.interceptor.GoogleAuthServerInterceptor</value>
        <description>Grpc authentication interceptor class</description>
    </property>

    <property>
        <name>proxy.grpc.ssl.enabled</name>
        <value>false</value>
        <description>Enable SSL of proxy</description>
    </property>

    <property>
        <name>proxy.grpc.ssl.certificate.file</name>
        <value></value>
        <description>SSL Certificate file</description>
    </property>

    <property>
        <name>proxy.grpc.ssl.upstream.enabled</name>
        <value>true</value>
        <description>Enable SSL support when connect to upstream endpoint</description>
    </property>

    <property>
        <name>proxy.grpc.ssl.key.file</name>
        <value></value>
        <description>Private Key file for SSL certificate</description>
    </property>

    <property>
        <name>proxy.thrift.threads.max</name>
        <value>1000</value>
        <description>Maximum Thrift server processing threads</description>
    </property>

    <property>
        <name>proxy.max.message.size.bytes</name>
        <value>${max_message_size_bytes}</value>
        <description>Max message size in bytes of inbound messages</description>
    </property>

    <property>
        <name>grpc.max.retries</name>
        <value>5</value>
        <description>Max number of GRPC retries</description>
    </property>
</configuration>
EOF
}

# The proxy jar has a version suffix.  Create a version-agnostic symlink, so
# that the systemd service file doesn't need awareness of the versioned path.
# If the image only supports one proxy version, then select that one.
# Otherwise, we need the control plane to pass a property that specifies which
# version to select.
function create_dpms_proxy_symlink() {
  if [[ "${#DPMS_PROXY_HIVE_VERSIONS[@]}" == 1 ]]; then
    local hive_version="${DPMS_PROXY_HIVE_VERSIONS[0]}"
  else
    local hive_version
    hive_version=$(get_dataproc_property dataproc.dpms-proxy.version)
  fi

  ln -s \
    "${DPMS_PROXY_INSTALL_DIR}/${hive_version}/hms-proxy-${hive_version}.jar" \
    "${DPMS_PROXY_INSTALL_DIR}/hms-proxy.jar"
}

# Configure Hive metastore URI setting to point at the proxy, either a single
# master node or multiple masters in an HA deployment.
function configure_hive() {
  set_property_hive_site 'hive.metastore.uris' "${CLUSTER_MASTER_METASTORE_URIS}"
}

function configure_multi_region_dpms() {
  local dpms_service
  dpms_service=$(get_dataproc_property "dataproc.dpms-proxy.dpms-service")
  if [[ -z "${dpms_service}" ]]; then
    log_and_fail "dpms-proxy" "Multi region DPMS found but dpms-service config was not set"
  fi
  set_property_dpms "multi.regional.dpms" "true"
  set_property_dpms "dpms.service" "${dpms_service}"
}

function configure_single_region_dpms() {
  # This is the destination URI of the metastore gRPC endpoint.  The proxy will
  # receive Thrift protocol locally and proxy to this gRPC endpoint.
  local proxy_uri
  proxy_uri=$(get_dataproc_property dataproc.dpms-proxy.proxy-uri)
  set_property_dpms "proxy.uri" "${proxy_uri}"
}

function pre_activate_dpms_proxy() {
  # Create logging directory
  install -d -m 0755 -o "${DPMS_PROXY_USER}" -g "${DPMS_PROXY_USER}" \
    "${DPMS_PROXY_LOG_DIR}"

  configure_dpms_proxy
  create_dpms_proxy_symlink

  local dpms_multi_region
  dpms_multi_region=$(get_dataproc_property_or_default "dataproc.dpms-proxy.multi-regional-dpms" "false")
  if [[ "${dpms_multi_region}" == "true" ]]; then
    configure_multi_region_dpms
  else
    configure_single_region_dpms
  fi

  if [[ "${DATAPROC_IMAGE_TYPE}" != "serverless-spark" ]]; then
    configure_hive
  fi
}

pre_activate_dpms_proxy
