#!/bin/bash

set -euo pipefail

set -a
source "$(dirname "$0")/../../bdutil_env.sh"
set +a
source "$(dirname "$0")/../../bdutil_logging.sh"
source "$(dirname "$0")/../../bdutil_metadata.sh"
source "$(dirname "$0")/../../bdutil_properties.sh"
source "$(dirname "$0")/../../bdutil_packages.sh"
source "$(dirname "$0")/../../bdutil_components.sh"
source "$(dirname "$0")/../../bdutil_misc.sh"
source "$(dirname "$0")/../../bdutil_versions.sh"

set -x

set_log_tag "pre-activate-component-hdfs"

readonly HDFS_ADMIN=hdfs
readonly HDFS_SITE_XML="${HADOOP_CONF_DIR}/hdfs-site.xml"

# Location of HDFS metadata on namenode
export HDFS_NAME_DIR=/hadoop/dfs/name
export HDFS_SECONDARY_NAME_DIR=/hadoop/dfs/namesecondary

CLUSTER_NAME=$(get_metadata_cluster_name)
export CLUSTER_NAME
DATAPROC_MASTER=$(get_metadata_master)
DATAPROC_MASTER_ADDITIONAL=$(get_metadata_master_additional || true)
MASTER_HOSTNAMES=($DATAPROC_MASTER ${DATAPROC_MASTER_ADDITIONAL//,/ })
NUM_MASTERS=${#MASTER_HOSTNAMES[@]}
NUM_WORKERS=$(get_metadata_worker_count)
DATANODE_PACKAGES=('hadoop-hdfs-datanode')
THIRD_MASTER_OPTIONAL_PACKAGES=('hadoop-hdfs-namenode' 'hadoop-hdfs-zkfc')

for ((i = 0; i < ${NUM_MASTERS}; i++)); do
  declare "MASTER_HOSTNAME_$i"="${MASTER_HOSTNAMES[$i]}"
  export MASTER_HOSTNAME_"${i}"
done

KERBEROS_ENABLED=$(is_component_selected "kerberos" \
 && echo "true" || echo "false")

export ENABLE_HDFS_PERMISSIONS="${KERBEROS_ENABLED}"

# If disks are mounted use all of them for HDFS data
DATA_DIRS_ARRAY=($(get_data_dirs))

# Location of HDFS data blocks on datanodes
HDFS_DIRS="${DATA_DIRS_ARRAY[*]/%/hadoop/dfs}"
HDFS_DIRS_ARRAY=(${HDFS_DIRS})
HDFS_DATA_DIRS="${DATA_DIRS_ARRAY[*]/%/hadoop/dfs/data}"
HDFS_DATA_DIRS_ARRAY=(${HDFS_DATA_DIRS})

function update_hostname_in_xml {
  local host_name
  local domain
  domain=$(dnsdomainname)

  host_name="${DATAPROC_MASTER}.${domain}"
  update_hostname_for_property "${HDFS_SITE_XML}" "dfs.namenode.http-address" "${host_name}"
  update_hostname_for_property "${HDFS_SITE_XML}" "dfs.namenode.https-address" "${host_name}"
  update_hostname_for_property "${HDFS_SITE_XML}" "dfs.namenode.secondary.http-address" "${host_name}"
  update_hostname_for_property "${HDFS_SITE_XML}" "dfs.namenode.secondary.https-address" "${host_name}"

  host_name="$(hostname -f)"
  update_hostname_for_property "${HDFS_SITE_XML}" "dfs.datanode.address" "${host_name}"
  update_hostname_for_property "${HDFS_SITE_XML}" "dfs.datanode.http.address" "${host_name}"
  update_hostname_for_property "${HDFS_SITE_XML}" "dfs.datanode.https.address" "${host_name}"
  update_hostname_for_property "${HDFS_SITE_XML}" "dfs.datanode.ipc.address" "${host_name}"
}

function update_hdfs_param_values() {
  # b/287074566
  # https://hadoop.apache.org/docs/current/hadoop-project-dist/hadoop-hdfs/hdfs-default.xml

  bdconfig set_property \
    --configuration_file "${HDFS_SITE_XML}" \
    --name 'dfs.namenode.replication.work.multiplier.per.iteration' \
    --value '20' \
    --clobber
  bdconfig set_property \
    --configuration_file "${HDFS_SITE_XML}" \
    --name 'dfs.namenode.replication.max-streams' \
    --value '20' \
    --clobber
  bdconfig set_property \
    --configuration_file "${HDFS_SITE_XML}" \
    --name 'dfs.namenode.replication.max-streams-hard-limit' \
    --value '40' \
    --clobber
}

# Do not create HDFS_NAME_DIR, or Hadoop will think it is already formatted
mkdir -p /hadoop/dfs "${HDFS_DATA_DIRS_ARRAY[@]}"

chown -L -R "${HDFS_ADMIN}:hadoop" /hadoop/dfs "${HDFS_DIRS_ARRAY[@]}"

# Make sure the data dirs have the expected permissions.
chmod -R "${HDFS_DATA_DIRS_PERM}" "${HDFS_DIRS_ARRAY[@]}"

# Set general Hadoop environment variables

# Calculate the memory allocations, MB, using 'free -m'. Floor to nearest MB.
TOTAL_MEM=$(free -m | awk '/^Mem:/{print $2}')
NAMENODE_MEM_MB=$(python -c \
    "print(int(${TOTAL_MEM} * ${HDFS_MASTER_MEMORY_FRACTION} / 2))")
SECONDARYNAMENODE_MEM_MB=${NAMENODE_MEM_MB}
if is_version_at_least "${DATAPROC_IMAGE_VERSION}" "2.1"; then
  GC_LOG_OPTS="-Xlog:gc*:stdout:time,level,tags"
else
  GC_LOG_OPTS="-XX:+PrintGCTimeStamps -XX:+PrintGCDateStamps -XX:+PrintGCDetails"
fi

cat >>"${HADOOP_CONF_DIR}/hadoop-env.sh" <<EOF
# Lessen stop-the-world collection:
GC_OPTS="-XX:+UseConcMarkSweepGC"
# Log GC details to stdout, these will be in diagnostic tarballs.
GC_LOGGING_OPTS="${GC_LOG_OPTS}"

# Add GC options to the namenode / secondary namenode.
HADOOP_NAMENODE_OPTS="\${GC_OPTS} \${GC_LOGGING_OPTS} \${HADOOP_NAMENODE_OPTS}"
HADOOP_SECONDARYNAMENODE_OPTS="\${GC_OPTS} \${GC_LOGGING_OPTS} \${HADOOP_SECONDARYNAMENODE_OPTS}"

# Increase the maximum NameNode / SecondaryNameNode heap.
HADOOP_NAMENODE_OPTS="-Xmx${NAMENODE_MEM_MB}m \${HADOOP_NAMENODE_OPTS}"
HADOOP_SECONDARYNAMENODE_OPTS="-Xmx${SECONDARYNAMENODE_MEM_MB}m \${HADOOP_SECONDARYNAMENODE_OPTS}"

HADOOP_DATANODE_OPTS="\${HADOOP_DATANODE_OPTS}"
EOF

# Convert from space to comma separated
export HDFS_DATA_DIRS="${HDFS_DATA_DIRS// /,}"

if [[ $NUM_MASTERS -gt 1 ]]; then
  readonly HDFS_TEMPLATE="hdfs-ha-template.xml"
else
  readonly HDFS_TEMPLATE="hdfs-template.xml"
fi

bdconfig merge_configurations \
  --configuration_file "${HDFS_SITE_XML}" \
  --source_configuration_file ${HDFS_TEMPLATE} \
  --resolve_environment_variables \
  --create_if_absent \
  --clobber

if ((NUM_WORKERS == 0)); then
  bdconfig set_property \
    --configuration_file "${HDFS_SITE_XML}" \
    --name 'dfs.replication' \
    --value '1' \
    --clobber
fi

# Mixins for scaling simplification.
if [[ $NUM_MASTERS -gt 1 ]]; then
  readonly HDFS_SIMPLIFICATION_MIXINS="hdfs-simplification-ha-mixins.xml"
else
  readonly HDFS_SIMPLIFICATION_MIXINS="hdfs-simplification-mixins.xml"
fi
bdconfig merge_configurations \
  --configuration_file "${HDFS_SITE_XML}" \
  --source_configuration_file "${BDUTIL_DIR}/conf/${HDFS_SIMPLIFICATION_MIXINS}" \
  --resolve_environment_variables \
  --create_if_absent \
  --clobber

update_hostname_in_xml
update_hdfs_param_values

merge_xml_properties "${CLUSTER_PROPERTIES_DIR}/hdfs.xml" \
  "${HDFS_SITE_XML}"

# Determine where the cluster membership files should be.
INCLUDE_PATH=$(create_or_validate_include_file_path)
EXCLUDE_PATH=$(create_or_validate_exclude_file_path)

# Tell the HDFS namenode where the membership files are.
# NB: the YARN resource manager should point to the same files if YARN is
# activated.
bdconfig set_property \
  --configuration_file "${HDFS_SITE_XML}" \
  --name 'dfs.hosts' \
  --value "${INCLUDE_PATH}" \
  --clobber
bdconfig set_property \
  --configuration_file "${HDFS_SITE_XML}" \
  --name 'dfs.hosts.exclude' \
  --value "${EXCLUDE_PATH}" \
  --clobber

function update_master_hostnames_to_fqdn() {
  local domain
  domain=$(dnsdomainname)

  # Replace all the master-hostname:port properties in the config to
  # master-fqdn:port
  local master_hostname
  local master_fqdn
  for master_hostname in "${MASTER_HOSTNAMES[@]}"; do
    master_fqdn="${master_hostname}.${domain}"
    sed -i -e "s/${master_hostname}:/${master_fqdn}:/g" \
      "${HDFS_SITE_XML}"
  done
}

if [[ "${KERBEROS_ENABLED}" == "true" ]]; then
  update_master_hostnames_to_fqdn
fi

# Mark datanode packages for uninstallation from nodes where datanode is not
# enabled, e.g., secondary workers.
if [[ $(get_metadata_datanode_enabled || echo "") != 'true' ]]; then
  mark_packages_to_uninstall "${DATANODE_PACKAGES[@]}"
fi

if  [[ "$(get_metadata_role)" == "Master" ]] \
    && [[ "${NUM_MASTERS}" == "3" ]] \
    && [[ "${MASTER_INDEX}" == "2" ]]; then
        name_node_list=$(get_property_in_xml "${HDFS_SITE_XML}" \
            "dfs.ha.namenodes.${CLUSTER_NAME}")
        if [[ "${name_node_list}" != "nn0,nn1,nn2" ]]; then
            mark_packages_to_uninstall "${THIRD_MASTER_OPTIONAL_PACKAGES[@]}"
        fi
fi
