#!/bin/bash

set -euo pipefail

source "$(dirname "$0")/../../bdutil_components.sh"
source "$(dirname "$0")/../../bdutil_env.sh"
source "$(dirname "$0")/../../bdutil_logging.sh"
source "$(dirname "$0")/../../bdutil_metadata.sh"
source "$(dirname "$0")/../../bdutil_properties.sh"
source "$(dirname "$0")/../../bdutil_versions.sh"

set -x

set_log_tag "pre-activate-component-hive-server2"

HIVE_CONF_DIR="/etc/hive/conf"
CLUSTER_PROPERTIES_DIR="/tmp/cluster/properties"

CONFIGBUCKET=$(get_metadata_bucket)
CLUSTER_UUID=$(get_metadata_cluster_uuid)
DATAPROC_MASTER=$(get_metadata_master)
DATAPROC_MASTER_ADDITIONAL=$(get_metadata_master_additional || true)
MASTER_HOSTNAMES=($DATAPROC_MASTER ${DATAPROC_MASTER_ADDITIONAL//,/ })
NUM_MASTERS=${#MASTER_HOSTNAMES[@]}

# For hive high availability, configure zookeeper quorum for hive server 2.
# See https://issues.apache.org/jira/browse/HIVE-1293 and
# https://cwiki.apache.org/confluence/display/Hive/HiveServer2+Clients#HiveServer2Clients-ConnectionURLWhenZooKeeperServiceDiscoveryIsEnabled
if [[ $NUM_MASTERS -gt 1 ]]; then
  ZOOKEEPER_QUORUM=$(printf ",%s:2181" "${MASTER_HOSTNAMES[@]}")
  ZOOKEEPER_QUORUM=${ZOOKEEPER_QUORUM:1}
  bdconfig set_property \
    --configuration_file /etc/hive/conf/hive-site.xml \
    --name 'hive.zookeeper.quorum' \
    --value "${ZOOKEEPER_QUORUM}" \
    --clobber

  bdconfig merge_configurations \
    --configuration_file /etc/hive/conf/hive-site.xml \
    --source_configuration_file "${BDUTIL_DIR}/conf/hive-ha-mixins.xml" \
    --resolve_environment_variables \
    --create_if_absent \
    --clobber
fi

# This is the location that Hive in Tez mode will use to upload Hive jars.
# Use the cluster's GCS folder to get rid of the dependency on HDFS.
bdconfig set_property \
  --configuration_file /etc/hive/conf/hive-site.xml \
  --name 'hive.user.install.directory' \
  --value "gs://${CONFIGBUCKET}/google-cloud-dataproc-metainfo/${CLUSTER_UUID}/hive/user-install-dir" \
  --clobber

if is_version_at_least "${DATAPROC_IMAGE_VERSION}" "2.1"; then
  set_property_hive_site 'hive.exec.input.listing.max.threads' '20'

  # The hive.fetch.task.conversion property controls an optimization that
  # fetches data directly into the client, without starting a MapReduce or Tez
  # application, for simple LIMIT queries. Its default value is "more", but
  # we've found that this is too aggressive. It can cause degraded performance
  # for LIMIT queries that also include filters that match very few rows. In
  # image 1.5 and 2.0, the control plane would set the property to "none", which
  # completely disables the optimization. However, starting in image 2.1, we
  # control this from the image code and set the property to "minimal". This
  # allows the optimization when the query's filters only include the partition
  # columns, which is trivially implemented in the client by partition pruning.
  set_property_hive_site 'hive.fetch.task.conversion' 'minimal'

  if is_component_selected trino; then
    set_property_hive_site 'metastore.storage.schema.reader.impl' \
      'org.apache.hadoop.hive.metastore.SerDeStorageSchemaReader'
  fi
fi

merge_xml_properties \
  "${CLUSTER_PROPERTIES_DIR}/hive.xml" \
  "${HIVE_CONF_DIR}/hive-site.xml"

merge_java_properties \
  "${CLUSTER_PROPERTIES_DIR}/hive-log4j2.properties" \
  "${HIVE_CONF_DIR}/hive-log4j2.properties"
