#!/bin/bash

set -euo pipefail

source "$(dirname "$0")/../../bdutil_env.sh"
source "$(dirname "$0")/../../bdutil_logging.sh"
source "$(dirname "$0")/../../bdutil_properties.sh"
source "$(dirname "$0")/../../bdutil_components.sh"
source "$(dirname "$0")/../../bdutil_versions.sh"
source "$(dirname "$0")/../../bdutil_gcs.sh"
source "$(dirname "$0")/../../bdutil_retry.sh"
source "${DATAPROC_DIR}/dataproc_env.sh"

set -x

set_log_tag "pre-activate-component-hudi"

readonly HUDI_HOME="/usr/lib/hudi"
readonly HUDI_CONF_DIR="/etc/hudi/conf"
readonly HUDI_CONF_FILE="${HUDI_CONF_DIR}/hudi-defaults.conf"
readonly HUDI_REGIONAL_BUCKET_PREFIX="gs://goog-dataproc-artifacts-"
readonly HUDI_GLOBAL_BUCKET="gs://hadoop-lib"
readonly HUDI_EFFECTIVE_VERSION=$(get_dataproc_property_or_default \
    "hudi.version" "${HUDI_VERSION}")
SPARK_MINOR_VERSION=$(ls /usr/lib/spark/jars/spark-core_*.jar)
SPARK_MINOR_VERSION=${SPARK_MINOR_VERSION##*-}
SPARK_MINOR_VERSION=${SPARK_MINOR_VERSION%.*.jar}
readonly SPARK_MINOR_VERSION
readonly HUDI_SPARK_JAR_NAME="hudi-spark${SPARK_MINOR_VERSION}-bundle_${SCALA_VERSION%.*}-${HUDI_EFFECTIVE_VERSION}.jar"
readonly HUDI_SPARK_JAR_PATH="${HUDI_HOME}/lib/${HUDI_SPARK_JAR_NAME}"
readonly HUDI_HADOOP_MR_JAR_PATH=${HUDI_HOME}/lib/hudi-hadoop-mr-bundle-${HUDI_EFFECTIVE_VERSION}.jar


# Use regional bucket if available, otherwise fall back to global bucket.
function get_hudi_base_gcs_dir() {
  # User-specified.
  local custom_dir
  custom_dir=$(get_dataproc_property_or_default "hudi.gcs.dir" "")
  if [[ -n "${custom_dir}" ]]; then
    echo "${custom_dir}"
    return 0
  fi

  local -r global_dir="${HUDI_GLOBAL_BUCKET}/hudi"

  # Legacy global region.
  if [[ "${REGION}" == 'global' ]]; then
    echo "${global_dir}"
    return 0
  fi

  # Regional bucket.
  local regional_dir="${HUDI_REGIONAL_BUCKET_PREFIX}${REGION}/hudi"
  if check_gcs_path_exists "${regional_dir}"; then
    echo "${regional_dir}"
    return 0
  fi

  # Fall back to global bucket.
  loginfo "WARNING: ${regional_dir} does not exist, fall back to ${global_dir}"
  echo "${global_dir}"
}

function get_hudi_versioned_gcs_dir() {
  local base_dir
  base_dir=$(get_hudi_base_gcs_dir)
  # Remove the trailing '/' from the base dir if any.
  echo "${base_dir%/}/${HUDI_EFFECTIVE_VERSION}"
}

HUDI_VERSIONED_GCS_DIR=$(get_hudi_versioned_gcs_dir)
readonly HUDI_VERSIONED_GCS_DIR

function download_hudi_cli() {
  local -r cli_gcs_dir="${HUDI_VERSIONED_GCS_DIR}/cli"

  # Currently Hudi CLI packaging is not complete, see
  # https://issues.apache.org/jira/browse/HUDI-4666
  mkdir -p "${HUDI_HOME}/cli"
  gcs_copy_with_metadata_preserved "${cli_gcs_dir}/*" "${HUDI_HOME}/cli/"
  chmod a+x ${HUDI_HOME}/cli/hudi-cli.sh
}

function download_hudi_spark() {
  gcs_copy_with_metadata_preserved \
      "${HUDI_VERSIONED_GCS_DIR}/lib/spark/${HUDI_SPARK_JAR_NAME}" \
      "${HUDI_SPARK_JAR_PATH}"
}

function download_hudi_lib() {
  mkdir -p "${HUDI_HOME}/lib"
  # Spark jars
  download_hudi_spark
  # Non-Spark jars
  gcs_copy_with_metadata_preserved \
      "${HUDI_VERSIONED_GCS_DIR}/lib/*.jar" "${HUDI_HOME}/lib/"
}

function download_hudi_tools() {
  mkdir -p "${HUDI_HOME}/tools"

  gcs_rsync_with_metadata_preserved \
      "${HUDI_VERSIONED_GCS_DIR}/tools/" "${HUDI_HOME}/tools/"
}

function download_hudi() {
  if [[ "${ROLE}" == "Master" ]]; then
    download_hudi_cli
    download_hudi_tools
  fi

  download_hudi_lib
}

function create_hudi_config() {
  mkdir -p "${HUDI_CONF_DIR}"
  # The default config is empty now, which means all are Hudi default. We can
  # add some Dataproc defaults when needed.
  touch "${HUDI_CONF_FILE}"
  chmod a+rx "${HUDI_CONF_DIR}"
  chmod a+r "${HUDI_CONF_FILE}"

  merge_java_properties \
    "${CLUSTER_PROPERTIES_DIR}/hudi.properties" \
    "${HUDI_CONF_FILE}"

  cat >>/etc/environment <<EOF

HUDI_CONF_DIR=/etc/hudi/conf
EOF
}

function configure_spark_for_hudi() {
  local -r spark_config_file="/etc/spark/conf/spark-defaults.conf"

  set_java_property \
    "${spark_config_file}" \
    "spark.driver.extraClassPath" \
    "${HUDI_SPARK_JAR_PATH}" \
    "append" \
    ":"
  set_java_property \
    "${spark_config_file}" \
    "spark.executor.extraClassPath" \
    "${HUDI_SPARK_JAR_PATH}" \
    "append" \
    ":"
  set_java_property \
    "${spark_config_file}" \
    "spark.serializer" \
    "org.apache.spark.serializer.KryoSerializer" \
    "skip"
  set_java_property \
    "${spark_config_file}" \
    "spark.sql.extensions" \
    "org.apache.spark.sql.hudi.HoodieSparkSessionExtension" \
    "skip"
  if is_version_at_least "${DATAPROC_IMAGE_VERSION}" '2.1'; then
    set_java_property \
      "${spark_config_file}" \
      "spark.sql.catalog.spark_catalog" \
      "org.apache.spark.sql.hudi.catalog.HoodieCatalog" \
      "skip"
    # Explicitly setting the default value until this fix is available
    # https://github.com/apache/hudi/pull/8082
    set_java_property \
      "${spark_config_file}" \
      "spark.sql.legacy.parquet.nanosAsLong" \
      "false" \
      "skip"
    # For spark 3.3.2+, we need to disable vectorized reader
    # https://github.com/apache/hudi/pull/8082/files#diff-2303f28311c7430a3a4958bdb3a2a80a59c9e113bf4277c612e318205e42e02cR78
    set_java_property \
      "${spark_config_file}" \
      "spark.sql.parquet.enableVectorizedReader" \
      "false" \
      "skip"
  fi
}

function configure_hive_for_hudi() {
  set_property_in_xml \
    "/etc/hive/conf/hive-site.xml" \
    "hive.aux.jars.path" \
    "file://${HUDI_HADOOP_MR_JAR_PATH}" \
    "append" \
    ":"
}

function configure_hudi() {
  create_hudi_config
  configure_spark_for_hudi
  configure_hive_for_hudi
}

function pre_activate_hudi() {
  download_hudi
  configure_hudi
}

pre_activate_hudi