#!/bin/bash

set -euo pipefail

source "$(dirname "$0")/../../bdutil_env.sh"
source "$(dirname "$0")/../../bdutil_logging.sh"
source "$(dirname "$0")/../../bdutil_metadata.sh"
source "$(dirname "$0")/../../bdutil_properties.sh"
source "$(dirname "$0")/../../bdutil_components.sh"
source "$(dirname "$0")/../../bdutil_gcs.sh"
source "$(dirname "$0")/../shared/proxy.sh"

set_log_tag "pre-activate-component-ranger"

readonly NODE_NAME="$(hostname -s)"
readonly RANGER_INSTALL_DIR='/usr/lib/ranger'

if is_component_selected "kerberos"; then
  readonly KMS_KEY_URI=$(get_dataproc_property ranger.kms.key.uri)
  if [[ -z "${KMS_KEY_URI}" ]]; then
    log_and_fail "ranger" "Ranger component is enabled but failed to read KMS key URI from Dataproc properties. \
  Set it through 'dataproc:ranger.kms.key.uri'." "${EXIT_CODE_CLIENT_ERROR}"
  fi
  readonly ADMIN_PASSWORD_URI=$(get_dataproc_property ranger.admin.password.uri)
  if [[ -z "${ADMIN_PASSWORD_URI}" ]]; then
    log_and_fail "ranger" "Ranger component is enabled but failed to read admin password URI \
  from Dataproc properties. Set it through 'dataproc:ranger.admin.password.uri'." "${EXIT_CODE_CLIENT_ERROR}"
  fi

  set +x
  readonly ADMIN_PASSWORD=$(decrypt_file_with_kms_key "${ADMIN_PASSWORD_URI}" "${KMS_KEY_URI}")
  if [[ -z "${ADMIN_PASSWORD}" ]]; then
    log_and_fail "ranger" "Admin password cannot be empty!" "${EXIT_CODE_CLIENT_ERROR}"
  fi
fi

function create_symlinks_to_hadoop_conf() {
  if [[ ! -d "${RANGER_INSTALL_DIR}/hadoop/etc" ]]; then
    mkdir -p "${RANGER_INSTALL_DIR}/hadoop/etc" "${RANGER_INSTALL_DIR}/hadoop/share/hadoop/hdfs/"
    ln -s /etc/hadoop/conf "${RANGER_INSTALL_DIR}/hadoop/etc/hadoop"
    ln -s /usr/lib/hadoop-hdfs/lib "${RANGER_INSTALL_DIR}/hadoop/share/hadoop/hdfs/"
  fi
}

function create_principals() {
  # add ranger user in group hadoop so it can access the spnego (http) keytab.
  useradd ranger || true
  usermod -a -G hadoop ranger

  set +x
  # add an admin principal using the password provided
  kadmin.local -q "addprinc -pw ${ADMIN_PASSWORD} admin"
  set -x
}


function ranger_pre_activate() {
  # Add user 'root' to hdfs supergroup 'hadoop'.
  # If Kerberized, add user 'dataproc'.
  usermod -a -G hadoop root
  if is_component_selected "kerberos"; then
    usermod -a -G hadoop dataproc
  fi

  # Some distributions of Python does not have the 'requests' module,
  # which is needed by Ranger, so make sure it is installed.
  # So far we've only seen this on Ubuntu machines, where pip is installed.
  if ! python -c "import requests"; then
    pip install requests
  fi

  local role
  role="$(get_metadata_role)"
  mkdir -p "${RANGER_INSTALL_DIR}" && cd "${RANGER_INSTALL_DIR}"

  if [[ "${role}" == 'Master' ]]; then
    if [[ "${NODE_NAME}" =~ ^.*(-m|-m-0)$ ]]; then
      if is_component_selected "kerberos"; then
        create_principals
      fi
    fi
    create_symlinks_to_hadoop_conf
  fi
}

ranger_pre_activate
