#!/bin/bash

set -euo pipefail

source "$(dirname "$0")/../../bdutil_components.sh"
source "$(dirname "$0")/../../bdutil_env.sh"
source "$(dirname "$0")/../../bdutil_logging.sh"
source "$(dirname "$0")/../../bdutil_metadata.sh"
source "$(dirname "$0")/../../bdutil_misc.sh"
source "$(dirname "$0")/../../bdutil_networking.sh"
source "$(dirname "$0")/../../bdutil_os.sh"
source "$(dirname "$0")/../../bdutil_properties.sh"
source "$(dirname "$0")/../../bdutil_retry.sh"
source "$(dirname "$0")/../../bdutil_versions.sh"
source "$(dirname "$0")/../shared/hdfs.sh"
source "$(dirname "$0")/../shared/hive-metastore.sh"
source "$(dirname "$0")/../shared/nvidia-drivers.sh"
source "$(dirname "$0")/../shared/spark.sh"
source "${DATAPROC_DIR}/dataproc_env.sh"

set -x

set_log_tag "pre-activate-component-spark"

readonly HCFS_EVENTLOG_DIR="${HCFS_ROOT_URI}/user/spark/eventlog"
readonly SPARK_STANDALONE_LOCAL_DATA_DIR="${SPARK_DATA_DIR}/local-dir"

function update_bq_connector() {
  local connector_version
  connector_version=$(get_metadata_spark_bq_connector_version || true)
  local connector_url
  connector_url=$(get_metadata_spark_bq_connector_url || true)

  local jar_name
  if [[ -z "${connector_url}" ]] && [[ -n "${connector_version}" ]]; then
    if [[ "${connector_version}" = 'latest' ]]; then
      log_and_fail \
        "spark" \
        "Invalid version specified in the 'SPARK_BQ_CONNECTOR_VERSION' metadata value: '${connector_version}'"
    fi

    if [[ "${DATAPROC_IMAGE_TYPE}" == "standard" ]] &&
        is_version_at_least "${DATAPROC_IMAGE_VERSION}" '2.2' &&
        is_version_at_least "${connector_version}" '0.33.0'; then
      jar_name="spark-3.4-bigquery-${connector_version}.jar"
    elif [[ "${DATAPROC_IMAGE_TYPE}" == "serverless-spark" ]] &&
        ([[ "${DATAPROC_IMAGE_VERSION}" == '1.2' ]] ||
          is_version_at_least "${DATAPROC_IMAGE_VERSION}" '2.2') &&
        is_version_at_least "${connector_version}" '0.33.0'; then
      local spark_version
      spark_version=$(dpkg -s spark-core | grep "^Version: " | cut -d' ' -f2 | cut -d'-' -f1)
      jar_name="spark-${spark_version%.*}-bigquery-${connector_version}.jar"
    elif [[ "${DATAPROC_IMAGE_TYPE}" == "serverless-spark" ]] &&
        is_version_at_least "${DATAPROC_IMAGE_VERSION}" '2.0'; then
      jar_name="spark-bigquery-with-dependencies_2.13-${connector_version}.jar"
    else
      jar_name="spark-bigquery-with-dependencies_2.12-${connector_version}.jar"
    fi

    connector_url="gs://spark-lib/bigquery/${jar_name}"
  fi

  if [[ -n "${connector_url}" ]]; then
    jar_name=$(basename "${connector_url}")

    if [[ "${DATAPROC_IMAGE_TYPE}" == "serverless-spark" ]] ||
        is_version_at_least "${DATAPROC_IMAGE_VERSION}" '2.2'; then
      local -r connector_dir="${SPARK_JARS_DIR}"
    else
      local -r connector_dir="${DATAPROC_COMMON_LIB_DIR}"
    fi

    # Do not upgrade connector if it already at the desired version
    if [[ -f "${connector_dir}/${jar_name}" ]]; then
      return
    fi

    # Remove old connector
    if ! rm "${connector_dir}"/spark-*bigquery-*; then
      echo "WARNING: failed to delete old Spark BigQuery connector - proceeding with upgrade"
    fi

    # Download a new connector with retries
    local retries=5
    until gsutil -q cp -P "${connector_url}" "${connector_dir}/" || [[ ${retries} -le 0 ]]; do
      ((retries -= 1))
      sleep 2
    done

    if [[ ! -f "${connector_dir}/${jar_name}" ]]; then
      log_and_fail "spark" "Failed to update Spark BigQuery connector, exiting"
    fi
  fi
}

# Comma-separated list of local data dirs for YARN mode.
function get_local_data_dirs_yarn() {
  local space_separated_dirs
  space_separated_dirs="$(get_yarn_nm_local_dirs)"

  local comma_separated_dirs="${space_separated_dirs// /,}"
  echo "${comma_separated_dirs}"
}

function init_local_dirs_standalone() {
  # Ensure Spark standalone daemons can create their subdirectories.
  mkdir -p "${SPARK_STANDALONE_LOCAL_DATA_DIR}"
  chown -R spark:spark "${SPARK_STANDALONE_LOCAL_DATA_DIR}"
  chmod -R 775 "${SPARK_STANDALONE_LOCAL_DATA_DIR}"
}

function init_local_dirs_common() {
  mkdir -p "${SPARK_TMP_DIR}" "${SPARK_WORK_DIR}" "${SPARK_LOG_DIR}"
  chown spark:spark -R "${SPARK_DATA_DIR}" "${SPARK_LOG_DIR}"
  chmod 1777 -R "${SPARK_DATA_DIR}" "${SPARK_LOG_DIR}"

  # Recreate the work dir symbolic link which was created by Bigtop.
  ln -sf "${SPARK_WORK_DIR}" "${SPARK_HOME}/work"
}

function configure_env_common() {
  cat >>"${SPARK_CONF_DIR}/spark-env.sh" <<EOF

export SPARK_WORKER_DIR=${SPARK_WORK_DIR}
export SPARK_LOG_DIR=${SPARK_LOG_DIR}
export SPARK_SUBMIT_OPTS="\${SPARK_SUBMIT_OPTS} -Dscala.usejavacp=true"

# Add Dataproc common libs (e.g. connectors) to Spark class path.
SPARK_DIST_CLASSPATH="\${SPARK_DIST_CLASSPATH}:${DATAPROC_COMMON_LIB_DIR}/*"
EOF
}

function configure_env_standalone() {
  if [[ "${ROLE}" == 'Master' ]]; then
    local -r daemon_memory='416m'
  else
    local -r daemon_memory='320m'
  fi

  local driver_memory
  driver_memory=$(get_java_property \
    "${SPARK_CONF_DIR}/spark-defaults.conf" spark.driver.memory)
  cat >>"${SPARK_CONF_DIR}/spark-env.sh" <<EOF

export STANDALONE_SPARK_MASTER_HOST=${MASTER_HOSTNAME_0}
export SPARK_LOCAL_DIRS=${SPARK_STANDALONE_LOCAL_DATA_DIR}

# Set heap for Spark master and worker daemons
export SPARK_DAEMON_MEMORY=${daemon_memory}
export SPARK_CONNECT_DAEMON_MEMORY=${driver_memory}
EOF

  if [[ "${ROLE}" == 'Worker' ]] && is_component_selected 'nvidia-drivers'; then
    local num_gpus
    num_gpus=$(get_num_gpus)
    if (( num_gpus == 0 )); then
      # If number of GPUs can not be determined then run `lspci` for debugging
      lspci
      log_and_fail \
        'spark' "Failed to detect number of GPUs attached: ${num_gpus}"
    fi
    cat >>"${SPARK_CONF_DIR}/spark-env.sh" <<EOF
export SPARK_WORKER_OPTS="-Dspark.worker.resource.gpu.amount=${num_gpus} -Dspark.worker.resource.gpu.discoveryScript=/usr/lib/spark/scripts/gpu/getGpusResources.sh"
EOF
  fi
}

function configure_env_yarn() {
  cat >>"${SPARK_CONF_DIR}/spark-env.sh" <<EOF

export SPARK_LOCAL_DIRS=${SPARK_TMP_DIR}

# Include MySQL JDBC jar
SPARK_DIST_CLASSPATH="\${SPARK_DIST_CLASSPATH}:/usr/share/java/mysql.jar"
EOF

  if is_version_at_least "${DATAPROC_IMAGE_VERSION}" "2.1"; then
    # Needed to start Spark 3.2
    cat >>"${SPARK_CONF_DIR}/spark-env.sh" <<EOF
export NM_HOST="\$(hostname -f)"
export NM_HTTP_PORT="8042"
EOF
  fi

  # TODO: Fix in Bigtop
  # YARN_USER_CLASSPATH was deprecated in favor of HADOOP_CLASSPATH:
  # https://issues.apache.org/jira/browse/HADOOP-11460
  cat >>"${HADOOP_CONF_DIR}/yarn-env.sh" <<EOF

# Include Spark shuffle service jar.
HADOOP_CLASSPATH="/usr/lib/spark/yarn/*:\$HADOOP_CLASSPATH"
EOF
  # END_TODO: Fix in Bigtop
}

function configure_arrow() {
  if [[ "${DATAPROC_IMAGE_TYPE}" == "standard" ]] &&
    ! is_version_at_least "${DATAPROC_IMAGE_VERSION}" '2.0'; then
    cat >>"${HADOOP_CONF_DIR}/yarn-env.sh" <<EOF

# For Dataproc 1.5 version, add PyArrow backwards compatibility property.
# See: https://spark.apache.org/docs/latest/sql-pyspark-pandas-with-arrow.html#compatibiliy-setting-for-pyarrow--0150-and-spark-23x-24x
export ARROW_PRE_0_15_IPC_FORMAT=1
EOF

    cat >>"${SPARK_CONF_DIR}/spark-env.sh" <<EOF

# For Dataproc 1.5 version, add PyArrow backwards compatibility property.
# See: https://spark.apache.org/docs/latest/sql-pyspark-pandas-with-arrow.html#compatibiliy-setting-for-pyarrow--0150-and-spark-23x-24x
export ARROW_PRE_0_15_IPC_FORMAT=1
EOF
  fi
}

function configure_defaults_standalone() {
  cat >>"${SPARK_CONF_DIR}/spark-defaults.conf" <<EOF

spark.master=spark://${MASTER_HOSTNAME_0}:${SPARK_MASTER_PORT}

# HCFS configuration
spark.hadoop.fs.defaultFS=${HCFS_ROOT_URI}

# GCS connector configuration
spark.hadoop.fs.gs.application.name.suffix=-dataproc zone/${ZONE} (GPN:dataproc)
spark.hadoop.fs.gs.project.id=${PROJECT}
EOF
}

function configure_defaults_yarn() {
  cat >>"${SPARK_CONF_DIR}/spark-defaults.conf" <<EOF

spark.yarn.historyServer.address=${MASTER_HOSTNAME_0}:18080

spark.dataproc.metrics.listener.metrics.collector.hostname=${MASTER_HOSTNAME_0}
EOF

  if is_version_at_least "${DATAPROC_IMAGE_VERSION}" '2.0' &&
    is_component_selected hive-server2; then
    cat >>"${SPARK_CONF_DIR}/spark-defaults.conf" <<EOF

# Enable using Hive as the metastore for Spark
spark.sql.catalogImplementation=hive
EOF
  fi

  if is_version_at_least "${DATAPROC_IMAGE_VERSION}" "2.1"; then
    cat >>"${SPARK_CONF_DIR}/spark-defaults.conf" <<EOF
spark.checkpoint.compress=true
EOF
  fi

  # If Docker-on-YARN is enabled, do not set spark.yarn.jars and
  # spark.yarn.archive, so Spark will create a zip file with all jars under
  # $SPARK_HOME/jars and upload it to the YARN distributed cache, and the job
  # container image doesn't have to include Spark.
  #
  # See https://spark.apache.org/docs/3.2.1/running-on-yarn.html#preparations
  # for more details.
  local enable_docker_yarn
  enable_docker_yarn=$(get_dataproc_property_or_default yarn.docker.enable "")
  if [[ -z "${enable_docker_yarn}" ]]; then
    # This is kept for backwards compatibility.
    enable_docker_yarn=$(get_dataproc_property_or_default docker.yarn.enable "")
  fi
  if [[ "${enable_docker_yarn}" == "true" ]]; then
    local -r name="yarn_docker_user"
    local -r id="1099"

    groupadd -g "${id}" "${name}"
    useradd -d "/home/${name}" -m "${name}" -g "${id}" -s /bin/bash

    cat >>"${SPARK_CONF_DIR}/spark-defaults.conf" <<EOF

# Spark settings for Docker on YARN
spark.submit.deployMode=cluster
spark.executorEnv.YARN_CONTAINER_RUNTIME_DOCKER_MOUNTS=/etc/passwd:/etc/passwd:ro,/etc/group:/etc/group:ro,/usr/local/share/google/dataproc/lib:/usr/local/share/google/dataproc/lib:ro,/usr/lib/jvm:/usr/lib/jvm:ro,/opt/conda:/opt/conda:ro,/usr/lib/hadoop:/usr/lib/hadoop:ro
spark.yarn.appMasterEnv.YARN_CONTAINER_RUNTIME_DOCKER_MOUNTS=/etc/passwd:/etc/passwd:ro,/etc/group:/etc/group:ro,/usr/local/share/google/dataproc/lib:/usr/local/share/google/dataproc/lib:ro,/usr/lib/jvm:/usr/lib/jvm:ro,/opt/conda:/opt/conda:ro,/usr/lib/hadoop:/usr/lib/hadoop:ro

EOF
  else
    cat >>"${SPARK_CONF_DIR}/spark-defaults.conf" <<EOF

spark.submit.deployMode=client
spark.yarn.jars=local:${SPARK_JARS_DIR}/*

EOF
  fi
}

function configure_broadcast_join() {
  local -r fraction=$1
  local -r max_mb=$2

  # Do not override spark.sql.autoBroadcastJoinThreshold
  # if already set by user/control plane
  local user_broadcast_join_threshold_mb
  user_broadcast_join_threshold_mb=$(get_java_property \
    "${SPARK_CONF_DIR}/spark-defaults.conf" \
    spark.sql.autoBroadcastJoinThreshold)
  if [[ -n ${user_broadcast_join_threshold_mb} ]]; then
    return
  fi
  local spark_executor_memory
  spark_executor_memory=$(get_java_property \
    "${SPARK_CONF_DIR}/spark-defaults.conf" spark.executor.memory)
  local spark_executor_memory_bytes
  spark_executor_memory_bytes=$(get_size_in_bytes "${spark_executor_memory}")
  # Calculate spark.sql.autoBroadcastJoinThreshold as fraction
  # of Spark executor memory in [10, max_mb] MiB range
  local broadcast_join_threshold_mb
  broadcast_join_threshold_mb=$(python -c \
    "print(min(max(int(${spark_executor_memory_bytes} * ${fraction} / 1024 / 1024), 10), ${max_mb}))")

  cat >>"${SPARK_CONF_DIR}/spark-defaults.conf" <<EOF

spark.sql.autoBroadcastJoinThreshold=${broadcast_join_threshold_mb}m
EOF
}

function configure_driver_max_result_size() {
  # Do not override spark.driver.maxResultSize
  # if already set by user/control plane
  local user_driver_max_result_size
  user_driver_max_result_size=$(get_java_property \
    "${SPARK_CONF_DIR}/spark-defaults.conf" spark.driver.maxResultSize)
  if [[ -n ${user_driver_max_result_size} ]]; then
    return
  fi
  local spark_driver_memory
  spark_driver_memory=$(get_java_property \
    "${SPARK_CONF_DIR}/spark-defaults.conf" spark.driver.memory)
  local spark_driver_memory_bytes
  spark_driver_memory_bytes=$(get_size_in_bytes "${spark_driver_memory}")
  # Calculate spark.driver.maxResultSize as 80% of Spark driver memory
  local spark_driver_max_result_mb
  spark_driver_max_result_mb=$(((spark_driver_memory_bytes * 8 / 10) / 1024 / 1024))

  cat >>"${SPARK_CONF_DIR}/spark-defaults.conf" <<EOF

spark.driver.maxResultSize=${spark_driver_max_result_mb}m
EOF
}

function configure_standalone_worker_memory() {
  local spark_executor_memory
  spark_executor_memory=$(get_java_property \
    "${SPARK_CONF_DIR}/spark-defaults.conf" spark.executor.memory)

  cat >>"${SPARK_CONF_DIR}/spark-env.sh" <<EOF
# Explicitly set how much memory to allow Spark executors to use on each worker
export SPARK_WORKER_MEMORY=${spark_executor_memory}
EOF
}

function configure_efm() {
  local spark_efm_property
  spark_efm_property=$(get_dataproc_property efm.spark.shuffle)
  if [[ "${spark_efm_property}" == "primary-worker" ]]; then
    cat >>"${SPARK_CONF_DIR}/spark-defaults.conf" <<EOF

# Enable Spark primary worker shuffle (EFM)
spark.efm.remote-shuffle.enabled=true
EOF

    # Set YARN properties to configure Spark's external shuffle service to act as
    # a primary worker shuffle service.
    bdconfig set_property \
      --configuration_file "${HADOOP_CONF_DIR}/yarn-site.xml" \
      --name "spark.efm.remote-shuffle.enabled" \
      --value "true" \
      --clobber
    bdconfig set_property \
      --configuration_file "${HADOOP_CONF_DIR}/yarn-site.xml" \
      --name "spark.efm.remote-shuffle.server.local-dirs" \
      --value "$(get_local_data_dirs_yarn)" \
      --clobber
  fi
}

function configure_lineage() {
  local lineage_enabled
  # If lineage not configured via property, fallback to metadata value
  lineage_enabled=$(get_dataproc_property_or_default \
    'dataproc.lineage.enabled' \
    "$(get_metadata_dataproc_lineage_enabled || echo 'false')")

  if [[ "${lineage_enabled}" == 'true' ]]; then
    cat >>"${SPARK_CONF_DIR}/spark-defaults.conf" <<EOF

spark.dataproc.job.lineage.enabled=true
spark.extraListeners=io.openlineage.spark.agent.OpenLineageSparkListener
spark.openlineage.transport.type=datacatalog
EOF
  else
    rm -f "${DATAPROC_COMMON_LIB_DIR}/openlineage-spark.jar"
  fi
}

function configure_event_log_dir() {
  local event_log_dir="${HCFS_EVENTLOG_DIR}"
  local persist_history_to_gcs
  if [[ "${DATAPROC_IMAGE_TYPE}" == "serverless-spark" ]]; then
    persist_history_to_gcs=true
  else
    persist_history_to_gcs=$(get_dataproc_property_or_default \
      job.history.to-gcs.enabled false)
  fi

  if [[ "${persist_history_to_gcs}" == "true" ]]; then
    loginfo "Enabling persisting Spark job history files to GCS."
    if [[ -n "${TEMP_BUCKET}" ]]; then
      local gcs_history_dir_path="${DATAPROC_WORKLOAD_UUID:-${CLUSTER_UUID}}/spark-job-history"
      event_log_dir="gs://${TEMP_BUCKET}/${gcs_history_dir_path}"
    else
      loginfo "[WARNING] Cannot find VALUE for METADATA 'dataproc-temp-bucket'. Will not persist job history files to the temp bucket."
    fi
  fi

  cat >>"${SPARK_CONF_DIR}/spark-defaults.conf" <<EOF

spark.eventLog.enabled=true
spark.eventLog.dir=${event_log_dir}
spark.history.fs.logDirectory=${event_log_dir}
EOF
}

function create_event_log_dir() {
  local event_log_dir
  event_log_dir=$(get_java_property \
    "${SPARK_CONF_DIR}/spark-defaults.conf" spark.eventLog.dir)

  if [[ "${event_log_dir}" == gs://* ]]; then
    loginfo "Creating Spark event log dir in GCS"

    event_log_dir=${event_log_dir#*//} # strip 'gs://' prefix
    event_log_dir=${event_log_dir%*/}  # strip '/' suffix
    local event_log_dir_bucket=${event_log_dir%%/*}
    local event_log_dir_path=${event_log_dir#*/}

    # Disable logs to prevent access token leak in log
    set +x

    # Can not use `get_auth_token` function here, because this function executes
    # in forked process asynchronously
    local token
    token=$(/usr/share/google/get_metadata_value service-accounts/default/token |
      jq -r ".access_token")

    # It is possible that we will have to impersonate a BigSpark BYOSA
    # service account to create the event log directory
    if [[ "${DATAPROC_IMAGE_TYPE}" == "serverless-spark" ]]; then
      local spark_byosa
      spark_byosa=$(get_dataproc_property spark.byosa.submitter.service-account)
      if [[ -n "${spark_byosa}" ]]; then
        # Use the VM service account to impersonate the BYOSA service account
        token=$(curl -H "Authorization: Bearer ${token}" -H "Content-Type:application/json" \
          -d '{"scope":["https://www.googleapis.com/auth/cloud-platform"]}' \
          "https://iamcredentials.googleapis.com/v1/projects/-/serviceAccounts/${spark_byosa}:generateAccessToken" |
          jq -r ".accessToken")
      fi
    fi

    # Ignore curl exit code because we rely on HTTP status code verification.
    set +e

    # Use curl to create a directory object - gcloud can not do this.
    # Set `ifGenerationMatch=0` precondition to not override
    # directory object if it already exists.
    local http_code
    http_code=$(curl -fs \
      --retry-connrefused --retry 5 --retry-max-time 30 \
      --data-binary @/dev/null \
      --output /dev/null \
      --write-out "%{http_code}" \
      -X POST \
      -H "Authorization: Bearer ${token}" \
      -H "Content-Type: application/octet-stream" \
      "https://storage.googleapis.com/upload/storage/v1/b/${event_log_dir_bucket}/o?uploadType=media&name=${event_log_dir_path}/&ifGenerationMatch=0")
    local -r exit_code=$?

    set -ex

    # 412 status code is a success - it means that the directory already exists
    # because `ifGenerationMatch=0` precondition was set.
    if [[ "${http_code}" -ge "200" && "${http_code}" -le "299" ]] ||
      [[ "${http_code}" == "412" ]]; then
      return 0
    fi

    log_and_fail "spark" \
      "Failed to create event log dir with ${http_code} HTTP response code and ${exit_code} exit code"
  elif is_in_cluster_hdfs "${event_log_dir}"; then
    # In-cluster HDFS, will create the dir in post-hdfs.
    loginfo "Delay creating ${event_log_dir}, HDFS is not ready yet"
  elif [[ "${event_log_dir}" == hdfs://* ]]; then
    # External HDFS, best effort
    create_hdfs_dir "${event_log_dir}" spark:spark 755 || true
  fi
}

function merge_user_properties() {
  merge_java_properties "${CLUSTER_PROPERTIES_DIR}/spark.properties" \
    "${SPARK_CONF_DIR}/spark-defaults.conf"
  merge_sh_env_vars "${CLUSTER_PROPERTIES_DIR}/spark-env.sh" \
    "${SPARK_CONF_DIR}/spark-env.sh"

  if [[ "${DATAPROC_IMAGE_TYPE}" == "standard" ]] &&
    ! is_version_at_least "${DATAPROC_IMAGE_VERSION}" "2.1"; then
    merge_java_properties "${CLUSTER_PROPERTIES_DIR}/spark-log4j.properties" \
      "${SPARK_CONF_DIR}/log4j.properties"
  else
    merge_java_properties "${CLUSTER_PROPERTIES_DIR}/spark-log4j2.properties" \
      "${SPARK_CONF_DIR}/log4j2.properties"
  fi
}

# Configure authentication so that only intended members could join the cluster.
function configure_authentication_standalone() {
  if [[ ! -f "${SPARK_AUTH_SECRET_FILE}" ]]; then
    loginfo "Error: ${SPARK_AUTH_SECRET_FILE} was not found."
    exit 1
  fi

  # Disable logs of secret
  set +x
  local auth_secret
  auth_secret="$(<"${SPARK_AUTH_SECRET_FILE}")"
  cat >>"${SPARK_CONF_DIR}/spark-defaults.conf" <<EOF

# Dataproc configuration override
spark.authenticate=true
spark.authenticate.secret=${auth_secret}
EOF
  set -x
}

function configure_metastore_standalone() {
  if is_component_selected "dpms-proxy"; then
    cat >>"${SPARK_CONF_DIR}/spark-defaults.conf" <<EOF

# Use DPMS via gRPC proxy as a Hive metastore for Spark
spark.hadoop.hive.metastore.uris=${CLUSTER_MASTER_METASTORE_URIS}
EOF
  fi
}

function configure_network_encryption() {
  cat >>"${SPARK_CONF_DIR}/spark-defaults.conf" <<EOF

# Turn on AES based encryption for Spark network communication
spark.network.crypto.enabled=true
spark.network.crypto.keyLength=256
spark.network.crypto.saslFallback=false
EOF
}

# Configure network routing rules for Spark standalone so that the assorted
# Spark UIs are only accessible from within the VM.
function configure_network_standalone() {
  for PORT in ${SPARK_UI_PORT} ${SPARK_MASTER_WEBUI_PORT} ${SPARK_WORKER_WEBUI_PORT}; do
    iptables -A INPUT -p tcp --dport "${PORT}" -s 127.0.0.1 -j ACCEPT
    iptables -A INPUT -p tcp --dport "${PORT}" -s "$(hostname -i)" -j ACCEPT
    iptables -A INPUT -p tcp --dport "${PORT}" -j DROP
  done
  save_iptables
}

function configure_authentication_and_encryption_on_yarn() {
  cat >>"${SPARK_CONF_DIR}/spark-defaults.conf" <<EOF

spark.authenticate=true
EOF

  # Set same properties for the Spark shuffle service
  set_property_yarn_site "spark.authenticate" "true"
  set_property_yarn_site "spark.network.crypto.enabled" "true"
  set_property_yarn_site "spark.network.crypto.keyLength" "256"
  set_property_yarn_site "spark.network.crypto.saslFallback" "false"
}

# Configures GCS Maven Central mirror.
#
# In serverless, clusters are always internal-ip-only, they don't have access
# to Maven Central, unless the network has been configured with NAT. But the
# GCS Maven Central mirrors are accessible if the subnet has Private Google
# Access enabled. We removed the default Maven Central for serverless to avoid
# any potential connection timeout errors.
#
# TODO: add to DPGCE after testing with serverless, and add the default
#  Maven Central after the GCS mirror.
function configure_gcs_maven_mirror() {
  # Choose region-specific mirror if possible.
  # https://storage-download.googleapis.com/maven-central/index.html
  local mirror_uri
  if [[ "${ZONE}" == europe-* ]]; then
    mirror_uri='https://maven-central-eu.storage-download.googleapis.com/maven2/'
  elif [[ "${ZONE}" == asia-* ]]; then
    mirror_uri='https://maven-central-asia.storage-download.googleapis.com/maven2/'
  else
    # The default US mirror.
    mirror_uri='https://maven-central.storage-download.googleapis.com/maven2/'
  fi

  cat >"${SPARK_CONF_DIR}/ivysettings.xml" <<EOF
<ivysettings>
    <settings defaultResolver="chain"/>
    <resolvers>
        <chain name="chain">
            <ibiblio name="gcs-maven-mirror" m2compatible="true" root="${mirror_uri}"/>
        </chain>
    </resolvers>
</ivysettings>
EOF

  cat >>"${SPARK_CONF_DIR}/spark-defaults.conf" <<EOF

# GCS Maven Central mirror
spark.jars.ivySettings=${SPARK_CONF_DIR}/ivysettings.xml
EOF
}

function configure_spark_gpus() {
  if [[ "${ROLE}" == "Master" && "${MASTER_INDEX}" == "0" ]]; then
    cat >>"${SPARK_CONF_DIR}/spark-defaults.conf" <<EOF

# Make Spark aware of executor GPUs
spark.executor.resource.gpu.amount=1
spark.executor.resource.gpu.discoveryScript=/usr/lib/spark/scripts/gpu/getGpusResources.sh
EOF
  fi
}

function configure_spark_rapids() {
  local use_default_jars
  use_default_jars=$(get_metadata_spark_rapids_use_default_jars || true)
  if [[ "${use_default_jars:-true}" != 'true' ]]; then
    return
  fi

  # Add default Rapids jars to classpath
  ln -s "${SPARK_CONNECTOR_DIR}"/rapids-*.jar "${SPARK_JARS_DIR}"

  if [[ "${ROLE}" == "Master" && "${MASTER_INDEX}" == "0" ]]; then
    cat >>"${SPARK_CONF_DIR}/spark-defaults.conf" <<EOF

# RAPIDS properties for Spark
spark.plugins=com.nvidia.spark.SQLPlugin
EOF
  fi
}

function configure_hive_scratch_dir_standalone() {
  local -r hive_scratch_dir="${SPARK_TMP_DIR}/hive"

  mkdir -p "${hive_scratch_dir}"
  chown -R spark:spark "${SPARK_TMP_DIR}"
  chmod -R 1777 "${SPARK_TMP_DIR}"

  cat >>"${SPARK_CONF_DIR}/spark-defaults.conf" <<EOF

spark.hadoop.hive.exec.local.scratchdir=${hive_scratch_dir}
spark.hadoop.hive.exec.scratchdir=${hive_scratch_dir}
EOF
}

function configure_spark_buffer() {
  local -r ingestion_enabled=$(get_java_property_or_default \
    "${SPARK_CONF_DIR}/spark-defaults.conf" \
    spark.dataproc.appContext.enabled false)
  if [[ "${ingestion_enabled}" == "true" ]]; then
    ln -s "${SPARK_CONNECTOR_DIR}/$(basename "${SPARK_DATAPROC_BUFFER_URI}")" "${SPARK_JARS_DIR}"
  fi
}

function pre_activate_spark() {
  update_bq_connector
  init_local_dirs_common
  configure_env_common
  configure_arrow
  configure_event_log_dir

  if [[ "${DATAPROC_IMAGE_TYPE}" == "serverless-spark" ]]; then
    init_local_dirs_standalone
    configure_defaults_standalone
    configure_metastore_standalone
    configure_gcs_maven_mirror
  else
    configure_env_yarn
    configure_defaults_yarn
    configure_efm
    configure_lineage

    if is_component_selected "kerberos"; then
      configure_network_encryption
      configure_authentication_and_encryption_on_yarn
    fi
  fi

  merge_user_properties

  # Configure authentication and other overrides after merging user-provided
  # properties, because we don't allow users to override these properties.
  if [[ "${DATAPROC_IMAGE_TYPE}" == "serverless-spark" ]]; then
    configure_env_standalone
    configure_hive_scratch_dir_standalone
    if [[ "${DATAPROC_IMAGE_VERSION}" == '1.2' ]] || is_version_at_least "${DATAPROC_IMAGE_VERSION}" '2.2'; then
      # Configure broadcast join threshold as 2% of
      # Spark executor memory with 200MiB max
      configure_broadcast_join '0.02' '200'
    else
      # Configure broadcast join threshold as 2% of
      # Spark executor memory with 500MiB max
      configure_broadcast_join '0.02' '500'
    fi
    configure_driver_max_result_size
    configure_network_standalone

    if [[ -v SPARK_DATAPROC_BUFFER_URI ]]; then
      configure_spark_buffer
    fi

    if is_component_selected 'nvidia-drivers'; then
      configure_spark_gpus
      configure_spark_rapids

      # Add default XGBoost jars to classpath
      if ! is_version_at_least "${DATAPROC_IMAGE_VERSION}" '2.0'; then
        ln -s "${SPARK_CONNECTOR_DIR}"/xgboost4j-*.jar "${SPARK_JARS_DIR}"
      fi
    fi

    if ! is_version_at_least "${DATAPROC_IMAGE_VERSION}" '3.0'; then
      configure_network_encryption
      configure_authentication_standalone
      configure_standalone_worker_memory
    fi
  elif is_version_at_least "${DATAPROC_IMAGE_VERSION}" '2.0'; then
    # Configure broadcast join threshold as 0.75% of
    # Spark executor memory with 200MiB max
    configure_broadcast_join '0.0075' '200'
  fi

  # Create the event log directory after merging properties to make sure
  # that we pick up the exact directory Spark will use.
  if [[ "${ROLE}" == "Master" && "${MASTER_INDEX}" == "0" ]]; then
    run_in_background --tag 'spark-create-event-log-dir' create_event_log_dir
  fi
}

pre_activate_spark
