#!/bin/bash

set -euo pipefail

source "$(dirname "$0")/../../bdutil_env.sh"
source "$(dirname "$0")/../../bdutil_logging.sh"
source "$(dirname "$0")/../../bdutil_properties.sh"
source "$(dirname "$0")/../../bdutil_components.sh"
source "$(dirname "$0")/../../bdutil_versions.sh"
source "${DATAPROC_DIR}/dataproc_env.sh"

set -x

set_log_tag "pre-activate-component-tez"

readonly TEZ_UI_WAR=$(ls "${TEZ_LIB_DIR}"/tez-ui-*.war)

# Configure ATS to serve Tez UI. Tez UI can be accessed at
# http://clustername-m:8188/tez or http://clustername-m-0:8188/tez.
function configure_ui_war() {
  local tmp_dir
  tmp_dir=$(mktemp -d)
  unzip -q "${TEZ_UI_WAR}" -d "${tmp_dir}"

  local -r tez_configs="${tmp_dir}/config/configs.env"
  local -r tez_ui_js="${tmp_dir}/assets/tez-ui.js"

  local ats_v2_port=8192
  if is_component_selected 'kerberos'; then
    ats_v2_port=8188
  fi

  local ats_v2_enabled=false
  if [[ -n "$(get_dataproc_property yarn.atsv2.bigtable.instance)" ]] ; then
    ats_v2_enabled=true
  fi

  # Replace localhost with the host name in Tez config (javascript).
  if [[ -f "${tez_configs}" ]]; then
    # Find the commented out line in the config that sets default the timeline
    # and rm host/port (http://localhost:8188 and http://localhost:8088)
    # Replace these lines with the actual VM hostname to support accessing
    # the UIs through an ssh tunnel
    sed -i "s#\(.*\)//atsV2Enabled: true\(.*\)#\1atsV2Enabled: ${ats_v2_enabled}\2#" "${tez_configs}"
    sed -i "s#\(.*\)//timeline: \"http://localhost:8188\"\(.*\)#\1timeline: \"http://${HOSTNAME}:8188\"\2#" "${tez_configs}"
    sed -i "s#\(.*\)//timelineV2: \"http://localhost:8192\"\(.*\)#\1timelineV2: \"http://${HOSTNAME}:${ats_v2_port}\"\2#" "${tez_configs}"
    sed -i "s#\(.*\)//rm: \"http://localhost:8088\"\(.*\)#\1rm: \"http://${HOSTNAME}:8088\"\2#" "${tez_configs}"
  fi

  # Replace localhost with the host name in tez_ui.js.
  if [[ -f "${tez_ui_js}" ]]; then
    # Find the line in the script that sets default the timeline and rm
    # host/port (http://localhost:8188 and http://localhost:8088)
    # Replace these lines with the actual VM hostname to support accessing
    # the UIs through an ssh tunnel
    sed -i "s#\"timeline\":\"localhost:8188\"#\"timeline\":\"${HOSTNAME}:8188\"#" "${tez_ui_js}"
    sed -i "s#\"timelineV2\":\"localhost:8192\"#\"timelineV2\":\"${HOSTNAME}:8192\"#" "${tez_ui_js}"
    sed -i "s#\"rm\":\"localhost:8088\"#\"rm\":\"${HOSTNAME}:8088\"#" "${tez_ui_js}"
  fi

  # If component gateway is enabled, add support for proxying Tez UI
  if is_component_selected 'knox'; then
    local cluster_ui_hostname
    cluster_ui_hostname=$(get_dataproc_property \
      dataproc.proxy.public.hostname |
      sed 's/\\:/:/') # Unescape the colon in the URL

    # Create a customized config file with a .js suffix for correct MIME type
    cat >"${tmp_dir}/config/configs.js" <<EOF
// This file overrides the Tez UI default config when using the Component Gateway
// Sets the target host and base path for API requests from client JS
ENV = {
  atsV2Enabled: ${ats_v2_enabled},
  hosts: {
    timeline: "${cluster_ui_hostname}",
    timelineV2: "${cluster_ui_hostname}",
    rm: "${cluster_ui_hostname}",
  },

  namespaces: {
    webService: {
      timeline: "gateway/default/apphistory/ws/v1/timeline",
      timelineV2: "gateway/default/apphistory/ws/v2/timeline",
      appHistory: "gateway/default/apphistory/ws/v1/applicationhistory",
      rm: "gateway/default/yarn/ws/v1/cluster",
      am: "gateway/default/yarn/proxy/{app_id}/ws/v{version:2}/tez",
    },
  },
};
EOF
  fi

  # cd into tmp_dir so we don't include the directory prefix in the zipped files
  cd "${tmp_dir}"
  zip -q "${TEZ_UI_WAR}" -r ./*
  cd ..
  rm -rf "${tmp_dir}"
  # Reset the edit time of the war to the same as other jars so Tez doesn't
  # complain about desynced resources on different nodes
  touch -d "@$(stat /usr/lib/tez/tez-common-*.jar --format=%Y)" "${TEZ_UI_WAR}"
}

function configure_yarn_for_tez() {
  # Declare UI named tez
  bdconfig set_property \
    --configuration_file "${HADOOP_CONF_DIR}/yarn-site.xml" \
    --name 'yarn.timeline-service.ui-names' --value 'tez' \
    --clobber
  # Set web archive file for UI called tez
  bdconfig set_property \
    --configuration_file "${HADOOP_CONF_DIR}/yarn-site.xml" \
    --name 'yarn.timeline-service.ui-on-disk-path.tez' --value "${TEZ_UI_WAR}" \
    --clobber
  # Serve UI called tez at /tez-ui
  bdconfig set_property \
    --configuration_file "${HADOOP_CONF_DIR}/yarn-site.xml" \
    --name 'yarn.timeline-service.ui-web-path.tez' --value '/tez-ui' \
    --clobber
}

function configure_tez() {
  local history_logging_service_class="ATSHistoryLoggingService"
  if [[ -n "$(get_dataproc_property yarn.atsv2.bigtable.instance)" ]] ; then
    history_logging_service_class="ATSV2HistoryLoggingService"
  fi

  # Configure the Tez UI to store history in ATS.
  # See https://tez.apache.org/tez-ui.html for more information.
  bdconfig set_property \
    --configuration_file "${TEZ_CONF_DIR}/tez-site.xml" \
    --name 'tez.history.logging.service.class' \
    --value "org.apache.tez.dag.history.logging.ats.${history_logging_service_class}" \
    --clobber
  bdconfig set_property \
    --configuration_file "${TEZ_CONF_DIR}/tez-site.xml" \
    --name 'tez.tez-ui.history-url.base' \
    --value "http://${MASTER_HOSTNAME_0}:8188/tez-ui/" \
    --clobber

  # Disable node blacklisting. Blacklisting makes sense for on-prem clusters,
  # where the main cause of nodes disappearing is actual node failures. In GCE,
  # due to live migration, nodes failing is pretty rare. More often, nodes will
  # disappear because a) preemptible vms getting preempted or b) (auto)-scaling.
  # In both cases, VMs can come back with the same name, meaning they may be
  # blacklisted upon return.
  bdconfig set_property \
    --configuration_file "${TEZ_CONF_DIR}/tez-site.xml" \
    --name 'tez.am.node-blacklisting.enabled' \
    --value 'false' \
    --clobber
}

function merge_user_properties() {
  if [[ -f "${TEZ_CONF_DIR}/tez-site.xml" ]]; then
    merge_xml_properties "${CLUSTER_PROPERTIES_DIR}/tez.xml" \
      "${TEZ_CONF_DIR}/tez-site.xml"
  fi
}

function pre_activate_tez() {
  configure_ui_war
  configure_yarn_for_tez
  configure_tez
  merge_user_properties
}

pre_activate_tez
