#!/bin/bash

set -euo pipefail

source "$(dirname "$0")/../../bdutil_env.sh"
source "$(dirname "$0")/../../bdutil_logging.sh"
source "$(dirname "$0")/../../bdutil_metadata.sh"
source "$(dirname "$0")/../../bdutil_misc.sh"
source "$(dirname "$0")/../../bdutil_os.sh"
source "$(dirname "$0")/../../bdutil_properties.sh"
source "$(dirname "$0")/../../bdutil_versions.sh"
source "$(dirname "$0")/../../bdutil_components.sh"
source "$(dirname "$0")/../../bdutil_collections.sh"
source "$(dirname "$0")/../../configure_bigtable_hbase_client.sh"

set -x

set_log_tag "pre-activate-component-yarn"

readonly ATSV2_BIGTABLE_RESOURCE="$(get_dataproc_property yarn.atsv2.bigtable.instance)"

function create_table() {
  local -r project="$1"
  local -r instance_id="$2"
  local -r table="$3"
  local -r families="$4"

  # If table doesn't exist then create table.
  if ! cbt -project "${project}" -instance "${instance_id}" ls "${table}" ; then
    cbt -project "${project}" -instance "${instance_id}" \
      createtable "${table}" "families=${families}"
    return $?
  fi

  # If table exists then create column families individually if they don't exist.
  IFS=',' read -ra families_array <<< "${families}"
  for family in "${families_array[@]}"; do
    if ! cbt -project "${project}" -instance "${instance_id}" ls "${table}" | grep "^${family}\s" ; then
      cbt -project "${project}" -instance "${instance_id}" \
        createfamily "${table}" "${family}"
      cbt -project "${project}" -instance "${instance_id}" \
        setgcpolicy "${table}" "${family}" maxversions=1
    fi
  done
}

function create_atsv2_schema() {
  local -r bigtable_project="$1"
  local -r bigtable_instance_id="$2"
  local -r prefix=$(get_property_in_xml "${HADOOP_CONF_DIR}/yarn-site.xml" \
    yarn.timeline-service.hbase-schema.prefix "prod.")

  create_table "${bigtable_project}" "${bigtable_instance_id}" \
    "${prefix}timelineservice.entity" "c,i,m"
  create_table "${bigtable_project}" "${bigtable_instance_id}" \
    "${prefix}timelineservice.application" "c,i,m"
  create_table "${bigtable_project}" "${bigtable_instance_id}" \
    "${prefix}timelineservice.app_flow" "m"
  create_table "${bigtable_project}" "${bigtable_instance_id}" \
    "${prefix}timelineservice.flowactivity" "i"
  create_table "${bigtable_project}" "${bigtable_instance_id}" \
    "${prefix}timelineservice.flowrun" "i"
  create_table "${bigtable_project}" "${bigtable_instance_id}" \
    "${prefix}timelineservice.subapplication" "c,i,m"
}

function configure_timelinereader_service() {
  local -r service_name="hadoop-yarn-timelinereader"
  #Update classpaths
  cat >>"${HADOOP_CONF_DIR}/hadoop-env.sh" <<EOF
export JAVA_HOME="${JAVA_HOME}"
EOF

  # Timelinereader service executes a script and it exits. Service is running
  # even after it exits. Service type as OneShot with RemainAfterExit keeps
  # status as active in this case.
  cat >"/etc/systemd/system/${service_name}.service" <<EOF
[Unit]
Description=Hadoop yarn timeline reader
Wants=network-online.target
After=network-online.target
[Service]
Type=oneshot
ExecStart=/bin/bash -c 'HADOOP_HOME=/usr/lib/hadoop HADOOP_LIBEXEC_DIR=/usr/lib/hadoop/libexec /usr/lib/hadoop-yarn/sbin/yarn-daemon.sh start timelinereader'
ExecStop=/bin/bash -c 'HADOOP_HOME=/usr/lib/hadoop HADOOP_LIBEXEC_DIR=/usr/lib/hadoop/libexec /usr/lib/hadoop-yarn/sbin/yarn-daemon.sh stop timelinereader'
RemainAfterExit=yes
[Install]
WantedBy=multi-user.target
EOF

  systemctl daemon-reload
}

function configure_atsv2() {
  # Dataproc frontend performs validation for Bigtable instance format.
  # Expected value is a fully-qualified GCP resource path, sample:
  # projects/<project_id>/instances/<instance_id>
  local -r project_id=$(echo "${ATSV2_BIGTABLE_RESOURCE}" | cut -d\/ -f2)
  local -r instance_id=$(echo "${ATSV2_BIGTABLE_RESOURCE}" | cut -d\/ -f4)

  # Only create ATSv2 schema from 1 node in the cluster, i.e., master-0.
  # This is done to avoid race condition where multiple nodes will try to create
  # tables "simultaneously" and latter requests will fail with `AlreadyExists`
  # exception.
  if [[ "$(get_metadata_role)" == "Master" ]] \
    && [[ "${MASTER_COUNT}" == "1" || "${MASTER_INDEX}" == "0" ]]; then

    if ! cbt -project "${project_id}" -instance "${instance_id}" ls ; then
      loginfo "Invalid Bigtable instance specified for ATSv2 ${ATSV2_BIGTABLE_RESOURCE}"
      return 1
    fi

    create_atsv2_schema "${project_id}" "${instance_id}"
  fi
  configure_bigtable_hbase_client "${project_id}" "${instance_id}"

  # Setting the default values to enable the ATSv2
  set_property_yarn_site "yarn.timeline-service.version" "2.0"
  set_property_yarn_site "yarn.timeline-service.enabled" "true"
  set_property_yarn_site "yarn.nodemanager.aux-services.timeline_collector.class" "org.apache.hadoop.yarn.server.timelineservice.collector.PerNodeTimelineCollectorsAuxService"
  set_property_yarn_site "yarn.system-metrics-publisher.enabled" "true"
  set_property_yarn_site "yarn.timeline-service.hbase.configuration.file" "file:${BIGTABLE_HBASE_CONF_DIR}/hbase-site.xml"
  set_property_yarn_site "yarn.timeline-service.hostname" "0.0.0.0"
  if is_component_selected 'kerberos'; then
    set_property_yarn_site "yarn.timeline-service.collector.bind-host" "${MY_FULL_HOSTNAME}"
    set_property_yarn_site "yarn.timeline-service.reader.webapp.https.address" "${MY_FULL_HOSTNAME}:8188"
    set_property_yarn_site "yarn.timeline-service.http-authentication.type" "simple"
  else
    set_property_yarn_site "yarn.timeline-service.reader.webapp.address" "${MY_HOSTNAME}:8192"
  fi
  set_property_mapred_site "mapreduce.job.emit-timeline-data" "true"

  # We need to enable timeline_collector and mapreduce_shuffle nodemanager aux
  # services as default configuration for ATSv2 to properly function.This logic
  # append timeline_collector and mapreduce_shuffle to the
  # yarn.nodemanager.aux-services.
  local aux_services_str
  aux_services_str=$(get_property_in_xml "${HADOOP_CONF_DIR}/yarn-site.xml" \
    yarn.nodemanager.aux-services "")
  if [[ -n "${aux_services_str}" ]] ; then
    IFS=',' read -r -a aux_services_arr <<< "$aux_services_str"
    local required_nm_aux_services=("timeline_collector" "mapreduce_shuffle")
    local diff_aux_services=($(difference required_nm_aux_services aux_services_arr))
    local aux_service_arr=($(union aux_services_arr diff_aux_services))
    printf -v aux_service_str '%s,' "${aux_service_arr[@]}"
    aux_service_str="${aux_service_str%,}"
    set_property_yarn_site "yarn.nodemanager.aux-services" "${aux_service_str}"
  fi
}

function configure_ui2() {
  set_property_yarn_site "yarn.webapp.ui2.enable" "true"
  set_property_yarn_site "yarn.webapp.ui2.war-file-path" "/usr/lib/hadoop-yarn/hadoop-yarn-ui.war"
  set_property_yarn_site "yarn.timeline-service.http-cross-origin.enabled" "true"
  set_property_yarn_site "yarn.resourcemanager.webapp.cross-origin.enabled" "true"
  set_property_yarn_site "yarn.nodemanager.webapp.cross-origin.enabled" "true"
}

# Symlink jersey-core-1x and jersey-client-1x jars from hadoop lib path to spark
# classpath. This is a hack required for ATSv2 client to function for spark jobs.
# More details on why this required is captured in b/211614373
# Upstream tracker for jersey upgrade to 2.x in hadoop: HADOOP-15984
function symlink_jersey1x_jars_to_spark_classpath() {
  local -r jersey_core_jar=$(find /usr/lib/hadoop/lib/ -type f -name "jersey-core-1.*.jar")
  local -r jersey_client_jar=$(find /usr/lib/hadoop-yarn/lib/ -type f -name "jersey-client-1.*.jar")

  ln -s "${jersey_core_jar}" /usr/lib/spark/jars/
  ln -s "${jersey_client_jar}" /usr/lib/spark/jars/
}

# ATSv2 functionality will be configured only if a valid ATSV2_BIGTABLE_RESOURCE
# (Bigtable instance Id) is provided while cluster creation.
if [[ -n "${ATSV2_BIGTABLE_RESOURCE}" ]] ; then
  configure_timelinereader_service
  configure_atsv2
  configure_ui2

  if is_version_at_least "${DATAPROC_IMAGE_VERSION}" '2.0'; then
    symlink_jersey1x_jars_to_spark_classpath
  fi
else
  rm -Rf "${DATAPROC_COMMON_LIB_DIR}/${BIGTABLE_HBASE_CLIENT_JAR}"
fi

function setup_yarn_sasl() {
  local realm
  realm=$(get_dataproc_property kerberos.realm)
  if [[ -z "${realm}" ]]; then
    realm=$(echo "$(dnsdomainname)" | awk '{print toupper($0)}')
  fi

  # Create a new yarn jaas file
  cat >"${HADOOP_CONF_DIR}/yarn_jaas.conf" <<EOF
Client {
  com.sun.security.auth.module.Krb5LoginModule required
  useKeyTab=true
  keyTab="${KEYTAB_DIR}/yarn.service.keytab"
  storeKey=true
  useTicketCache=false
  principal="yarn/$(hostname -f)@${realm}";
};
EOF


  # Append to /etc/hadoop/conf/yarn-env.sh to make ResourceManager use
  # the ZooKeeper SASL properties and the JAAS configuration file:
  cat >>"${HADOOP_CONF_DIR}/yarn-env.sh" <<EOF

YARN_RESOURCEMANAGER_OPTS="-Dzookeeper.sasl.client=true -Dzookeeper.sasl.client.username=zookeeper \
-Djava.security.auth.login.config=${HADOOP_CONF_DIR}/yarn_jaas.conf \
-Dzookeeper.sasl.clientconfig=Client \${YARN_RESOURCEMANAGER_OPTS}"
EOF
}

# If the cluster is kerberized and is HA mode i.e.: multi-master cluster
if is_component_selected "kerberos"; then
  # Regular (1 master, n workers) or single-node cluster.
  if [[ "${MASTER_COUNT}" == "1" || "${WORKER_COUNT}" == "0" ]]; then
    # do nothing for single node
    echo "Number of master is 1 or worker count is 0 - SASL setup not required."
  else
    # Setup only on master nodes.
    if [[ "$(get_metadata_role)" == "Master" ]] ; then
      bdconfig set_property --configuration_file "${HADOOP_CONF_DIR}"/yarn-site.xml \
        --name 'yarn.resourcemanager.zk-acl' --value 'sasl:yarn:cdraw,world:anyone:r' --clobber
      setup_yarn_sasl
    fi
  fi
fi
