#!/bin/bash

set -euo pipefail

source "$(dirname "$0")/../../bdutil_env.sh"
source "$(dirname "$0")/../../bdutil_versions.sh"
source "$(dirname "$0")/../../bdutil_logging.sh"
source "$(dirname "$0")/../../bdutil_properties.sh"
source "$(dirname "$0")/../../bdutil_packages.sh"
source "$(dirname "$0")/../../bdutil_components.sh"
source "$(dirname "$0")/../shared/zookeeper.sh"
source "${DATAPROC_DIR}/dataproc_env.sh"

set -x

set_log_tag "pre-activate-component-zookeeper"

readonly ZOOKEEPER_DATA_DIR="/var/lib/zookeeper"

function configure_zookeeper() {
  for ((i = 0; i < ${#MASTER_HOSTNAMES[@]}; i++)); do
    echo "server.${i}=${MASTER_HOSTNAMES[i]}:2888:3888" >>"${ZOOKEEPER_CONFIG_FILE}"
  done

  # Automatically purge snapshots and transaction logs every week.
  echo "autopurge.purgeInterval=168" >>"${ZOOKEEPER_CONFIG_FILE}"

  if [[ "${ROLE}" == "Master" ]]; then
    MY_ID=$(uname -n | sed -e 's/.*-m-//')
    # Above does not work if it's a standalone zookeeper server
    if [[ "$MY_ID" =~ -m ]]; then
      MY_ID=1
    fi
    # Make sure the owner of the myid file is zookeeper (b/224950450).
    sudo -u zookeeper bash -c "echo ${MY_ID} >${ZOOKEEPER_DATA_DIR}/myid"
    if is_version_at_least "${DATAPROC_IMAGE_VERSION}" "2.1"; then
      sudo -u zookeeper mkdir -p "${ZOOKEEPER_DATA_DIR}/version-2"
      echo "admin.serverPort=38080" >>"${ZOOKEEPER_CONFIG_FILE}"
      # display zookeeper client warnings or errors on console
      add_console_warn_property
      add_logdir_property
      # required for zookeeper client to write to logs
      touch "${ZOOKEEPER_LOG_FILE}"
      chmod 777 "${ZOOKEEPER_LOG_FILE}"
    fi
  fi
}

function merge_user_properties() {
  merge_java_properties \
    "${CLUSTER_PROPERTIES_DIR}/zookeeper.properties" \
    "${ZOOKEEPER_CONFIG_FILE}"
  merge_java_properties \
    "${CLUSTER_PROPERTIES_DIR}/zookeeper-log4j.properties" \
    "${ZOOKEEPER_CONF_DIR}/log4j.properties"
}

# Once the config are set, and hdfs_zkfs needs to be formatted, before SASL is effective.
function secure_zookeeper_with_sasl() {
  local realm
  realm=$(get_dataproc_property kerberos.realm)
  if [[ -z "${realm}" ]]; then
    realm=$(echo "$(dnsdomainname)" | awk '{print toupper($0)}')
  fi

  local -r hdfs_jaas_file="${HADOOP_CONF_DIR}"/hdfs_jaas.conf
  local -r zk_jaas_file="/etc/zookeeper/conf/jaas.conf"
  local -r dataproc_jaas_file="${HADOOP_CONF_DIR}"/dataproc_jaas.conf

  loginfo "Securing Zookeeper with SASL in a HA environment"
  # Add property for sasl in core-site.xml so that every entry under
  # zookeepers /hadoop-ha znode is protected with sasl (kerberos)
  bdconfig set_property --configuration_file '/etc/hadoop/conf/core-site.xml' \
    --name 'ha.zookeeper.acl' --value 'sasl:hdfs:cdraw,world:anyone:r' --clobber
  loginfo 'Setting up SASL properties for hdfs and zookeeper'

 # Create a new hadoop hdfs_jaas.conf file
  cat > "${hdfs_jaas_file}" <<EOF
Client {
  com.sun.security.auth.module.Krb5LoginModule required
  useKeyTab=true
  keyTab="/etc/security/keytab/hdfs.service.keytab"
  storeKey=true
  useTicketCache=false
  principal="hdfs/$(hostname -f)@${realm}";
};
EOF

  cat >> "${HADOOP_CONF_DIR}/hadoop-env.sh" <<EOF
HADOOP_ZKFC_OPTS="-Dzookeeper.sasl.client=true -Dzookeeper.sasl.client.username=zookeeper \
-Djava.security.auth.login.config=${hdfs_jaas_file} \
-Dzookeeper.sasl.clientconfig=Client \${HADOOP_ZKFC_OPTS}"
EOF

  # Create a new dataproc daatproc_jaas.conf file
  cat > "${dataproc_jaas_file}" <<EOF
Client {
  com.sun.security.auth.module.Krb5LoginModule required
  useKeyTab=true
  keyTab="/etc/security/keytab/dataproc.service.keytab"
  storeKey=true
  useTicketCache=false
  principal="dataproc/$(hostname -f)@${realm}";
};
EOF

  bdconfig set_property --configuration_file "${HADOOP_CONF_DIR}"/core-site.xml \
        --name 'dataproc.agent.ha.zk-acl' --value 'sasl:dataproc:cdraw,world:anyone:r' --clobber

  # Append to the existing zk jaas.conf file
  cat >>"${zk_jaas_file}" <<EOF
Client {
  com.sun.security.auth.module.Krb5LoginModule required
  useKeyTab=true
  keyTab="${KEYTAB_DIR}/zookeeper.service.keytab"
  storeKey=true
  useTicketCache=false
  principal="zookeeper/$(hostname -f)@${realm}";
};
EOF
}

function pre_activate_zookeeper() {
  if [[ "${ROLE}" == 'Master' ]] && is_component_selected 'zookeeper' && is_component_selected "kerberos" ; then
    secure_zookeeper_with_sasl
  fi
  configure_zookeeper
  merge_user_properties
  if [[ "${ROLE}" != "Master" ]]; then
    mark_packages_to_uninstall "zookeeper-server"
  fi
}

pre_activate_zookeeper
