#!/bin/bash

function create_and_activate_environment() {
  local -r conda_install_path=$1
  local -r conda_bin_dir=$2
  local -r conda_env_name=$3
  local -r conda_env_config=$4

  local installer
  if [[ -f "${conda_bin_dir}/mamba" ]]; then
    installer="${conda_bin_dir}/mamba"
  else
    installer="${conda_bin_dir}/conda"
  fi

  retry_constant_custom 4 1 "${installer}" env create \
      --quiet \
      --name="${conda_env_name}" \
      --file="${conda_env_config}" \
      --force
  source "${conda_bin_dir}/activate" "${conda_env_name}"

  # Change PYSPARK_PYTHON path to point to python in custom Conda environment
  # Append python profile with code to activate env, this activates custom Conda
  # environment on startup of bash shell

  cat >>/etc/profile.d/effective-python.sh <<EOT

# Use custom Conda environment for PySpark
export PYSPARK_PYTHON="${conda_install_path}/envs/${conda_env_name}/bin/python"

# Activate custom Conda environment
conda activate "${conda_env_name}"
EOT
}

function install_conda_packages() {
  local -r conda_bin_dir=$1
  local -r conda_packages=$2

  if [[ -n "${conda_packages}" ]]; then
    local -a packages
    IFS=',' read -r -a packages <<< "${conda_packages}"

    # Conda will upgrade dependencies only if required, and fail if conflict
    # resolution with existing packages is not possible.
    local installer
    if [[ -f "${conda_bin_dir}/mamba" ]]; then
      installer="${conda_bin_dir}/mamba"
    else
      installer="${conda_bin_dir}/conda"
    fi
    "${installer}" install "${packages[@]}" --yes
  fi
}

function install_pip_packages() {
  local -r pip_packages=$1
  if [[ -n "${pip_packages}" ]]; then
    local -a packages
    IFS=',' read -r -a packages <<< "${pip_packages}"

    # Pip will upgrade dependencies only if required. Pip does not check for
    # conflicts and may result in inconsistent environment.
    pip install -U --upgrade-strategy only-if-needed "${packages[@]}"
  fi
}

function customize_conda_env() {
  local -r conda_install_path=$1
  local -r conda_bin_dir=$2
  local -r conda_env_config_uri=$(get_dataproc_property conda.env.config.uri)
  local -r conda_packages=$(get_dataproc_property conda.packages)
  local -r pip_packages=$(get_dataproc_property pip.packages)

  if [[ -n "${conda_env_config_uri}" ]]; then
    local temp_config_file
    temp_config_file=$(mktemp /tmp/conda_env_XXX.yaml)
    download_resource "${conda_env_config_uri}" "${temp_config_file}"
    local conda_env_name
    conda_env_name="$(grep 'name: ' "${temp_config_file}" | awk '{print $2}')"
    if [[ -z "${conda_env_name}" ]]; then
      conda_env_name="custom"
    fi
    create_and_activate_environment "${conda_install_path}" "${conda_bin_dir}" "${conda_env_name}" "${temp_config_file}"
  else
    retry_constant_custom 4 1 install_conda_packages "${conda_bin_dir}" "${conda_packages}"
    retry_constant_custom 4 1 install_pip_packages "${pip_packages}"
  fi
}
