#!/bin/bash

set -euo pipefail

source "${BDUTIL_DIR}/components/shared/spark.sh"

export CONTAINERD_CONFIG_FILE="/etc/containerd/config.toml"
export CONTAINERD_LOG_DIR="/var/dataproc/log"
export CONTAINERD_LOG_FILE="containerd.log"
export PAUSE_IMAGE_FILE="${DATAPROC_DIR}/images/pause-image.tar"

# BYOSA Image Pull request file is used to log the status of the image pull
# request background process that is spawned during cluster creation.
# Executors read this and wait until the request is completed.
export BYOSA_IMAGE_PULL_STATUS_FILE="/var/log/byosa_image_pull_status.log"

function envs_to_json() {
  local -a envs_in=("$@")
  local -a envs_json
  local env_json
  local env
  for env in "${envs_in[@]}"; do
    # Convert KEY=VALUE to {"key": "KEY", "value": "VALUE"}
    # Use jq to escape special chars in KEY and VALUE
    env_json="$(echo -n "${env}" | jq -s -R)"
    envs_json+=("{\"key\":${env_json/=/\",\"value\":\"}}")
  done
  local IFS=','
  echo "[${envs_json[*]}]" | jq
}

function args_to_json() {
  local -a args_in=("$@")
  local -a args_json
  local arg
  for arg in "${args_in[@]}"; do
    # Convert VALUE to "VALUE", escaping JSON special characters
    # and append to the array
    args_json+=("$(echo -n "${arg}" | jq -s -R)")
  done
  # Join JSON arg objects in a JSON array string (delimited by `,`)
  local IFS=','
  echo "[${args_json[*]}]" | jq
}

function mounts_to_json() {
  local -a mounts_in=("$@")
  local -a mounts_json
  local mount_json
  local mount
  for mount in "${mounts_in[@]}"; do
    # Convert HOST_PATH:CONTAINER_PATH to
    # {"host_path": "HOST_PATH", "container_path": "CONTAINER_PATH"}
    mount_json=$(echo "${mount}" |
      perl -pe 's/(\S+.*?):(.*)/{"host_path": "\1", "container_path": "\2"}/')
    mounts_json+=("${mount_json}")
  done
  # Join JSON mount objects in a JSON array string (delimited by `,`)
  local IFS=','
  echo "[${mounts_json[*]}]" | jq
}

function create_containerd_config() {
  local -r working_dir=$1
  local -r container_name=$2
  local -r image=$3
  local -n _envs=$4
  local -n _args=$5
  local -n _mounts=$6
  local -r container_memory_bytes=$7

  local -r container_env_file="${working_dir}/dataproc-container-env.sh"
  if [[ -f "${container_env_file}" ]]; then
    local -a driver_env_variables
    mapfile -t driver_env_variables <"${container_env_file}"
    local env_var
    for env_var in "${driver_env_variables[@]}"; do
      _envs+=("${env_var}")
    done
    # Clean up after reading env vars
    rm "${container_env_file}"
  fi

  if [[ "${container_name}" == executor-* ]]; then
    local auth_secret
    auth_secret="$(<"${SPARK_AUTH_SECRET_FILE}")"
    # The executor JVM process depends on env variable _SPARK_AUTH_SECRET,
    # it doesn't read from the Spark config.
    _envs+=("_SPARK_AUTH_SECRET=${auth_secret}")
  fi

  local envs_json
  envs_json=$(envs_to_json "${_envs[@]}")
  local args_json
  args_json=$(args_to_json "${_args[@]}")
  local mounts_json
  mounts_json=$(mounts_to_json "${_mounts[@]}")

  # TODO: Pass these vars as function args,
  # so do not need to be global (non-local)
  readonly POD_NAME="${container_name}-pod"
  readonly POD_UID="${container_name}"
  readonly POD_CONFIG="/tmp/${container_name}-pod.json"
  # For config options see PodSandboxConfig message in
  # In the following, `network: 2` is the enum value of NODE,
  # i.e. share the node's network with the pod containers
  cat >"${POD_CONFIG}" <<EOF
{
  "metadata": {
    "name": "${POD_NAME}",
    "uid": "${POD_UID}",
    "namespace": "default"
  },
  "linux": {
    "security_context": {
      "namespace_options" : {
        "network": 2
      }
    }
  },
  "log_directory": "${CONTAINERD_LOG_DIR}"
}
EOF

  # For config options see ContainerConfig message in
  readonly CONTAINER_CONFIG="/tmp/${container_name}-container.json"
  cat >"${CONTAINER_CONFIG}" <<EOF
{
  "metadata": {
    "name": "${container_name}"
  },
  "image": {
    "image": "${image}"
  },
  "command": [
    "/entrypoint.sh"
  ],
  "args": ${args_json},
  "working_dir": "${working_dir}",
  "envs": ${envs_json},
  "mounts": ${mounts_json},
  "log_path": "${CONTAINERD_LOG_FILE}",
  "linux": {
    "resources": {
      "memory_limit_in_bytes": ${container_memory_bytes},
      "oom_score_adj": 500
    },
    "security_context": {
      "run_as_user": {
        "value": ${SPARK_UID}
      },
      "run_as_group": {
        "value": ${SPARK_GID}
      }
    }
  }
}
EOF
}

# Waits for one second and kills the process with PID $TAIL_PID
function kill_tail() {
  local -r exit_code="$1"

  if [[ -n "${TAIL_PID:-}" ]]; then
    # Wait till all output is flushed
    sleep 1
    kill "${TAIL_PID}"
  fi

  exit "$exit_code"
}

# Returns whether container is stopped according to the inspect command
function container_stopped() {
  local -r con_id="$1"

  local state
  state=$(crictl inspect \
    --output go-template --template '{{.status.state}}' "${con_id}")
  [[ "${state}" != "CONTAINER_RUNNING" ]]
}

# Returns true if BYOSA_IMAGE_PULL_STATUS_FILE exists which means the image pull
# was started during the cluster creation.
function check_image_pull_start() {
  [[ -r "$BYOSA_IMAGE_PULL_STATUS_FILE" ]]
}

# Waits on the image pull async process to complete by reading the output.
function get_image_pull_output() {
  echo "Reading the ${BYOSA_IMAGE_PULL_STATUS_FILE} file."
  tail -f "${BYOSA_IMAGE_PULL_STATUS_FILE}" | sed '/DONE/ q'
}

# Checks if the image pull was successful by inspecting on the image.
# The inspect should return status details if successful.
function check_image_pull_successful() {
  local -r image="$1"
  local username
  username=$(crictl inspecti \
    --output go-template --template '{{.status.username}}' "${image}")
  [[ "${username}" == "spark" ]]
}
