# HCFS dirs which need to be created during cluster startup.
# Each item is in "<path> <owner> <mode>" format. Owner and mode will be ignored
# when defaultFS is GCS or Kerberos is not enabled.
#
# `pig` is not a Linux user, so add it here for backwards compatibility.
HCFS_DIRS=(
  "/ hdfs:hadoop 1777"
  "/tmp hdfs:hadoop 1777"
  "/tmp/hadoop-yarn/staging yarn:hadoop 1777"
  "/tmp/hadoop-yarn/staging/history yarn:hadoop 755"
  "/user hdfs:hadoop 755"
  "/var hdfs:hadoop 775"
  "/var/tmp hdfs:hadoop 1777"
)

KERBEROS_ENABLED=$(is_component_selected "kerberos" \
 && echo "true" || echo "false")
WEBHDFS_ENABLED=$(get_property_in_xml "/etc/hadoop/conf/hdfs-site.xml" \
    dfs.webhdfs.enabled "true")
WEBHDFS_BASE_URI=""

function wait_for_hdfs() {
  local -r hostname=$1
  loginfo 'Waiting for NameNode to listen on RPC port'
  local namenode_port_binding_timeout
  namenode_port_binding_timeout=$(get_dataproc_property_or_default \
    startup.component.service-binding-timeout.hadoop-hdfs-namenode 300)
  wait_for_port "hadoop-hdfs-namenode" "${hostname}" 8020 \
    "${namenode_port_binding_timeout}"
}

# User `hdfs` is the admin of HDFS, need to run HDFS commands as `hdfs`.
function login_as_hdfs_if_kerberos_enabled() {
  if is_component_selected "kerberos"; then
    sudo -u hdfs \
      kinit -kt "${KEYTAB_DIR}/hdfs.service.keytab" "hdfs/${MY_FULL_HOSTNAME}"
  fi
}

function encode_path_for_cli() {
  local path="$1"
  path=$(printf '%q' "${path}")
  # Revert any double escaped backslashes (\\\\) back to escaped backslash
  # (\\).
  if [[ "${path}" == *\\\\\\\\* ]]; then
    path=$(echo "${path}" | sed 's/\\\\\\\\/\\\\/g')
  fi
  echo "${path}"
}

function encode_path_for_webhdfs() {
  local encoded
  encoded=$(perl -MURI::Escape -s -e 'print uri_escape($query);' -- -query="$1")
  # Revert %2F back to '/'
  echo "${encoded//%2F//}"
}

# Runs `hadoop fs <args>` as user `hdfs`.
function run_hdfs_command_as_hdfs() {
  sudo -u hdfs hadoop fs "$@"
}

# Initializes the WebHDFS URI. In HA mode, WebHDFS could be running on any of
# the master node, WebHDFS calls must be routed to the active.
function init_webhdfs_uri() {
  local protocol
  local port
  if [[ "${KERBEROS_ENABLED}" == "true" ]]; then
    protocol="https"
    port=$(get_property_in_xml "/etc/hadoop/conf/hdfs-site.xml" \
      dfs.namenode.https-address "0.0.0.0:9871" | cut -d':' -f2)
  else
    protocol="http"
    port=$(get_property_in_xml "/etc/hadoop/conf/hdfs-site.xml" \
      dfs.namenode.http-address "0.0.0.0:9870" | cut -d':' -f2)
  fi

  local active_host
  if [[ "${MASTER_COUNT}" -gt "1" ]]; then
    active_host=$(hdfs haadmin -getAllServiceState \
      | awk '{if ($2 == "active") print $1;}' \
      | cut -d: -f1)
    # Remove the trailing "."
    active_host=${active_host%.}
  else
    active_host=${HOSTNAME}
  fi

  if [[ -n "${active_host}" ]]; then
    WEBHDFS_BASE_URI="${protocol}://${active_host}:${port}/webhdfs/v1"
  fi
}

function webhdfs() {
  local -r uri="$1"

  local -a command
  command=(
    curl -X PUT \
    -i -fsS --retry-connrefused --max-time 60 --retry 3 --retry-max-time 30)
  if [[ "${KERBEROS_ENABLED}" == "true" ]]; then
    command+=(--negotiate -u: -k)
  fi
  command+=("${uri}")

  sudo -u hdfs "${command[@]}"
}

function webhdfs_mkdir() {
  local -r dir="$1"
  local -r owner="$2"

  if [[ -n "${owner}" ]]; then
    webhdfs "${WEBHDFS_BASE_URI}${dir}?op=MKDIRS&user.name=${owner}"
  else
    webhdfs "${WEBHDFS_BASE_URI}${dir}?op=MKDIRS"
  fi
}

function webhdfs_chown() {
  local -r dir="$1"
  local -r owner_group="$2"

  local owner
  local group
  owner=$(echo "${owner_group}" | cut -d: -f1)
  group=$(echo "${owner_group}" | cut -d: -f2)

  webhdfs \
    "${WEBHDFS_BASE_URI}${dir}?op=SETOWNER&owner=${owner}&group=${group}"
}

function webhdfs_chmod() {
  local -r dir="$1"
  local -r mode="$2"

  webhdfs "${WEBHDFS_BASE_URI}${dir}?op=SETPERMISSION&permission=${mode}"
}

# Add HDFS home dirs for all Linux users.
function add_user_dirs() {
  # System users with ID <1000
  local -a system_users=(
    hdfs mapred yarn spark pig hive hbase zookeeper solr zeppelin)

  # Real users with ID >1000
  local -a real_users
  real_users=($(getent passwd | awk -F: '1000 < $3 && $3 < 6000 { print $1}'))

  local -a users=("${system_users[@]}" "${real_users[@]}")

  local user
  for user in "${users[@]}"; do
    HCFS_DIRS+=("/user/${user} ${user}:${user} 700")
  done
}

function wait_all() {
  local -a pids=("$@")

  for pid in "${pids[@]}"; do
    wait "${pid}"
  done
}

function create_hcfs_dirs() {
  local use_webhdfs="$1"
  local set_permission="$2"

  # Make sure WebHDFS is available, especially in HA mode.
  if [[ "${use_webhdfs}" == "true" ]]; then
    init_webhdfs_uri
    if [[ -z "${WEBHDFS_BASE_URI}" ]]; then
      echo "WARNING: WebHDFS is not available"
      use_webhdfs="false"
    fi
  fi

  local -a dirs
  local -a owner_groups
  local -a modes

  add_user_dirs

  local row
  local -a columns
  local encoded_path
  for row in "${HCFS_DIRS[@]}"; do
    columns=(${row})
    if [[ "${use_webhdfs}" == "true" ]]; then
      encoded_path=$(encode_path_for_webhdfs "${columns[0]}")
    else
      encoded_path=$(encode_path_for_cli "${columns[0]}")
    fi
    dirs+=("${encoded_path}")
    owner_groups+=("${columns[1]}")
    modes+=("${columns[2]}")
  done

  # WebHDFS is preferred for lower latency. The HDFS CLI takes several seconds
  # to initialize for each run.
  if [[ "${use_webhdfs}" == "true" ]]; then
    echo "Creating HCFS dirs with WebHDFS ..."
    local batch_size=20
    local -a pids=()
    for ((i=0; i<"${#dirs[@]}"; i++)); do
      if [[ "${set_permission}" == "true" ]]; then
        (webhdfs_mkdir "${dirs[i]}" "" \
          && webhdfs_chown "${dirs[i]}" "${owner_groups[i]}" \
          && webhdfs_chmod "${dirs[i]}" "${modes[i]}") &
        pids+=("$!")
      else
        webhdfs_mkdir "${dirs[i]}" "hdfs" &
        pids+=("$!")
      fi
      # Parallel runs in batches.
      if (((i+1)%${batch_size}==0)); then
        wait_all "${pids[@]}"
        pids=()
      fi
    done
    wait_all "${pids[@]}"
  else
    # Create dirs in batch (to reduce latency).
    echo "Creating HCFS dirs ..."
    run_hdfs_command_as_hdfs -mkdir -p "${dirs[@]}"

    if [[ "${set_permission}" == "true" ]]; then
      echo "Setting HCFS dir permissions ..."
      for ((i=0; i<"${#dirs[@]}"; i++)); do
        run_hdfs_command_as_hdfs -chown "${owner_groups[$i]}" "${dirs[$i]}"
        run_hdfs_command_as_hdfs -chmod "${modes[$i]}" "${dirs[$i]}"
      done
    fi
  fi
}

# Creates system and user dirs and set permissions in in-cluster HDFS.
function init_internal_hdfs() {
  wait_for_hdfs "${MY_HOSTNAME}"

  create_hcfs_dirs "${WEBHDFS_ENABLED}" "${KERBEROS_ENABLED}"

  # b/240671362 This logic is not good but kept for backwards compatibility.
  # The assumption is that non-kerberos clusters are single-tenant, and there
  # might be jobs which assume some dirs are writable, we don't want to break
  # them.
  if [[ "${KERBEROS_ENABLED}" == "false" ]]; then
    retry_constant_short run_hdfs_command_as_hdfs -chmod -R 1777 /
  fi
}

# Best effort to create system and user dirs in external HDFS.
function init_external_hdfs() {
  create_hcfs_dirs "false" "false" || true
}

# Creates system and user dirs in GCS.
function init_gcs_fs() {
  retry_constant_custom 30 10 create_hcfs_dirs "false" "false"
}

# Creates HCFS (hdfs:// or gs://) directories and sets permissions for a new
# cluster. This function should be called from the (1st) master node.
#
# Note that HCFS_ROOT_URI could point to an in-cluster HDFS, an external HDFS,
# or GCS.
function init_hcfs_dirs() {
  if is_in_cluster_hdfs "${HCFS_ROOT_URI}"; then
    init_internal_hdfs
  elif [[ "${HCFS_ROOT_URI}" == hdfs://* ]]; then
    init_external_hdfs
  elif [[ "${HCFS_ROOT_URI}" == gs://* ]]; then
    init_gcs_fs
  fi
}

# Checks if a URI is in-cluster HDFS URI.
#
# For HA clusters, in-cluster HDFS root is `hdfs://<cluster-name>`, e.g.,
# `hdfs://cluster1`.
#
# For non-HA clusters, in-cluster HDFS root is
# - `hdfs://<master-hostname>` for non-kerberos clusters, e.g.,
#   `hdfs://cluster1-m`;
# - `hdfs://<master-dns-name>` for kerberos clusters, e.g.,
#   `hdfs://cluster1-m.us-west1-a.c.project1.internal`.
#
# Note that
# - master hostname does NOT always have cluster name as its prefix.
# In some clusters (e.g., personal auth, Dataplex), the VM names include a
# special prefix, see b/265629436.
# - `hdfs:///` is a valid in-cluster URI
function is_in_cluster_hdfs() {
  local uri="$1"

  if [[ "${uri}" != hdfs://* ]]; then
    return 1
  fi

  local host_name
  # Remove prefix `hdfs://`, `hdfs//cluster1-m/foo/bar` -> `cluster1-m/foo/bar`
  host_name="${uri#hdfs://}"
  # Remove sub dirs, `cluster1-m/foo/bar` -> `cluster1-m`
  host_name="${host_name%%/*}"
  # Remove port, `cluster1-m:1234` -> `cluster1-m`
  host_name="${host_name%%:*}"

  [[ "${host_name}" == "${DATAPROC_MASTER}" ]] \
    || [[ "${host_name}" == "${DATAPROC_MASTER_FQDN}" ]] \
    || [[ "${host_name}" == "${CLUSTER_NAME}" ]] \
    || [[ -z "${host_name}" ]]
}

function create_hdfs_dir() {
  local dir="$1"
  local owner="$2"
  local mode="$3"

  run_hdfs_command_as_hdfs -mkdir -p "${dir}"
  run_hdfs_command_as_hdfs -chown "${owner}" "${dir}"
  run_hdfs_command_as_hdfs -chmod "${mode}" "${dir}"
}
