#!/bin/bash

readonly CLOUD_SQL_PROXY_BIN_NAME="cloud_sql_proxy"
readonly CLOUD_SQL_PROXY_DOWNLOAD_DIR="/opt/cloud_sql_proxy"
readonly CLOUD_SQL_PROXY_DOWNLOAD_PATH="${CLOUD_SQL_PROXY_DOWNLOAD_DIR}/${CLOUD_SQL_PROXY_BIN_NAME}"

# Maintain cloud-sql service ports for different components here
readonly RANGER_CLOUD_SQL_PORT="3307"
readonly DRUID_CLOUD_SQL_PORT="3308"

function install_cloud_proxy() {
  if [[ -f "${CLOUD_SQL_PROXY_DOWNLOAD_PATH}" ]]; then
    echo "Cloud SQL proxy already available at '${CLOUD_SQL_PROXY_DOWNLOAD_PATH}'."
    return
  fi
  local -r download_url="https://dl.google.com/cloudsql/cloud_sql_proxy.linux.amd64"
  mkdir -p "${CLOUD_SQL_PROXY_DOWNLOAD_DIR}"
  curl -fsS --retry-connrefused --retry 10 --retry-max-time 30 \
    "${download_url}" -o "${CLOUD_SQL_PROXY_DOWNLOAD_PATH}"
}

# configure_cloud_sql_proxy_for_component druid cloud-sql-proxy-druid google.com:hadoop-cloud-dev:us-central1:animeshvn-sql-test 3308 false
function configure_cloud_sql_proxy_for_component() {
  if [[ $# != 4 ]]; then
    log_and_fail "${component}" "Wrong number of arguments passed to configure cloud-sql=proxy service."
  fi

  local component=$1
  local cloud_sql_instance=$2
  local cloud_sql_port=$3
  local cloud_sql_private_ip=$4
  local cloud_sql_service_name="cloud-sql-proxy-${component}"
  local proxy_dir="/var/run/cloud_sql_proxy_${component}"
  local proxy_bin="/usr/local/bin/${CLOUD_SQL_PROXY_BIN_NAME}"
  local proxy_log_dir="/var/log/cloud-sql-proxy"
  local init_script="/usr/lib/systemd/system/${cloud_sql_service_name}.service"
  if [[ ! -f "${proxy_bin}" ]]; then
    cp "${CLOUD_SQL_PROXY_DOWNLOAD_PATH}" "${proxy_bin}" ||
      log_and_fail "${component}" "Cannot find Cloud SQL Proxy executable."
  fi
  chmod +x "${proxy_bin}"
  mkdir -p "${proxy_dir}"
  mkdir -p "${proxy_log_dir}"
  local proxy_instance_flag
  proxy_instance_flag="-instances=${cloud_sql_instance}=tcp:${cloud_sql_port}"
  if [[ "${cloud_sql_private_ip}" == "true" ]]; then
    proxy_instance_flag+=" --ip_address_types=PRIVATE"
  fi
  cat >"${init_script}" <<EOF
[Unit]
Description=Google Cloud SQL Proxy Service For ${component}
After=local-fs.target network-online.target
After=google.service
Before=shutdown.target
[Service]
Type=simple
ExecStart=/bin/sh -c '${proxy_bin} \
  -dir=${proxy_dir} \
  ${proxy_instance_flag} >> ${proxy_log_dir}/${cloud_sql_service_name}.log 2>&1'
[Install]
WantedBy=multi-user.target
EOF

  chmod a+rw "${init_script}"
  systemctl enable "${cloud_sql_service_name}"
  systemctl start "${cloud_sql_service_name}"
}
