set -euo pipefail

source "$(dirname "$0")/../../bdutil_misc.sh"
source "$(dirname "$0")/../../bdutil_metadata.sh"
source "$(dirname "$0")/../../bdutil_versions.sh"
source "$(dirname "$0")/../shared/proxy.sh"
source "$(dirname "$0")/../shared/mysql.sh"

set -x

readonly NODE_NAME="$(hostname -s)"
readonly RANGER_ADMIN_PORT="6080"
readonly RANGER_INSTALL_DIR='/usr/lib/ranger'
readonly SOLR_HOME='/usr/lib/solr'
readonly MASTER_ADDITIONAL="$(get_metadata_master_additional)"
readonly INSTANCE_NAME_PREFIX="$(get_instance_name_prefix)"
readonly FQDN="$(hostname -f)"
readonly DOMAIN="$(dnsdomainname)"
readonly MASTER="$(get_metadata_master)"
readonly MASTER_FQDN="${MASTER}.${DOMAIN}"
readonly RANGER_ADMIN_INSTALL_PROPERTIES="${RANGER_INSTALL_DIR}/ranger-admin/install.properties"

if is_component_selected "kerberos"; then
  readonly REALM=$(grep "default_realm = " /etc/krb5.conf | awk '{print $NF}')
fi

KMS_KEY_URI=$(get_dataproc_property ranger.kms.key.uri)
if [[ -z "${KMS_KEY_URI}" ]]; then
  log_and_fail "ranger" "Ranger component is enabled but failed to read KMS key URI from Dataproc properties. \
Set it through 'dataproc:ranger.kms.key.uri'." "${EXIT_CODE_CLIENT_ERROR}"
fi

ADMIN_PASSWORD_URI=$(get_dataproc_property ranger.admin.password.uri)
if [[ -z "${ADMIN_PASSWORD_URI}" ]]; then
  log_and_fail "ranger" "Ranger component is enabled but failed to read admin password URI \
from Dataproc properties. Set it through 'dataproc:ranger.admin.password.uri'." "${EXIT_CODE_CLIENT_ERROR}"
fi

set +x
ADMIN_PASSWORD=$(decrypt_file_with_kms_key "${ADMIN_PASSWORD_URI}" "${KMS_KEY_URI}")
if [[ -z "${ADMIN_PASSWORD}" ]]; then
  log_and_fail "ranger" "Admin password cannot be empty!" "${EXIT_CODE_CLIENT_ERROR}"
fi

DB_ADMIN_PASSWORD_URI=$(get_dataproc_property ranger.db.admin.password.uri)
if [[ -z "${DB_ADMIN_PASSWORD_URI}" ]]; then
  DB_ADMIN_PASSWORD="rangerpass"
  if is_version_at_least "${DATAPROC_IMAGE_VERSION}" "2.2"; then
    # ranger 2.3.0 onwards password should contain at least one
    # small letter, capital letter and number.
    DB_ADMIN_PASSWORD="rangerPass1234"
  fi
else
  DB_ADMIN_PASSWORD=$(decrypt_file_with_kms_key "${DB_ADMIN_PASSWORD_URI}" "${KMS_KEY_URI}")
  if [[ -z "${DB_ADMIN_PASSWORD}" ]]; then
    log_and_fail "ranger" "Database ranger admin password cannot be empty!" "${EXIT_CODE_CLIENT_ERROR}"
  fi
fi
# read from common function from shared/mysql.sh
DB_ROOT_PASSWORD="$(read_mysql_password)"
set -x

CLOUD_SQL_INSTANCE=$(get_dataproc_property ranger.cloud-sql.instance.connection.name)
MYSQL_PORT=3306
MYSQL_HOST="localhost"
MYSQL_USER_HOST="localhost"
if [[ -n "${CLOUD_SQL_INSTANCE}" ]]; then
  MYSQL_PORT="${RANGER_CLOUD_SQL_PORT}"
  MYSQL_HOST="127.0.0.1"
  MYSQL_USER_HOST="%"
fi

# Get a property value to indicate whether the same ranger database is to be used
# when the value is false then ranger database will be created.
function get_ranger_reuse_schema_value() {
  RANGER_REUSE_SCHEMA_PROP=$(get_dataproc_property ranger.reuse.schema)
  if [[ -z "${RANGER_REUSE_SCHEMA_PROP}" ]]; then
    RANGER_REUSE_SCHEMA_PROP="false"
  fi
  RANGER_REUSE_SCHEMA=$(echo ${RANGER_REUSE_SCHEMA_PROP} | tr '[:upper:]' '[:lower:]')
  echo "${RANGER_REUSE_SCHEMA}"
}

function check_for_cloud_sql_instance() {
  if [[ -n "${CLOUD_SQL_INSTANCE}" ]]; then
    DB_ROOT_PASSWORD_URI=$(get_dataproc_property ranger.cloud-sql.root.password.uri)
    if [[ -z "${DB_ROOT_PASSWORD_URI}" ]]; then
      if [[ "${RANGER_REUSE_SCHEMA}" == 'false' ]]; then
        log_and_fail "ranger" "When using Cloud SQL instance as the ranger database, you must specify the property 'dataproc:ranger.cloud-sql.root.password.uri'."
      fi
    else
      echo "Decrypting DB_ROOT_PASSWORD ${DB_ROOT_PASSWORD_URI} with ${KMS_KEY_URI}"
      set +x
      DB_ROOT_PASSWORD=$(decrypt_file_with_kms_key "${DB_ROOT_PASSWORD_URI}" "${KMS_KEY_URI}")
      if [[ -z "${DB_ROOT_PASSWORD}" ]]; then
        if [[ "${RANGER_REUSE_SCHEMA}" == 'false' ]]; then
          log_and_fail "ranger" "Cloud SQL root password cannot be empty."
        fi
      fi
    fi
    set -x
    CLOUD_SQL_PRIVATE_IP=$(get_dataproc_property ranger.cloud-sql.use-private-ip)
  fi
}

function set_property_in_xml() {
  bdconfig set_property \
    --configuration_file "$1" \
    --name "$2" --value "$3" \
    --create_if_absent \
    --clobber ||
    log_and_fail "ranger" "Unable to set $2"
}

function set_property_hive_audit() {
  set_property_in_xml "${RANGER_INSTALL_DIR}/hive/conf/ranger-hive-audit.xml" "$1" "$2"
}

function set_property_hdfs_audit() {
  set_property_in_xml "${RANGER_INSTALL_DIR}/hadoop/etc/hadoop/ranger-hdfs-audit.xml" "$1" "$2"
}

function set_property_yarn_audit() {
  set_property_in_xml "${RANGER_INSTALL_DIR}/hadoop/etc/hadoop/ranger-yarn-audit.xml" "$1" "$2"
}

function set_property_ranger_admin() {
  set_property_in_xml "${RANGER_INSTALL_DIR}/ranger-admin/ews/webapp/WEB-INF/classes/conf.dist/ranger-admin-site.xml" "$1" "$2"
}

function remove_property_ranger_admin() {
  bdconfig remove_property --configuration_file "${RANGER_INSTALL_DIR}/ranger-admin/ews/webapp/WEB-INF/classes/conf.dist/ranger-admin-site.xml" \
    --name "$1"
}

function run_ranger_admin() {
  if is_component_selected "kerberos"; then
    local curl_command=('curl' '-k' '--negotiate' '-u' ':')
    local -r login="export SOLR_INCLUDE=/etc/default/solr; "
  else
    local curl_command=('curl')
    local -r login=""
  fi

  # Add common retry flags to curl command
  curl_command+=('-fsS' '--retry-connrefused' '--retry' '300' '--retry-delay' '1')

  if is_version_at_least "${DATAPROC_IMAGE_VERSION}" "2.1"; then
    # LRUCache,FastLRUCache & LFUCache are removed in solr 9.0
    # https://solr.apache.org/guide/solr/latest/upgrade-notes/major-changes-in-solr-9.html#deprecations-and-removals
    sed -i 's/FastLRUCache/CaffeineCache/g' ${RANGER_INSTALL_DIR}/ranger-admin/contrib/solr_for_audit_setup/conf/solrconfig.xml
    sed -i 's/LRUCache/CaffeineCache/g' ${RANGER_INSTALL_DIR}/ranger-admin/contrib/solr_for_audit_setup/conf/solrconfig.xml
    sed -i 's/LFUCache/CaffeineCache/g' ${RANGER_INSTALL_DIR}/ranger-admin/contrib/solr_for_audit_setup/conf/solrconfig.xml
  fi

  if [[ "${MASTER_ADDITIONAL}" != "" ]]; then
    host_1="${INSTANCE_NAME_PREFIX}-m-1.${DOMAIN}"
    host_2="${INSTANCE_NAME_PREFIX}-m-2.${DOMAIN}"
    sed -i "s/^audit_solr_zookeepers=/audit_solr_zookeepers=${FQDN}:2181,${host_1}:2181,${host_2}:2181\/solr/" \
      "${RANGER_ADMIN_INSTALL_PROPERTIES}"
    sed -i 's/^audit_solr_urls=/audit_solr_urls=none/' \
      "${RANGER_ADMIN_INSTALL_PROPERTIES}"

    "${curl_command[@]}" "http://${FQDN}:8983/solr"
    runuser -l solr -s /bin/bash -c "${login}${SOLR_HOME}/bin/solr create_collection -c ranger_audits -d ${RANGER_INSTALL_DIR}/ranger-admin/contrib/solr_for_audit_setup/conf -shards 1 -replicationFactor 3"
  else
    sed -i "s/^audit_solr_urls=/audit_solr_urls=http:\/\/${MASTER_FQDN}:8983\/solr\/ranger_audits/" \
      "${RANGER_ADMIN_INSTALL_PROPERTIES}"
    "${curl_command[@]}" "http://${MASTER_FQDN}:8983/solr"
    runuser -l solr -s /bin/bash -c "${login}${SOLR_HOME}/bin/solr create_core -c ranger_audits -d ${RANGER_INSTALL_DIR}/ranger-admin/contrib/solr_for_audit_setup/conf -shards 1 -replicationFactor 1"
  fi

  echo "Running ./set_globals.sh and ./setup.sh"
  pushd "${RANGER_INSTALL_DIR}/ranger-admin" && ./set_globals.sh && ./setup.sh

  if is_component_selected "kerberos"; then
    sed -i '/^JAVA_OPTS=.*/a JAVA_OPTS=" ${JAVA_OPTS} -Djavax.net.ssl.trustStore=/etc/hadoop/conf/ssl/truststore.jks"' "${RANGER_INSTALL_DIR}"/ranger-admin/ews/ranger-admin-services.sh
  fi
  cat >/etc/systemd/system/ranger-admin.service <<EOF
[Unit]
Description=Apache Ranger Admin
ConditionPathExists=${RANGER_INSTALL_DIR}/ranger-admin
After=syslog.target network.target remote-fs.target nss-lookup.target systemd-journald-dev-log.socket solr.service
Before=multi-user.target
Conflicts=shutdown.target

[Service]
WorkingDirectory=${RANGER_INSTALL_DIR}/ranger-admin
User=root
Type=forking

Restart=on-failure
RestartSec=5

ExecStart=/usr/bin/ranger-admin start -f
ExecStop=/usr/bin/ranger-admin stop

[Install]
WantedBy=multi-user.target
EOF
  retry_constant systemctl daemon-reload
  retry_constant systemctl enable ranger-admin
  retry_constant systemctl start ranger-admin
  popd
}

function add_usersync_plugin() {
  mkdir -p /var/log/ranger-usersync && chown ranger /var/log/ranger-usersync &&
    chgrp ranger /var/log/ranger-usersync

  local -r install_properties="${RANGER_INSTALL_DIR}/ranger-usersync/install.properties"
  if is_component_selected "kerberos"; then
    sed -i "s/^usersync_principal=/usersync_principal=rangerusersync\/${FQDN}@${REALM}/" \
      "${install_properties}"
    sed -i "s/^usersync_keytab=/usersync_keytab=\/etc\/security\/keytab\/rangerusersync.service.keytab/" \
      "${install_properties}"
  fi

  sed -i 's/^logdir=logs/logdir=\/var\/log\/ranger-usersync/' \
    "${install_properties}"
  sed -i "s/^POLICY_MGR_URL =/POLICY_MGR_URL = http:\/\/${FQDN}:${RANGER_ADMIN_PORT}/" \
    "${install_properties}"

  sed -i 's/^MIN_UNIX_USER_ID_TO_SYNC = .*/MIN_UNIX_USER_ID_TO_SYNC = 0/' \
    "${install_properties}"
  sed -i 's/^MIN_UNIX_GROUP_ID_TO_SYNC = .*/MIN_UNIX_GROUP_ID_TO_SYNC = 0/' \
    "${install_properties}"

  # shorten sync_interval to 1 minute
  sed -i 's/^SYNC_INTERVAL =/SYNC_INTERVAL = 1/' "${install_properties}"

  set +x
  sed -i "s/^rangerUsersync_password=/rangerUsersync_password=${ADMIN_PASSWORD}/" \
    "${install_properties}"
  set -x
  pushd "${RANGER_INSTALL_DIR}/ranger-usersync" && ./setup.sh
  cat >/etc/systemd/system/ranger-usersync.service <<EOF
[Unit]
Description=Apache Ranger UserSync
ConditionPathExists=${RANGER_INSTALL_DIR}/ranger-usersync
After=syslog.target network.target remote-fs.target nss-lookup.target systemd-journald-dev-log.socket solr.service ranger-admin.service
Before=multi-user.target
Conflicts=shutdown.target

[Service]
WorkingDirectory=${RANGER_INSTALL_DIR}/ranger-usersync
User=root
Type=forking

Restart=on-failure
RestartSec=5

ExecStart=/usr/bin/ranger-usersync start -f
ExecStop=/usr/bin/ranger-usersync stop

[Install]
WantedBy=multi-user.target
EOF

  # We need to create a symlink for ranger-usersync to work in HA cluster because of this bug
  # https://www.mail-archive.com/dev@ranger.apache.org/msg28660.html
  # This jar is required because in the core-site.xml the property
  # "hadoop.security.token.service.use_ip" is set to false for HA cluster forcing a dns lookup
  ln -s /usr/lib/hadoop/lib/dnsjava-2.1.7.jar /usr/lib/ranger/ranger-usersync/lib/dnsjava-2.1.7.jar

  retry_constant systemctl daemon-reload
  retry_constant systemctl start ranger-usersync
  popd
}

function failover_back_to_namenode() {
  # due to multiple restarts of hadoop-hdfs-namenode, the namenode on nn1 
  # may become active while waiting for nn0 to install the ranger setup.sh
  if [[ "${MASTER_ADDITIONAL}" != "" && "${NODE_NAME}" =~ ^.*(-m-0|-m-1)$ ]]; then
    if [[ "$(hdfs haadmin -getServiceState nn1)" == "active" ]]; then
      /usr/bin/su -s /bin/bash hdfs -c 'hdfs haadmin -failover nn1 nn0 2>/dev/null ||true'
    fi
  fi
}

function apply_common_plugin_configuration() {
  local plugin_name="${1}"
  local service_name="${2}"
  sed -i "s/^POLICY_MGR_URL=/POLICY_MGR_URL=http:\/\/${MASTER_FQDN}:${RANGER_ADMIN_PORT}/" \
    "${RANGER_INSTALL_DIR}/${plugin_name}/install.properties"
  sed -i "s/^REPOSITORY_NAME=/REPOSITORY_NAME=${service_name}/" \
    "${RANGER_INSTALL_DIR}/${plugin_name}/install.properties"
  sed -i 's/^XAAUDIT.SOLR.ENABLE=false/XAAUDIT.SOLR.ENABLE=true/' \
    "${RANGER_INSTALL_DIR}/${plugin_name}/install.properties"
  sed -i "s/^XAAUDIT.SOLR.URL=NONE/XAAUDIT.SOLR.URL=http:\/\/${MASTER_FQDN}:8983\/solr\/ranger_audits/" \
    "${RANGER_INSTALL_DIR}/${plugin_name}/install.properties"
  sed -i 's/^XAAUDIT.SOLR.USER=NONE/XAAUDIT.SOLR.USER=solr/' \
    "${RANGER_INSTALL_DIR}/${plugin_name}/install.properties"
  if [[ "${MASTER_ADDITIONAL}" != "" ]]; then
    sed -i "s/^XAAUDIT.SOLR.ZOOKEEPER=NONE/XAAUDIT.SOLR.ZOOKEEPER=${INSTANCE_NAME_PREFIX}-m-0:2181,${INSTANCE_NAME_PREFIX}-m-1:2181,${INSTANCE_NAME_PREFIX}-m-2:2181\/solr/" \
      "${RANGER_INSTALL_DIR}/${plugin_name}/install.properties"
  fi
}

function add_hdfs_plugin() {
  apply_common_plugin_configuration "ranger-hdfs-plugin" "hadoop-dataproc"

  if is_version_at_least "${DATAPROC_IMAGE_VERSION}" "2.1"; then
    sed -i "s#^COMPONENT_INSTALL_DIR_NAME=.*#COMPONENT_INSTALL_DIR_NAME=hadoop#" \
      "${RANGER_INSTALL_DIR}/ranger-hdfs-plugin/install.properties"
  fi

  pushd ranger-hdfs-plugin && ./enable-hdfs-plugin.sh && popd

  if is_component_selected "kerberos"; then
    set_property_hdfs_audit "xasecure.audit.solr.is.enabled" "true"
    set_property_hdfs_audit "xasecure.audit.destination.solr.force.use.inmemory.jaas.config" "true"
    set_property_hdfs_audit "xasecure.audit.jaas.Client.loginModuleControlFlag" "required"
    set_property_hdfs_audit "xasecure.audit.jaas.Client.loginModuleName" "com.sun.security.auth.module.Krb5LoginModule"
    set_property_hdfs_audit "xasecure.audit.jaas.Client.option.keyTab" "/etc/security/keytab/hdfs.service.keytab"
    set_property_hdfs_audit "xasecure.audit.jaas.Client.option.principal" "hdfs/_HOST@${REALM}"
    set_property_hdfs_audit "xasecure.audit.jaas.Client.option.serviceName" "solr"
    set_property_hdfs_audit "xasecure.audit.jaas.Client.option.storeKey" "false"
    set_property_hdfs_audit "xasecure.audit.jaas.Client.option.useKeyTab" "true"
  fi

  if [[ "${NODE_NAME}" =~ ^.*(-m|-m-0)$ ]]; then
    systemctl daemon-reload
    systemctl stop hadoop-hdfs-namenode.service
    retry_constant_short systemctl start hadoop-hdfs-namenode.service

    # Notify cluster that plugin is installed on master.
    until hadoop fs -touchz hdfs://"${MASTER}"/tmp/ranger-hdfs-plugin-ready &>/dev/null; do
      sleep 10
      failover_back_to_namenode
    done

    set +x
    if is_component_selected "kerberos"; then
      cat >service-hdfs.json <<EOF
{
    "configs": {
        "username": "admin",
        "password": "${ADMIN_PASSWORD}",
        "hadoop.security.authentication": "kerberos",
        "hadoop.security.authorization": "Yes",
        "fs.default.name": "hdfs://${NODE_NAME}:8020",
        "dfs.datanode.kerberos.principal": "hdfs/_HOST@${REALM}",
        "dfs.namenode.kerberos.principal": "hdfs/_HOST@${REALM}",
        "dfs.secondary.namenode.kerberos.principal": "hdfs/_HOST@${REALM}",
        "hadoop.rpc.protection": "privacy",
        "hadoop.security.auth_to_local": "RULE:[1:\$1](.*)s/(.*)/\$1/g     RULE:[2:\$1](.*)s/(.*)/\$1/g",
        "policy.download.auth.users": "hdfs"
    },
    "description": "Hadoop hdfs service",
    "isEnabled": true,
    "name": "hadoop-dataproc",
    "type": "hdfs",
    "version": 1
}
EOF
    else
      cat >service-hdfs.json <<EOF
{
    "configs": {
        "username": "admin",
        "password": "${ADMIN_PASSWORD}",
        "hadoop.security.authentication": "Simple",
        "hadoop.security.authorization": "No",
        "fs.default.name": "hdfs://${NODE_NAME}:8020"
    },
    "description": "Hadoop hdfs service",
    "isEnabled": true,
    "name": "hadoop-dataproc",
    "type": "hdfs",
    "version": 1
}
EOF
    fi

    curl --retry 5 --user "admin:${ADMIN_PASSWORD}" -H "Content-Type: application/json" \
      -X POST -d @service-hdfs.json "http://${FQDN}:${RANGER_ADMIN_PORT}/service/public/v2/api/service"
    set -x
    rm service-hdfs.json
  elif [[ "${NODE_NAME}" =~ ^.*(-m-1)$ ]]; then
    # Waiting until hdfs plugin will be configured on m-0
    until hadoop fs -ls hdfs://"${MASTER}"/tmp/ranger-hdfs-plugin-ready &>/dev/null; do
      sleep 10
      failover_back_to_namenode
    done
    retry_constant_short systemctl enable hadoop-hdfs-namenode.service
    systemctl daemon-reload
    systemctl stop hadoop-hdfs-namenode.service
    retry_constant_short systemctl start hadoop-hdfs-namenode.service
    # re-check after the restart of hadoop namenode service.
    failover_back_to_namenode
  fi
}

function grant_hive_user_access() {
  if is_component_selected "kerberos"; then
    additional_user='dataproc'
  else
    additional_user='hive'
  fi
  # Force a user sync
  systemctl stop ranger-usersync
  retry_constant_short systemctl start ranger-usersync
  local python_binary='python2'
  if is_version_at_least "${DATAPROC_IMAGE_VERSION}" "2.1"; then
    python_binary="/opt/conda/miniconda3/bin/python"
  fi

  set +x
  "$python_binary" - <<END
import json
import requests
import logging
from time import sleep
logging.basicConfig(level=logging.DEBUG)
from requests.auth import HTTPBasicAuth
res=requests.get('http://${FQDN}:${RANGER_ADMIN_PORT}/service/public/v2/api/policy/4',
  auth=HTTPBasicAuth('admin', '${ADMIN_PASSWORD}'))
policy=res.json()
for item in policy['policyItems']:
  if 'users' in item:
    item['users'].append('${additional_user}')
for i in range(0,300):
  r = requests.put('http://${FQDN}:${RANGER_ADMIN_PORT}/service/public/v2/api/policy/4',
    auth=HTTPBasicAuth('admin', '${ADMIN_PASSWORD}'),
    headers={'content-type':'application/json'}, json=policy)
  if r.ok:
    break
  else:
    sleep(1)
END
  set -x
}

function add_hive_plugin() {
  apply_common_plugin_configuration "ranger-hive-plugin" "hive-dataproc"

  if is_version_at_least "${DATAPROC_IMAGE_VERSION}" "2.1"; then
  sed -i "s#^COMPONENT_INSTALL_DIR_NAME=.*#COMPONENT_INSTALL_DIR_NAME=hive#" \
    "${RANGER_INSTALL_DIR}/ranger-hive-plugin/install.properties"
  fi


  set_property_in_xml /etc/hive/conf/hive-site.xml "hive.security.authorization.sqlstd.confwhitelist.append" "tez.application.tags"

  mkdir -p hive &&
    ln -s /etc/hive/conf hive &&
    ln -s /usr/lib/hive/lib hive

  pushd ranger-hive-plugin && ./enable-hive-plugin.sh && popd

  if is_component_selected "kerberos"; then
    set_property_hive_audit "xasecure.audit.solr.is.enabled" "true"
    set_property_hive_audit "xasecure.audit.destination.solr.force.use.inmemory.jaas.config" "true"
    set_property_hive_audit "xasecure.audit.jaas.Client.loginModuleControlFlag" "required"
    set_property_hive_audit "xasecure.audit.jaas.Client.loginModuleName" "com.sun.security.auth.module.Krb5LoginModule"
    set_property_hive_audit "xasecure.audit.jaas.Client.option.keyTab" "/etc/security/keytab/hive.service.keytab"
    set_property_hive_audit "xasecure.audit.jaas.Client.option.principal" "hive/_HOST@${REALM}"
    set_property_hive_audit "xasecure.audit.jaas.Client.option.serviceName" "solr"
    set_property_hive_audit "xasecure.audit.jaas.Client.option.storeKey" "false"
    set_property_hive_audit "xasecure.audit.jaas.Client.option.useKeyTab" "true"
  fi

  if [[ "${NODE_NAME}" =~ ^.*(-m|-m-0)$ ]]; then
    systemctl daemon-reload
    systemctl stop hive-server2.service
    retry_constant_short systemctl start hive-server2.service

    # Notify cluster that hive plugin is installed on master.
    until hadoop fs -touchz hdfs://"${MASTER}"/tmp/ranger-hive-plugin-ready &>/dev/null; do
      sleep 10
      failover_back_to_namenode
    done

    set +x
    if is_component_selected "kerberos"; then
      cat >service-hive.json <<EOF
{
    "configs": {
        "username": "admin",
        "password": "${ADMIN_PASSWORD}",
        "jdbc.driverClassName" : "org.apache.hive.jdbc.HiveDriver",
        "jdbc.url" : "jdbc:hive2://${FQDN}:10000/default;principal=hive/${FQDN}@${REALM}",
        "policy.download.auth.users" : "hive"
    },
    "description": "Hadoop HIVE service",
    "isEnabled": true,
    "name": "hive-dataproc",
    "type": "hive",
    "version": 1
}
EOF
    else
      cat >service-hive.json <<EOF
{
    "configs": {
        "username": "admin",
        "password": "${ADMIN_PASSWORD}",
        "jdbc.driverClassName" : "org.apache.hive.jdbc.HiveDriver",
        "jdbc.url" : "jdbc:mysql://localhost"
    },
    "description": "Hadoop HIVE service",
    "isEnabled": true,
    "name": "hive-dataproc",
    "type": "hive",
    "version": 1
}
EOF
    fi

    curl --retry 5 --user "admin:${ADMIN_PASSWORD}" -H "Content-Type: application/json" \
      -X POST -d @service-hive.json "http://${FQDN}:${RANGER_ADMIN_PORT}/service/public/v2/api/service"
    grant_hive_user_access
    set -x
    rm service-hive.json
  elif [[ "${NODE_NAME}" =~ ^.*(-m-1|-m-2)$ ]]; then
    # Waiting until hive plugin will be configured on m-0
    until hadoop fs -ls hdfs://"${MASTER}"/tmp/ranger-hive-plugin-ready &>/dev/null; do
      sleep 10
      failover_back_to_namenode
    done
    systemctl daemon-reload
    systemctl stop hive-server2.service
    retry_constant_short systemctl start hive-server2.service
  fi
}

function add_yarn_plugin() {
  apply_common_plugin_configuration "ranger-yarn-plugin" "yarn-dataproc"
  pushd ranger-yarn-plugin && ./enable-yarn-plugin.sh && popd

  if is_component_selected "kerberos"; then
    set_property_yarn_audit "xasecure.audit.solr.is.enabled" "true"
    set_property_yarn_audit "xasecure.audit.destination.solr.force.use.inmemory.jaas.config" "true"
    set_property_yarn_audit "xasecure.audit.jaas.Client.loginModuleControlFlag" "required"
    set_property_yarn_audit "xasecure.audit.jaas.Client.loginModuleName" "com.sun.security.auth.module.Krb5LoginModule"
    set_property_yarn_audit "xasecure.audit.jaas.Client.option.keyTab" "/etc/security/keytab/yarn.service.keytab"
    set_property_yarn_audit "xasecure.audit.jaas.Client.option.principal" "yarn/_HOST@${REALM}"
    set_property_yarn_audit "xasecure.audit.jaas.Client.option.serviceName" "solr"
    set_property_yarn_audit "xasecure.audit.jaas.Client.option.storeKey" "false"
    set_property_yarn_audit "xasecure.audit.jaas.Client.option.useKeyTab" "true"

    yarn_url="https://${FQDN}:8090"
    authentication_type="kerberos"
  else
    yarn_url="http://${NODE_NAME}:8088"
    authentication_type="simple"
  fi

  if [[ "${NODE_NAME}" =~ ^.*(-m|-m-0)$ ]]; then
    systemctl daemon-reload
    systemctl stop hadoop-yarn-resourcemanager.service
    retry_constant_short systemctl start hadoop-yarn-resourcemanager.service

    # Notify cluster that yarn plugin is installed on master.
    until hadoop fs -touchz hdfs://"${MASTER}"/tmp/ranger-yarn-plugin-ready &>/dev/null; do
      sleep 10
      failover_back_to_namenode
    done

    set +x
    cat >service-yarn.json <<EOF
{
    "configs": {
        "username": "admin",
        "password": "${ADMIN_PASSWORD}",
        "yarn.url": "${yarn_url}",
        "policy.download.auth.users": "yarn",
        "hadoop.security.authentication": "${authentication_type}"
    },
    "description": "Hadoop YARN service",
    "isEnabled": true,
    "name": "yarn-dataproc",
    "type": "yarn",
    "version": 1
}
EOF
    curl --retry 5 --user "admin:${ADMIN_PASSWORD}" -H "Content-Type: application/json" \
      -X POST -d @service-yarn.json "http://${MASTER_FQDN}:${RANGER_ADMIN_PORT}/service/public/v2/api/service"
    set -x
    rm service-yarn.json
  elif [[ "${NODE_NAME}" =~ ^.*(-m-1|-m-2)$ ]]; then
    # Waiting until yarn plugin will be configured on m-0
    until hadoop fs -ls hdfs://"${MASTER}"/tmp/ranger-yarn-plugin-ready &>/dev/null; do
      sleep 10
      failover_back_to_namenode
    done
    systemctl daemon-reload
    systemctl stop hadoop-yarn-resourcemanager.service
    retry_constant_short systemctl start hadoop-yarn-resourcemanager.service
  fi
}

function add_presto_plugin() {
  apply_common_plugin_configuration "ranger-presto-plugin" "presto-dataproc"

  mkdir -p presto &&
    ln -s /etc/presto/conf presto &&
    ln -s /usr/lib/presto/lib presto

  pushd ranger-presto-plugin && ./enable-presto-plugin.sh && popd
  
  cp /usr/lib/presto/lib/failureaccess-1.0.1.jar /usr/lib/presto/plugin/ranger

  if [[ "${NODE_NAME}" =~ ^.*(-m|-m-0)$ ]]; then
    systemctl daemon-reload
    systemctl stop presto.service
    retry_constant_short systemctl start presto.service

    # Notify cluster that presto plugin is installed on master.
    until hadoop fs -touchz hdfs://"${MASTER}"/tmp/ranger-presto-plugin-ready &>/dev/null; do
      sleep 10
      failover_back_to_namenode
    done

    set +x
    cat >service-presto.json <<EOF
{
    "configs": {
        "username": "admin",
        "password": "${ADMIN_PASSWORD}",
        "jdbc.driverClassName": "io.prestosql.jdbc.PrestoDriver",
        "jdbc.url": "jdbc:presto://localhost"
    },
    "description": "presto service",
    "isEnabled": true,
    "name": "presto-dataproc",
    "type": "presto",
    "version": 1
}
EOF

    curl --retry 5 --user "admin:${ADMIN_PASSWORD}" -H "Content-Type: application/json" \
      -X POST -d @service-presto.json "http://${FQDN}:${RANGER_ADMIN_PORT}/service/public/v2/api/service"
    set -x
    rm service-presto.json
  elif [[ "${NODE_NAME}" =~ ^.*(-m-1|-m-2)$ ]]; then
    # Waiting until presto plugin will be configured on m-0
    until hadoop fs -ls hdfs://"${MASTER}"/tmp/ranger-presto-plugin-ready &>/dev/null; do
      sleep 10
      failover_back_to_namenode
    done
    systemctl daemon-reload
    systemctl stop presto.service
    retry_constant_short systemctl start presto.service
  fi
}
