set -euo pipefail

# The script provides functions to install the Ranger GCS plugin.
# Some variables are defined in shared/ranger.sh. Please source it before
# using functions in this script.

source '/etc/environment'
source "${BDUTIL_DIR}/bdutil_os.sh"
source "${BDUTIL_DIR}/components/shared/kerberos.sh"
source "${BDUTIL_DIR}/components/shared/mysql.sh"
source "${BDUTIL_DIR}/bdutil_services.sh"

set -x

# The component name for logging.
readonly RANGER_GCS_PLUGIN_COMPONENT_NAME='Ranger GCS Plugin'
readonly RANGER_GCS_PLUGIN_PLUGIN_SRC='/usr/lib/dataproc-ranger-gcs-plugin'
# The dir to put all plugin config files.
readonly RANGER_GCS_PLUGIN_CONF_DIR='/etc/dataproc-ranger-gcs-plugin/conf'
readonly RANGER_GCS_PLUGIN_GCS_KEYTAB='/etc/security/keytab/gcs.service.keytab'
# The dir to put all the plugin lib
readonly RANGER_GCS_PLUGIN_BASE_DIR='/usr/lib/dataproc-ranger-gcs-plugin'
readonly RANGER_GCS_PLUGIN_LIB_DIR="${RANGER_GCS_PLUGIN_BASE_DIR}/lib"

function get_ranger_admin_password() {
  local admin_password
  if [[ -f "${RANGER_ADMIN_INSTALL_PROPERTIES}" ]] ; then
    admin_password=$(grep 'rangerAdmin_password' "${RANGER_ADMIN_INSTALL_PROPERTIES}" | cut -b 22-);
  else
    log_and_fail "${RANGER_GCS_PLUGIN_COMPONENT_NAME}" "Cannot locate ${RANGER_ADMIN_INSTALL_PROPERTIES}. Failed to retrieve Ranger admin password." "${EXIT_CODE_CLIENT_ERROR}"
  fi
  echo "${admin_password}"
}

function copy_ranger_gcs_plugin_service_jar_to_ranger_admin() {
  echo "Running copy_ranger_gcs_plugin_service_jar_to_ranger_admin"
  local -r jar_destination="${RANGER_INSTALL_DIR}/ranger-admin/ews/webapp/WEB-INF/classes/ranger-plugins/gcs"
  mkdir -p "${jar_destination}"
  cp "${RANGER_GCS_PLUGIN_PLUGIN_SRC}/ranger-gcs-service-"*".jar" "${jar_destination}"
  chown -R ranger:ranger "${jar_destination}"
}

function install_ranger_gcs_plugin_servicedef() {
  echo "Running install_ranger_gcs_plugin_servicedef"
  local admin_password
  set +x
  admin_password="$(get_ranger_admin_password)"
  set -x

  local
  host_name="$(hostname -f)"
  # RANGER_ADMIN_PORT is defined in shared/ranger.sh
  local -r service_def_endpoint="http://${host_name}:${RANGER_ADMIN_PORT}/service/plugins/definitions"
  set +x
  curl --retry 10 --retry-max-time 30 -u "admin:${admin_password}" -X POST \
    -H 'Accept: application/json' -H 'Content-Type: application/json' \
    --data "@${RANGER_GCS_PLUGIN_PLUGIN_SRC}/ranger-servicedef-gcs.json" "${service_def_endpoint}"
  set -x

  # Create a default service
  local -r service_data="$(cat <<-EOD
          {
                  "configs": {},
                  "description": "GCS service",
                  "isEnabled": true,
                  "name": "gcs-dataproc",
                  "type": "GCS",
                  "version": 1
          }
EOD
  )"

  local -r service_endpoint="http://${host_name}:${RANGER_ADMIN_PORT}/service/public/v2/api/service"
  set +x
  curl --retry 10 --retry-max-time 30 -u "admin:${admin_password}" -X POST \
    -H 'Accept: application/json' -H 'Content-Type: application/json' \
    --data "${service_data}" "${service_endpoint}"
  echo ''
  set -x
}

function install_ranger_gcs_plugin_default_policy() {
  echo "Running function install_ranger_gcs_plugin_default_policy"
  local admin_password
  set +x
  admin_password="$(get_ranger_admin_password)"
  set -x

  local
  host_name="$(hostname -f)"

  local dataproc_bucket
  dataproc_bucket="$(/usr/share/google/get_metadata_value attributes/dataproc-bucket)"
  local dataproc_temp_bucket
  dataproc_temp_bucket="$(/usr/share/google/get_metadata_value attributes/dataproc-temp-bucket)"

  echo "Creating policy file gcs-dataproc service for staging bucket"
  local -r hadoop_dataproc_bucket_all_policy="$(cat <<-EOD
      {
        "service": "gcs-dataproc",
        "name": "hadoop group and the dataproc user to access dataproc staging bucket",
        "policyType": 0,
        "policyPriority": 0,
        "description": "hadoop group and the dataproc user to access dataproc staging bucket",
        "isAuditEnabled": true,
        "resources": {
          "bucket": {
            "values": [
              "${dataproc_bucket}"
            ],
            "isExcludes": false,
            "isRecursive": false
          },
          "object-path": {
            "values": [
              "*"
            ],
            "isExcludes": false,
            "isRecursive": true
          }
        },
        "policyItems": [
          {
            "accesses": [
              {
                "type": "LIST_OBJECTS",
                "isAllowed": true
              },
              {
                "type": "READ_OBJECTS",
                "isAllowed": true
              },
              {
                "type": "WRITE_OBJECTS",
                "isAllowed": true
              },
              {
                "type": "MODIFY_OBJECTS",
                "isAllowed": true
              },
              {
                "type": "DELETE_OBJECTS",
                "isAllowed": true
              },
              {
                "type": "GET_BUCKETS",
                "isAllowed": true
              },
              {
                "type": "CREATE_BUCKETS",
                "isAllowed": true
              },
              {
                "type": "DELETE_BUCKETS",
                "isAllowed": true
              },
              {
                "type": "LIST_BUCKETS",
                "isAllowed": true
              },
              {
                "type": "GET_METADATA_OBJECTS",
                "isAllowed": true
              }
            ],
            "users": [
              "dataproc"
            ],
            "groups": [
              "hadoop"
            ],
            "roles": [],
            "conditions": [],
            "delegateAdmin": false
          },
          {
            "accesses": [
              {
                "type": "GET_METADATA_OBJECTS",
                "isAllowed": true
              }
            ],
            "users": [
              "{USER}"
            ],
            "groups": [],
            "roles": [],
            "conditions": [],
            "delegateAdmin": false
          }
        ],
        "denyPolicyItems": [],
        "allowExceptions": [],
        "denyExceptions": [],
        "dataMaskPolicyItems": [],
        "rowFilterPolicyItems": [],
        "options": {},
        "validitySchedules": [],
        "policyLabels": [],
        "isDenyAllElse": true,
        "isEnabled": true
      }
EOD
  )"

  echo "Creating policy file gcs-dataproc service for temp bucket"
  local -r hadoop_dataproc_temp_bucket_all_policy="$(cat <<-EOD
      {
        "service": "gcs-dataproc",
        "name": "hadoop group and dataproc user to access dataproc temp bucket",
        "policyType": 0,
        "policyPriority": 0,
        "description": "hadoop group and dataproc user to access dataproc temp bucket",
        "isAuditEnabled": true,
        "resources": {
          "bucket": {
            "values": [
              "${dataproc_temp_bucket}"
            ],
            "isExcludes": false,
            "isRecursive": false
          },
          "object-path": {
            "values": [
              "*"
            ],
            "isExcludes": false,
            "isRecursive": true
          }
        },
        "policyItems": [
          {
            "accesses": [
              {
                "type": "LIST_OBJECTS",
                "isAllowed": true
              },
              {
                "type": "READ_OBJECTS",
                "isAllowed": true
              },
              {
                "type": "WRITE_OBJECTS",
                "isAllowed": true
              },
              {
                "type": "MODIFY_OBJECTS",
                "isAllowed": true
              },
              {
                "type": "DELETE_OBJECTS",
                "isAllowed": true
              },
              {
                "type": "GET_BUCKETS",
                "isAllowed": true
              },
              {
                "type": "CREATE_BUCKETS",
                "isAllowed": true
              },
              {
                "type": "DELETE_BUCKETS",
                "isAllowed": true
              },
              {
                "type": "LIST_BUCKETS",
                "isAllowed": true
              },
              {
                "type": "GET_METADATA_OBJECTS",
                "isAllowed": true
              }
            ],
            "users": [
              "dataproc"
            ],
            "groups": [
              "hadoop"
            ],
            "roles": [],
            "conditions": [],
            "delegateAdmin": false
          },
          {
            "accesses": [
              {
                "type": "GET_METADATA_OBJECTS",
                "isAllowed": true
              }
            ],
            "users": [
              "{USER}"
            ],
            "groups": [],
            "roles": [],
            "conditions": [],
            "delegateAdmin": false
          }
        ],
        "denyPolicyItems": [],
        "allowExceptions": [],
        "denyExceptions": [],
        "dataMaskPolicyItems": [],
        "rowFilterPolicyItems": [],
        "options": {},
        "validitySchedules": [],
        "policyLabels": [],
        "isDenyAllElse": true,
        "isEnabled": true
      }
EOD
  )"

  echo "Creating policy file gcs-dataproc service for all - bucket, object-path"
  local -r allow_get_metadata_objects_all_policy="$(cat <<-EOD
      {
        "service": "gcs-dataproc",
        "name": "all - bucket, object-path",
        "policyType": 0,
        "policyPriority": 0,
        "description": "policy for all bucket and objectpath",
        "isAuditEnabled": true,
        "resources": {
          "bucket": {
            "values": [
              "*"
            ],
            "isExcludes": false,
            "isRecursive": false
          },
          "object-path": {
            "values": [
              "*"
            ],
            "isExcludes": false,
            "isRecursive": true
          }
        },
        "policyItems": [
          {
            "accesses": [
              {
                "type": "LIST_OBJECTS",
                "isAllowed": true
              },
              {
                "type": "READ_OBJECTS",
                "isAllowed": true
              },
              {
                "type": "WRITE_OBJECTS",
                "isAllowed": true
              },
              {
                "type": "MODIFY_OBJECTS",
                "isAllowed": true
              },
              {
                "type": "DELETE_OBJECTS",
                "isAllowed": true
              },
              {
                "type": "GET_BUCKETS",
                "isAllowed": true
              },
              {
                "type": "CREATE_BUCKETS",
                "isAllowed": true
              },
              {
                "type": "DELETE_BUCKETS",
                "isAllowed": true
              },
              {
                "type": "LIST_BUCKETS",
                "isAllowed": true
              },
              {
                "type": "GET_METADATA_OBJECTS",
                "isAllowed": true
              }
            ],
            "users": [
              "rangerlookup",
              "dataproc"
            ],
            "groups": [],
            "roles": [],
            "conditions": [],
            "delegateAdmin": false
          },
          {
            "accesses": [
              {
                "type": "GET_METADATA_OBJECTS",
                "isAllowed": true
              }
            ],
            "users": [
              "{USER}"
            ],
            "groups": [],
            "roles": [],
            "conditions": [],
            "delegateAdmin": false
          }
        ],
        "denyPolicyItems": [],
        "allowExceptions": [],
        "denyExceptions": [],
        "dataMaskPolicyItems": [],
        "rowFilterPolicyItems": [],
        "options": {},
        "validitySchedules": [],
        "policyLabels": [],
        "isDenyAllElse": false,
        "isEnabled": true
      }
EOD
  )"

  local -r policy_endpoint="http://${host_name}:${RANGER_ADMIN_PORT}/service/public/v2/api/policy"
  echo "policy_endpoint is ${policy_endpoint}"

  local -r numRetries=30
  echo "Adding policy hadoop_dataproc_bucket_all_policy"
  set +x
  curl --retry "${numRetries}" --retry-max-time 30 -u "admin:${admin_password}" -X POST \
    -H 'Accept: application/json' -H 'Content-Type: application/json' \
    --data "${hadoop_dataproc_bucket_all_policy}" "${policy_endpoint}"
  echo ''

  echo "Adding policy hadoop_dataproc_temp_bucket_all_policy"
  curl --retry "${numRetries}" --retry-max-time 30 -u "admin:${admin_password}" -X POST \
    -H 'Accept: application/json' -H 'Content-Type: application/json' \
    --data "${hadoop_dataproc_temp_bucket_all_policy}" "${policy_endpoint}"
  echo ''

  # Remove the default policy for all bucket all path and upload a new one
  # Assuming the default one only contains a condition for rangerlookup
  # We can't use mergeIfExists since it is in v2.1
  # delete then upload since updating crashes Ranger for some reason
  # TODO this is blocking us form supporting external Ranger admin
  NAME='all - bucket, object-path'
  echo "Deleting current policy ${NAME}"
  curl --retry "${numRetries}" --retry-max-time 30 -u "admin:${admin_password}" -X DELETE \
    -H 'Accept: application/json' -H 'Content-Type: application/json' \
    "${policy_endpoint}?servicename=gcs-dataproc&policyname=${NAME// /%20}"
  echo ''

  echo "Adding policy allow_get_metadata_objects_all_policy"
  curl --retry "${numRetries}" --retry-max-time 30 -u "admin:${admin_password}" -X POST \
    -H 'Accept: application/json' -H 'Content-Type: application/json' \
    --data "${allow_get_metadata_objects_all_policy}" \
    "${policy_endpoint}"
  echo ''
  set -x
}

function register_ranger_gcs_service_to_ranger_admin() {
  echo "register_ranger_gcs_service_to_ranger_admin"
  copy_ranger_gcs_plugin_service_jar_to_ranger_admin
  install_ranger_gcs_plugin_servicedef
  install_ranger_gcs_plugin_default_policy
}

# Create custom IAM roles
function create_iam_role() {
  local -r role=$1
  local -r role_title=$2
  local -r permissions=$3
  local -r project="$(get_metadata_project_id)"
  local -r role_exists=$(gcloud iam roles list --project="${project}" | grep "${role}")
  if [[ -z "${role_exists}" ]]; then
    gcloud iam roles create "${role}" --title="${role_title}" --permissions="${permissions}" --stage=GA --project="${project}"
    echo "IAM Role ${role} Created."
  else
    echo "IAM Role ${role} Exists. Skipping."
  fi
}

function create_ranger_gcs_plugin_iam_roles() {
  create_iam_role 'gcsRangerAuthorizationServiceListObjectsRole' 'GCS Objects List Role' 'storage.objects.list'
  create_iam_role 'gcsRangerAuthorizationServiceWriteObjectsRole' 'GCS Objects Write Role' 'storage.objects.create'
  create_iam_role 'gcsRangerAuthorizationServiceEditObjectsRole' 'GCS Objects Edit Role' 'storage.objects.create,storage.objects.get'
  create_iam_role 'gcsRangerAuthorizationServiceGetObjectsRole' 'GCS Objects Read Role' 'storage.objects.get'
  create_iam_role 'gcsRangerAuthorizationServiceDeleteObjectsRole' 'GCS Objects Delete Role' 'storage.objects.delete'
  create_iam_role 'gcsRangerAuthorizationServiceGetBucketsRole' 'GCS Buckets Read Role' 'storage.buckets.get'
  create_iam_role 'gcsRangerAuthorizationServiceCreateBucketsRole' 'GCS Buckets Create Role' 'storage.buckets.create'
  create_iam_role 'gcsRangerAuthorizationServiceDeleteBucketsRole' 'GCS Buckets Delete Role' 'storage.buckets.delete'
  create_iam_role 'gcsRangerAuthorizationServiceListBucketsRole' 'GCS Buckets List Role' 'storage.buckets.list'
  echo "Done creating all necessary iam roles for ranger_gcs_plugin"
}

function set_ranger_gcs_plugin_property() {
  local -r name=$1
  local -r value=$2
  local -r file=$3
  bdconfig set_property \
      --configuration_file "${RANGER_GCS_PLUGIN_CONF_DIR}/${file}" \
      --name "${name}" \
      --value "${value}" \
      --clobber
}

function set_property_in_core_site() {
  local -r name=$1
  local -r value=$2
  bdconfig set_property \
      --configuration_file "/etc/hadoop/conf/core-site.xml" \
      --name "${name}" \
      --value "${value}" \
      --clobber
}

function set_property_ranger_gcs_site() {
  local -r name=$1
  local -r value=$2
  set_ranger_gcs_plugin_property "${name}" "${value}" 'ranger-gcs-site.xml'
}

function set_property_ranger_gcs_security() {
  local -r name=$1
  local -r value=$2
  set_ranger_gcs_plugin_property "${name}" "${value}" 'ranger-gcs-security.xml'
}

function set_property_ranger_gcs_audit() {
  local -r name=$1
  local -r value=$2
  set_ranger_gcs_plugin_property "${name}" "${value}" 'ranger-gcs-audit.xml'
}

function get_ranger_gcs_plugin_mysql_password_from_gcs() {
  local -r kms_key_uri=$(get_dataproc_property ranger.gcs.plugin.mysql.kms.key.uri)
  local -r root_password_uri=$(get_dataproc_property ranger.gcs.plugin.mysql.password.uri)
  if [[ -z "${root_password_uri}" ]]; then
    log_and_fail "Ranger ${RANGER_GCS_PLUGIN_COMPONENT_NAME} Plugin" "Ranger GCS Plugin is enabled but failed to read root principal \
  password URI from Dataproc properties. Set it through 'dataproc:ranger.gcs.plugin.mysql.password.uri'." "${EXIT_CODE_CLIENT_ERROR}"
  fi

  local -r root_password=$(decrypt_with_kms_key "${root_password_uri}" "${kms_key_uri}")
  status=$?
  if [[ $status -ne 0 ]]; then
    log_and_fail "${RANGER_GCS_PLUGIN_COMPONENT_NAME}" "Ranger GCS Plugin failed to get encrypted file from Google Cloud Storage and/or failed to decrypt using \
    the KMS key. Please make sure the KMS key ${kms_key_uri} exists, the VM service account has \
    'storage.objects.get' permission to ${root_password_uri}, and \
    'cloudkms.cryptoKeyVersions.useToDecrypt' permission to ${kms_key_uri}." "${EXIT_CODE_CLIENT_ERROR}"
  fi

  if [[ -z "${root_password}" ]]; then
    log_and_fail "${RANGER_GCS_PLUGIN_COMPONENT_NAME}" "Root principal password cannot be empty!" "${EXIT_CODE_CLIENT_ERROR}"
  fi

  echo "${root_password}"
}

function restart_master_services() {
  services=('hadoop-hdfs-namenode' 'hadoop-hdfs-secondarynamenode' 'hadoop-yarn-resourcemanager' 'hive-server2' 'hive-metastore' 'hadoop-yarn-timelineserver' 'hadoop-mapreduce-historyserver' 'spark-history-server')
  systemctl daemon-reload
  for service in "${services[@]}"; do
    if (systemctl is-enabled --quiet "${service}"); then
      echo "Restarting Service ${service} on $(hostname)"
      retry_constant_custom 60 1 systemctl stop "${service}"
      retry_constant_custom 60 1 systemctl start "${service}"
    fi
  done
}

function restart_services() {
  local role
  role="$(get_metadata_role)"
  # Worker services only start after the startup scripts are done.
  if [[ "${role}" == 'Master' ]]; then
    echo "Restarting master services"
    restart_master_services
  fi
}

function install_ranger_gcs_plugin_client() {
  echo "Installing  Ranger GCS Plugin Client"
  # Linking client jar to dataproc common lib dir is done in build-image.sh so
  # that the agent can pick up the jar without a restart during start up.
  set_property_in_core_site 'fs.gs.delegation.token.binding' 'com.google.cloud.hadoop.ranger.gcs.hadoop.fs.AuthorizationServerDelegationTokenBinding'
  set_property_in_core_site 'fs.gs.auth.null.enable' 'true'
  set_property_in_core_site 'fs.gs.auth.service.account.enable' 'false'
  set_property_in_core_site 'authorization.service.host' 'localhost'
  set_property_in_core_site 'authorization.service.port' '6081'
  set_property_in_core_site 'authorization.service.authentication.spnego.keytab' "${RANGER_GCS_PLUGIN_GCS_KEYTAB}"
  # REALM is defined in shared/ranger.sh
  set_property_in_core_site 'authorization.service.authentication.spnego.principal' "gcs/${FQDN}@${REALM}"

  # Restart hadoop services so that they can pick up new library and config.
  # This doesn't restart the agent.
  echo "Installing Restarting Hadoop services"
  restart_services
}

function install_gcs_root_principal_on_kerberos() {
  echo "Running install_gcs_root_principal_on_kerberos"
  local kerberos_password
  set +x
  kerberos_password="$(get_kerberos_root_password)"
  # Sanity check.
  if [[ -z "${kerberos_password}" ]]; then
    log_and_fail "${RANGER_GCS_PLUGIN_COMPONENT_NAME}" 'Failed to get Kerberos password.' "${EXIT_CODE_CLIENT_ERROR}"
  fi
  # FQDN is defined in shared/ranger.sh
  kadmin -p root -w "${kerberos_password}" -q "addprinc -randkey gcs/${FQDN}"
  kadmin -p root -w "${kerberos_password}" -q "ktadd -k ${RANGER_GCS_PLUGIN_GCS_KEYTAB} gcs/${FQDN}"
  set -x
  chmod 400 "${RANGER_GCS_PLUGIN_GCS_KEYTAB}"
}

function copy_ranger_gcs_plugin_auth_service_artifacts() {
  # Copy jars
  mkdir -p "${RANGER_GCS_PLUGIN_LIB_DIR}"
  chmod 751 "${RANGER_GCS_PLUGIN_LIB_DIR}"
  # Clear whatever inside the lib dir since users may be able to create this dir and put things in it before we create it
  # Keep the ":?" part just in case the variable for some reason is empty or unset
  rm -rf "${RANGER_GCS_PLUGIN_LIB_DIR:?}/*"
  cp "${RANGER_GCS_PLUGIN_PLUGIN_SRC}/ranger-gcs-authorization-service-"*".jar" "${RANGER_GCS_PLUGIN_LIB_DIR}/ranger-gcs-plugin-authorization-server.jar"

  # Copy and modify the configuration files
  mkdir -p "${RANGER_GCS_PLUGIN_CONF_DIR}"
  cp "${RANGER_GCS_PLUGIN_PLUGIN_SRC}/ranger-gcs-site.xml" "${RANGER_GCS_PLUGIN_CONF_DIR}/ranger-gcs-site.xml"
  cp "${RANGER_GCS_PLUGIN_PLUGIN_SRC}/ranger-gcs-audit.xml" "${RANGER_GCS_PLUGIN_CONF_DIR}/ranger-gcs-audit.xml"
  cp "${RANGER_GCS_PLUGIN_PLUGIN_SRC}/ranger-gcs-security.xml" "${RANGER_GCS_PLUGIN_CONF_DIR}/ranger-gcs-security.xml"
  chmod 700 "${RANGER_GCS_PLUGIN_CONF_DIR}/ranger-gcs-site.xml"
  chmod 700 "${RANGER_GCS_PLUGIN_CONF_DIR}/ranger-gcs-audit.xml"
  chmod 700 "${RANGER_GCS_PLUGIN_CONF_DIR}/ranger-gcs-security.xml"
  local -r mysql_root_user='root'
  local -r mysql_auth_db_name='authsvc'
  local -r mysql_auth_user='authsvc'
  local role

  role="$(/usr/share/google/get_metadata_value attributes/dataproc-role)"
  set +x
  local mysql_root_password
  local mysql_auth_password
  mysql_auth_password="$(get_ranger_gcs_plugin_mysql_password_from_gcs)"
  set -x

  if [[ -n "${CLOUD_SQL_INSTANCE}" ]]; then
    kms_key_uri=$(get_dataproc_property ranger.kms.key.uri)
    db_root_password_uri=$(get_dataproc_property ranger.cloud-sql.root.password.uri)
    set +x
    db_root_password=$(decrypt_file_with_kms_key "${db_root_password_uri}" "${kms_key_uri}")
    mysql_root_password=${db_root_password}
    set -x
  else
     set +x
     mysql_root_password="$(read_mysql_password)"
     set -x
  fi

  # Create the database only once from the master node.
  # In HA clusters Namenode it is created on m-0. 
  if [[ "${role}" == 'Master' && "${NODE_NAME}" =~ ^.*(-m|-m-0)$ ]]; then
    echo "Creating database ${mysql_auth_db_name} on ${role} using ${mysql_root_user} for mysql host ${MYSQL_HOST}:${MYSQL_PORT}"
    set +x
    echo "CREATE DATABASE IF NOT EXISTS ${mysql_auth_db_name}"
    mysql -u "${mysql_root_user}" -p"${mysql_root_password}" -h "${MYSQL_HOST}" -P "${MYSQL_PORT}" -e "CREATE DATABASE IF NOT EXISTS ${mysql_auth_db_name};"

    echo "CREATE USER IF NOT EXISTS ${mysql_auth_user}"
    mysql -u "${mysql_root_user}" -p"${mysql_root_password}" -h "${MYSQL_HOST}" -P "${MYSQL_PORT}" -e "CREATE USER IF NOT EXISTS '${mysql_auth_user}'@'%' IDENTIFIED BY '${mysql_auth_password}'"

    echo "GRANT ALL PRIVILEGES ON ${mysql_auth_db_name} TO ${mysql_auth_user}"
    mysql -u "${mysql_root_user}" -p"${mysql_root_password}" -h "${MYSQL_HOST}" -P "${MYSQL_PORT}" -e "GRANT ALL PRIVILEGES ON ${mysql_auth_db_name}.* TO '${mysql_auth_user}'@'%'"
    set -x
  else
    echo "Skipping Creating database ${mysql_auth_db_name} on host $(hostname)"
  fi

  local metadatastore_jdbc_uri
  if [[ -n "${CLOUD_SQL_INSTANCE}" ]]; then
    metadatastore_jdbc_uri="jdbc:mysql://${MYSQL_HOST}:${MYSQL_PORT}/metastore"
    bdconfig set_property \
      --configuration_file /etc/hive/conf/hive-site.xml \
      --name 'javax.jdo.option.ConnectionURL' \
      --value "${metadatastore_jdbc_uri}" \
      --clobber

    if [[ "${role}" == "Master" ]]; then
      echo "Restarting hive-server2 and hive-metastore as well"
      retry_constant_short systemctl restart hive-server2
      retry_constant_short systemctl restart hive-metastore
    fi
  fi

  echo "Setting properties for Ranger GCS"
  set_property_ranger_gcs_site 'gcp.project.id' "$(get_metadata_project_id)"
  set_property_ranger_gcs_site 'authorization.service.host' 'localhost'
  set_property_ranger_gcs_site 'authorization.service.port' '6081'
  set_property_ranger_gcs_site 'authorization.service.authentication.spnego.keytab' "${RANGER_GCS_PLUGIN_GCS_KEYTAB}"
  set_property_ranger_gcs_site 'authorization.service.authentication.spnego.principal' "gcs/${FQDN}@${REALM}"
  set_property_ranger_gcs_site 'authorization.service.proxyuser.yarn.users' "*"
  set_property_ranger_gcs_site 'authorization.service.proxyuser.hive.users' "*"
  set_property_ranger_gcs_site 'authorization.service.proxyuser.hdfs.users' "*"

  # MASTER is defined in shared/ranger.sh
  local jdbc_driver_url
  if [[ -n "${CLOUD_SQL_INSTANCE}" ]]; then
    jdbc_driver_url="jdbc:mysql://${MYSQL_HOST}:${MYSQL_PORT}/${mysql_auth_db_name}"
  else
    jdbc_driver_url="jdbc:mysql://${MASTER}:${MYSQL_PORT}/${mysql_auth_db_name}"
  fi  
  echo "Setting authorization.service.database.jdbc.driver.url to ${jdbc_driver_url}"
  if [[ -n "${jdbc_driver_url}" ]]; then
    set_property_ranger_gcs_site 'authorization.service.database.jdbc.driver.url' "${jdbc_driver_url}"
    set_property_ranger_gcs_site 'authorization.service.database.jdbc.driver.user' "${mysql_auth_user}"
    set +x
    set_property_ranger_gcs_site 'authorization.service.database.jdbc.driver.password' "${mysql_auth_password}"
    set -x
  fi

  # RANGER_ADMIN_PORT is defined in shared/ranger.sh
  set_property_ranger_gcs_security 'ranger.plugin.gcs.service.name' 'gcs-dataproc'
  set_property_ranger_gcs_security 'ranger.plugin.gcs.policy.rest.url' "http://${MASTER}:${RANGER_ADMIN_PORT}"
  set_property_ranger_gcs_security 'ranger.plugin.gcs.policy.cache.dir' '/etc/ranger/gcs/policycache'

  local -r solr_server_host="${MASTER}"
  local -r solr_server_port='8983'
  set_property_ranger_gcs_audit 'xasecure.audit.solr.is.enabled' 'true'
  set_property_ranger_gcs_audit 'xasecure.audit.destination.solr.urls' "http://${solr_server_host}:${solr_server_port}/solr/ranger_audits"
  set_property_ranger_gcs_audit 'xasecure.audit.destination.solr.user' 'solr'
  set_property_ranger_gcs_audit 'xasecure.audit.solr.is.enabled' 'true'
  set_property_ranger_gcs_audit 'xasecure.audit.destination.solr.force.use.inmemory.jaas.config' 'true'
  set_property_ranger_gcs_audit 'xasecure.audit.jaas.Client.loginModuleControlFlag' 'required'
  set_property_ranger_gcs_audit 'xasecure.audit.jaas.Client.loginModuleName' 'com.sun.security.auth.module.Krb5LoginModule'
  set_property_ranger_gcs_audit 'xasecure.audit.jaas.Client.option.keyTab' '/etc/security/keytab/gcs.service.keytab'
  set_property_ranger_gcs_audit 'xasecure.audit.jaas.Client.option.principal' "gcs/_HOST@${REALM}"
  set_property_ranger_gcs_audit 'xasecure.audit.jaas.Client.option.serviceName' 'solr'
  set_property_ranger_gcs_audit 'xasecure.audit.jaas.Client.option.storeKey' 'false'
  set_property_ranger_gcs_audit 'xasecure.audit.jaas.Client.option.useKeyTab' 'true'
}

function update_native_hadoop_apps_configs_for_ranger_gcs_plugin() {
  echo "Running update_native_hadoop_apps_configs_for_ranger_gcs_plugin"
  bdconfig set_property \
    --configuration_file '/etc/hive/conf/hive-site.xml' \
    --name 'hive.user.install.directory' \
    --value 'hdfs:///user/' \
    --clobber

  allowlist_config=$(bdconfig get_property_value \
    --configuration_file '/etc/hive/conf/hive-site.xml' \
    --name 'hive.security.authorization.sqlstd.confwhitelist.append')
  if [[ ${allowlist_config} == 'None' ]]; then
    allowlist_config=""
  else
    allowlist_config+="|"
  fi

  bdconfig set_property \
    --configuration_file '/etc/hive/conf/hive-site.xml' \
    --name 'hive.security.authorization.sqlstd.confwhitelist.append' \
    --value "${allowlist_config}tez.job.fs-servers|mapreduce.job.hdfs-servers" \
    --clobber

  local -r dataproc_temp_bucket="$(/usr/share/google/get_metadata_value attributes/dataproc-temp-bucket)"
  bdconfig set_property \
    --configuration_file '/etc/hadoop/conf/mapred-site.xml' \
    --name 'mapreduce.job.hdfs-servers' \
    --value "gs://${dataproc_temp_bucket}" \
    --clobber
}

function start_ranger_gcs_plugin_auth_service() {
  if [[ ! $(ls "${RANGER_GCS_PLUGIN_LIB_DIR}/"slf4j-api-*) ]]; then
    ln -s /usr/lib/hadoop/lib/slf4j-api-*.jar "${RANGER_GCS_PLUGIN_LIB_DIR}/"
    ln -s /usr/lib/hadoop/lib/slf4j-reload4j-*.jar "${RANGER_GCS_PLUGIN_LIB_DIR}/"
  fi
  cat <<EOF >"${RANGER_GCS_PLUGIN_CONF_DIR}/log4j.properties"
  log4j.rootLogger=INFO, LOGFILE
  log4j.appender.LOGFILE=org.apache.log4j.RollingFileAppender
  log4j.appender.LOGFILE.File=/var/log/ranger-gcs-plugin-authorization-server.log
  log4j.appender.LOGFILE.MaxFileSize=10MB
  log4j.appender.LOGFILE.MaxBackupIndex=10
  log4j.appender.LOGFILE.append=true
  log4j.appender.LOGFILE.layout=org.apache.log4j.PatternLayout
  log4j.appender.LOGFILE.layout.ConversionPattern=%d{yyyy-MM-dd HH:mm:ss,SSS} [%t] %-5p %c %x - %m%n
EOF

  local java_executable
  java_executable="$(which java)"
  cat <<EOF >"/etc/systemd/system/ranger-gcs-plugin-authorization-server.service"
  [Unit]
  Description=Ranger GCS Plugin Authorization Service
  ConditionPathExists=${RANGER_GCS_PLUGIN_BASE_DIR}

  [Service]
  Type=simple
  Environment=CONFIG_FILE=${RANGER_GCS_PLUGIN_CONF_DIR}/ranger-gcs-site.xml
  WorkingDirectory=${RANGER_GCS_PLUGIN_BASE_DIR}
  ExecStart=${java_executable} -cp ${RANGER_GCS_PLUGIN_LIB_DIR}/*:${RANGER_GCS_PLUGIN_CONF_DIR} -Dlog4j.configurationFile=log4j.properties com.google.cloud.hadoop.ranger.gcs.authorization.service.AuthorizationService
  Restart=on-failure
  RestartSec=10

  [Install]
  WantedBy=multi-user.target
EOF

  # Enable the service.
  echo "Starting ranger-gcs-plugin-authorization-server"
  systemctl daemon-reload
  enable_and_start_service "ranger-gcs-plugin-authorization-server"
  retry_constant_short systemctl start ranger-gcs-plugin-authorization-server
}

function install_ranger_gcs_plugin_authorization_service() {
  install_gcs_root_principal_on_kerberos

  # these jars are not available in dataproc 2.1 yet
  if [[ -d ${RANGER_GCS_PLUGIN_BASE_DIR} ]]; then
    echo "Running copy_ranger_gcs_plugin_auth_service_artifacts"
    copy_ranger_gcs_plugin_auth_service_artifacts
  else
    echo "Ranger service artifacts ranger_gcs_plugin_auth unavailable"
  fi

  echo "Executing update_native_hadoop_apps_configs_for_ranger_gcs_plugin"
  update_native_hadoop_apps_configs_for_ranger_gcs_plugin
  echo "Executing start_ranger_gcs_plugin_auth_service"
  start_ranger_gcs_plugin_auth_service
  echo "Completed Installation of Ranger gcs Plugin Authorization Service"
}
