# Note: this script should be sourced into startup-script.sh.

# Optional step of startup-script.sh to configure components at VM boot time
# before activating them.
function pre_activate_components() {
  local components=("$@")
  run_components_scripts "pre-activate" "${components[@]}"
}

# Called from startup-script.sh to activate components
# at VM boot time.
#
# Usage:
# activate_component "my-component"
#
# Where
# - COMPONENT the component to activate
function activate_component() {
  local -r component=$1
  local -r activate_script="${BDUTIL_DIR}/components/activate/${component}.sh"

  if [[ -f "${activate_script}" ]]; then
    echo "Running component activation script: ${activate_script}"
    mkdir -p "${DATAPROC_COMPONENTS_TMP_DIR}/activate"
    touch "${DATAPROC_COMPONENTS_TMP_DIR}/activate/${component}.running"
    local exit_code=0
    local -r start=$(date +%s.%N)
    bash -e "${activate_script}" || exit_code=$?
    local -r end=$(date +%s.%N)

    # Log human-readable runtime in seconds for debugging purposes.
    local -r runtime_s=$((${end%.*}-${start%.*}))
    echo "Component ${component} took ${runtime_s}s to activate"

    # Write start and end time to a file.
    # This file is read by the agent, which reports this info to the frontend.
    local -r time_file="${DATAPROC_COMPONENTS_TMP_DIR}/activate/${component}.time"
    touch "${time_file}"
    cat > "${time_file}" <<EOF
${start}
${end}
EOF

    if [[ ${exit_code} -ne 0 ]]; then
      echo "${exit_code}" \
        >"${DATAPROC_COMPONENTS_TMP_DIR}/activate/${component}.failed"
      log_and_fail "${component}" \
        "Component ${component} failed to activate" \
        ${exit_code}
    fi
    echo "Done running component activation script: ${activate_script}"
    touch "${DATAPROC_COMPONENTS_TMP_DIR}/activate/${component}.done"
  else
    log_and_fail "${component}" \
      "Component ${component} does not have activate file"
  fi
}

# Activate selected components.
function activate_components() {
  local components=("$@")

  mkdir -p "${DATAPROC_COMPONENTS_TMP_DIR}/activate"
  for component in "${components[@]}"; do
    # Exclude 'kerberos' component because it activated manually in startup script
    if [[ "${component}" != 'kerberos' ]]; then
      loginfo "Activating: ${component}"
      run_in_background --tag "activate-component-${component}" \
        activate_component "${component}"
    fi
  done
}

# Performs parts of component activation that require hdfs to be functional.
# This is invoked from post-hdfs-startup-script.sh
#
# This is an optional step, so not every component should have
# a post-hdfs file.
function post_hdfs_activate_component() {
  local -r component=$1
  local -r post_hdfs_script="${BDUTIL_DIR}/components/post-hdfs/${component}.sh"

  if [[ -f "${post_hdfs_script}" ]]; then
    echo "Running component activate post-hdfs script: ${post_hdfs_script}"
    touch "${DATAPROC_COMPONENTS_TMP_DIR}/post-hdfs/${component}.running"
    local exit_code=0
    local -r start=$(date +%s.%N)
    bash -e "${post_hdfs_script}" || exit_code=$?
    local -r end=$(date +%s.%N)

    # Log human-readable runtime in seconds for debugging purposes.
    local -r runtime_s=$((${end%.*}-${start%.*}))
    echo "Component ${component} took ${runtime_s}s to activate post-hdfs"

    # Write start and end time to a file.
    # This file is read by the agent, which reports this info to the frontend.
    local -r time_file="${DATAPROC_COMPONENTS_TMP_DIR}/post-hdfs/${component}.time"
    touch "${time_file}"
    cat > "${time_file}" <<EOF
${start}
${end}
EOF

    if [[ ${exit_code} -ne 0 ]]; then
      echo "${exit_code}" \
        >"${DATAPROC_COMPONENTS_TMP_DIR}/post-hdfs/${component}.failed"
      log_and_fail "${component}" \
        "Component ${component} failed to activate post-hdfs"
    fi
    touch "${DATAPROC_COMPONENTS_TMP_DIR}/post-hdfs/${component}.done"
  else
    echo "Component ${component} doesn't have a post-hdfs script"
  fi
}

# Run post-hdfs activation scripts of components.
function post_hdfs_activate_components() {
  local components=("$@")

  mkdir -p "${DATAPROC_COMPONENTS_TMP_DIR}/post-hdfs"
  for component in "${components[@]}"; do
    loginfo "Activating post-hdfs component ${component}"
    run_in_background --tag "post-hdfs-activate-component-${component}" \
      post_hdfs_activate_component "${component}"
  done
}

# Performs pre-uninstallation of non-debian component.
#
# This is an optional step, so not every component should have
# a pre-uninstall file.
function pre_uninstall_component() {
  local -r component=$1
  local -r pre_uninstall_script="${BDUTIL_DIR}/components/pre-uninstall/${component}.sh"

  if [[ -f "${pre_uninstall_script}" ]]; then
    echo "Running component pre-uninstall script: ${pre_uninstall_script}"
    touch "${DATAPROC_COMPONENTS_TMP_DIR}/pre-uninstall/${component}.running"
    local exit_code=0
    bash -e "${pre_uninstall_script}" || exit_code=$?
    if [[ ${exit_code} -ne 0 ]]; then
      echo "${exit_code}" \
        >"${DATAPROC_COMPONENTS_TMP_DIR}/pre-uninstall/${component}.failed"
      log_and_fail "${component}" \
        "Component ${component} failed to pre-uninstall"
    fi
    touch "${DATAPROC_COMPONENTS_TMP_DIR}/pre-uninstall/${component}.done"
  else
    echo "Component ${component} doesn't have a pre-uninstall script"
  fi
}

# Run pre-unstall scripts of components to mark deb packages as to-uninstall.
function pre_uninstall_components() {
  local components=("$@")

  mkdir -p "${DATAPROC_COMPONENTS_TMP_DIR}/pre-uninstall"
  for component in "${components[@]}"; do
    loginfo "Pre-uninstalling component ${component}"
    pre_uninstall_component "${component}"
  done
}

# Performs uninstallation of non-debian component.
function uninstall_component() {
  local -r component=$1
  local -r uninstall_script="${BDUTIL_DIR}/components/uninstall/${component}.sh"

  if [[ -f "${uninstall_script}" ]]; then
    echo "Running component uninstall script: ${uninstall_script}"
    touch "${DATAPROC_COMPONENTS_TMP_DIR}/uninstall/${component}.running"
    local exit_code=0
    bash -e "${uninstall_script}" || exit_code=$?
    if [[ ${exit_code} -ne 0 ]]; then
      echo "${exit_code}" \
        >"${DATAPROC_COMPONENTS_TMP_DIR}/uninstall/${component}.failed"
      log_and_fail "${component}" \
        "Component ${component} failed to uninstall"
    fi
    touch "${DATAPROC_COMPONENTS_TMP_DIR}/uninstall/${component}.done"
  else
    echo "Component ${component} doesn't have an uninstall script"
  fi
}

# Run uninstall scripts of components to uninstall non-deb packages or do
# post-uninstall cleanup for deb packages.
function uninstall_components() {
  local components=("$@")

  mkdir -p "${DATAPROC_COMPONENTS_TMP_DIR}/uninstall"
  for component in "${components[@]}"; do
    loginfo "Uninstalling component ${component}"
    uninstall_component "${component}"
  done
}
