#!/usr/bin/env bash

# Configure stackdriver monitoring and logging agents in docker containers. This
# is only applicable if the Image version is at least 2.2 and
# `dataproc:dataproc.observability.containerised.legacy.agents.enable` property is set to
# true. It requires DOCKER to be present as a default component.
# Since the UCP metrics agent binary is also extracted from Docker image, we have the
# logic to extract it inside /etc/ucp/monitoring directory

# TODO: Configure cpu and memory limits for the containers if needed.

set -eux
source "${BDUTIL_DIR}/bdutil_metadata.sh"

MONITORING_IMAGE=gcr.io/cloud-dataproc/observability/monitoring-agent:latest
LOGGING_IMAGE=gcr.io/cloud-dataproc/observability/logging-agent:latest
# TODO: Unpin the version to auto pick newer one from artifact registry
UCP_OBSERVABILITY_BASE_PATH=/etc/ucp
UCP_METRICS_AGENT_VERSION="0.86.0"
UCP_METRICS_AGENT_PATH="${UCP_OBSERVABILITY_BASE_PATH}/monitoring"
UCP_LOGGING_AGENT_PATH="${UCP_OBSERVABILITY_BASE_PATH}/logging"

function fetch_monitoring_agent_image() {
  docker pull $MONITORING_IMAGE
}

function fetch_logging_agent_image() {
  docker pull $LOGGING_IMAGE
}

function delete_observability_agents_docker_images() {
  systemctl start docker

  docker image rm $MONITORING_IMAGE
  docker image rm $LOGGING_IMAGE
}

function configure_containerised_monitoring_agent() {

  MONITORING_AGENT_CONTAINER_NAME='monitoring-agent'
  monitoring_args=()
  monitoring_args+=("--name=${MONITORING_AGENT_CONTAINER_NAME}")
  monitoring_args+=("--privileged")
  monitoring_args+=("--volume=/etc/stackdriver/collectd.d:/etc/stackdriver/collectd.d")
  monitoring_args+=("--volume=/opt/stackdriver/collectd/etc/collectd.d/:/opt/stackdriver/collectd/etc/collectd.d/")
  monitoring_args+=("--volume=/etc/stackdriver/collectd.conf:/etc/stackdriver/collectd.conf")
  monitoring_args+=("--volume=/var/log/:/var/log/")
  monitoring_args+=("--volume=/etc/passwd:/etc/passwd")
  monitoring_args+=("--network=host")
  monitoring_args+=("--pid=host")
  monitoring_args+=("${MONITORING_IMAGE}")

  cat >"/etc/systemd/system/stackdriver-agent-docker.service"<<EOF
[Unit]
Description=Stackdriver monitoring agent container
After=docker.service
Requires=docker.service

[Service]
Type=simple
User=root
ExecStart=docker run --rm ${monitoring_args[@]}
Restart=always
RestartSec=5
SuccessExitStatus=143

[Install]
WantedBy=multi-user.target
EOF
}

function configure_containerised_monitoring_agent_for_s8s() {
  local -r monitoring_agent_container_name='monitoring-agent'
  local -r cached_monitoring_agent_image='dataproc.local/monitoring-agent'
  local -r monitoring_args=(
    '--privileged'
    '--mount type=bind,src=/etc,dst=/etc,options=rbind:ro'
    '--mount type=bind,src=/opt/stackdriver/collectd/etc/collectd.d/,dst=/opt/stackdriver/collectd/etc/collectd.d/,options=rbind:ro'
    '--mount type=bind,src=/etc/stackdriver,dst=/etc/stackdriver,options=rbind:ro'
    '--mount type=bind,src=/var/log/,dst=/var/log/,options=rbind:ro'
    '--net-host'
    "${cached_monitoring_agent_image}"
    "${monitoring_agent_container_name}"
  )

  cat >"/etc/systemd/system/stackdriver-agent-ctr.service"<<EOF
[Unit]
Description=Stackdriver Monitoring Agent (containerized)
After=containerd.service
Requires=containerd.service

[Service]
Type=simple
User=root
ExecStart=ctr --namespace=k8s.io run --rm ${monitoring_args[@]}
ExecStop=ctr --namespace=k8s.io task kill -s SIGKILL ${monitoring_agent_container_name}
Restart=always
RestartSec=5
SuccessExitStatus=137

[Install]
WantedBy=multi-user.target
EOF
}

function configure_containerised_logging_agent() {
  LOGGING_AGENT_CONTAINER_NAME='logging-agent'

  logging_args=()
  logging_args+=("--name=${LOGGING_AGENT_CONTAINER_NAME}")
  logging_args+=("--volume=/var/log/:/var/log/")
  logging_args+=("--volume=/var/tmp/:/var/tmp/")
  logging_args+=("--volume=/etc/google-fluentd:/etc/google-fluentd")
  logging_args+=("--network=host")
  logging_args+=("--pid=host")
  # Hadoop configs and libraries are needed to enable job logging plugin
  logging_args+=("-e JAVA_HOME=${JAVA_HOME}")
  logging_args+=("--volume=${JAVA_HOME}:${JAVA_HOME}")
  logging_args+=("--volume=/etc/hadoop/conf.empty:/etc/hadoop/conf")
  logging_args+=("--volume=/usr/bin/hdfs:/usr/bin/hdfs")
  logging_args+=("--volume=/usr/lib/hadoop:/usr/lib/hadoop")
  logging_args+=("--volume=/usr/lib/hadoop-hdfs:/usr/lib/hadoop-hdfs")
  logging_args+=("--volume=/usr/lib/hadoop-yarn:/usr/lib/hadoop-yarn")
  logging_args+=("--volume=/usr/lib/hadoop-mapreduce:/usr/lib/hadoop-mapreduce")
  logging_args+=("--volume=/usr/lib/bigtop-utils:/usr/lib/bigtop-utils")
  if is_component_selected "kerberos"; then
    logging_args+=("--volume=/etc/krb5.conf:/etc/krb5.conf")
  fi
  logging_args+=("${LOGGING_IMAGE}")

  cat >"/etc/systemd/system/google-fluentd-docker.service"<<EOF
[Unit]
Description=Stackdriver logging agent container
After=docker.service
Requires=docker.service

[Service]
Type=simple
User=root
ExecStart=docker run --rm ${logging_args[@]}
Restart=always
RestartSec=5
SuccessExitStatus=143

[Install]
WantedBy=multi-user.target
EOF
}

function enable_containerised_observability_agents() {
  # Configure the monitoring and logging agents in here,
  # to determine whether kerberos is enabled or not.
  if [[ "${DATAPROC_IMAGE_TYPE}" == "standard" ]]; then
    configure_containerised_monitoring_agent
    configure_containerised_logging_agent

    systemctl start docker

    systemctl daemon-reload
    systemctl enable stackdriver-agent-docker
    systemctl enable google-fluentd-docker
  elif [[ "${DATAPROC_IMAGE_TYPE}" == "serverless-spark" ]] && is_debian12; then
    configure_containerised_monitoring_agent_for_s8s
    systemctl daemon-reload
    # enable and start the service after containerd is started
  fi
}

function configure_ucp_monitoring_agent() {
  local otel_args=()

  local -r config_files=$(find ${UCP_METRICS_AGENT_PATH}/configs/ -type f -name "*.yaml")
  for yaml_file in $config_files; do
    otel_args+=("--config ${yaml_file}")
  done

    local -r monitoring_agent_service_file=/etc/systemd/system/otel-metrics-agent.service
  cat >${monitoring_agent_service_file} <<EOF
[Unit]
Description=OpenTelemetry based agent for Dataproc monitoring over Google Cloud Monitoring
After=network.target

[Service]
Type=simple
Restart=on-failure
RestartSec=1
User=root
ExecStart=${UCP_METRICS_AGENT_PATH}/bin/otelsvc ${otel_args[@]}

[Install]
WantedBy=multi-user.target
EOF
}

function configure_ucp_logging_agent() {
  local -r fluentbit_service_file=/etc/systemd/system/fluentbit-ucp.service
  cat >${fluentbit_service_file} <<EOF
[Unit]
Description=Fluent bit agent for Dataproc logging over Google Cloud Logging
After=network.target

[Service]
Type=simple
Restart=on-failure
RestartSec=1
User=root
ExecStart=/opt/fluent-bit/bin/fluent-bit -c ${UCP_LOGGING_AGENT_PATH}/configs/fluentbit.conf

[Install]
WantedBy=multi-user.target
EOF
}

function enable_ucp_observability_agents() {
  # Configure the monitoing and logging agents in here,
  # to determine whether kerberos is enabled or not.
  configure_ucp_monitoring_agent
  configure_ucp_logging_agent

  systemctl daemon-reload
  systemctl enable otel-metrics-agent fluentbit-ucp
}

function install_ucp_monitoring_agent() {
  local -r gce_zone="$(get_metadata_zone)"
  local -r gce_region="${gce_zone%*\-*}" # Extract region from fully-qualified zone name, <region>-<zone>
  local -r ucp_artifact_registry="${gce_region}-docker.pkg.dev"
  gcloud -q auth configure-docker "${ucp_artifact_registry}"
  local -r image_name="${ucp_artifact_registry}/ucp-secure/ucp/otel-metrics-agent_java:${UCP_METRICS_AGENT_VERSION}"
  docker pull "${image_name}"
  mkdir -p "${UCP_METRICS_AGENT_PATH}/bin"
  mkdir "${UCP_METRICS_AGENT_PATH}/configs"
  local -r container_name="extract_otelsvc"
  docker create --name "${container_name}" "${image_name}"
  docker cp "${container_name}:./otelsvc" "${UCP_METRICS_AGENT_PATH}/bin"
  docker cp "${container_name}:./configs/java_extra_metrics.groovy" "${UCP_METRICS_AGENT_PATH}/configs"
  docker cp "${container_name}:./opt/opentelemetry-java-contrib-jmx-metrics.jar" "${UCP_METRICS_AGENT_PATH}/configs"
  docker rm -f ${container_name}
}

function delete_ucp_agent_libs() {
  # Uninstalls the fluent bit package and removes the UCP observability agents base installation directory
  os_uninstall_packages fluent-bit
  rm -rf "${UCP_OBSERVABILITY_BASE_PATH}"
}

function delete_unused_legacy_or_ucp_agents() {
  if use_ucp_observability_agents; then
    delete_observability_agents_docker_images
  elif is_containerised_legacy_agents_supported; then
    delete_ucp_agent_libs
  fi
}
