# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Generates the config files which will be needed by the hadoop servers such
# as 'slaves' listing all worker hostnames, 'masters' listing the master,
# and the xml files which go under the 'conf/' directory of the hadoop
# installation.

set -e
loginfo 'Running configure_hadoop.sh'

# Used for hadoop.tmp.dir
export HADOOP_TMP_DIR=/hadoop/tmp
mkdir -p ${HADOOP_TMP_DIR}

# Basic configuration not related to GHFS or HDFS.
# Rough rule-of-thumb settings for default maps/reduces taken from
# http://wiki.apache.org/hadoop/HowManyMapsAndReduces
export DEFAULT_NUM_MAPS=$((${NUM_WORKERS} * 10))
export DEFAULT_NUM_REDUCES=$((${NUM_WORKERS} * 4))

NUM_CORES="$(grep -c processor /proc/cpuinfo)"
export NUM_CORES
MAP_SLOTS=$(python -c "print(int(${NUM_CORES} // ${CORES_PER_MAP_TASK}))")
export MAP_SLOTS
REDUCE_SLOTS=$(python -c "print(int(${NUM_CORES} // ${CORES_PER_REDUCE_TASK}))")
export REDUCE_SLOTS

# Calculate the memory allocations, MB, using 'free -m'. Floor to nearest MB.
TOTAL_MEM_MB=$(free -m | awk '/^Mem:/{print $2}')
HADOOP_MR_MASTER_MEM_MB=$(python -c \
    "print(int(${TOTAL_MEM_MB} * ${HADOOP_MASTER_MAPREDUCE_MEMORY_FRACTION}))")

# MapReduce v2 (and YARN) Configuration
if [[ -x configure_mrv2_mem.py ]]; then
  TEMP_ENV_FILE=$(mktemp /tmp/mrv2_XXX_tmp_env.sh)
  ./configure_mrv2_mem.py \
    --output_file "${TEMP_ENV_FILE}" \
    --total_memory "${TOTAL_MEM_MB}" \
    --available_memory_ratio "${NODEMANAGER_MEMORY_FRACTION}" \
    --total_cores "${NUM_CORES}" \
    --cores_per_map "${CORES_PER_MAP_TASK}" \
    --cores_per_reduce "${CORES_PER_REDUCE_TASK}" \
    --cores_per_app_master "${CORES_PER_APP_MASTER}"
  source "${TEMP_ENV_FILE}"
  # Leave TMP_ENV_FILE around for debugging purposes.
fi

# Give Hadoop heap 1/4 of available memory, limit the max to 32GiB.
#
# Note that this also affects Hive Server2 and Metastore, because
# HADOOP_CLIENT_OPTS is used by them.
HADOOP_CLIENT_MEM_MB=$(python -c \
    "print(min(32 * 1024, int(${TOTAL_MEM_MB} / 4)))")

cat >>"${HADOOP_CONF_DIR}/hadoop-env.sh" <<EOF
# Set maximum Hadoop client heap. This also affects Hive Server2 and Metastore.
HADOOP_CLIENT_OPTS="-Xmx${HADOOP_CLIENT_MEM_MB}m \${HADOOP_CLIENT_OPTS}"

# Add Dataproc common libs (e.g. connectors) to Hadoop class path.
export HADOOP_CLASSPATH=\${HADOOP_CLASSPATH}:${DATAPROC_COMMON_LIB_DIR}/*
EOF

if is_version_at_least "${DATAPROC_IMAGE_VERSION}" "2.1"; then
  GC_LOG_OPTS="-Xlog:gc*:stdout:time,level,tags"
else
  GC_LOG_OPTS="-XX:+PrintGCTimeStamps -XX:+PrintGCDateStamps -XX:+PrintGCDetails"
fi

cat >>"${HADOOP_CONF_DIR}/yarn-env.sh" <<EOF

# Add Dataproc common libs (e.g. connectors) to YARN class path.
# (required for Hadoop 3: https://jira.apache.org/jira/browse/HADOOP-11656).
export HADOOP_CLASSPATH=\${HADOOP_CLASSPATH}:${DATAPROC_COMMON_LIB_DIR}/*

# Lessen stop-the-world collection:
GC_OPTS="-XX:+UseConcMarkSweepGC"
# Log GC details to stdout, these will be in diagnostic tarballs.
GC_LOGGING_OPTS="${GC_LOG_OPTS}"
export YARN_TIMELINESERVER_OPTS="\${GC_OPTS} \${GC_LOGGING_OPTS} \${YARN_TIMELINESERVER_OPTS}"
EOF

cat >>"${HADOOP_CONF_DIR}/mapred-env.sh" <<EOF
# Lessen stop-the-world collection:
GC_OPTS="-XX:+UseConcMarkSweepGC"
# Log GC details to stdout, these will be in diagnostic tarballs.
GC_LOGGING_OPTS="${GC_LOG_OPTS}"
export HADOOP_JOB_HISTORYSERVER_OPTS="\${GC_OPTS} \${GC_LOGGING_OPTS} \${HADOOP_JOB_HISTORYSERVER_OPTS}"
EOF

DATA_DIRS_ARRAY=($(get_data_dirs))
MAPRED_DIRS="${DATA_DIRS_ARRAY[*]/%/hadoop/mapred}"
MAPRED_DIRS_ARRAY=(${MAPRED_DIRS})
MAPRED_LOCAL_DIRS="${DATA_DIRS_ARRAY[*]/%/hadoop/mapred/local}"
MAPRED_LOCAL_DIRS_ARRAY=(${MAPRED_LOCAL_DIRS})
YARN_DIRS="${DATA_DIRS_ARRAY[*]/%/hadoop/yarn}"
YARN_DIRS_ARRAY=(${YARN_DIRS})
NODEMANAGER_LOCAL_DIRS="$(get_yarn_nm_local_dirs)"
NODEMANAGER_LOCAL_DIRS_ARRAY=(${NODEMANAGER_LOCAL_DIRS})

mkdir -p "${MAPRED_LOCAL_DIRS_ARRAY[@]}" "${NODEMANAGER_LOCAL_DIRS_ARRAY[@]}"

chgrp hadoop -L -R /hadoop "${HADOOP_TMP_DIR}"
chown -L -R mapred:hadoop "${MAPRED_DIRS_ARRAY[@]}"
chown -L -R yarn:hadoop "${YARN_DIRS_ARRAY[@]}"
chmod g+rwx -R \
  /hadoop \
  "${MAPRED_DIRS_ARRAY[@]}" \
  "${YARN_DIRS_ARRAY[@]}"
chmod 777 -R ${HADOOP_TMP_DIR}

export MAPRED_LOCAL_DIRS="${MAPRED_LOCAL_DIRS// /,}"
export NODEMANAGER_LOCAL_DIRS="${NODEMANAGER_LOCAL_DIRS// /,}"

YARN_ENV_FILE=${YARN_CONF_DIR:-$HADOOP_CONF_DIR}/yarn-env.sh
if [[ -f ${YARN_ENV_FILE} ]]; then
  cat >>"${YARN_ENV_FILE}" <<EOF
YARN_RESOURCEMANAGER_OPTS="-Xmx${HADOOP_MR_MASTER_MEM_MB}m \
    \${YARN_RESOURCEMANAGER_OPTS}"
# Increase maximum YARN client heap
YARN_CLIENT_OPTS="-Xmx${HADOOP_CLIENT_MEM_MB}m \${YARN_CLIENT_OPTS}"
EOF
fi

if [[ "${NUM_MASTERS}" -gt 1 ]]; then
  readonly CORE_TEMPLATE="core-ha-template.xml"
  readonly YARN_TEMPLATE="yarn-ha-template.xml"
else
  readonly CORE_TEMPLATE="core-template.xml"
  readonly YARN_TEMPLATE="yarn-template.xml"
fi

function merge_hadoop_configurations() {
  local -r config=$1
  local -r template=$2
  bdconfig merge_configurations \
    --configuration_file "${HADOOP_CONF_DIR}/${config}" \
    --source_configuration_file "${template}" \
    --resolve_environment_variables \
    --create_if_absent \
    --clobber
}

function set_hadoop_property() {
  local -r config=$1
  local -r property_name=$2
  local -r property_value=$3
  bdconfig set_property \
    --configuration_file "${HADOOP_CONF_DIR}/${config}" \
    --name "${property_name}" \
    --value "${property_value}" \
    --clobber
}

# enable strict container memory enforcement through cgroups for yarn.
function config_yarn_memory_control() {
  mkdir -p /sys/fs/cgroup/cpu/hadoop-yarn
  chown -R yarn /sys/fs/cgroup/cpu/hadoop-yarn

  mkdir -p /sys/fs/cgroup/memory/hadoop-yarn
  chown -R yarn /sys/fs/cgroup/memory/hadoop-yarn

  # CGroups is a Linux kernel feature and is exposed via the LinuxContainerExecutor.
  set_hadoop_property 'yarn-site.xml' \
    'yarn.nodemanager.container-executor.class' \
    'org.apache.hadoop.yarn.server.nodemanager.LinuxContainerExecutor'
  # If you wish to use CGroups, the resource-handler-class must be set to CGroupsLCEResourceHandler.
  set_hadoop_property 'yarn-site.xml' \
    'yarn.nodemanager.linux-container-executor.resources-handler.class' \
    'org.apache.hadoop.yarn.server.nodemanager.util.CgroupsLCEResourcesHandler'
  # The cgroups hierarchy under which to place YARN processes.
  set_hadoop_property 'yarn-site.xml' \
    'yarn.nodemanager.linux-container-executor.cgroups.hierarchy' \
    '/hadoop-yarn'
  # Whether the LCE should attempt to mount cgroups if not found.
  set_hadoop_property 'yarn-site.xml' \
    'yarn.nodemanager.linux-container-executor.cgroups.mount' \
    'false'
  # The Unix group of the NodeManager. It should match the setting in “container-executor.cfg”.
  set_hadoop_property 'yarn-site.xml' \
    'yarn.nodemanager.linux-container-executor.group' \
    'hadoop'
  set_hadoop_property 'yarn-site.xml' \
    'yarn.nodemanager.resource.memory.enabled' \
    'true'
  # Configured to run jobs as the user submitting the job
  set_hadoop_property 'yarn-site.xml' \
    'yarn.nodemanager.linux-container-executor.nonsecure-mode.local-user' \
    'yarn'

  local -r config_file="${HADOOP_CONF_DIR}/container-executor.cfg"
  cat >"${config_file}" <<EOF
yarn.nodemanager.linux-container-executor.group=hadoop
banned.users=hdfs,mapred,bin
allowed.system.users=yarn
min.user.id=1000#Prevent other super-users
EOF


  # container-executor need same owner and permission.
  chown root:hadoop /usr/lib/hadoop-yarn/bin/container-executor
  chmod 6050 /usr/lib/hadoop-yarn/bin/container-executor

  # enable strict container memory enforcement through cgroups
  set_hadoop_property 'yarn-site.xml' \
    'yarn.nodemanager.pmem-check-enabled' \
    'true'
  set_hadoop_property 'yarn-site.xml' \
    'yarn.nodemanager.resource.memory.enforced' \
    'true'
}

# Set nofile and nproc limit for 'yarn' user.
function set_limits_for_yarn()  {
  local -r nofile="$1"
  local -r nproc="$2"
  cat >/etc/security/limits.d/z-yarn.conf << EOF
yarn   - nofile ${nofile}
yarn   - nproc ${nproc}
EOF
  echo "${nproc}" > /proc/sys/kernel/pid_max
}

merge_hadoop_configurations 'core-site.xml' "${CORE_TEMPLATE}"
merge_hadoop_configurations 'mapred-site.xml' 'mapred-template.xml'
merge_hadoop_configurations 'yarn-site.xml' "${YARN_TEMPLATE}"
merge_hadoop_configurations \
  'capacity-scheduler.xml' 'capacity-scheduler-template.xml'

# Until https://issues.apache.org/jira/browse/MAPREDUCE-5653 is resolved,
# explicitly create a distcp-default.xml which matches mapreduce.*.memory.mb
# settings placed in mapred-site.xml, otherwise child memory opts will be
# greater than requested container sizes, causing distcp to occasionally
# fail with YARN-enforced memory limits.
merge_hadoop_configurations 'distcp-default.xml' 'distcp-template.xml'

set_hadoop_property 'mapred-site.xml' 'mapreduce.application.classpath' \
  "\$HADOOP_MAPRED_HOME/share/hadoop/mapreduce/*,
    \$HADOOP_MAPRED_HOME/share/hadoop/mapreduce/lib/*,
    ${DATAPROC_COMMON_LIB_DIR}/*"

# b/287074566
set_hadoop_property 'mapred-site.xml' 'mapreduce.client.submit.file.replication' '2'

set_hadoop_property 'yarn-site.xml' 'yarn.application.classpath' \
  "\$HADOOP_CONF_DIR,\$HADOOP_COMMON_HOME/*,\$HADOOP_COMMON_HOME/lib/*,
    \$HADOOP_HDFS_HOME/*,\$HADOOP_HDFS_HOME/lib/*,\$HADOOP_MAPRED_HOME/*,
    \$HADOOP_MAPRED_HOME/lib/*,\$HADOOP_YARN_HOME/*,\$HADOOP_YARN_HOME/lib/*,
    ${DATAPROC_COMMON_LIB_DIR}/*"

if is_version_at_least "${DATAPROC_IMAGE_VERSION}" '2.0'; then
  cat >>"${HADOOP_CONF_DIR}/yarn-env.sh" <<EOF
# Add Hadoop native libraries
export LD_LIBRARY_PATH=\${LD_LIBRARY_PATH}:/usr/lib/hadoop/lib/native
export HADOOP_MAPRED_HOME="/usr/lib/hadoop-mapreduce"
EOF

  set_hadoop_property 'yarn-site.xml' 'yarn.nodemanager.env-whitelist' \
    'PATH,JAVA_HOME,HADOOP_COMMON_HOME,HADOOP_HDFS_HOME,HADOOP_CONF_DIR,CLASSPATH_PREPEND_DISTCACHE,HADOOP_YARN_HOME,HADOOP_MAPRED_HOME,LD_LIBRARY_PATH,LANG,TZ'
  # Decrease resource manager time to mark a nodemanager LOST.
  # The default value is 10 minutes - decreasing the value helps the
  # resource manager recognize that a PVM has been preempted more quickly.
  # The resource manager will check for expiration every 1/3 of the expiry interval.
  #
  # Some considerations on the expiry interval:
  # 1. The expiry interval must be >= the heartbeat interval
  # (yarn.resourcemanager.nodemanagers.heartbeat-interval-ms, which is 1 second
  # by default).
  # 2. yarn.resourcemanager.nm-tokens.master-key-rolling-interval-secs and
  # yarn.resourcemanager.container-tokens.master-key-rolling-interval-secs
  # must be at least 3x the expiry interval.
  # 3. setting expiry interval too short can cause nodes status thrashing as a busy
  # cluster (eg GC, network glitch) may miss NM heartbeat and prematurely mark a NM as LOST.
  set_hadoop_property 'yarn-site.xml' 'yarn.nm.liveness-monitor.expiry-interval-ms' \
    '120000'
fi

set_hadoop_property 'yarn-site.xml' 'yarn.log-aggregation-enable' 'false'

ZK_QUORUM="${MASTER_HOSTNAME_0}:2181,${MASTER_HOSTNAME_1}:2181,${MASTER_HOSTNAME_2}:2181"

if [[ "${NUM_MASTERS}" -gt 1 ]]; then
  set_hadoop_property 'core-site.xml' 'hadoop.zk.address' "${ZK_QUORUM}"
fi

# When closing a file, allow 1 block to be committed (not completed).
# This prevents mapreduce jobs from failing during hdfs decommission when
# trying to write the final history event at the end of the job, resulting in
# an "Unknown Job" error.
set_hadoop_property 'hdfs-site.xml' \
  'dfs.namenode.file.close.num-committed-allowed' '1'

############### Configure ATS ################
# Enable cross origin requests to the timeline server. Also retain the default
# StaticUserWebFilter so that users don't need to authenticate to the webui,
# and instead can make POST/PUT requests as dr.who, the anonymous user.
set_hadoop_property 'core-site.xml' 'hadoop.http.filter.initializers' \
  'org.apache.hadoop.security.HttpCrossOriginFilterInitializer,org.apache.hadoop.http.lib.StaticUserWebFilter'
set_hadoop_property 'yarn-site.xml' \
  'yarn.resourcemanager.webapp.cross-origin.enabled' 'true'
set_hadoop_property 'yarn-site.xml' \
  'yarn.timeline-service.http-cross-origin.enabled' 'true'

# Point clients to the timeline server. It always runs on a single master.
set_hadoop_property 'yarn-site.xml' 'yarn.timeline-service.enabled' 'true'
set_hadoop_property 'yarn-site.xml' \
  'yarn.timeline-service.hostname' "${MASTER_HOSTNAME_0}"
set_hadoop_property 'yarn-site.xml' \
  'yarn.timeline-service.bind-host' '0.0.0.0'

# Publish YARN's own metrics into the timeline server
set_hadoop_property 'yarn-site.xml' \
  'yarn.resourcemanager.system-metrics-publisher.enabled' 'true'

# Allow apps to publish their own metrics into ATS
set_hadoop_property 'yarn-site.xml' \
  'yarn.timeline-service.generic-application-history.enabled' 'true'

# Disabling this option will allow existing and new clients to detect an IP
# change and continue to locate the new host's token.
if is_version_at_least "${DATAPROC_IMAGE_VERSION}" '2.1'; then
  set_hadoop_property 'core-site.xml' \
    'hadoop.security.token.service.use_ip' 'false'
fi

# Configure capacity scheduler for non-preemptible node label
AM_ON_PRIMARY_WORKER_ENABLED=$(get_dataproc_property am.primary_only)
DATAPROC_DATANODE_ENABLED=$(get_metadata_datanode_enabled || true)

if [[ "${AM_ON_PRIMARY_WORKER_ENABLED}" == 'true' ]]; then
  set_hadoop_property 'yarn-site.xml' 'yarn.node-labels.enabled' 'true'
  CONFIGBUCKET=$(get_metadata_bucket)
  CLUSTER_UUID=$(get_metadata_cluster_uuid)
  # GCS doesn't support append.
  set_hadoop_property 'yarn-site.xml' \
    'yarn.node-labels.fs-store.impl.class' 'org.apache.hadoop.yarn.nodelabels.NonAppendableFSNodeLabelStore'
  set_hadoop_property 'yarn-site.xml' \
    'yarn.node-labels.fs-store.root-dir' "gs://${CONFIGBUCKET}/google-cloud-dataproc-metainfo/${CLUSTER_UUID}/node-labels"
  set_hadoop_property 'yarn-site.xml' \
    'yarn.node-labels.configuration-type' 'distributed'

  set_hadoop_property 'capacity-scheduler.xml' \
    'yarn.scheduler.capacity.root.accessible-node-labels' 'non-preemptible'
  set_hadoop_property 'capacity-scheduler.xml' \
    'yarn.scheduler.capacity.root.accessible-node-labels.non-preemptible.capacity' \
    '100'
  set_hadoop_property 'capacity-scheduler.xml' \
    'yarn.scheduler.capacity.root.default.accessible-node-labels' \
    'non-preemptible'
  set_hadoop_property 'capacity-scheduler.xml' \
    'yarn.scheduler.capacity.root.default.accessible-node-labels.non-preemptible.capacity' \
    '100'

  if [[ "${DATAPROC_DATANODE_ENABLED}" == 'true' ]]; then
    set_hadoop_property 'yarn-site.xml' \
      'yarn.nodemanager.node-labels.provider' 'config'
    set_hadoop_property 'yarn-site.xml' \
      'yarn.nodemanager.node-labels.provider.configured-node-partition' \
      'non-preemptible'
  fi
fi

# Determine where the cluster membership files should be.
INCLUDE_PATH=$(create_or_validate_include_file_path)
EXCLUDE_PATH=$(create_or_validate_exclude_file_path)

# Tell the YARN resource manager where the membership files are.
# NB: the HDFS namenode should point to the same files if HDFS is activated.
set_hadoop_property 'yarn-site.xml' 'yarn.resourcemanager.nodes.include-path' \
  "${INCLUDE_PATH}"
set_hadoop_property 'yarn-site.xml' 'yarn.resourcemanager.nodes.exclude-path' \
  "${EXCLUDE_PATH}"

# Configure capacity scheduler for driver-pool node label
MASTER_RUN_DRIVER_LOCATION=$(/usr/share/google/get_metadata_value \
    'attributes/master-run-driver-location' || echo "")

if [[ "${MASTER_RUN_DRIVER_LOCATION}" == 'YARN' ]]; then
  set_hadoop_property 'yarn-site.xml' 'yarn.node-labels.enabled' 'true'
  CONFIGBUCKET=$(get_metadata_bucket)
  CLUSTER_UUID=$(get_metadata_cluster_uuid)
  # GCS doesn't support append.
  set_hadoop_property 'yarn-site.xml' \
    'yarn.node-labels.fs-store.impl.class' 'org.apache.hadoop.yarn.nodelabels.NonAppendableFSNodeLabelStore'
  set_hadoop_property 'yarn-site.xml' \
    'yarn.node-labels.fs-store.root-dir' "gs://${CONFIGBUCKET}/google-cloud-dataproc-metainfo/${CLUSTER_UUID}/node-labels"
  set_hadoop_property 'yarn-site.xml' \
    'yarn.node-labels.configuration-type' 'distributed'


  set_hadoop_property 'capacity-scheduler.xml' \
  'yarn.scheduler.capacity.root.queues' 'dataproc-driverpool-driver-queue,default'
  # Maximum percent of resources in the cluster which can be used to run
  # application masters. 1 means 100%, we are running exclusively app masters in
  # the driver pool queue so we want to make sure they can use 100% of resources available.
  # This is not appropriate to set outside this specific queue, because our goal
  # is to pin drivers to specific nodes.
  set_hadoop_property 'capacity-scheduler.xml' \
    'yarn.scheduler.capacity.root.dataproc-driverpool-driver-queue.maximum-am-resource-percent' \
    '1'

  # Configure a default queue for application containers and a driver pool queue
  # for drivers. The sum of capacities for the child queues under root must not
  # exceed 100, so set each value to 50.
  set_hadoop_property 'capacity-scheduler.xml' \
    'yarn.scheduler.capacity.root.default.capacity' '50'
  set_hadoop_property 'capacity-scheduler.xml' \
    'yarn.scheduler.capacity.root.dataproc-driverpool-driver-queue.capacity' '50'

  # Set the queues' maximum capacities to 100 to allow bursting to full
  # utilization if idle.
  set_hadoop_property 'capacity-scheduler.xml' \
    'yarn.scheduler.capacity.root.default.maximum-capacity' '100'
  set_hadoop_property 'capacity-scheduler.xml' \
    'yarn.scheduler.capacity.root.dataproc-driverpool-driver-queue.maximum-capacity' '100'

  # However, utilization by a single user is still constrained to the value
  # defined for capacity (not maximum-capacity). Set user-limit-factor to 2x
  # capacity, allowing a single user to achieve full resource utilization.
  set_hadoop_property 'capacity-scheduler.xml' \
    'yarn.scheduler.capacity.root.default.user-limit-factor' '2'
  set_hadoop_property 'capacity-scheduler.xml' \
    'yarn.scheduler.capacity.root.dataproc-driverpool-driver-queue.user-limit-factor' '2'

  # yarn.scheduler.capacity.<queue-path>.accessible-node-labels.<label>.capacity
  # defines the queue capacity for accessing nodes that belong to partition
  # “label”. The default is 0.
  set_hadoop_property 'capacity-scheduler.xml' \
    'yarn.scheduler.capacity.root.accessible-node-labels.dataproc-driverpool.capacity' \
    '100'
  # yarn.scheduler.capacity.<queue-path>.accessible-node-labels defines the node
  # labels that the queue can access. ” ” (a blank space) means that the queue
  # can access only the Default partition.
  set_hadoop_property 'capacity-scheduler.xml' \
    'yarn.scheduler.capacity.root.default.accessible-node-labels' \
    ' '
  set_hadoop_property 'capacity-scheduler.xml' \
    'yarn.scheduler.capacity.root.dataproc-driverpool-driver-queue.accessible-node-labels' \
    'dataproc-driverpool, '
  set_hadoop_property 'capacity-scheduler.xml' \
    'yarn.scheduler.capacity.root.dataproc-driverpool-driver-queue.accessible-node-labels.dataproc-driverpool.capacity' \
    '100'
  # yarn.scheduler.capacity.<queue-path>.default-node-label-expression defines
  # the queue’s default node label expression. If applications submitted to the
  # queue without specifying node label in their resource requests, it will use
  # default node label
  set_hadoop_property 'capacity-scheduler.xml' \
    'yarn.scheduler.capacity.root.dataproc-driverpool-driver-queue.default-node-label-expression' \
    'dataproc-driverpool'

  NODE_GROUP_ROLES=$(/usr/share/google/get_metadata_value attributes/node-group-roles ||
    echo "")
  DRIVER_POOL_ROLE='DRIVER'

  if [[ "${NODE_GROUP_ROLES}" != "" && ",${NODE_GROUP_ROLES}," = *",${DRIVER_POOL_ROLE},"* ]]; then
    set_hadoop_property 'yarn-site.xml' \
      'yarn.nodemanager.node-labels.provider' 'config'
    set_hadoop_property 'yarn-site.xml' \
      'yarn.nodemanager.node-labels.provider.configured-node-partition' \
      'dataproc-driverpool'
    # Cgroups does not support in yarn in earlier image versions
    if is_version_at_least "${DATAPROC_IMAGE_VERSION}" '2.0'; then
      config_yarn_memory_control
    fi
  fi
  # Increase the max number of tasks allowed for 'yarn' user
  # Increase the system-wide pid max up to 4,194,304
  # (the maximum allowed in 64-bit Linux),
  # consequently the pid_max in cgroup for the yarn user is now 1,384,119.
  set_limits_for_yarn 1000000 4194304
fi

# Mixins for scaling simplification.
if [[ $NUM_MASTERS -gt 1 ]]; then
  readonly YARN_SIMPLIFICATION_MIXINS="yarn-simplification-ha-mixins.xml"
else
  readonly YARN_SIMPLIFICATION_MIXINS="yarn-simplification-mixins.xml"
fi
merge_hadoop_configurations \
  'yarn-site.xml' "${BDUTIL_DIR}/conf/${YARN_SIMPLIFICATION_MIXINS}"


function join_by_char() {
  local IFS="$1"
  shift
  echo "$*"
}

# We use "Bind mounting" way to let Docker container access /etc/passwd and /etc/group
function configure_docker_on_yarn() {
  echo "Configuring container-executor.cfg"

  local additional_registries
  additional_registries="$(get_dataproc_property docker.additional.registries)"

  local regional_artifact_registry
  regional_artifact_registry="$(get_metadata_dataproc_region)-docker.pkg.dev"

  local yarn_docker_image
  local yarn_docker_image_registry
  yarn_docker_image=$(get_dataproc_property_or_default yarn.docker.image '')
  yarn_docker_image_registry="${yarn_docker_image%%/*}"

  local trusted_registries="local,gcr.io,us.gcr.io,eu.gcr.io,asia.gcr.io,marketplace.gcr.io,staging-k8s.gcr.io,asia-docker.pkg.dev,europe-docker.pkg.dev,us-docker.pkg.dev,${regional_artifact_registry}"

  if [[ -n "${yarn_docker_image_registry}" ]]; then
    trusted_registries="${trusted_registries},${yarn_docker_image_registry}"
  fi

  if [[ -n "${additional_registries}" ]]; then
    trusted_registries="${trusted_registries},${additional_registries}"
  fi

  # Allow read-only mounting of password files, Java home, Conda home, Hadoop
  # dirs, Spark dirs, and Dataproc common lib dir.
  local -r ro_mounts_array=(
    /etc/passwd
    /etc/group
    /etc/hadoop/conf
    /etc/spark/conf
    /opt/conda
    /usr/lib/jvm
    /usr/lib/hadoop
    /usr/lib/hadoop-hdfs
    /usr/lib/hadoop-mapreduce
    /usr/lib/hadoop-yarn
    /usr/lib/spark
    /usr/local/share/google/dataproc/lib
  )
  local ro_mounts
  ro_mounts=$(join_by_char ',' "${ro_mounts_array[@]}")

  # Allow read-write mounting of data dirs.
  local -r rw_mounts_array=(
    "${DATA_DIRS_ARRAY[@]}"
    "${MAPRED_DIRS_ARRAY[@]}"
    "${MAPRED_LOCAL_DIRS_ARRAY[@]}"
    "${NODEMANAGER_LOCAL_DIRS_ARRAY[@]}"
    "${YARN_DIRS_ARRAY[@]}"
    /hadoop/spark
    /var/log/hadoop-yarn
    /var/log/hadoop-mapreduce
    /var/log/hadoop-hdfs
    /var/log/spark
  )
  local rw_mounts
  rw_mounts=$(join_by_char ',' "${rw_mounts_array[@]}")

  local -r config_file="${HADOOP_CONF_DIR}/container-executor.cfg"
  cat >"${config_file}" <<EOF
yarn.nodemanager.local-dirs="${NODEMANAGER_LOCAL_DIRS}"
yarn.nodemanager.log-dirs=/var/log/hadoop-yarn/userlogs
yarn.nodemanager.linux-container-executor.group=hadoop
banned.users=hdfs,mapred,bin
min.user.id=1000
allowed.system.users=yarn_docker_user,yarn,nobody

[docker]
module.enabled=true
docker.binary=/usr/bin/docker
docker.allowed.capabilities=ALL
docker.allowed.networks=bridge,host,none
docker.allowed.ro-mounts=${ro_mounts}
docker.allowed.rw-mounts=${rw_mounts}
docker.privileged-containers.enabled=false
docker.trusted.registries=${trusted_registries}
EOF

  set_hadoop_property "yarn-site.xml" 'yarn.nodemanager.container-executor.class' 'org.apache.hadoop.yarn.server.nodemanager.LinuxContainerExecutor'
  set_hadoop_property "yarn-site.xml" 'yarn.nodemanager.linux-container-executor.group' 'hadoop'
  set_hadoop_property "yarn-site.xml" 'yarn.nodemanager.linux-container-executor.nonsecure-mode.limit-users' 'false'
  set_hadoop_property "yarn-site.xml" 'yarn.nodemanager.runtime.linux.allowed-runtimes' 'default,docker'
  set_hadoop_property "yarn-site.xml" 'yarn.nodemanager.runtime.linux.type' 'docker'
  set_hadoop_property "yarn-site.xml" 'yarn.nodemanager.runtime.linux.docker.allowed-container-networks' 'host'
  set_hadoop_property "yarn-site.xml" 'yarn.nodemanager.runtime.linux.docker.default-container-network' 'host'
  set_hadoop_property "yarn-site.xml" 'yarn.nodemanager.runtime.linux.docker.privileged-containers.allowed' "false"
  set_hadoop_property "yarn-site.xml" 'yarn.nodemanager.runtime.linux.docker.privileged-containers.acl' ''
  set_hadoop_property "yarn-site.xml" 'yarn.nodemanager.runtime.linux.docker.capabilities' 'ALL'

  # The container-executor binary relies on POSIX sticky bits and required to be
  # owned by root:hadoop.
  chown root:hadoop /usr/lib/hadoop-yarn/bin/container-executor
  chmod 6050 /usr/lib/hadoop-yarn/bin/container-executor
  chown yarn /hadoop/yarn/nm-local-dir
}

# Configure Running Docker container on YARN feature. No Kerberos support.
# TODO: fix this for Rocky Linux
if ! is_component_selected 'kerberos'; then
  ENABLE_DOCKER_YARN="$(get_dataproc_property yarn.docker.enable)"
  if [[ -z "${ENABLE_DOCKER_YARN}" ]]; then
    # This is kept for backwards compatiblitiy.
    ENABLE_DOCKER_YARN="$(get_dataproc_property docker.yarn.enable)"
  fi
  if is_version_at_least "${DATAPROC_IMAGE_VERSION}" '2.0' &&
      is_component_selected 'docker-ce' &&
      [[ "${ENABLE_DOCKER_YARN}" == "true" ]]; then
    configure_docker_on_yarn
  fi
fi
