# Copyright 2020 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Generates the cryptographic key pair that can be used to securely
# send secrets to the VM. The public key is written to the serial port output
# of the second serial port, and the private key is written to a directory
# that only the root user can access.

readonly METADATA_HOST="metadata.google.internal"
readonly INSTANCE_METADATA_PATH="computeMetadata/v1/instance"
readonly IDENTITY_KEY="service-accounts/default/identity"
readonly GUEST_ATTRIBUTE_PATH="${INSTANCE_METADATA_PATH}/guest-attributes/dataproc-signed-keys"
readonly SERIAL_DEVICE="/dev/ttyS3"
readonly GENERATE_KEYS_SCRIPT="/usr/bin/google-dataproc-generate-cluster-keys.sh"
readonly TINKEY_BINARY_PATH="/usr/local/bin/tinkey"

cluster_keys_enabled=$(get_dataproc_property \
  dataproc.encryption.keygen.enabled)
cluster_keys_rotation_hours=$(get_dataproc_property_or_default \
  dataproc.encryption.keygen.rotation_hours 6)

cat >"${GENERATE_KEYS_SCRIPT}" <<EOF
#!/bin/bash
source "\${BDUTIL_DIR}/bdutil_helpers.sh"

# The create_rsa_key method generates a single RSA public/private key
# pair for the specified user with the specified key size.
#
# Usage:
#   create_rsa_key "\${USER}" "\${HOME}" "\${KEYSIZE}"
function create_rsa_key() {
  local -r owner="\${1}"
  local -r home="\${2}"
  local -r keysize="\${3}"
  local -r keydir="\${home}/.keys"
  local -r keyname="rsa\${keysize}"

  mkdir -p "\${keydir}"
  chown -R "\${owner}" "\${keydir}"
  chmod -R go-rwx "\${keydir}"

  HOME="\${home}" openssl genrsa -out "\${keydir}/\${keyname}.private.pem"  "\${keysize}"
  HOME="\${home}" openssl rsa -in "\${keydir}/\${keyname}.private.pem" -pubout -outform PEM \
    -out "\${keydir}/\${keyname}.public.pem"
}

# The create_ecies_key method generates a single ECIES public/private key
# pair for the specified user.
#
# Usage:
#   create_ecies_key "\${USER}" "\${HOME}"
function create_ecies_key() {
  local -r owner="\${1}"
  local -r home="\${2}"
  local -r keydir="\${home}/.keys"
  local -r keyname="ecies"

  mkdir -p "\${keydir}"
  chown -R "\${owner}" "\${keydir}"
  chmod -R go-rwx "\${keydir}"

  HOME="\${home}" "${TINKEY_BINARY_PATH}" create-keyset --out \
    "\${keydir}/\${keyname}.private.json" \
    --key-template ECIES_P256_HKDF_HMAC_SHA256_AES128_GCM

  HOME="\${home}" "${TINKEY_BINARY_PATH}" create-public-keyset --in \
    "\${keydir}/\${keyname}.private.json" \
    --out "\${keydir}/\${keyname}.public.json"
}

# The create_keys method generates a default set of RSA and ECIES public/private
# key pairs for a given user.
#
# For consistency, the set of keys generated here is intended to match
# those listed in this doc:
# https://cloud.google.com/kms/docs/algorithms#asymmetric_encryption_algorithms
#
# Usage:
#   create_keys "\${USER}" "\${HOME}"
function create_keys() {
  local -r owner="\${1}"
  local -r home="\${2}"

  create_rsa_key "\${owner}" "\${home}" 2048
  create_rsa_key "\${owner}" "\${home}" 3072
  create_rsa_key "\${owner}" "\${home}" 4096
  create_ecies_key "\${owner}" "\${home}"
  mkdir -p "\${home}/.historical_keys"
  cp -R "\${home}/.keys" "\${home}/.historical_keys/\$(date "+%Y%m%d%H%M")"
}

# The sign_key method generates a signature that can be used to
# verify that the given public key was generated on this specific VM instance.
#
# The resulting signature is a GCE VM Identity token
# (see https://cloud.google.com/compute/docs/instances/verifying-instance-identity)
# where the audience of the token is a sha256 hash of the key.
#
# Usage:
#   key_sig="\$(sign_key "\${key_id}" "\${key_contents}" "\${key_extension}")"
function sign_key() {
  # Generate a hash of the key which is what we will sign
  local -r key_id="\${1}"
  local key_contents="\${2}"
  local -r key_extension="\${3}"

  if [[ "\${key_extension}" == "json" ]]; then
    key_contents="\$(echo "\${key_contents}" | jq -r '."key"[0]."keyData"."value"')"
  fi

  local -r key_sum="\$(echo "\${key_contents}" | sha256sum)"

  # Trim the trailing hyphen and newline
  local -r key_hash="\${key_sum:0:64}"

  # Fetch a VM identity with the key hash as the audience
  local -r audience="\${key_id}/\${key_hash}"
  /usr/share/google/get_metadata_value \
    "${IDENTITY_KEY}?audience=\${audience}&format=full"
}

# The publish_keys method publishes public keys and corresponding signatures
# for all of the current public/private key pairs for a specific user.
#
# Usage:
#   publish_keys "\${USER}" "\${HOME}"
function publish_keys() {
  local -r owner="\${1}"
  local -r home="\${2}"
  local -r keydir="\${home}/.keys"

  for keyfile in "\${keydir}"/*.public.*; do
    local keyname="\${keyfile#\${keydir}/}"
    local keyname="\${keyname%.public.*}"
    local key_extension="\${keyfile#\${keydir}/*.public.}"
    local key_id="\${keyname}-\${owner}"
    local key_contents
    key_contents="\$(cat "\${keyfile}")"
    local key_sig
    key_sig="\$(sign_key "\${key_id}" "\${key_contents}" "\${key_extension}")"
    if [[ "\$?" == 0 ]]; then
      # If we successfully generated a signed key, then publish the public key
      # contents and signature to the VM guest attributes.
      #
      # Since the getGuestAttributes API does not return the VM ID, including
      # the signature allows the caller to verify that the attribute was set
      # by the specific instance they want without any additional API calls.
      retry_constant \
          curl -X PUT --data "\${key_contents}" \
          -H "Metadata-Flavor: Google" \
          "http://${METADATA_HOST}/${GUEST_ATTRIBUTE_PATH}/\${key_id}"
      retry_constant \
          curl -X PUT --data "\${key_sig}" \
          -H "Metadata-Flavor: Google" \
          "http://${METADATA_HOST}/${GUEST_ATTRIBUTE_PATH}/sig-\${key_id}"
      # Additionally write the signature to a serial port that only root has
      # write access to. This allows an external user to verify the key was
      # written by a superuser on the current VM.
      echo "\${key_id}:\${key_sig}" >> ${SERIAL_DEVICE}
      echo "" >> ${SERIAL_DEVICE}
    else
      # If the key generation for a specific key failed, then log that but
      # continue with generating the rest of the keys.
      loginfo "Failure generating the key \${key_id}"
    fi
  done
}

function mount_cluster_keys_ramdisk() {
  mount | grep "${DATAPROC_GENERATED_KEYS_DIR}"
  if [[ "\$?" != "0" ]]; then
    mkdir -p "${DATAPROC_GENERATED_KEYS_DIR}"
    mount -t tmpfs -o size=1g tmpfs "${DATAPROC_GENERATED_KEYS_DIR}"
  fi
  mkdir "${DATAPROC_GENERATED_KEYS_DIR}/users/"
  chmod a+rx "${DATAPROC_GENERATED_KEYS_DIR}"
  chmod a+rx "${DATAPROC_GENERATED_KEYS_DIR}/users/"
}

function create_keys_for_user() {
  local -r user="\$1"
  local -r keys_dir="${DATAPROC_GENERATED_KEYS_DIR}/users/\${user}"
  mkdir -p "\${keys_dir}"
  chown -R "\${user}:\${user}" "\${keys_dir}"
  chmod 700 "\${keys_dir}"
  create_keys "\${user}" "\${keys_dir}"
  publish_keys "\${user}" "\${keys_dir}"
}

mount_cluster_keys_ramdisk
users_with_keys="root dataproc"
for user in \${users_with_keys}; do
  create_keys_for_user "\${user}"
done
EOF
chmod u+rwx "${GENERATE_KEYS_SCRIPT}"

cat >/usr/lib/systemd/system/google-dataproc-generate-keys.service <<EOF
[Unit]
Description=Generate Dataproc Cluster Encryption Keys
[Service]
Type=simple
EnvironmentFile=/etc/environment
ExecStart=${GENERATE_KEYS_SCRIPT}
[Install]
WantedBy=multiuser.target
EOF

cat >/usr/lib/systemd/system/google-dataproc-generate-keys.timer <<EOF
[Unit]
Description=Generate Dataproc Cluster Encryption Keys Every ${cluster_keys_rotation_hours} Hours
After=network.target

[Timer]
OnBootSec=0
OnUnitActiveSec=${cluster_keys_rotation_hours}h
Unit=google-dataproc-generate-keys.service

[Install]
WantedBy=timers.target
EOF

if [[ "${cluster_keys_enabled}" == "true" ]]; then
  systemctl daemon-reload
  systemctl enable google-dataproc-generate-keys.service
  systemctl enable google-dataproc-generate-keys.timer
  # Run the key generation script manually once at startup to ensure that the
  # first set of keys is available as soon as possible.
  ${GENERATE_KEYS_SCRIPT} || true
  systemctl start google-dataproc-generate-keys.timer
else
  # The cluster keys property is not enabled, so revert
  # the locking-down of the fourth serial port.
  rm /etc/udev/rules.d/80-ttyS3.rules
  udevadm trigger
  chown root:dialout /dev/ttyS3
fi
