# Copyright 2020 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Generates the cryptographic key pair that can be used to securely
# send secrets to the VM. The public key is written to the serial port output
# of the second serial port, and the private key is written to a directory
# that only the root user can access.

readonly RUN_SCRIPT_PATH="/usr/local/share/google/dataproc/metadata-proxy.sh"
readonly POLLING_SCRIPT_PATH="/usr/local/share/google/dataproc/poll-metadata.sh"
readonly PROXY_SERVICE_CONF='/usr/lib/systemd/system/metadata-proxy.service'
readonly POLLING_SERVICE_CONF='/usr/lib/systemd/system/metadata-credentials-polling.service'
readonly METADATA_ADDRESS='169.254.169.254'
readonly METADATA_PORT='80'
readonly METADATA_PASSTHROUGH_PORT='987'
readonly METADATA_PROXY_PORT='988'
readonly TOKEN_SOURCE_METADATA='TOKEN FROM METADATA ATTRIBUTES'
readonly TOKEN_SOURCE_GCS='TOKEN FROM GCS'

CONFIGBUCKET=$(/usr/share/google/get_metadata_value \
  attributes/dataproc-bucket)
readonly CONFIGBUCKET
CLUSTER_UUID=$(/usr/share/google/get_metadata_value \
  attributes/dataproc-cluster-uuid)
readonly CLUSTER_UUID
INSTANCE_UUID=$(/usr/share/google/get_metadata_value id)
readonly INSTANCE_UUID
readonly TOKEN_METADATA_ATTRIBUTE="attributes/dataproc-injected-credentials"

METADATA_PROXY_ENABLED=$(get_dataproc_property dataproc.metadata.proxy.enabled)
readonly METADATA_PROXY_ENABLED
DATAPROC_EXCLUSIVE_USER=$(get_dataproc_property dataproc.exclusive.user)
readonly DATAPROC_EXCLUSIVE_USER
PERSONAL_AUTH_OVERRIDE_USER_DISPLAY_NAME=$(get_dataproc_property dataproc.personal-auth.override-user-display-name)
readonly PERSONAL_AUTH_OVERRIDE_USER_DISPLAY_NAME

function get_token_source() {
  local initial_credentials
  initial_credentials="$(/usr/share/google/get_metadata_value \
    "${TOKEN_METADATA_ATTRIBUTE}")"
  readonly initial_credentials
  if [[ -n "${initial_credentials}" ]]; then
    # The VM was created with the initial credential populated in the
    # metadata attributes, so we will always read credentials from the
    # metadata server.
    echo "${TOKEN_SOURCE_METADATA}"
  else
    # The VM was not created with the initial credential populated in the
    # metadata attributes, so fall back to using the legacy GCS credential
    # propagation.
    #
    # TODO: Remove this option once we turn off support
    # for GCS-based token propagation.
    echo "${TOKEN_SOURCE_GCS}"
  fi
}

function get_token_path() {
  if [[ "${METADATA_PROXY_TOKEN_SOURCE}" == "${TOKEN_SOURCE_METADATA}" ]]; then
    # The VM was created with the initial credential populated in the
    # metadata attributes, so we will always read credentials from the
    # metadata server.
    local -r token_path_on_disk="/root/.dataproc/access-token"
    mkdir -p "$(dirname "${token_path_on_disk}")"
    touch "${token_path_on_disk}"
    echo "${token_path_on_disk}"
  else
    # The VM was not created with the initial credential populated in the
    # metadata attributes, so fall back to using the legacy GCS credential
    # propagation.
    #
    # TODO: Remove this fallthrough once we turn off support
    # for GCS-based token propagation.
    echo "gs://${CONFIGBUCKET}/google-cloud-dataproc-metainfo/${CLUSTER_UUID}/${INSTANCE_UUID}/access-token"
  fi
}

function create_credentials_polling_service() {
  # Set up a service to continuously poll the metadata server for injected
  # credentials. We do this in two steps:
  #
  # 1. Issue a hanging get for credentials that will wait until the credentials
  #    are changed, and then return to us the updated credentials.
  # 2. Issue a non-blocking get that will fetch the current value of the
  #    credentials.
  #
  # It is important for the hanging get to happen first so that we avoid a
  # race condition where the do the initial get, the value changes, then we
  # issue the request to wait for a change and never see the change that
  # happened between the two calls.
  cat >"${POLLING_SCRIPT_PATH}" <<EOF
#!/bin/bash
while true; do
  (sleep 1 && /usr/share/google/get_metadata_value \
    "${TOKEN_METADATA_ATTRIBUTE}" > "${TOKEN_PATH}") &
  /usr/share/google/get_metadata_value \
    "${TOKEN_METADATA_ATTRIBUTE}?wait_for_change=true" > "${TOKEN_PATH}"
done
EOF
  chmod u+rwx "${POLLING_SCRIPT_PATH}"

  cat >"${POLLING_SERVICE_CONF}" <<EOF
[Unit]
Description=Polling Injected Access Tokens From Metadata Attributes
[Service]
Type=simple
ExecStart="${POLLING_SCRIPT_PATH}"
Restart=on-failure
[Install]
WantedBy=multi-user.target
EOF
  chmod a+rw "${POLLING_SERVICE_CONF}"
}

function get_expected_username() {
  if [[ -n "${PERSONAL_AUTH_OVERRIDE_USER_DISPLAY_NAME}" ]]; then
    echo "${PERSONAL_AUTH_OVERRIDE_USER_DISPLAY_NAME}"
  elif [[ -n "${DATAPROC_EXCLUSIVE_USER}" ]]; then
    echo "${DATAPROC_EXCLUSIVE_USER}"
  else
    echo "<unknown>"
  fi
}

function configure_java_util_logging() {
  mkdir -p /etc/dataproc
  cat >/etc/dataproc/metadata-proxy.logging.properties <<'EOF'
.level=INFO

handlers=java.util.logging.FileHandler

java.util.logging.FileHandler.level=ALL
java.util.logging.FileHandler.formatter=java.util.logging.SimpleFormatter
# Limit each log file to 10MB.
java.util.logging.FileHandler.limit=10485760
java.util.logging.FileHandler.count=10
java.util.logging.FileHandler.append=true
java.util.logging.FileHandler.pattern=/var/log/metadata-proxy.%g.log

java.util.logging.SimpleFormatter.format=%1$tb %1$td, %1$tY %1$tl:%1$tM:%1$tS %1$Tp %2$s %4$s: %5$s%6$s%n
EOF
}

function create_proxy_service() {
  local -r agent_jar="/usr/local/share/google/dataproc/dataproc-agent.jar"
  local hadoop_classpath
  if [[ "${DATAPROC_IMAGE_TYPE}" == "serverless-spark" ]]; then
    hadoop_classpath="/usr/lib/spark/jars/*"
  else
    hadoop_classpath="$(hadoop classpath)"
  fi
  local max_heap=""
  if [[ "${DATAPROC_IMAGE_TYPE}" == "serverless-spark" ]]; then
    # Set max heap size for the metadata proxy to make sure we don't
    # unexpectedly overrun. 216 MB is based on the total memory usage of
    # the proxy on DPGCE personal auth clusters
    max_heap="-Xmx216m"
  fi
  cat >"${RUN_SCRIPT_PATH}" <<EOF
#!/bin/bash
/usr/bin/java -cp "${agent_jar}:${hadoop_classpath}" ${max_heap} \
  -Djava.util.logging.config.file=/etc/dataproc/metadata-proxy.logging.properties \
  com.google.cloud.hadoop.services.agent.metadata.StandaloneMetadataProxyMain \
  file "${METADATA_PROXY_PORT}" "${TOKEN_PATH}" \
  \${DATAPROC_GENERATED_KEYS_DIR}/users/root/ \
  "${METADATA_ADDRESS}:${METADATA_PASSTHROUGH_PORT}" \
  "$(get_expected_username)"
EOF
  chmod u+rwx "${RUN_SCRIPT_PATH}"

  cat >"${PROXY_SERVICE_CONF}" <<EOF
[Unit]
Description=Access Token Relay
[Service]
Type=simple
EnvironmentFile=/etc/environment
ExecStartPre="${IPTABLES_RESTORE_SCRIPT}"
ExecStart="${RUN_SCRIPT_PATH}"
Restart=on-failure
[Install]
WantedBy=multi-user.target
EOF
  chmod a+rw "${PROXY_SERVICE_CONF}"
}

function create_metadata_routes() {
  iptables -t nat -I OUTPUT -p tcp -d "${METADATA_ADDRESS}" \
    --dport "${METADATA_PASSTHROUGH_PORT}" -j DNAT \
    --to-destination "${METADATA_ADDRESS}:${METADATA_PORT}"
  iptables -t nat -I OUTPUT -p tcp -d "${METADATA_ADDRESS}" \
    --dport "${METADATA_PORT}" \
    -m owner --uid-owner 1-65535 -j DNAT \
    --to-destination "127.0.0.1:${METADATA_PROXY_PORT}"
  save_iptables
}

function start_metadata_proxy_and_credentials_polling() {
  systemctl daemon-reload
  systemctl enable metadata-proxy
  systemctl enable metadata-credentials-polling
  systemctl restart metadata-proxy
  systemctl restart metadata-credentials-polling
  systemctl status -l metadata-proxy
  systemctl status -l metadata-credentials-polling
}

if [[ "${METADATA_PROXY_ENABLED}" == "true" ]]; then
  METADATA_PROXY_TOKEN_SOURCE="$(get_token_source)"
  readonly METADATA_PROXY_TOKEN_SOURCE
  TOKEN_PATH="$(get_token_path)"
  readonly TOKEN_PATH

  configure_java_util_logging
  create_proxy_service
  create_credentials_polling_service
  create_metadata_routes
  start_metadata_proxy_and_credentials_polling
fi
