# Copyright 2020 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
set -euxo pipefail

readonly MULTI_USER_METADATA_ADDRESS='169.254.169.254'
readonly MULTI_USER_METADATA_PORT='80'
readonly MULTI_USER_METADATA_PASSTHROUGH_PORT='8081'

readonly METADATA_PROXY_PORT_BASE='801'

function create_proxy_service() {
  local -r port_to_sa="$1"
  local -r run_script_path="/usr/local/share/google/dataproc/metadata-proxy-multi-user.sh"
  local -r service_conf="/usr/lib/systemd/system/metadata-proxy-multi-user.service"
  local -r cloud_platform_scope="https://www.googleapis.com/auth/cloud-platform"
  local max_heap=""
  if [[ "${DATAPROC_IMAGE_TYPE}" == "serverless-spark" ]]; then
    # Set max heap size for the metadata proxy to make sure we don't
    # unexpectedly overrun. 216 MB is based on the total memory usage of
    # the proxy on DPGCE personal auth clusters
    max_heap="-Xmx216m"
  fi
  cat >"${run_script_path}" <<EOF
#!/bin/bash
/usr/bin/java -cp /usr/local/share/google/dataproc/dataproc-agent.jar ${max_heap} \
  -Djava.util.logging.config.file=/etc/dataproc/metadata-proxy-multi-user.logging.properties \
  com.google.cloud.hadoop.services.agent.metadata.StandaloneMetadataProxyMain \
  impersonate \
  "${MULTI_USER_METADATA_ADDRESS}:${MULTI_USER_METADATA_PASSTHROUGH_PORT}" \
  "${cloud_platform_scope}" \
  "${port_to_sa}"
EOF
  chmod u+rwx "${run_script_path}"

  cat >"${service_conf}" <<EOF
[Unit]
Description=Access Token Relay for Secure Multi-tenancy
[Service]
Type=simple
ExecStart="${run_script_path}"
Restart=on-failure
[Install]
WantedBy=multi-user.target
EOF
  chmod a+rw "${service_conf}"
}

function start_metadata_proxy() {
  systemctl daemon-reload
  systemctl enable metadata-proxy-multi-user
  systemctl start metadata-proxy-multi-user
  # Instead of checking the service status through `systemctl status ...`,
  # check the file which marks the successful start of not only the service,
  # but also all the threads.
  retry_constant_short test -f /var/lib/google/dataproc/metadata-proxy-multi-user-started ||
    err "metadata-proxy-multi-user service failed to start."
}

function create_metadata_routes_overall() {
  # Allow root to pass through to the real metadata server through the
  # pass-through port
  iptables -t nat -I OUTPUT -p tcp -d "${MULTI_USER_METADATA_ADDRESS}" \
    --dport "${MULTI_USER_METADATA_PASSTHROUGH_PORT}" \
    -m owner --uid-owner 0 \
    -j DNAT \
    --to-destination "${MULTI_USER_METADATA_ADDRESS}:${MULTI_USER_METADATA_PORT}"
  # Block any other user's access to the pass-through port, thus blocking access
  # to the real metadata server through the pass-through port.
  iptables -A OUTPUT -p tcp -d "${MULTI_USER_METADATA_ADDRESS}" \
    --dport "${MULTI_USER_METADATA_PASSTHROUGH_PORT}" -j REJECT
}

function create_metadata_routes_for_user() {
  local -r uid="$1"
  local -r port="$2"

  # Allow user with specific UID to access its own assigned metadata proxy
  iptables -t nat -I OUTPUT -p tcp -d "${MULTI_USER_METADATA_ADDRESS}" \
    --dport "${MULTI_USER_METADATA_PORT}" \
    -m owner --uid-owner "${uid}" -j DNAT \
    --to-destination "127.0.0.1:${port}"
  # Block any other user's access to this metadata proxy port, only allow root
  # and the assigned user
  iptables -I OUTPUT -p tcp -d 127.0.0.1 --dport "${port}" \
    -m owner --uid-owner 0 \
    -j ACCEPT
  iptables -I OUTPUT -p tcp -d 127.0.0.1 --dport "${port}" \
    -m owner --uid-owner "${uid}" \
    -j ACCEPT
  iptables -A OUTPUT -p tcp -d 127.0.0.1 --dport "${port}" -j REJECT
}

function err() {
  echo "[$(date +'%Y-%m-%dT%H:%M:%S%z')]: $*" >&2
  return 1
}

function get_port_info() {
  local -r port=$1
  ss -a "( sport = :${port} )" | sed '1d'
}

function port_is_free() {
  local -r port=$1
  local ss_lines
  ss_lines=$(get_port_info "${port}" | wc -l)
  return $((ss_lines > 0))
}

function configure_java_util_logging() {
  mkdir -p /etc/dataproc
  cat >/etc/dataproc/metadata-proxy-multi-user.logging.properties <<'EOF'
.level=INFO

handlers=java.util.logging.FileHandler

java.util.logging.FileHandler.level=ALL
java.util.logging.FileHandler.formatter=java.util.logging.SimpleFormatter
# Limit each log file to 10MB.
java.util.logging.FileHandler.limit=10485760
java.util.logging.FileHandler.count=10
java.util.logging.FileHandler.append=true
java.util.logging.FileHandler.pattern=/var/log/metadata-proxy-multi-user.%g.log

java.util.logging.SimpleFormatter.format=%1$tb %1$td, %1$tY %1$tl:%1$tM:%1$tS %1$Tp %2$s %4$s: %5$s%6$s%n
EOF
}

function configure_multi_user_metadata_proxy() {
  create_metadata_routes_overall

  local user_mappings
  user_mappings=$(get_dataproc_property \
    dataproc.beta.secure.multi-tenancy.user.mapping)
  if [[ -z "${user_mappings}" ]]; then
    user_mappings=$(get_dataproc_property internal.spark.byosa.user.mapping)
  fi
  local -a user_list
  IFS=',' read -r -a user_list <<<"${user_mappings}"

  # Usable port range is [METADATA_PROXY_PORT_BASE, 1024)
  local -r len="${#user_list[@]}"
  if (( "${len}" > 1024 - "${METADATA_PROXY_PORT_BASE}" )); then
    err "Not enough privileged ports for metadata proxies."
  fi

  local username
  local uid
  local port=${METADATA_PROXY_PORT_BASE}
  local -a port_to_sa

  for pair in "${user_list[@]}"; do
    IFS=':' read -r user sa <<< "${pair}"
    username="$(echo "${user}" | cut -d@ -f1)"
    # will fail to get UID if user does not exist, thus failing the whole
    # script.
    uid=$(id -u "${username}")
    while ! port_is_free "${port}"; do
      port=$((port+1))
    done
    if ((port > 1023)); then
      err "Cannot run multi_user_metadata_proxy on unprivileged ports."
    fi
    port_to_sa+=("${port}:${sa}")
    create_metadata_routes_for_user "${uid}" "${port}"
    port=$((port+1))
  done
  save_iptables

  configure_java_util_logging
  local port_to_sa_str
  port_to_sa_str="$(IFS=',' ; echo "${port_to_sa[*]}")"
  create_proxy_service "${port_to_sa_str}"
  start_metadata_proxy
}

configure_multi_user_metadata_proxy
