#!/usr/bin/env bash

# Configure stackdriver monitoring :
# 1. Moving the collectd plugin configuration files for metric sources enabled
#    from bdutil/conf to the collectd path /opt/stackdriver/collectd/etc/collect.d/
# 2. Enabling jmx for the metric sources using jmx for metric collection.
# 3. Running collectd-plugin-generator if metrics are overridden for metric sources.

set -eu

source "${BDUTIL_DIR}/bdutil_metadata.sh"
source "${BDUTIL_DIR}/bdutil_properties.sh"
source "${BDUTIL_DIR}/bdutil_metrics.sh"
source "${BDUTIL_DIR}/bdutil_misc.sh"
source "${BDUTIL_DIR}/configure_containerised_agents.sh"

readonly DATAPROC_ETC_DIR='/etc/google-dataproc'
readonly HADOOP_CONF_DIR='/etc/hadoop/conf'
readonly HIVE_CONF_DIR="/etc/hive/conf"
readonly REGION_NAME="$(get_metadata_dataproc_region)"
readonly ROLE=$(get_metadata_role)
readonly SPARK_CONF_DIR="/etc/spark/conf"
readonly COLLECTD_S8S_PLUGIN_DIR="${BDUTIL_DIR}/conf/collectd/s8s"
readonly FLINK_CONF='/usr/lib/flink/conf'
YARN_QUEUES="$(get_property_in_xml ${HADOOP_CONF_DIR}/capacity-scheduler.xml 'yarn.scheduler.capacity.root.queues' || true)"
readonly YARN_QUEUES
MASTER_RUN_DRIVER_LOCATION=$(/usr/share/google/get_metadata_value \
        'attributes/master-run-driver-location' || echo "")
readonly MASTER_RUN_DRIVER_LOCATION

STACKDRIVER_MONITORING_JMX_ENABLED=$(get_dataproc_property \
  dataproc.monitoring.stackdriver.jmx.enable)

function setup_jmx_sources() {
  if [[ "${STACKDRIVER_MONITORING_JMX_ENABLED}" != 'false' ]]; then
    if is_metric_source_enabled hadoop || is_metric_source_enabled "${YARN_METRIC_SOURCE}" ; then
      open_jmx_port_for_yarn
    fi

    if is_metric_source_enabled hadoop || is_metric_source_enabled "${HDFS_METRIC_SOURCE}" ; then
      open_jmx_port_for_hdfs
    fi

    if is_metric_source_enabled "${SPARK_HISTORY_SERVER_METRIC_SOURCE}" ; then
      open_jmx_port_for_spark_history_server
    fi

    if is_metric_source_enabled "${HIVESERVER2_METRIC_SOURCE}" || is_metric_source_enabled "${HIVEMETASTORE_METRIC_SOURCE}" ; then
      open_jmx_port_for_hive
    fi
  fi
}

function open_jmx_port_for_hdfs() {
  # Turn on the ports for JMX in HDFS NameNode, Secondary NameNode, DataNode.
  cat >>${HADOOP_CONF_DIR}/hadoop-env.sh <<'EOF'

# Stackdriver monitoring agent plugin
function jmx_opts() {
  local JMX_PORT="$1"
  echo "-Dcom.sun.management.jmxremote.authenticate=false \
      -Dcom.sun.management.jmxremote.ssl=false \
      -Djava.rmi.server.hostname=localhost \
      -Dcom.sun.management.jmxremote.port=$JMX_PORT \
      -Dcom.sun.management.jmxremote.rmi.port=$JMX_PORT"
}
HADOOP_NAMENODE_OPTS="$(jmx_opts 8551) $HADOOP_NAMENODE_OPTS"
export HADOOP_NAMENODE_OPTS
HADOOP_SECONDARYNAMENODE_OPTS="$(jmx_opts 8552) $HADOOP_SECONDARYNAMENODE_OPTS"
export HADOOP_SECONDARYNAMENODE_OPTS
HADOOP_DATANODE_OPTS="$(jmx_opts 8553) $HADOOP_DATANODE_OPTS"
export HADOOP_DATANODE_OPTS
EOF
}

function open_jmx_port_for_yarn() {
  # Turn on the port for JMX in Yarn ResourceManager.
  cat >>${HADOOP_CONF_DIR}/yarn-env.sh <<'EOF'

# Stackdriver monitoring agent plugin
function jmx_opts() {
  local JMX_PORT="$1"
  echo "-Dcom.sun.management.jmxremote.authenticate=false \
      -Dcom.sun.management.jmxremote.ssl=false \
      -Djava.rmi.server.hostname=localhost \
      -Dcom.sun.management.jmxremote.port=$JMX_PORT \
      -Dcom.sun.management.jmxremote.rmi.port=$JMX_PORT"
}
YARN_RESOURCEMANAGER_OPTS="$(jmx_opts 8554) $YARN_RESOURCEMANAGER_OPTS"
export YARN_RESOURCEMANAGER_OPTS
EOF
}

function open_jmx_port_for_hive() {
  cat>>${HIVE_CONF_DIR}/hive-env.sh <<'EOF'

# Added by configure_oss_monitoring.sh to enable jmx for collecting metrics from hiveserver2 and/or hive-metastore
function jmx_opts() {
  local JMX_PORT="$1"
  echo "-Dcom.sun.management.jmxremote.authenticate=false \
      -Dcom.sun.management.jmxremote.ssl=false \
      -Djava.rmi.server.hostname=localhost \
      -Dcom.sun.management.jmxremote.port=$JMX_PORT \
      -Dcom.sun.management.jmxremote.rmi.port=$JMX_PORT"
}
if [ "$SERVICE" = "hiveserver2" ]; then
  HADOOP_CLIENT_OPTS="$(jmx_opts 8005) $HADOOP_CLIENT_OPTS"
  export HADOOP_CLIENT_OPTS
fi

HIVE_METASTORE_HADOOP_OPTS="$(jmx_opts 8009) $HIVE_METASTORE_HADOOP_OPTS"
export HIVE_METASTORE_HADOOP_OPTS
EOF
}

function open_jmx_port_for_spark_history_server() {
  cat>>${SPARK_CONF_DIR}/spark-env.sh <<'EOF'

# Added by configure_oss_monitoring.sh to enable jmx for collecting metrics from history server
function jmx_opts() {
  local JMX_PORT="$1"
  echo "-Dcom.sun.management.jmxremote.authenticate=false \
      -Dcom.sun.management.jmxremote.ssl=false \
      -Djava.rmi.server.hostname=localhost \
      -Dcom.sun.management.jmxremote.port=$JMX_PORT \
      -Dcom.sun.management.jmxremote.rmi.port=$JMX_PORT"
}
SPARK_HISTORY_OPTS="$(jmx_opts 8008) $SPARK_HISTORY_OPTS"
export SPARK_HISTORY_OPTS
EOF
}

function init_collectd_config_dir() {
  mkdir -p /etc/stackdriver/collectd.d/
  mkdir -p /opt/stackdriver/collectd/etc/collectd.d/
  cp "${BDUTIL_DIR}/conf/collectd/gce/dataproc_gce_default.conf" "/etc/stackdriver/collectd.conf"
}

function update_default_s8s_collectd_config_files() {
  sed -i "s/REGION/${REGION_NAME}/g" "${COLLECTD_S8S_PLUGIN_DIR}/../dataproc_s8s_default.conf"
  sed -i "s/WORKLOAD_TYPE/${DATAPROC_WORKLOAD_TYPE}/g" "${COLLECTD_S8S_PLUGIN_DIR}/../dataproc_s8s_default.conf"
  sed -i "s/WORKLOAD_ID/${DATAPROC_WORKLOAD_ID}/g" "${COLLECTD_S8S_PLUGIN_DIR}/../dataproc_s8s_default.conf"
  cp "${COLLECTD_S8S_PLUGIN_DIR}/../dataproc_s8s_default.conf" "/etc/stackdriver/collectd.conf"
}

function set_property_in_flink_conf() {
  local -r name=$1
  local -r value=$2
  set_property_in_yaml "${FLINK_CONF}/flink-conf.yaml" "${name}" "${value}"
}

function update_flink_statsd_configs() {
  set_property_in_flink_conf "metrics.reporters" "stsd"
  set_property_in_flink_conf "metrics.reporter.stsd.factory.class" "org.apache.flink.metrics.statsd.StatsDReporterFactory"
  set_property_in_flink_conf "metrics.reporter.stsd.host" "localhost"
  set_property_in_flink_conf "metrics.reporter.stsd.port" "8125"
  set_property_in_flink_conf "metrics.reporter.stsd.interval" "10 SECONDS"
  set_property_in_flink_conf "metrics.scope.jm" "flink.jobmanager"
  set_property_in_flink_conf "metrics.scope.jm-job" "flink.jobmanager.job.job_id.<job_id>"
  set_property_in_flink_conf "metrics.scope.tm" "flink.taskmanager.tm_id.<tm_id>"
  set_property_in_flink_conf "metrics.scope.tm-job" "flink.taskmanager.job.job_id.<job_id>.tm_id.<tm_id>"
  set_property_in_flink_conf "metrics.scope.task" "flink.task.job_id.<job_id>.tm_id.<tm_id>"
  set_property_in_flink_conf "metrics.scope.operator" "flink.operator.job_id.<job_id>.tm_id.<tm_id>"
}

function init_opentelemetry_config_dir() {
   mkdir -p "${UCP_METRICS_AGENT_PATH}"/configs
}

function configure_spark_for_otel() {
  # Update spark namespace delimiter
  local -r search_string="namespace=app_name:\${spark.app.name}.app_id:\${spark.app.id}"
  local -r replace_string="namespace=app_name##\${spark.app.name}.app_id##\${spark.app.id}"
  sed -i "s/${search_string}/${replace_string}/g" "${SPARK_CONF_DIR}/spark-defaults.conf"
}

# Update opentelemetry config file based on metric source
function update_opentelemetry_config_files() {
  loginfo 'Running configure_oss_monitoring:update_opentelemetry_config_files.'

  local -r plugin_gen_dir="${BDUTIL_DIR}/monitoring_plugin_generator"
  local -r otel_out_dir="${plugin_gen_dir}/otel-out"
  setup_jmx_sources
  configure_spark_for_otel

  local statsd_services=()
  if is_metric_source_enabled "${SPARK_METRIC_SOURCE}"; then
    statsd_services+=("spark")
  fi

  if is_metric_source_enabled "${FLINK_METRIC_SOURCE}"; then
    statsd_services+=("flink")
  fi

  local jmx_enabled_services=()
  if is_metric_source_enabled hadoop || is_metric_source_enabled "${YARN_METRIC_SOURCE}"; then
    jmx_enabled_services+=("yarn")
  fi

  if is_metric_source_enabled hadoop || is_metric_source_enabled "${HDFS_METRIC_SOURCE}"; then
    jmx_enabled_services+=("hdfs")
  fi

  if is_metric_source_enabled "${SPARK_HISTORY_SERVER_METRIC_SOURCE}"; then
    jmx_enabled_services+=("sparkHistoryServer")
  fi

  if is_metric_source_enabled "${HIVESERVER2_METRIC_SOURCE}"; then
    jmx_enabled_services+=("hiveserver2")
  fi

  if is_metric_source_enabled "${HIVEMETASTORE_METRIC_SOURCE}"; then
    jmx_enabled_services+=("hivemetastore")
  fi

  local service_configs_to_generate=( "${jmx_enabled_services[@]}" "${statsd_services[@]}" )
  # Run default plugin generator for default metrics
  (
    # Activate the virtualenv
    source "${DATAPROC_DIR}/system/virtualenv/bin/activate"
    # Run jinja to generate the default opentelemetry config.
    # File name should be the same as the default file name for service for the override to work correctly.
    # shellcheck disable=SC2046 "Need unquoted expansion to pass kwargs to python"
    included_services=$(join_by_char ',' "${service_configs_to_generate[@]}")
    python "${plugin_gen_dir}/dataproc_monitoring_plugin_generator.py" \
        --metrics_file "${plugin_gen_dir}/dataproc-default-metrics.txt" \
        --plugin_generator_config_dir "${plugin_gen_dir}" \
        --output_dir "${otel_out_dir}" \
        --templates_dir "${plugin_gen_dir}" \
        --mode "default during startup" \
        --config_parameters_kwargs queue_name="${YARN_QUEUES}" \
        --include_services "${included_services}" \
        --agent "opentelemetry"
    deactivate
  )


  # Apply metric overrides, if any
  local metric_override_file
  metric_override_file=$(get_dataproc_property dataproc.metric.override.file)
  if [[ -n ${metric_override_file} ]]; then
    local config_bucket
    config_bucket=$(get_metadata_bucket)

    local cluster_uuid
    cluster_uuid=$(get_metadata_cluster_uuid)

    local cluster_staging_folder
    cluster_staging_folder="gs://${config_bucket}/google-cloud-dataproc-metainfo/${cluster_uuid}"

    local metric_override_file_uri
    metric_override_file_uri="${cluster_staging_folder}/${metric_override_file}"

    local collectd_conf_files_dir
    collectd_conf_files_dir="${BDUTIL_DIR}/conf/collectd/gce"

    # Download the customized metric list file from gcs
    download_resource "${metric_override_file_uri}" \
        "${plugin_gen_dir}/dataproc-metric-overrides"

    # If yarn metrics are overridden AND if driver pool is enabled,
    # add driver pool partition metrics into the overrides file
    local_metrics_override_file="${plugin_gen_dir}"/dataproc-metric-overrides
    is_yarn_metrics_overridden=$(grep "yarn:" "$local_metrics_override_file" || echo "")

    if [[ "${MASTER_RUN_DRIVER_LOCATION}" == "YARN" && $is_yarn_metrics_overridden ]]; then
    cat >>"$local_metrics_override_file" <<'EOF'

yarn:ResourceManager:DriverPoolsQueueMetrics:AvailableMB
yarn:ResourceManager:DriverPoolsQueueMetrics:PendingContainers
EOF
    fi

    # Run plugin generator again for metrics override
    # Start a subshell so the virtual environment doesn't affect wherever the script was run from
    (
      # Quit the bash strict mode, because virtualenv script might cause unbound
      # variable error.
      set +u

      # Activate the virtualenv
      source "${DATAPROC_DIR}/system/virtualenv/bin/activate"


      # We need to exclude jmx services on non-master node similar to collectd
      # implementation. But also need to explicitly mention include_services
      # but plugin generator does not support both of these parameters
      # together hence we just pass include_services.
      # include_services = enabled_services - exclude_services
      local exclude_services=""
      if [[ "${ROLE}" != 'Master' ]]; then
        exclude_services=("${HDFS_METRIC_SOURCE}" "${YARN_METRIC_SOURCE}" "${HIVESERVER2_METRIC_SOURCE}" "${SPARK_HISTORY_SERVER_METRIC_SOURCE}" "${HIVEMETASTORE_METRIC_SOURCE}")
      fi

      for service in "${exclude_services[@]}"; do
        service_configs_to_generate=("${service_configs_to_generate[@]/$service}")
      done

      included_services=$(join_by_char ',' "${service_configs_to_generate[@]}")

      # Run jinja to generate the opentelemetry config and replace the default one.
      # File name should be the same as the default file name for service for the override to work correctly.
      # shellcheck disable=SC2046 "Need unquoted expansion to pass kwargs to python"
      python "${plugin_gen_dir}/dataproc_monitoring_plugin_generator.py" \
          --metrics_file "${plugin_gen_dir}/dataproc-metric-overrides" \
          --plugin_generator_config_dir "${plugin_gen_dir}" \
          --output_dir "${otel_out_dir}" \
          --conf_files_dir "${collectd_conf_files_dir}" \
          --templates_dir "${plugin_gen_dir}" \
          --include_services "${included_services}" \
          --mode "override" \
          --config_parameters_kwargs queue_name="${YARN_QUEUES}" \
          --agent "opentelemetry"

      deactivate
    )
  fi


  # Copy otel base configuration
  cp "${plugin_gen_dir}/dataproc_otel_base.yaml" "${UCP_METRICS_AGENT_PATH}/configs/otel_base_config.yaml"

  # Copy individual OSS otel config when respective metric source is enabled

  if is_metric_source_enabled "${SPARK_METRIC_SOURCE}"; then
    cp "${otel_out_dir}/otel_spark_config.yaml" "${UCP_METRICS_AGENT_PATH}/configs/otel_spark_config.yaml"
  fi

  if is_metric_source_enabled "${FLINK_METRIC_SOURCE}"; then
    cp "${otel_out_dir}/otel_flink_config.yaml" "${UCP_METRICS_AGENT_PATH}/configs/otel_flink_config.yaml"
  fi

  # Update the target groovy file with content of groovy configs generated
  # by plugin generator for each jmx metric source (e.g. hdfs, yarn, etc.)
  local -r jmx_groovy_file="${UCP_METRICS_AGENT_PATH}/configs/java_extra_metrics.groovy"
  truncate -s 0 "${jmx_groovy_file}"

  if is_metric_source_enabled hadoop || is_metric_source_enabled "${YARN_METRIC_SOURCE}"; then
    cat "${otel_out_dir}/otel_yarn_config.groovy" >> "${jmx_groovy_file}"
  fi

  if is_metric_source_enabled hadoop || is_metric_source_enabled "${HDFS_METRIC_SOURCE}"; then
    cat "${otel_out_dir}/otel_hdfs_config.groovy" >> "${jmx_groovy_file}"
  fi

  if is_metric_source_enabled "${SPARK_HISTORY_SERVER_METRIC_SOURCE}"; then
    cat "${otel_out_dir}/otel_shs_config.groovy" >> "${jmx_groovy_file}"
  fi

  if is_metric_source_enabled "${HIVESERVER2_METRIC_SOURCE}"; then
    cat "${otel_out_dir}/otel_hs2_config.groovy" >> "${jmx_groovy_file}"
  fi

  if is_metric_source_enabled "${HIVEMETASTORE_METRIC_SOURCE}"; then
    cat "${otel_out_dir}/otel_hms_config.groovy" >> "${jmx_groovy_file}"
  fi

  # Copy otel_jmx_plugin.yaml into otel configs directory
  # if we have any jmx metric source enabled
  if (( ${#jmx_enabled_services[@]} != 0 )); then
    cp "${otel_out_dir}/otel_jmx_plugin.yaml" "${UCP_METRICS_AGENT_PATH}/configs/otel_jmx_plugin.yaml"
  else
    # Remove groovy file if no jmx metric sources are enabled
    rm "${jmx_groovy_file}"
  fi
}

# Update collectd config file based on metric source
function update_collectd_config_files() {
  loginfo 'Running configure_oss_monitoring:update_collectd_config_files.'

  local -r collectd_plugin_dir="/opt/stackdriver/collectd/etc/collectd.d/"
  setup_jmx_sources

  # Default collectd.conf file at /etc/stackdriver/collectd.conf contains plugins
  # (like cpu, agent, disk, interface, load, memory, pagefile, processes, swap)
  # that are enabled by default.
  # We ignore these plugins (except agent) when monitoring_agent_defaults is not
  # enabled. By replacing the collectd.conf with the minimal configuration at
  # collectd_without_monitoring_agent_defaults.conf.
  # Serverless Spark pushes all metrics against per-Dataproc Serverless workload
  # resource type. This requires us to use a separate collectd.conf config file.
  # We also need to populate region and batch_id/session_id for s8s collectd
  # conf files.
  if [[ "${DATAPROC_IMAGE_TYPE}" == "serverless-spark" ]]; then
    update_default_s8s_collectd_config_files
    local node_id
    node_id=${DATAPROC_WORKLOAD_ID}-${HOSTNAME/gdpic-$(get_metadata_cluster_name)-/}
    if is_metric_source_enabled "${MONITORING_AGENT_DEFAULTS_SOURCE}"; then
      sed -i "s/NODE_ID/$node_id/g" "${COLLECTD_S8S_PLUGIN_DIR}"/*
      cp "${COLLECTD_S8S_PLUGIN_DIR}"/* "${collectd_plugin_dir}"
    fi
  elif ! is_metric_source_enabled "${MONITORING_AGENT_DEFAULTS_SOURCE}"; then
    cp "${BDUTIL_DIR}/conf/collectd_without_monitoring_agent_defaults.conf" "/etc/stackdriver/collectd.conf"
  else
    cp "${BDUTIL_DIR}/conf/collectd_default_filtered_write.conf" "/etc/stackdriver/collectd.conf"
  fi

  # Copy spark collectd file when spark metrics are enabled
  if is_metric_source_enabled "${SPARK_METRIC_SOURCE}"; then
    local -r collectd_spark_file="${BDUTIL_DIR}/conf/${SPARK_COLLECTD_CONF_FILE_NAME}"
    cp "${collectd_spark_file}" "${collectd_plugin_dir}/${SPARK_COLLECTD_CONF_FILE_NAME}"
  fi

  # Copy flink collectd file when flink metrics are enabled
  if is_metric_source_enabled "${FLINK_METRIC_SOURCE}"; then
    update_flink_statsd_configs
    local -r collectd_flink_file="${BDUTIL_DIR}/conf/${FLINK_COLLECTD_CONF_FILE_NAME}"
    cp "${collectd_flink_file}" "${collectd_plugin_dir}/${FLINK_COLLECTD_CONF_FILE_NAME}"
  fi

  local load_jmx_plugin=false

  # For now, the only jmx connections in hadoop/yarn/hdfs are services that are only enabled in Master
  # We will enhance this when we start adding support for other connections that are applicable to workers too.
  # If we let these jmx conf files get installed in workers, the /var/log/syslog is polluted with
  # lot of connection refused error messages.
  # "GenericJMXConfConnection: Creating MBean server connection failed:
  # java.io.IOException: Failed to retrieve RMIServer stub: javax.naming.ServiceUnavailableException"
  if [[ "${ROLE}" == 'Master' ]]; then
    # If hadoop is enabled, copy over both hdfs and yarn config file to
    # "/opt/stackdriver/collectd/etc/collectd.d/" path.
    if is_metric_source_enabled hadoop || is_metric_source_enabled "${HDFS_METRIC_SOURCE}"; then
      load_jmx_plugin=true
      local -r collectd_hdfs_file="${BDUTIL_DIR}/conf/${HDFS_COLLECTD_CONF_FILE_NAME}"
      cp "${collectd_hdfs_file}" "${collectd_plugin_dir}/${HDFS_COLLECTD_CONF_FILE_NAME}"
    fi

    if is_metric_source_enabled hadoop || is_metric_source_enabled "${YARN_METRIC_SOURCE}"; then
      load_jmx_plugin=true
      # Because of the addition of ChildQueueMetrics to the default metric list,
      # we have to generate yarn collect configuration during startup even if
      # there are no overrides. This is because we need to read the queue name
      # from the property which is only known during the startup.
      (
      # Quit the bash strict mode, because virtualenv script might cause unbound
      # variable error.
      set +u
      local -r plugin_gen_dir="${BDUTIL_DIR}/monitoring_plugin_generator"

      # If driver pools enabled, add driver pools partition metrics into default
      # metrics list.
      if [[ "${MASTER_RUN_DRIVER_LOCATION}" == "YARN" ]]; then
      cat >>"${plugin_gen_dir}"/dataproc-default-metrics.txt <<'EOF'
yarn:ResourceManager:DriverPoolsQueueMetrics:AvailableMB
yarn:ResourceManager:DriverPoolsQueueMetrics:PendingContainers
EOF
      fi

      # Activate the virtualenv
      source "${DATAPROC_DIR}/system/virtualenv/bin/activate"
      # Run jinja to generate the collectd plugin.
      # File name should be the same as the default file name for service for the override to work correctly.
      # shellcheck disable=SC2046 "Need unquoted expansion to pass kwargs to python"
      python "${plugin_gen_dir}/dataproc_monitoring_plugin_generator.py" \
          --metrics_file "${plugin_gen_dir}/dataproc-default-metrics.txt" \
          --plugin_generator_config_dir "${plugin_gen_dir}" \
          --output_dir "${collectd_plugin_dir}" \
          --templates_dir "${plugin_gen_dir}" \
          --output_file_name_kwargs $(get_metric_source_to_file_name_kwargs) \
          --mode "default during startup" \
          --include_services "yarn" \
          --config_parameters_kwargs queue_name="${YARN_QUEUES}"
      deactivate
      )
    fi

    if is_metric_source_enabled "${SPARK_HISTORY_SERVER_METRIC_SOURCE}"; then
      load_jmx_plugin=true
      local -r collectd_shs_file="${BDUTIL_DIR}/conf/${SPARK_HISTORY_SERVER_COLLECTD_CONF_FILE_NAME}"
      cp "${collectd_shs_file}" "${collectd_plugin_dir}/${SPARK_HISTORY_SERVER_COLLECTD_CONF_FILE_NAME}"
    fi

    if is_metric_source_enabled "${HIVESERVER2_METRIC_SOURCE}"; then
      load_jmx_plugin=true
      local -r collectd_hiveserver2_file="${BDUTIL_DIR}/conf/${HIVESERVER2_COLLECTD_CONF_FILE_NAME}"
      cp "${collectd_hiveserver2_file}" "${collectd_plugin_dir}/${HIVESERVER2_COLLECTD_CONF_FILE_NAME}"
    fi

    if is_metric_source_enabled "${HIVEMETASTORE_METRIC_SOURCE}"; then
      load_jmx_plugin=true
      local -r collectd_hivemetastore_file="${BDUTIL_DIR}/conf/${HIVEMETASTORE_COLLECTD_CONF_FILE_NAME}"
      cp "${collectd_hivemetastore_file}" "${collectd_plugin_dir}/${HIVEMETASTORE_COLLECTD_CONF_FILE_NAME}"
    fi
  fi
  # Copy the config file to load the jmx plugin.
  # Note that the file should be prefixed with "00" to make it the first one
  # per lexicographical order. We do this in order to load the plugin ahead of
  # the other .conf files using JMX plugin. Having multiple files load the same
  # jmx plugin gives warnings (b/211579403#comment5)
  # Refer: https://collectd.org/documentation/manpages/collectd.conf.5.shtml#global_options
  if ${load_jmx_plugin}; then
    local -r collectd_load_jmx_plugin_file="${BDUTIL_DIR}/conf/collectd_load_jmx_plugin.conf"
    cp "${collectd_load_jmx_plugin_file}" "${collectd_plugin_dir}/00_load_jmx_plugin.conf"
  fi

  # Apply metric overrides, if any
  local metric_override_file
  metric_override_file=$(get_dataproc_property dataproc.metric.override.file)
  if [[ -n ${metric_override_file} ]]; then
    local config_bucket
    config_bucket=$(get_metadata_bucket)

    local cluster_uuid
    cluster_uuid=$(get_metadata_cluster_uuid)

    local cluster_staging_folder
    cluster_staging_folder="gs://${config_bucket}/google-cloud-dataproc-metainfo/${cluster_uuid}"

    local metric_override_file_uri
    metric_override_file_uri="${cluster_staging_folder}/${metric_override_file}"

    local plugin_gen_dir
    plugin_gen_dir="${BDUTIL_DIR}/monitoring_plugin_generator"

    local collectd_conf_files_dir
    collectd_conf_files_dir="${BDUTIL_DIR}/conf/collectd/gce"

    local default_collectd_file
    default_collectd_file="/etc/stackdriver/collectd.conf"

    # Download the customized metric list file from gcs
    download_resource "${metric_override_file_uri}" \
        "${plugin_gen_dir}/dataproc-metric-overrides"

    # If yarn metrics are overridden AND if driver pool is enabled,
    # add driver pool partition metrics into the overrides file
    local_metrics_override_file="${plugin_gen_dir}"/dataproc-metric-overrides
    is_yarn_metrics_overridden=$(grep "yarn:" "$local_metrics_override_file" || echo "")

    if [[ "${MASTER_RUN_DRIVER_LOCATION}" == "YARN" && $is_yarn_metrics_overridden ]]; then
    cat >>"$local_metrics_override_file" <<'EOF'

yarn:ResourceManager:DriverPoolsQueueMetrics:AvailableMB
yarn:ResourceManager:DriverPoolsQueueMetrics:PendingContainers
EOF
    fi

    # Start a subshell so the virtual environment doesn't affect wherever the script was run from
    (
      # Quit the bash strict mode, because virtualenv script might cause unbound
      # variable error.
      set +u

      # Activate the virtualenv
      source "${DATAPROC_DIR}/system/virtualenv/bin/activate"

      local exclude_services=""
      if [[ "${ROLE}" != 'Master' ]]; then
        exclude_services="${HDFS_METRIC_SOURCE},${YARN_METRIC_SOURCE},${HIVESERVER2_METRIC_SOURCE},${SPARK_HISTORY_SERVER_METRIC_SOURCE},${HIVEMETASTORE_METRIC_SOURCE}"
      fi
      # Run jinja to generate the collectd plugin and replace the default one.
      # File name should be the same as the default file name for service for the override to work correctly.
      # shellcheck disable=SC2046 "Need unquoted expansion to pass kwargs to python"
      # shellcheck disable=SC2128 "exclude_services is not an array but a string"
      python "${plugin_gen_dir}/dataproc_monitoring_plugin_generator.py" \
          --metrics_file "${plugin_gen_dir}/dataproc-metric-overrides" \
          --plugin_generator_config_dir "${plugin_gen_dir}" \
          --output_dir "${collectd_plugin_dir}" \
          --conf_files_dir "${collectd_conf_files_dir}" \
          --default_collectd_file "${default_collectd_file}" \
          --templates_dir "${plugin_gen_dir}" \
          --output_file_name_kwargs $(get_metric_source_to_file_name_kwargs) \
          --exclude_services "${exclude_services}" \
          --mode "override" \
          --config_parameters_kwargs queue_name="${YARN_QUEUES}"

      deactivate
    )
  fi

}