local socket = require "socket.http"
local https = require "ssl.https"
local ltn12 = require "ltn12"
local lunajson = require "lunajson"

local cluster_props = {}
local job_id_cache = {}
local job_id_cache_size = 0
local address_port_regexp = "[^/]+:([0-9]+)"
local default_rm_http_port = 8088
local default_rm_https_port = 8090
local get_job_id_attempts = {}
local max_rm_http_attempts = 5

--[[
  See https://docs.fluentbit.io/manual/v/1.9-pre/pipeline/outputs/stackdriver
  record is a lua table with various fields corresponding to jsonPayload in a cloud log entry
  Any top-level field will be directly copied into jsonPayload except:
  1. if field name equals Severity_Key mentioned in stackdriver output plugin parameters, its
  value will be represented as severity of the log entry in cloud logging
  2. if field name equals "logging.googleapis.com/logName", its value will become LOGNAME of
  the log entry in cloud logging prefixed with project name. Ex: projects/<PROJECT>/logs/<LOGNAME>
  3. if field name equals "logging.googleapis.com/labels", it is a lua table. The values
  are copied inside labels of the log entry (these are on top of the monitored resource specific labels)
  4. if field name equals "logging.googleapis.com/monitored_resource", it is a lua table with "labels"
  key. The values inside labels are copied inside resource.labels of the log entry. These correspond to
  identifying the monitored resource whose type is specified as resource.type of the log entry

  Example input record:
  { "message" = "Starting health monitor daemon for []",
    "class" = "com.google.cloud.hadoop.services.agent.node.DaemonMonitorService",
    "method" = "doStart",
    "severity" = "INFO"
  }

  Output record of process method:
  { "message" = "Starting health monitor daemon for []",
    "class" = "com.google.cloud.hadoop.services.agent.node.DaemonMonitorService",
    "method" = "doStart",
    "severity" = "INFO",
    "logging.googleapis.com/labels" = {
      "compute.googleapis.com/resource_id" = "1622993124111963422",
      "compute.googleapis.com/resource_name" = "service-test-w-1",
      "compute.googleapis.com/zone" = "us-central1-c"
    },
    "logging.googleapis.com/monitored_resource" = {
      "labels" = {
        "cluster_name" = "service-test",
        "cluster_uuid" = "123444sa-3445-2444-dc43-65dhdhh3354",
        "region" = "us-central1"
      }
    }
  }

  How this log entry looks in Cloud Logging:
  {
    jsonPayload: {
      message: "Starting health monitor daemon for []",
      class: "com.google.cloud.hadoop.services.agent.node.DaemonMonitorService",
      method: "doStart"
    },
    labels: {
      compute.googleapis.com/resource_id: "1622993124111963422",
      compute.googleapis.com/resource_name: "service-test-w-1",
      compute.googleapis.com/zone: "us-central1-c"
    },
    resource: {
      labels: {
        cluster_name: "service-test",
        cluster_uuid: "123444sa-3445-2444-dc43-65dhdhh3354",
        project_id: "test-project", -- Inferred from GCE credentials for stackdriver output plugin
        region: "us-central1"
      },
      type: "cloud_dataproc_cluster" -- Inferred from Resource mentioned in stackdriver output plugin parameters
    },
    severity: "INFO" -- This is set from value of Severity_Key mentioned in stackdriver output plugin parameters.
    -- If Severity_Key lookup does not match any fields inside output record, severity shall not be set
    -- For example, if Severity_Key equals "severity1" then the field severity: "INFO" would remain inside
    -- jsonPayload itself and this log's severity would not be set (hence assumed DEFAULT)
  }
--]]
function process(tag, timestamp, record)
  -- Tags starting with re-emitted are ignored because they have already been processed
  -- prefix hyphen character with '%' for regex find
  if string.find(tag, "^re%-emitted") then
    return 0, timestamp, record
  end

  local resource_prefix = "logging.googleapis.com/monitored_resource"
  local labels_key = "logging.googleapis.com/labels"
  local log_name_key = "logging.googleapis.com/logName"
  local sep = "/"
  local file_name_key = "filename"
  local dataproc_job_service_prefix = "dataproc.googleapis.com"

  if record[resource_prefix] == nil then
    record[resource_prefix] = {}
    record[resource_prefix]["labels"] = {}
  end

  -- Set GCE VM labels inside record
  local gce_labels_prefix = "compute.googleapis.com"
  record[labels_key] = {}
  record[labels_key][gce_labels_prefix .. "/resource_id"] = get_property_value("id")
  record[labels_key][gce_labels_prefix .. "/resource_name"] = get_hostname()
  record[labels_key][gce_labels_prefix .. "/zone"] = get_zone(sep)

  -- Obtain the following using shell script /usr/share/google/get_metadata_value "attributes/dataproc-region"
  local cluster_name = get_property_value("attributes/dataproc-cluster-name")
  local cluster_uuid = get_property_value("attributes/dataproc-cluster-uuid")
  local region = get_property_value("attributes/dataproc-region")
  record[resource_prefix]["labels"]["cluster_name"] = cluster_name
  record[resource_prefix]["labels"]["cluster_uuid"] = cluster_uuid
  record[resource_prefix]["labels"]["region"] = region

  -- We do not need further filename based processing for earlyOOM records
  if tag == "google.dataproc.oom-killer" then
    return 2, timestamp, record
  end

  -- prefix hyphen character with '%' for regex substitution
  local prefixed_hostname = "%-" .. string.gsub(get_hostname(), "%-", "%%%-")
  if record[file_name_key] ~= nil then
    local orig_file_path = record[file_name_key]
    local t = split_and_build_table(orig_file_path, sep)
    record[file_name_key] = t[#t]
    -- prefix hyphen character with '%' for regex find
    if string.find(record[file_name_key], "startup%-script") or
        string.find(record[file_name_key], "dataproc%-initialization%-script%-") then
      record[log_name_key] = "google.dataproc.startup"
      if record["message"] == nil or record["message"] == "" then
        return -1, timestamp, record -- Drop the records without meaningful message
      end
      if record["subprocess_tag"] and record["pid"] then
        record["message"] = record["subprocess_tag"] .. "[" .. record["pid"] .. "]: " .. record["message"]
        -- Remove repeated data
        record["subprocess_tag"] = nil
        record["pid"] = nil
      end
    elseif string.find(orig_file_path, "/hadoop%-yarn/userlogs/") then -- prefix hyphen character with '%' for regex find
      application_id = t[#t-2]
      record["application"] = application_id
      record["container"] = t[#t-1]
      record["container_logname"] = t[#t]
      record[file_name_key] = table.concat({application_id, record["container"], record["container_logname"]}, ".")

      -- Logic to map to cloud_dataproc_job resource
      if tag == "dataproc.job.yarn.container" then
        -- Erase the job_id_cache table and reset job_id_cache_size if size of cache exceeds 200. Since there is no easy
        -- way to track size of job_id_cache table with string application_id keys, keep a count of keys separately
        if job_id_cache_size > 200 then
          job_id_cache = {}
          job_id_cache_size = 0
        end

        local job_id, job_uuid
        if job_id_cache[application_id] ~= nil then
          job_id = job_id_cache[application_id][1]
          job_uuid = job_id_cache[application_id][2]
        else
          local app_json = get_app(application_id)
          local inner_json = app_json["app"] or {}
          local tags = split_and_build_table(inner_json["applicationTags"] or "", ",")

          for _, tag in pairs(tags) do
            if not string.find(tag, "^dataproc_job_attempt_timestamp_") then
              if string.match(tag, "dataproc_job_([%a%d]([%a%d%-%_]+[%a%d]))") then
                job_id = string.gsub(tag, "^dataproc_job_", "")
              elseif string.match(tag, "dataproc_uuid_([%a%d]([%a%d%-%_]+[%a%d]))") then
                job_uuid = string.gsub(tag, "^dataproc_uuid_", "")
              end
            end
          end
          if job_id and job_uuid then
            job_id_cache[application_id] = {job_id, job_uuid} -- table with indices 1 for job_id, 2 for job_uuid
            job_id_cache_size = job_id_cache_size + 1
          else
            local invalid_attempts = (get_job_id_attempts[application_id] or 0) + 1
            get_job_id_attempts[application_id] = invalid_attempts
            if invalid_attempts >= max_rm_http_attempts then
              job_id_cache[application_id] = {nil, nil} -- table with indices 1 for job_id, 2 for job_uuid
              job_id_cache_size = job_id_cache_size + 1
            end
          end
        end

        if job_id and job_uuid then
          record[resource_prefix]["labels"] = {} -- replace the cluster labels with job labels
          record[resource_prefix]["labels"]["job_id"] = job_id
          record[resource_prefix]["labels"]["job_uuid"] = job_uuid
          record[resource_prefix]["labels"]["region"] = region

          -- Set cluster_name, cluster_uuid inside record, besides GCE VM labels
          record[labels_key][dataproc_job_service_prefix .. "/cluster_name"] = cluster_name
          record[labels_key][dataproc_job_service_prefix .. "/cluster_uuid"] = cluster_uuid
        else
          -- If job_id or job_uuid is missing, return the record with cluster resource labels only
          -- We will emit this under the cloud_dataproc_cluster resource by renaming the tag to re-emitted-yarn-userlogs
          -- After renaming, the extra key route_job_record_to_cluster will be erased by fluentbit pipeline
          record["route_job_record_to_cluster"] = "true"
          record[log_name_key] = "yarn-userlogs"
        end
      end
    elseif string.find(orig_file_path, "/google%-dataproc%-job/") then -- prefix hyphen character with '%' for regex find
      record[resource_prefix]["labels"] = {} -- replace the cluster labels with job labels
      record[resource_prefix]["labels"]["job_id"] = t[#t-2]
      record[resource_prefix]["labels"]["job_uuid"] = t[#t-1]
      record[resource_prefix]["labels"]["region"] = region
      record[file_name_key] = nil

      -- Set cluster_name, cluster_uuid inside record, besides GCE VM labels
      record[labels_key][dataproc_job_service_prefix .. "/cluster_name"] = cluster_name
      record[labels_key][dataproc_job_service_prefix .. "/cluster_uuid"] = cluster_uuid
    else
      record[log_name_key] = string.gsub(t[#t], "%.log", "") -- prefix dot character with '%' for regex substitution
      record[log_name_key] = string.gsub(record[log_name_key], prefixed_hostname, "")
    end
  end

  return 2, timestamp, record
end

function get_app(app)
  local rm_hosts = { get_property_value("attributes/dataproc-master") }
  local addl_hosts = split_and_build_table(get_property_value("attributes/dataproc-master-additional"), ",")
  for _, addl_host in ipairs(addl_hosts) do
    table.insert(rm_hosts, addl_host)
  end

  for _, rm_host in ipairs(rm_hosts) do
    local resp = {}
    if use_https() then
      _, code, headers, status = https.request{
        url="https://" .. rm_host .. "." .. get_domain() .. ":" .. rm_https_port() .. "/ws/v1/cluster/apps/" .. app,
        protocol="sslv23",
        method="GET",
        sink=ltn12.sink.table(resp)
      }
    else
      _, code, headers, status = socket.request{
        host=rm_host,
        path="/ws/v1/cluster/apps/" .. app,
        port=rm_http_port(),
        scheme="http",
        method="GET",
        sink=ltn12.sink.table(resp)
      }
    end
    if code == 200 then
      local result_str = table.concat(resp)
      local status_ok, app_json = pcall(lunajson.decode, result_str)
      if status_ok then
        return app_json
      end
    end
  end
  return {}
end

function split_and_build_table(input, sep)
  local t = {}
  for str in string.gmatch(input, "([^"..sep.."]+)") do
    table.insert(t, str)
  end
  return t
end

function get_property_value(property)
  if cluster_props[property] == nil then
    cluster_props[property] = command("/usr/share/google/get_metadata_value " .. property)
  end
  return cluster_props[property]
end

function get_zone(sep)
  if cluster_props["zone"] == nil then
    local t = split_and_build_table(command("/usr/share/google/get_metadata_value zone"), sep)
    cluster_props["zone"] = t[#t]
  end
  return cluster_props["zone"]
end

function use_https()
  if cluster_props["use_https"] == nil then
    cluster_props["use_https"] = read_property_safe("yarn.http.policy") == "HTTPS_ONLY"
  end
  return cluster_props["use_https"]
end

function rm_http_port()
  if cluster_props["rm_http_port"] == nil then
    local rm_http_address = read_property_safe("yarn.resourcemanager.webapp.address")
    local http_port = default_rm_http_port
    _, _, port = string.find(rm_http_address, address_port_regexp)
    if port ~= nil then
      http_port = port
    end
    cluster_props["rm_http_port"] = http_port
  end
  return cluster_props["rm_http_port"]
end

function rm_https_port()
  if cluster_props["rm_https_port"] == nil then
    local rm_https_address = read_property_safe("yarn.resourcemanager.webapp.https.address")
    local https_port = default_rm_https_port
    _, _, port = string.find(rm_https_address, address_port_regexp)
    if port ~= nil then
      https_port = port
    end
    cluster_props["rm_https_port"] = https_port
  end
  return cluster_props["rm_https_port"]
end

function read_property_safe(property)
  local status_ok, result = pcall(command, "echo `hdfs getconf -confKey " .. property .. "`")
  if status_ok then
    return result.gsub(result, "\n$", "")
  end
  return ""
end

function get_domain()
  if cluster_props["domain"] == nil then
    cluster_props["domain"] = string.gsub(command("echo `dnsdomainname`"), "\n$", "")
  end
  return cluster_props["domain"]
end

function get_hostname()
  if cluster_props["hostname"] == nil then
    cluster_props["hostname"] = string.gsub(command("/bin/hostname"), "\n$", "")
  end
  return cluster_props["hostname"]
end

function command(cmd)
  local handle = io.popen(cmd)
  local result = handle:read("*a")
  handle:close()
  return result
end
