# Copyright 2019 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

require_relative 'yarn_rest_client'


# This module defines functionality for resolving Dataproc Job IDs and UUIDs
# from YARN Application IDs.
module DataprocJobIdResolver

  # The structure for resolved job ids.
  ResolvedJobId = Struct.new(:job_id, :job_uuid)

  # A DataprocJobId resolver based on REST requests to YARN resource managers.
  #  This assumes that Dataproc Jobs that have corresponding YARN applications
  #  have application tags attached to them identifying the application's Job.
  class YarnBasedJobIdResolver

    JOB_ID_APP_TAG_REGEXP =
        /dataproc_job_(?<id>[a-zA-Z0-9]([a-zA-Z0-9\-\_]{0,98}[a-zA-Z0-9])?)/.freeze
    JOB_UUID_APP_TAG_REGEXP =
        /dataproc_uuid_(?<uuid>[a-zA-Z0-9]([a-zA-Z0-9\-\_]{0,98}[a-zA-Z0-9])?)/.freeze
    DATAPROC_APP_TAG_DIVIDER = ','.freeze

    def initialize(yarn_client)
      @yarn_client = yarn_client
    end

    # Get the ResolvedJobId for a given Application ID.
    #
    #  This assumes that Dataproc Jobs have been appropriately tagged in the
    #  YARN Resource Manager. If the Job was not tagged or the Job ID and Job
    #  UUID are unable to be resolved for any reason, this method will return
    #  nil.
    def resolve(app_id)
      job_id = nil
      job_uuid = nil
      app_json = @yarn_client.get_app(app_id)
      if app_json
        app = app_json[Yarn::ApplicationJsonConstants::APP_FIELD_NAME]
        app_tags = app[
          Yarn::ApplicationJsonConstants::APP_NESTED_APP_TAG_FIELD_NAME]
        if app_tags
          app_tags.split(DATAPROC_APP_TAG_DIVIDER).each do |tag|
            matched_job_id_tag = JOB_ID_APP_TAG_REGEXP.match(tag)
            job_id = matched_job_id_tag[:id] if matched_job_id_tag
            matched_job_uuid_tag = JOB_UUID_APP_TAG_REGEXP.match(tag)
            job_uuid = matched_job_uuid_tag[:uuid] if matched_job_uuid_tag
          end
        end
      end
      ResolvedJobId.new(job_id, job_uuid) if job_id && job_uuid
    end
  end
end
