# Copyright 2019 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

require 'fluent/plugin/filter'
require 'lru_redux'
require_relative 'dataproc_job_id_resolver_provider'
require_relative 'dataproc_yarn_job_id_resolver'


module Fluent
  module Plugin

    # Fluentd filter plugin for converting YARN Application IDs into Dataproc Job
    # IDs on a Dataproc Cluster.
    class YarnAppToDataprocJob < Filter

      # Register the plugin for use by configs
      Fluent::Plugin.register_filter('yarn_app_to_dataproc_job', self)

      DEFAULT_APP_ID_KEY = 'app_id'.freeze
      DEFAULT_JOB_ID_KEY = 'job_id'.freeze
      DEFAULT_JOB_UUID_KEY = 'job_uuid'.freeze
      DEFAULT_MAX_CACHE_SIZE = 200.freeze
      DEFAULT_CACHE_TTL_SEC = (10 * 60).freeze

      APP_ID_KEY_DESCRIPTION = ('The record key for the YARN application ' +
          "ID, defaults to #{DEFAULT_APP_ID_KEY}.").freeze
      JOB_ID_KEY_DESCRIPTION = ('The record key for the Dataproc Job ID, ' +
          "defaults to #{DEFAULT_JOB_ID_KEY}.").freeze
      JOB_UUID_KEY_DESCRIPTION = ('The record key for the Dataproc Job ' +
          "UUID, defaults to #{DEFAULT_JOB_UUID_KEY}.").freeze
      MAX_CACHE_SIZE_DESCRIPTION = ('The maximum number of mappings cached ' +
          "internally, defaults to #{DEFAULT_MAX_CACHE_SIZE}.").freeze
      CACHE_TTL_DESCRIPTION = ('The TTL in seconds of mappings cached ' +
          "internally, defaults to #{DEFAULT_CACHE_TTL_SEC}.").freeze

      # Define plugin params

      desc APP_ID_KEY_DESCRIPTION
      config_param :app_id_key, :string, default: DEFAULT_APP_ID_KEY

      desc JOB_ID_KEY_DESCRIPTION
      config_param :job_id_key, :string, default: DEFAULT_JOB_ID_KEY

      desc JOB_UUID_KEY_DESCRIPTION
      config_param :job_uuid_key, :string, default: DEFAULT_JOB_UUID_KEY

      desc MAX_CACHE_SIZE_DESCRIPTION
      config_param :max_cache_size, :integer, default: DEFAULT_MAX_CACHE_SIZE

      desc CACHE_TTL_DESCRIPTION
      config_param :cache_ttl_seconds, :integer, default: DEFAULT_CACHE_TTL_SEC

      def start
        super
        @log = $log
        @job_id_resolver = DataprocJobIdResolver::Provider.default
        @job_id_cache = LruRedux::TTL::ThreadSafeCache.
            new(@max_cache_size, @cache_ttl_seconds)
      end

      def filter(tag, time, record)
        # Only modify a record if it contains the App ID, but not Job ID or UUID.
        if record[@job_id_key].to_s.empty? &&
           record[@job_uuid_key].to_s.empty? &&
           !record[@app_id_key].to_s.empty?
          app_id = record[@app_id_key]
          resolved_job_ids = @job_id_cache.getset(app_id) {
            @job_id_resolver.resolve(app_id)
          }
          # Only set the the ID and UUID if they were correctly resolved.
          if resolved_job_ids != nil
            record[@job_id_key] = resolved_job_ids.job_id
            record[@job_uuid_key] = resolved_job_ids.job_uuid
          end
        end
        record
      end
    end
  end
end
