#!/usr/bin/env rspec

require 'fluent/test'
require 'fluent/test/driver/filter'
require 'fluent/test/helpers'
require 'rspec'
require 'timecop'
require_relative 'dataproc_job_id_resolver_provider'
require_relative 'filter_yarn_app_to_dataproc_job'


RSpec.describe Fluent::Plugin::YarnAppToDataprocJob do

  include Fluent::Test::Helpers

  before(:example) do
    # Create a stub resolver that returns nil unless otherwise allowed
    @job_id_resolver = instance_double('JobIdResolver')
    # Override the static default provider to inject our stub
    allow(DataprocJobIdResolver::Provider).to receive(:default).
        and_return(@job_id_resolver)
  end

  def create_driver(conf='')
    Fluent::Test::Driver::Filter.new(Fluent::Plugin::YarnAppToDataprocJob).
        configure(conf)
  end

  def run(driver, *records)
    driver.run do
      records.each do |record|
        time = event_time
        driver.feed('test', time, record)
      end
    end
    driver.filtered_records
  end

  context 'default plugin configuration' do
    before(:example) do
      @driver = create_driver
    end

    it 'should successfully transform records with default settings' do
      record = {
        'message' => 'SOS',
        'app_id' => 'app_1'
      }
      job_ids = {
        'job_id' => 'job-1',
        'job_uuid' => '1234-5678-9012'
      }
      allow(@job_id_resolver).to receive(:resolve).with('app_1').
          and_return(DataprocJobIdResolver::ResolvedJobId.
              new('job-1', '1234-5678-9012'))
      filtered_records = run(@driver, record)
      expect(filtered_records.size).to eq(1)
      expect(filtered_records[0]).to eq(record.merge(job_ids))
    end

    it 'should not call resolver twice when value is cached' do
      record_1 = {
        'message' => 'lululu',
        'app_id' => 'app_1',
      }
      record_2 = {
        'message' => 'I got some apples',
        'app_id' => 'app_1',
      }
      job_ids = {
        'job_id' => 'job-1',
        'job_uuid' => '1234-5678-9012'
      }
      expect(@job_id_resolver).to receive(:resolve).with('app_1').
          and_return(DataprocJobIdResolver::ResolvedJobId.
              new('job-1', '1234-5678-9012')).once
      filtered_records = run(@driver, record_1, record_2)
    end

    it 'should not modify records when job ids do not resolve' do
      record = {
        'message' => 'Whats the Jobs API?',
        'app_id' => 'app_1'
      }
      allow(@job_id_resolver).to receive(:resolve).with('app_1').
          and_return(nil)
      filtered_records = run(@driver, record)
      expect(filtered_records.size).to eq(1)
      expect(filtered_records[0]).to eq(record)
    end

    it 'should not call resolver twice when cached value is nil' do
      record_1 = {
        'message' => 'lululu',
        'app_id' => 'app_1',
      }
      record_2 = {
        'message' => 'I got some apples',
        'app_id' => 'app_1',
      }
      expect(@job_id_resolver).to receive(:resolve).with('app_1').
          and_return(nil).once
      filtered_records = run(@driver, record_1, record_2)
    end

    it 'should not modify records with no app id key' do
      record_without_app_id_key = {
        'message' => 'yarn crashed or sommat',
        'other_key' => 'other_value'
      }
      filtered_records = run(@driver, record_without_app_id_key)
      expect(filtered_records.size).to eq(1)
      expect(filtered_records[0]).to eq(record_without_app_id_key)
      expect(@job_id_resolver).to_not receive(:resolve).with(any_args)
    end

    it 'should not modify records with preexisting job id key' do
      record_with_preexisting_job_id_key = {
        'message' => 'your mr job is still chugging along',
        'app_id' => 'app_1',
        'other_key' => 'other_value',
        'job_id' => 'job-1'
      }
      filtered_records = run(@driver, record_with_preexisting_job_id_key)
      expect(filtered_records.size).to eq(1)
      expect(filtered_records[0]).to eq(record_with_preexisting_job_id_key)
      expect(@job_id_resolver).to_not receive(:resolve).with(any_args)
    end

    it 'should not modify records with preexisting job uuid key' do
      record_with_preexisting_job_uuid_key = {
        'message' => 'your spark job almost done promise',
        'app_id' => 'app_1',
        'other_key' => 'other_value',
        'job_uuid' => '1234-5678-9012'
      }
      filtered_records = run(@driver, record_with_preexisting_job_uuid_key)
      expect(filtered_records.size).to eq(1)
      expect(filtered_records[0]).to eq(record_with_preexisting_job_uuid_key)
      expect(@job_id_resolver).to_not receive(:resolve).with(any_args)
    end
  end

  context 'plugin configs' do
    context 'for record keys' do

      before(:example) do
        # All key tests have the same built in resolution
        @app_id_value = 'app_1'
        @job_id_value = 'job-1'
        @job_uuid_value = '1234-5678-9012'
        allow(@job_id_resolver).to receive(:resolve).with(@app_id_value).
            and_return(DataprocJobIdResolver::ResolvedJobId.
                new(@job_id_value, @job_uuid_value))
      end

      context 'on input' do
        before(:example) do
          @job_ids = {
            'job_id' => @job_id_value,
            'job_uuid' => @job_uuid_value
          }
        end

        it 'should correctly handle custom app id keys' do
          config = %[
            app_id_key 'application'
          ]
          driver = create_driver(config)
          record_with_custom_app_id_key = {
            'message' => 'hello',
            'application' => 'app_1'
          }
          filtered_records = run(driver, record_with_custom_app_id_key)
          expect(filtered_records.size).to eq(1)
          expect(filtered_records[0]).
              to eq(record_with_custom_app_id_key.merge(@job_ids))
        end
      end

      context 'on output' do
        before(:example) do
          @record = {
            'message' => 'yolo',
            'app_id' => @app_id_value
          }
        end

        it 'should correctly handle custom job id keys' do
          config = %[
            job_id_key 'user-defined-job-id'
          ]
          driver = create_driver(config)
          job_ids = {
            'user-defined-job-id' => @job_id_value,
            'job_uuid' => @job_uuid_value
          }
          filtered_records = run(driver, @record)
          expect(filtered_records.size).to eq(1)
          expect(filtered_records[0]).to eq(@record.merge(job_ids))
        end

        it 'should correctly handle custom job uuid keys' do
          config = %[
            job_uuid_key 'dataproc-defined-job-uuid'
          ]
          driver = create_driver(config)
          job_ids = {
            'job_id' => @job_id_value,
            'dataproc-defined-job-uuid' => @job_uuid_value
          }
          filtered_records = run(driver, @record)
          expect(filtered_records.size).to eq(1)
          expect(filtered_records[0]).to eq(@record.merge(job_ids))
        end
      end
    end

    context 'for cache properties' do
      before(:example) do
        @resolved = DataprocJobIdResolver::ResolvedJobId.new(
          'job-1', '1234-5678-9012')
      end

      it 'should respect max cache size' do
        config = %[
          max_cache_size 1
        ]
        driver = create_driver(config)
        record_1_for_app_1 = {
          'message' => 'why am I putting messages again?',
          'app_id' => 'app_1'
        }
        record_2_for_app_1 = {
          'message' => 'I mean, I guess its for realism?',
          'app_id' => 'app_1'
        }
        record_3_for_app_1 = {
          'message' => 'Yean, thats what Ill say...',
          'app_id' => 'app_1'
        }
        record_for_app_2 = {
          'message' => 'Hopefully no one asks...',
          'app_id' => 'app_2'
        }
        expect(@job_id_resolver).to receive(:resolve).with('app_1').
            and_return(@resolved).twice
        expect(@job_id_resolver).to receive(:resolve).with('app_2').
            and_return(@resolved).once
        run(
            driver,
            record_1_for_app_1,
            record_for_app_2,
            record_2_for_app_1,
            record_3_for_app_1)
      end

      it 'should respect cache ttl' do
        config = %[
          cache_ttl_seconds 10
        ]
        driver = create_driver(config)
        record_1 = {
          'message' => 'earlier',
          'app_id' => 'app_1'
        }
        record_2 = {
          'message' => 'early',
          'app_id' => 'app_1'
        }
        record_3 = {
          'message' => 'later',
          'app_id' => 'app_1'
        }
        expect(@job_id_resolver).to receive(:resolve).with('app_1').
            and_return(@resolved).twice
        Timecop.freeze(Time.now)
        begin
          Timecop.freeze(Time.now)
          driver.run do
            driver.feed('test', event_time, record_1)
            Timecop.freeze(Time.now + 2)
            driver.feed('test', event_time, record_2)
            Timecop.freeze(Time.now + 11)
            driver.feed('test', event_time, record_3)
          end
        ensure
          Timecop.return
        end
      end
    end
  end
end
