#!/usr/bin/env rspec

require 'logger'
require 'rspec'
require 'webmock/rspec'
require_relative 'yarn_rest_client'


RSpec.describe Yarn::RestClient do

  def stub_expected_request_for_host_from_file(host, path, status)
    stub_request(:get, "http://#{host}:1337/ws/v1/cluster/apps/#{@app_id}").
        to_return(body: File.new(path), status: status)
  end

  def stub_expected_request_for_host_to_raise(host, exception)
    stub_request(:get, "http://#{host}:1337/ws/v1/cluster/apps/#{@app_id}").
        to_raise(exception)
  end

  before(:example) do
    # Disable out-bound network communications during tests
    WebMock.disable_net_connect!
    @app_id = 'application_1548790533163_0001'
    @rm_rest_port = 1337
  end

  context 'non-ha cluster' do
    before(:example) do
      @rm_host = 'my-cluster-m'
    end

    context 'non-kerberized cluster' do
      before(:example) do
        @yarn_client = Yarn::RestClient.new(
          [@rm_host], @rm_rest_port, false, nil)
      end

      def stub_expected_request_from_file(path, status)
        stub_expected_request_for_host_from_file(@rm_host, path, status)
      end

      def stub_expected_request_to_raise(exception)
        stub_expected_request_for_host_to_raise(@rm_host, exception)
      end

      it 'correctly returns a parsed json object on success' do
        stub_expected_request_from_file('data/yarn_app_normal.json', 200)
        json_app = @yarn_client.get_app(@app_id)
        expect(json_app).to_not be_nil
        expect(json_app[Yarn::ApplicationJsonConstants::APP_FIELD_NAME]).
            to_not be_nil
      end

      it 'returns nil on bad request' do
        stub_expected_request_from_file(
            'data/malformed_app_id_response.json', 400)
        expect(@yarn_client.get_app(@app_id)).to be_nil
      end

      it 'returns nil on redirect' do
        stub_expected_request_from_file('data/redirect_response.txt', 307)
        expect(@yarn_client.get_app(@app_id)).to be_nil
      end

      it 'returns nil on not found' do
        stub_expected_request_from_file('data/app_does_not_exist.json', 404)
        expect(@yarn_client.get_app(@app_id)).to be_nil
      end

      it 'returns nil on host not found' do
        stub_expected_request_to_raise(SocketError.
            new('Failed to open TCP connection to my-cluster-m:1337'))
        expect(@yarn_client.get_app(@app_id)).to be_nil
      end

      it 'returns nil on connection refused' do
        stub_expected_request_to_raise(Errno::ECONNREFUSED.
            new('Failed to open TCP connection to my-cluster-m:1337'))
        expect(@yarn_client.get_app(@app_id)).to be_nil
      end
    end

    context 'kerberized cluster' do
      before(:example) do
        @ca_file = '/etc/hadoop/conf/ssl/truststore.pem'
        @yarn_client = Yarn::RestClient.new(
            [@rm_host], @rm_rest_port, true, @ca_file)
      end

      def stub_expected_request_from_file(path, status)
        stub_request(
            :get, "https://#{@rm_host}:1337/ws/v1/cluster/apps/#{@app_id}").
          to_return(body: File.new(path), status: status)
      end

      def stub_expected_request_to_raise(exception)
        stub_request(
            :get, "https://#{@rm_host}:1337/ws/v1/cluster/apps/#{@app_id}").
          to_raise(exception)
      end

      it 'correctly returns a parsed json object on success' do
        stub_expected_request_from_file('data/yarn_app_normal.json', 200)
        json_app = @yarn_client.get_app(@app_id)
        expect(json_app).to_not be_nil
        expect(json_app[Yarn::ApplicationJsonConstants::APP_FIELD_NAME]).
            to_not be_nil
      end

      it 'returns nil on bad request' do
        stub_expected_request_from_file(
            'data/malformed_app_id_response.json', 400)
        expect(@yarn_client.get_app(@app_id)).to be_nil
      end

      it 'returns nil on redirect' do
        stub_expected_request_from_file('data/redirect_response.txt', 307)
        expect(@yarn_client.get_app(@app_id)).to be_nil
      end

      it 'returns nil on not found' do
        stub_expected_request_from_file('data/app_does_not_exist.json', 404)
        expect(@yarn_client.get_app(@app_id)).to be_nil
      end

      it 'returns nil on host not found' do
        stub_expected_request_to_raise(SocketError.
            new('Failed to open TCP connection to my-cluster-m:1337'))
        expect(@yarn_client.get_app(@app_id)).to be_nil
      end

      it 'returns nil on connection refused' do
        stub_expected_request_to_raise(Errno::ECONNREFUSED.
            new('Failed to open TCP connection to my-cluster-m:1337'))
        expect(@yarn_client.get_app(@app_id)).to be_nil
      end
    end
  end

  context 'ha cluster' do
    before(:example) do
      @rm_host_0 = 'my-cluster-m-0'
      @rm_host_1 = 'my-cluster-m-1'
      @yarn_client = Yarn::RestClient.new(
          [@rm_host_0, @rm_host_1], @rm_rest_port, false, nil)
    end

    context 'second host would succeed' do
      before(:example) do
        # Setup success from second host...
        stub_expected_request_for_host_from_file(
            @rm_host_1, 'data/yarn_app_normal.json', 200)
      end

      it 'doesnt hit second host when first succeeds' do
        stub_expected_request_for_host_from_file(
            @rm_host_0, 'data/yarn_app_normal.json', 200)
        json_app = @yarn_client.get_app(@app_id)
        expect(json_app).to_not be_nil
        expect(json_app[Yarn::ApplicationJsonConstants::APP_FIELD_NAME]).
            to_not be_nil
        expect(WebMock).not_to have_requested(:get, @rm_host_1)
      end

      it 'returns nil on bad request' do
        stub_expected_request_for_host_from_file(
            @rm_host_0, 'data/malformed_app_id_response.json', 400)
        expect(@yarn_client.get_app(@app_id)).to be_nil
        expect(WebMock).not_to have_requested(:get, @rm_host_1)
      end

      it 'returns successfuly from second host on first host redirect' do
        stub_expected_request_for_host_from_file(
            @rm_host_0, 'data/redirect_response.txt', 307)
        json_app = @yarn_client.get_app(@app_id)
        expect(json_app).to_not be_nil
        expect(json_app[Yarn::ApplicationJsonConstants::APP_FIELD_NAME]).
            to_not be_nil
      end

      it 'returns successfuly from second host on first host 404' do
        stub_expected_request_for_host_from_file(
          @rm_host_0, 'data/app_does_not_exist.json', 404)
        json_app = @yarn_client.get_app(@app_id)
        expect(json_app).to_not be_nil
        expect(json_app[Yarn::ApplicationJsonConstants::APP_FIELD_NAME]).
            to_not be_nil
      end

      it 'returns successfuly from second host on first host not found' do
        stub_expected_request_for_host_to_raise(@rm_host_0, SocketError.
            new('Failed to open TCP connection to my-cluster-m-1:1337'))
        json_app = @yarn_client.get_app(@app_id)
        expect(json_app).to_not be_nil
        expect(json_app[Yarn::ApplicationJsonConstants::APP_FIELD_NAME]).
            to_not be_nil
      end

      it 'returns successfuly from second host on first host conn refused' do
        stub_expected_request_for_host_to_raise(@rm_host_0, Errno::ECONNREFUSED.
            new('Failed to open TCP connection to my-cluster-m-1:1337'))
        json_app = @yarn_client.get_app(@app_id)
        expect(json_app).to_not be_nil
        expect(json_app[Yarn::ApplicationJsonConstants::APP_FIELD_NAME]).
            to_not be_nil
      end
    end
  end
end
