# Copyright 2019 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


module Yarn

  # A module for obtaining the value of YARN properties.
  module PropertiesProvider

    ADDRESS_PORT_REGEXP = /[^\/]+:(?<port>[0-9]+)/
    DEFAULT_RM_HTTP_PORT = 8088.freeze
    DEFAULT_RM_HTTPS_PORT = 8090.freeze

    # Get the port number for the YARN Resource Manager's REST https endpoint.
    #
    #  Attempts to read the yarn property 'yarn.resourcemanager.webapp.address'
    #  and parse the port number from the property value. If unable to read the
    #  property or a value was not set, defaults to 8088
    def self.rm_http_port
      yarn_rm_webapp_address = read_property_safe(
          'yarn.resourcemanager.webapp.address')
      matched_port = ADDRESS_PORT_REGEXP.match(yarn_rm_webapp_address)
      matched_port ? matched_port[:port].to_i : DEFAULT_RM_HTTP_PORT
    end

    # Get the port number for the YARN Resource Manager's REST https endpoint.
    #  Attempts to read the yarn property
    #  'yarn.resourcemanager.webapp.https.address' and parse the port number
    # from the property value. If unable to read the property or a value was not
    #  set, defaults to 8090
    def self.rm_https_port
      yarn_rm_webapp_address = read_property_safe(
          'yarn.resourcemanager.webapp.https.address')
      matched_port = ADDRESS_PORT_REGEXP.match(yarn_rm_webapp_address)
      matched_port ? matched_port[:port].to_i : DEFAULT_RM_HTTPS_PORT
    end

    def self.use_https
      read_property_safe('yarn.http.policy') == 'HTTPS_ONLY'
    end

    private

    def self.read_property_safe(property)
      `hdfs getconf -confKey #{property}`.to_s.strip
    rescue Errno::ENOENT => e
      # no hdfs command line tool on PATH
      $log.error(
          'Unable to use hdfs command line tool for reading properties.',
          error: e)
      ''
    end
  end
end
