# Copyright 2019 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

require 'json'
require 'net/http'
require_relative 'yarn_application_json_constants'


module Yarn

  # A simple YARN Resource Manager REST Client.
  class RestClient

    YARN_APPS_REST_PATH = '/ws/v1/cluster/apps'.freeze
    HTTP_TIMEOUT_SECONDS = 5.freeze

    # Initialize the RestClient to use the given list of RM hosts and port.
    def initialize(rm_hosts, rm_rest_port, use_https, ca_file)
      @log = $log
      @rm_hosts = rm_hosts
      @rm_rest_port = rm_rest_port
      @use_https = use_https
      @ca_file = ca_file
    end

    # Get a YARN Application's JSON definition by its Application ID.
    #
    #  This method returns a parsed JSON object of the YARN representation
    #  of an application. Fields should be accessed via
    #  Yarn::ApplicationJsonConstants keys.
    def get_app(app_id)
      # Try each Resource Manager hosts until we get a good, parsable response
      @rm_hosts.each do |rm_host|
        response = attempt_get_app(rm_host, app_id)
        case response
        when Net::HTTPBadRequest
          # If the App ID was malformed don't retry on other hosts, return
          # nil immediately.
          @log.error("Bad request to YARN Resource Manager: #{response.body}")
          return nil
        when Net::HTTPSuccess
          # If we got a successful response, attempt to return the parsed body.
          begin
            return JSON.parse(response.body)
          rescue JSON::ParserError => e
            @log.warn(
                'Error parsing http response from YARN Resource Manager.',
                error: e)
          end
        end
      end
      # We didn't get a successful, parsable response from any host, default nil
      nil
    end

    private

    def attempt_get_app(rm_host, app_id)
      Net::HTTP.start(rm_host, @rm_rest_port,
                      :read_timeout => HTTP_TIMEOUT_SECONDS,
                      :use_ssl => @use_https,
                      :ca_file => @ca_file) do |http|
        http.get("#{YARN_APPS_REST_PATH}/#{app_id}")
      end
    rescue SocketError => e
      @log.error "Unable to reach master node #{rm_host}.", error: e
      nil
    rescue Errno::ECONNREFUSED => e
      @log.error "Connection to resource manager on node #{rm_host} refused.",
                 error: e
      nil
    end
  end
end
