# Copyright 2018 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

set -euxo pipefail

# This script contains utility functions that install Conscrypt and generate
# related config templates java.security.default and java.security.conscrypt
# under /usr/local/share/google/dataproc. The config templates can be applied
# optionally during cluster initialization.

DATAPROC_DIR=${DATAPROC_DIR:-/usr/local/share/google/dataproc}
if is_arm64; then
  readonly CONSCRYPT_JAR="conscrypt-openjdk-2.6-8b3fb8c-linux-aarch_64.jar"
else
  readonly CONSCRYPT_JAR="conscrypt-openjdk-${CONSCRYPT_VERSION}-linux-x86_64.jar"
fi
readonly CONSCRYPT_JAR_URL="gs://dataproc-image-dependencies/conscrypt/${CONSCRYPT_JAR}"

# Customize Java security properties.
# These patches can be regenerated by `diff -u0 <original> <fixed>``
function get_default_java_security_patch() {
  cat <<'EOF'
--- conf/security/java.security 2022-08-16 07:26:11.000000000 +0000
+++ /usr/local/share/google/dataproc/java.security.default      2022-09-20 12:29:46.878891054 +0000
@@ -150 +150,12 @@
-securerandom.source=file:/dev/random
+################## Dataproc note
+# /dev/random blocks in virtualized environments due to lack of
+# good entropy sources, which makes SecureRandom use impractical.
+# In particular, that affects the performance of HTTPS that relies
+# on SecureRandom.'
+#
+# Due to that, /dev/urandom is used as the default.
+#
+# See http://www.2uo.de/myths-about-urandom/ for some background
+# on security of /dev/urandom on Linux.
+################## End Dataproc note
+securerandom.source=file:/dev/urandom
EOF
}

function get_conscrypt_java_security_patch() {
  cat <<'EOF'
--- conf/security/java.security 2022-08-16 07:26:11.000000000 +0000
+++ /usr/local/share/google/dataproc/java.security.conscrypt    2022-09-20 12:28:58.527835576 +0000
@@ -63,12 +63,16 @@
-security.provider.1=SUN
-security.provider.2=SunRsaSign
-security.provider.3=SunEC
-security.provider.4=SunJSSE
-security.provider.5=SunJCE
-security.provider.6=SunJGSS
-security.provider.7=SunSASL
-security.provider.8=XMLDSig
-security.provider.9=SunPCSC
-security.provider.10=JdkLDAP
-security.provider.11=JdkSASL
-security.provider.12=SunPKCS11
+################## Dataproc note
+# Set Conscrypt/BoringSSL as the default security provider.
+################## End Dataproc note
+security.provider.1=org.conscrypt.OpenSSLProvider
+security.provider.2=SUN
+security.provider.3=SunRsaSign
+security.provider.4=SunEC
+security.provider.5=SunJSSE
+security.provider.6=SunJCE
+security.provider.7=SunJGSS
+security.provider.8=SunSASL
+security.provider.9=XMLDSig
+security.provider.10=SunPCSC
+security.provider.11=JdkLDAP
+security.provider.12=JdkSASL
+security.provider.13=SunPKCS11
@@ -150 +154,12 @@
-securerandom.source=file:/dev/random
+################## Dataproc note
+# /dev/random blocks in virtualized environments due to lack of
+# good entropy sources, which makes SecureRandom use impractical.
+# In particular, that affects the performance of HTTPS that relies
+# on SecureRandom.'
+#
+# Due to that, /dev/urandom is used as the default.
+#
+# See http://www.2uo.de/myths-about-urandom/ for some background
+# on security of /dev/urandom on Linux.
+################## End Dataproc note
+securerandom.source=file:/dev/urandom
EOF
}

function _install_conscrypt() {
  mkdir "${DATAPROC_DIR}/conscrypt"
  pushd "${DATAPROC_DIR}/conscrypt"

  download_resource \
    "${CONSCRYPT_JAR_URL}" \
    "${DATAPROC_DIR}/conscrypt/${CONSCRYPT_JAR}" \
    "${DATAPROC_COMMON_LIB_DIR}/conscrypt.jar"

  # Extract native library from the jar, otherwise each java process will need
  # to extract it at runtime impacting performance and consuming disk storage
  if is_arm64; then
    local -r suffix='aarch_64'
  else
    local -r suffix='x86_64'
  fi

  local -r native_lib="META-INF/native/libconscrypt_openjdk_jni-linux-${suffix}.so"
  jar -xf "./${CONSCRYPT_JAR}" "${native_lib}"
  mv "${native_lib}" ./
  rm -rf "META-INF/"

  popd
}
