# Copyright 2018 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

set -euxo pipefail

# This script contains utility functions that install Conscrypt and generate
# related config templates java.security.default and java.security.conscrypt
# under /usr/local/share/google/dataproc. The config templates can be applied
# optionally during cluster initialization.

DATAPROC_DIR=${DATAPROC_DIR:-/usr/local/share/google/dataproc}
readonly CONSCRYPT_JAR="conscrypt-openjdk-${CONSCRYPT_VERSION}-linux-x86_64.jar"
readonly CONSCRYPT_DIR_URL="${OSS_MAVEN_REPO}/org/conscrypt/conscrypt-openjdk/${CONSCRYPT_VERSION}"
readonly CONSCRYPT_JAR_URL="${CONSCRYPT_DIR_URL}/${CONSCRYPT_JAR}"

# Customize Java security properties.
# These patches can be regenerated by `diff -u0 <original> <fixed>``
function get_default_java_security_patch() {
  cat <<'EOF'
--- security/java.security
+++ /usr/local/share/google/dataproc/java.security.default
@@ -117 +117,12 @@
-securerandom.source=file:/dev/random
+################## Dataproc note
+# /dev/random blocks in virtualized environments due to lack of
+# good entropy sources, which makes SecureRandom use impractical.
+# In particular, that affects the performance of HTTPS that relies
+# on SecureRandom.'
+#
+# Due to that, /dev/urandom is used as the default.
+#
+# See http://www.2uo.de/myths-about-urandom/ for some background
+# on security of /dev/urandom on Linux.
+################## End Dataproc note
+securerandom.source=file:/dev/urandom
EOF
}

function get_conscrypt_java_security_patch() {
  cat <<'EOF'
--- security/java.security
+++ /usr/local/share/google/dataproc/java.security.conscrypt
@@ -68,9 +68,14 @@
-security.provider.1=sun.security.provider.Sun
-security.provider.2=sun.security.rsa.SunRsaSign
-security.provider.3=sun.security.ec.SunEC
-security.provider.4=com.sun.net.ssl.internal.ssl.Provider
-security.provider.5=com.sun.crypto.provider.SunJCE
-security.provider.6=sun.security.jgss.SunProvider
-security.provider.7=com.sun.security.sasl.Provider
-security.provider.8=org.jcp.xml.dsig.internal.dom.XMLDSigRI
-security.provider.9=sun.security.smartcardio.SunPCSC
+#
+################## Dataproc note
+# Set Conscrypt/BoringSSL as the default security provider.
+################## End Dataproc note
+security.provider.1=org.conscrypt.OpenSSLProvider
+security.provider.2=sun.security.provider.Sun
+security.provider.3=sun.security.rsa.SunRsaSign
+security.provider.4=sun.security.ec.SunEC
+security.provider.5=com.sun.net.ssl.internal.ssl.Provider
+security.provider.6=com.sun.crypto.provider.SunJCE
+security.provider.7=sun.security.jgss.SunProvider
+security.provider.8=com.sun.security.sasl.Provider
+security.provider.9=org.jcp.xml.dsig.internal.dom.XMLDSigRI
+security.provider.10=sun.security.smartcardio.SunPCSC
@@ -117 +122,12 @@
-securerandom.source=file:/dev/random
+################## Dataproc note
+# /dev/random blocks in virtualized environments due to lack of
+# good entropy sources, which makes SecureRandom use impractical.
+# In particular, that affects the performance of HTTPS that relies
+# on SecureRandom.'
+#
+# Due to that, /dev/urandom is used as the default.
+#
+# See http://www.2uo.de/myths-about-urandom/ for some background
+# on security of /dev/urandom on Linux.
+################## End Dataproc note
+securerandom.source=file:/dev/urandom
EOF
}

function _install_conscrypt() {
  mkdir "${DATAPROC_DIR}/conscrypt"

  pushd "${DATAPROC_DIR}/conscrypt"
  curl -fsSO --retry-connrefused --retry 10 --retry-max-time 30 \
    "${CONSCRYPT_JAR_URL}"
  # Extract native library from the jar, otherwise each java process will need
  # to extract it at runtime impacting performance and consuming disk storage
  local -r native_lib="META-INF/native/libconscrypt_openjdk_jni-linux-x86_64.so"
  jar -xf "./${CONSCRYPT_JAR}" "${native_lib}"
  mv "${native_lib}" ./
  rm -rf "META-INF/"
  popd
}
