#!/usr/bin/env bash

# Script which helps launching Jupyter server in a container.
#
# Usage: jupyter-containerd-runner.sh
#
# The container image is from `dataproc:dataproc.spark.container.image` if
# the property is provided, otherwise from the `SPARK_CONTAINER_IMAGE_DEFAULT`
# env variable which is set in `/etc/environment` at image build time.

set -euo pipefail

source /etc/environment
source "${BDUTIL_DIR}/bdutil_components.sh"
source "${BDUTIL_DIR}/bdutil_logging.sh"
source "${BDUTIL_DIR}/bdutil_metadata.sh"
source "${BDUTIL_DIR}/bdutil_misc.sh"
source "${BDUTIL_DIR}/bdutil_retry.sh"
source "${BDUTIL_DIR}/bdutil_properties.sh"
source "${BDUTIL_DIR}/bdutil_versions.sh"
source "${BDUTIL_DIR}/components/shared/conda.sh"
source "${BDUTIL_DIR}/components/shared/containerd.sh"
source "${BDUTIL_DIR}/components/shared/jupyter.sh"
source "${BDUTIL_DIR}/components/shared/miniconda3.sh"
source "${BDUTIL_DIR}/components/shared/spark.sh"

# Always stop the tail -f command (see below) when we exit this script
trap 'kill_tail $?' EXIT

function main() {
  local -r container_name="jupyter-server"
  local -r container_type="jupyter-server"
  local -r working_dir="${JUPYTER_DATAPROC_DIR}"

  local -a envs
  envs=(
    "CUSTOM_CONFIGS_ENABLED=false"
    "CUSTOM_INIT_SCRIPT_ENABLED=false"
    "DEFAULT_PYSPARK_PYTHON=${CONDA_HOME}/bin/python"
    "JAVA_HOME=${JAVA_HOME}"
    "SPARK_HOME=${SPARK_HOME}"
  )

  local pypi_repo
  pypi_repo=$(get_dataproc_property internal.s8s.pypi.repository)
  if [[ -n "${pypi_repo}" ]]; then
    envs+=("DATAPROC_PYPI_REPOSITORY=${pypi_repo}")
  fi

  if is_component_selected 'nvidia-drivers'; then
    envs+=(
      "NVIDIA_DRIVER_CAPABILITIES=all"
      "NVIDIA_VISIBLE_DEVICES=all"
    )
  fi

  if [[ "${DATAPROC_IMAGE_VERSION}" == '1.2' ]] ||
      is_version_at_least "${DATAPROC_IMAGE_VERSION}" '2.2'; then
    envs+=("DEFAULT_LANG=C.UTF-8")
  fi

  local -a args
  args=(
    "${container_type}"
    "/bin/bash"
    "-c"
    "${CONDA_HOME}/bin/jupyter notebook"
  )

  # In addition to Jupyter itself, also need to mount Java, Conda, Spark into
  # the container, because Jupyter PySpark kernel needs to run spark-submit
  # inside the container.
  local -a mounts
  mounts=(
    # Conda is not portable. Some files in Conda recorded the original
    # installation path, so the path must be the same in the container.
    "${CONDA_HOME}:${CONDA_HOME}"
    "${JAVA_HOME}:${JAVA_HOME}"
    # ${JAVA_HOME}/lib/security/cacerts is a symbolic link to
    # /etc/ssl/certs/adoptium/cacerts in Temurin JDK.
    "/etc/ssl/certs/adoptium/:/etc/ssl/certs/adoptium/"
    "${SPARK_HOME}:${SPARK_HOME}"
    "${SPARK_CONF_DIR}:${SPARK_CONF_DIR}"
    "${SPARK_DATA_DIR}:${SPARK_DATA_DIR}"
    "${SPARK_LOG_DIR}:${SPARK_LOG_DIR}"
    "${JUPYTER_DATAPROC_DIR}:${JUPYTER_DATAPROC_DIR}"
    "${JUPYTER_ETC_DIR}:${JUPYTER_ETC_DIR}"
    "${JUPYTER_LOG_DIR}:${JUPYTER_LOG_DIR}"
    "${DATAPROC_COMMON_LIB_DIR}:${DATAPROC_COMMON_LIB_DIR}"
    "${BDUTIL_DIR}/spark-entrypoint.sh:/entrypoint.sh"
    "/home/spark/.ipython:/home/spark/.ipython"
  )

  # Mount the NFS local path to the container if it exists
  if [[ -v NFS_PACKAGE_MOUNT_LOCAL_PATH ]]; then
    envs+=("NFS_PACKAGE_MOUNT_LOCAL_PATH=${NFS_PACKAGE_MOUNT_LOCAL_PATH}")
    mounts+=(
      "${NFS_PACKAGE_MOUNT_LOCAL_PATH}:${NFS_PACKAGE_MOUNT_LOCAL_PATH}"
    )
  fi

  local image
  image=$(get_dataproc_property_or_default \
    "dataproc.spark.container.image" "${SPARK_CONTAINER_IMAGE_DEFAULT}")
  pull_image_using_containerd "${image}"

  create_containerd_config \
    "${working_dir}" \
    "${container_name}" \
    "${image}" \
    envs \
    args \
    mounts \
    "$(get_spark_container_memory)"

  # Stop and/or remove existing container if any.
  # This is necessary to make the script retriable.
  local container_id
  container_id=$(crictl ps --name "${container_name}" -q -a)
  local pod_id
  pod_id=$(crictl pods --name "${POD_NAME}" -q)
  if [[ -n "${container_id}" || -n "${pod_id}" ]]; then
    crictl stop "${container_id}" >/dev/null 2>&1 || true
    crictl rm "${container_id}" >/dev/null 2>&1 || true
    crictl stopp "${pod_id}" >/dev/null 2>&1 || true
    crictl rmp "${pod_id}" >/dev/null 2>&1 || true
  fi

  crictl runp "${POD_CONFIG}" >/dev/null
  pod_id=$(crictl pods --name "${POD_NAME}" -q)
  crictl create "${pod_id}" "${CONTAINER_CONFIG}" "${POD_CONFIG}" >/dev/null
  container_id=$(crictl ps --name "${container_name}" -q -a)
  crictl start "${container_id}" >/dev/null

  # Wait for log file creation by crictl start
  echo "Waiting for container log creation"
  retry_constant_short ls -l "${CONTAINERD_LOG_DIR}/${CONTAINERD_LOG_FILE}" &>/dev/null

  # Add read permission to the container log
  sudo /usr/local/bin/grant_container_log_read.sh

  # Stream the container log file to the Jupyter log file, removing the
  # timestamp and STDOUT/STDERR fields added by containerd.
  # We run tail -f on the container logs in the background and crictl attach in
  # the foreground. attach alone causes us to miss any logs before the call to
  # attach.  Likewise, tailing the log with `crictl logs` has several issues,
  # such as not emitting STDERR output and missing the last few log lines,
  # because it exits as soon as the container exits.  Combining tail with attach
  # solves these problems.  Below we also use the stdbuf command to not buffer
  # output (otherwise no output will be printed till buffer is full).
  tail -f "${CONTAINERD_LOG_DIR}/${CONTAINERD_LOG_FILE}" |
    stdbuf -oL cut -d " " -f 4- >>"${JUPYTER_LOG_FILE}" &
  TAIL_PID=$!
  crictl attach "${container_id}" &>/dev/null

  # crictl attach does not output the exit status of the container.  We must
  # instead get this from crictl inspect.
  echo "Waiting for container to stop"
  retry_quiet container_stopped "${container_id}"

  local exit_code
  exit_code=$(crictl inspect \
    --output go-template --template '{{.status.exitCode}}' "${container_id}")
  exit "$exit_code"
}

main "$@"
