r"""Generates various collectd conf files based on the service and the plugin types.

blaze build //cloud/hadoop/services/images/bdutil/monitoring_plugin_generator:all

blaze-bin/cloud/hadoop/services/images/bdutil/monitoring_plugin_generator/dataproc_monitoring_plugin_generator \
--metrics_file \
cloud/hadoop/services/images/bdutil/monitoring_plugin_generator/dataproc-default-metrics.txt \
--plugin_generator_config_dir \
cloud/hadoop/services/images/bdutil/monitoring_plugin_generator/ \
--output_dir cloud/hadoop/services/images/bdutil/conf/ \
--templates_dir cloud/hadoop/services/images/bdutil/monitoring_plugin_generator/ \
--output_file_name_kwargs yarn=collectd_yarn_jmx_metrics.conf \
hdfs=collectd_hdfs_jmx_metrics.conf spark=collectd_spark_yarn_metrics.conf \
--mode "default" \
--config_parameters_kwargs queue_name=default,dataproc-driverpool-driver-queue

"""

import argparse
import collections
import copy
import enum
import logging
import os
from typing import Any, Dict, Mapping, Sequence, Set

import jinja2
import yaml


class MonitoringAgent(enum.Enum):
  COLLECTD = "collectd"
  OPENTELEMETRY = "opentelemetry"


DEFAULT_JINJA_TEMPLATES_DIR = "."
DEFAULT_PLUGIN_GENERATOR_CONFIG_DIR = "."
DEFAULT_CONF_FILES_DIR = "../conf/collectd/gce"

## Assuming CollectD as monitoring agent
PLUGIN_GENERATOR_CONFIG_FILE_NAME = (
    "dataproc_collectd_jmx_plugin_generator_config.yaml"
)
JMX_PLUGIN_JINJA_TEMPLATE_FILE_NAME = "dataproc_collectd_jmx_plugin.conf.jinja"
SERVICE_TO_STATSD_PLUGIN_TEMPLATE_FILE_MAP = {
    "spark": "dataproc_collectd_statsd_spark_plugin.conf.jinja",
    "flink": "dataproc_collectd_statsd_flink_plugin.conf.jinja",
}
JMX_PLUGIN_OTEL_TEMPLATE_FILE_NAME = "dataproc_otel_jmx_plugin.yaml.jinja"

SERVICES_FOR_JMX_PLUGIN = [
    "hdfs",
    "sparkHistoryServer",
    "yarn",
    "hiveserver2",
    "hivemetastore",
]
SERVICE_FOR_DEFAULT_PLUGIN = ["monitoringAgentDefaults"]

SERVICE_TO_OUTPUT_FILE_NAME_MAP = {
    "hdfs": "collectd_hdfs_jmx_metrics.conf",
    "sparkHistoryServer": "collectd_shs_jmx_metrics.conf",
    "yarn": "collectd_yarn_jmx_metrics.conf",
    "hiveserver2": "collectd_hiveserver2_jmx_metrics.conf",
    "hivemetastore": "collectd_hivemetastore_jmx_metrics.conf",
    "spark": "collectd_spark_statsd_metrics.conf",
    "flink": "collectd_flink_statsd_metrics.conf",
    "cpu": "cpu.conf",
    "disk": "disk.conf",
    "memory": "memory.conf",
    "interface": "interface.conf",
    "network": "network.conf",
    "processes": "processes.conf",
}

# literals for keys
CONNECTIONS = "connections"
CONNECTION_NAME = "name"
GROUPS = "groups"
GROUP_NAME = "name"
MBEAN = "mbean"
LABEL = "label"
CONFIG_PARAMETER = "config_parameter"
METRICS = "metrics"
METRIC_ATTRIBUTE = "attribute"
METRIC_ATTRIBUTE_PREFIX = "prefix"


logging.basicConfig(
    format=(
        "%(asctime)s,%(msecs)d %(levelname)-8s [%(filename)s:%(lineno)d]"
        " %(message)s"
    ),
    datefmt="%Y-%m-%d:%H:%M:%S",
)


class PluginType(enum.Enum):
  UNKNOWN = 1
  JMX = 2
  STATSD = 3
  SYSTEM = 4


class Metric:
  """High level representation of the input metric strings.

  Attributes:
    service: The service whose metric is being collected.
    input: The argument that is used to initialize the Metric object.
    plugin_type: The plugin that is responsible for collecting the metric.
  """

  def __init__(self, metric_string):
    """Initializes the Metric object.

    Parses the metric string and populates the attributes.
    Args:
      metric_string: The full name of the metric that is passed as a string by
        the user.
    """
    self.service = metric_string.split(":")[0]
    self.input = metric_string
    if self.service in SERVICES_FOR_JMX_PLUGIN:
      self.plugin_type = PluginType.JMX
    elif self.service in SERVICE_TO_STATSD_PLUGIN_TEMPLATE_FILE_MAP:
      self.plugin_type = PluginType.STATSD
    elif self.service in SERVICE_FOR_DEFAULT_PLUGIN:
      self.plugin_type = PluginType.SYSTEM
    else:
      self.plugin_type = PluginType.UNKNOWN

  def __str__(self):
    return (
        "Metric: { service: "
        + self.service
        + " , plugin_type : "
        + str(self.plugin_type)
        + " , input : "
        + self.input
        + " }"
    )


class SystemMetric(Metric):
  """Class representing System Metrics.

  System metrics are the health metrics pertaining to the instance that is being
  used.

  Attributes:
    prefix: Prefix indicates the component of the system that the metric belongs
      to. This is the division level that is publicly visible.
    metric: The name of the metric.
    plugin: Plugin that pushes the metric. This is relevant for the collectd
      config, and does not have a 1-1 correspondence with metric prefix.
    type_regex: The target regex corresponding to the metric. The combination of
      type_regex and plugin is able to uniquely identify the metric.
  """

  def __init__(self, metric_string):
    """Initializes the SystemMetric object.

    Parses the metric string and populates the attributes based on predefined
    dictionaries.

    Args:
      metric_string: The full name of the metric that is passed as a string by
        the user.
    """
    super().__init__(metric_string)
    self.prefix = metric_string.split(":")[1].strip()
    self.metric = metric_string.split(":")[2].strip()

    full_metric = f"{self.prefix}/{self.metric}"

    metric_to_plugin_dict = {
        "cpu/utilization": "cpu",
        "cpu/load_1m": "load",
        "cpu/load_5m": "load",
        "cpu/load_15m": "load",
        "disk/bytes_Used": "df",
        "disk/percent_Used": "df",
        "disk/read_Bytes_Count": "disk",
        "disk/write_Bytes_Count": "disk",
        "disk/operation_Count": "disk",
        "disk/operation_Time": "disk",
        "disk/merged_Operations": "disk",
        "disk/pending_Operations": "disk",
        "disk/io_Time": "disk",
        "disk/weighted_Io_Time": "disk",
        "interface/traffic": "interface",
        "interface/packets": "interface",
        "interface/errors": "interface",
        "memory/bytes_Used": "memory",
        "memory/percent_Used": "memory",
        "network/tcp_Connections": "tcpconns",
        "processes/count_By_State": "processes",
        "processes/fork_Count": "processes",
        "processes/vm_Usage": "processes",
        "processes/rss_Usage": "processes",
        "processes/cpu_Time": "processes",
        "processes/disk": "processes",
    }
    if full_metric in metric_to_plugin_dict:
      self.plugin = metric_to_plugin_dict[full_metric]

    metric_to_type_dict = {
        "cpu/utilization": "percent",
        "cpu/load_1m": "load",
        "cpu/load_5m": "load",
        "cpu/load_15m": "load",
        "disk/bytes_Used": "df_complex",
        "disk/percent_Used": "percent_bytes",
        "disk/read_Bytes_Count": "disk_octets",
        "disk/write_Bytes_Count": "disk_octets",
        "disk/operation_Count": "disk_ops",
        "disk/operation_Time": "disk_time",
        "disk/merged_Operations": "disk_merged",
        "disk/pending_Operations": "pending_operations",
        "disk/io_Time": "disk_io_time",
        "disk/weighted_Io_Time": "disk_io_time",
        "interface/traffic": "if_octets",
        "interface/packets": "if_packets",
        "interface/errors": "if_errors",
        "memory/bytes_Used": "memory",
        "memory/percent_Used": "percent",
        "network/tcp_Connections": "tcp_connections",
        "processes/count_By_State": "ps_state",
        "processes/fork_Count": "fork_rate",
        "processes/vm_Usage": "ps_vm",
        "processes/rss_Usage": "ps_rss",
        "processes/cpu_Time": "ps_cputime",
        "processes/disk": "disk_octets",
    }
    if full_metric in metric_to_type_dict:
      self.type_regex = metric_to_type_dict[full_metric]

  @staticmethod
  def is_valid(metric_string: str = "") -> bool:
    """Determines if the metric provided by the user has the correct number of ":" separated details.

    Args:
      metric_string: The full name of the metric that is passed as a string by
        the user.

    Returns:
      True if the metric is valid, and false otherwise.
    """

    return len(metric_string.split(":")) == 3

  def get_help_message(self) -> str:
    """Provides a template to help the user understand the correct format in which to provide the metric.

    Returns:
      The helper message in string format.
    """

    return (
        "Collectd metrics should be given in the format "
        "service:prefix:metric. "
        "e.g.: ["
        "monitoringAgentDefaults:memory:bytes_Used , "
        "monitoringAgentDefaults:interface:errors ]"
    )

  def __str__(self):
    """Returns:String representation of the metric."""
    return (
        f"SystemMetric : {{ service :{self.service} , prefix: {self.prefix} ,"
        f" metric:{self.metric}}}"
    )


class JmxMetric(Metric):
  """Class representing Jmx Metrics."""

  def __init__(self, metric_string):
    super().__init__(metric_string)
    self.connection = metric_string.split(":")[1].strip()
    self.mbean = metric_string.split(":")[2].strip()
    self.metric = metric_string.split(":")[3].strip()

  @staticmethod
  def is_valid(metric_string):
    if len(metric_string.split(":")) == 4:
      return True
    return False

  @staticmethod
  def get_help_message():
    return (
        "JMX metrics should be given in the format "
        "service:connection:mbeanName:attribute. "
        "e.g.: ["
        "hdfs:NameNode:FSNamesystem:ExpiredHeartbeats , "
        "yarn:ResourceManager:ClusterMetrics:NumActiveNMs ]"
    )

  @staticmethod
  def get_lookup_map(jmx_metric_objs):
    metric_to_is_valid_nested_map = nested_dict()
    for m in jmx_metric_objs:
      metric_to_is_valid_nested_map[m.service][m.connection][m.mbean][
          m.metric
      ] = False
    return metric_to_is_valid_nested_map

  @staticmethod
  def get_invalid_metrics(lookup_nested_dict, metric_string=""):
    invalid_metrics = []
    for k, v in lookup_nested_dict.items():
      if isinstance(v, dict):
        invalid_metrics.extend(
            JmxMetric.get_invalid_metrics(v, metric_string + ":" + k)
        )
      elif not v:
        invalid_metrics.append(metric_string.lstrip(":") + ":" + k)
    return invalid_metrics

  @staticmethod
  def log_warning_for_invalid_jmx_metrics_if_any(metric_to_is_valid_nested_map):
    """Logs warning string for invalid input metric strings.

    Args:
      metric_to_is_valid_nested_map: Nested dict created for input metric
        strings of format service:connection:mbean:metric
    """
    invalid_metrics = JmxMetric.get_invalid_metrics(
        metric_to_is_valid_nested_map
    )

    if invalid_metrics:
      err = ", ".join(invalid_metrics)
      logging.warning(
          "Following input metrics are invalid/excluded for this run. Check"
          " against the configuration present at %s/%s. %s",
          DEFAULT_PLUGIN_GENERATOR_CONFIG_DIR,
          PLUGIN_GENERATOR_CONFIG_FILE_NAME,
          err,
      )

  def __str__(self):
    return (
        "JmxMetric : { service :"
        + self.service
        + " , connection: "
        + self.connection
        + " , mbean:"
        + self.mbean
        + " , metric:"
        + self.metric
        + " }"
    )


class ParseKwargs(argparse.Action):

  def __call__(self, parser, namespace, values, option_string=None):
    setattr(namespace, self.dest, dict())
    for value in values:
      key, value = value.split("=")
      getattr(namespace, self.dest)[key] = value


def nested_dict():
  return collections.defaultdict(nested_dict)


def get_plugin_config_for_input_metrics(
    service,
    plugin_config_for_service,
    metric_to_is_valid_nested_map,
    config_parameters_kv_map,
    monitoring_agent: str,
):
  """Returns the pruned plugin generator config to be passed to the jinja template generation for the input metric passed to the script.

  Also updates the map metric_to_is_valid_nested_map with True if the
  input metric was found in the plugin_config_for_service.
  Args:
    service: service for which the config is being fetched.
    plugin_config_for_service: plugin generator configuration for the service
      read from the config yaml file
    metric_to_is_valid_nested_map: Nested dict for the input metric per service
      in format [connection][mbean][metric] = boolean. The value denotes whether
      matching configuration was found for the metric in
      plugin_config_for_service. All values will be False since the
      configuration is yet to be evaluated.
    config_parameters_kv_map: Dict with values for the config parameters.
    monitoring_agent: Monitoring agent to assume (OpenTelemetry or CollectD)
  """
  cloned_plugin_config = copy.deepcopy(plugin_config_for_service)
  filtered_connections = []
  for connection_metadata in cloned_plugin_config[CONNECTIONS]:
    connection_name = connection_metadata[CONNECTION_NAME]
    if connection_name in metric_to_is_valid_nested_map:
      filtered_groups = []
      for group in connection_metadata[GROUPS]:
        group_name = group[GROUP_NAME]
        if group_name in metric_to_is_valid_nested_map[connection_name]:
          filtered_metrics = get_filtered_metrics(
              service,
              connection_name,
              group,
              config_parameters_kv_map,
              metric_to_is_valid_nested_map[connection_name][group_name],
              monitoring_agent,
          )
          if filtered_metrics:
            filtered_groups.append(group)
            group[METRICS] = filtered_metrics
      if filtered_groups:
        filtered_connections.append(connection_metadata)
        connection_metadata[GROUPS] = filtered_groups
  if filtered_connections:
    cloned_plugin_config[CONNECTIONS] = filtered_connections
    return cloned_plugin_config
  return None


def get_filtered_metrics(
    service: str,
    connection_name: str,
    group: Dict[str, Any],
    config_parameters_kv_map: Dict[str, str],
    metric_to_is_valid_map: Dict[str, bool],
    monitoring_agent: str,
) -> ...:
  """Returns a list of metrics that will be passed to the jinja template generation.

  Also updates the input map with True if the
  input metric was found in the given subset group of metrics.
  Args:
    service: service for which the metrics belong.
    connection_name: connection name of the group to be evaluated
    group: Dict of the group for which the metrics belong to
    config_parameters_kv_map: Dict with values for the config parameters.
    metric_to_is_valid_map: Dict for the input metric in format [metric] =
      boolean. This is the map obtained from the nested map at
      metric_to_is_valid_nested_map[connection][mbean]. The value denotes
      whether matching configuration was found for the metric in
      plugin_config_for_service. All values will be False since the
      configuration is yet to be evaluated.
    monitoring_agent: Monitoring agent to assume (OpenTelemetry or CollectD)
  """
  filtered_metrics = []
  if monitoring_agent == MonitoringAgent.OPENTELEMETRY.value and group.get(
      "composite_data"
  ):
    attribute_metric_map = {}
    for metric in copy.copy(group[METRICS]):
      attribute = metric[METRIC_ATTRIBUTE].split(".")[0]
      attribute_metric_map[attribute] = {
          **metric,
          "name": attribute,
          "attribute": attribute,
      }
      group[METRICS].remove(metric)

    filtered_metrics = list(attribute_metric_map.values())

  for metric in group[METRICS]:
    if METRIC_ATTRIBUTE_PREFIX in metric:
      matching_attributes = [
          attribute
          for attribute in metric_to_is_valid_map.keys()
          if attribute.startswith(metric[METRIC_ATTRIBUTE_PREFIX])
      ]
      for attribute in matching_attributes:
        matching_metric = copy.deepcopy(metric)
        matching_metric[METRIC_ATTRIBUTE] = attribute
        matching_metric.pop(METRIC_ATTRIBUTE_PREFIX)
        if not has_valid_config_parameters(group, config_parameters_kv_map):
          logging.warning(
              "Ignoring %s:%s:%s:%s because the config parameter %s "
              "is not passed to the script",
              service,
              connection_name,
              group[GROUP_NAME],
              attribute,
              group[CONFIG_PARAMETER],
          )
          continue
        filtered_metrics.append(matching_metric)
        metric_to_is_valid_map[attribute] = True
    elif (
        METRIC_ATTRIBUTE in metric
        and metric[METRIC_ATTRIBUTE] in metric_to_is_valid_map
    ):
      if not has_valid_config_parameters(group, config_parameters_kv_map):
        logging.warning(
            "Ignoring %s:%s:%s:%s because the config parameter %s "
            "is not passed to the script",
            service,
            connection_name,
            group[GROUP_NAME],
            metric[METRIC_ATTRIBUTE],
            group[CONFIG_PARAMETER],
        )
        continue
      filtered_metrics.append(metric)
      metric_to_is_valid_map[metric[METRIC_ATTRIBUTE]] = True

  return filtered_metrics


def has_valid_config_parameters(group, config_parameters_kv_map):
  if CONFIG_PARAMETER not in group:
    return True
  if (
      config_parameters_kv_map is None
      or group[CONFIG_PARAMETER] not in config_parameters_kv_map
  ):
    return False
  return True


def write_config_file(
    service, rendered_config, output_dir, service_to_output_file_name_map
):
  """Writes the rendered string from jinja template to the output file.

  Args:
    service: Service for which the plugin configuration needs to be written.
    rendered_config: Rendered string from Jinja template
    output_dir: Directory where the plugin configuration should be written.
    service_to_output_file_name_map: Map of service to the output file name.
  """
  output_file_name = None
  if (
      service_to_output_file_name_map is None
      or service not in service_to_output_file_name_map
  ):
    output_file_name = SERVICE_TO_OUTPUT_FILE_NAME_MAP[service]
    logging.warning(
        "Output file name not provided for %s. Assuming default: %s",
        service,
        output_file_name,
    )
  else:
    output_file_name = service_to_output_file_name_map[service]
  f = open(output_dir + "/" + output_file_name, "w")
  f.write(rendered_config)
  f.close()


def apply_config_parameters(filtered_config, config_parameters_kv_map):
  """Substitutes the config parameter placeholders in the config.

  Eg: a metric with mbean having "queue_name" config_parameter like this,
  "Hadoop:name=QueueMetrics,q0=root,q1={{queue_name}},service=ResourceManager"
  and config_parameters_kv_map = { "queue_name" : "default,driver-pool" }
  will be cloned into two different metrics as follows :
  "Hadoop:name=QueueMetrics,q0=root,q1=default,service=ResourceManager",
  "Hadoop:name=QueueMetrics,q0=root,q1=dataproc-driverpool-driver-queue,service=ResourceManager"

  Args:
    filtered_config: config which needs to be substituted with config
      parameters.
    config_parameters_kv_map: map of config parameters and its value.
  """

  if not config_parameters_kv_map:
    return
  for connection_metadata in filtered_config[CONNECTIONS]:
    processed_group = []
    for group in connection_metadata[GROUPS]:
      if CONFIG_PARAMETER not in group:
        processed_group.append(group)
        continue
      config_parameter = group[CONFIG_PARAMETER]
      values_for_config_parameter = config_parameters_kv_map[
          config_parameter
      ].split(",")
      for v in values_for_config_parameter:
        cloned_group = copy.deepcopy(group)
        cloned_group[GROUP_NAME] = cloned_group[GROUP_NAME] + "_" + v
        cloned_group[MBEAN] = cloned_group[MBEAN].replace(
            "{{" + config_parameter + "}}", v
        )
        cloned_group[LABEL] = cloned_group[LABEL].replace(
            "{{" + config_parameter + "}}", v
        )
        processed_group.append(cloned_group)
    connection_metadata[GROUPS] = processed_group


def generate_jmx_plugin_config_file(
    jmx_metric_objs: Sequence[JmxMetric],
    plugin_generator_config,
    templates_dir,
    output_dir,
    service_to_output_file_name_map,
    config_parameters_kv_map,
    services_to_be_excluded,
    mode,
    monitoring_agent=MonitoringAgent.COLLECTD.value,
):
  """Generates and writes the jmx plugin configuration files for metric_objs.

  Args:
    jmx_metric_objs: Input metrics of type `JmxMetric`.
    plugin_generator_config: configuration used to map the input metric to the
      template input needed for generating the plugin config.
    templates_dir: Directory containing the jinja templates for service that
      come under jmx plugin.
    output_dir: Directory where the generated jmx plugin configuration files are
      to be stored.
    service_to_output_file_name_map: Map of service to the output file name to
      be used for the generated jmx plugin configuration files.
    config_parameters_kv_map: map of config parameters and their values.
    services_to_be_excluded: services to be excluded from the statsd plugin
      generation.
    mode: Mode to be substituted in the generated plugin configuration file to
      different between the files generated during various phases - like image
      build/image startup, etc.
    monitoring_agent: Monitoring agent to assume (OpenTelemetry or CollectD)
  """
  env = jinja2.Environment(
      loader=jinja2.FileSystemLoader(templates_dir),
      trim_blocks=True,
      keep_trailing_newline=True,
      lstrip_blocks=True,
      extensions=["jinja2.ext.do"],
  )
  template = env.get_template(JMX_PLUGIN_JINJA_TEMPLATE_FILE_NAME)

  metric_to_is_valid_nested_map = JmxMetric.get_lookup_map(jmx_metric_objs)

  for service in metric_to_is_valid_nested_map.keys():
    if service in services_to_be_excluded:
      continue
    filtered_config = get_plugin_config_for_input_metrics(
        service,
        plugin_generator_config[service],
        metric_to_is_valid_nested_map[service],
        config_parameters_kv_map,
        monitoring_agent,
    )
    if filtered_config is None:
      logging.warning("None of the metrics in % is valid. Skipping.", service)
      continue

    apply_config_parameters(filtered_config, config_parameters_kv_map)
    rendered_config = template.render(
        service=service,
        service_prefix=plugin_generator_config[service]["prefix"],
        config=filtered_config,
        mode=mode,
    )

    write_config_file(
        service, rendered_config, output_dir, service_to_output_file_name_map
    )
  JmxMetric.log_warning_for_invalid_jmx_metrics_if_any(
      metric_to_is_valid_nested_map
  )
  pass


def generate_otel_jmx_plugin_config_file(
    jmx_metric_objs: Sequence[JmxMetric],
    plugin_generator_config: Dict[str, Any],
    templates_dir: str,
    output_dir: str,
    service_to_output_file_name_map: Mapping[str, str],
    config_parameters_kv_map: Dict[str, str],
    services_to_be_excluded: Set[str],
    mode: str,
    monitoring_agent=MonitoringAgent.COLLECTD.value,
):
  """Generates and writes the jmx plugin configuration files for metric_objs.

  Args:
    jmx_metric_objs: Input metrics of type `JmxMetric`.
    plugin_generator_config: configuration used to map the input metric to the
      template input needed for generating the plugin config.
    templates_dir: Directory containing the jinja templates for service that
      come under jmx plugin.
    output_dir: Directory where the generated jmx plugin configuration files are
      to be stored.
    service_to_output_file_name_map: Map of service to the output file name to
      be used for the generated jmx plugin configuration files.
    config_parameters_kv_map: map of config parameters and their values.
    services_to_be_excluded: services to be excluded from the statsd plugin
      generation.
    mode: Mode to be substituted in the generated plugin configuration file to
      different between the files generated during various phases - like image
      build/image startup, etc.
    monitoring_agent: Monitoring agent to assume (OpenTelemetry or CollectD)
  """
  env = jinja2.Environment(
      loader=jinja2.FileSystemLoader(templates_dir),
      trim_blocks=True,
      keep_trailing_newline=True,
      lstrip_blocks=True,
      extensions=["jinja2.ext.do"],
  )
  template = env.get_template(JMX_PLUGIN_OTEL_TEMPLATE_FILE_NAME)

  metric_to_is_valid_nested_map = JmxMetric.get_lookup_map(jmx_metric_objs)
  jmx_configs = []
  for service in copy.copy(metric_to_is_valid_nested_map).keys():
    if service in services_to_be_excluded:
      continue

    filtered_config = get_plugin_config_for_input_metrics(
        service,
        plugin_generator_config[service],
        metric_to_is_valid_nested_map[service],
        config_parameters_kv_map,
        monitoring_agent,
    )

    if filtered_config is None:
      logging.warning("None of the metrics in % is valid. Skipping.", service)
      continue

    jmx_configs.append({
        "service": service,
        "service_prefix": plugin_generator_config[service]["prefix"],
        **filtered_config,
    })

    rendered_config = template.render(
        config=jmx_configs,
        mode=mode,
    )

    write_config_file(
        "otel_jmx_plugin",
        rendered_config,
        output_dir,
        service_to_output_file_name_map,
    )


def generate_statsd_plugin_config_file(
    metric_objs: Sequence[Metric],
    templates_dir,
    output_dir,
    service_to_output_file_name_map,
    services_to_be_excluded,
    mode,
):
  """Generates and writes the statsd plugin configuration files for metric_objs.

  Args:
    metric_objs: Input metrics of type `Metric`
    templates_dir: Directory containing the jinja templates for service that
      come under statsd plugin.
    output_dir: Directory where the generated statsd plugin configuration files
      are to be stored.
    service_to_output_file_name_map: Map of service to the output file name to
      be used for the generated statsd plugin configuration files.
    services_to_be_excluded: services to be excluded from the statsd plugin
      generation.
    mode: Mode to be substituted in the generated plugin configuration file to
      different between the files generated during various phases - like image
      build/image startup, etc
  """
  service_to_metrics_map = collections.defaultdict(list)
  for m in metric_objs:
    # m.input looks like spark:driver:DAGScheduler:job.allJobs for driver and
    # spark:executor:executor:bytesRead for executor. We want to pass the
    # different components of the metric name as separate strings to the
    # jinja template
    metric_type_instance = m.input.split(":")
    service_to_metrics_map[m.service].append(metric_type_instance)

  for service in service_to_metrics_map:
    if service in services_to_be_excluded:
      continue
    if service not in SERVICE_TO_STATSD_PLUGIN_TEMPLATE_FILE_MAP.keys():
      logging.warning(
          "Template filename not found. %s is not configured in the generator",
          service,
      )
      continue

    template_file = SERVICE_TO_STATSD_PLUGIN_TEMPLATE_FILE_MAP[service]
    env = jinja2.Environment(
        loader=jinja2.FileSystemLoader(templates_dir),
        keep_trailing_newline=True,
        trim_blocks=True,
        lstrip_blocks=True,
        extensions=["jinja2.ext.do"],
    )
    template = env.get_template(template_file)

    rendered_config = template.render(
        metrics=list(service_to_metrics_map[service]), mode=mode
    )

    write_config_file(
        service, rendered_config, output_dir, service_to_output_file_name_map
    )
  pass


def handle_system_metrics(
    pref: str,
    metrics_list: Sequence[SystemMetric],
    conf_files_dir: str,
    output_dir: str,
    service_to_output_file_name_map: Mapping[str, str],
) -> None:
  """Handles creation of config files for disk and cpu prefixes.

  Args:
    pref: The prefix domain of the metric.
    metrics_list: The metrics that need to be ingested for the specified prefix.
    conf_files_dir: The directory in which the template config files for
      different prefixes is located.
    output_dir: The directory in which the created config files need to be put.
    service_to_output_file_name_map: The file names for the created config
      files.
  """
  # prefix_to_regex_map contains a set of all target regexes (each
  # corresponding to a metric) corresponding to different plugins.
  # Sorting rearranges this set in an ordered format so that the
  # string obtained by concatenating the target regexes has a specific order.
  # This order does not affect functionality of the config file, but removes the
  # flakiness of the test associated to proper creation of the config file.
  prefix_to_type_map = collections.defaultdict(set)
  prefix_to_target_regex_map = {}

  # stub string "STUB_${prefix}" placed in the collectd file is replaced with a
  # regex string created according to the metrics selected and then written to
  # output directory.
  for metric in metrics_list:
    plugin = metric.plugin
    target_regex = metric.type_regex
    prefix_to_type_map[plugin].add(target_regex)

  for plugin in prefix_to_type_map:
    prefix_to_target_regex_map[plugin] = "|".join(
        sorted(prefix_to_type_map[plugin])
    )
    if len(prefix_to_type_map[plugin]) > 1:
      prefix_to_target_regex_map[plugin] = (
          "(" + prefix_to_target_regex_map[plugin] + ")"
      )

  with open(conf_files_dir + "/" + SERVICE_TO_OUTPUT_FILE_NAME_MAP[pref]) as r:
    text = r.read()
    for plugin in prefix_to_type_map:
      text = text.replace(
          "STUB_" + plugin.upper(), prefix_to_target_regex_map[plugin]
      )
  write_config_file(pref, text, output_dir, service_to_output_file_name_map)


def generate_monitoring_plugin_config_file(
    system_metric_objs: Sequence[SystemMetric],
    output_dir: str,
    conf_files_dir: str,
    default_collectd_file: str,
    service_to_output_file_name_map: Mapping[str, str],
) -> None:
  """Generates and writes the collectd plugin configuration files for metric_objs.

  Args:
    system_metric_objs: Input metrics of type `SystemMetric`.
    output_dir: Directory where the generated collectd plugin configuration
      files are to be stored.
    conf_files_dir: Directory where template config files are stored.
    default_collectd_file: The preinstalled collectd plugin file that pushes all
      system metrics.
    service_to_output_file_name_map: Map of service to the output file name to
      be used for the generated collectd plugin configuration files.
  """

  # replacing the default collectd file with minimal configuration config file.
  with open(conf_files_dir + "/dataproc_gce_default.conf", "r") as r:
    text = r.read()
  with open(default_collectd_file, "w") as w:
    w.write(text)

  # creating a dict that maps prefixes to their selected metrics.
  prefix_to_metrics_map = collections.defaultdict(list)
  for m in system_metric_objs:
    if hasattr(m, "plugin"):
      prefix_to_metrics_map[m.prefix].append(m)
    else:
      logging.warning(
          "The metric %s is not supported by stackdriver-agent", m.input
      )

  # edit the collectd files for the metrics selected in overrides and write them
  # in the output folder. Cpu and Disk prefixes are handled separately because
  # they contain metrics corresponding to multiple plugins.
  for pref in prefix_to_metrics_map:
    handle_system_metrics(
        pref,
        prefix_to_metrics_map[pref],
        conf_files_dir,
        output_dir,
        service_to_output_file_name_map,
    )


def parse_args():
  """Returns the parsed argument passed to the script.

  --agent : Targeting agent to generate output files, one of "opentelemetry" or
  "collectd", default: "collectd".
  --metrics_file : a valid file with the input
  metrics to be overridden.
  --templates_dir : Directory containing the jinja
  templates for various plugins to be generated.
  --plugin_generator_config_dir : Directory containing the yaml configuration
  file needed for the plugin-gen to
  work.
  --conf_files_dir : Directory containing the template system
  collectd.conf files.
  --default_collectd_file : The preinstalled collectd
  plugin file that pushes all system metrics
  --output_dir : Directory where the
  generated plugin configuration files should be stored.
  --output_file_name_kwargs : Multiple k=v separated by space where k is the
  service and v is the generated plugin output file name.
  --exclude_services : services to exclude for plugin generation.
  --include_services : services to include for plugin generation.
  --mode : default/override - to differentiate the output files generated
  at various stages (build image/image startup/etc).
  --config_parameters_kwargs : Multiple k=v separated by space where k is the
  variable at the configuration yaml and v is the one to be substituted with.

  Raises:
    ArgumentError
  """
  parser = argparse.ArgumentParser(
      description=(
          "Generate collectd configuration files to collect dataproc OSS"
          " metrics"
      )
  )
  parser.add_argument("--metrics_file", type=open, required=True)
  parser.add_argument(
      "--templates_dir", type=str, default=DEFAULT_JINJA_TEMPLATES_DIR
  )
  parser.add_argument(
      "--plugin_generator_config_dir",
      type=str,
      default=DEFAULT_PLUGIN_GENERATOR_CONFIG_DIR,
  )
  parser.add_argument(
      "--conf_files_dir", type=str, default=DEFAULT_CONF_FILES_DIR
  )
  parser.add_argument(
      "--default_collectd_file",
      type=str,
      default="/etc/stackdriver/collectd.conf",
  )
  parser.add_argument("--output_dir", type=str, required=True)
  parser.add_argument(
      "--output_file_name_kwargs", nargs="*", action=ParseKwargs
  )
  parser.add_argument(
      "--config_parameters_kwargs", nargs="*", action=ParseKwargs
  )
  group = parser.add_mutually_exclusive_group()
  group.add_argument("--exclude_services", nargs="?", const=None)
  group.add_argument("--include_services", nargs="?", const=None)
  parser.add_argument("--mode", type=str, default="default")
  parser.add_argument(
      "--agent", type=str, default=MonitoringAgent.COLLECTD.value
  )
  return parser.parse_args()


def get_metric_objs(input_metric_file):
  return [Metric(e.strip()) for e in input_metric_file.readlines()]


def get_metric_objs_for_jmx_with_invalid_format(metric_objs):
  return [
      e
      for e in metric_objs
      if e.plugin_type == PluginType.JMX and not JmxMetric.is_valid(e.input)
  ]


def get_services_to_be_excluded(args):
  """Returns the services to be excluded for the run by reading the program arguments."""
  if args.exclude_services is not None:
    return {s.strip() for s in args.exclude_services.split(",")}
  elif args.include_services is not None:
    services_to_be_included = {
        s.strip() for s in args.include_services.split(",")
    }
    return SERVICE_TO_OUTPUT_FILE_NAME_MAP.keys() - services_to_be_included
  return set()


def load_plugin_generator_config(config_dir):
  try:
    with open(config_dir + "/" + PLUGIN_GENERATOR_CONFIG_FILE_NAME) as file:
      return yaml.safe_load(file)
  except FileNotFoundError as e:
    logging.error(
        "Expecting config yaml file %s at the directory %s. Got error: %s",
        PLUGIN_GENERATOR_CONFIG_FILE_NAME,
        config_dir,
        e.strerror,
    )
    exit(1)


def apply_opentelemetry_overrides():
  """Applies opentelemetry overrides to conf related constants."""
  global JMX_PLUGIN_JINJA_TEMPLATE_FILE_NAME
  global SERVICE_TO_STATSD_PLUGIN_TEMPLATE_FILE_MAP
  global SERVICE_TO_OUTPUT_FILE_NAME_MAP

  JMX_PLUGIN_JINJA_TEMPLATE_FILE_NAME = "dataproc_otel_jmx.groovy.jinja"

  SERVICE_TO_STATSD_PLUGIN_TEMPLATE_FILE_MAP = {
      "spark": "dataproc_otel_spark_config.yaml.jinja",
      "flink": "dataproc_otel_flink_config.yaml.jinja",
  }

  SERVICE_TO_OUTPUT_FILE_NAME_MAP = {
      "spark": "otel_spark_config.yaml",
      "flink": "otel_flink_config.yaml",
      "yarn": "otel_yarn_config.groovy",
      "hdfs": "otel_hdfs_config.groovy",
      "sparkHistoryServer": "otel_shs_config.groovy",
      "hiveserver2": "otel_hs2_config.groovy",
      "hivemetastore": "otel_hms_config.groovy",
  }


def get_metric_objects(
    metrics: Sequence[Metric],
    expected_plugin_type: PluginType,
    metric_type_class: ...,
) -> Sequence[Any]:
  """Return metric object instances of passed metric class."""

  typed_metric = (
      lambda e: metric_type_class(e.input) if metric_type_class else e
  )
  validate = (
      lambda e: metric_type_class.is_valid(e.input)
      if metric_type_class
      else True
  )

  return [
      typed_metric(e)
      for e in metrics
      if e.plugin_type == expected_plugin_type and validate(e)
  ]


def main():
  args = parse_args()

  if not os.path.exists(args.output_dir):
    os.makedirs(args.output_dir)

  if args.agent == MonitoringAgent.OPENTELEMETRY.value:
    apply_opentelemetry_overrides()

  # strips the \n at end of each line in the input file
  metric_objs = get_metric_objs(args.metrics_file)
  if not metric_objs:
    logging.warning("No metrics passed. Exiting")
    exit(0)

  plugin_generator_config = load_plugin_generator_config(
      args.plugin_generator_config_dir
  )

  services_to_be_excluded = get_services_to_be_excluded(args)

  jmx_metric_objs: Sequence[JmxMetric] = get_metric_objects(
      metric_objs, PluginType.JMX, JmxMetric
  )

  if jmx_metric_objs:
    config_args = {
        "jmx_metric_objs": jmx_metric_objs,
        "plugin_generator_config": plugin_generator_config,
        "templates_dir": args.templates_dir,
        "output_dir": args.output_dir,
        "service_to_output_file_name_map": args.output_file_name_kwargs,
        "config_parameters_kv_map": args.config_parameters_kwargs,
        "services_to_be_excluded": services_to_be_excluded,
        "mode": args.mode,
        "monitoring_agent": args.agent,
    }

    print(config_args["monitoring_agent"])

    generate_jmx_plugin_config_file(**config_args)

    if args.agent == MonitoringAgent.OPENTELEMETRY.value:
      config_args["service_to_output_file_name_map"] = {
          "otel_jmx_plugin": "otel_jmx_plugin.yaml"
      }

      generate_otel_jmx_plugin_config_file(**config_args)

  statsd_metric_objs: Sequence[Metric] = get_metric_objects(
      metric_objs, PluginType.STATSD, None
  )
  if statsd_metric_objs:
    generate_statsd_plugin_config_file(
        statsd_metric_objs,
        args.templates_dir,
        args.output_dir,
        args.output_file_name_kwargs,
        services_to_be_excluded,
        args.mode,
    )

  system_metric_objs: Sequence[SystemMetric] = get_metric_objects(
      metric_objs, PluginType.SYSTEM, SystemMetric
  )

  if system_metric_objs:
    generate_monitoring_plugin_config_file(
        system_metric_objs,
        args.output_dir,
        args.conf_files_dir,
        args.default_collectd_file,
        args.output_file_name_kwargs,
    )

  metrics_with_unsupported_service = get_metric_objects(
      metric_objs, PluginType.UNKNOWN, None
  )
  if metrics_with_unsupported_service:
    err_message = " , ".join(
        [e.input for e in metrics_with_unsupported_service]
    )
    logging.warning(
        "Ignoring metric inputs with unsupported services [ %s ]", err_message
    )

  jmx_metrics_with_invalid_format = get_metric_objs_for_jmx_with_invalid_format(
      metric_objs
  )

  if jmx_metrics_with_invalid_format:
    err_message = " , ".join([e.input for e in jmx_metrics_with_invalid_format])
    logging.warning(
        "Ignoring metric inputs with invalid format [ %s ]. %s",
        err_message,
        JmxMetric.get_help_message(),
    )


if __name__ == "__main__":
  main()
