#!/bin/bash
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Mounts any attached persistent and ephemeral disks non-boot disks

# N.B.: this script is run directly by systemd without sourcing bdutil_env.sh
# or dataproc_env.sh, so do not rely on variables/functions declared by them.

set -x
set -e

# Get standard environment variables to use DATAPROC_IMAGE_TYPE
source "/etc/environment"

PATH="$PATH:/usr/share/google"

function mount_disks_gce() {
  local disk_paths mounted_disks
  # Get a list of disks from the metadata server.
  disk_paths=$(get_metadata_value 'disks/')
  mounted_disks=()

  local disk_type disk_path disk_index disk_name disk_interface
  local disk_id device_path datamount
  for disk_path in ${disk_paths}; do
    disk_type=$(get_metadata_value "disks/${disk_path}type")
    # In EAP release, Optane (aka AEP) is implemented as disks.
    if [[ "${disk_type}" == "AEP-NVDIMM" ]]; then
      echo "Type of disk is ${disk_type}; will not attempt to mount it."
      continue
    fi

    # Use the metadata server to determine the official index/name of each disk.
    disk_index=$(get_metadata_value "disks/${disk_path}index")
    disk_name=$(get_metadata_value "disks/${disk_path}device-name")
    # Index '0' is the boot disk and is thus already mounted.
    if [[ "${disk_index}" == '0' ]]; then
      echo "Boot disk is ${disk_name}; will not attempt to mount it."
      continue
    fi
  
    # Local SSDs can be attached using any of following two interfaces;
    # SCSI (default) and NVME.
    # NVME is chosen for better performance, if images support NVMe and
    # include all optimizations.
    # Details: https://cloud.google.com/compute/docs/disks/local-ssd#performance
    disk_interface=$(get_metadata_value "disks/${disk_path}interface")
    # NVME Local SSDs will have different id than device-name, unlike SCSI.
    # Also device-name cannot be specified by user, so it will always be in
    # form of 'local-ssd-x'.
    # ex: device-name will be 'local-ssd-0';
    # disk id under /dev/disk/by-id will be 'google-local-nvme-ssd-0'
    if [[ "${disk_interface}" == 'NVME' ]]; then
      disk_name="${disk_name//local-ssd/local-nvme-ssd}"
    fi

    # The metadata-specified 'name' can be converted to a disk 'id' by prepending
    # 'google-' and finding it under /dev/disk/by-id.
    disk_id="/dev/disk/by-id/google-${disk_name}"
    device_path=$(readlink -f "${disk_id}")
    echo "Resolved disk name '${disk_name}' to expected path '${device_path}'."

    # We will name the mount-point after the official 'disk index'; this means
    # there will be no mounted disk with suffix '0' since '0' is the boot disk.
    datamount="/mnt/${disk_index}"
    mkdir -p "${datamount}"
    mounted_disks+=(${datamount})
    echo "Mounting '${device_path}' under mount point '${datamount}'..."
    safe_format_and_mount -m 'mkfs.ext4 -F' "${device_path}" "${datamount}"

    # Do *not* add to fstab, because this script will be re-run on boot
    # to re-mount disks.
  done

  if ((${#mounted_disks[@]})); then
    # Ensure Hadoop daemons can create their subdirectories
    # E.g. HDFS data dirs, YARN scratch dirs
    mkdir -p "${mounted_disks[@]/%//hadoop}"
    chown root:hadoop "${mounted_disks[@]/%//hadoop}"
    chmod 775 "${mounted_disks[@]/%//hadoop}"

    # If disks are mounted use the first one to hold target of symlink /hadoop
    local -r mounted_hadoop_dir=${mounted_disks[0]}/hadoop
    if [[ ! -d /hadoop ]]; then
      ln -sf "${mounted_hadoop_dir}" /hadoop
    fi
  fi
}

function mount_disks_s8s {
  local -r ssd_device="/dev/md0"
  local -r mount_point="/var/dataproc"
  readarray -t ssds < <(ls -A1 /dev/disk/by-id/google-local-nvme-ssd-*)
  if [[ ${#ssds[@]} -gt 0 ]]; then
    mdadm --create ${ssd_device} --level=0 --force \
      --raid-devices="${#ssds[@]}" "${ssds[@]}"

    mkfs.ext4 -F ${ssd_device}
    mkdir -p ${mount_point}
    mount ${ssd_device} ${mount_point}
    echo UUID="$(blkid -s UUID -o value ${ssd_device})" \
        ${mount_point} ext4 discard,defaults,nofail 0 2 | sudo tee -a /etc/fstab
    # Create a spark scratch directory
    mkdir -p "${mount_point}/spark"
    chmod 775 "${mount_point}/spark"
  else
    # If no data disks available, just make sure the mount point exists
    mkdir -p "${mount_point}"
  fi
  touch "${mount_point}/sentinel"
}

if [[ "${DATAPROC_IMAGE_TYPE}" == "serverless-spark" ]]; then
  mount_disks_s8s
else
  mount_disks_gce
fi
