# Copyright 2020 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Helper functions for use in other libexec/ core setup files

source "${BDUTIL_DIR}/os/shared.sh"

APT_SENTINEL='apt.lastupdate'

function os_update_repo() {
  retry_constant_short apt-get update -y -qq
  touch "${APT_SENTINEL}"
}

function os_upgrade_packages() {
  retry_constant_short apt-get upgrade -y -qq
}

function os_update_repo_if_needed() {
  # If the apt sentinel doesn't exist or if it hasn't ben touched in the
  # last hour, update apt
  if [[ "$(find "$(dirname ${APT_SENTINEL})" -maxdepth 1 \
    -name "$(basename ${APT_SENTINEL})" -newermt '-1 hour')" == "" ]]; then
    os_update_repo
  fi
}

# Installs a number of debian applications. With retries.
function os_install_packages() {
  local packages=("$@")
  os_update_repo_if_needed

  local extra_args=()
  if [[ "${DATAPROC_IMAGE_TYPE}" == 'serverless-spark' ]]; then
    extra_args+=('--no-install-recommends')
  fi

  retry_constant_short apt-get -y install "${extra_args[@]}" "${packages[@]}"
}

function os_uninstall_packages() {
  local -r packages_to_uninstall=("$@")

  loginfo "Uninstalling packages: ${packages_to_uninstall[*]}"
  # Safe to leave inactive packages installed in case of failure
  # Use autoremove to also clean up unused dependencies.
  local -r uninstall_cmd="DEBIAN_FRONTEND=noninteractive \
    apt-get autoremove -y --purge ${packages_to_uninstall[*]}"
  retry_constant_short bash -c "$uninstall_cmd" || true
}

function os_is_package_installed() {
  dpkg -l "$1" >/dev/null 2>&1
}

function os_installed_package_version() {
  dpkg -s "$1" | sed -n 's/^Version: \([0-9\.]*\).*/\1/p'
}

function os_lock_package_version() {
  local packages=("$@")
  retry_constant_short apt-mark hold "${packages[@]}"
}

# Locks package to specific version
# Use this util when package needs to be locked before install and needs
# guarantee that no other version of package will get installed.
# Specify specific package version with glob if required
# e.g. if only major,minor package version needs to be locked, call this with
# os_lock_package_to_specific_version docker-ce 19.3.*
function os_lock_package_to_specific_version() {
  local -r package=$1
  local -r version=$2
  # https://manpages.debian.org/stretch/apt/apt_preferences.5.en.html
  cat >"/etc/apt/preferences.d/${package}.pref" <<EOF
Package: ${package}
Pin: version ${version}
Pin-Priority: 600
EOF
}

# Please refer following for public keys.
# https://cran.r-project.org/bin/linux/ubuntu/
# https://cran.r-project.org/bin/linux/debian/
function os_setup_r_repo() {
  if is_ubuntu; then
    local -r repo_key="E298A3A825C0D65DFD57CBB651716619E084DAB9"
  else
    local -r repo_key="95C0FAF38DB3CCAD0C080A7BDC78B2DDEABC47B7"
  fi

  retry_fibonacci apt-key adv --no-tty \
    --keyserver "hkp://keyserver.ubuntu.com:80" \
    --recv-keys "${repo_key}"

  cat >/etc/apt/sources.list.d/cran-r.list <<EOF
deb ${R_APT_SITE}/bin/linux/${OS_ID} ${OS_CODENAME}-${R_REPO}/
EOF
}

function os_lock_docker_package() {
  # Needs exact "epoch" in version number to lock package
  os_lock_package_to_specific_version "docker-ce" "5:${DOCKER_VERSION}.*"
  os_lock_package_to_specific_version "docker-ce-cli" "5:${DOCKER_VERSION}.*"
}

function os_lock_r_package() {
  os_lock_package_to_specific_version "r-base" "${R_VERSION}.*"
  os_lock_package_to_specific_version "r-recommended" "${R_VERSION}.*"
}

function add_regional_bigtop_repo() {
  # For test images which do not use gs://dataproc-bigtop-repo for Bigtop repo
  # regional_bigtop_repo_uri will be evaluated to empty.
  local -r region=$1
  local -r dataproc_repo_file="/etc/apt/sources.list.d/dataproc.list"

  if is_test_bigtop_repo "${dataproc_repo_file}" ; then
    echo 'Skip adding regional Bigtop repo (this seems to be a test image).'
    return 0
  fi

  local regional_bigtop_repo_uri
  regional_bigtop_repo_uri=$(cat ${dataproc_repo_file} |
    sed "s#dataproc-bigtop-repo#goog-dataproc-bigtop-repo-${region}#" |
    grep "deb .*goog-dataproc-bigtop-repo-${region}.* dataproc contrib" |
    cut -d ' ' -f 2 |
    head -1)

  if [[ "${regional_bigtop_repo_uri}" == */ ]]; then
    local -r bigtop_key_uri="${regional_bigtop_repo_uri}archive.key"
  else
    local -r bigtop_key_uri="${regional_bigtop_repo_uri}/archive.key"
  fi

  # Retry for 15 seconds max to not slow-down cluster creation
  if curl -fsS --retry-connrefused --retry 3 --retry-delay 5 \
      "${bigtop_key_uri}" | apt-key add -; then
    # Prepend the regional Bigtop repo which is preferred over the global one.
    echo "Adding regional Bigtop repo for ${region} in APT sources."
    cat >/tmp/dataproc.list <<EOF
# Regional Bigtop repo (used first).
deb ${regional_bigtop_repo_uri} dataproc contrib
deb-src ${regional_bigtop_repo_uri} dataproc contrib

# Global Bigtop repo (fallback).
EOF
    cat "${dataproc_repo_file}" >>/tmp/dataproc.list
    mv -f /tmp/dataproc.list "${dataproc_repo_file}"
  else
    echo "Failed to add regional repo ${regional_bigtop_repo_uri}"
  fi
}
