#!/usr/bin/env bash

# Debian-specific functions used to build a Dataproc VM image.

set -Eeuxo pipefail

# Necessary for installing multiple packages
export DEBIAN_FRONTEND='noninteractive'

ARCH_SUFFIX="$(get_arch_suffix)"
readonly ARCH_SUFFIX

OS_CODENAME="$(. /etc/os-release; echo "${VERSION_CODENAME}")"
export OS_CODENAME

OS_DATAPROC_SPECIFIC_COMMON_PACKAGES=(
  libapr1
  libatlas3-base
  libjansi-java
  libsnappy-dev
  libssl-dev
  libzstd-dev
  openssl
  uuid-runtime
)

if [[ ${DATAPROC_IMAGE_TYPE} == 'serverless-spark' ]] && ! is_debian11; then
  OS_DATAPROC_SPECIFIC_COMMON_PACKAGES+=(
    "linux-headers-cloud-${ARCH_SUFFIX}/${OS_CODENAME}-backports"
    "linux-image-cloud-${ARCH_SUFFIX}/${OS_CODENAME}-backports"
  )
elif is_debian; then
  OS_DATAPROC_SPECIFIC_COMMON_PACKAGES+=(
    "linux-headers-cloud-${ARCH_SUFFIX}" "linux-image-cloud-${ARCH_SUFFIX}")
else
  OS_DATAPROC_SPECIFIC_COMMON_PACKAGES+=("linux-headers-$(uname -r)")
fi

if is_debian10 || is_ubuntu18; then
  OS_DATAPROC_SPECIFIC_COMMON_PACKAGES+=(
    libopenblas-dev
    netcat
    python-numpy
    python-pip
    python-requests
    python-setuptools
  )
else
  OS_DATAPROC_SPECIFIC_COMMON_PACKAGES+=(
    build-essential
    libopenblas0
    netcat-openbsd
    python3-numpy
    python3-pip
    python3-requests
    python3-setuptools
  )
fi

if is_debian10; then
  # linux-kbuild-5.10 is a dependency of linux-headers-amd64, if we don't
  # install it explicitly, by default the same version as linux-headers-amd64
  # will be installed, but we want it to be its own latest version, so add it
  # explicitly. We can update the pinned version when there is a new version.
  # https://packages.debian.org/buster/linux-kbuild-5.10
  OS_DATAPROC_SPECIFIC_COMMON_PACKAGES+=(linux-kbuild-5.10)
fi

if is_arm64 && is_ubuntu; then
  # These packages are not pre-installed in ubuntu arm image
  OS_DATAPROC_SPECIFIC_COMMON_PACKAGES+=(unzip zip python)
fi

export JAVA_HOME=/usr/lib/jvm/${JAVA_PACKAGE}-${ARCH_SUFFIX}
if is_java8; then
  export JAVA_CONFIG_DIR=${JAVA_HOME}/jre/lib
else
  export JAVA_CONFIG_DIR=${JAVA_HOME}/conf
fi

if [[ -n "${ADDITIONAL_JAVA_PACKAGE}" ]]; then
  export ADDITIONAL_JAVA_HOME=/usr/lib/jvm/${ADDITIONAL_JAVA_PACKAGE}-${ARCH_SUFFIX}
  export ADDITIONAL_JAVA_CONFIG_DIR=${ADDITIONAL_JAVA_HOME}/conf
fi

REPO_CACHE_DIR="/var/cache/apt"

if is_debian10; then
  PACKAGE_INSTALL_FLAGS=('-t' "${OS_CODENAME}-backports")
fi

function os_setup_env() {
  loginfo "No env needs to be setup for Debian."
}

function os_prepare_packages_repo() {
  mv_and_link "${REPO_CACHE_DIR}" "${BUILD_DIR}/apt-cache"
  # Download APT packages to enable HTTPS repo
  loginfo "Enabling APT to download from HTTPS repository"
  os_update_repo
  os_install_packages dirmngr apt-transport-https dpkg-dev software-properties-common
}

function os_setup_bigtop_repo() {
  local -r bigtop_repo_url=$1
  curl -fsS --retry-connrefused --retry 10 --retry-max-time 30 \
    "${bigtop_repo_url}/archive.key" | apt-key add -
  cat >/etc/apt/sources.list.d/dataproc.list <<EOF
deb ${bigtop_repo_url} dataproc contrib
deb-src ${bigtop_repo_url} dataproc contrib
EOF

  # Prefer Dataproc Bigtop packages over others
  # n=dataproc refers to our codename see:
  # https://manpages.debian.org/stretch/apt/apt_preferences.5.en.html
  cat >>/etc/apt/preferences.d/dataproc <<EOF
Package: *
Pin: release n=dataproc
Pin-Priority: 600
EOF
}

# Secondary Bigtop repo is used only for test images.
function os_setup_secondary_bigtop_repo() {
  loginfo "Setting up secondary BigTop repository"
  local secondary_bigtop_repo_url=$1
  curl -fsS --retry-connrefused --retry 10 --retry-max-time 30 \
    "${secondary_bigtop_repo_url}/archive.key" | apt-key add -
  cat >/etc/apt/sources.list.d/_dataproc.list <<EOF
deb ${secondary_bigtop_repo_url} dataproc contrib
deb-src ${secondary_bigtop_repo_url} dataproc contrib
EOF

  # Prefer secondary Dataproc Bigtop packages over others
  # n=dataproc refers to our codename see:
  # https://manpages.debian.org/stretch/apt/apt_preferences.5.en.html
  cat >>/etc/apt/preferences.d/_dataproc <<EOF
Package: *
Pin: release n=dataproc
Pin-Priority: 700
EOF
}

function os_setup_google_instance_configs() {
  # Ubuntu does not have instance_configs.cfg config file by default
  # Adding minimum accounts and daemon configs relevant to Dataproc
  # Config info here: https://github.com/GoogleCloudPlatform/guest-agent#configuration
  if is_ubuntu; then
    cat >/etc/default/instance_configs.cfg <<EOF
[Daemons]
accounts_daemon = true

[MetadataScripts]
default_shell = /bin/bash
shutdown = true
startup = true
EOF
  fi
}

function os_setup_java_repo() {
  # This broadly follows the instructions from
  # https://adoptium.net/installation/linux/#_deb_installation_on_debian_or_ubuntu
  local -r key_url="https://packages.adoptium.net/artifactory/api/gpg/key/public"
  curl -fsS --retry-connrefused --retry 10 --retry-max-time 30 "${key_url}" |
    apt-key add -
  cat >/etc/apt/sources.list.d/adoptium.list <<EOF
deb https://packages.adoptium.net/artifactory/deb/ ${OS_CODENAME} main
EOF
  apt-get update
  # There are some throttling issues in adoptium repo while installing JDK.
  # Installing from downloaded package.
  if [[ "${DATAPROC_IMAGE_TYPE}" == "standard" ]] && ! is_arm64 ; then
    if is_version_at_least "${DATAPROC_IMAGE_VERSION}" '2.1'; then
      local jdk_package_name="temurin-11-jdk_11.0.20.1.0+1_amd64.deb"
      download_resource \
        "gs://dataproc-image-dependencies/jdk/${jdk_package_name}" \
        "${BUILD_DIR}"
      os_install_packages "${BUILD_DIR}/${jdk_package_name}"
    fi
    if is_version_at_least "${DATAPROC_IMAGE_VERSION}" '2.2'; then
      local additional_jdk_package_name="temurin-17-jdk_17.0.9.0.0+9_amd64.deb"
      download_resource \
        "gs://dataproc-image-dependencies/jdk/${additional_jdk_package_name}" \
        "${BUILD_DIR}"
      os_install_packages "${BUILD_DIR}/${additional_jdk_package_name}"
    fi
  fi
}

function os_setup_docker_repo() {
  local -r key_url="https://download.docker.com/linux/${OS_ID}/gpg"
  curl -fsS --retry-connrefused --retry 10 --retry-max-time 30 "${key_url}" |
    apt-key add -

  cat >/etc/apt/sources.list.d/docker.list <<EOF
deb https://download.docker.com/linux/${OS_ID} ${OS_CODENAME} stable
EOF
}

function os_setup_additional_repos() {
  loginfo "No additional repos are required for Debian"
}

function os_setup_python() {
  if [[ -f /usr/bin/python3 ]] && ! [[ -f /usr/bin/python ]]; then
    # Ubuntu 22 uses an explicit python3.
    ln -s /usr/bin/python3 /usr/bin/python
  fi
  if [[ "${DATAPROC_IMAGE_TYPE}" == "standard" ]] &&
    is_version_at_least "${DATAPROC_IMAGE_VERSION}" '2.2'; then
    # Install python 3.9 excplicitly to run hue component, which compiles
    # cleanly only on python3.9.
    # There is no installer available for the latest python version.
    # Therefore has to be installed via source.
    local -r python_pkg_name="Python-${HUE_PYTHON_VERSION}"
    download_resource \
      "https://storage.googleapis.com/dataproc-bigtop-toolchain/${python_pkg_name}.tar.xz" \
      "${BUILD_DIR}/${python_pkg_name}.tar.xz"
    if is_ubuntu; then
      # Enable source repository first to download build dependencies in Ubuntu
      local -r src_entry="deb-src http://archive.ubuntu.com/ubuntu/ ${OS_CODENAME} main"
      cat >> /etc/apt/sources.list <<EOF
${src_entry}
EOF
      apt-get update
    fi
    # Install dependencies
    apt-get -y build-dep python3
    os_install_packages build-essential gdb lcov pkg-config \
      libbz2-dev libffi-dev libgdbm-dev libgdbm-compat-dev liblzma-dev \
      libncurses5-dev libreadline6-dev libsqlite3-dev libssl-dev \
      lzma lzma-dev tk-dev uuid-dev zlib1g-dev default-libmysqlclient-dev
    # Compile and install python
    tar -xf "${BUILD_DIR}/${python_pkg_name}.tar.xz" -C "${BUILD_DIR}" && \
      cd "${BUILD_DIR}/${python_pkg_name}" && \
      ./configure --enable-optimizations && \
      make -j 2 && make altinstall &&
      /usr/local/bin/pip3.9 install virtualenv=="${HUE_VIRTUALENV_VERSION}" \
        virtualenv-make-relocatable=="${HUE_VIRTUALENV_MAKE_RELOCATABLE_VERSION}" \
        mysqlclient

    if is_ubuntu; then
      # Disable to keep default state
      sed -i "s@${src_entry}@#${src_entry}@g" /etc/apt/sources.list
      apt-get update
    fi
  else
    loginfo "Skip setting up Python for Debian"
  fi
}

function os_configure_java() {
  # Prevent Java from being upgraded after the image is built:
  apt-mark hold "${JAVA_PACKAGE}"

  if [[ -n ${ADDITIONAL_JAVA_PACKAGE} ]]; then
    apt-mark hold "${ADDITIONAL_JAVA_PACKAGE}"
    # Reset the alternatives to default JAVA_HOME/bin
    local default_java_cmds
    mapfile -t default_java_cmds < <(ls "${JAVA_HOME}/bin")
    local cmd
    for cmd in "${default_java_cmds[@]}"; do
      update-alternatives --set "$cmd" "${JAVA_HOME}/bin/${cmd}"
    done
  fi
}

function os_install_scala() {
  local scala_version=$1
  download_resource \
    "gs://dataproc-image-dependencies/scala/scala-${scala_version}.deb" \
    "${BUILD_DIR}"
  dpkg -i --path-exclude='/usr/share/doc/*' "${BUILD_DIR}/scala-${scala_version}.deb"
}

function os_update_ssh_service() {
  # Return code 255 will prevent ssh.service from auto restarting on failure.
  wrap_systemd_service ssh /etc/ssh/sshd_not_to_be_run 255 ExecStart
}

# b/113161214: Delay unattended updates/upgrades by 1 hour in Debian/Ubuntu
function os_delay_unattended_upgrades() {
  sed -i '/ExecStart/i ExecStartPre=/bin/sleep 3600' /lib/systemd/system/apt-daily.service
  sed -i '/ExecStart/i ExecStartPre=/bin/sleep 3600' /lib/systemd/system/apt-daily-upgrade.service

  # Disable snap update by default, it will be enabled in startup script.
  if is_ubuntu; then
    # Manually refresh to get the latest packages, if the base Ubuntu image is
    # old, this will potentially reduce the auto refreshes at runtime.
    sudo snap install core20
    snap refresh

    if is_ubuntu22; then
      # In Ubuntu 22, after falling back to CGroup v1, masked snapd is
      # delaying/blocking the complete startup of nodes.
      # So keeping the snapd running, but holding the refresh
      snap refresh -hold
    else
      # Set a refresh hold to prevent auto refresh during startup, startup script
      # will reset it to 1 hour after starting snapd. But note that Snap has a
      # max hold limit of 60 days, when expired, it will auto refresh when snapd
      # starts regardless of the hold.
      snap set system refresh.hold="$(date -uIs -d '+60 days')"

      systemctl stop snapd
      systemctl mask snapd
    fi
  fi
}

function os_cleanup_packages_cache() {
  # Clean up /var/cache/apt
  apt-get clean
  unlink_and_mv "${REPO_CACHE_DIR}"
}

function os_cleanup_logs() {
  if [[ "${DATAPROC_IMAGE_TYPE}" == "serverless-spark" ]]; then
    rm -rf /var/log/*
  else
    rm -f /var/log/syslog /var/log/daemon.log
  fi
}
