# Copyright 2020 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Helper functions for use in other libexec/ core setup files

source "${BDUTIL_DIR}/os/shared.sh"

YUM_SENTINEL='yum.lastupdate'

function os_update_repo() {
  retry_constant_short yum -y update
  touch "${YUM_SENTINEL}"
}

function os_update_repo_if_needed() {
  # If the yum sentinel doesn't exist or if it hasn't ben touched in the
  # last hour, update yum
  if [[ "$(find "$(dirname ${YUM_SENTINEL})" -maxdepth 1 \
    -name "$(basename ${YUM_SENTINEL})" -newermt '-1 hour')" == "" ]]; then
    os_update_repo
  fi
}

function os_upgrade_packages() {
  retry_constant_short yum -y upgrade
}

# Installs a number of Rocky Linux applications with retries.
function os_install_packages() {
  local packages=("$@")
  os_update_repo_if_needed
  retry_constant_short yum -y install "${packages[@]}"
}

function os_uninstall_packages() {
  local -r packages_to_uninstall=("$@")

  loginfo "Uninstalling packages: ${packages_to_uninstall[*]}"
  # Safe to leave inactive packages installed in case of failure
  # Yum removes unused dependencies and config file automatically
  local -r uninstall_cmd="yum remove -y ${packages_to_uninstall[*]}"
  retry_constant_short bash -c "$uninstall_cmd" || true
}

function os_is_package_installed() {
  rpm -qi "$1" >/dev/null 2>&1
}

function os_installed_package_version() {
  rpm -qi "$1" | sed -n 's/^Version.*: \([0-9\.]*\).*/\1/p'
}

function os_lock_package_version() {
  local packages=("$@")

  # `versionlock` commanad is available as a plugin, need to install it first.
  retry_constant_short yum -y install 'dnf-command(versionlock)' \
    && retry_constant_short yum versionlock add "${packages[@]}"
}

# Locks package to specific version
# Use this util when package needs to be locked before install and needs
# guarantee that no other version of package will get installed.
# Specify specific package version with glob if required
# e.g. if only major,minor package version needs to be locked, call this with
# os_lock_package_to_specific_version docker-ce 19.3.*
function os_lock_package_to_specific_version() {
  local -r package=$1
  local -r version=$2
  local output
  output=$(yum versionlock add "${package}-${version}" 2>&1)
  local -r status=$?

  if [[ "${status}" -eq 0 ]]; then
    if echo "${output}" | grep -q "Adding versionlock on"; then
      return 0
    fi
  fi
  return 1
}

function os_lock_docker_package() {
  os_lock_package_to_specific_version "docker-ce" "${DOCKER_VERSION}.*"
  os_lock_package_to_specific_version "docker-ce-cli" "${DOCKER_VERSION}.*"
}

function os_lock_r_package() {
  os_lock_package_to_specific_version "R" "${R_VERSION}.*"
}

function os_lock_jdk_package() {
  os_lock_package_to_specific_version "${JAVA_PACKAGE}" "11.0.20.1.0.1-1.x86_64"
}

function add_regional_bigtop_repo() {
  local -r region=$1
  local -r dataproc_repo_file="/etc/yum.repos.d/dataproc.repo"
  local -r dataproc_fallback_repo_file="/tmp/dataproc_fallaback.repo"

  if is_test_bigtop_repo "${dataproc_repo_file}" ; then
    echo 'Skip adding regional Bigtop repo (this seems to be a test image).'
    return 0
  fi

  local regional_bigtop_key_uri
  regional_bigtop_key_uri=$(cat ${dataproc_repo_file} |
    sed "s#dataproc-bigtop-repo#goog-dataproc-bigtop-repo-${region}#" |
    grep "gpgkey" |
    cut -d '=' -f 2 |
    head -1)

  # Verify if repo exists. Checking for repo key is used as a proxy for same.
  if ! curl -fsS --head --retry-connrefused --retry 3 --retry-delay 5 \
      "${regional_bigtop_key_uri}" ; then
    echo "Failed to add regional repo for region: ${region}"
    return 0
  fi

  # Create copy of dataproc repo
  cat "${dataproc_repo_file}" >> "${dataproc_fallback_repo_file}"
  # replace dataproc-bigtop-repo to regional repo
  sed -i "s#dataproc-bigtop-repo#goog-dataproc-bigtop-repo-${region}#" "${dataproc_repo_file}"

  # Change repo name and description for fallback repo
  sed -i "s/\[dataproc\]/\[dataproc-fallback\]/" "${dataproc_fallback_repo_file}"
  sed -i 's/^\(name\s*=\s*\).*$/\1Dataproc Fallback Repo/' "${dataproc_fallback_repo_file}"

  # Add an empty line at the end of the file, to add fallback repo
  echo "" >> "${dataproc_repo_file}"
  # Append fallback repo to main repo file
  cat "${dataproc_fallback_repo_file}" >> "${dataproc_repo_file}"
  rm "${dataproc_fallback_repo_file}"
}
