#!/usr/bin/env bash

# Rocky-specific functions used to build a Dataproc VM image.

set -Eeuxo pipefail

# These packages are available on all Dataproc instances.
# Components in this set are assumed to be libraries and not services.
OS_DATAPROC_SPECIFIC_COMMON_PACKAGES=(
  apr-devel
  atlas
  cyrus-sasl
  cyrus-sasl-plain
  gcc
  gcc-c++
  libzstd-devel
  netcat
  openblas-devel
  openssl-devel
  snappy-devel
  uuid)

if is_version_at_least "${DATAPROC_IMAGE_VERSION}" "2.1"; then
  OS_DATAPROC_SPECIFIC_COMMON_PACKAGES+=(
    python3-numpy python3-pip python3-requests python3-setuptools)
else
  OS_DATAPROC_SPECIFIC_COMMON_PACKAGES+=(
    python2-numpy python2-pip python2-requests python2-setuptools)
fi
if is_rocky9; then
  OS_DATAPROC_SPECIFIC_COMMON_PACKAGES+=(initscripts)
fi

export JAVA_HOME=/usr/lib/jvm/${JAVA_PACKAGE}
if is_java8; then
  export JAVA_CONFIG_DIR=${JAVA_HOME}/jre/lib
else
  export JAVA_CONFIG_DIR=${JAVA_HOME}/conf
fi

if [[ -n "${ADDITIONAL_JAVA_PACKAGE}" ]]; then
  export ADDITIONAL_JAVA_HOME=/usr/lib/jvm/${ADDITIONAL_JAVA_PACKAGE}
  export ADDITIONAL_JAVA_CONFIG_DIR=${ADDITIONAL_JAVA_HOME}/conf
fi

REPO_CACHE_DIR="/var/cache/dnf"

function os_setup_env() {
  # umask on Rocky Linux is 077 by default, change it to 022
  # to be consistent with Debian/Ubuntu system
  sed -i 's/^\(UMASK\s*\s*\).*$/\1022/' /etc/login.defs
  # Delete code that sets umask for all uid and gid
  sed -i '/# By default, we want umask to get set/,+8d' /etc/bashrc
  sed -i '/# By default, we want umask to get set/,+8d' /etc/profile
  # Add umask default setting
  cat >/etc/profile.d/configure_umask.sh <<EOF
umask 022
EOF

  # Set prompt to be consistent with debian
  # This changes bash prompt to format user@host:current_dir#
  cat >/etc/profile.d/configure_prompt.sh <<'EOF'
export PS1='\u@\h:\w\$ '
EOF

  # SELinux is enabled by default which may interfere with some applications,
  # make it a permissive mode, which provides warning for access-control violations
  # Additional info here: https://www.redhat.com/en/topics/linux/what-is-selinux
  setenforce 0
  sed -i 's/^\(SELINUX\s*=\s*\).*$/\1permissive/' /etc/selinux/config
}

function os_prepare_packages_repo() {
  mv_and_link "${REPO_CACHE_DIR}" "${BUILD_DIR}/dnf-cache"
  # Check https://fedoraproject.org/wiki/EPEL
  # This enables downloading packages like R from https://cloud.r-project.org/bin/linux/redhat/
  local epel_version
  if is_rocky8; then
    epel_version=8
  elif is_rocky9; then
    epel_version=9
  fi
  os_install_packages "https://dl.fedoraproject.org/pub/epel/epel-release-latest-${epel_version}.noarch.rpm"
  os_install_packages yum-plugin-versionlock

  if is_rocky9; then
     # This package is available only in rocky-9 devel and all services needs
     # this package to have '/lib/lsb/init-functions'.
     # Installing from devel repo as per instructions in
     # https://rockylinux.pkgs.org/9/rockylinux-devel-x86_64/redhat-lsb-4.1-56.el9.x86_64.rpm.html
     os_install_packages --enablerepo=devel redhat-lsb
  fi
}

function os_setup_bigtop_repo() {
  loginfo "Setting up BigTop repository"
  local bigtop_repo_url=$1
  # Prefer Dataproc Bigtop packages over others
  # [1] is highest, default without the line is [99]
  cat >/etc/yum.repos.d/dataproc.repo <<EOF
[dataproc]
name=Dataproc Repo
baseurl=${bigtop_repo_url}
enabled=1
gpgcheck=1
gpgkey=${bigtop_repo_url}/RPM-GPG-KEY-dataproc
priority=2
EOF
}

# Secondary Bigtop repo is used only for test images.
function os_setup_secondary_bigtop_repo() {
  loginfo "Setting up secondary BigTop repository"
  local secondary_bigtop_repo_url=$1
  cat >/etc/yum.repos.d/_dataproc.repo <<EOF
[secondary-dataproc]
name=Secondary Dataproc Repo
baseurl=${secondary_bigtop_repo_url}
enabled=1
gpgcheck=1
gpgkey=${secondary_bigtop_repo_url}/RPM-GPG-KEY-dataproc
priority=1
EOF
}

function os_setup_google_instance_configs() {
  loginfo "Rocky Linux does not need to setup google-instance configs... Skipping"
}

function os_setup_java_repo() {
  cat >/etc/yum.repos.d/adoptium.repo <<EOF
[adoptium]
name=Adoptium repo
baseurl=https://adoptium.jfrog.io/artifactory/rpm/rocky/8/$(arch)/
enabled=1
gpgcheck=1
gpgkey=https://adoptium.jfrog.io/artifactory/api/security/keypair/default-gpg-key/public
EOF
}

function os_setup_r_repo() {
  loginfo "Rocky Linux does not need to separately setup R repo... Skipping"
}

function os_setup_docker_repo() {
  local repo_version=8
  if is_rocky9; then
    repo_version=9
  fi
  cat >/etc/yum.repos.d/docker.repo <<EOF
[docker-ce-stable]
name=Docker CE Stable - $(arch)
baseurl=https://download.docker.com/linux/centos/${repo_version}/$(arch)/stable
enabled=1
gpgcheck=1
gpgkey=https://download.docker.com/linux/centos/gpg
priority=1
EOF
}

function os_setup_additional_repos() {
  if is_rocky8; then
    local powertools="PowerTools"
    if dnf repolist all | grep -q "^powertools"; then
      powertools="powertools"
    fi
    dnf config-manager --set-enabled "${powertools}"
  elif is_rocky9; then
    # In Rocky Linux 9 powertools repo is known as crb.
    local crb="crb"
    dnf config-manager --set-enabled "${crb}"
  fi
}

function os_setup_python() {
  if [[ -f /usr/bin/python2 ]] && ! [[ -f /usr/bin/python ]]; then
    # Rocky Linux 8 uses an explicit python2 or python3, but few applications
    # still rely on python to be available. Make it work for now
    alternatives --set python /usr/bin/python2
    # Make sure pip works for compatibility, because Rocky Linux
    # uses explicit pip2 or pip3, default pip to pip2
    ln -s /usr/bin/pip2 /usr/bin/pip
  else
    if ! [[ -f /usr/bin/python ]]; then
      # Rocky Linux 9 uses an explicit python3. Python2 is not available in
      # Rocky Linux 9.
      ln -s /usr/bin/python3 /usr/bin/python
    fi
    if ! [[ -f /usr/bin/pip ]]; then
      # since pip2 not available in Rocky Linux 9, pip default to pip3.
      ln -s /usr/bin/pip3 /usr/bin/pip
    fi
  fi

  if [[ "${DATAPROC_IMAGE_TYPE}" == "standard" ]] &&
    is_version_at_least "${DATAPROC_IMAGE_VERSION}" '2.2'; then
    # Install python 3.9 excplicitly to run hue component, which compiles
    # cleanly only on python3.9.
    # There is no installer available for the latest python version.
    # Therefore has to be installed via source.
    local -r python_pkg_name="Python-${HUE_PYTHON_VERSION}"
    download_resource \
      "https://storage.googleapis.com/dataproc-bigtop-toolchain/${python_pkg_name}.tar.xz" \
      "${BUILD_DIR}/${python_pkg_name}.tar.xz"
    #Install dependencies
    os_install_packages yum-utils python3-devel mysql-devel 
    yum-builddep -y python3
    # Compile and install python
    tar -xf "${BUILD_DIR}/${python_pkg_name}.tar.xz" -C "${BUILD_DIR}" && \
      cd "${BUILD_DIR}/${python_pkg_name}" && \
      ./configure --enable-optimizations && \
      make -j 2 && make altinstall && \
      /usr/local/bin/pip3.9 install virtualenv=="${HUE_VIRTUALENV_VERSION}" \
        virtualenv-make-relocatable=="${HUE_VIRTUALENV_MAKE_RELOCATABLE_VERSION}" \
        mysqlclient
  fi
}

function os_configure_java() {
  # Prevent Java from being upgraded after the image is built:
  yum versionlock add "${JAVA_PACKAGE}"

  if [[ -n ${ADDITIONAL_JAVA_PACKAGE} ]]; then
    yum versionlock add "${ADDITIONAL_JAVA_PACKAGE}"
    # Reset the alternatives to default JAVA_HOME/bin
    default_java_cmds=( $(ls "${JAVA_HOME}"/bin) )
    for cmd in "${default_java_cmds[@]}"; do
      # Ignore if not present
      alternatives --set "$cmd" "$JAVA_HOME"/bin/"$cmd" || true
    done
  fi
}

function os_install_scala() {
  local scala_version=$1
  download_resource \
    "gs://dataproc-image-dependencies/scala/scala-${scala_version}.rpm" \
    "${BUILD_DIR}"
  yum -y --nodocs install "${BUILD_DIR}/scala-${scala_version}.rpm"
}

function os_update_ssh_service() {
  # Return code 255 will prevent ssh.service from auto restarting on failure.
  wrap_systemd_service sshd /etc/ssh/sshd_not_to_be_run 255 ExecStart
}

function os_delay_unattended_upgrades() {
  loginfo "No unattended upgrades in Rocky Linux... Skipping"
}

function os_cleanup_packages_cache() {
  # Clean up /var/cache/dnf
  yum clean all
  unlink_and_mv "${REPO_CACHE_DIR}"
}

function os_cleanup_logs() {
  rm -f /var/log/messages
}

function os_prevent_tmp_folder_deletion() {
  # By default the /tmp folder gets deleted every 10 days in RedHat/Centos/(Rocky)
  # This behavior interferes with processes that rely on files in /tmp. 
  # eg of such is dataproc-launcher.py. In Ubuntu and Debian the files do not get deleted.
  # Duplicating the behavior of Debian and Ubuntu by setting the /tmp folder properties
  loginfo "Setting /tmp folder file deletion property"
  systemctl stop systemd-tmpfiles-clean.timer
  sed -i "s/q \/tmp 1777 root root 10d/D \/tmp 1777 root root -/g" /usr/lib/tmpfiles.d/tmp.conf
  sed -i "s/^q \/var\/tmp 1777 root root 30d/#q \/var\/tmp 1777 root root 30d/g" /usr/lib/tmpfiles.d/tmp.conf
  systemctl daemon-reload
  systemctl start systemd-tmpfiles-clean.timer
  loginfo "Successfully restarted systemd tmpfiles cleaner"
}
