# Set the default root password for MySQL.
function configure_password() {
  # start mysqld first to get temp password
  systemctl start "${MYSQL_SERVICE}"

  set +x

  # Rocky Linux does not have debconf-set-selections to select password during
  # install hence grab temp password with service start from logs
  #
  # https://dev.mysql.com/doc/refman/8.0/en/mysql-oci-marketplace-connecting.html
  echo "Getting MySQL tmp password"
  local temp_mysql_password
  temp_mysql_password=$(grep 'temporary password' /var/log/mysqld.log | tail -1 | awk '{split($0,a,": "); print a[2]}')

  echo "Updating MySQL tmp password to default password"
  # Change the tmp password to the default mysql password
  mysql -u root -p"${temp_mysql_password}" \
    --connect-expired-password \
    -e "ALTER USER 'root'@'localhost' IDENTIFIED BY '${MYSQL_DEFAULT_ROOT_PASSWORD}'; \
        FLUSH PRIVILEGES;"

  # In MySQL 5.7, uninstall validate_password plugin to allow weak passwords for
  # backwards compatibility.
  #
  # In MySQL 8 validate_password is no longer a plugin, trying to uninstall it
  # will fail, instead we will set validate_password.policy to LOW later in
  # /etc/mysql/my.cnf to get the same effect.
  if [[ "${MYSQL_VERSION}" == "5.7" ]]; then
    echo "Uninstalling validate_password plugin"
    mysql -u root -p"${MYSQL_DEFAULT_ROOT_PASSWORD}" \
      -e "uninstall plugin validate_password;"
  fi

  set -x
}

function add_config_dir() {
  cat >>/etc/my.cnf <<EOF

!includedir ${MYSQL_CONFIG_DIR}
EOF
}

function lock_mysql_connector() {
  # Do not upgrade MySQL connector after install as we symlink MySQL jar
  # and it can change after upgrade
  os_lock_package_version mysql-connector-j
}

function configure_mysqld() {
    echo "
[mysqld]

# Allow connections (e.g., Hive Metastore) from other master nodes to master 0
# in HA mode
bind-address = 0.0.0.0
" >> /etc/mysql/my.cnf

  # Settings for backwards compatibility with MySQL 5.7
  if is_version_at_least "${MYSQL_VERSION}" "8.0"; then
    echo "
# Users can provide MySQL passwords for some components (e.g., Ranger), we have
# to allow weak ones for backwards compatibility.
validate_password.policy=LOW

# Disable binary logging (otherwise Ranger will fail to create functions)
skip-log-bin
" >> /etc/mysql/my.cnf
  fi
}

lock_mysql_connector
configure_password
add_config_dir
configure_mysqld
