#!/usr/bin/env bash

# Script which helps launching Spark containers of different types.
#
# Usage: spark-containerd-runner.sh \
#        <working-dir> <container-name> <container-type> [container args...]
#
# The container image is from `dataproc:dataproc.spark.container.image` if
# the property is provided, otherwise from the `SPARK_CONTAINER_IMAGE_DEFAULT`
# env variable which is set in `/etc/environment` at image build time.

set -euo pipefail

source /etc/environment
source "${BDUTIL_DIR}/bdutil_components.sh"
source "${BDUTIL_DIR}/bdutil_logging.sh"
source "${BDUTIL_DIR}/bdutil_metadata.sh"
source "${BDUTIL_DIR}/bdutil_misc.sh"
source "${BDUTIL_DIR}/bdutil_retry.sh"
source "${BDUTIL_DIR}/bdutil_properties.sh"
source "${BDUTIL_DIR}/bdutil_versions.sh"
source "${BDUTIL_DIR}/components/shared/conda.sh"
source "${BDUTIL_DIR}/components/shared/containerd.sh"
source "${BDUTIL_DIR}/components/shared/miniconda3.sh"
source "${BDUTIL_DIR}/components/shared/spark.sh"

# Conda of the VM is always mounted to into containers at DEFAULT_CONDA_HOME.
# It is used by the entrypoint.sh if the container image doesn't define
# PYSPARK_PYTHON.
readonly DEFAULT_CONDA_HOME="${CONDA_HOME}"
readonly DEFAULT_PYSPARK_PYTHON="${DEFAULT_CONDA_HOME}/bin/python"

# Always stop the tail -f command (see below) when we exit this script
trap 'kill_tail $?' EXIT

function main() {
  if [[ $# -lt 3 ]]; then
    local script
    script=$(basename "$0")
    echo "Usage: ${script} <working-dir> <container-name> <container-type> [container args...]"
    exit 1
  fi

  local -r working_dir="$1"
  local -r container_name="$2"
  local -r container_type="$3"

  shift 3

  if [[ -z "${working_dir}" ]]; then
    working_dir="${SPARK_WORK_DIR}"
  fi

  local -a envs
  envs=(
    "CUSTOM_CONFIGS_ENABLED=false"
    "CUSTOM_INIT_SCRIPT_ENABLED=false"
    "DEFAULT_PYSPARK_PYTHON=${DEFAULT_PYSPARK_PYTHON}"
    "JAVA_HOME=${JAVA_HOME}"
    "SPARK_HOME=${SPARK_HOME}"
  )

  local pypi_repo
  pypi_repo=$(get_dataproc_property internal.s8s.pypi.repository)
  if [[ -n "${pypi_repo}" ]]; then
    envs+=("DATAPROC_PYPI_REPOSITORY=${pypi_repo}")
  fi

  if is_component_selected 'nvidia-drivers'; then
    envs+=(
      "NVIDIA_DRIVER_CAPABILITIES=all"
      "NVIDIA_VISIBLE_DEVICES=all"
    )
  fi

  if [[ "${DATAPROC_IMAGE_VERSION}" == '1.2' ]] ||
      is_version_at_least "${DATAPROC_IMAGE_VERSION}" '2.2'; then
    envs+=("DEFAULT_LANG=C.UTF-8")
  fi

  local -a args=("${container_type}" "$@")

  local -a mounts=(
    # Conda is not portable. Some files in Conda recorded the original
    # installation path, so the path must be the same in the container.
    "${DEFAULT_CONDA_HOME}:${DEFAULT_CONDA_HOME}"
    "${JAVA_HOME}:${JAVA_HOME}"
    # ${JAVA_HOME}/lib/security/cacerts is a symbolic link to
    # /etc/ssl/certs/adoptium/cacerts in Temurin JDK.
    "/etc/ssl/certs/adoptium/:/etc/ssl/certs/adoptium/"
    "${SPARK_HOME}:${SPARK_HOME}"
    "${SPARK_CONF_DIR}:${SPARK_CONF_DIR}"
    "${SPARK_DATA_DIR}:${SPARK_DATA_DIR}"
    "${SPARK_LOG_DIR}:${SPARK_LOG_DIR}"
    "${working_dir}:${working_dir}"
    "${DATAPROC_COMMON_LIB_DIR}:${DATAPROC_COMMON_LIB_DIR}"
    "${BDUTIL_DIR}/spark-entrypoint.sh:/entrypoint.sh"
  )

  # Mount the NFS local path to the container if it exists
  if [[ -v NFS_PACKAGE_MOUNT_LOCAL_PATH ]]; then
    envs+=("NFS_PACKAGE_MOUNT_LOCAL_PATH=${NFS_PACKAGE_MOUNT_LOCAL_PATH}")
    mounts+=(
      "${NFS_PACKAGE_MOUNT_LOCAL_PATH}:${NFS_PACKAGE_MOUNT_LOCAL_PATH}"
    )
  fi

  local image
  image=$(get_dataproc_property_or_default \
    "dataproc.spark.container.image" "${SPARK_CONTAINER_IMAGE_DEFAULT}")
  pull_image_using_containerd "${image}"

  create_containerd_config \
    "${working_dir}" \
    "${container_name}" \
    "${image}" \
    envs \
    args \
    mounts \
    "$(get_spark_container_memory)"

  crictl runp "${POD_CONFIG}" >/dev/null
  local pod_id
  pod_id=$(crictl pods --name "${POD_NAME}" -q)
  crictl create "${pod_id}" "${CONTAINER_CONFIG}" "${POD_CONFIG}" >/dev/null
  local container_id
  container_id=$(crictl ps --name "${container_name}" -q -a)
  crictl start "${container_id}" >/dev/null

  echo "Waiting for container log creation"
  retry_constant_short ls -l "${CONTAINERD_LOG_DIR}/${CONTAINERD_LOG_FILE}" &>/dev/null

  # Add read permission to the container log
  sudo /usr/local/bin/grant_container_log_read.sh

  # Stream the container log file to STDOUT and STDERR, removing the
  # timestamp and STDOUT/STDERR fields added by containerd.  We run tail -f on
  # the container logs in the background and crictl attach in
  # the foreground. attach alone causes us to miss any logs before the call to
  # attach.  Likewise, tailing the log with `crictl logs` has several issues,
  # such as not emitting STDERR output and missing the last few log lines,
  # because it exits as soon as the container exits.  Combining tail with
  # attach solves these problems.  Below we also use the stdbuf command to not
  # buffer output (otherwise no output will be printed till buffer is full).
  tail -f "${CONTAINERD_LOG_DIR}/${CONTAINERD_LOG_FILE}" |
    tee >(grep --line-buffered 'stderr F' | stdbuf -oL cut -d " " -f 4- >&2) |
    grep --line-buffered 'stdout F' | stdbuf -oL cut -d " " -f 4- &
  TAIL_PID=$!
  crictl attach "${container_id}" &>/dev/null

  # crictl attach does not output the exit status of the container.  We must
  # instead get this from crictl inspect.
  retry_quiet container_stopped "${container_id}"

  local exit_code
  exit_code=$(crictl inspect \
    --output go-template --template '{{.status.exitCode}}' "${container_id}")
  exit "$exit_code"
}

main "$@"
