#!/usr/bin/env bash

# Dataproc Spark container entrypoint.
#
# This script can start different Spark processes based on the first argument of
# the command line. Possible values include "master", "worker", "driver",
# "driver-py", "driver-r", "executor", "spark-submit".
#
# Custom images shouldn't override this entrypoint, instead consider adding a
# custom init script at /opt/init-script.sh for startup time customizations.

set -euo pipefail

readonly CUSTOM_CONFIGS_ENABLED="${CUSTOM_CONFIGS_ENABLED:-false}"
readonly CUSTOM_INIT_SCRIPT_ENABLED="${CUSTOM_INIT_SCRIPT_ENABLED:-false}"
readonly HEARTBEAT_FILE="${HEARTBEAT_FILE:-}"
readonly SPARK_CONF_DIR=/etc/spark/conf
readonly SPARK_EXTRA_CLASSPATH="${SPARK_EXTRA_CLASSPATH:-}"
readonly CREATE_SPARK_EVENT_LOG_DIR_ENABLED="${CREATE_SPARK_EVENT_LOG_DIR_ENABLED:-false}"

function parse_args() {
  if [[ $# -eq 0 ]]; then
    echo "Requires at least one argument."
    exit 1
  fi

  readonly CONTAINER_TYPE="$1"

  case "${CONTAINER_TYPE}" in
  driver | driver-py | driver-r | executor)
    CMD=("${SPARK_CONF_DIR}/entrypoint.sh" "$@")
    ;;
  spark-submit)
    CMD=("$@")
    ;;
  jupyter-server)
    shift 1
    CMD=("$@")
    ;;
  */java)
    CMD=($(get_java_command "$@"))
    ;;
  *)
    echo "Unknown command provided, proceeding in pass-through mode..."
    CMD=("$@")
    ;;
  esac
}

function configure_charset() {
  if [[ -n "${DEFAULT_LANG:-}" && -z ${LANG:-} ]]; then
    export LANG="${DEFAULT_LANG}"
  fi
}

# Configures Python.
#
# To override the default Python, custom container images can include its own
# Python environment and set PYSPARK_PYTHON accordingly.
function configure_python() {
  if [[ -z "${PYSPARK_PYTHON:-}" ]]; then
    # DEFAULT_PYSPARK_PYTHON only needs to be specified if PYSPARK_PYTHON is
    # not.
    if [[ -z "${DEFAULT_PYSPARK_PYTHON:-}" ]]; then
      echo "Error: PYSPARK_PYTHON and DEFAULT_PYSPARK_PYTHON are not defined"
      return 1
    fi
  fi

  export PYSPARK_PYTHON="${PYSPARK_PYTHON:-${DEFAULT_PYSPARK_PYTHON}}"
  echo "PYSPARK_PYTHON=${PYSPARK_PYTHON}"

  local python_env_bin
  python_env_bin=$(dirname "${PYSPARK_PYTHON}")
  # Make sure that ${python_env_bin} is in the PATH and takes precedence.
  export PATH="${python_env_bin}:${PATH}"

  # If the NFS package local mount directory exists, add to PYTHONPATH
  if [[ -n "${NFS_PACKAGE_MOUNT_LOCAL_PATH:-}" ]]; then
    if [[ -n "${PYTHONPATH:-}" ]]; then
      export PYTHONPATH="${NFS_PACKAGE_MOUNT_LOCAL_PATH}:${PYTHONPATH}"
    else
      export PYTHONPATH="${NFS_PACKAGE_MOUNT_LOCAL_PATH}"
    fi
  fi

  # Add the mounted pyspark package to PYTHONPATH, which takes precedence over
  # installed Python environment packages.
  if [[ -n "${PYTHONPATH:-}" ]]; then
    export PYTHONPATH="${SPARK_HOME}/python:${PYTHONPATH}"
  else
    export PYTHONPATH="${SPARK_HOME}/python"
  fi

  # Generate the .pip conf file if (1) users don't config it through custom
  # image and (2) the env variable is configured or NFS repo mounted.
  if [[ ! -f "/home/spark/.pip/pip.conf" ]] &&
    [[ -n "${DATAPROC_PYPI_REPOSITORY:-}" || -n "${NFS_PACKAGE_MOUNT_LOCAL_PATH:-}" ]]; then
    echo "Generating /home/spark/.pip/pip.conf"
    mkdir -p "/home/spark/.pip"
    cat >"/home/spark/.pip/pip.conf" <<EOF
[global]
EOF

    if [[ -n "${DATAPROC_PYPI_REPOSITORY:-}" ]]; then
      echo "Configuring index-url as '${DATAPROC_PYPI_REPOSITORY}'"
      cat >>"/home/spark/.pip/pip.conf" <<EOF
index-url = ${DATAPROC_PYPI_REPOSITORY}
EOF
    fi

    if [[ -n "${NFS_PACKAGE_MOUNT_LOCAL_PATH:-}" ]]; then
      echo "Configuring target as '${NFS_PACKAGE_MOUNT_LOCAL_PATH}'"
      cat >>"/home/spark/.pip/pip.conf" <<EOF
target = ${NFS_PACKAGE_MOUNT_LOCAL_PATH}
EOF
    fi
  fi
}

# Configures R.
#
# To override the default R, custom container images can include their own
# R environment and set R_HOME accordingly.
function configure_r() {
  if [[ -n "${R_HOME:-}" ]]; then
    echo "R_HOME=${R_HOME}"
    export R_HOME

    # Make sure that ${R_HOME} is in the PATH and takes precedence.
    export PATH="${R_HOME}/bin:${PATH}"
  fi
}

# Configures Java.
function configure_java() {
  if [[ -n "${JAVA_HOME:-}" ]]; then
    echo "JAVA_HOME=${JAVA_HOME}"
    # Make sure ${JAVA_HOME}/bin is in PATH and takes precedence.
    export PATH="${JAVA_HOME}/bin:${PATH}"
  else
    echo "JAVA_HOME is not defined"
    return 1
  fi
}

function configure_spark() {
  # Make sure ${SPARK_HOME}/bin is in PATH and takes precedence.
  export PATH="${SPARK_HOME}/bin:${PATH}"

  echo "SPARK_EXTRA_CLASSPATH=${SPARK_EXTRA_CLASSPATH}"

  if [[ -n "${SPARK_EXTRA_CLASSPATH}" ]]; then
    cat >>"${SPARK_CONF_DIR}/spark-env.sh" <<EOF

# Extra Spark classpath from the custom image.
SPARK_DIST_CLASSPATH="\${SPARK_DIST_CLASSPATH}:${SPARK_EXTRA_CLASSPATH}"
EOF
  fi
}

function configure_jupyter() {
  # If the custom image doesn't define PYSPARK_DRIVER_PYTHON, use the `ipython`
  # in the same dir as PYSPARK_PYTHON. PYSPARK_PYTHON should have been
  # configured earlier.
  if [[ -z "${PYSPARK_DRIVER_PYTHON:-}" ]]; then
    PYSPARK_DRIVER_PYTHON="$(dirname "${PYSPARK_PYTHON}")/ipython"
  fi
  export PYSPARK_DRIVER_PYTHON
  echo "PYSPARK_DRIVER_PYTHON=${PYSPARK_DRIVER_PYTHON}"

  if [[ "${PYSPARK_DRIVER_PYTHON}" != */ipython ]]; then
    echo "Error: PYSPARK_DRIVER_PYTHON (${PYSPARK_DRIVER_PYTHON}) is not ipython"
    return 1
  fi

  if [[ ! -f "${PYSPARK_DRIVER_PYTHON}" ]]; then
    echo "Error: ${PYSPARK_DRIVER_PYTHON} was not found"
    return 1
  fi

  if [[ -n "${JUPYTER_CONFIG_PATH:-}" ]]; then
    echo "Error: JUPYTER_CONFIG_PATH must not be set in the image"
    return 1
  fi

  if [[ -n "${JUPYTER_CONFIG_DIR:-}" ]]; then
    echo "Error: JUPYTER_CONFIG_DIR must not be set in the image"
    return 1
  fi

  if [[ -n "${JUPYTER_PATH:-}" ]]; then
    echo "Error: JUPYTER_PATH must not be set in the image"
    return 1
  fi

  if [[ -n "${JUPYTER_DATA_DIR:-}" ]]; then
    echo "Error: JUPYTER_DATA_DIR must not be set in the image"
    return 1
  fi
}

function merge_configs() {
  local -r src=$1
  local -r dest=$2
  local -r header=${3:-"\n# Properties from the custom image."}

  if [[ ! -f ${src} ]]; then
    echo "Skipping merging ${src}, file does not exist."
    return 0
  fi

  echo -e "${header}" >>"${dest}"
  cat "${src}" >>"${dest}"

  echo "Merged ${src} into ${dest}."
}

# Merges the configs from the customer image into the Spark configs. The former
# takes precedence.
function merge_spark_configs() {
  echo 'Merging Spark configs'
  merge_configs \
    /opt/spark/conf/spark-defaults.conf "${SPARK_CONF_DIR}/spark-defaults.conf"
  merge_configs \
    /opt/spark/conf/log4j.properties "${SPARK_CONF_DIR}/log4j.properties"
  merge_configs \
    /opt/spark/conf/spark-env.sh "${SPARK_CONF_DIR}/spark-env.sh"
}

function run_custom_init_script() {
  if [[ -f /opt/init-script.sh ]]; then
    echo "Running custom init script"
    bash -x /opt/init-script.sh "$@"
  else
    echo "Skipping custom init script, file does not exist."
  fi
}

function run_heartbeat_process() {
  echo "Running heartbeat loop"
  while true; do
    touch "${HEARTBEAT_FILE}"
    sleep 1
  done
}

# Spark worker launches executors with `java -cp <classpath> ...`, but the
# classpath doesn't include the env variable SPARK_EXTRA_CLASSPATH which could
# be provided in the container image. So here we need to add it.
function get_java_command() {
  # Keep the original args if SPARK_EXTRA_CLASSPATH is not provided.
  if [[ -z "${SPARK_EXTRA_CLASSPATH:-''}" ]]; then
    echo "$*"
    return 0
  fi

  local -a args
  local arg
  local classpath_arg="false"
  for arg in "$@"; do
    if [[ ' -cp -classpath --classpath ' == *" ${arg} "* ]]; then
      classpath_arg="true"
      args+=("${arg}")
    elif [[ "${classpath_arg}" == "true" ]]; then
      classpath_arg="false"
      args+=("${arg}:${SPARK_EXTRA_CLASSPATH}")
    else
      args+=("${arg}")
    fi
  done
  echo "${args[*]}"
}

function get_java_property() {
  local -r property_file=$1
  local -r property_name=$2
  local property_value
  property_value=$(grep "^${property_name}=" "${property_file}" |
    tail -n 1 | cut -d '=' -f 2- | sed -r 's/\\([#!=:])/\1/g')
  echo "${property_value}"
}

function create_event_log_dir() {
  # The event log directory may be in the spark configuration file.
  local event_log_dir
  event_log_dir=$(get_java_property \
    "${SPARK_CONF_DIR}/spark-defaults.conf" spark.eventLog.dir)

  # The event log directory may be in an argument. If so, it takes priority
  # over anything from the spark configuration file.
  local arg
  for arg in "$@"; do
    if [[ "${arg}" == spark.eventLog.dir=* ]]; then
      event_log_dir="${arg:19}" # 19 == len(spark.eventLog.dir=).
    fi
  done

  if [[ "${event_log_dir}" == gs://* ]]; then
    event_log_dir=${event_log_dir#*//} # strip 'gs://' prefix
    event_log_dir=${event_log_dir%*/}  # strip '/' suffix
    local event_log_dir_bucket=${event_log_dir%%/*}
    local event_log_dir_path=${event_log_dir#*/}

    # Read the auth token from the Metadata server.
    local auth_token
    auth_token=$(curl -fs --retry-connrefused --retry 5 --retry-max-time 30 \
      -H "Metadata-Flavor: Google" \
      "http://metadata.google.internal/computeMetadata/v1/instance/service-accounts/default/token" |
      python3 -c 'import json, sys; print(json.load(sys.stdin)["access_token"])')

    # Ignore curl exit code because we rely on HTTP status code verification.
    set +e

    # Use curl to create a directory object.
    # Set `ifGenerationMatch=0` precondition to not override
    # directory object if it already exists.
    local http_code
    http_code=$(curl -fs \
      --retry-connrefused --retry 5 --retry-max-time 30 \
      --data-binary @/dev/null \
      --output /dev/null \
      --write-out "%{http_code}" \
      -X POST \
      -H "Authorization: Bearer ${auth_token}" \
      -H "Content-Type: application/octet-stream" \
      "https://storage.googleapis.com/upload/storage/v1/b/${event_log_dir_bucket}/o?uploadType=media&name=${event_log_dir_path}/&ifGenerationMatch=0")
    local -r exit_code=$?

    set -e

    # 412 status code is a success - it means that the directory already exists
    # because `ifGenerationMatch=0` precondition was set.
    if [[ "${http_code}" -ge "200" && "${http_code}" -le "299" ]] ||
        [[ "${http_code}" == "412" ]]; then
      return 0
    fi

    echo "Failed to create event log dir with ${http_code} HTTP response code and ${exit_code} exit code"
    return 1
  fi
}

function main() {
  parse_args "$@"
  configure_charset
  configure_python
  configure_r
  configure_java
  configure_spark
  if [[ "${CONTAINER_TYPE}" == "jupyter-server" ]]; then
    configure_jupyter
  fi
  if [[ "${CUSTOM_CONFIGS_ENABLED}" == "true" ]]; then
    merge_spark_configs
  fi
  if [[ "${CUSTOM_INIT_SCRIPT_ENABLED}" == "true" ]]; then
    run_custom_init_script "$@"
  fi
  if [[ "${CREATE_SPARK_EVENT_LOG_DIR_ENABLED}" == "true" && "$1" == "spark-submit" ]]; then
    # Only create the eventLog directory on the Driver, not each Executor.
    create_event_log_dir "$@"
  fi
  if [[ -n "${HEARTBEAT_FILE}" ]]; then
    run_heartbeat_process &
  fi
  exec /usr/bin/tini -s -- "${CMD[@]}"
}

main "$@"
