# This script is sourced at the end of image build to verify the image is in
# expected state.

# Verifies that Dataproc image components are in expected state.
function verify_components() {
  local -r components=("$@")

  local component
  for component in "${components[@]}"; do
    local verify_script="${BDUTIL_DIR}/components/verify/${component}.sh"
    if [[ -f "${verify_script}" ]]; then
      loginfo "Running component verify script: ${verify_script}"
      source "${verify_script}"
    else
      loginfo "Component ${component} does not have verify script"
    fi
  done
}

function verify_java() {
  local -a java_alternatives
  local expected_jdks=( "${JAVA_PACKAGE}" )
  if [[ -n "${ADDITIONAL_JAVA_PACKAGE}" ]]; then
    expected_jdks+=( "${ADDITIONAL_JAVA_PACKAGE}" )
  fi
  mapfile -t java_alternatives < <(update-alternatives --display java | grep priority)
  if [[ ${#java_alternatives[@]} != "${#expected_jdks[@]}" ]]; then
    (IFS=$'\n'; echo -e "ERROR: found more than ${#expected_jdks[@]} Java installed:\n${java_alternatives[*]}")
    exit 1
  fi

  for jdk in "${expected_jdks[@]}"; do
    if ! [[ "${java_alternatives[*]}" =~ "${jdk}" ]]; then
      echo -e "ERROR: '${jdk}' Java should be installed but found:\n${java_alternatives[*]}"
      exit 1
    fi
  done

  if [[ -n "${ADDITIONAL_JAVA_PACKAGE}" ]]; then
    # verify that JAVA_PACKAGE is the default.
    default_java_cmds=( $(ls "${JAVA_HOME}"/bin) )
    for cmd in "${default_java_cmds[@]}"; do
      alternative=$(update-alternatives --query "${cmd}" | grep 'Value:' || true)
      if [[ -z "$alternative" ]]; then
         continue
      fi
      if ! [[ "$alternative" =~ "${JAVA_HOME}/bin/${cmd}" ]]; then
        echo -e "ERROR: '${jdk}' should be default!  but found: '${alternative}'"
        exit 1
      fi
    done
  fi

}

# Verify OS specific artifacts are in expected state.
function verify_os_specific_artifacts() {
  loginfo "Running script ${BDUTIL_DIR}/os/${OS_ID}/verify_image.sh"
  source "${BDUTIL_DIR}/os/${OS_ID}/verify_image.sh"
}

function verify_services_disabled() {
  local -r services=("$@")

  for service in "${services[@]}"; do
    if ! is_service_installed "${service}"; then
      loginfo "Error: service ${service} is not installed"
      return 1
    fi

    if is_service_enabled "${service}"; then
      loginfo "Error: service ${service} is enabled unexpectedly"
      return 1
    fi

    loginfo "Info: service ${service} is disabled as expected"
  done
}

function verify_services_not_installed() {
  local -r services=("$@")

  for service in "${services[@]}"; do
    if is_service_installed "${service}"; then
      loginfo "Error: service ${service} is installed unexpectedly"
      return 1
    fi

    loginfo "Info: service ${service} is not installed"
  done
}

# go/log4j-cyexit-cert
function verify_no_vulnerable_jars() {
  download_resource \
    https://www.gstatic.com/log4j_jarscanner/2hamster-scanner-linux \
    /tmp/jar-scanner
  chmod +x /tmp/jar-scanner

  local -a jars
  mapfile -t jars < <(
      /tmp/jar-scanner --codeyellow_pa="GCP" --codeyellow_product="Dataproc/Image" 2>&1 |
          grep "vulnerable-jar" |
          sed -e 's#.*vulnerable-jar: path="\(.*.jar\)" .*#\1#')

  rm -rf /tmp/jar-scanner

  if [[ "${#jars[@]}" -gt 0 ]]; then
    echo "ERROR: found ${#jars[@]} vulnerable jars: ${#jars[*]}"
    exit 1
  fi

  echo "No vulnerable jars found"
}

# Validates the collectd default conf installation in bdutil done at ${BDUTIL_DIR}/generate_collectd_plugin.sh during image build.
function verify_collectd_default_conf_installation() {
  local -r collectd_plugin_install_dir="${BDUTIL_DIR}/conf"
  local missing_files=()
  local file
  for file in "${COLLECTD_CONF_FILE_NAMES[@]}"; do
    if [[ ! -f "${collectd_plugin_install_dir}/$file" ]]; then
      missing_files+=("${file}")
    fi
  done
  if [[ "${#missing_files[@]}" -gt 0 ]]; then
    printf "ERROR: Missing default collectd conf files in bdutil. Following files are expected in '%s': %s" "${collectd_plugin_install_dir}" "${missing_files[*]}"
    exit 1
  fi
}

function verify_image() {
  verify_components "${DATAPROC_OPTIONAL_COMPONENTS[@]}"
  verify_java
  verify_no_vulnerable_jars
  verify_os_specific_artifacts
  verify_collectd_default_conf_installation
}

verify_image
